package com.sapportals.wcm.protocol.webdav;

import java.util.*;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.sap.tc.logging.Location;

import com.sapportals.wcm.util.uri.IUriReference;
import com.sapportals.wcm.util.uri.UriFactory;

/**
 * Maps {@link IUriReference}s to {@link DavStatus} objects.
 * <p>
 * Copyright (c) SAP AG 2002-2003
 * @author julian.reschke@greenbytes.de
 * @version $Id: DavMultiStatus.java,v 1.2 2004/02/19 18:31:05 jre Exp $
 */

public class DavMultiStatus {

  private static final Location log = Location.getLocation(DavMultiStatus.class);

  private Map map = Collections.EMPTY_MAP;

  /**
   * Builds empty map.
   */
  public DavMultiStatus() {
  }

  /**
   * Builds map from {@link Document} containing a WebDAV
   * <a href="http://greenbytes.de/tech/webdav/rfc2518.html#ELEMENT_multistatus">DAV:multistatus</a> body.
   * Problems found in the document will be logged, but no exception will
   * be reported (in the worst case, an empty object is created).
   * @param doc Document with document element DAV:multistatus.
   */
  public DavMultiStatus(Document doc) {
    if (doc == null) throw new IllegalArgumentException();

    this.map = new HashMap();

    Element docelem = doc.getDocumentElement();
    if (Consts.N_multistatus.equalsNode(docelem)) {
      Iterator rit = DomElement.getChildren(docelem, Consts.N_response).iterator();

      while (rit.hasNext()) {
        Element response = (Element) rit.next();
        List hrefs = DomElement.getChildren(response, Consts.N_href);
        if (hrefs.size() == 1) {
          String href = DomElement.getText((Element) hrefs.get(0));
          try {
            IUriReference uri = UriFactory.parseUriReference(href);
            DavStatus status = new DavStatus(response);
            this.map.put(uri, status);
          }
          catch (java.net.MalformedURLException ex) {
            log.debugT(href + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(ex));
          }
          catch (com.sapportals.wcm.WcmException ex) {
            log.debugT(DomSerializer.toString(response, false) + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(ex));
          }
        }
        else {
          log.debugT("DAV:response element must have exactly on DAV:href element");
        }
      }
    }
    else {
      log.debugT("document element must be DAV:multistatus.");
    }
  }

  // map interface methods

  /**
   * Check whether information for the given URI reference is present.
   * @param uri URI reference
   * @return <code>true</code> when information present.
   */
  public boolean containsKey(IUriReference uri) {
    return this.map.containsKey(uri);
  }

  /**
   * Get status information for the given URI reference.
   * @param uri URI reference
   * @return {@link DavStatus} for this URI reference or <code>null</code>.
   */
  public DavStatus get(IUriReference uri) {
    return (DavStatus)this.map.get(uri);
  }

  /**
   * Check whether the map is empty.
   * @return <code>true</code> is the map is empty.
   */
  public boolean isEmpty() {
    return this.map.isEmpty();
  }

  /**
   * A information about a specific URI reference to the map.
   * @param uri URI reference
   * @param status status information
   */
  public void put(IUriReference uri, DavStatus status) {
    init();
    this.map.put(uri, status);
  }

  /**
   * Return the size of the map.
   * @return the size
   */
  public int size() {
    return this.map.size();
  }

  /**
   * Add a DAV:multistatus body element describing the error map.
   * @param doc the document to which the DAV:multistatus element will
   * be added as document element (must be empty)
   * @return HTTP status code 207
   */
  public int addMultiStatusBody(Document doc) {

    Element ms = DomElement.makeNew(doc, Consts.N_multistatus);
    doc.appendChild(ms);

    Iterator it = this.map.keySet().iterator();

    while (it.hasNext()) {
      IUriReference uri = (IUriReference)it.next();
      DavStatus status = (DavStatus)this.map.get(uri);
      ms.appendChild(status.getResponseElement(doc, uri));
    }

    return DavStatus.SC_MULTI_STATUS;
  }

  /**
   * Add a DAV:error element describing the first error in the map.
   * @param doc the document to which the DAV:error element will
   * be added as document element (must be empty)
   * @return HTTP status extracted from the error element
   */
  public int addErrorElement(Document doc) {
    DavStatus status = (DavStatus)this.map.values().iterator().next();
    doc.appendChild(status.getErrorElement(doc, null));
    return status.getStatus();
  }

  private void init() {
    synchronized(this) {
      if (this.map.size() == 0) {
        this.map = new HashMap();
      }
    }
  }

}