package com.sapportals.wcm.protocol.webdav;

import java.util.*;

import org.w3c.dom.*;

import com.sap.tc.logging.Location;

import com.sapportals.wcm.util.http.DateFormat;
import com.sapportals.wcm.util.name.IName;

/**
 * Utility functions for working with {@link org.w3c.dom.Node}s.
 * <p>
 * Note: most methods are static, some are also available as non-static
 * methods of DomElement (which in this case is only a container for
 * a {@link org.w3c.dom.Document}).
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author julian.reschke@greenbytes.de
 * @version $Id: DomElement.java,v 1.3 2004/04/19 12:48:06 jre Exp $
 */

public class DomElement {

  private static final Location log = Location.getLocation(DomElement.class);

  private final Document owner;

  public DomElement(Document owner) {
    this.owner = owner;
  }

  public static Comment makeSafeComment(Document doc, String text) {
    
    StringBuffer sb = new StringBuffer(text.length());
    
    char last = 0;
    
    for (int i = 0; i < text.length(); i++) {
      char c = text.charAt(i);
     
      if (c == '-') {
        if (last != '-') {
          last = c;
          sb.append(c);
        }
      }
      else if (c >= 32 && c < 128) {
        last = c;
        sb.append(c);
      }
    }
    
    return doc.createComment(sb.toString());
  }
  
  
  /**
   * Create a new element node with optional content
   * @param name name for the element
   * @param content optional content (content of type {@link org.w3c.dom.Element}
   * is added as a child node, {@link IInternationalString}s are added as {@link Text} node
   * while setting the attribute xml:lang, otherwise the string value is added as a
   * {@link Text} node)
   * @return newly created element
   */
  public Element makeNew(IName name, Object content) {
    return makeNew(this.owner, name, content);
  }

  /**
   * Create a new element node
   * @param name name for the element
   * @return newly created element
   */
  public Element makeNew(IName name) {
    return makeNew(this.owner, name, null);
  }

  /**
    * Create a new element node (with specific prefix) and optional content
    * @param doc the owner document
    * @param namespace namespace name
    * @param name element's local name
    * @param content optional content (content of type {@link org.w3c.dom.Element}
    * is added as a child node, {@link IInternationalString}s are added as {@link Text} node
    * while setting the attribute xml:lang, otherwise the string value is added as a
    * {@link Text} node)
    * @return newly created element
    */

  public static Element makeNew(Document doc, String namespace, String prefix, String name, Object content) {

    Element e = null;

    // special treatment for some namespaces for more compact serialization

    if (Consts.NS_SAP.equals(namespace)) {
      e = doc.createElementNS(namespace, "sap:" + name);
      setNsDeclOnContainer(e, "xmlns:sap", Consts.NS_SAP);
    }
    else if (Consts.NS_SOAP_ENC.equals(namespace)) {
      e = doc.createElementNS(namespace, "soap-enc:" + name);
      setNsDeclOnContainer(e, "xmlns:soap-enc", Consts.NS_SOAP_ENC);
    }
    else if (Consts.NS_WEBDAV.equals(namespace)) {
      e = doc.createElementNS(namespace, "D:" + name);
    }
    else if ("xs".equals(prefix) && Consts.NS_XSD.equals(namespace)) {
      e = doc.createElementNS(namespace, prefix + ":" + name);
      setNsDeclOnContainer(e, "xmlns:xs", Consts.NS_XSD);
    }
    else {
      e = doc.createElementNS("".equalsIgnoreCase(namespace) ? null : namespace, name);
    }

    if (content != null) {
      if (content instanceof org.w3c.dom.Node) {
        e.appendChild(DomElement.importNode(doc, (Node)content, true));
      }
      else if (content instanceof IInternationalString) {
        IInternationalString is = (IInternationalString)content;
        e.appendChild(doc.createTextNode(is.getText()));
        if (is.getLanguageTag() != null && is.getLanguageTag().length() > 0) {
          e.setAttributeNS(Consts.NS_XML, "xml:lang", is.getLanguageTag());
        }
      }
      else {
        e.appendChild(doc.createTextNode(content.toString()));
      }
    }

    return e;
  }

  /**
   * Create a new element node
   * @param doc the owner document
   * @param namespace namespace name
   * @param name element's local name
   * @return newly created element
   */
  public static Element makeNew(Document doc, String namespace, String name) {
    return makeNew(doc, namespace, null, name, null);
  }

  /**
   * Create a new element node with optional content
   * @param doc the owner document
   * @param namespace namespace name
   * @param name element's local name
   * @param content optional content (content of type {@link org.w3c.dom.Element}
   * is added as a child node, {@link IInternationalString}s are added as {@link Text} node
   * while setting the attribute xml:lang, otherwise the string value is added as a
   * {@link Text} node)
   * @return newly created element
   */
  public static Element makeNew(Document doc, String namespace, String name, Object content) {
    return makeNew(doc, namespace, null, name, content);
  }

  /**
   * Create a new element node with the given name.
   * @param doc the owner document
   * @param name element's name
   * @return newly created element
   */
  public static Element makeNew(Document doc, IName name) {
    return makeNew(doc, name.getNamespace(), null, name.getName(), null);
  }

  /**
   * Create a new element node with the given qualified name (respecting the namespace prefix).
   * @param doc the owner document
   * @param name element's name
   * @return newly created element
   */
  public static Element makeNew(Document doc, QualifiedName name) {
    return makeNew(doc, name.getNamespaceUri(), name.getPrefix(), name.getLocalName(), null);
  }

  /**
   * Create a new element node with the given name and
   * optional content
   * @param doc the owner document
   * @param name element's name
   * @param content optional content (content of type {@link org.w3c.dom.Element}
   * is added as a child node, {@link IInternationalString}s are added as {@link Text} node
   * while setting the attribute xml:lang, otherwise the string value is added as a
   * {@link Text} node)
   * @return newly created element
   */
  public static Element makeNew(Document doc, IName name, Object content) {
    return makeNew(doc, name.getNamespace(), null, name.getName(), content);
  }

  /**
   * Create a new element node with the given name
   * (respecting the prefix) and optional content
   * @param doc the owner document
   * @param name element's name
   * @param content optional content (content of type {@link org.w3c.dom.Element}
   * is added as a child node, {@link IInternationalString}s are added as {@link Text} node
   * while setting the attribute xml:lang, otherwise the string value is added as a
   * {@link Text} node)
   * @return newly created element
   */
  public static Element makeNew(Document doc, QualifiedName name, Object content) {
    return makeNew(doc, name.getNamespaceUri(), name.getPrefix(), name.getLocalName(), content);
  }

  /**
   * Create a new {@link org.w3c.dom.Element} in the "DAV:" namespace
   * @param doc the owner document
   * @param name element's local name
   * @return a new {@link org.w3c.dom.Element}
   * @deprecated
   */
  /*
    private static Element makeNewDAV (Document doc, String name)
    {
      return makeNew (doc, Consts.NS_WEBDAV, null, name, null);
    }
  */
  /**
   * Create a new {@link org.w3c.dom.Element}  in the "DAV:" namespace.
   * @param doc the owner document
   * @param name element's local name
   * @param content optional content (content of type {@link org.w3c.dom.Element}
   * is added as a child node, otherwise the String value is added as a
   * {@link org.w3c.dom.Text} node
   * @return a new {@link org.w3c.dom.Element}
   * @deprecated
   */
  /*
    private static Element makeNewDAV (Document doc, String name, java.lang.Object content)
    {
      return makeNew (doc, Consts.NS_WEBDAV, name, null, content);
    }
  */
  /**
   * Compare a node's namespace and name
   * @param n node
   * @param namespace namespace name
   * @param name name
   * @return whether namespace and name match
   */
  public static boolean match(Node n, String namespace, String name) {
    if (n == null)
      return false;

    String ns = n.getNamespaceURI();

    if (ns == null) {
      return (ns == namespace && n.getLocalName().equals(name));
    }
    else {
      return (ns.equals(namespace) && n.getLocalName().equals(name));
    }
  }

  /**
   * Compare a node's namespace and name
   * @param n node
   * @param name name
   * @return whether namespaceUri is <code>DAV:</code> and the name matches
   */

  public static boolean match(Node n, IName name) {
    return match(n, name.getNamespace(), name.getName());
  }

  /**
    * Attaches <a href="http://www.w3.org/TR/1998/NOTE-XML-data-0105/">XML DATA</a> datatype information to an element to support
    * the special requirements of the Microsoft Webfolder client.
    * @param e the element to which the datatype should be attached
    * @param dt the datatype
    */
  public static void attachXmlDataDataType(Element e, String dt) {
    e.setAttributeNS(Consts.NS_NOTE_XMLDATA_DATATYPES, "dt:dt", dt);
    setNsDeclOnContainer(e, "xmlns:dt", Consts.NS_NOTE_XMLDATA_DATATYPES);
  }

  /**
    * Attaches property flag to a property.
    * @param e the element to which the datatype should be attached
    * @param flag name of flag
    * @param value boolean value of flag
    */
  public static void attachPropertyFlag(Element e, String flag, boolean value) {
    e.setAttributeNS(Consts.NS_SAP_PROPERTYFLAGS, "pf:" + flag, value ? "true" : "false");
    setNsDeclOnContainer(e, "xmlns:pf", Consts.NS_SAP_PROPERTYFLAGS);
  }

  /**
    * Attaches <a href="http://www.w3.org/TR/xmlschema-2/">XML Schema datatype</a> information to an element.
    * @param e the element to which the datatype should be attached
    * @param dt the datatype
    */

  public static void attachXsdDataType(Element e, QualifiedName dt) {
    String ns = dt.getNamespaceOrNull();

    e.setAttributeNS(Consts.NS_XSI, "xsi:type", dt.getPrefix() + ":" + dt.getLocalName());
    setNsDeclOnContainer(e, "xmlns:xsi", Consts.NS_XSI);

    if (Consts.NS_XSD.equals(ns)) {
      setNsDeclOnContainer(e, "xmlns:xs", Consts.NS_XSD);
    }
    else {
      setNsDeclOnContainer(e, "xmlns:" + dt.getPrefix(), dt.getNamespace());
    }
  }
  

  /**
    * Attaches <a href="http://www.w3.org/TR/SOAP/#_Toc478383522">SOAP array type</a> information to an element.
    * @param e the element to wich the datatype should be attached
    * @param dt the datatype
    * @param size size of array
    */
  public static void attachSoapArrayType(Element e, QualifiedName dt, int size) {
    attachXsdDataType(e, XsdDataTypes.SOAP_ARRAY);

    // special case for multivalued properties that aren't marshalled as SOAP arrays
    if (XsdDataTypes.SAP_MAPPED_RID.equals(dt) || XsdDataTypes.SAP_MAPPED_RID_LIST.equals(dt)) {
      attachXsdDataType(e, XsdDataTypes.SAP_MAPPED_RID_LIST);
      return;
    }

    String ns = dt.getNamespaceOrNull();

    e.setAttributeNS(Consts.NS_SOAP_ENC, "soap-enc:arrayType", dt.getPrefix() + ":" + dt.getLocalName() + "[" + size + "]");
    setNsDeclOnContainer(e, "xmlns:soap-enc", Consts.NS_SOAP_ENC);

    if (Consts.NS_XSD.equals(ns)) {
      setNsDeclOnContainer(e, "xmlns:xs", Consts.NS_XSD);
    }
    else {
      setNsDeclOnContainer(e, "xmlns:" + dt.getPrefix(), dt.getNamespace());
    }
  }

  /**
   * Checks whether there are Element children in the DAV: namespace.
   *
   * @param elem the element to check
   * @return whether or not elem has a child elem in the DAV: namespace
   */
  public static boolean hasElementChildrenInDavNs(Element elem) {

    for (Node n = elem.getFirstChild(); n != null; n = n.getNextSibling()) {
      if (n.getNodeType() == Node.ELEMENT_NODE) {
        Element e = (Element)n;

        if (Consts.NS_WEBDAV.equals(e.getNamespaceURI()))
          return true;
      }
    }

    return false;
  }

  /**
   * Adds a <a href="http://greenbytes.de/tech/webdav/rfc2518.html#ELEMENT_status">DAV:status</a>
   * child element with the given content.
   *
   * @param prop Element to which the DAV:status shall be added
   * @param status the HTTP status to set
   */

  public static void addDavStatus(Element prop, String status) {
    Element e = DomElement.makeNew(prop.getOwnerDocument(), Consts.N_status, status);
    prop.appendChild(e);
  }

  /**
   * Gets the typed value from text content of an element based on it's
   * <a href="http://www.w3.org/TR/xmlschema-1/#xsi_type">xsi:type</a> attribute.
   * Supported types are <a href="http://www.w3.org/TR/xmlschema-2/#boolean">xs:boolean</a>
   * (mapped to {@link Boolean}),
   * <a href="http://www.w3.org/TR/xmlschema-2/#dateTime">xs:dateTime</a>
   * (mapped to {@link Date}),
   * <a href="http://www.w3.org/TR/xmlschema-2/#long">xs:long</a>
   * (mapped to {@link Long}) and
   * <a href="http://www.w3.org/TR/xmlschema-2/#int">xs:int</a>
   * (mapped to {@link Integer}).
   * Other types are treated as string values. If the text content can not
   * be parsed according to the data type information, it's treated as string
   * as well.
   * @param e containing element
   * @param df date format parser (for ISO8601 dates)
   * @return typed value
   */
  public static Object getTypedValue(Element e, DateFormat df) {

    String content = DomElement.getText(e);
    Name type = XsdDataTypes.getXsdTypeInfo(e);

    try {
      if (XsdDataTypes.XSD_DATETIME.equals(type)) {
        return df.parseISO8601(content);
      }
      else if (XsdDataTypes.XSD_INT.equals(type)) {
        return new Integer(content);
      }
      else if (XsdDataTypes.XSD_LONG.equals(type)) {
        return new Long(content);
      }
      else if (XsdDataTypes.XSD_BOOLEAN.equals(type)) {
        return Boolean.valueOf(content);
      }
      else {
        return content;
      }
    }
    catch (java.text.ParseException ex) {
      //$JL-EXC$
      return content;
    }
  }

  /**
   * Get the (concatenated) text content of an element (only visits direct
   * children of type {@link Node#TEXT_NODE} and {@link Node#CDATA_SECTION_NODE}).
   *
   * @param e parent element
   * @return text content (empty string when no text present)
   */

  public static String getText(Node e) {

    StringBuffer res = new StringBuffer();

    e.normalize();

    for (Node n = e.getFirstChild(); n != null; n = n.getNextSibling()) {

      int type = n.getNodeType();
      if (type == Node.TEXT_NODE || type == Node.CDATA_SECTION_NODE) {
        res.append(n.getNodeValue());
      }
    }

    return res.toString();
  }

  public static Set getHrefs(Node e) {

    Set result = new HashSet();

    for (Node n = e.getFirstChild(); n != null; n = n.getNextSibling()) {
      if (Consts.N_href.equalsNode(n)) {
        result.add(DomElement.getText(n));
      }
    }

    return result;
  }

  public static String getSingleHref(Node e) throws java.lang.IllegalArgumentException {

    String href = null;

    for (Node n = e.getFirstChild(); n != null; n = n.getNextSibling()) {
      if (Consts.N_href.equalsNode(n)) {
        if (href != null) {
          throw new java.lang.IllegalArgumentException("only single DAV:href child element expected");
        }
        href = DomElement.getText(n);
      }
    }

    return href;
  }

  /**
   * Gets an {@link IInternationalString} from a node. For non-elements,
   * this is just the text content without language identification. For
   * elements, an attempts is made to find an xml:lang attribute in scope
   * to determine the language.
   * @param e containing node
   * @return text contents and laguage information as {@link IInternationalString}
   */
  public static IInternationalString getInternationalText(Node e) {
    String text = getText(e);

    if (e.getNodeType() != Node.ELEMENT_NODE) {
      return new InternationalString((String) null, text);
    }
    else {
      String language = null;
      Element el = (Element)e;
      Element search = el;

      while (language == null && search != null) {
        language = search.getAttributeNS(Consts.NS_XMLNS, "lang");
        search = (Element)search.getParentNode();
      }

      return new InternationalString(language, text);
    }
  }

  /**
   * Checks whether an element node contains only text content (being
   * nodes of type {@link Node#TEXT_NODE} or {@link Node#CDATA_SECTION_NODE}}.
   * @param e containing element
   * @return <code>true</code> if only text content
   */
  public static boolean hasOnlyTextContent(Node e) {

    for (Node n = e.getFirstChild(); n != null; n = n.getNextSibling()) {

      int type = n.getNodeType();

      if (type != Node.TEXT_NODE && type != Node.CDATA_SECTION_NODE) {
        return false;
      }
    }

    return true;
  }

  /** Get NodeList containing all descendants
   * with a certain element name.
   * @param parent the parent of the elements to find
   * @param name the name of the elements to find
   * @return NodeList of matching child elements
   */
  public static NodeList getElements(Element parent, IName name) {
    return parent.getElementsByTagNameNS(name.getNamespace(), name.getName());
  }

  /**
   * Get all children with matching namespace and name.
   * @return List of matching children (or empty list).
   */
  public static List getChildren(Node parent, String namespaceUri, String name) {

    List result = null;

    for (Node c = parent.getFirstChild(); c != null; c = c.getNextSibling()) {
      if (match(c, namespaceUri, name)) {
        if (result == null)
          result = new ArrayList();
        result.add(c);
      }
    }

    return result == null ? java.util.Collections.EMPTY_LIST : result;
  }

  /**
   * Gets all child element nodes with matching {@link IName}.
   * @param parent parent node from hwere to match
   * @param name element name to match for
   * @return list of matching child elements (never <code>null</code>).
   */
  public static List getChildren(Node parent, IName name) {
    return getChildren(parent, name.getNamespace(), name.getName());
  }

  /**
   * Determines if node has child with matching {@link IName}.
   * @param parent parent node from hwere to match
   * @param name element name to match for
   * @return if matching children exist
   */
  public static boolean hasChild(Node parent, IName name) {
    for (Node c = parent.getFirstChild(); c != null; c = c.getNextSibling()) {
      if (match(c, name)) {
        return true;
      }
    }

    return false;
  }

  /**
   * Get the value of the first instance of a top-level processing instruction by name.
   *
   * @param doc to be checked
   * @param target PI target
   * @return the value or <code>null</code>
   */

  public static String getProcessingInstruction(Document doc, String target) {
    for (Node n = doc.getFirstChild(); n != null; n = n.getNextSibling()) {
      if (n.getNodeType() == Node.PROCESSING_INSTRUCTION_NODE) {
        if (n.getNodeName().equals(target))
          return n.getNodeValue();
      }
    }

    return null;
  }

  /**
   * Set the text element content of an element, allowing <code>null</code>
   * and removing all previous content. Checks for valid character content.
   * @param e the element
   * @param val the text content
   * @throws DOMException with code {@link DOMException#INVALID_CHARACTER_ERR} when invalid content present
   */

  public static void setUnsafeText(Element e, String val) throws DOMException {
    setText(e, makeSafeXml(val));
  }

  /**
   * Check the string's content for <a href="http://www.w3.org/TR/2000/REC-xml-20001006#charsets">XML compliance</a>
   * (whether it can be used as value for a text node). Rejects the string when it includes
   * invalid control characters or invalid surrogate pairs.
   * @param val string value
   * @return the string itself
   * @throws DOMException with code {@link DOMException#INVALID_CHARACTER_ERR} when invalid content present
   */
  public static String makeSafeXml(String val) {

    if (val != null) {

      int len = val.length();

      for (int i = 0; i < len; i++) {
        int c = val.charAt(i);
        c &= 0xffff;

        if (c < 32 && (c != 9 && c != 10 && c != 13)) {
          throw new org.w3c.dom.DOMException(
            DOMException.INVALID_CHARACTER_ERR,
            "invalid XML content (illegal control characters) in: " + dump(val));
        }
        else if (c >= 0xd800 && c <= 0xdbff) {
          if (i + 1 >= len) {
            throw new org.w3c.dom.DOMException(
              DOMException.INVALID_CHARACTER_ERR,
              "invalid XML content (truncated surrogate pair) in: " + dump(val));
          }
          else {
            int c2 = (val.charAt(++i)) & 0xffff;
            if (c2 < 0xdc00 || c > 0xdfff) {
              throw new org.w3c.dom.DOMException(
                DOMException.INVALID_CHARACTER_ERR,
                "invalid XML content (invalid surrogate pair) in: " + dump(val));
            }
          }
        }
      }
    }

    return val;
  }

  /**
   * Set the text element content of an element, allowing <code>null</code>
   * and removing all previous content.
   * @param e the element
   * @param val the text content
   */
  public static void setText(Element e, String val) {

    // remove existing child nodes
    if (e.hasChildNodes()) {
      NodeList nl = e.getChildNodes();
      for (int i = 0; i < nl.getLength(); i++) {
        e.removeChild(nl.item(i));
      }
    }

    if (val != null) {
      e.appendChild(e.getOwnerDocument().createTextNode(val));
    }
  }

  /**
   * Set the text element content of an element to a stringified long value,
   * removing all previous content.
   * @param e the element
   * @param val the long value
   */
  public static void setText(Element e, long val) {
    setText(e, Long.toString(val));
  }

  /**
   * Inherit all attributes in the XML namespace, according to
   * <a href="http://www.w3.org/TR/2001/REC-xml-c14n-20010315#DocSubsets">Canonical XML</a>
   * @param elem anchor element (attributes are added)
   */
  public static void inheritXmlAttributes(Element elem) {
    NamedNodeMap attrs = elem.getAttributes();
    Node parent = elem.getParentNode();

    while (parent != null && parent.getNodeType() == Node.ELEMENT_NODE) {
      NamedNodeMap pattrs = parent.getAttributes();

      for (int i = 0; i < pattrs.getLength(); i++) {
        Attr a = (Attr)pattrs.item(i);
        String ns = a.getNamespaceURI();

        if (Consts.NS_XML.equals(ns)) {
          // found one

          if (null == attrs.getNamedItemNS(ns, a.getLocalName())) {
            Attr n = elem.getOwnerDocument().createAttributeNS(ns, a.getNodeName());
            n.setNodeValue(a.getNodeValue());
            attrs.setNamedItemNS(n);
          }
        }
      }

      parent = parent.getParentNode();
    }
  }

  /**
   * Wrapper method for {@link Document#importNode(Node, boolean)}, providing logging of
   * fatal errors (seen in early Xerces versions).
   * @param owner owner document
   * @param imp node to import
   * @param deep <code>true</code> when importing recursively
   * @return imported node
   */
  public static Node importNode(Document owner, Node imp, boolean deep) {
    // The code below was needed by some historic version of Xerces because
    // it was throwing an exception upon importNode, while everything seemed
    // to be ok. Setting the errorChecking to false seemed to solve this
    // problem. As of 1.4.2, the problem seems to have disappeared. (JRE)

    //if (owner instanceof org.apache.xerces.dom.DocumentImpl)
    //  ((org.apache.xerces.dom.DocumentImpl) owner).setErrorChecking(false);

    try {
      return owner.importNode(imp, deep);
    }
    catch (DOMException ex) {
      log.fatalT(
        "It seems that the currently loaded DOM implementation has "
          + "a bug in importNode. This is a known problem for instance in older version "
          + "of Xerces-1. Version 1.4.3 seems to work fine.");
      throw ex;
    }

    //if (owner instanceof org.apache.xerces.dom.DocumentImpl)
    //  ((org.apache.xerces.dom.DocumentImpl) owner).setErrorChecking(true);
  }

//  /**
//   * Check whether all of the node's childs have the same name (and return it).
//   * @param p parent node
//   * @return the name common to all children (or <code>null</code>)
//   * @deprecated multivalued marshalling doesn't require common child names anymore
//   */
  /*  private static IName childElementsWithCommonName (Node p)
    {
      Name result = null;
  
      for (Node n = p.getFirstChild(); n != null; n = n.getNextSibling())
      {
        if (n.getNodeType() == Node.ELEMENT_NODE)
        {
          if (result == null)
            result = new Name (n);
          else if (! result.equalsNode (n))
            return null;
        }
      }
  
      return result;
    }
    */

  private static String dump(String s) {
    String res = "";

    for (int i = 0; i < s.length(); i++) {
      int x = s.charAt(i);

      x &= 0xffff;

      if (x == '\\')
        res += "\\\\";
      else if (x > 127 || x < 32)
        res += "\\u" + Integer.toHexString(x);
      else
        res += s.charAt(i);
    }

    return res;
  }
  
  /**
   * Convenience method for setting namespace declarations. Will set the declaration
   * both on element itself and on document element. Setting it on the element
   * itself is necessary because we may be streaming the result and the document element
   * itself already has been serialized.
   */
  private static void setNsDeclOnContainer(Element e, String name, String value) {
  
    Element container = e.getOwnerDocument().getDocumentElement();
    
    if (container != null) {
      container.setAttributeNS(Consts.NS_XMLNS, name, value);
    }
    
    if (container != e) {
      e.setAttributeNS(Consts.NS_XMLNS, name, value);
    }
    
  }

}