package com.sapportals.wcm.protocol.webdav;

import java.util.MissingResourceException;
import java.util.ResourceBundle;

import com.sap.tc.logging.Location;
import com.sapportals.wcm.util.logging.LoggingFormatter;

/** Implementation of {@link IInternationalString}.
 *
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author julian.reschke@greenbytes.de
 * @version $Revision: 1.4 $
 */
public class InternationalString implements IInternationalString {

  private final String languageTag;
  private final String text;

  private static final Location log = Location.getLocation(InternationalString.class);

  /**
   * Build InternationalString from tag and content.
   * @param languageTag language tag
   * @param text text content
   * @returns new InternationalString
   */
  public InternationalString (String languageTag, String text) {
    this.languageTag = languageTag;
    this.text = text;
  }

  /**
   * Build InternationalString from entry in bundle. When key is not found in
   * the bundle, log the problem and use the key instead.
   * @param bundle ResourceBundle from which to extract content and language
   * @param key key in bundle to use
   * @returns new InternationalString
   */
  public InternationalString (ResourceBundle bundle, String key) {
    this(bundle, null, key);
  }

  /**
   * Build InternationalString from entry in bundle. When key is not found in
   * the bundle, log the problem and use the key instead.
   * @param bundle ResourceBundle from which to extract content and language
   * @param secondaryBundle secondary (fallback) bundle
   * @param key key in bundle to use
   * @returns new InternationalString
   */
  public InternationalString (ResourceBundle bundle, ResourceBundle secondaryBundle, String key) {
    String tmp;
    String lang;
    
    lang = bundle.getLocale().getLanguage();
    try {
      tmp = bundle.getString(key);
    }
    catch (MissingResourceException ex) {
      // $JL-EXC$
      // stupid jlin
      if (secondaryBundle == null) {
        log.warningT("missing resource - " + LoggingFormatter.extractCallstack(ex));
        // fall back to key
        tmp = key;         
      }
      else {
        lang = secondaryBundle.getLocale().getLanguage();        
        try {
          tmp = secondaryBundle.getString(key);
        }
        catch (MissingResourceException ex2) {
          log.warningT("missing resource - " + LoggingFormatter.extractCallstack(ex));
          // fall back to key
          tmp = key;         
        }
      }
    }
    
    this.languageTag = lang;
    this.text = tmp;
  }

  public String getText () {
    return this.text;
  }

  public String getLanguageTag () {
    return this.languageTag;
  }

  /**
   * @return "[tag]text"
   */
  public String toString () {
    return "[" + this.languageTag + "] " + this.text;
  }

  public boolean matchesLocale (java.util.Locale loc, boolean exact) {

    if (log.beDebug()) {
      log.debugT("matching: " + this.languageTag + " " + loc.getLanguage());
    }

    if (this.languageTag == null || loc == null) {
      return false;
    }
    else {
      if (exact) {
        String country = loc.getCountry();
        if ("".equals(country)) {
          return this.languageTag.equalsIgnoreCase(loc.getLanguage());
        }
        else {
          return this.languageTag.equalsIgnoreCase(loc.getLanguage() + "_" + loc.getCountry());
        }
      }
      else {
        return this.languageTag.substring(0, 2).equalsIgnoreCase(loc.getLanguage());
      }
    }
  }

}



