/*
 * Copyright (c) 2004 by SAP AG. All Rights Reserved.
 *
 * SAP, mySAP, mySAP.com and other SAP products and
 * services mentioned herein as well as their respective
 * logos are trademarks or registered trademarks of
 * SAP AG in Germany and in several other countries all
 * over the world. MarketSet and Enterprise Buyer are
 * jointly owned trademarks of SAP AG and Commerce One.
 * All other product and service names mentioned are
 * trademarks of their respective companies.
 *
 * @version $Id: QPCodec.java,v 1.2 2004/02/19 18:31:05 jre Exp $
 */
package com.sapportals.wcm.protocol.webdav;

import java.io.IOException;
import java.io.UnsupportedEncodingException;

/**
 * Manages Quoted Printable Encoding/Decoding according to RFC 2045 Ch. 6.7<p>
 *
 * Copyright (c) SAP AG 2001-2004
 *
 * @author stefan.eissing@greenbytes.de
 * @version $Id: QPCodec.java,v 1.2 2004/02/19 18:31:05 jre Exp $
 */
public final class QPCodec {

  private final static String[] EscapeSequence = new String[]{
  // control 00 - 0f
    "=00", "=01", "=02", "=03", "=04", "=05", "=06", "=07", "=08", "=09", "=0A", "=0B", "=0C", "=0D", "=0E", "=0F",
  // control 10 - 1f
    "=10", "=11", "=12", "=13", "=14", "=15", "=16", "=17", "=18", "=19", "=1A", "=1B", "=1C", "=1D", "=1E", "=1F",
  // sp    !     "      #      $      %      &      '      (     )     *     +      ,      -      .      /
    "=20", null, "=22", "=23", "=24", "=25", "=26", "=27", null, null, null, "=2B", "=2C", null, null, null,
  // 0   1      2      3      4      5      6      7      8      9      :      ;      <      =      >      ?
    null, null, null, null, null, null, null, null, null, null, "=3A", "=3B", "=3C", "=3D", "=3E", "=3F",
  // @   A      B      C      D      E      F      G      H      I      J      K      L      M      N      O
    "=40", null, null, null, null, null, null, null, null, null, null, null, null, null, null, null,
  // P   Q      R      S      T      U      V      W      X      Y      Z      [      \      ]      ^      _
    null, null, null, null, null, null, null, null, null, null, null, null, "=5C", null, "=5E", null,
  // `   a      b      c      d      e    f      g     h    i     j     k     l     m     n     o
    "=60", null, null, null, null, null, null, null, null, null, null, null, null, null, null, null,
  // p   q      r      s     t     u     v     w      x      y    z      {      |      }      ~      7f
    null, null, null, null, null, null, null, null, null, null, null, "=7B", "=7C", "=7D", "=7E", "=7F",
    };

  private final static String[] encodedBytes;
  static {
    String tmp[] = new String[256];
    for (int i = 0; i < tmp.length; i++) {
      String hex = Integer.toHexString(i).toUpperCase();
      tmp[i] = "=" + (hex.length() == 1 ? "0" : "") + hex;
    }

    encodedBytes = tmp;
  }

  public static String Encode(String value, String encoding) throws UnsupportedEncodingException {
    byte[] bytes = value.getBytes(encoding);
    StringBuffer sb = new StringBuffer(bytes.length + 10);
    for (int i = 0, n = bytes.length; i < n; ++i) {
      byte b = bytes[i];
      if (b < 0) {
        // bit 8 set
        sb.append(getEncoding(b));
      }
      else {
        String escape = QPCodec.EscapeSequence[b];
        if (escape != null) {
          sb.append(escape);
        }
        else {
          sb.append((char)b);
        }
      }
    }
    return sb.toString();
  }

  /**
   * Converts a string in quoted printable format back to its original form.
   * quoted bytes are in the given character encoding.
   * @param value to decode
   * @param encoding to use for conversion bytes to characters
   * @return decoded value
   * @throws UnsupportedEncodingException when character encoding is not supported
   * @throws IOException when value is in invalid form
   */
  public static String Decode(String value, String encoding) 
  throws UnsupportedEncodingException, IOException {
    byte[] bytes = new byte[value.length()];
    int pos = 0;
    for (int i = 0, n = value.length(); i < n; ++i) {
      char c = value.charAt(i);
      if (c < 0 || c > 127) {
        throw new IOException("character "+c+" is not valid in quoted printables at offset "+
          i+" in: "+value);
      }
      
      if (c == '=') {
        if (n >= i + 2) {
          char c1 = value.charAt(i + 1);
          char c2 = value.charAt(i + 2);
          if (c1 == '\r' && c2 == '\n') {
            // soft linebreak, nothing to add to output
            i += 2;
          }
          else {
            int hnibble = Character.digit(c1, 16);
            int lnibble = Character.digit(c2, 16);
            if (hnibble >= 0 && lnibble >= 0) {
              bytes[pos++] = (byte)((hnibble << 4) | lnibble);
              i += 2;
            }
            else {
              throw new IOException("invalid escape sequence at offset "+i+" in: "+value);
            }
          }
        }
        else {
          throw new IOException("unterminated escape sequence at offset "+i+" in: "+value);
        }
      }
      else {
        bytes[pos++] = (byte)c;
      }
    }
    
    return new String(bytes, 0, pos, encoding);
  }

  //------------ private -------------------------------------------------------

  public final static String getEncoding(byte b) {
    return encodedBytes[b & 0x00ff];
  }
}
