package com.sapportals.wcm.protocol.webdav;

import org.w3c.dom.Node;

import com.sap.tc.logging.Location;

/**
 * A <a href="http://www.w3.org/TR/1999/REC-xml-names-19990114/#ns-qualnames">qualified and prefixex XML Name</a>,
 * consisting of namespace name, prefix and local name.
 *
 * <p>
 * Copyright (c) SAP AG 2001-2004 
 * @author julian.reschke@greenbytes.de
 * @version $Id: QualifiedName.java,v 1.2 2004/02/19 18:31:05 jre Exp $
 */

public class QualifiedName extends Name {

  private static Location log = Location.getLocation(QualifiedName.class);

  /**
   * gets the prefix from a prefixed name
   * @param prefixedName prefixed name
   * @return the prefix  or <code>null</code>
   */
  public static String getPrefix(String prefixedName) {
    int pos = prefixedName.indexOf(":");
    return (pos >= 0) ? prefixedName.substring(0, pos) : null;
  }

  /**
   * gets the local name from a prefixed name
   * @param prefixedName prefixed Name
   * @return the local name part
   */
  public static String getLocalName(String prefixedName) {
    int pos = prefixedName.indexOf(":");
    return (pos >= 0) ? prefixedName.substring(pos + 1) : prefixedName;
  }

  /**
   * Builds a QualifiedName from a string in the following formats:
   *
   * <ul>
   *  <li>{<a href="http://www.w3.org/TR/1999/REC-xml-names-19990114/#dt-NSName">namespaceUri</a>}<a href="http://www.w3.org/TR/1999/REC-xml-names-19990114/#NT-LocalPart">localPart</a></li>
   *  <li>xmlns:<a href="http://www.w3.org/TR/1999/REC-xml-names-19990114/#NT-Prefix">prefix</a></li>
   *  <li>xml:<a href="http://www.w3.org/TR/1999/REC-xml-names-19990114/#NT-LocalPart">localPart</a></li>
   *  <li><a href="http://www.w3.org/TR/1999/REC-xml-names-19990114/#NT-NCName">NCName</a></li>
   * </ul>
   *
   * @param name input
   * @throws java.text.ParseException when format not OK
   */

  public static QualifiedName getInstance(String name) throws java.text.ParseException {
    int pos;
    String l_name = name;
    String l_ns = null;
    String l_prefix = null;

    // try to dissect the name into a namespace and a name

    if (name.startsWith("{")) {
      pos = name.indexOf("}");

      if (pos > 0) {
        l_ns = name.substring(1, pos);
        l_name = name.substring(pos + 1);
      }
      else {
        log.errorT("QualiedName( + " + name + "): illegal format");
        throw new java.text.ParseException(name, 0);
      }
    }
    else if (name.startsWith("xml:")) {
      l_prefix = "xml:";
      l_ns = Consts.NS_XML;
      l_name = name.substring(4);
    }
    else if (name.startsWith("xmlns:")) {
      l_prefix = "xmlns:";
      l_ns = Consts.NS_XMLNS;
      l_name = name.substring(6);
    }
    else if ((pos = name.indexOf(':')) >= 0) {
      log.errorT("QualiedName( + " + name + "): illegal format");
      throw new java.text.ParseException(name, pos);
    }

    return new QualifiedName(l_ns, l_name, l_prefix);
  }

  private final String m_prefix;

  /**
   * @param namespaceUri namespace URI
   * @param localName
   * @param prefix
   */

  public QualifiedName(String namespaceUri, String localName, String prefix) {
    super(namespaceUri, localName);
    m_prefix = prefix;
  }

  /**
   * @param namespaceUri namespace URI
   * @param qname a <a href="http://www.w3.org/TR/1999/REC-xml-names-19990114/#ns-qualnames">QName</a>
   */

  public QualifiedName(String namespaceUri, String qname) {
    this(namespaceUri, getLocalName(qname), getPrefix(qname));
  }

  /**
   * @param node DOM element
   */
  public QualifiedName(Node node) {
    this(node.getNamespaceURI(), node.getLocalName(), node.getPrefix());
  }

  public String getPrefix() {
    return m_prefix;
  }
}