package com.sapportals.wcm.protocol.webdav;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.OutputStreamWriter;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.sap.tc.logging.Location;
import com.sapportals.wcm.IWcmConst;
import com.sapportals.wcm.repository.*;
import com.sapportals.wcm.util.content.IContent;
import com.sapportals.wcm.util.http.Base64;
import com.sapportals.wcm.util.http.DateFormat;
import com.sapportals.wcm.util.logging.LoggingFormatter;
import com.sapportals.wcm.util.uri.IHierarchicalUri;

/**
 * Simple RSS provider.
 * <p>
 * Copyright (c) SAP AG 2004
 * @author julian.reschke@greenbytes.de
 */
public class RSSProvider {
  
  private static final Location log = Location.getLocation(RSSProvider.class);
  
  private static final String NS_RDF = "http://www.w3.org/1999/02/22-rdf-syntax-ns#";
  private static final String NS_RSS = "http://purl.org/rss/1.0/";
  private static final String NS_DC = "http://purl.org/dc/elements/1.1/";
  private static final String NS_CRT = IWcmConst.SAP_WCM_NAMESPACE + com.sapportals.wcm.IWcmConst.NAMESPACE_SEPARATOR + "crt";
  
  private static final Name N_dc_creator = new Name(NS_DC, "dc:creator");
  private static final Name N_dc_date = new Name(NS_DC, "dc:date");

  private static final Name N_rdf_RDF = new Name(NS_RDF, "rdf:RDF");

  private static final Name N_rss_channel = new Name(NS_RSS, "channel");
  private static final Name N_rss_description = new Name(NS_RSS, "description");
  private static final Name N_rss_item = new Name(NS_RSS, "item");
  private static final Name N_rss_link = new Name(NS_RSS, "link");
  private static final Name N_rss_title = new Name(NS_RSS, "title");

  private static IPropertyName N_crt_configurationerror;
  private static IPropertyName N_crt_startuperror;
  private static IPropertyName N_crt_state;
  private static IPropertyName N_dav_comment;
  
  static {
    try {
      N_crt_configurationerror = new PropertyName(NS_CRT, "configurationerror");
      N_crt_startuperror = new PropertyName(NS_CRT, "startuperror");
      N_crt_state = new PropertyName(NS_CRT, "state");
      N_dav_comment = new PropertyName(Consts.N_comment);
    }
    catch (ResourceException e) {
      throw new RuntimeException(e.getMessage());
    }
  }

  private RSSProvider() {
  }
  
  public static RSSProvider getInstance() {
    return new RSSProvider();
  }
  
  public IContent getContentForResource(IResource res, IHierarchicalUri baseUri) throws ResourceException {
    
    Document doc = DomBuilder.newDocument(N_rdf_RDF);
    Element rdf = doc.getDocumentElement();

    String link, desc, disp;
    DateFormat df = new DateFormat();

    rdf.setAttributeNS(Consts.NS_XMLNS, "xmlns:dc", NS_DC);
    rdf.setAttributeNS(Consts.NS_XMLNS, "xmlns:rdf", NS_RDF);
    
    Element channel = DomElement.makeNew(doc, N_rss_channel);    
    link = baseUri.mapToAbsoluteUri(res.getRID()).toExternalForm();
    channel.setAttributeNS(NS_RDF, "rdf:about", link);
    desc = res.getDescription();
    if (desc != null && desc.length() > 0) channel.appendChild(DomElement.makeNew(doc, N_rss_description, desc));
    channel.appendChild(DomElement.makeNew(doc, N_rss_link, link));
    disp = res.getDisplayName();
    
    IVersionHistoryResource vhr = (IVersionHistoryResource) res.as(IVersionHistoryResource.class);
    if (vhr == null) {
      channel.appendChild(DomElement.makeNew(doc, N_rss_title, "RSS feed for " + (disp.length() > 0 ? disp : res.getName())));
    }
    else {
      channel.appendChild(DomElement.makeNew(doc, N_rss_title, "RSS feed version history " + (disp.length() > 0 ? disp : res.getName())));
    }
    
    rdf.appendChild(channel);

    if (res.isCollection()) {
      addChildren(baseUri, doc, rdf, df, res.getName() + "/", ((ICollection)res).getChildren());
    }
 
    if (res.isVersioned()) {
      IVersionControlledResource vcr = (IVersionControlledResource)res.as(IVersionControlledResource.class);
      if (vcr != null) {
        IResource t = vcr.getVersionHistoryResource();
        IVersionHistoryResource vhrx = t == null ? null : (IVersionHistoryResource)t.as(IVersionHistoryResource.class);
        if (vhrx != null) {
          addChildren(baseUri, doc, rdf, df, res.getName() + " version ", vhrx.getVersionResourceSet());
        }
      }
    }
    
    ByteArrayOutputStream os = new ByteArrayOutputStream();
    try {
      DomSerializer.write(new OutputStreamWriter(os, "UTF-8"), doc, true);
    }
    catch (Exception e) {
      throw new ResourceException(e);
    }

    byte[] ba = os.toByteArray();

    String etag = null;
    try {
      etag = Base64.encode(MessageDigest.getInstance("MD5").digest(ba));
    }
    catch (NoSuchAlgorithmException e1) {
      log.debugT(LoggingFormatter.extractCallstack(e1));
    }

    Content content = new Content(new ByteArrayInputStream(ba), "application/rss+xml", os.size());
    if (etag != null) content.setEntityTag(etag);

    return content;
  }

  private void addChildren(IHierarchicalUri baseUri, Document doc, Element rdf, DateFormat df, String pref, IResourceList children) throws ResourceException, NotSupportedException, AccessDeniedException {

    String link;
    String desc;
    String disp;

    for (IResourceListIterator it = children.listIterator(); it.hasNext(); ) {
      IProperty pt;
      IResource r = it.next();
      Element item = DomElement.makeNew(doc, N_rss_item);
      link = baseUri.mapToAbsoluteUri(r.getRID()).toExternalForm();
      item.setAttributeNS(NS_RDF, "rdf:about", link);
      item.appendChild(DomElement.makeNew(doc, N_rss_link, link));
      
      desc = r.getDescription();
      if (desc != null && desc.length() > 0) item.appendChild(DomElement.makeNew(doc, N_rss_description, desc));

      disp = r.getDisplayName();
      if (disp.length() == 0) disp = pref + r.getName();
      pt = r.getProperty(N_crt_state);
      if (pt != null) disp += " (" + pt.getValueAsString() + ")";
      pt = r.getProperty(N_crt_configurationerror);
      if (pt != null && pt.getValueAsString().length() > 0) disp += " (config error: " + pt.getValueAsString() + ")";
      pt = r.getProperty(N_crt_startuperror);
      if (pt != null && pt.getValueAsString().length() > 0) disp += " (startup error: " + pt.getValueAsString() + ")";
      pt = r.getProperty(N_dav_comment);
      if (pt != null && pt.getValueAsString().length() > 0) disp += " (checkin comment: " + pt.getValueAsString() + ")";
      
      item.appendChild(DomElement.makeNew(doc, N_rss_title, (disp.length() > 0 ? disp : pref + r.getName())));

      String creator = r.getLastModifiedBy();
      if (creator != null && creator.length() > 0) item.appendChild(DomElement.makeNew(doc, N_dc_creator, creator));

      String dat = df.formatISO8601(r.getLastModified());
      item.appendChild(DomElement.makeNew(doc, N_dc_date, dat));
      
      rdf.appendChild(item); 
    }
  }
  
  
  
}
