package com.sapportals.wcm.protocol.webdav;

import com.sap.tc.logging.Location;

import com.sapportals.wcm.service.urlgenerator.IUriMapper;
import com.sapportals.wcm.util.uri.*;

/** Maps uri references to WCM Paths (RID) in a context of a base Uri.
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author stefan.eissing@greenbytes.de
 * @version $Revision: 1.3 $
 */
public class UriContext {

  private static final Location log = Location.getLocation(UriContext.class);

  private final RID m_basePath;
  private final IUriMapper m_mapper;
  private final IHierarchicalUri m_baseUri;
  private final IUriReference m_baseRef;

  /**
   * Create a new UriContext which follows the given mapping and has
   * the given WCM Path as base.
   * @param basePath anchor for mapping in WCM Path space
   * @param mapper for general mapping between WCM Path and Uri space
   */
  public UriContext(RID basePath, IUriMapper mapper) {
    m_basePath = basePath;
    m_mapper = mapper;
    m_baseUri = m_mapper.toAbsoluteUri(m_basePath);
    if (m_baseUri == null) {
      log.errorT("base path("+m_basePath+") mapped to null, ShouldNotHappen!, mapper: "+m_mapper);
    }
    m_baseRef = new UriReference(m_baseUri.getPath(), m_baseUri.getQuery(), null);
  }

  /**
   * Retrieve the base WCM Path of this context
   * @return base WCM Path of this context
   */
  public RID getBasePath() {
    return m_basePath;
  }

  /**
   * Retrieve the base uri as reference without scheme and host name
   * @return base uri as reference without scheme and host name
   */
  public IUriReference getBaseReference() {
    return m_baseRef;
  }

  /**
   * Retrieve the base uri of this context
   * @return the base uri of this context
   */
  public IHierarchicalUri getBaseUri() {
    return m_baseUri;
  }

  /**
   * Maps the external string representation of an uri reference to
   * the the local WCM Path.
   * @param externalForm of uri reference
   * @return mapped WCM Path or null if no mapping possible
   */
  public RID mapWcmPath(String externalForm) {
    try {
      IUriReference ref = UriFactory.parseUriReference(externalForm);
      return mapWcmPath(ref);
    }
    catch (java.net.MalformedURLException ex) {
      log.debugT("parsing Uri reference" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(ex));
    }
    return null;
  }

  /**
   * Maps the uri reference to the the local WCM Path.
   * @param ref uri reference
   * @return mapped WCM Path or null if no mapping possible
   */
  public RID mapWcmPath(IUriReference ref) {
    IUri uri = m_baseUri.resolve(ref);
    if (uri instanceof IHierarchicalUri) {
      RID rid = m_mapper.toRID((IHierarchicalUri)uri);
      if (rid == null) {
        rid = UriMetaMap.getInstance().toRID(this.m_basePath, (IHierarchicalUri)uri);
      }
      return rid;
    }
    return null;
  }
}


