package com.sapportals.wcm.protocol.webdav;

import java.util.ArrayList;
import java.util.List;

import com.sapportals.wcm.service.urlgenerator.IUriMapper;
import com.sapportals.wcm.util.uri.*;

/**
 * Default implementation of {@link IUriMapper}.
 * <p>
 * Copyright (c) SAP AG 2001-2005
 * @author stefan.eissing@greenbytes.de
 * @version $Revision: 1.4 $
 */
public class UriMapper implements IUriMapper {

  // private static final com.sap.tc.logging.Location log = com.sap.tc.logging.Location.getLocation(UriMapper.class.getName());

  private final IHierarchicalUri rootUri;
  private final RID rootRID;
  private final String rootUriPath;
  private final int rootUriPathLen;
  private final int ridPathLen;
  private final boolean checkTrailingSlash;
  private final boolean caseSensitive;
  private final RID rootRIDNoSlash;
  private final List aliases;

  /**
   * Create a new mapping between root path and uri
   * @param rootPath of mapping
   * @param rootUri uri of mapping
   */
  public UriMapper(RID rootPath, IHierarchicalUri rootUri) {
    this(rootPath, rootUri, null);
  }
  
  /**
   * Create a new mapping between root path and uri
   * @param rootPath of mapping
   * @param rootUri uri of mapping
   * @param aliases list of {@link IHierarchicalUri}s pointing to same server
   */
  public UriMapper(RID rootPath, IHierarchicalUri rootUri, List aliases) {
    this(rootPath, rootUri, null, true); 
  }
  
  /**
   * Create a new mapping between root path and uri
   * @param rootPath of mapping
   * @param rootUri uri of mapping
   * @param aliases list of {@link IHierarchicalUri}s pointing to same server
   * @param caseSensitive mapping of RIDs/URIs
   */
  public UriMapper(RID rootPath, IHierarchicalUri rootUri, 
      List aliases, boolean caseSensitive) {
    this.rootUri = (rootUri.getPath().length() > 0)? rootUri : rootUri.appendPath("/");
    this.rootRID  = rootPath;
    this.rootRIDNoSlash = this.rootRID.removeTrailingSlash();
    this.rootUriPath = this.rootUri.getPath();
    this.rootUriPathLen = this.rootUriPath.length();
    this.ridPathLen = this.rootRID.getPath().length();
    this.checkTrailingSlash = (this.rootUriPathLen > 0
      && (this.rootUriPath.charAt(this.rootUriPathLen-1) == '/')
      && (this.rootUri.getQuery() == null));
    this.aliases = getHierarchicalUris(aliases);
    this.caseSensitive = caseSensitive;
  }

  public boolean isGeneric() {
    return false;
  }

  public boolean isCaseSensitive() {
  	return this.caseSensitive;
  }
  
  public IHierarchicalUri getRootUri() {
    return this.rootUri;
  }

  public RID getRootRID() {
    return this.rootRID;
  }

  public IHierarchicalUri toAbsoluteUri(RID path) {
    if (isRootRID(path)) {
      return this.rootUri;
    }

    path = getRelativePath(path);
    if (path != null) {
      return this.rootUri.mapToAbsoluteUri(path);
    }
    return null;
  }

  public IUriReference toAbsolutePath(RID path) {
    if (isRootRID(path)) {
      return new UriReference(this.rootUri.getPath(), this.rootUri.getQuery(), null);
    }

    path = getRelativePath(path);
    if (path != null) {
      return this.rootUri.mapToAbsolutePath(path);
    }
    return null;
  }

  public RID toRID(IHierarchicalUri uri) {
    if (this.rootUri.isAncestorOf(uri)) {
      String rpath = URICodec.Decode(uri.getPath().substring(this.rootUriPathLen));
      return this.rootRID.add(hideQueryInternal(RID.getRID(rpath, uri.getQuery())));
    }
    else if (this.rootUri.equals(uri)) {
      return this.rootRID;
    }
    else if (this.checkTrailingSlash
      && this.rootUri.getAuthority().equals(uri.getAuthority())
      && this.rootUri.getScheme().equals(uri.getScheme())
      && uri.getQuery() == null) {
      // Same scheme, query and authority, see if the path differ only by
      // an ending slash character. Can only happen if rootUri
      // ends with slash and uri does not. Otherwise isAncestor()
      // would have returned true.
      //
      String upath = uri.getPath();
      if (this.rootUriPath.startsWith(upath) && (this.rootUriPathLen == upath.length() + 1)) {
        return this.rootRIDNoSlash;
      }
    }
    
    if (this.aliases != null) {
      //default mapping did not work, maybe the uri is a child of some alias we have
      for (int i = 0, n = this.aliases.size(); i < n; ++i) {
        IHierarchicalUri root = (IHierarchicalUri)this.aliases.get(i);
        int rootUriPathLen = root.getPath().length();
        if (root.isAncestorOf(uri)) {
          String rpath = URICodec.Decode(uri.getPath().substring(rootUriPathLen));
          return getRootRID().add(hideQueryInternal(RID.getRID(rpath, uri.getQuery())));
        }
        else if (root.equals(uri)) {
          return getRootRID();
        }
        else {
          if (root.getAuthority().equals(uri.getAuthority())
            && root.getScheme().equals(uri.getScheme())
            && uri.getQuery() == null) {
            if (rootUriPathLen > 0
                && (root.getPath().charAt(rootUriPathLen-1) == '/')
                && (root.getQuery() == null)) {
              // Same scheme, query and authority, see if the path differ only by
              // an ending slash character. Can only happen if rootUri
              // ends with slash and uri does not. Otherwise isAncestor()
              // would have returned true.
              //
              String upath = uri.getPath();
              if (root.getPath().startsWith(upath) && (rootUriPathLen == upath.length() + 1)) {
                return getRootRID().removeTrailingSlash();
              }
            }
          }
        }
      }
      return null;
    }
    return null;
  }

  public String toString() {
    return "UriMapper["+this.rootRID+" <-> "+this.rootUri+"]";
  }

  //------------------------- protected ---------------------------------

  protected RID hideQueryInternal(RID path) {
    return path;
  }

  protected RID discoverQueryInternal(RID path) {
    return path;
  }

  //------------------------- private -----------------------------------

  private boolean isRootRID(RID rid) {
  	return (this.rootRID.endsWithSlash() == rid.endsWithSlash())
      && (this.rootRID.equals(rid)
      || ((!this.caseSensitive) && this.rootRID.getPath().equalsIgnoreCase(rid.getPath())));
  }

  private boolean isBelowRoot(RID path) {
    if (this.caseSensitive) {
      return this.rootRID.isAncestorOf(path);
    }
    else {
      String rootPath = this.rootRID.getPath().toLowerCase();
      String rpath = path.getPath().toLowerCase();
      if ((rootPath.length() < rpath.length()) && rpath.startsWith(rootPath)) {
        return this.rootRID.endsWithSlash() 
          || rpath.charAt(rootPath.length()) == RID.PATH_SEPARATOR;
      }
    }
    return false;
  }
  
  private RID getRelativePath(RID path) {
    if (isBelowRoot(path)) {
      path = discoverQueryInternal(path);
      String rpath = path.getPath().substring(this.ridPathLen);
      return RID.getRID(rpath, path.getQuery());
    }
    else if (path.getQuery() != null) {
      RID qless = RID.getRID(path.getPath(), null);
      if (this.rootRID.equals(qless)) {
        return RID.getRID("", path.getQuery()); 
      }
    }
    return null;
  }

  private static List getHierarchicalUris(List aliases) {
    if (aliases != null && !aliases.isEmpty()) {
      List result = new ArrayList(aliases.size());
      for (int i = 0, n = aliases.size(); i < n; ++i) {
        Object o = aliases.get(i);
        if (o instanceof IHierarchicalUri) {
          result.add(o);
        }
      }
      return result;
    }
    return null;
  }

}


