package com.sapportals.wcm.protocol.webdav;

import java.util.*;

import com.sap.tc.logging.Location;

import com.sapportals.wcm.service.urlgenerator.IUriMapper;
import com.sapportals.wcm.util.uri.IHierarchicalUri;
import com.sapportals.wcm.util.uri.RID;

/**
 * Handles system wide mappings or RIDs to Uris.
 * <p>
 * Copyright (c) SAP AG 2002-2003
 * @author stefan.eissing@greenbytes.de
 * @version $Revision: 1.2 $
 */
public final class UriMetaMap {

  private static final Location log = Location.getLocation(UriMetaMap.class);

  private static final UriMetaMap g_instance = new UriMetaMap();

  /**
   * Get singleton instance of UriMetaMap.
   * @return single instance of UriMetaMap
   */
  public static UriMetaMap getInstance() {
    return g_instance;
  }

  private final MapperMap mappers;
  private final GenericMapperMap genericMappers;

  private UriMetaMap() {
    this.mappers = new MapperMap();
    this.genericMappers = new GenericMapperMap();
  }

  /**
   * Map the given uri to a local RID.
   * @param base RID context for mapping
   * @param uri to map to local RID
   * @return mapped RID or <code>null</code> if no mapping is available
   */
  public RID toRID(RID base, IHierarchicalUri uri) {
    IUriMapper mapper = this.mappers.getFor(uri);
    if (mapper != null) {
      return mapper.toRID(uri);
    }

    // No specific Mapper available, try the best-fit generic one
    if (base != null) {
      mapper = this.genericMappers.getFor(base);
      if (mapper != null) {
        RID rid = mapper.toRID(uri);
        if (rid != null) {
          return rid;
        }
      }
    }

    // Either we had no base, or the best-fit generic mapper
    // refuses to map this uri. Try all generic mappers
    return this.genericMappers.toRID(uri);
  }

  /**
   * Register a mapper.
   * @param mapper to install
   */
  public void register(IUriMapper mapper) {
    if (mapper.isGeneric()) {
      this.genericMappers.register(mapper);
    }
    else {
      this.mappers.register(mapper);
    }
  }

  /**
   * Deregister a mapping.
   * @param mapper to remove.
   */
  public void deregister(IUriMapper mapper) {
    if (mapper.isGeneric()) {
      this.genericMappers.unregister(mapper);
    }
    else {
      this.mappers.unregister(mapper);
    }
  }

  // -------------------------------- private ---------------------------------

  private static final class MapperMap {

    /** preserves registration sequence so that lookups via Uris always
     * find the same (first matching) mapper.
     */
    private final List mappers;
    private final Map rid2mapper;

    public MapperMap() {
      this.mappers = new ArrayList();
      this.rid2mapper = new HashMap();
    }

    public synchronized IUriMapper getFor(RID rid) {
      return (IUriMapper)this.rid2mapper.get(rid);
    }

    public synchronized IUriMapper getFor(IHierarchicalUri uri) {
      for (Iterator iter = this.mappers.iterator(); iter.hasNext(); ) {
        IUriMapper mapper = (IUriMapper)iter.next();
        IHierarchicalUri root = mapper.getRootUri();
        if (root.equals(uri) || root.isAncestorOf(uri)) {
          return mapper;
        }
      }
      return null;
    }

    private synchronized void register(IUriMapper mapper) {
      IUriMapper existing = getFor(mapper.getRootRID());
      if (existing != null) {
        log.warningT("replacing existing mapping "+existing);
        unregister(existing);
      }
      this.mappers.add(mapper);
      this.rid2mapper.put(mapper.getRootRID(), mapper);
    }

    private synchronized void unregister(IUriMapper mapper) {
      this.mappers.remove(mapper);
      this.rid2mapper.remove(mapper.getRootRID());
    }
  }

  private static final class GenericMapperMap {

    private final List mappers;
    private final Map rid2mappers;

    public GenericMapperMap() {
      this.mappers = new ArrayList();
      this.rid2mappers = new HashMap();
    }

    /**
     * Gives the best matching or first installed generic mapper.
     * @param base to determine best match
     * @return generic mapper or <code>null</code> if none is installed.
     */
    public synchronized IUriMapper getFor(RID base) {
      IUriMapper mapper = (IUriMapper)this.rid2mappers.get(base);
      if (mapper != null) {
        return mapper;
      }

      for (Iterator i = this.rid2mappers.values().iterator(); i.hasNext(); ) {
        mapper = (IUriMapper)i.next();
        if (mapper.getRootRID().isAncestorOfOrSelf(base)) {
          return mapper;
        }
      }
      return null;
    }

    public synchronized RID toRID(IHierarchicalUri uri) {
      for (int i = 0, n = this.mappers.size(); i < n; ++i) {
        IUriMapper mapper = (IUriMapper)this.mappers.get(i);
        RID rid = mapper.toRID(uri);
        if (rid != null) {
          return rid;
        }
      }
      return null;
    }

    public synchronized void register(IUriMapper mapper) {
      this.mappers.add(mapper);
      this.rid2mappers.put(mapper.getRootRID(), mapper);
    }

    public synchronized void unregister(IUriMapper mapper) {
      this.mappers.remove(mapper);
      this.rid2mappers.remove(mapper.getRootRID());
    }
  }
}


