package com.sapportals.wcm.protocol.webdav;

import java.text.ParseException;
import java.util.Date;

import org.w3c.dom.Document;
import org.w3c.dom.Node;

import com.sapportals.wcm.protocol.webdav.dompath.DomPath;
import com.sapportals.wcm.repository.*;
import com.sapportals.wcm.util.http.DateFormat;
import com.sapportals.wcm.util.http.IResponse;
import com.sapportals.wcm.util.uri.RID;

/** Maps HTTP codes to {@link ResourceException}s.
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author stefan.eissing@greenbytes.de
 * @version $Id: WDHttp.java,v 1.4 2004/07/22 14:36:13 jre Exp $
 */
public class WDHttp {

  // private static final com.sap.tc.logging.Location log = com.sap.tc.logging.Location.getLocation(WDHttp.class.getName());

  private static final DomPath DP_Error = new DomPath("D:" + Consts.EN_error, DomPath.DEFAULT_NAMESPACE_BINDINGS);

  /**
   * Construct a resource exception for the give HTTP response
   * code and message.
   * @param uc UriContext of the concerned resource
   * @param response from the remote system
   * @param message for this error
   * @return new ResourceException
   */
  public static ResourceException exception(UriContext uc, IResponse response, String message) {
    return exception(uc, response, message, null);
  }

  /**
   * Construct a resource exception for the give HTTP response
   * code and message.
   * @param uc UriContext of the concerned resource
   * @param response from the remote system
   * @param message for this error
   * @return new ResourceException
   */
  public static ResourceException exception(UriContext uc, IResponse response, String message, RID target) {
    ResourceException ex = getError(response, uc, message, target);
    if (ex == null) {
      ex = exception(uc.getBasePath(), response.getStatus(), null, message, response);
    }
    return ex;
  }

  /**
   * Construct a resource exception for the given HTTP response
   * code and message.
   * @param rid of the concerned resource
   * @param httpCode from the remote system
   * @param message for this error
   * @return new ResourceException
   */
  public static ResourceException exception(RID rid, int httpCode, String message) {
    return exception(rid, httpCode, null, message, null);
  }

  public static ResourceException exception(RID rid, int httpCode, Name condition, String message) {
    return exception(rid, httpCode, condition, message, null);
  }

  /**
   * Construct a resource exception for the given HTTP response
   * code and message.
   * @param rid of the concerned resource
   * @param response for this response
   * @return new ResourceException
   */
  public static ResourceException exception(RID rid, IResponse response) {
    return exception(rid, response.getStatus(), null, response.getStatusDescription(), response);
  }

  private static ResourceException exception(RID rid, int httpCode, Name condition, String message, IResponse response) {

    // we may not need to consider the status code if we have a condition code
    if (condition != null) {

      if (DavConditions.N_segment_must_identity_member.equals(condition)) {
        return new InvalidPositionException(null, rid);
      }
    }

    switch (httpCode) {
      case 301:
      case 302:
      case 303:
      case 305:
      case 307:
      case 404:
        return new ResourceNotFoundException(message, rid);

      case 400:
        return new InvalidArgumentException(message, rid);

      case 401:
        return new AuthorizationRequiredException(message);

      case 403:
        return new AccessDeniedException(message, rid, null, null);

      case 405:
        return new MethodNotAllowedException(message, rid);

      case 407:
        return new AuthorizationRequiredException(message);

      case 408:
        return new IOErrorException(message, rid);

      case 409:
        return new InvalidArgumentException(message, rid);

      case 412:
        return new LockNotFoundException(message, rid);

      case 423:
        return new LockedException(message, rid, null);

      case 501:
        return new NotSupportedException(message, rid);

      case 503:
        return new ServiceUnavailableException(message, getRetryAfter(response), rid);

      case 506:
        return new LinkLoopException(message, rid);

      default:
        return new IOErrorException(message, rid);
    }
  }

  public static ResourceException getError(IResponse response, UriContext uc, String message, RID target) {
    Document doc = response.getDocument();
    if (doc != null) {
      Node error = DP_Error.matchFirst(doc);
      if (error != null) {
        if (DomElement.hasChild(error, DavConditions.N_can_overwrite)) {
          return new NameAlreadyExistsException(message, uc.getBasePath());
        }
        else if (DomElement.hasChild(error, DavConditions.N_resource_must_be_null)) {
          return new NameAlreadyExistsException(message, uc.getBasePath());
        }
        else if (DomElement.hasChild(error, DavConditions.N_cannot_modify_checked_in_parent)) {
          return new NotCheckedOutException(message, uc.getBasePath().parent());
        }
        else if (DomElement.hasChild(error, DavConditions.N_collection_must_be_ordered)) {
          return new NotOrderedException(uc.getBasePath());
        }
        else if (DomElement.hasChild(error, DavConditions.N_cannot_modify_version_controlled_property)) {
          return new NotCheckedOutException(message, uc.getBasePath());
        }
        else if (DomElement.hasChild(error, DavConditions.N_must_be_checked_in)) {
          return new CheckedOutException(message, uc.getBasePath(), null);
        }
        else if (DomElement.hasChild(error, DavConditions.N_must_be_checked_out)) {
          return new NotCheckedOutException(message, uc.getBasePath());
        }
        else if (DomElement.hasChild(error, DavConditions.N_cannot_copy_history)) {
          return new NotSupportedException("cannot copy version history resources", uc.getBasePath());
        }
        else if (DomElement.hasChild(error, DavConditions.N_cannot_rename_history)) {
          return new NotSupportedException("cannot rename version history resources", uc.getBasePath());
        }
        else if (DomElement.hasChild(error, DavConditions.N_preserve_versioning_properties)) {
          return new NotSupportedException("cannot preserve versioning properties", uc.getBasePath());
        }
        else if (DomElement.hasChild(error, DavConditions.N_SAP_destination_must_be_in_collection)) {
          return new InvalidTargetException(target);
        }
        else if (DomElement.hasChild(error, DavConditions.N_SAP_parent_collection_must_exist)) {
          return new ResourceNotFoundException(uc.getBasePath().parent());
        }
        else if (DomElement.hasChild(error, DavConditions.N_SAP_resource_supports_entity_body)) {
          return new NotSupportedException(message, uc.getBasePath());
        }
        else if (DomElement.hasChild(error, DavConditions.N_must_be_checked_out_version_controlled_resource)) {
          return new NotSupportedException("must be version controlled and checked out", uc.getBasePath());
        }
        else if (DomElement.hasChild(error, DavConditions.N_cycle_allowed)) {
          return new LinkLoopException(null);
        }
        else if (DomElement.hasChild(error, DavConditions.N_name_allowed)) {
          if (response.getStatus() == 409) {
            return new UnassignableNameException(message); // this one is retryable
          }
          else {
            return new InvalidNameException(message, null);
          }
        }
      }
    }
    return null;
  }

  private static Date getRetryAfter(IResponse response) {
    if (response != null) {
      String retryAfter = response.getHeader("Retry-After");
      if (retryAfter != null) {
        try {
          int seconds = Integer.parseInt(retryAfter);
          if (seconds >= 0) {
            long rtime = System.currentTimeMillis() + (seconds * 1000L);
            return new Date(rtime);
          }
        }
        catch (NumberFormatException e) {
          // $JL-EXC$
          // well, it's not in seconds format, may it's a HTTP date
          // fall through
        }
        try {
          DateFormat df = new DateFormat();
          return df.parseHTTP(retryAfter);
        }
        catch (ParseException e) {
          // $JL-EXC$
          // well, it's not a date either. Give up
        }
      }
    }
    return null;
  }

}
