package com.sapportals.wcm.protocol.webdav;

import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import org.w3c.dom.Attr;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import com.sap.tc.logging.Location;
import com.sapportals.wcm.repository.IProperty;
import com.sapportals.wcm.repository.ResourceException;
import com.sapportals.wcm.repository.enum.PropertyType;
import com.sapportals.wcm.util.logging.LoggingFormatter;
import com.sapportals.wcm.util.name.IName;

/**
 * Support for <a href="http://www.w3.org/TR/xmlschema-2/">XSD data types</a>.
 *
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author julian.reschke@greenbytes.de
 * @version $Id: XsdDataTypes.java,v 1.4 2004/05/07 10:01:46 jre Exp $
 */

public class XsdDataTypes {

  private static final Location log = Location.getLocation(XsdDataTypes.class);

  public static final QualifiedName XSI_TYPE = new QualifiedName(Consts.NS_XSI, "xs:type");

  /**
   * <a href="http://www.w3.org/TR/xmlschema-2/#boolean">xs:boolean</a>
   */
  public static final QualifiedName XSD_BOOLEAN = new QualifiedName(Consts.NS_XSD, "xs:boolean");

  /**
   * <a href="http://www.w3.org/TR/xmlschema-2/#dateTime">xs:dateTime</a>
   */
  public static final QualifiedName XSD_DATETIME = new QualifiedName(Consts.NS_XSD, "xs:dateTime");

  /**
   * <a href="http://www.w3.org/TR/xmlschema-2/#int">xs:int</a>
   */
  public static final QualifiedName XSD_INT = new QualifiedName(Consts.NS_XSD, "xs:int");

  /**
   * <a href="http://www.w3.org/TR/xmlschema-2/#long">xs:long</a>
   */
  public static final QualifiedName XSD_LONG = new QualifiedName(Consts.NS_XSD, "xs:long");

  /**
   * <a href="http://www.w3.org/TR/xmlschema-2/#string">xs:string</a>
   */
  public static final QualifiedName XSD_STRING = new QualifiedName(Consts.NS_XSD, "xs:string");

  /**
   * <a href="http://www.w3.org/TR/xmlschema-2/#anyURI">xs:anyURI</a>.
   * Derived from {@link XsdDataTypes#XSD_STRING}.
   */
  // public static final QualifiedName XSD_ANYURI = new QualifiedName(Consts.NS_XSD, "xs:anyURI");

  /**
   * <a href="http://www.w3.org/TR/SOAP/#_Toc478383522">soap-enc:array</a>
   * <p>
   * This type is used for marshalling multi-valued properties. 
   */
  public static final QualifiedName SOAP_ARRAY = new QualifiedName(Consts.NS_SOAP_ENC, "soap-enc:Array");

  /**
   * Holds a serialized form of an XML name using the format "{<em>namespace-uri</em>}<em>local-name</em>".
   * Derived from {@link XsdDataTypes#XSD_STRING}.
   */
  public static final QualifiedName SAP_EXPANDED_NAME = new QualifiedName(Consts.NS_SAP_DATATYPES, "sapportals-datatypes:expanded-name");

  /**
   * Holds a URI that represents a CM-internal resource (mapped from a {@link com.sapportals.wcm.util.uri.RID}).
   * Derived from {@link XsdDataTypes#XSD_STRING}.
   */
  public static final QualifiedName SAP_MAPPED_RID = new QualifiedName(Consts.NS_SAP_DATATYPES, "sapportals-datatypes:mapped-rid");

  /**
   * Holds a list of URIs that represent a CM-internal resources (mapped from a {@link com.sapportals.wcm.util.uri.RID} multivalued).
   * Derived from {@link XsdDataTypes#XSD_STRING}.
   */
  public static final QualifiedName SAP_MAPPED_RID_LIST = new QualifiedName(Consts.NS_SAP_DATATYPES, "sapportals-datatypes:mapped-rid-list");

  /**
   * Holds a {@link com.sapportals.wcm.util.uri.RID}. Usually, this type is serialized as {@link XsdDataTypes#SAP_MAPPED_RID}.
   * Derived from {@link XsdDataTypes#XSD_STRING}.
   */
  public static final QualifiedName SAP_RID = new QualifiedName(Consts.NS_SAP_DATATYPES, "sapportals-datatypes:rid");

  /**
   * Holds XML markup.
   */
  public static final QualifiedName SAP_XMLMARKUP = new QualifiedName(Consts.NS_SAP_DATATYPES, "sapportals-datatypes:markup");

  public static final String XN_TYPE = "{" + XsdDataTypes.XSI_TYPE.getNamespace() + "}" + XsdDataTypes.XSI_TYPE.getLocalName();
  public static final String XN_EXPANDED_NAME = "{" + XsdDataTypes.SAP_RID.getNamespace() + "}" + XsdDataTypes.SAP_EXPANDED_NAME.getLocalName();
  public static final String XN_RID = "{" + XsdDataTypes.SAP_RID.getNamespace() + "}" + XsdDataTypes.SAP_RID.getLocalName();
  // public static final String XN_URI = "{" + XsdDataTypes.XSD_ANYURI.getNamespace() + "}" + XsdDataTypes.XSD_ANYURI.getLocalName();
  
  public static final Properties ATTRS_TYPE_EXPANDED_NAME;
  public static final Properties ATTRS_TYPE_RID;
  

  private static final Map xsdToSap;
  private static final Map xnToXsd;
  
  static {
    xsdToSap = new HashMap(7);
    xsdToSap.put(XSD_BOOLEAN, PropertyType.BOOLEAN);
    xsdToSap.put(XSD_DATETIME, PropertyType.DATE);
    xsdToSap.put(XSD_INT, PropertyType.INTEGER);
    xsdToSap.put(XSD_LONG, PropertyType.LONG);
    xsdToSap.put(XSD_STRING, PropertyType.STRING);

    xsdToSap.put(SAP_EXPANDED_NAME, PropertyType.STRING);
    xsdToSap.put(SAP_RID, PropertyType.STRING);
    xsdToSap.put(SAP_MAPPED_RID, PropertyType.STRING);
    xsdToSap.put(SAP_MAPPED_RID_LIST, PropertyType.STRING);

    xsdToSap.put(SAP_XMLMARKUP, PropertyType.XML);
    
    xnToXsd = new HashMap();
    xnToXsd.put(XN_RID, SAP_RID);
    xnToXsd.put(XN_EXPANDED_NAME, SAP_EXPANDED_NAME);

    Properties tmp = new Properties();
    tmp.setProperty(XN_TYPE, XN_EXPANDED_NAME);
    ATTRS_TYPE_EXPANDED_NAME = ImmutableProperties.getInstance(tmp);

    tmp = new Properties();
    tmp.setProperty(XN_TYPE, XN_RID);
    ATTRS_TYPE_RID = ImmutableProperties.getInstance(tmp);
  }

  public static IName getXsdTypeInfo(String expandedName) {
    IName name = (IName)XsdDataTypes.xnToXsd.get(expandedName);
    if (name == null && expandedName != null) {
      try {  
        name = new Name(expandedName);  
      }
      catch (Exception ex) {
        // $JL-EXC$ 
        // seems not to be a valid expanded name
      }
    }
    return name;
  }
  
  /**
   * Get XSD type information for the element.
   * @param e element node
   * @return datatype name and namespace
   */
  public static QualifiedName getXsdTypeInfo(Element e) {
    // m_logger.error (DomSerializer.toString (e, false));

    Attr attr = e.getAttributeNodeNS(Consts.NS_XSI, "type");
    return attr == null ? null : getXsdTypeInfoFromValue(e, attr.getNodeValue());
  }

  public static QualifiedName getXsdTypeInfoFromValue(Element baseElem, String qname) {

    String localName = QualifiedName.getLocalName(qname);
    String prefix = QualifiedName.getPrefix(qname);
    String ns = null;

    // if we have a prefix, try to resolve it
    if (prefix != null) {
      Node search = baseElem;
      String attrName = "xmlns:" + prefix;
      while (search != null && search.getNodeType() == Node.ELEMENT_NODE) {
        Attr a = ((Element)search).getAttributeNode(attrName);
        if (a != null) {
          ns = a.getNodeValue();
          break;
        }
        search = search.getParentNode();
      }

      // as a fallback, try the documentElement of the owner document
      if (ns == null) {
        Element de = baseElem.getOwnerDocument().getDocumentElement();
        if (de != null) {
          Attr a = de.getAttributeNode(attrName);
          if (a != null) {
            ns = a.getNodeValue();
          }
        }
      }
    }

    return new QualifiedName(ns, localName, prefix);
  }

  public static boolean parseBoolean(String val) throws java.text.ParseException {
    if ("0".equals(val) || "false".equals(val)) { 
      return false;
    }
    else if ("1".equals(val) || "true".equals(val)) { 
      return true;
    }
    else {
      throw new java.text.ParseException(val, 0);
    }
  }

  public static QualifiedName mapSapToXsdType(PropertyType pt) {
    if (pt.equals(PropertyType.BOOLEAN)) {
      return XSD_BOOLEAN;
    }
    else if (pt.equals(PropertyType.DATE)) {
      return XSD_DATETIME;
    }
    else if (pt.equals(PropertyType.INTEGER)) {
      return XSD_INT;
    }
    else if (pt.equals(PropertyType.LONG)) {
      return XSD_LONG;
    }
    else if (pt.equals(PropertyType.XML)) {
      return SAP_XMLMARKUP;
    }
    else {
      return XSD_STRING;
    }
  }

  public static QualifiedName mapInternalTypeToXsdType(IProperty prop) {
    try {
      String val = prop.getAttribute(XsdDataTypes.XN_TYPE);
      if (XsdDataTypes.XN_EXPANDED_NAME.equals(val)) {
        return XsdDataTypes.SAP_EXPANDED_NAME;      
      }
      else if (XsdDataTypes.XN_RID.equals(val)) {
        return XsdDataTypes.SAP_MAPPED_RID;      
      }
      else {
        return null;
      }
    }
    catch (ResourceException ex) {
      log.warningT("getting Property Attributes: " + LoggingFormatter.extractCallstack(ex));
      return null;
    }
  }


  public static PropertyType mapXsdToSapType(IName xsdType) {
    PropertyType pt = (PropertyType)xsdToSap.get(xsdType);
    return pt != null ? pt : PropertyType.STRING;
  }

  public static PropertyType mapXsdToSapTypeOrNull(IName xsdType) {
    return (PropertyType)xsdToSap.get(xsdType);
  }
}