package com.sapportals.wcm.protocol.webdav.cache;

import java.util.*;

/**
 * DiscardingCache is a {@link IGenerationCache} for all kinds of Java objects.
 * <p>
 * DiscardingCache honors the {@link IDiscardable} interface. Objects placed
 * in the cache will be checked for implementing IDiscardable before
 * removing them from the cache.
 * <p>
 * DiscardingCache is multithread-safe.
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author stefan.eissing@greenbytes.de
 * @version $Revision: 1.2 $
 */
public class DiscardingCache implements IGenerationCache, IDiscardable {

//  private static final com.sapportals.wcm.util.log.Category log =
//    com.sapportals.wcm.util.log.Category.getInstance(DiscardingCache.class.getName());

  private static Timer m_timer;

  private final synchronized static Timer getTimer() {
    if (m_timer == null) {
      m_timer = new Timer(true);
    }
    return m_timer;
  }

  private long m_generationTime;
  private Map m_generation1;
  private Map m_generation2;
  private boolean m_registered;
  private TimerTask m_client;

  /** Construct a new Cache with the given time between generations
   * @param generationTime time in milliseconds
   */
  public DiscardingCache(long generationTime) {
    this(generationTime, new HashMap(), new HashMap());
  }

  /** Construct a new Cache with the given time between generations
   * and the given Maps for the two generations. Both maps should
   * be of the same type and size (if applicable).
   *
   * @param generationTime time in milliseconds
   * @param map1 map for a generation
   * @param map2 map for a generation
   */
  public DiscardingCache(long generationTime, Map map1, Map map2) {
    m_generationTime = generationTime;
    m_generation1 = map1;
    m_generation2 = map2;
    setGenerationTime(m_generationTime);
  }

  /** Remove all entries from the cache
   */
  public synchronized void clear() {
    clear(m_generation2);
    clear(m_generation1);
  }

  /** Clear the given map. If any object in the map implements
   * IDiscardable, call its <code>discard()</code> method.
   */
  private synchronized void clear(Map map) {
    Iterator iter = map.entrySet().iterator();
    while (iter.hasNext()) {
      Object value = ((java.util.Map.Entry)iter.next()).getValue();
      if (value instanceof IDiscardable) {
        ((IDiscardable)value).discard();
      }
    }
    map.clear();
  }

  /** This cache itself has been discarded. Clear the cache.
   */
  public void discard() {
    if (m_client != null) {
      m_client.cancel();
    }
    clear();
  }

  /** Retrieve the value for the given key
   * @return value for key or <code>null</code>
   */
  public synchronized Object get(Object key) {
    Object value = m_generation1.get(key);
    if (value == null) {
      value = m_generation2.get(key);
      if (value != null) {
        // This key/value pair needs rejuvenation. Simulate
        // a put() to add it to the current generation and
        // remove if from the old generation. This ensures that
        // the old generation only holds old keys.
        //
        put(key, value);
        m_generation2.remove(key);
      }
    }
    return value;
  }

  /**
   * @return the generation time for this cache
   */
  public synchronized long getGenerationTime() {
    return m_generationTime;
  }

  /** Put the value in the cache under the given key
   * @param key for lookup of value
   * @param value to place in cache
   */
  public synchronized void put(Object key, Object value) {
    m_generation1.put(key, value);
    register();
  }

  /** Remove the given key and its value from the cache.
   */
  public synchronized void remove(Object key) {
    Object value = m_generation1.get(key);
    if (value == null) {
      value = m_generation2.get(key);
    }
    if (value != null) {
      if (value instanceof IDiscardable)
        ((IDiscardable)value).discard();
      m_generation1.remove(key);
      m_generation2.remove(key);
    }
  }

  /** Adjust the generation time of this cache.
   * @param generationTime to set
   */
  public void setGenerationTime(long generationTime) {
    m_generationTime = generationTime;
    if (m_registered) {
      m_client.cancel();
      register();
    }
  }

  // --------------- private /protected ----------------------------------

  private void register() {
    if (!m_registered) {
      m_registered = true;
      m_client = new TimerTask() {
        public boolean cancel() {
          m_registered = false;
          return super.cancel();
        }

        public void run() {
          if (!stepGeneration()) {
            this.cancel();
          }
        }
      };

      getTimer().schedule(m_client, m_generationTime, m_generationTime);
    }
  }

  /** Time between two generations has been elapsed.
   * @return if this task shall continue running
   */
  private synchronized boolean stepGeneration() {
    if (m_generation1.size() > 0 || m_generation2.size() > 0) {
      clear(m_generation2);

      Map tmp = m_generation2;
      m_generation2 = m_generation1;
      m_generation1 = tmp;
    }

    return m_generation2.size() != 0;
  }

}



