package com.sapportals.wcm.protocol.webdav.dompath;

import java.util.*;

import org.w3c.dom.Attr;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import com.sapportals.wcm.protocol.webdav.Name;

/**
 * Implements <a href="http://www.w3.org/TR/xpath">XPath</a> on
 * any level-2 <a href="http://www.w3.org/TR/DOM-Level-2-Core/">DOM</a>.
 * <p>
 * For a complete description of the XPath language see
 * <a href="http://www.w3.org/TR/xpath">XPath</a>. DomPath uses
 * this language to select nodes in a document.
 * <p>
 * <b>Usage</b><p>
 * You create a DomPath object with<pre>
 * DomPath dp = new DomPath("XPath expression", NamespaceMap);</pre>
 * and use it either to retrieve all matching Nodes:<pre>
 * List resultNodes = dp.match(document);</pre>
 * or to find the first matching Node:<pre>
 * Node result = dp.matchFirst(document);</pre>
 * Some examples:<pre>
 * DomPath dp = new DomPath(".//chapter", NamespaceMap);</pre>
 * will match all elements named "chapter"<pre>
 * DomPath dp = new DomPath(".//chapter/@title", NamespaceMap);</pre>
 * will match all title attributes of all chapter elements<pre>
 * DomPath dp = new DomPath(".//chapter[@title = 'Introduction']", NamespaceMap);</pre>
 * will match all chapter elements which have a title attribute with value "Introduction".
 * <p>
 * The NamespaceMap maps prefixes to namespace URIs. If you want to find
 * all elements in a certain namespace, say "http://www.sap.com/test", you
 * would associate a prefix, say "st", with the URI in the map:<pre>
 * Map nsmap = new HashMap();
 * nsmap.put("st", "http://www.sap.com/test");
 * DomPath dp = new DomPath(".//st:*", nsmap);</pre>
 * would find all elements in the namespace "http://www.sap.com/test"
 * in any supplied document. Note that the prefix you choose for the DomPath can
 * be completely different from the prefix used in the matched document. The
 * matching is done on the URI, not on the prefix.
 * <p>
 * There is a predefined, unmodifiable Map as constant in DomPath which
 * already has some associations between prefixes and common namespace URIs.
 * <p>
 * <b>Multithreading</b><p>
 * DomPath objects are multithread-safe. So, if you have a fixed Xpath
 * expression which you need to evaluate over and over again, make the
 * DomPath object a static member of your class.
 * <p>
 * <b>Extension Functions</b><p>
 * not supported yet.
 *
 * <p><b>Limitations</b>
 * The following lists all known deviations from XPath 1.0 standard:
 * <ul>
 * <li>The axis "following", "following-sibling", "preceding",
 *    "preceding-sibling" and "namespace" are not supported.</li>
 * <li>The node test "processing-instruction()" does not take a parameter.</li>
 * <li>The function "id(...)" is not supported.</li>
 * </ul></p>
 *
 * <p>
 * Copyright (c) SAP AG 2001-2003
 * @author stefan.eissing@greenbytes.de
 * @version $Id: DomPath.java,v 1.1.1.1 2004/02/18 13:00:22 sync Exp $
 */
public class DomPath {

  private static final com.sap.tc.logging.Location log = com.sap.tc.logging.Location.getLocation(DomPath.class.getName());

  /**
   * Default bindings of prefixes to namespace URIs for some common
   * namespaces<pre>
   * xml   -> http://www.w3.org/XML/1998/namespace
   * xmlns -> http://www.w3.org/2000/xmlns/
   * xs    -> http://www.w3.org/2001/XMLSchema
   * xsi   -> http://www.w3.org/2001/XMLSchema-instance
   * D     -> DAV:
   * </pre>
   */
  public static final Map DEFAULT_NAMESPACE_BINDINGS;

  static {
    Map map = new HashMap(13);
    map.put("xml", "http://www.w3.org/XML/1998/namespace");
    map.put("xmlns", "http://www.w3.org/2000/xmlns/");
    map.put("xs", "http://www.w3.org/2001/XMLSchema");
    map.put("xsi", "http://www.w3.org/2001/XMLSchema-instance");
    map.put("D", "DAV:");
    DEFAULT_NAMESPACE_BINDINGS = Collections.unmodifiableMap(map);
  }

  private final IExpression m_expr;
  private final IStep m_step;
  private final boolean m_containsPredicate;

  /**
   * Create a new DomPath with given XPath expression and
   * namespace mapping. This contructor throws {@link
   * java.lang.IllegalArgumentException} silently on
   * not-wellformed XPath expressions.
   *
   * @param path the XPath expression to match with
   * @param namespaces mapping of prefixes to namespace URIs
   */
  public DomPath(String path, Map namespaces) {

    IExpression expr = null;
    try {
      DomPathParser parser = new DomPathParser();
      expr = parser.parse(path, namespaces);
    }
    catch (java.text.ParseException ex) {
      log.errorT("unable to parse given path("+path+")" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(ex));
      throw new java.lang.IllegalArgumentException(ex.getMessage());
    }

    m_expr = expr;
    if (m_expr instanceof LocationExpression) {
      m_step = ((LocationExpression)m_expr).getStep();
      m_containsPredicate = m_step.containsPredicate();
    }
    else {
      m_step = null;
      m_containsPredicate = false;
    }
  }

  /**
   * Match the expression in the given document, start from given element.
   * If the path is an absolute path (starting with '/'), the search begins
   * nevertheless at the document root.
   * @param start the Node from which to start matching
   * @return List of matched Node objects, will be empty if none found
   */
  public List match(Node start) {
    Node[] result = matchNodes(start);
    return toList(result);
  }

  /**
   * Like {@link #match(org.w3c.dom.Node)}, but checks for required size of result list
   * before returning it.
   * @param start node from which to start matching
   * @param minOccur minimal required size of result list
   * @param maxOccur maximum allowed size of result list
   * @return List of matched Node objects, will be empty if none found
   * @throws DomPathAssertionFailedException when number of results outside range
   */
  public List matchOrDie(Node start, int minOccur, int maxOccur, String message)
    throws DomPathAssertionFailedException {
    Node[] result = matchNodes(start);

    if (result.length < minOccur || result.length > maxOccur) {
      throw new DomPathAssertionFailedException (message);
    }

    return toList(result);
  }

  /**
   * Match the expression in the given document, start from given element.
   * If the path is an absolute path (starting with '/'), the search begins
   * nevertheless at the document root.
   * @param start the Node from which to start matching
   * @return array of Node objects
   */
  public Node[] matchNodes(Node start) {
    Node[] result = null;
    if (m_step != null) {
      Node[] nodes = new Node[] { start };
      result = m_step.match(nodes, PathContext.EMPTY_SET);
    }
    else if (m_expr != null) {
      IContext eval = m_expr.evaluate(new PathContext(start));
      if (eval.getType() == IContext.TYPE_NODESET) {
        result = eval.getNodes();
      }
    }

    return result;
  }


  /**
   * Find the first Node which matches the expression.
   * @param start the Node to start the match from
   * @return the first found Node object or <code>null</code>
   */
  public Node matchFirst(Node start) {
    if (m_step != null) {
      if (m_containsPredicate) {
        Node[] nodes = new Node[] { start };
        Node[] result = m_step.match(nodes, PathContext.EMPTY_SET);
        return (result.length > 0)? result[0] : null;
      }
      else {
        return m_step.matchFirst(start);
      }
    }
    else if (m_expr != null) {
      IContext result = m_expr.evaluate(new PathContext(start));
      Node[] nodes = result.getNodes();
      if (nodes != null && nodes.length > 0) {
        return nodes[0];
      }
    }

    return null;
  }

  /**
   * Matches the XPath expression against the start Node and returns
   * a mapping of the resulting nodes against their
   * {@link com.sapportals.wcm.protocol.webdav.Name}s. If the Names are not unique
   * an expception is thrown.
   * @param start Node to evalualte expression against
   * @return map of matched nodes to their Name
   * @throws DomPathException if named mapping is not unique
   */
  public Map matchUniqueNames(Node start) throws DomPathException {
    Node[] nodes = matchNodes(start);

    if (nodes != null && nodes.length > 0) {
      Map result = new HashMap(nodes.length*2);
      for (int i = 0; i < nodes.length; ++i) {
        Node node = nodes[i];
        Name name = new Name(node);
        if (result.put(name, node) != null) {
          throw new DomPathException("name "+name+" not unique");
        }
      }
      return result;
    }

    return Collections.EMPTY_MAP;
  }

  /**
   * Matches the XPath expression against the start Node and returns
   * a mapping of the resulting nodes against their
   * {@link com.sapportals.wcm.protocol.webdav.Name}s. If the Names are not unique
   * the first occurance is returned in the Map.
   * @param start Node to evalualte expression against
   * @return map of matched nodes to their Name
   */
  public Map matchFirstNames(Node start) {
    Node[] nodes = matchNodes(start);

    if (nodes != null && nodes.length > 0) {
      Map result = new HashMap(nodes.length*2);
      for (int i = 0; i < nodes.length; ++i) {
        Node node = nodes[i];
        Name name = new Name(node);
        Object o = result.put(name, node);
        if (o != null) {
          result.put(name, o);
        }
      }
      return result;
    }

    return Collections.EMPTY_MAP;
  }

  /**
   * Delete all matched Nodes, start from given element.
   * @param start the Node from which to start matching
   * @return List of deleted Node objects, will be empty if none found
   */
  public List delete(Node start) {
    Node[] nodes = matchNodes(start);
    for (int i = 0; i < nodes.length; ++i) {
      Node node = nodes[i];

      switch (node.getNodeType()) {
      case Node.DOCUMENT_NODE:
        // no can do
        break;

      case Node.ATTRIBUTE_NODE:
        Attr attr = (Attr)node;
        Element owner = attr.getOwnerElement();
        if (owner != null) {
          owner.removeAttributeNode(attr);
        }
        break;

      default:
        Node parent = node.getParentNode();
        if (parent != null) {
          parent.removeChild(node);
        }
        break;
      }
    }

    return toList(nodes);
  }

  /**
   * Delete all matched Nodes, start from given element.
   * @param start the Node from which to start matching
   * @return Array of deleted Node objects, will be empty if none found
   */
  public Node[] deleteNodes(Node start) {
    Node[] nodes = matchNodes(start);
    for (int i = 0; i < nodes.length; ++i) {
      Node node = nodes[i];

      switch (node.getNodeType()) {
      case Node.DOCUMENT_NODE:
        // no can do
        break;

      case Node.ATTRIBUTE_NODE:
        Attr attr = (Attr)node;
        Element owner = attr.getOwnerElement();
        if (owner != null) {
          owner.removeAttributeNode(attr);
        }
        break;

      default:
        Node parent = node.getParentNode();
        if (parent != null) {
          parent.removeChild(node);
        }
        break;
      }
    }

    return nodes;
  }

  /**
   * Return if node matches.
   * @param start the Node to start the match from
   * @return if node matches
   */
  public boolean matches(Node start) {
    return matchFirst(start) != null;
  }

  /**
   * Converts the DomPath to its String representation, merely useful
   * for debuggging.
   * @return String representation of parsed XPath expression
   */
  public String toString() {
    if (m_step == null)
      return super.toString();
    else
      return m_step.toString();
  }

  private final List toList(Node[] nodes) {
    if (nodes == null || nodes.length == 0) {
      return Collections.EMPTY_LIST;
    }
    else {
      return new FixedNodeList(nodes);
    }
  }
}
