package com.sapportals.wcm.protocol.webdav.dompath;

import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.w3c.dom.Node;

/**
 * <p>
 * Copyright (c) SAP AG 2001-2003
 * @author stefan.eissing@greenbytes.de
 * @version $Id: DomPathParser.java,v 1.1.1.1 2004/02/18 13:00:22 sync Exp $
 */
class DomPathParser {

  private static final com.sap.tc.logging.Location log = com.sap.tc.logging.Location.getLocation(DomPathParser.class.getName());

  private Map m_namespaces;
  private Lexer m_lex;

  public DomPathParser() {
  }

  IExpression parse(String path, Map namespaces) throws ParseException {
    m_namespaces = namespaces;
    m_lex = new Lexer(path);

    IExpression expression = expr();

    if (log.beDebug()) {
      log.debugT("path("+path+") is "+expression.toString());
    }

    int token = m_lex.getNext();
    if (token != Lexer.EOF) {
      log.errorT(expression.toString());
      unexpected(token);
    }

    return expression;
  }

  private IStep locationStep() throws ParseException {
    IStep step = null;
    int token = m_lex.getNext();
    m_lex.unget();

    switch (token) {
    case Lexer.OPERATOR:
      String name = m_lex.getString();
      if (name.charAt(0) == '/') {
        step = absoluteLocation();
        break;
      }
      else
        m_lex.unget();
      // fall through
    default:
      step = relativeLocation();
      break;
    }
    return step;
  }

  private IStep absoluteLocation() throws ParseException {
    int token = m_lex.getNext();
    String name = m_lex.getString();
    if (token != Lexer.OPERATOR
        || (!name.equals("/") && !name.equals("//"))) {
      unexpected(token);
    }

    IStep root = new RootStep(null);
    IStep current = root;
    if (name.equals("//")) {
      current = new SelfOrDescendantStep(null);
      root.setNext(current);
    }

    token = m_lex.getNext();
    m_lex.unget();

    if (token != Lexer.EOF) {
      IStep next = relativeLocation();
      current.setNext(next);
    }

    return root;
  }

  private IStep relativeLocation() throws ParseException {

    IStep first = step();

    int token = m_lex.getNext();
    switch (token) {
    case Lexer.OPERATOR:
      String name = m_lex.getString();
      if (name.equals("/")) {
        IStep next = relativeLocation();
        first.setNext(next);
      }
      else if (name.equals("//")) {
        IStep current = null;
        if (first instanceof SelfStep) {
          first = new SelfOrDescendantStep(null, true);
          current = first;
        }
        else {
          current = new SelfOrDescendantStep(null);
          first.setNext(current);
        }

        token = m_lex.getNext();
        m_lex.unget();

        if (token != Lexer.EOF) {
          IStep next = relativeLocation();
          current.setNext(next);
        }
      }
      else
        m_lex.unget();
      break;
    default:
      m_lex.unget();
      break;
    }

    return first;
  }

  /**
   *  Step ::= AbrbreviatedStep | AxisSpecifier NodeTest Predicate*
   */
  private IStep step() throws ParseException {
    IStep s = null;

    int token = m_lex.getNext();
    if (token == Lexer.STRING) {
      String name = m_lex.getString();

      if (".".equals(name)) {                // Abbreviated Step?
        s = new SelfStep(null);
      }
      else if ("..".equals(name)) {          // Abbreviated Step?
        s = new ParentStep(null);
      }
    }

    if (s == null) {                         // AxisSpecifier NodeTest Predicate*
      m_lex.unget();
      String axis = axisSpecifier();

      INodeTest nodeTest = nodeTest();

      if (axis.equals("child")) {
        s = new ChildStep(nodeTest);
      }
      else if (axis.equals("attribute")) {
        s = new AttributeStep(nodeTest);
      }
      else if (axis.equals("parent")) {
        s = new ParentStep(nodeTest);
      }
      else if (axis.equals("self")) {
        s = new SelfStep(nodeTest);
      }
      else if (axis.equals("descendant")) {
        s = new SelfOrDescendantStep(nodeTest, false);
      }
      else if (axis.equals("descendant-or-self")) {
        s = new SelfOrDescendantStep(nodeTest, true);
      }
      else if (axis.equals("ancestor")) {
        s = new SelfOrAncestorStep(nodeTest, false);
      }
      else if (axis.equals("ancestor-or-self")) {
        s = new SelfOrAncestorStep(nodeTest, true);
      }
      else {
        throw new ParseException("Axisspecifier not supported: "+axis, m_lex.getIndex());
      }
    }

    IExpression expression = predicateList();
    if (expression != null) {
      s.setPredicate(expression);
    }

    return s;
  }

  /**
   * AxisSpecifier ::= AxisName '::' | AbbreviatedAxisSpecifier
   * AbbreviatedAxisSpecifier ::= '@'?
   */
  private String axisSpecifier() throws ParseException {
    int token = m_lex.getNext();

    switch (token) {
    case Lexer.AT:
      return "attribute";

    case Lexer.AXIS_NAME:
      String axis = m_lex.getString();
      token = m_lex.getNext();
      if (token != Lexer.DDCOLON) {
        unexpected(token);
      }
      return axis;

    default:
      m_lex.unget();
      return "child";
    }
  }

  /**
   * NodeTest ::= NameTest | NodeType '('')'
   *
   */
  private INodeTest nodeTest() throws ParseException {
    int token = m_lex.getNext();
    switch (token) {

    case Lexer.NODE_TYPE:
      String testname = m_lex.getString();

      token = m_lex.getNext();
      if (token != Lexer.OBRACE) {
        throw new ParseException("expected '()', found "+token, m_lex.getIndex());
      }
      token = m_lex.getNext();
      if (token != Lexer.CBRACE) {
        throw new ParseException("expected '()', found "+token, m_lex.getIndex());
      }

      INodeTest nodeTest = null;
      if (testname.equals("comment")) {
        nodeTest = new NodeTest(Node.COMMENT_NODE);
      }
      else if (testname.equals("text")) {
        nodeTest = new NodeTest(Node.TEXT_NODE);
      }
      else if (testname.equals("processing-instruction")) {
        nodeTest = new NodeTest(Node.PROCESSING_INSTRUCTION_NODE);
      }
      else if (testname.equals("node")) {
        nodeTest = new NodeTest();
      }

      if (nodeTest == null) {
        throw new ParseException("node test for "+testname+" not implemented", m_lex.getIndex());
      }
      return nodeTest;

    case Lexer.NAME_TEST:
      String ns = null;
      String name = m_lex.getString();
      int index = name.indexOf(':');
      if (index >= 0) {
        String prefix = name.substring(0, index);
        ns = (String)m_namespaces.get(prefix);
        if (ns == null) {
          throw new ParseException("undefined namespace prefix: "+prefix, m_lex.getIndex());
        }
        name = name.substring(index+1);
      }
      return new NameTest(ns, name);

    case Lexer.STRING:
      return new NameTest(null, m_lex.getString());

    default:
      unexpected(token);
      return null;
    }
  }

  /**
   * PredicateList ::= Predicate*
   */
  private IExpression predicateList() throws ParseException {
    PredicateExpression list = null;
    PredicateExpression current = null;

    while (m_lex.getNext() == Lexer.OANGULAR) {
      m_lex.unget();
      PredicateExpression expr = predicate();
      if (list == null) {
        list = expr;
      }
      else {
        current.setNext(expr);
      }
      current = expr;
    }
    m_lex.unget();

    return list;
  }

  /**
   * Predicate ::= '[' Expr ']'
   */
  private PredicateExpression predicate() throws ParseException {

    int token = m_lex.getNext();
    if (token != Lexer.OANGULAR) {
      unexpected(token);
    }

    PredicateExpression expression = new PredicateExpression(expr());

    token = m_lex.getNext();
    if (token != Lexer.CANGULAR) {
      log.errorT(expression.toString());
      unexpected(token);
    }

    return expression;
  }

  /**
   * Expr           ::= OrExpr
   */
  private IExpression expr() throws ParseException {
    IExpression expression = orExpr();
    return expression;
  }

  /**
   * OrExpr         ::= AndExpr | OrExpr 'or' AndExpr
   */
  private IExpression orExpr() throws ParseException {
    IExpression expression = andExpr();
    int token = m_lex.getNext();
    if (token == Lexer.OPERATOR && m_lex.getString().equals("or")) {
      return new OrExpression(expression, orExpr());
    }
    else {
      m_lex.unget();
      return expression;
    }
  }

  /**
   * AndExpr        ::= EqualityExpr | AndExpr 'and' EqualityExpr
   */
  private IExpression andExpr() throws ParseException {
    IExpression expression = equalityExpr();
    int token = m_lex.getNext();
    if (token == Lexer.OPERATOR && m_lex.getString().equals("and")) {
      return new AndExpression(expression, andExpr());
    }
    else {
      m_lex.unget();
      return expression;
    }
  }

  /**
   * EqualityExpr   ::= RelationalExpr | EqualityExpr ['='|'!='] RelationalExpr
   */
  private IExpression equalityExpr() throws ParseException {
    IExpression expression = relationalExpr();
    int token = m_lex.getNext();
    String name = m_lex.getString();
    if (token == Lexer.OPERATOR
        && (name.equals("=") || name.equals("!="))) {
      return new EqualExpression(expression, name, equalityExpr());
    }
    else {
      m_lex.unget();
      return expression;
    }
  }

  /**
   * RelationalExpr ::= AdditiveExpr | RelationalExpr ['<'|'>'|'<='|'>='] AdditiveExpr
   */
  private IExpression relationalExpr() throws ParseException {
    IExpression expression = additiveExpr();
    int token = m_lex.getNext();
    String name = m_lex.getString();
    if (token == Lexer.OPERATOR
        && (name.equals("<") || name.equals(">") || name.equals("<=") || name.equals(">="))) {
      return new RelationalExpression(expression, name, relationalExpr());
    }
    else {
      m_lex.unget();
      return expression;
    }
  }

  /**
   * AdditiveExpr   ::= MultiExpr | AdditiveExpr ['+'|'-'] MultiExpr
   */
  private IExpression additiveExpr() throws ParseException {
    IExpression expression = multiExpr();
    int token = m_lex.getNext();
    String name = m_lex.getString();
    if (token == Lexer.OPERATOR && (name.equals("+") || name.equals("-"))) {
      return new AdditiveExpression(expression, name, additiveExpr());
    }
    else {
      m_lex.unget();
      return expression;
    }
  }

  /**
   * MultiExpr      ::= UnaryExpr | MultiExpr ['*'|'div'|'mod'] UnaryExpr
   */
  private IExpression multiExpr() throws ParseException {
    IExpression expression = unaryExpr();
    int token = m_lex.getNext();
    String name = m_lex.getString();
    if (token == Lexer.OPERATOR && (name.equals("*") || name.equals("div") || name.equals("mod"))) {
      return new MultiExpression(expression, name, multiExpr());
    }
    else {
      m_lex.unget();
      return expression;
    }
  }

  /**
   * UnaryExpr      ::= UnionExpr | '-' UnaryExpr
   */
  private IExpression unaryExpr() throws ParseException {
    int token = m_lex.getNext();
    String name = m_lex.getString();
    if (token == Lexer.OPERATOR && name.equals("-")) {
      return new NegateExpression(unaryExpr());
    }
    else {
      m_lex.unget();
      return unionExpr();
    }
  }

  /**
   * UnionExpr      ::= PathExpr | UnionExpr '|' PathExpr
   */
  private IExpression unionExpr() throws ParseException {
    IExpression expression = pathExpr();
    int token = m_lex.getNext();
    String name = m_lex.getString();
    if (token == Lexer.OPERATOR && name.equals("|")) {
      return new UnionExpression(expression, unionExpr());
    }
    else {
      m_lex.unget();
      return expression;
    }
  }

  /**
   * PathExpr       ::= LocationPath | FilterExpr [ ['/'|'//'] RelativeLocationPath ]
   * VariableRef    ::= '$' QName
   *
   * Terminals that can start which Expr:
   * '('      -->  PrimaryExpr           --> FilterExpr
   * '$'      -->  VariableReference     --> FilterExpr
   * NUMBER   -->  PrimaryExpr           --> FilterExpr
   * LITERAL  -->  PrimaryExpr           --> FilterExpr
   * FUNCTION_NAME-->  FunctionCall      --> FilterExpr
   * '/'      -->  AbsoluteLocationPath  --> LocationPath
   * '//'      -->  AbsoluteLocationPath --> LocationPath
   * AXIS_NAME-->  RelativeLocationPath  --> LocationPath
   * '.'      -->  RelativeLocationPath  --> LocationPath
   * '..'     -->  RelativeLocationPath  --> LocationPath
   * '@'      -->  RelativeLocationPath  --> LocationPath
   * NAME_TEST-->  RelativeLocationPath  --> LocationPath
   * NODE_TYPE-->  RelativeLocationPath  --> LocationPath
   *
   */
  private IExpression pathExpr() throws ParseException {
    int token = m_lex.getNext();
    m_lex.unget();
    switch (token) {
    case Lexer.OBRACE:
    case Lexer.VARIABLE:
    case Lexer.NUMBER:
    case Lexer.LITERAL:
    case Lexer.FUNCTION_NAME:
      return filterExpr();
    default:
      return locationPath();
    }
  }

  private IExpression locationPath() throws ParseException {
    return new LocationExpression(locationStep());
  }

  /**
   * FilterExpr     ::= PrimaryExpr | FilterExpr Predicate
   */
  private IExpression filterExpr() throws ParseException {
    IExpression expression = primaryExpr();
    int token = m_lex.getNext();
    m_lex.unget();
    if (token == Lexer.OANGULAR) {
      return new FilterExpression(expression, predicateList());
    }
    return expression;
  }

  /**
   * PrimaryExpr    ::= VariableRef | '(' Expr ')' | Literal | Number | FunctionCall
   */
  private IExpression primaryExpr() throws ParseException {
    int token = m_lex.getNext();
    String name = m_lex.getString();
    switch (token) {
    case Lexer.VARIABLE:
      throw new ParseException("variable references not supported", m_lex.getIndex());

    case Lexer.OBRACE:
      IExpression expression = expr();
      token = m_lex.getNext();
      if (token != Lexer.CBRACE) {
        unexpected(token);
      }
      return expression;

    case Lexer.LITERAL:
    case Lexer.NUMBER:
      return new ConstantExpression(token, name);

    case Lexer.FUNCTION_NAME:
      return functionCall(name);

    default:
      unexpected(token);
      return null;
    }
  }

  /**
   * FunctionCall   ::= FunctionName '(' (Argument (',' Argument)*)? ')'
   * Argument       ::= Expr
   */
  private IExpression functionCall(String name) throws ParseException  {
    int token = m_lex.getNext();
    if (token != Lexer.OBRACE) {
      unexpected(token);
    }

    token = m_lex.getNext();
    boolean firstArg = true;
    List list = null;
    boolean finished = (token == Lexer.CBRACE);
    while (!finished) {
      if (firstArg) {
        firstArg = false;
        m_lex.unget();
        list = new ArrayList(10);
      }
      list.add(expr());

      token = m_lex.getNext();
      finished = (token != Lexer.COMMA);
    }

    if (token != Lexer.CBRACE) {
      unexpected(token);
    }

    return new FunctionExpression(name, list);
  }

  private void unexpected(int token) throws ParseException {
    if (log.beDebug()) {
    }
    String buffer = m_lex.getString();
    int index = m_lex.getIndex();
    throw new ParseException("unexpected token: "+token
                             +" ("+buffer+") at "+index, index);

  }
}
