package com.sapportals.wcm.protocol.webdav.dompath;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.w3c.dom.Attr;
import org.w3c.dom.Element;
import org.w3c.dom.Node;


/**
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author stefan.eissing@greenbytes.de
 * @version $Id: FunctionExpression.java,v 1.2 2004/03/03 17:44:19 jre Exp $
 */
class FunctionExpression implements IExpression {

  private static final int F_LAST = 1;
  private static final int F_POSITION = 2;
  private static final int F_COUNT = 3;
  private static final int F_ID = 4;
  private static final int F_LOCAL_NAME = 5;
  private static final int F_NAMESPACE_URI = 6;
  private static final int F_NAME = 7;
  private static final int F_STRING = 8;
  private static final int F_CONCAT = 9;
  private static final int F_STARTS_WITH = 10;
  private static final int F_CONTAINS = 11;
  private static final int F_SUBSTRING_BEFORE = 12;
  private static final int F_SUBSTRING_AFTER = 13;
  private static final int F_SUBSTRING = 14;
  private static final int F_STRING_LENGTH = 15;
  private static final int F_NORMALIZE_SPACE = 16;
  private static final int F_TRANSLATE = 17;
  private static final int F_BOOLEAN = 18;
  private static final int F_NOT = 19;
  private static final int F_TRUE = 20;
  private static final int F_FALSE = 21;
  private static final int F_LANG = 22;
  private static final int F_NUMBER = 23;
  private static final int F_SUM = 24;
  private static final int F_FLOOR = 25;
  private static final int F_CEILING = 26;
  private static final int F_ROUND = 27;

  private static final Map nameCode;
  private static final int[] argCount = new int[] {
    0,
    0, // F_LAST
    0, // F_POSITION
    1, // F_COUNT
    1, // F_ID
    -1,// F_LOCAL_NAME
    -1,// F_NAMESPACE_URI
    -1,// F_NAME
    -1,// F_STRING
    -17,// F_CONCAT
    2, // F_STARTS_WITH
    2, // F_CONTAINS
    2, // F_SUBSTRING_BEFORE
    2, // F_SUBSTRING_AFTER
    -3,// F_SUBSTRING
    -1,// F_STRING_LENGTH
    -1,// F_NORMALIZE_SPACE
    3, // F_TRANSLATE
    1, // F_BOOLEAN
    1, // F_NOT
    0, // F_TRUE
    0, // F_FALSE
    1, // F_LANG
    -1,// F_NUMBER
    1, // F_SUM
    1, // F_FLOOR
    1, // F_CEILING
    1, // F_ROUND
  };

  static {
    nameCode = new HashMap(31);
    nameCode.put("last", new Integer(F_LAST));
    nameCode.put("position", new Integer(F_POSITION));
    nameCode.put("count", new Integer(F_COUNT));
    nameCode.put("id", new Integer(F_ID));
    nameCode.put("local-name", new Integer(F_LOCAL_NAME));
    nameCode.put("namespace-uri", new Integer(F_NAMESPACE_URI));
    nameCode.put("name", new Integer(F_NAME));
    nameCode.put("string", new Integer(F_STRING));
    nameCode.put("concat", new Integer(F_CONCAT));
    nameCode.put("starts-with", new Integer(F_STARTS_WITH));
    nameCode.put("contains", new Integer(F_CONTAINS));
    nameCode.put("substring-before", new Integer(F_SUBSTRING_BEFORE));
    nameCode.put("substring-after", new Integer(F_SUBSTRING_AFTER));
    nameCode.put("substring", new Integer(F_SUBSTRING));
    nameCode.put("string-length", new Integer(F_STRING_LENGTH));
    nameCode.put("normalize-space", new Integer(F_NORMALIZE_SPACE));
    nameCode.put("translate", new Integer(F_TRANSLATE));
    nameCode.put("boolean", new Integer(F_BOOLEAN));
    nameCode.put("not", new Integer(F_NOT));
    nameCode.put("true", new Integer(F_TRUE));
    nameCode.put("false", new Integer(F_FALSE));
    nameCode.put("lang", new Integer(F_LANG));
    nameCode.put("number", new Integer(F_NUMBER));
    nameCode.put("sum", new Integer(F_SUM));
    nameCode.put("floor", new Integer(F_FLOOR));
    nameCode.put("ceiling", new Integer(F_CEILING));
    nameCode.put("round", new Integer(F_ROUND));
  }

  private final List m_args;
  private final String m_name;
  private final int m_opcode;
  private final int m_args_len;

  FunctionExpression(String name, List args) {
    m_name = name;
    m_args = args;
    m_args_len = (m_args != null)? m_args.size() : 0;

    Integer code = (Integer)nameCode.get(m_name);
    if (code == null) {
      throw new IllegalArgumentException("unknown function: "+m_name);
    }
    m_opcode = code.intValue();

    int args_needed = argCount[m_opcode];
    switch (args_needed) {
    case -3:
      if (m_args_len < 2 || m_args_len > 3) {
        throw new IllegalArgumentException(
          "function "+m_name+" needs 2 or 3 arguments, got "+m_args_len);
      }
      break;
    case -2:
      if (m_args_len < 1 || m_args_len > 2) {
        throw new IllegalArgumentException(
          "function "+m_name+" needs 1 or 2 arguments, got "+m_args_len);
      }
      break;
    case -1:
      if (m_args_len > 1) {
        throw new IllegalArgumentException(
          "function "+m_name+" needs 0 or 1 argument, got "+m_args_len);
      }
      break;

    case -17:
      if (m_args_len < 2) {
        throw new IllegalArgumentException(
          "function "+m_name+" needs minimum of 2 arguments, got "+m_args_len);
      }
      break;

    default:
       if (args_needed != m_args_len) {
         throw new IllegalArgumentException(
           "function "+m_name+" needs "+args_needed+" arguments, got "+m_args_len);
       }
       break;
    }
  }

  public IContext evaluate(IContext context) {
    Node node = null;
    String tmp = null;
    int len;
    IContext eval0 = null;
    IContext eval1 = null;
    IContext eval2 = null;
    if (m_args_len > 0) {
      eval0 = ((IExpression)m_args.get(0)).evaluate(context);
      if (m_args_len > 1) {
        eval1 = ((IExpression)m_args.get(1)).evaluate(context);
        if (m_args_len > 2) {
          eval2 = ((IExpression)m_args.get(2)).evaluate(context);
        }
      }
    }

    switch (m_opcode) {

    case F_BOOLEAN:
      return context.newInstance(eval0.toBoolean());

    case F_CEILING:
      return context.newInstance(Math.ceil(eval0.toNumber()));

    case F_CONCAT:
      {
        StringBuffer sb = new StringBuffer();
        sb.append(eval0.toString());
        sb.append(eval1.toString());
        if (eval2 != null) {
          sb.append(eval2.toString());
          if (m_args_len > 3) {
            for (int i = 3; i < m_args_len; ++i) {
              IContext eval = ((IExpression)m_args.get(i)).evaluate(context);
              sb.append(eval.toString());
            }
          }
        }
        //System.err.println("concat makes: "+sb.toString());
        return context.newInstance(sb.toString());
      }

    case F_CONTAINS:
      return context.newInstance(eval0.toString().indexOf(eval1.toString()) >= 0);

    case F_COUNT:
      return context.newInstance(eval0.getSize());

    case F_FALSE:
      return context.newInstance(false);

    case F_FLOOR:
      return context.newInstance((int)eval0.toNumber());

    case F_LANG:
      node = context.getCurrentNode();
      boolean match = false;
      if (node != null && node.getNodeType() == Node.ELEMENT_NODE) {
        tmp = eval0.toString().toLowerCase();
        Element element = (Element)node;
        Attr attr = element.getAttributeNodeNS(Consts.NS_XML, "lang");
        while (attr == null && element != null) {
          //dumpAttr(element);
          Node parent = context.getParent(element);
          if (parent.getNodeType() == Node.ELEMENT_NODE) {
            element = (Element)parent;
            attr = element.getAttributeNodeNS(Consts.NS_XML, "lang");
          }
          else {
            element = null;
            attr = null;
          }
        }

        if (attr != null) {
          String attrValue = attr.getNodeValue().toLowerCase();
          if (attrValue.equals(tmp)) {
            match = true;
          }
          else if (attrValue.startsWith(tmp)) {
            match = (attrValue.length() > tmp.length()
                     && tmp.charAt(attrValue.length()) == '_');
          }
        }
      }
      return context.newInstance(match);

    case F_LAST:
      return context.newInstance(context.getSize());

    case F_LOCAL_NAME:
      if (eval0 == null) {
        node = context.getCurrentNode();
      }
      else {
        Node[] nodes = eval0.getNodes();
        if (nodes != null && nodes.length > 0) {
          node = nodes[0];
        }
      }
      if (node != null) {
        tmp = node.getLocalName();
        if (tmp != null) {
          return context.newInstance(tmp);
        }
      }
      return context.newInstance("");

    case F_NAME:
      if (eval0 == null) {
        node = context.getCurrentNode();
      }
      else {
        Node[] nodes = eval0.getNodes();
        if (nodes != null && nodes.length > 0) {
          node = nodes[0];
        }
      }
      if (node != null) {
        tmp = node.getNodeName();
        if (tmp != null) {
          return context.newInstance(tmp);
        }
      }
      return context.newInstance("");

    case F_NAMESPACE_URI:
      if (eval0 == null) {
        node = context.getCurrentNode();
      }
      else {
        Node[] nodes = eval0.getNodes();
        if (nodes != null && nodes.length > 0) {
          node = nodes[0];
        }
      }
      if (node != null) {
        tmp = node.getNamespaceURI();
        if (tmp != null) {
          return context.newInstance(tmp);
        }
      }
      return context.newInstance("");

    case F_NORMALIZE_SPACE:
      if (eval0 == null) {
        node = context.getCurrentNode();
        if (node != null) {
          tmp = context.toString(node);
        }
        else {
          tmp = "";
        }
      }
      else {
        tmp = eval0.toString();
      }
      //System.err.println("normalize-space: ("+tmp+")");
      len = tmp.length();
      {
        StringBuffer sb = new StringBuffer(len);
        boolean prefixSpace = false;
        boolean seenWord = false;
        for (int i = 0; i < len; ++i) {
          char c = tmp.charAt(i);
          if (!Character.isWhitespace(c)) {
            if (prefixSpace) {
              sb.append(" ");
              prefixSpace = false;
            }
            sb.append(c);
            seenWord = true;
          }
          else {
            prefixSpace = seenWord;
          }
        }
        //System.err.println("normalize-space -> ("+sb.toString()+")");
        return context.newInstance(sb.toString());
      }

    case F_NOT:
      return context.newInstance(!eval0.toBoolean());

    case F_NUMBER:
      if (eval0 == null) {
        node = context.getCurrentNode();
        if (node != null) {
          return context.newInstance(context.toNumber(node));
        }
        else {
          return context.newInstance(0);
        }
      }
      else {
        return context.newInstance(eval0.toNumber());
      }

    case F_POSITION:
      return context.newInstance(context.getPosition());

    case F_ROUND:
      double d = eval0.toNumber();
      if (Double.isNaN(d) || Double.isInfinite(d)) {
        return context.newInstance(d);
      }
      return context.newInstance(Math.round(d));

    case F_STARTS_WITH:
      return context.newInstance(eval0.toString().startsWith(eval1.toString()));

    case F_STRING:
      if (eval0 == null) {
        node = context.getCurrentNode();
        if (node != null) {
          //System.err.println("string() current: "+context.toString((Node)result));
          return context.newInstance(context.toString(node));
        }
        else {
          return context.newInstance("");
        }
      }
      else {
        return context.newInstance(eval0.toString());
      }

    case F_STRING_LENGTH:
      {
        if (eval0 == null) {
          node = context.getCurrentNode();
          if (node != null) {
            //System.err.println("string() current: "+context.toString((Node)result));
            tmp = context.toString(node);
          }
          else {
            return context.newInstance(0);
          }
        }
        else {
          tmp = eval0.toString();
        }
        UnicodeString utmp = new UnicodeString(tmp);
        return context.newInstance(utmp.length());
      }

    case F_SUBSTRING:
      {
        String s0 = eval0.toString();
        double d1 = eval1.toNumber();
        if (Double.isNaN(d1)) {
          return context.newInstance("");
        }
        int start = (int)Math.round(d1);
        if (start < 0)
          start = 0;
        else
          start -= 1;

        UnicodeString us0 = new UnicodeString(s0);
        len = us0.length();
        if (start >= len)
          return context.newInstance("");

        int end = len;
        if (eval1 != null) {
          d1 = eval1.toNumber();
          if (Double.isNaN(d1)) {
            return context.newInstance("");
          }
          end = start + (int)Math.round(d1);
          if (end > len)
            end = len;
          return context.newInstance(us0.substr(start, end));
        }
        else {
          return context.newInstance(us0.substr(start));
        }
      }

    case F_SUBSTRING_AFTER:
      {
        String s0 = eval0.toString();
        String s1 = eval1.toString();
        int index = s0.indexOf(s1);
        if (index > 0) {
          return context.newInstance(s0.substring(index+s1.length()));
        }
        else
          return context.newInstance("");
      }

    case F_SUBSTRING_BEFORE:
      {
        String s0 = eval0.toString();
        int index = s0.indexOf(eval1.toString());
        if (index > 0) {
          return context.newInstance(s0.substring(0, index));
        }
        else
          return context.newInstance("");
      }

    case F_SUM:
      {
        double sum = 0.0;
        Node[] nodes = eval0.getNodes();
        if (nodes != null) {
          for (int i = 0; i < nodes.length; ++i) {
            sum += context.toNumber(nodes[i]);
          }
        }
        //System.err.println("sum is "+sum);
        return context.newInstance(sum);
      }

    case F_TRANSLATE:
      {
        UnicodeString us0 = new UnicodeString(eval0.toString());
        return context.newInstance(us0.translate(eval1.toString(), eval2.toString()));
      }

    case F_TRUE:
      return context.newInstance(true);

    default:
      throw new IllegalArgumentException("not implemented: "+m_name);
    }
  }

  public boolean evaluateBoolean(IContext context) {
    // TODO: optimize for certain functions
    return evaluate(context).toBoolean();
  }

  public int getType() {
    switch (m_opcode) {

    case F_BOOLEAN:
    case F_CONTAINS:
    case F_FALSE:
    case F_LANG:
    case F_NOT:
    case F_STARTS_WITH:
    case F_TRUE:
      return IContext.TYPE_BOOLEAN;

    case F_CEILING:
    case F_COUNT:
    case F_FLOOR:
    case F_LAST:
    case F_NUMBER:
    case F_POSITION:
    case F_ROUND:
    case F_STRING_LENGTH:
    case F_SUM:
      return IContext.TYPE_NUMBER;

    case F_CONCAT:
    case F_LOCAL_NAME:
    case F_NAME:
    case F_NAMESPACE_URI:
    case F_NORMALIZE_SPACE:
    case F_STRING:
    case F_SUBSTRING:
    case F_SUBSTRING_AFTER:
    case F_SUBSTRING_BEFORE:
    case F_TRANSLATE:
      return IContext.TYPE_STRING;

    default:
      throw new IllegalArgumentException("not implemented: "+m_name);
    }
  }

  public String toString() {
    StringBuffer sb = new StringBuffer();
    sb.append(m_name);
    if (m_args != null && !m_args.isEmpty()) {
      sb.append('(');
      int len = m_args.size();
      for (int i = 0; i < len; ++i) {
        if (i > 0) {
          sb.append(", ");
        }
        sb.append(m_args.get(i));
      }
      sb.append(')');
    }
    else {
      sb.append("()");
    }
    return sb.toString();
  }


//  private void dumpAttr(Element elem) {
//    NamedNodeMap map = elem.getAttributes();
//    for (int i = 0; i < map.getLength(); ++i) {
//      System.err.println("attr of "+elem+" is "+map.item(i));
//    }
//  }
}
