package com.sapportals.wcm.protocol.webdav.dompath;

import java.util.List;

import org.w3c.dom.Node;

/**
 * Represents a value in <a href="http://www.w3.org/TR/xpath">XPath</a>
 * expression evaluation.
 * <p>
 * In {@link com.sapportals.wcm.protocol.webdav.dompath.DomPath} all (sub-)expressions
 * are evaluated to IContext objects. This is transparent to users
 * of DomPath. However if you want to write an extension function
 * you might want to work with IContext direclty and accept
 * IContext objects as parameters to your function.
 * <p>
 * An IContext always represents one of five possible types: node set,
 * number, string, boolean or object.<br>
 * Node sets are instances of {@link java.util.List} of which all
 * list members must be of type {@link org.w3c.dom.Node}.<br>
 * Numbers are any instance of {@link java.lang.Number}, however
 * internally all number values are kept as double.<br>
 * Objects can be any {@link java.lang.Object}. IContext converts
 * these objects by using <code>toString()</code> for determining
 * needed string, number or boolean values.
 * <p>
 * Conversions of IContext objects to String, double and boolean
 * are available and done according to the XPath string(), number()
 * and boolean() functions as described in
 * <a href="http://www.w3.org/TR/xpath">XPath</a>.
 * <p>
 * Similar to conversions, comparision is done according to XPath.
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author stefan.eissing@greenbytes.de
 * @version $Revision: 1.2 $
 */
interface IContext {

  /** IContext represents a node set */
  public static final int TYPE_NODESET = 1;
  /** IContext represents a number */
  public static final int TYPE_NUMBER = 2;
  /** IContext represents a string */
  public static final int TYPE_STRING = 3;
  /** IContext represents a boolean */
  public static final int TYPE_BOOLEAN = 4;
  /** IContext represents none of the other */
  public static final int TYPE_OBJECT = 5;

  // Construction -----------------------------

  /**
   * Create a new IContext which represents the given node set
   * @param nodeSet list of {@link org.w3c.dom.Node} objects
   * @return new IContext representing node set
   */
  public IContext newInstance(List nodeSet);

  /**
   * Create a new IContext which represents the given node set
   * @param nodes array of {@link org.w3c.dom.Node} objects
   * @return new IContext representing node set
   */
  public IContext newInstance(Node[] nodes);

  /**
   * Create a new IContext which represents the given node set
   * @param node content of node set
   * @return new IContext representing node set
   */
  public IContext newInstance(Node node);

  /**
   * Create a new IContext which represents the given boolean value
   * @param value the boolean for the new context
   * @return new IContext representing boolean
   */
  public IContext newInstance(boolean value);

  /**
   * Create a new IContext which represents the given number value
   * @param value the int for the new context
   * @return new IContext representing number
   */
  public IContext newInstance(int value);

  /**
   * Create a new IContext which represents the given number value
   * @param value the double for the new context
   * @return new IContext representing number
   */
  public IContext newInstance(double value);

  /**
   * Create a new IContext which represents the given string value
   * @param value the string for the new context
   * @return new IContext representing string
   */
  public IContext newInstance(String value);

  /**
   * Create a new IContext which represents the given value. The
   * value is introspected to determine the type of the new IContext.
   * <ul>
   * <li>If the Object is instanceof java.util.List, the IContext will
   * have TYPE_NODESET.</li>
   * <li>If the Object is instanceof java.lang.String, the IContext
   * will have TYPE_STRING.</li>
   * <li>If the Object is instanceof java.lang.Number, the IContext
   * will have TYPE_NUMBER.</li>
   * <li>If the Object is instanceof java.lang.Boolean, the IContext
   * will have TYPE_BOOLEAN.</li>
   * <li>otherwise the new IContext will have TYPE_OBJECT.
   * </ul>
   * @param value the Object for the new context
   * @return new IContext
   */
  public IContext newInstance(Object value);


  // Access -----------------------------------

  /**
   * Type of this IContext.
   * @return the type of the IContext
   */
  public int getType();

  /**
   * Returns the node set or <code>null</code> if this IContext
   * is not of type TYPE_NODESET.
   * @return the node set or <code>null</code>
   */
  public Node[] getNodes();

  /**
   * Retrieves object at current position.
   * @return the object at the current position.
   */
  public Object getCurrent();

  /**
   * Retrieves Node at current position.
   * @return the node at the current position or <code>null</code>
   */
  public Node getCurrentNode();

  /**
   * Determines number of objects.
   * @return number of objects in this context
   */
  public int getSize();

  /**
   * Returns current position
   * @return current position
   */
  public int getPosition();

  /**
   * Increment the current position and indicate success.
   * @return if there is a next position
   */
  public boolean nextPosition();

  /**
   * Reset the current position to the beginngin
   */
  public void resetPosition();


  // Comparisions ------------------------------

  /**
   * Determines equality according to XPath.
   * @param other IContext to compare this one with
   * @return if this IContext equals other according to XPath.
   */
  public boolean equals(IContext other);

  /**
   * Determines > according to XPath.
   * @param other IContext to compare this one with
   * @return if this IContext > other according to XPath
   */
  public boolean greaterThan(IContext other);

  /**
   * Determines >= according to XPath.
   * @param other IContext to compare this one with
   * @return if this IContext >= other according to XPath
   */
  public boolean greaterOrEqual(IContext other);

  /**
   * Compares this Context with the string representation of a node set.
   * @param strings to compare this one with
   * @return if this IContext equals the list of String objects according to XPath.
   */
  public boolean equals(String[] strings);


  // Conversions -------------------------------

  /**
   * Applies XPath string() function to this context.
   * @return the value as String according to XPath string() function
   */
  public String toString();

  /**
   * Applies XPath number() function to this context.
   * @return the value as double according to XPath number() function
   */
  public double toNumber();

  /**
   * Applies XPath boolean() function to this context.
   * @return the value as boolean according to XPath boolean() function
   */
  public boolean toBoolean();

  /**
   * Returns the mimumum numeric value present in this IContext.
   * @return the minimum numeric value present in this IContext
   */
  public double getMin();

  /**
   * Returns the maximum numeric value present in this IContext.
   * @return the maximum numeric value present in this IContext
   */
  public double getMax();


  // Node utilities ----------------------------

  /**
   * Determine the parent Node of the given node
   * @param node to retrieve parent for
   * @return parent Node or <code>null</null>
   */
  public Node getParent(Node node);

  /**
   * Determine the String value of a Node according to XPath string() function
   * @param node to convert
   * @return XPath:string() of node
   */
  public String toString(Node node);

  /**
   * Determine the numeric value of a Node according to XPath number() function
   * @param node to convert
   * @return XPath:number() of node
   */
  public double toNumber(Node node);

}
