package com.sapportals.wcm.protocol.webdav.dompath;

import java.text.ParseException;
import java.util.HashSet;
import java.util.Set;

/**
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author stefan.eissing@greenbytes.de
 * @version $Id: Lexer.java,v 1.2 2004/03/03 17:44:19 jre Exp $
 */
class Lexer {

  static final int NONE = -1;
  static final int EOF = 0;
  static final int STRING = 2;
  static final int LITERAL = 3;
  static final int NUMBER = 4;
  static final int DCOLON = 5;
  static final int DDCOLON = 6;
  static final int OBRACE = 7;
  static final int CBRACE = 8;
  static final int OANGULAR = 9;
  static final int CANGULAR = 10;
  static final int AT = 11;
  static final int AXIS_NAME = 12;
  static final int NODE_TYPE = 13;
  static final int NAME_TEST = 14;
  static final int OPERATOR = 15;
  static final int FUNCTION_NAME = 16;
  static final int VARIABLE = 17;
  static final int COMMA = 18;

  private static final String SPECIAL = "()[]@.,:/|+-=!<>,$";

  private static final Set m_axisNames;
  private static final Set m_nodeTypes;
  private static final Set m_operators;

  static {
    m_axisNames = new HashSet(17);
    m_axisNames.add("ancestor");
    m_axisNames.add("ancestor-or-self");
    m_axisNames.add("attribute");
    m_axisNames.add("child");
    m_axisNames.add("descendant");
    m_axisNames.add("descendant-or-self");
    m_axisNames.add("following");
    m_axisNames.add("following-sibling");
    m_axisNames.add("namespace");
    m_axisNames.add("parent");
    m_axisNames.add("preceding");
    m_axisNames.add("preceding-sibling");
    m_axisNames.add("self");

    m_nodeTypes = new HashSet(7);
    m_nodeTypes.add("comment");
    m_nodeTypes.add("text");
    m_nodeTypes.add("processing-instruction");
    m_nodeTypes.add("node");

    m_operators = new HashSet(7);
    m_operators.add("and");
    m_operators.add("or");
    m_operators.add("mod");
    m_operators.add("div");
  }

  private final String m_path;
  private final int m_len;
  private int m_index;
  private final StringBuffer m_buffer;

  private int m_lastToken;
  private String m_lastString;

  private int m_nextToken;
  private String m_nextString;

  private int m_prevToken;
  // private String m_prevString;

  Lexer(String path) {
    m_path = path;
    m_len = m_path.length();
    m_index = 0;
    m_buffer = new StringBuffer(m_len);

    m_lastToken = NONE;
    m_nextToken = NONE;
    m_prevToken = NONE;
  }

  int getNext() throws ParseException {
    if (m_nextToken != NONE) {
      m_lastToken = m_nextToken;
      m_lastString = m_nextString;
      m_nextToken = NONE;
      return m_lastToken;
    }

    m_buffer.setLength(0);
    m_prevToken = m_lastToken;
    // m_prevString = m_lastString;

    for (;m_index < m_len; ++m_index) {
      char c = m_path.charAt(m_index);
      switch (c) {
      case ' ':
      case '\t':
      case '\n':
      case '\r':
        break;

      case '(':
        ++m_index;
        m_lastToken = OBRACE;
        m_lastString = "(";
        return m_lastToken;

      case ')':
        ++m_index;
        m_lastToken = CBRACE;
        m_lastString = ")";
        return m_lastToken;

      case '[':
        ++m_index;
        m_lastToken = OANGULAR;
        m_lastString = "[";
        return m_lastToken;

      case ']':
        ++m_index;
        m_lastToken = CANGULAR;
        m_lastString = "]";
        return m_lastToken;

      case '@':
        ++m_index;
        m_lastToken = AT;
        m_lastString = "@";
        return m_lastToken;

      case '$':
        ++m_index;
        m_lastToken = VARIABLE;
        appendNCName();
        return m_lastToken;

      case ',':
        ++m_index;
        m_lastToken = COMMA;
        m_lastString = ",";
        return m_lastToken;

      case '*':
        switch (m_prevToken) {
        case NONE:
        case AT:
        case DDCOLON:
        case OBRACE:
        case OANGULAR:
        case OPERATOR:
          ++m_index;
          m_lastToken = NAME_TEST;
          m_lastString = "*";
          return m_lastToken;

        default:
          break;
        }
        // fall through
      case '-':
      case '+':
      case '=':
      case '!':
      case '<':
      case '>':
      case '|':
      case '/':
        return getOperator();

      case ':':
        ++m_index;
        if (m_index < m_len && m_path.charAt(m_index) == ':') {
          ++m_index;
          m_lastToken = DDCOLON;
          m_lastString = "::";
        }
        else {
          m_lastToken = DCOLON;
          m_lastString = "/";
        }
        return m_lastToken;

      case '"':
        return getLiteral('"');
      case '\'':
        return getLiteral('\'');

      case '.':
        int index = m_index+1;
        if (index < m_len && Character.isDigit(m_path.charAt(index))) {
          return getNumber();
        }
        else {
          return getStringToken();
        }

      case '0':
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
        return getNumber();

      default:
        return getStringToken();
      }

    }

    m_lastToken = EOF;
    m_lastString = null;
    return EOF;
  }

  String getString() {
    return m_lastString;
  }

  int getIndex() {
    return m_index;
  }

  void unget() {
    m_nextToken = m_lastToken;
    m_nextString = m_lastString;
  }

  private int getLiteral(char end) throws ParseException {
    ++m_index;
    for (;m_index < m_len; ++m_index) {
      char c = m_path.charAt(m_index);
      if (c == end) {
        ++m_index;
        m_lastToken = LITERAL;
        m_lastString = m_buffer.toString();
        return m_lastToken;
      }

      m_buffer.append(c);
    }

    throw new ParseException("unterminated literal", m_index);
  }

  private int getStringToken() throws ParseException {
    for (;m_index < m_len; ++m_index) {
      char c = m_path.charAt(m_index);
      switch (c) {
      case ' ':
      case '\t':
      case '\n':
      case '\r':
      case '/':
      case '[':
      case ']':
      case '(':
      case ')':
      case ':':
      case ',':
      case '|':
      case '@':
      case '=':
      case '!':
        m_lastToken = STRING;
        m_lastString = m_buffer.toString();

        switch (m_prevToken) {
        case NONE:
        case AT:
        case DDCOLON:
        case OBRACE:
        case OANGULAR:
        case OPERATOR:
          break;

        default:
          if (!m_operators.contains(m_lastString)) {
            throw new ParseException("unknown operator '"+m_lastString+"'", m_index);
          }
          m_lastToken = OPERATOR;
          return m_lastToken;
        }

        String follows = nextNonWS();
        if (follows.startsWith("::") && m_axisNames.contains(m_lastString)) {
          m_lastToken = AXIS_NAME;
        }
        else if (follows.startsWith("(")) {
          if (m_nodeTypes.contains(m_lastString))
            m_lastToken = NODE_TYPE;
          else
            m_lastToken = FUNCTION_NAME;
        }
        else if (c == ':' && m_index < m_len) {
          char c2 = m_path.charAt(m_index+1);
          if (c2 == '*') {
            m_index += 2;
            m_buffer.append(":*");
            m_lastToken = NAME_TEST;
            m_lastString = m_buffer.toString();
          }
          else if (!Character.isDigit(c2) && !isSpecial(c2)) {
            // It must be a NameTest token
            ++m_index;
            m_buffer.append(':');
            appendNCName();
            m_lastString = m_buffer.toString();
            follows = nextNonWS();
            if (follows.startsWith("("))
              m_lastToken = FUNCTION_NAME;
            else
              m_lastToken = NAME_TEST;
          }
        }
        return m_lastToken;

      default:
        m_buffer.append(c);
        break;
      }
    }

    if (m_buffer.length() > 0) {
      m_lastToken = STRING;
      m_lastString = m_buffer.toString();
      return m_lastToken;
    }

    throw new ParseException("unterminated literal", m_index);
  }

  private int getNumber() throws ParseException {
    boolean period = false;
    for (;m_index < m_len; ++m_index) {
      char c = m_path.charAt(m_index);
      if (c == '.' && !period) {
        period = true;
        m_buffer.append(c);
      }
      else if (!Character.isDigit(c)) {
        m_lastToken = NUMBER;
        m_lastString = m_buffer.toString();
        return m_lastToken;
      }
      else {
        m_buffer.append(c);
      }
    }

    if (m_buffer.length() > 0) {
      m_lastToken = NUMBER;
      m_lastString = m_buffer.toString();
      return m_lastToken;
    }

    throw new ParseException("unterminated literal", m_index);
  }

  private int getOperator() throws ParseException {
    char c = m_path.charAt(m_index);
    switch (c) {
    case '*':
    case '-':
    case '+':
    case '=':
    case '|':
      ++m_index;
      m_lastToken = OPERATOR;
      m_lastString = String.valueOf(c);
      break;

    case '/':
      ++m_index;
      m_lastToken = OPERATOR;
      if (m_index < m_len && m_path.charAt(m_index) == '/') {
        ++m_index;
        m_lastString = "//";
      }
      else {
        m_lastString = "/";
      }
      break;

    case '<':
    case '>':
      ++m_index;
      m_lastToken = OPERATOR;
      if (m_index < m_len && m_path.charAt(m_index) == '=') {
        ++m_index;
        m_buffer.append(c).append('=');
      }
      else {
        m_buffer.append(c);
      }
      m_lastString = m_buffer.toString();
      break;

    case '!':
      ++m_index;
      if (m_index < m_len && m_path.charAt(m_index) == '=') {
        ++m_index;
        m_lastToken = OPERATOR;
        m_lastString = "!=";
        break;
      }
      // fall through
    default:
      throw new ParseException("unrecognized operator '"+c+"'", m_index);
    }

    return m_lastToken;
  }

  private final boolean isSpecial(char c) {
    return SPECIAL.indexOf(c) >= 0;
  }

  private final String nextNonWS() {
    int index = m_index;
    while (index < m_len) {
      if (!Character.isWhitespace(m_path.charAt(index))) {
        return m_path.substring(index);
      }
      ++index;
    }

    return "";
  }

  private void appendNCName() {
    for (;m_index < m_len; ++m_index) {
      char c = m_path.charAt(m_index);
      if (c != '-' && (Character.isWhitespace(c) || isSpecial(c))) {
        break;
      }
      m_buffer.append(c);
    }
  }
}
