package com.sapportals.wcm.protocol.webdav.dompath;

import java.util.Collections;
import java.util.List;

import org.w3c.dom.Attr;
import org.w3c.dom.Node;


/**
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author stefan.eissing@greenbytes.de
 * @version $Id: PathContext.java,v 1.3 2004/06/03 15:36:42 jre Exp $
 */
class PathContext implements IContext {

  public static final IContext TRUE;
  public static final IContext FALSE;
  public static final IContext EMPTY_SET;
  public static final IContext EMPTY_STRING;
  public static final IContext N_0;
  public static final IContext N_1;
  public static final IContext N_2;
  public static final IContext N_3;

  private static final Node[] EMPTY_NODES = new Node[0];
  private static final String[] EMPTY_STRINGS = new String[0];

  static {
    TRUE = new PathContext(true);
    FALSE = new PathContext(false);
    EMPTY_SET = new PathContext(Collections.EMPTY_LIST);
    EMPTY_STRING = new PathContext("");
    N_0 = new PathContext(0);
    N_1 = new PathContext(1);
    N_2 = new PathContext(2);
    N_3 = new PathContext(3);
  }

  public static final Node GetParent(Node node) {
    switch (node.getNodeType()) {
    case Node.ATTRIBUTE_NODE:
      return ((Attr)node).getOwnerElement();
    default:
      return node.getParentNode();
    }
  }

  public static final double ToNumber(String s) {
    try {
      return Double.parseDouble(s);
    }
    catch (NumberFormatException ex) {
      //$JL-EXC$
      return Double.NaN;
    }
  }

  private final int m_type;

  private final Node[] m_nodes;
  private double m_number;
  private String m_string;
  private boolean m_boolean;
  private Object m_object;

  private int m_position;

  private boolean m_knownMinMax;
  private double m_maximum;
  private double m_minimum;
  private String[] m_strings;

  private PathContext(Object start) {
    m_type = TYPE_OBJECT;
    m_nodes = null;
    m_object = start;
  }

  public PathContext(Node start) {
    m_type = TYPE_NODESET;
    m_nodes = new Node[1];
    m_nodes[0] = start;
  }

  public PathContext(List start) {
    m_type = TYPE_NODESET;
    m_nodes = (Node[])start.toArray(EMPTY_NODES);
  }

  public PathContext(Node[] nodes) {
    m_type = TYPE_NODESET;
    m_nodes = nodes;
  }

  public PathContext(double value) {
    m_type = TYPE_NUMBER;
    m_nodes = null;
    m_number = value;
  }

  public PathContext(String value) {
    m_type = TYPE_STRING;
    m_nodes = null;
    m_string = value;
  }

  public PathContext(boolean value) {
    m_type = TYPE_BOOLEAN;
    m_nodes = null;
    m_boolean = value;
  }

  public Node[] getNodes() {
    if (m_nodes != null) {
      return m_nodes;
    }
    return EMPTY_NODES;
  }

  public int getType() {
    return m_type;
  }

  public Object getCurrent() {
    if (m_position == 0) m_position = 1;
    switch (m_type) {
    case TYPE_NODESET:
      int pos = m_position-1;
      return (m_nodes.length > pos)? m_nodes[pos] : null;
    case TYPE_NUMBER:
      return new Double(m_number);
    case TYPE_STRING:
      return m_string;
    case TYPE_BOOLEAN:
      return new Boolean(m_boolean);
    default:
      return m_object;
    }
  }

  /**
   * Retrieves Node at current position.
   * @return the node at the current position or <code>null</code>
   */
  public Node getCurrentNode() {
    if (m_type == TYPE_NODESET) {
      if (m_position == 0) m_position = 1;
      int pos = m_position-1;
      return (m_nodes.length > pos)? m_nodes[pos] : null;
    }
    return null;
  }

  public Node getParent(Node node) {
    return GetParent(node);
  }

  public int getSize() {
    return (m_type == IContext.TYPE_NODESET)? m_nodes.length : 1;
  }

  public int getPosition() {
    return m_position;
  }

  public boolean nextPosition() {
    return ++m_position <= getSize();
  }

  public void resetPosition() {
    m_position = 0;
  }

  public IContext newInstance(Node[] nodes) {
    if (nodes == null || nodes.length == 0) {
      return EMPTY_SET;
    }
    return new PathContext(nodes);
  }

  public IContext newInstance(List startSet) {
    if (startSet.isEmpty()) {
      return EMPTY_SET;
    }
    return new PathContext(startSet);
  }

  public IContext newInstance(Node node) {
    if (node == null) {
      return EMPTY_SET;
    }
    return new PathContext(node);
  }

  public IContext newInstance(Object start) {
    return new PathContext(start);
  }

  public IContext newInstance(boolean value) {
    return (value? TRUE : FALSE);
  }

  public IContext newInstance(int value) {
    switch (value) {
    case 0:
      return N_0;
    case 1:
      return N_1;
    case 2:
      return N_2;
    case 3:
      return N_3;
    default:
      return new PathContext(value);
    }
  }

  public IContext newInstance(double value) {
    return new PathContext(value);
  }

  public IContext newInstance(String value) {
    if (value == null || value.length() == 0) {
      return EMPTY_STRING;
    }
    return new PathContext(value);
  }


  public boolean toBoolean() {
    switch (m_type) {
    case TYPE_NODESET:
      return m_nodes.length > 0;
    case TYPE_NUMBER:
      return m_number != 0;
    case TYPE_STRING:
      return m_string.length() > 0;
    case TYPE_BOOLEAN:
      return m_boolean;
    default:
      return m_object.equals(Boolean.TRUE);
    }
  }

  public String toString() {
    switch (m_type) {
    case TYPE_NODESET:
      return (m_nodes.length > 0)? toString(m_nodes[0]) : "";
    case TYPE_NUMBER:
      return String.valueOf(m_number);
    case TYPE_STRING:
      return m_string;
    case TYPE_BOOLEAN:
      return m_boolean? "true" : "false";
    default:
      return m_object.toString();
    }
  }

  public double toNumber() {
    switch (m_type) {
    case TYPE_NODESET:
      return toNumber(toString());
    case TYPE_NUMBER:
      return m_number;
    case TYPE_STRING:
      return toNumber(m_string);
    case TYPE_BOOLEAN:
      return m_boolean? 1.0 : 0.0;
    default:
      return toNumber(m_object.toString());
    }
  }

  public boolean equals(IContext other) {
    switch (m_type) {
    case TYPE_NODESET:
      if (m_nodes.length == 0 || other.getSize() == 0) {
        return false;
      }
      return other.equals(getStrings());
    case TYPE_BOOLEAN:
      return m_boolean == other.toBoolean();
    case TYPE_NUMBER:
      if (other.getType() == IContext.TYPE_NODESET) return other.equals(this);
      return m_number == other.toNumber();
    case TYPE_STRING:
      if (other.getType() == IContext.TYPE_NODESET) return other.equals(this);
      return m_string.equals(other.toString());
    default:
      if (other.getType() == IContext.TYPE_NODESET) return other.equals(this);
      return m_object.equals(other.getCurrent());
    }
  }

  public boolean equals(String[] strings) {
    switch (m_type) {
    case TYPE_NODESET:
      String[] my = getStrings();
      if (my.length == strings.length) {
        for (int i = 0; i < strings.length; ++i) {
          if (!my[i].equals(strings[i])) {
            return false;
          }
        }
        return true;
      }
      return false;
    case TYPE_NUMBER:
      for (int i = 0; i < strings.length; ++i) {
        if (m_number == toNumber(strings[i]))
          return true;
      }
      return false;
    case TYPE_STRING:
      for (int i = 0; i < strings.length; ++i) {
        if (m_string.equals(strings[i]))
          return true;
      }
      return false;
    case TYPE_BOOLEAN:
      return m_boolean == (strings.length > 0);
    default:
      for (int i = 0; i < strings.length; ++i) {
        if (m_object.equals(strings[i]))
          return true;
      }
      return false;
    }
  }

  public boolean greaterThan(IContext other) {
    if (m_nodes != null) {
      if (m_nodes.length == 0 || other.getSize() == 0) {
        return false;
      }
    }

    return this.getMax() > other.getMin();
  }

  public boolean greaterOrEqual(IContext other) {
    if (m_nodes != null) {
      if (m_nodes.length == 0 || other.getSize() == 0) {
        return false;
      }
    }

    return this.getMax() >= other.getMin();
  }

  public final double getMax() {
    if (!m_knownMinMax) {
      calcMinMax();
    }
    return m_maximum;
  }

  public final double getMin() {
    if (!m_knownMinMax) {
      calcMinMax();
    }
    return m_minimum;
  }

  private static void addTextDescendants(Node node, StringBuffer sb) {
    for (Node child = node.getFirstChild(); child != null; child = child.getNextSibling()) {
      switch (child.getNodeType()) {
      case Node.TEXT_NODE:
      case Node.CDATA_SECTION_NODE:
        sb.append(child.getNodeValue());
        break;
      case Node.ELEMENT_NODE:
        addTextDescendants(child, sb);
        break;
      default:
        break;
      }
    }
  }

  private final void calcMinMax() {
    if (m_nodes != null) {
      String[] strings = getStrings();
      double min = Double.POSITIVE_INFINITY;
      double max = Double.NEGATIVE_INFINITY;

      for (int i = 0; i < strings.length; ++i) {
         double d = toNumber(strings[i]);
         if (d > max) {
           max = d;
         }
         if (d < min) {
           min = d;
         }
      }
      m_maximum = max;
      m_minimum = min;
    }
    else {
      m_maximum = this.toNumber();
      m_minimum = m_maximum;
    }
    m_knownMinMax = true;
  }

  private final String[] getStrings() {
    if (m_nodes != null) {
      if (m_strings == null) {
        calcStrings();
      }
      return m_strings;
    }
    return EMPTY_STRINGS;
  }

  private final void calcStrings() {
    m_strings = new String[m_nodes.length];
    for (int i = 0; i < m_nodes.length; ++i) {
      m_strings[i] = toString(m_nodes[i]);
    }
  }

  public final double toNumber(String s) {
    return ToNumber(s);
  }

  public final double toNumber(Node node) {
    return ToNumber(toString(node));
  }

  public final String toString(Node node) {
    switch (node.getNodeType()) {
    case Node.DOCUMENT_NODE:
    case Node.ELEMENT_NODE:
      StringBuffer sb = new StringBuffer();
      addTextDescendants(node, sb);
      return sb.toString();
    default:
      return node.getNodeValue();
    }
  }


}
