package com.sapportals.wcm.protocol.webdav.dompath;

import java.util.List;

import org.w3c.dom.Node;

/**
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author stefan.eissing@greenbytes.de
 * @version $Id: Result.java,v 1.2 2004/03/03 17:44:19 jre Exp $
 */
final class Result {

  private static final Node[] EMPTY_NODES = new Node[0];

  private final IContext m_context;
  private Node[] m_nodes;
  private int m_length;
  private Node m_node;

  public Result(IContext context) {
    m_context = context;
  }

  /**
   * Add a list of {@link org.w3c.dom.Node}s to the result node set.
   * Note that the result set is a set and thus will not keep duplicate Nodes.
   * @param nodes list of Node objects
   */
  public final void add(List nodes) {
    int len = nodes.size();
    ensureAdditionalCapacity(len);

    for (int i = 0; i < len; ++i) {
      addInternal((Node)nodes.get(i));
    }
  }

  /**
   * Add an array of {@link org.w3c.dom.Node}s to the result node set.
   * Note that the result set is a set and thus will not keep duplicate Nodes.
   * @param nodes of Node objects
   */
  public final void add(Node[] nodes) {
    ensureAdditionalCapacity(nodes.length);

    for (int i = 0; i < nodes.length; ++i) {
      addInternal(nodes[i]);
    }
  }

  /**
   * Add a single {@link org.w3c.dom.Node} to the result node set.
   * Note that the result set is a set and thus will not keep duplicate Nodes.
   * @param node new result Node
   */
  public final void add(Node node) {
    if (node != null) {
      ensureAdditionalCapacity(1);
      addInternal(node);
    }
  }

  public final void clear() {
    m_node = null;
    m_length = 0;
  }

  public final Node[] getNodes() {
    if (m_nodes == null) {
      if (m_node != null) {
        return new Node[] { m_node };
      }
      else {
        return EMPTY_NODES;
      }
    }
    else if (m_length == 0) {
      return EMPTY_NODES;
    }
    else {
      if (m_length == m_nodes.length) {
        return m_nodes;
      }
      else {
        Node[] nodes = new Node[m_length];
        System.arraycopy(m_nodes, 0, nodes, 0, m_length);
        return nodes;
      }
    }
  }

  public final boolean isEmpty() {
    return (m_length == 0 && m_node == null);
  }

  /**
   * Makes the currently collected result node set the node set
   * represented by this IContext and clears the result node set.
   * <p>
   * This is useful if you have to manipulate an TYPE_NODESET IContext
   * object over and over again, since it saves you to create a new
   * IContext for every iteration.
   * <p>
   * <b>Caution:</b>the current value of the IContext is destroyed. So,
   * only apply this to a IContext which you created yourself or else.
   */
  public final IContext toContext() {
    return m_context.newInstance(getNodes());
  }

  private final void ensureAdditionalCapacity(int additional) {
    if (m_nodes == null) {
      if (additional == 1 && m_node == null) {
        return;
      }
      else {
        m_nodes = new Node[additional+10];
        m_length = 0;
        if (m_node != null) {
          m_nodes[0] = m_node;
          m_length = 1;
          m_node = null;
        }
      }
    }
    else {
      int needed = m_length + additional;
      if (needed > m_nodes.length) {
        int newSize = 2 * m_nodes.length;
        Node[] nodes = new Node[(needed > newSize)? needed : newSize];
        System.arraycopy(m_nodes, 0, nodes, 0, m_length);
        m_nodes = nodes;
      }
    }
  }

  private final void addInternal(Node node) {
    if (m_nodes == null) {
      m_node = node;
    }
    else {
      int nodeHash = node.hashCode();
      for (int i = 0; i < m_length; ++i) {
        Node other = m_nodes[i];
        if (nodeHash == other.hashCode() && node.equals(other)) {
          return;
        }
      }

      m_nodes[m_length++] = node;
    }
  }
}
