/*
 * Copyright (c) 2004 by SAP AG. All Rights Reserved.
 *
 * SAP, mySAP, mySAP.com and other SAP products and
 * services mentioned herein as well as their respective
 * logos are trademarks or registered trademarks of
 * SAP AG in Germany and in several other countries all
 * over the world. MarketSet and Enterprise Buyer are
 * jointly owned trademarks of SAP AG and Commerce One.
 * All other product and service names mentioned are
 * trademarks of their respective companies.
 *
 * @version $Id: HTMLLogin.java,v 1.4 2004/09/13 14:26:09 jre Exp $
 */
package com.sapportals.wcm.protocol.webdav.html;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.Collections;
import java.util.List;

import com.sapportals.wcm.protocol.webdav.QPCodec;
import com.sapportals.wcm.repository.IOErrorException;
import com.sapportals.wcm.repository.InvalidArgumentException;
import com.sapportals.wcm.repository.ResourceException;
import com.sapportals.wcm.util.html.HTMLException;
import com.sapportals.wcm.util.html.IHTMLReader;
import com.sapportals.wcm.util.http.ContentType;
import com.sapportals.wcm.util.http.HttpRequest;
import com.sapportals.wcm.util.http.UserInfo;
import com.sapportals.wcm.util.uri.IHierarchicalUri;
import com.sapportals.wcm.util.uri.IUriReference;
import com.sapportals.wcm.util.uri.RID;
import com.sapportals.wcm.util.uri.UriQuery;
import com.sapportals.wcm.util.uri.UriReference;

/**
 * Represents a HTML login form. <p>
 *
 * Copyright (c) SAP AG 2001-2004
 *
 * @author stefan.eissing@greenbytes.de
 * @version $Id: HTMLLogin.java,v 1.4 2004/09/13 14:26:09 jre Exp $
 */
public final class HTMLLogin implements IHTMLLogin {

  private static final String CRLF = "\r\n";
  
  /**
   * Extracts a login form from the HTML reader. Either the id of the form is
   * specified or the form is autodetected. In case autodetection does not work or
   * the given form id does not exist, a HTMLFormNotFoundException is thrown.
   * @param reader to read HTML form from
   * @param base uri of html file
   * @param formID of form to find or <code>null</code> for autodetection
   * @return found HTML login form
   * @throws HTMLFormNotFoundException when form could not be detected
   * @throws HTMLException on generic HTML errors
   * @throws IOException on errors reading the HTML
   */
  public static IHTMLLogin getInstance(IHTMLReader reader, IHierarchicalUri base, String formID) 
    throws HTMLFormNotFoundException, HTMLException, IOException {
    HTMLLoginDetector detector = HTMLLoginDetector.getInstance(base, formID);
    detector.setParent(reader);
    detector.parse();
    return detector.getLogin();
  }
  
  static IHTMLLogin getInstance(HTMLForm form) {
    return new HTMLLogin(form);
  }
  
  private final HTMLForm form;
  
  private HTMLLogin(HTMLForm form) {
    this.form = form;
  }
  
  //------------ IHTMLLogin ----------------------------------------------------
  
  public IHierarchicalUri getAction() {
    return this.form.action;
  }

  public String getEncType() {
    return this.form.encType;
  }

  public String getID() {
    return this.form.id;
  }

  public List getInputs() {
    return Collections.unmodifiableList(this.form.inputs);
  }

  public String getMethod() {
    return this.form.method;
  }

  public HttpRequest createRequest(UserInfo ui) throws ResourceException {
    HttpRequest request = new HttpRequest();
    if ("GET".equalsIgnoreCase(this.form.method)) {
      request.setMethod("GET");
      UriQuery uq = createQuery(ui.getUser(), ui.getPassword());
      IUriReference ref = new UriReference(this.form.action.getPath(), uq.toExternalForm(), null);
      request.setURI(ref);
    }
    else {
      request.setMethod("POST");
      request.setURI(new UriReference(this.form.action.getPath(), null, null));
      ContentType ct = ContentType.valueOf(this.form.encType);
      if (ct.getEncoding() == null) {
        ct = ContentType.valueOf(ct.getMimeType(), "UTF-8");
      }
      if (HTMLForm.URLENCODING.equalsIgnoreCase(ct.getMimeType())) {
        UriQuery uq = createQuery(ui.getUser(), ui.getPassword());
        String body = uq.toExternalForm();
        request.setBody(body);        
      }
      else if (HTMLForm.MULTIPART.equalsIgnoreCase(ct.getMimeType())) {
        String boundary = calcBoundary(ui.getUser(), ui.getPassword());
        if (boundary == null) {
          throw new InvalidArgumentException("unable to calculate boundary string");        
        }
        ct = ContentType.valueOf(ct.getMimeType()+"; boundary="+boundary, ct.getEncoding());
        StringBuffer sb = new StringBuffer();
        for (int i = 0, n = this.form.inputs.size(); i < n; ++i) {
          HTMLInput input = (HTMLInput)this.form.inputs.get(i);
          String value = input.getValue();
          if (HTMLInput.PASSWORD.equalsIgnoreCase(input.getType())) {
            value = ui.getPassword();      
          }
          else if (HTMLInput.TEXT.equalsIgnoreCase(input.getType())) {
            value = ui.getUser();
          }
          if (value != null) {
            sb.append("--").append(boundary).append(CRLF);
            sb.append("Content-Disposition: form-data; name=\"");
            sb.append(input.getName()).append('\"').append(CRLF);
            if (isAscii(value)) {
              // no more headers necessary
              sb.append(CRLF);
              sb.append(value).append(CRLF);
            }
            else {
              try {
                sb.append("Content-Type: text/plain; charset=UTF-8").append(CRLF);
                sb.append("Content-Transfer-Encoding: quoted-printable").append(CRLF);
                sb.append(CRLF);
                sb.append(QPCodec.Encode(value, "UTF-8")).append(CRLF);
              }
              catch (UnsupportedEncodingException ex)  {
                // what kind of system is this?
                throw new IOErrorException(ex.getMessage(), ex, (RID)null);
              }
            }
          }
          else if (HTMLInput.HIDDEN.equalsIgnoreCase(input.getType())) {
            sb.append("--").append(boundary).append(CRLF);
            sb.append("Content-Disposition: form-data; name=\"");
            sb.append(input.getName()).append('\"').append(CRLF);
            sb.append(CRLF).append(CRLF);
          }
        }
        sb.append("--").append(boundary).append("--").append(CRLF);
        request.setBody(sb.toString());
      }
      else {
        throw new InvalidArgumentException("unknown html form encoding: "+ct.getMimeType());        
      }      
      request.setContentType(ct.getContentType());
    }
    return request;
  }

  //------------ private -------------------------------------------------------
  
  private UriQuery createQuery(String username, String password) {
    UriQuery uq = new UriQuery();
    for (int i = 0, n = this.form.inputs.size(); i < n; ++i) {
      HTMLInput input = (HTMLInput)this.form.inputs.get(i);
      String value = input.getValue();
      if (HTMLInput.PASSWORD.equalsIgnoreCase(input.getType())) {
        value = password;      
      }
      else if (HTMLInput.TEXT.equalsIgnoreCase(input.getType())) {
        value = username;
      }
      if (value != null) {
        uq.set(input.getName(), value);
      }
      else if (HTMLInput.HIDDEN.equalsIgnoreCase(input.getType())) {
        uq.set(input.getName(), "");
      }
    }
    return uq;      
  }
  
  private static final String[] Boundaries = {
    "AaB03x",
    "WHat_Ever_19991221",
    "Bxjsuw8",
    "djsIS98d",
    "KDHS8wehd",  
  };
  
  private String calcBoundary(String username, String password) {
    loop: for (int i = 0; i < Boundaries.length; ++i) {
      String boundary = Boundaries[i];
      if (username.indexOf(boundary) >= 0) {
        continue loop;
      }
      if (password.indexOf(boundary) >= 0) {
        continue loop;
      }
      for (int j = 0, m = this.form.inputs.size(); j < m; ++i) {
        HTMLInput input = (HTMLInput)this.form.inputs.get(j);
        String value = input.getValue();
        if (value != null && value.indexOf(boundary) >= 0) {
          continue loop;
        }
      }
      return boundary;
    }
    return null;
  }
  
  private boolean isAscii(String s) {
    for (int i = 0, n = s.length(); i < n; ++i) {
      char c = s.charAt(i);
      if (c <= 0 || c > 127) {
        return false;
      }
    }
    return true;
  }
}
