/*
 * Copyright (c) 2004 by SAP AG. All Rights Reserved.
 *
 * SAP, mySAP, mySAP.com and other SAP products and
 * services mentioned herein as well as their respective
 * logos are trademarks or registered trademarks of
 * SAP AG in Germany and in several other countries all
 * over the world. MarketSet and Enterprise Buyer are
 * jointly owned trademarks of SAP AG and Commerce One.
 * All other product and service names mentioned are
 * trademarks of their respective companies.
 *
 * @version $Id: HTMLLoginDetector.java,v 1.5 2004/09/13 14:26:09 jre Exp $
 */
package com.sapportals.wcm.protocol.webdav.html;

import java.io.IOException;
import java.io.StringWriter;
import java.net.MalformedURLException;
import java.util.HashSet;
import java.util.Set;

import com.sap.tc.logging.Location;
import com.sapportals.wcm.util.html.HTMLException;
import com.sapportals.wcm.util.html.HTMLFilterImpl;
import com.sapportals.wcm.util.html.IHTMLElement;
import com.sapportals.wcm.util.html.IHTMLElementStart;
import com.sapportals.wcm.util.logging.LoggingFormatter;
import com.sapportals.wcm.util.uri.IHierarchicalUri;
import com.sapportals.wcm.util.uri.IUri;
import com.sapportals.wcm.util.uri.IUriReference;
import com.sapportals.wcm.util.uri.UriFactory;

/**
 * Detects HTML Login Forms.<p>
 *
 * Copyright (c) SAP AG 2001-2004
 *
 * @author stefan.eissing@greenbytes.de
 * @version $Id: HTMLLoginDetector.java,v 1.5 2004/09/13 14:26:09 jre Exp $
 */
final class HTMLLoginDetector extends HTMLFilterImpl {

  private static final Location log = Location.getLocation(HTMLLoginDetector.class);

  private static final int S_OUTSIDE_FORM = 1;
  private static final int S_INSIDE_FORM  = 2;
  
  private static final Set INTERESTING_INPUT_TYPES;
  static {
    Set set = new HashSet();
    set.add(HTMLInput.TEXT);
    set.add(HTMLInput.PASSWORD);
    set.add(HTMLInput.HIDDEN);
    INTERESTING_INPUT_TYPES = set;
  }
  
  public static HTMLLoginDetector getInstance(IHierarchicalUri base, String formID) {
    return new HTMLLoginDetector(base, formID);
  }
  
  private final IHierarchicalUri base;
  private final String formID;
  private int state;
  private HTMLForm current;
  private HTMLForm found;
  private StringWriter sw;
  
  private HTMLLoginDetector(IHierarchicalUri base, String formID) {
    this.base = base;
    this.formID = (formID != null && formID.length() > 0)? formID.toLowerCase() : null;
    this.state = S_OUTSIDE_FORM;
    this.sw = new StringWriter(8000);
  }
  
  public IHTMLLogin getLogin() throws HTMLFormNotFoundException {
    String page = null;
    if (this.sw != null) {
      StringBuffer sb = this.sw.getBuffer();
      page = sb.toString();
      sb.setLength(0);
      this.sw = null;
    }
    if (this.found == null) {
      if (page != null) {
        log.errorT("unable to detect login form ("+
          ((this.formID != null)? this.formID : "default")
          +") at "+this.base+":"
          +page);
      }
      throw new HTMLFormNotFoundException("could not find login form in "+this.base);
    }
    return HTMLLogin.getInstance(this.found);
  }
  
  //----------------------------------------------------------------------------

  public void characters(char[] ch, int start, int length)
    throws HTMLException {
    this.sw.write(ch, start, length);
    super.characters(ch, start, length);
  }

  public void startElement(IHTMLElementStart element) throws HTMLException {
    try {
      element.writeTo(this.sw);
    }
    catch (IOException ex) {
      throw new HTMLException(ex);
    }
    
    switch (this.state) {
      case S_OUTSIDE_FORM:
        if (element.getNameLength() == 4 
          && element.hasName("form") 
          && !element.isEmpty()
          && formStarts(element)) {
          this.state = S_INSIDE_FORM;
        }
        break;
      case S_INSIDE_FORM:
        if (element.getNameLength() == 5 && element.hasName("input")) {
          formInput(element);
        }
        break;
      default:
        break;
    }
    super.startElement(element);
  }

	public void endElement(IHTMLElement element) throws HTMLException {
    try {
      element.writeTo(this.sw);
    }
    catch (IOException ex) {
      throw new HTMLException(ex);
    }

    switch (this.state) {
      case S_OUTSIDE_FORM:
        break;
      case S_INSIDE_FORM:
        if (element.getNameLength() == 4 && element.hasName("form")) {
          formEnds();
          this.state = S_OUTSIDE_FORM;
        }
        break;
      default:
        break;
    }
		super.endElement(element);
	}


  //----------------------------------------------------------------------------
  
  private boolean formStarts(IHTMLElementStart element) {
    String id = null;
    IHierarchicalUri action = null;
    String method = "GET";
    String encType = HTMLForm.DEFAULT_ENC_TYPE;
    
    // Check if the form is interesting for us
    int index = element.getIndexOf("id");
    if (index < 0) {
      index = element.getIndexOf("name");
    }
    if (index >= 0) {
      id = element.getValueOf(index).toLowerCase();
    }
    
    if (this.formID != null && (id == null || !id.equalsIgnoreCase(this.formID))) {
      // not interested in this form
      return false;
    }
    
    index = element.getIndexOf("action");
    if (index >= 0) {
      try {
        IUriReference actionRef = UriFactory.parseUriReference(element.getValueOf(index));
        IUri uri = this.base.resolve(actionRef);
        if (uri instanceof IHierarchicalUri) {
          action = (IHierarchicalUri)uri;
        }
      }
      catch (MalformedURLException ex) {
        log.infoT("parsing action url: " + LoggingFormatter.extractCallstack(ex));
        //ignore
      }
    }
    if (action == null) {
      return false;
    }
    
    index = element.getIndexOf("method");
    if (index >= 0) {
      method = element.getValueOf(index).toUpperCase();
    }
    if (!"GET".equals(method) && !"POST".equals(method)) {
      return false;
    }
    
    index = element.getIndexOf("enctype");
    if (index >= 0) {
      encType = element.getValueOf(index).toLowerCase();
    }
    
    // Ok, we have a form with action uriref and valid id/method/enctype attributes
    this.current = HTMLForm.getInstance(id, action, method, encType);
            
    return true;
  }
  
  private void formEnds() {
    if (this.current.isValidLogin()) {
      this.found = this.current;
    }
    this.current = null;
  }
  
  private void formInput(IHTMLElementStart element) {
    if (this.current != null) {
      String type = HTMLInput.TEXT;
      int index = element.getIndexOf("type");
      if (index >= 0) {
        type = element.getValueOf(index).toLowerCase();
      }
      if (INTERESTING_INPUT_TYPES.contains(type)) {
        index = element.getIndexOf("name");
        if (index >= 0) {
          String name = element.getValueOf(index);
          String value = null;
          index = element.getIndexOf("value");
          if (index >= 0) {
            value = element.getValueOf(index);
          }
          this.current.addInput(name, type, value);
        }
      }
    }
  }
  
}

