package com.sapportals.wcm.protocol.webdav.repository;

import java.util.Properties;

import com.sap.security.api.ticket.TicketVerifier;
import com.sap.tc.logging.Location;
import com.sapportals.portal.security.usermanagement.IPortalLogin;
import com.sapportals.portal.security.usermanagement.IUser;
import com.sapportals.wcm.WcmException;
import com.sapportals.wcm.protocol.webdav.repository.landscape.IServerInfo;
import com.sapportals.wcm.protocol.webdav.repository.landscape.ServerInfoBroker;
import com.sapportals.wcm.repository.IOErrorException;
import com.sapportals.wcm.repository.IResourceContext;
import com.sapportals.wcm.repository.ResourceException;
import com.sapportals.wcm.util.http.*;
import com.sapportals.wcm.util.logging.LoggingFormatter;
import com.sapportals.wcm.util.uri.IHierarchicalUri;
import com.sapportals.wcm.util.uri.IUri;
import com.sapportals.wcm.util.uri.RID;
import com.sapportals.wcm.util.uri.URICodec;

/**
 * Manages http requester and user authorization against server.
 * <p>
 * Copyright (c) SAP AG 2002-2004
 * @author stefan.eissing@greenbytes.de
 * @version $Id: ServerRequestBroker.java,v 1.6 2004/07/28 11:24:40 sei Exp $
 */
public final class ServerRequestBroker {

  private static final Location log = Location.getLocation(ServerRequestBroker.class);

  /** class name of IWDRequesterFactory, property cache.requesterFactory */
  // private static final String PROP_REQUESTER_FACTORY = "cache.requesterFactory";

  /** Repository configuration parameter for cache timeout, cache.timeout */
  // private static final String PROP_CACHE_TIMEOUT = "cache.timeout";

  /** Repository configuration parameter for cache timeout, cache.timeout */
  private static final String PROP_HTTP_TIMEOUT = "http.timeout";

  private static final String PROP_PROXY_SYSTEM = "http.proxySystem";
  private static final String PROP_PROXY_HOST = "http.proxyHost";
  private static final String PROP_USE_DEFAULT_PROXY = "http.useDefaultProxy";

  private final static String TRANSIENT_USER_ATTRIBUTE_TICKET_STRING = "MYSAPSSO2_STRING";

  /**
   * Get a ServerRequestBroker for use against the server, given configuration properties
   * and base uri.
   * @param base uri of server
   * @param config configuration properties
   * @param server to work against (maybe <code>null</code>)
   * @return ServerRequestBroker
   * @throws ResourceException on configuration errors
   */
  public static ServerRequestBroker getInstance(IHierarchicalUri base, Properties config, IServerInfo server)
  throws ResourceException {
    return getBroker(base, config, server);
  }
  
  public static ServerRequestBroker getInstance(IHierarchicalUri base, ServerRequestBroker template)
  throws ResourceException {
    return getBroker(base, template);
  }

  private final IRequesterFactory factory;
  private final IServerInfo server;
  private final UserInfo defaultUI;
  private final ProxyInfo pi;
  private final IServerInfo proxyServer;
  private final UserInfo defaultProxyUI;
  private final boolean sameLoginForAllUsers;
  private final boolean sameProxyLoginForAllUsers;
  private final boolean sameUserDomain;

  private ServerRequestBroker(IRequesterFactory factory, IServerInfo server, ProxyInfo pi, IServerInfo proxyServer) {
    this.server = server;
    this.factory = factory;
    this.pi = pi;
    this.proxyServer = proxyServer;

    if (this.server != null) {
      this.defaultUI = this.server.getDefaultUserMapping();
      this.factory.setMaxConnections(this.server.getMaxHttpConnections());
    }
    else {
      this.defaultUI = UserInfo.ANONYMOUS;
    }

    if (this.proxyServer != null) {
      defaultProxyUI = proxyServer.getDefaultUserMapping();
    }
    else {
      defaultProxyUI = UserInfo.ANONYMOUS;
    }

    this.sameLoginForAllUsers = !UserInfo.ANONYMOUS.equals(this.defaultUI);
    this.sameProxyLoginForAllUsers = !UserInfo.ANONYMOUS.equals(this.defaultProxyUI);

    boolean sameDomain = false;
    if (this.server != null) {
      sameDomain = this.server.isSameUserDomain();
    }

    this.sameUserDomain = sameDomain;
  }

  /**
   * Get a IWDRequester suitable for this server and the given resource
   * context. Will handle user mapping for authentication.
   * @param context request should work in
   * @return suitable requester
   */
  public IRequester getRequester(IResourceContext context) throws ResourceException {
    try {
      return this.factory.getRequester(mapContext(context));
    }
    catch (ResourceException ex) {
      log.debugT("while getting requester" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(ex));
      throw ex;
    }
    catch (WcmException ex) {
      log.debugT("while getting requester" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(ex));
      throw new IOErrorException("while getting requester for " + this.getBase(), ex, (RID)null);
    }
  }

  /**
   * Get a IWDContext suitable for the given IResourceContext.
   * Or better: for the IUser of the IResourceContext.
   * @param rcontext with IUSer
   * @return suitable IWDContext
   */
  public IContext mapContext(IResourceContext rcontext) {
    IContext context = null;
    boolean isNew = false;
    if (this.sameLoginForAllUsers) {
      context = this.factory.getContext(this.defaultUI);
      if (context.isNew()) {
        isNew = true;
        context.setUserInfo(this.defaultUI);
      }
    }
    else if (rcontext != null) {
      IUser user = rcontext.getUser();
      if (user != null) {
        context = this.factory.getContext(user.getId());
        if (context.isNew()) {
          isNew = true;
          UserInfo mapped = UserInfo.ANONYMOUS;
          if (this.server != null) {
            mapped = this.server.getUserMappingFor(user);
            context.setUserInfo(mapped);

            if (log.beInfo()) {
              log.infoT("mapping user "+user.getId()+" to "+context.getUserInfo());
            }
          }

          setTicketHandler(rcontext, context, mapped);
        }
        else if ( ! this.sameUserDomain && this.server != null) {
          // Set again so we do not miss change of credentials
          context.setUserInfo(this.server.getUserMappingFor(user));
        }

        if (this.sameUserDomain) {
          ICredentials credentials = context.getSpecialCredentials();
          
          if (credentials == null) {
            if (log.beInfo()) {
              log.infoT("should not happen: old context in sameUserDomain has no specialCredentials: "+user.getId());
            }
            UserInfo mapped = UserInfo.ANONYMOUS;
            if (this.server != null) {
              mapped = this.server.getUserMappingFor(user);
            }
            setTicketHandler(rcontext, context, mapped);
          }
          else if (credentials instanceof TicketCredentials) {
            // We have to set the user object anew, the transient ticket data
            // is immutable so we will miss ticket changes otherwise
            ((TicketCredentials)credentials).update(user);
          }
        }
      }
    }

    if (this.proxyServer != null && isNew) {
      if (this.sameProxyLoginForAllUsers) {
        context.setProxyUserInfo(this.defaultProxyUI);
      }
      else {
        IUser user = rcontext.getUser();
        if (user != null) {
          context.setProxyUserInfo(this.proxyServer.getUserMappingFor(user));
        }
      }
    }

    return context;
  }

  /**
   * Return IServerInfo this broker works with.
   * @return IServerInfo or <code>null</code> if no such info is set.
   */
  public IServerInfo getServerInfo() {
    return this.server;
  }

  /**
   * Get the base uri handled by this broker
   * @return base uri handled by this broker
   */
  public IHierarchicalUri getBase() {
    return this.factory.getBase();  
  }
  
  //------------------------ private ------------------------------------------
  
  private static ServerRequestBroker getBroker(IHierarchicalUri base, ServerRequestBroker template)
  throws ResourceException {
    try {
      IServerInfo proxyServer = template.proxyServer;
      ProxyInfo pi = template.pi;

      IRequesterFactory factory = RequesterFactory.newInstance();
      factory.setBaseURL(base.toString());
      factory.setSoTimeout(template.factory.getSoTimeout());
      factory.setProxyInfo(pi);

      return new ServerRequestBroker(factory, null, pi, proxyServer);
    }
    catch (Exception ex) {
      log.errorT("cannot instantiate factory" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(ex));
      throw new ResourceException("cannot instantiate factory", ex);
    }
  }

  private static ServerRequestBroker getBroker(IHierarchicalUri base, Properties config, IServerInfo server)
  throws ResourceException {
    try {
      IRequesterFactory factory = RequesterFactory.newInstance();
      factory.setBaseURL(base.toString());
      ProxyInfo pi = getProxyInfo(config);

      String psystem = pi.getProxySystemID(base.getScheme());
      IServerInfo proxyServer = (psystem != null)? ServerInfoBroker.getInfoAbout(psystem) : null;
      if (proxyServer != null) {
        pi.setProxyUri(base, proxyServer.getUri());
      }

      factory.setProxyInfo(pi);

      if (log.beDebug()) {
        log.debugT("Server "+base+" has proxy "+proxyServer+" (id:"+psystem+")");
      }

      int httpTimeout = 0;
      String tmp = config.getProperty(PROP_HTTP_TIMEOUT);
      if (tmp != null && tmp.length() > 0) {
        try {
          httpTimeout = Integer.parseInt(tmp);
        }
        catch (NumberFormatException ex) {
          log.warningT("http.timeout not a number: " + tmp + " - " + LoggingFormatter.extractCallstack(ex));
          // ignore
        }
      }
      factory.setSoTimeout(httpTimeout);

      return new ServerRequestBroker(factory, server, pi, proxyServer);
    }
    catch (Exception ex) {
      log.errorT("cannot instantiate factory" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(ex));
      throw new ResourceException("cannot instantiate factory", ex);
    }
  }

  private static ProxyInfo getProxyInfo(Properties config) {
    // If we are using defaults and no host or system is defined,
    // initialize from system properties
    //
    String psystem = config.getProperty(PROP_PROXY_SYSTEM);
    String phost = config.getProperty(PROP_PROXY_HOST);
    boolean useDefaults = "true".equalsIgnoreCase(config.getProperty(PROP_USE_DEFAULT_PROXY));
    if (useDefaults
        && (psystem == null || psystem.length() == 0)
        && (phost == null || phost.length() == 0)) {
      return new ProxyInfo();
    }
    else {
      return new ProxyInfo(config);
    }
  }

  /**
   * Reset the login ticket as cookie value into context
   * @param rcontext resource context with user
   * @param context http context with cookies
   * @return if cookie was set
   */
  private boolean setTicketHandler(IResourceContext rcontext, IContext context, UserInfo mappedUser) {
    if (this.sameUserDomain) {
      IUser user = rcontext.getUser();
      if (UserInfo.ANONYMOUS.equals(context.getUserInfo())) {
        // We have no user information for the remote system. However
        // this context is using login tickets and thus, implicitly,
        // each context is different from the ANONYMOUS context.
        // So, we fake a user info to make the context look different.
        //
        // stefan.eissing@greenbytes.de: this introduced the following bug:
        // - when ticket authentication fails, the userinfo inside the context
        //   is *not* anonymous, tricking the http requester to send the given
        //   username with empty password as login.
        // - the login with empty password usually fails, and after n tries,
        //   the user account is locked.
        //
        if (log.beInfo()) {
          log.infoT("giving otherwise anonymous context a user id: "+user.getId());
        }
        context.setUserInfo(UserInfo.valueOf(user.getId(), ""));
      }

      boolean neverLogin = UserInfo.ANONYMOUS.equals(mappedUser);
      String ticket = getTicket(user);
      context.setSpecialCredentials(new TicketCredentials(user, context, ticket, neverLogin));
      if (ticket != null) {
        // It's the same user domain and we got a login ticket - use it.
        //
        ICookie cookie = MyCookie.valueOf(IPortalLogin.SSOTICKET_ALIAS, ticket);
        context.setCookie(cookie);
        if (log.beDebug()) {
          log.debugT("setting cookie in new context: "+cookie);
        }
        return true;
      }
    }
    return false;
  }

  private static final String getTicket(IUser user) {
    if (log.beInfo()) {
      log.infoT("getting ticket for user "+user);
    }

    if (user == null) {
      return null;
    }

    String  ticket = (String)user.getTransientAttribute(ServerRequestBroker.TRANSIENT_USER_ATTRIBUTE_TICKET_STRING);
    if (ticket == null) {
      // ep5.0
      try {
        TicketVerifier t = (TicketVerifier)user.getTransientAttribute(IPortalLogin.SSOTICKET_USER_ATTRIBUTE);
        if (t != null) {
          ticket = t.getTicket();
        }
      }
      catch (Exception e) {
        log.infoT("getting ticket: " + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(e));
      }
    }
    if (ticket != null) {
      ticket = URICodec.EncodePath(ticket);
    }

    if (log.beInfo()) {
      log.infoT("ticket lookup for user "+user.getId()+" returned: "+ticket);
    }

    return ticket;
  }

  private static final class MyCookie implements ICookie {

    private static ICookie valueOf(String key, String value) {
      return new MyCookie(key, value);
    }

    private final String key;
    private final String header;

    private MyCookie(String key, String value) {
      this.key = key;
      this.header = this.key + "=" + value;
    }

    public boolean isExpired() {
      return false;
    }

    public boolean appliesTo(IUri uri) {
      return true;
    }

    public String getKey() {
      return this.key;
    }

    public int getVersion() {
      return 0;
    }

    public String getHeaderValue() {
      return this.header;
    }

    public String toString() {
      return "MyCookie["+this.header+"]";
    }
  }

  private static final class TicketCredentials implements ICredentials {

    private IUser user;
    private final IContext context;
    private String lastTicket;
    private final boolean neverLogin;

    public TicketCredentials(IUser user, IContext context, String ticket, boolean neverLogin) {
      this.user = user;
      this.context = context;
      this.lastTicket = ticket;
      this.neverLogin = neverLogin;
    }

    public String getName() {
      return "LoginTickets";
    }

    public void update(IUser u) {
      setUser(u);
    }
    
    public void got(IRequester requester, IResponse response) throws WcmException {
      // nothing to do here
    }

    public boolean setup(IRequester requester, String header) throws WcmException {
      return setup(requester, header, false);
    }

    public boolean setup(IRequester requester, String header, boolean retry) throws WcmException {
      // If we do not want to user the given user credentials to ever login
      // (this.neverLogin == true), we always return true in this method even
      // if we have retries on the very same ticket string.
      // The reason is that we know the userinfo in the context to be invalid
      // and never want to use it for login requests with other authentication
      // schemes. That might lock the user account after n failures
      //
      String ticket = ServerRequestBroker.getTicket(getUser());
      if (ticket != null) {
        if (retry && ticket.equals(this.lastTicket)) {
          return this.neverLogin;
        }

        // First try or new ticket value
        //
        ICookie cookie = MyCookie.valueOf(IPortalLogin.SSOTICKET_ALIAS, ticket);
        this.context.setCookie(cookie);
        this.lastTicket = ticket;
        return true;
      }
      return this.neverLogin;
    }

    public boolean apply(IRequester requester, String uri, IRequest request, String headerName) throws WcmException {
      String ticket = ServerRequestBroker.getTicket(getUser());
      if (ticket != null) {
        if ( ! this.lastTicket.equals(ticket)) {
          ICookie cookie = MyCookie.valueOf(IPortalLogin.SSOTICKET_ALIAS, ticket);
          this.context.setCookie(cookie);
          this.lastTicket = ticket;
        }
        return true;
      }
      return false;
    }

    public void startUse(IRequester requester) {
      //nop
    }

    public void endUse(IRequester requester) {
      //nop
    }

    public boolean canTriggerAuthentication(IRequester requester) {
      return false;
    }
    
    private synchronized IUser getUser() {
      return this.user;
    }

    private synchronized void setUser(IUser user) {
      this.user = user;
    }
  }
}
