package com.sapportals.wcm.protocol.webdav.repository;

import java.text.ParseException;
import java.util.*;

import org.w3c.dom.*;

import com.sap.tc.logging.Location;
import com.sapportals.wcm.protocol.webdav.*;
import com.sapportals.wcm.repository.*;
import com.sapportals.wcm.repository.enum.PropertyType;
import com.sapportals.wcm.util.http.DateFormat;
import com.sapportals.wcm.util.logging.LoggingFormatter;
import com.sapportals.wcm.util.uri.*;

/**
 * Helper functions for WebDAV repository interaction regarding property
 * objects.
 *
 * <p>
 * Attributes on the property element are persisted by mapping from/to
 * repository property attributes. As the XML format used in WebDAV
 * allows namespaced property names, while the repository only allows
 * simple names, the following mapping is used:
 * <table>
 * <tr>
 *   <td><i>name</i>=<i>value</i></td><td>as is</td>
 * </tr>
 * <tr>
 *   <td>xml:<i>localName</i>=<i>value</i></td><td>as is</td>
 * </tr>
 * <tr>
 *   <td>xmlns:<i>prefix</i>=<i>value</i></td><td>as is</td>
 * </tr>
 * <tr>
 *   <td><i>prefix</i>:<i>localName</i>=<i>value</i></td>
 *   <td>{<i>namespaceUri</i>}<i>localName</i>=<i>value</i> <b>and</b> <br />
 *   xmlns:<i>prefix</i>=<i>namespaceUri</i></td>
 * </tr>
 * </table>
 *
 * </p>
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author stefan.eissing@greenbytes.de
 * @author julian.reschke@greenbytes.de
 * @version $Id: WDProperties.java,v 1.3 2004/04/19 12:48:06 jre Exp $
 */

public class WDProperties {

  private static final Location log = Location.getLocation(WDProperties.class);

  public static void getElementAttributes(Element elem, IMutableProperty property) throws ResourceException {

    NamedNodeMap nnm = elem.getAttributes();

    for (int i = 0; i < nnm.getLength(); i++) {

      Node attr = nnm.item(i);

      if (attr.getNodeType() == Node.ATTRIBUTE_NODE) {

        String ns = getNamespaceUri((Attr)attr);
        String val = attr.getNodeValue();

        //log.errorT("got attr ns '" + attr.getNamespaceURI() + "' prefix '" + attr.getPrefix() + "' name '" + attr.getNodeName() + "' localname '" + attr.getLocalName() + "'");

        if (ns == null || ns.equals("")) {
          property.setAttribute(attr.getNodeName(), val);
        }
        else if (ns.equals(Consts.NS_XML)) {
          property.setAttribute("xml:" + attr.getLocalName(), val);
        }
        else if (ns.equals(Consts.NS_XMLNS)) {
          // do not save the default namespace declaration
          if (!attr.getNodeName().equals("xmlns"))
            property.setAttribute(attr.getNodeName(), val);
        }
        else {
          property.setAttribute(new QualifiedName(ns, attr.getLocalName()).toString(),val);

          //          log.errorT("GET setting attribute: " + new QualifiedName (ns, attr.getLocalName()).toString() + "->" +
          //            attr.getNodeValue());

          // optionally store the namespace prefix in a separate attribute
          String prefix = attr.getPrefix();
          if (prefix != null && !prefix.equals("")) {
            property.setAttribute("xmlns:" + prefix, ns);
            // log.errorT("GET adding xmlns: " + "xmlns:" + prefix + "->" + ns);
          }
        }
      }
    }

    if (log.beDebug()) {
      Properties aprops = property.getAttributes();
      log.debugT("props are: " + ((aprops != null) ? aprops.toString() : "null"));
    }
  }

  public static void setElementAttributes(Element elem, IProperty property) throws ResourceException {
    Properties props = property.getAttributes();

    if (props != null) {
      Enumeration col = props.keys();
      Map namespaces = new HashMap();

      while (col.hasMoreElements()) {
        String key = (String)col.nextElement();
        if (key.startsWith("xmlns:")) {
          namespaces.put(props.getProperty(key), key.substring(6));
        }
      }

      col = props.keys();

      while (col.hasMoreElements()) {
        String key = (String)col.nextElement();
        String value = props.getProperty(key);

        //log.errorT("SET: found: " + key + "->" + value);
        //log.errorT("SET: translated to: " + qname.getNamespaceUri() + " " + qname.getLocalName());

        try {
          QualifiedName qname = QualifiedName.getInstance(key);

          String ns = qname.getNamespaceUri();

          if (ns == null || ns.equals("")) {
            // don't dare to touch a default xmlns if already there
            if (!(key.equals("xmlns")) || elem.getAttributeNode("xmnls") == null)
              elem.setAttribute(key, DomElement.makeSafeXml(value));
          }
          else if (Consts.NS_XML.equals(ns)) {
            elem.setAttribute("xml:" + qname.getLocalName(), DomElement.makeSafeXml(value));
          }
          else if (Consts.NS_XMLNS.equals(ns)) {
            // don't need to do anything. This is a namespace declaration
            // that we've already used by putting it into the HashMap of
            // prefix to namespaces mappings.
          }
          else if (XsdDataTypes.XN_TYPE.equals(key)) {
            // extended type information -- should have been processed earlier
          }
          else {
            String prefix = (String)namespaces.get(ns);
            // log.errorT("SET prefix for " + qname.toString() + " is " + prefix);
            if (prefix != null) {
              elem.setAttributeNS(qname.getNamespaceUri(), prefix + ":" + qname.getLocalName(), DomElement.makeSafeXml(value));
            }
            else {
              String newprefix = getPrefixFor(namespaces, ns);
              elem.setAttributeNS(qname.getNamespaceUri(), newprefix + ":" + qname.getLocalName(), DomElement.makeSafeXml(value));
            }
          }

        }
        catch (DOMException ex) {
          log.debugT("can't create attribute for " + key);
          elem.appendChild(
            elem.getOwnerDocument().createComment("unprocessed attribute: " + key + ":" + DomElement.makeSafeXml(value)));
        }
        catch (java.text.ParseException ex) {
          log.debugT("can't parse attribute value " + key + " - " + LoggingFormatter.extractCallstack(ex));
        }
      }
    }
  }

  private static String getPrefixFor(Map namespaces, String ns) {
    int i = 0;
    while (namespaces.get("ns" + i) != null) {
      i++;
    }
    String pref = "ns" + i;
    namespaces.put(ns, pref);
    return pref;
  }

  /**
   * Add a Property definition as child of an element.
   * @param parent where to append property element
   * @param pname name to use for property
   * @param prop the property to add
   * @param df date format object to use
   * @throws ResourceException when property is not valid
   */
  public static void appendProperty(Element parent, IPropertyName pname, IProperty prop, DateFormat df, IHierarchicalUri baseUri) throws ResourceException {

    Document doc = parent.getOwnerDocument();
    Element propElement = DomElement.makeNew(doc, pname);
    appendPropertyValue(propElement, prop, df, baseUri, true);
    parent.appendChild(propElement);
  }

  public static void appendPropertyValue(Element parent, IProperty prop, DateFormat df, boolean includeHiddenFlag)
    throws ResourceException {
    appendPropertyValue(parent, prop, df, null, includeHiddenFlag);
  }
  
  /**
   * Add a Property Value as child of an element.
   * @param parent where to append property value and set attributes on
   * @param prop the property to add
   * @param df date format object to use
   * @throws ResourceException when property is not valid
   */
  public static void appendPropertyValue(Element parent, IProperty prop, DateFormat df, IHierarchicalUri baseUri, boolean includeHiddenFlag)
    throws ResourceException {

    Document doc = parent.getOwnerDocument();
    PropertyType propType = prop.getType();

    QualifiedName qname = XsdDataTypes.mapSapToXsdType(propType);
    
    Properties attrs = prop.getAttributes();
    String internalType = attrs == null ? null : (String)attrs.get(XsdDataTypes.XN_TYPE);

    // if we have internal type information, and we know that type, use that one to report the datatype
    if (XsdDataTypes.XN_EXPANDED_NAME.equals(internalType)) {
      qname = XsdDataTypes.SAP_EXPANDED_NAME;
    }
    else if (XsdDataTypes.XN_RID.equals(internalType)) {
      qname = baseUri == null ? XsdDataTypes.SAP_RID : XsdDataTypes.SAP_MAPPED_RID_LIST;
    }

    if (prop.isMultivalued()) {

      List values = prop.getValues();
      DomElement.attachSoapArrayType(parent, qname, values.size());

      for (int i = 0; i < values.size(); i++) {

        if (propType.equals(PropertyType.BOOLEAN)) {
          Boolean val = (Boolean)values.get(i);
          parent.appendChild(DomElement.makeNew(doc, qname, val.equals(Boolean.TRUE) ? "true" : "false"));
        }
        else if (propType.equals(PropertyType.DATE)) {
          java.util.Date val = (java.util.Date)values.get(i);
          parent.appendChild(DomElement.makeNew(doc, qname, df.formatISO8601(val)));
        }
        else if (propType.equals(PropertyType.XML)) {
          try {
            Document doc2 = DomBuilder.parse(values.get(i).toString());
            Element container = DomElement.makeNew(doc, qname);

            for (Node n = doc2.getDocumentElement().getFirstChild(); n != null; n = n.getNextSibling()) {
              container.appendChild(DomElement.importNode(doc, n, true));
            }
            // log.errorT("adding: " + DomSerializer.toString (container, true));

            parent.appendChild(container);
          }
          catch (Exception ex) {
            log.errorT("Property parse error - " + LoggingFormatter.extractCallstack(ex));
          }
        }
        else if (PropertyType.STRING.equals(propType) && XsdDataTypes.XN_RID.equals(internalType)) {
          RID ref = RID.getRID(values.get(i).toString());
          if (ref.isAbsolute()) {
            if (baseUri != null) {
              parent.appendChild(DomElement.makeNew(doc, Consts.N_href, baseUri.mapToAbsolutePath(ref)));    
            }
            else {
              parent.appendChild(DomElement.makeNew(doc, XsdDataTypes.SAP_RID, ref.toString()));
            }
          }
          else {
            parent.appendChild(DomElement.makeNew(doc, Consts.N_href, ref.toExternalForm()));    
          }
        }
        else if (PropertyType.STRING.equals(propType) && XsdDataTypes.XN_EXPANDED_NAME.equals(internalType)) {
          String xn = values.get(i).toString();
          parent.appendChild(DomElement.makeNew(doc, XsdDataTypes.SAP_EXPANDED_NAME, xn));
        }
        else {
          parent.appendChild(DomElement.makeNew(doc, qname, DomElement.makeSafeXml(values.get(i).toString())));
        }
      }
    }
    else {

      if (!propType.equals(PropertyType.STRING) && !propType.equals(PropertyType.XML)) {
        if (qname != null) {
          DomElement.attachXsdDataType(parent, qname);
        }
      }
            
      if (PropertyType.DATE.equals(propType)) {
        parent.appendChild(doc.createTextNode(df.formatISO8601(prop.getDateValue())));
      }
      else if (PropertyType.XML.equals(propType)) {
        try {
          Document doc2 = DomBuilder.parse(prop.getValueAsString());

          for (Node n = doc2.getDocumentElement().getFirstChild(); n != null; n = n.getNextSibling()) {
            parent.appendChild(DomElement.importNode(doc, n, true));
          }
        }
        catch (Exception ex) {
          log.errorT(
            "error building DOM from XML property: " + prop.getPropertyName() + " - " + LoggingFormatter.extractCallstack(ex));
        }
      }
      else if (XsdDataTypes.XN_EXPANDED_NAME.equals(internalType)) {
        DomElement.attachXsdDataType(parent, XsdDataTypes.SAP_EXPANDED_NAME);
        DomElement.setUnsafeText(parent, prop.getValueAsString());
      }
      else if (XsdDataTypes.XN_RID.equals(internalType)) {
        RID ref = RID.getRID(prop.getValueAsString());
        if (ref.isAbsolute() && baseUri != null) {
          parent.appendChild(DomElement.makeNew(parent.getOwnerDocument(), Consts.N_href, baseUri.mapToAbsolutePath(ref)));    
          DomElement.attachXsdDataType(parent, XsdDataTypes.SAP_MAPPED_RID);
        }
        else if (ref.isAbsolute()) {
          DomElement.setUnsafeText(parent, ref.toExternalForm());
          DomElement.attachXsdDataType(parent, XsdDataTypes.SAP_RID);
        }
        else {
          parent.appendChild(DomElement.makeNew(parent.getOwnerDocument(), Consts.N_href, ref.toExternalForm()));    
          DomElement.attachXsdDataType(parent, XsdDataTypes.SAP_MAPPED_RID);
        }
      }
      else {
        DomElement.setUnsafeText(parent, prop.getValueAsString());
      }
    }

    WDProperties.setElementAttributes(parent, prop);
    
    if (includeHiddenFlag) {
      DomElement.attachPropertyFlag(parent, "hidden", prop.getPropertyDef().isHidden());
    }
  }

  /**
   * Create a Property from its XML/WebDAV serialization. Use the given
   * PropertName. The element is the property name element. This method tries
   * to extract type information according to
   * <a href="http://greenbytes.de/tech/webdav/draft-reschke-webdav-property-datatypes-latest.html">Datatypes for
   * WebDAV properties</a>.
   * @param pname the name to use in the property
   * @param e the element with the property name in WebDAV XML serialization
   * @param df date formatter
   * @param tolerateTypeError when set to <code>true</code>, tolerate errors
   * in parsing data types and fall back to simple types like string or XML.
   * @return created property or <code>null</code>
   */
  public static IMutableProperty getProperty(IPropertyName pname, Element e, DateFormat df, IHierarchicalUri baseUri, boolean tolerateTypeError)
    throws java.text.ParseException, ResourceException {
    PropertyType pt = null;
    boolean multiValued = false;

    e.normalize();

    QualifiedName xsdType = XsdDataTypes.getXsdTypeInfo(e);
    QualifiedName internalType = null;

    if (xsdType != null && xsdType.getNamespaceOrNull() == null) {

      if (!tolerateTypeError) {
        log.debugT("cannot resolve datatype prefix prefix '" + xsdType.getPrefix() + "'");
        throw new java.text.ParseException("cannot resolve datatype prefix prefix '" + xsdType.getPrefix() + "'", 0);
      }
      else {
        xsdType = null;
      }
    }

    if (xsdType != null) {
      Attr soapArrayType = null;

      //log.debugT("WDPropertyMap: set - found datatype " + xsdType + " for element" + new Name (e));

      if (XsdDataTypes.SOAP_ARRAY.equals(xsdType)
        && (soapArrayType = e.getAttributeNodeNS(Consts.NS_SOAP_ENC, "arrayType")) != null) {

        // special case: this is multi-valued. Need to extract real data type
        // from soap-enc:arrayType attribute

        String dt = soapArrayType.getNodeValue();
        int pos = dt.indexOf("[");
        if (pos != -1) {
          dt = dt.substring(0, pos);
        }

        //log.debugT("getting data type for: " + dt);

        QualifiedName dataType = XsdDataTypes.getXsdTypeInfoFromValue(e, dt);

        if (dataType.getNamespaceOrNull() == null) {
          log.debugT("cannot resolve datatype prefix prefix '" + dataType.getPrefix() + "'");
          throw new java.text.ParseException("cannot resolve datatype prefix prefix '" + dataType.getPrefix() + "'", 0);
        }

        //log.debugT("-> " + dataType);
        pt = XsdDataTypes.mapXsdToSapTypeOrNull(dataType);

        // If property type is unknown, we can't persist it as multivalued
        // and will have to fall back to XMLMarkup
        if (pt != null) {
          multiValued = true;
          internalType = dataType;
        }
        else {
          pt = PropertyType.XML;
        }
      }
      else {
        internalType = xsdType;
        pt = XsdDataTypes.mapXsdToSapType(xsdType);
        
        if (XsdDataTypes.SAP_MAPPED_RID_LIST.equals(internalType)) {
          multiValued = true;
        }
      }
    }
    else {
      if (DomElement.hasOnlyTextContent(e) || XsdDataTypes.SAP_MAPPED_RID.equals(internalType)) {
        pt = PropertyType.STRING;
      }
      else {
        pt = PropertyType.XML;
      }
    }

    if (multiValued) {
      return internalGetPropertyMultiValued(pname, pt, e, df, baseUri, internalType);
    }
    else {
      return internalGetProperty(pname, pt, e, df, baseUri, internalType);
    }
  }

  /**
   * Create a Property from its XML/WebDAV serialization. Use the given
   * PropertyName and type. The element is the property name element.
   * @param pname the name to use in the property
   * @param pt the property type to use
   * @param e the element with the property name in WebDAV XML serialization
   * @param df date formatter
   * @return created property or <code>null</code>
   */
  public static IMutableProperty getProperty(IPropertyName pname, PropertyType pt, Element e, DateFormat df)
    throws java.text.ParseException, ResourceException {
    return internalGetProperty(pname, pt, e, df, null, null);
  }
    
  private static IMutableProperty internalGetProperty(IPropertyName pname, PropertyType pt, Element e, DateFormat df,
    IHierarchicalUri baseUri, QualifiedName internalType)
    throws java.text.ParseException, ResourceException {

    String text = null;
    IMutableProperty prop = null;
    Boolean isHidden = null;
    Boolean isProtected = null;

    Attr a = e.getAttributeNodeNS(Consts.NS_SAP_PROPERTYFLAGS, "hidden");
    if (a != null) {
      isHidden = XsdDataTypes.parseBoolean(a.getValue()) ? Boolean.TRUE : Boolean.FALSE;
    }

    a = e.getAttributeNodeNS(Consts.NS_SAP_PROPERTYFLAGS, "protected");
    if (a != null) {
      isProtected = XsdDataTypes.parseBoolean(a.getValue()) ? Boolean.TRUE : Boolean.FALSE;
    }

    // take system defaults if the flags weren't sent
    if (isProtected == null || isHidden == null) {

      WDPropertyDef wpdf = WDPropertyDef.mapInverse(pname);
      if (wpdf != null) {
        if (isHidden == null) {
          isHidden = wpdf.isHidden() ? Boolean.TRUE : Boolean.FALSE;
        }
        if (isProtected == null) {
          isProtected = wpdf.isProtected() ? Boolean.TRUE : Boolean.FALSE;
        }
      }
    }

    PropertyDef pd =
      PropertyDef.valueOf(
        pt,
        false,
        isProtected != null ? isProtected.booleanValue() : false,
        isHidden != null ? isHidden.booleanValue() : false);

    if (PropertyType.BOOLEAN.equals(pt)) {
      text = DomElement.getText(e);
      if (text.length() == 0)
        return null;
      prop = new MutableProperty(pname, pd, getBoolean(text));
    }
    else if (PropertyType.DATE.equals(pt)) {
      text = DomElement.getText(e);
      if (text.length() == 0)
        return null;
      prop = new MutableProperty(pname, pd, getDate(text, df));
    }
    else if (PropertyType.INTEGER.equals(pt)) {
      text = DomElement.getText(e);
      if (text.length() == 0)
        return null;
      prop = new MutableProperty(pname, pd, getInteger(text));
    }
    else if (PropertyType.LONG.equals(pt)) {
      text = DomElement.getText(e);
      if (text.length() == 0)
        return null;
      prop = new MutableProperty(pname, pd, getLong(text));
    }
    else if (PropertyType.STRING.equals(pt)) {
      if (XsdDataTypes.SAP_MAPPED_RID.equals(internalType)) {
        for (Node n = e.getFirstChild(); n != null; n = n.getNextSibling()) {
          if (Consts.N_href.equalsNode(n)) {
            if (prop != null) {
              throw new ParseException("invalid mapped_rid content", 0);
            }
            text = DomElement.getText(n);
            prop = new MutableProperty(pname, pd, uriReferenceToRID(baseUri, text).toString()); 
          }
        }
      }
      else {
        text = DomElement.getText(e);
        prop = new MutableProperty(pname, pd, text);
      }
    }
    else if (PropertyType.XML.equals(pt)) {
      prop = new MutableProperty(pname, pd, getXMLMarkup(e));
    }
    else {
      log.fatalT("unknown property type: " + pt);
    }

    if (prop != null) {
      removeDataTypeAttributes(e, false);
      removePropertyFlagAttributes(e);
      WDProperties.getElementAttributes(e, prop);

      if (XsdDataTypes.SAP_EXPANDED_NAME.equals(internalType)) {
        prop.setAttribute(XsdDataTypes.XN_TYPE, XsdDataTypes.XN_EXPANDED_NAME);
      }
      else if (XsdDataTypes.SAP_MAPPED_RID.equals(internalType)) {
        prop.setAttribute(XsdDataTypes.XN_TYPE, XsdDataTypes.XN_RID);
      }
      else if (XsdDataTypes.SAP_RID.equals(internalType)) {
        prop.setAttribute(XsdDataTypes.XN_TYPE, XsdDataTypes.XN_RID);
      }
    }

    return prop;
  }

  //---------------------------------- private ------------------------------------

  private static Boolean getBoolean(String s) throws java.text.ParseException {
    return XsdDataTypes.parseBoolean(s.trim()) ? Boolean.TRUE : Boolean.FALSE;
  }

  private static Integer getInteger(String s) throws java.text.ParseException {
    try {
      return new Integer(s.trim());
    }
    catch (java.lang.NumberFormatException ex) {
      throw new java.text.ParseException(s, 0);
    }
  }

  private static Long getLong(String s) throws java.text.ParseException {
    try {
      return new Long(s.trim());
    }
    catch (java.lang.NumberFormatException ex) {
      throw new java.text.ParseException(s, 0);
    }
  }

  private static Date getDate(String s, DateFormat df) throws java.text.ParseException {
    String text = s.trim();

    if (text.length() > 0) {
      if (Character.isDigit(text.charAt(0))) {
        try {
          return df.parseISO8601(text);
        }
        catch (java.text.ParseException ex) {
          log.warningT("not in ISO format - " + LoggingFormatter.extractCallstack(ex));
          return df.parseHTTP(text);
        }
      }
      else {
        try {
          return df.parseHTTP(text);
        }
        catch (java.text.ParseException ex) {
          //$JL-EXC$
          return df.parseISO8601(text);
        }
      }
    }
    return null;
  }

  private static XMLMarkup getXMLMarkup(Element e) throws java.text.ParseException {
    return new XMLMarkup(DomSerializer.toString(e, false));
  }

  private static IMutableProperty internalGetPropertyMultiValued(IPropertyName pname, PropertyType pt, Element parent, DateFormat df,
    IHierarchicalUri baseUri, QualifiedName internalType)
    throws java.text.ParseException, ResourceException {

    IMutableProperty prop = null;
    List values = new ArrayList();
    String text;

    Boolean isHidden = null;
    Boolean isProtected = null;

    Attr a = parent.getAttributeNodeNS(Consts.NS_SAP_PROPERTYFLAGS, "hidden");
    if (a != null) {
      isHidden = XsdDataTypes.parseBoolean(a.getValue()) ? Boolean.TRUE : Boolean.FALSE;
    }

    a = parent.getAttributeNodeNS(Consts.NS_SAP_PROPERTYFLAGS, "protected");
    if (a != null) {
      isProtected = XsdDataTypes.parseBoolean(a.getValue()) ? Boolean.TRUE : Boolean.FALSE;
    }

    // take system defaults if the flags weren't sent
    if (isProtected == null || isHidden == null) {

      WDPropertyDef wpdf = WDPropertyDef.mapInverse(pname);
      if (wpdf != null) {
        if (isHidden == null) {
          isHidden = wpdf.isHidden() ? Boolean.TRUE : Boolean.FALSE;
        }
        if (isProtected == null) {
          isProtected = wpdf.isProtected() ? Boolean.TRUE : Boolean.FALSE;
        }
      }
    }

    PropertyDef pd =
      PropertyDef.valueOf(
        pt,
        true,
        isProtected != null ? isProtected.booleanValue() : false,
        isHidden != null ? isHidden.booleanValue() : false);

    for (Node e = parent.getFirstChild(); e != null; e = e.getNextSibling()) {

      if (e.getNodeType() != Node.ELEMENT_NODE)
        continue;

      if (PropertyType.BOOLEAN.equals(pt)) {
        text = DomElement.getText(e);
        values.add(getBoolean(text));
      }
      else if (PropertyType.DATE.equals(pt)) {
        text = DomElement.getText(e);
        values.add(getDate(text, df));
      }
      else if (PropertyType.INTEGER.equals(pt)) {
        text = DomElement.getText(e);
        values.add(getInteger(text));
      }
      else if (PropertyType.LONG.equals(pt)) {
        text = DomElement.getText(e);
        values.add(getLong(text));
      }
      else if (PropertyType.STRING.equals(pt)) {
        text = DomElement.getText(e);
        if (XsdDataTypes.SAP_MAPPED_RID_LIST.equals(internalType)) {
          values.add(uriReferenceToRID(baseUri, text).toString());
        }
        else {
          values.add(text);
        }
      }
      else if (PropertyType.XML.equals(pt)) {
        values.add(getXMLMarkup((Element)e));
      }
      else {
        log.fatalT("unknown property type: " + pt);
      }
    }

    prop = new MutableProperty(pname, pd, values);
    removeDataTypeAttributes(parent, true);
    removePropertyFlagAttributes(parent);
    WDProperties.getElementAttributes(parent, prop);

    if (XsdDataTypes.SAP_EXPANDED_NAME.equals(internalType)) {
      prop.setAttribute(XsdDataTypes.XN_TYPE, XsdDataTypes.XN_EXPANDED_NAME);
    }
    else if (XsdDataTypes.SAP_MAPPED_RID_LIST.equals(internalType)) {
      prop.setAttribute(XsdDataTypes.XN_TYPE, XsdDataTypes.XN_RID);
    }

    return prop;
  }

  // workaround for DOM L2 implementation bugs
  private static String getNamespaceUri(Attr a) {
    String ns = a.getNamespaceURI();
    if (ns == null || ns.equals("")) {
      String nodeName = a.getNodeName();
      if ("xmlns".equals(nodeName) || nodeName.startsWith("xmlns:")) {
        ns = Consts.NS_XMLNS;
      }
    }
    return ns;
  }

  private static void removeDataTypeAttributes(Element e, boolean multiValued) {

    e.removeAttributeNS(Consts.NS_XSI, "type");
    NamedNodeMap attrs = e.getAttributes();
    for (int j = attrs.getLength() - 1; j >= 0; j--) {
      Attr a = (Attr)attrs.item(j);
      String ns = getNamespaceUri(a);
      if (Consts.NS_XMLNS.equals(ns) && (a.getNodeValue().equals(Consts.NS_XSD) || a.getNodeValue().equals(Consts.NS_XSI))) {
        e.removeAttributeNode(a);
      }
    }

    // remove SOAP array encoding info when needed
    if (multiValued) {
      e.removeAttributeNS(Consts.NS_SOAP_ENC, "arrayType");
      attrs = e.getAttributes();
      for (int j = attrs.getLength() - 1; j >= 0; j--) {
        Attr a = (Attr)attrs.item(j);
        String ns = getNamespaceUri(a);

        if (Consts.NS_XMLNS.equals(ns) && a.getNodeValue().equals(Consts.NS_SOAP_ENC)) {
          e.removeAttributeNode(a);
        }
      }
    }
  }

  private static void removePropertyFlagAttributes(Element e) {

    e.removeAttributeNS(Consts.NS_SAP_PROPERTYFLAGS, "hidden");
    e.removeAttributeNS(Consts.NS_SAP_PROPERTYFLAGS, "protected");

    NamedNodeMap attrs = e.getAttributes();
    for (int j = attrs.getLength() - 1; j >= 0; j--) {
      Attr a = (Attr)attrs.item(j);
      String ns = getNamespaceUri(a);
      if (Consts.NS_XMLNS.equals(ns) && (a.getNodeValue().equals(Consts.NS_SAP_PROPERTYFLAGS))) {
        e.removeAttributeNode(a);
      }
    }
  }


  private static RID uriReferenceToRID(IHierarchicalUri baseUri, String uriref) throws ParseException {
    try {
      IUriReference ref = UriFactory.parseUriReference(uriref);
      String path = ref.getPath();
      String query = ref.getQuery();
      RID mapped = ref.isAbsolute()
        ? baseUri.mapToResourceID(ref)
        : RID.getRID(path == null ? "" : URICodec.Decode(ref.getPath()), query == null ? null : URICodec.Decode(ref.getQuery()));
              
      if (mapped == null) {
        throw new ParseException("can't map relative URI", 0); 
      }
      
      return mapped;
    }
    catch (java.net.MalformedURLException ex) {
      log.fatalT(LoggingFormatter.extractCallstack(ex));
      throw new ParseException(ex.getMessage(), 0);
    }    
  }

}