package com.sapportals.wcm.protocol.webdav.repository;

import java.util.Date;
import java.util.Set;

import com.sapportals.portal.security.usermanagement.IUMPrincipal;
import com.sapportals.wcm.protocol.webdav.repository.acl.IPrivilege;
import com.sapportals.wcm.protocol.webdav.repository.acl.WDAcl;
import com.sapportals.wcm.repository.*;
import com.sapportals.wcm.repository.enum.LinkType;
import com.sapportals.wcm.repository.enum.OrderType;
import com.sapportals.wcm.util.content.IContent;
import com.sapportals.wcm.util.uri.IUri;
import com.sapportals.wcm.util.uri.RID;
import com.sapportals.wcm.util.uri.URL;

/**
 * A resource info object serves as a front-end for obtaining information
 * about a {@link com.sapportals.wcm.repository.IResource} object. The resource info
 * keeps the obtained information and never asks the repository twice.
 *
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author stefan.eissing@greenbytes.de
 * @author julian.reschke@greenbytes.de
 * @version $Id: WDResourceInfo.java,v 1.4 2004/06/18 13:07:01 jre Exp $
 */
public class WDResourceInfo {

//  private static final com.sapportals.wcm.util.log.Category log =
//    com.sapportals.wcm.util.log.Category.getInstance (WDResourceInfo.class.getName());

  private final IResource resource;

  private ISupportedOptionSet supportedOptions;

  private Date creationDate;
  private Date lastModified;

  private Object properties;

  private RID rid;
  private String description;
  private String displayName;
  private RID owner;
  private String ownerDisplayname;
  private LinkType linkType;
  private URL targetURL;
  private String revisionID;
  private IResourceList versionResources;

  private WDAcl acl;
  private IPrivilege supportedPrivilege;
  private Set currentUserPrivileges;

  private IContent content;

  private Boolean readOnly;
  private Boolean isHidden;
  private Boolean isVersioned;
  private Boolean isRevision;
  private Boolean isCheckedOut;

  public WDResourceInfo(IResource resource) {
    this(resource, null);
  }

  public WDResourceInfo(IResource resource, IResourceList versionResources) {
    this.resource = resource;
    this.versionResources = versionResources;
  }

  /**
   * Returns a collection of supported operations for the repository the resource
   * belongs to.
   * @return The set of supported operations
   */
  public ISupportedOptionSet getSupportedOptions() {
    if (this.supportedOptions == null) {
      this.supportedOptions = this.resource.getSupportedOptions();
    }
    return this.supportedOptions;
  }

  public WDAcl getAcl() throws ResourceException {
    if (this.acl == null) {
      this.acl = new WDAcl(this.resource);
    }
    return this.acl;
  }

  /**
   * Get name of the resource. The resource name ist the last part if its URI
   * path.
   * @return The name of the resource
   */
  public String getName() throws ResourceException {
    return this.resource.getName();
  }

  /**
   * Get value of the system property CreationDate
   * @return Property value, can be null if property is not supported by the repository.
   */
  public Date getCreationDate() throws ResourceException {
    if (this.creationDate == null) {
      this.creationDate = getDate(WDPropertyDef.creationdate.getInternalName(), null);
    }
    return this.creationDate;
  }

  /**
   * Get value of the system property LastModified
   * @return Property value, can be null if property is not supported by the repository.
   */
  public Date getLastModified() throws ResourceException {
    if (this.lastModified == null) {
      this.lastModified = getDate(WDPropertyDef.getlastmodified.getInternalName(), null);
      //this.lastModified = this.resource.getLastModified();
    }
    return this.lastModified;
  }

  /**
   * Get value of the system property Description
   * @return Property value, can be null if property is not supported by the repository.
   */
  public String getDescription() throws ResourceException {
    if (this.description == null) {
      this.description = getString(PropertyName.createDescription(), null);
      //this.description = this.resource.getDescription();
    }
    return this.description;
  }

  /**
   * Get value of the system property Displayname
   * @return Property value, can be null if property is not supported by the repository.
   */
  public String getDisplayName() throws ResourceException {
    if (this.displayName == null) {
      this.displayName = getString(WDPropertyDef.displayname.getInternalName(), null);
      //this.displayName = this.resource.getDisplayName();
    }
    return this.displayName;
  }


  private IUri orderingType = null;

  public IUri getOrderingType() throws ResourceException {

    if (orderingType == null) {

      IOrderableCollection oc = (IOrderableCollection)this.resource.as(IOrderableCollection.class);
      if (oc != null) {
        orderingType = oc.getOrderingType();
      }
      else {
        ICollectionVersionResource cv = (ICollectionVersionResource)this.resource.as(ICollectionVersionResource.class);
        if (cv != null) {
          orderingType = cv.getOrderingType();
        }
        else if (this.resource.isCollection()) {
          return ((ICollection)this.resource).getOrderType().equals(OrderType.MANUAL)
            ? com.sapportals.wcm.protocol.webdav.Consts.ORDERING_TYPE_CUSTOM
            : com.sapportals.wcm.protocol.webdav.Consts.ORDERING_TYPE_UNORDERED;
        }
        else {
          return com.sapportals.wcm.protocol.webdav.Consts.ORDERING_TYPE_UNORDERED;
        }
      }
    }

    return orderingType;
  }

  /**
   * @return Property value, can be null if property is not supported by the repository.
   */
  public RID getOwner() throws ResourceException {

    if (this.owner == null) {
      IUMPrincipal principal = getAcl().getOwner();
      if (principal != null) {
        this.owner = WDPrincipals.getRid(principal);
      }
    }
    return this.owner;
  }

  /**
   * @return Property value, can be null if property is not supported by the repository.
   */
  public String getOwnerDisplayname() throws ResourceException {

    if (this.ownerDisplayname == null) {
      IUMPrincipal principal = getAcl().getOwner();
      if (principal != null) {
        this.ownerDisplayname = principal.getDisplayName();
      }
    }
    return this.ownerDisplayname;
  }

  /**
   * Get the RID of this resource
   * @return The RID
   */
  public RID getRID() throws ResourceException {
    if (this.rid == null) {
      this.rid = this.resource.getAccessRID();
      if (this.rid.length() == 0) {
        this.rid = this.resource.getRID();
      }
    }
    return this.rid;
  }

  /**
   * Get value of the system property Readonly
   * @return Property value
   */
  public boolean isReadOnly() throws ResourceException {
    if (this.readOnly == null) {
      this.readOnly = new Boolean(getBoolean(PropertyName.createReadOnly(), false));
      //this.readOnly = new Boolean(this.resource.isReadOnly());
    }
    return this.readOnly.booleanValue();
  }

  /**
   * Get value of the system property Hidden
   * @return Property value
   */
  public boolean isHidden() throws ResourceException {
    if (this.isHidden == null) {
      this.isHidden = getBoolean(WDPropertyDef.ishidden.getInternalName(), false) ? Boolean.TRUE : Boolean.FALSE;
      //this.isHidden = new Boolean(this.resource.isHidden());
    }
    return this.isHidden.booleanValue();
  }

  /**
   * Returns true if the resource is a collection (or an internal link to a collection)
   * @return Boolean value
   */
  public boolean isCollection() {
    return this.resource.isCollection();
  }

  /**
   * Returns true if the resource is a "real" collection (and not possibly a link to a collection)
   * @return Boolean value
   */
  public boolean isRealCollection() {
    try {
      return this.resource.isCollection() && (! LinkType.INTERNAL.equals(this.getLinkType()));
    }
    catch (ResourceException ex) {
      //$JL-EXC$
      // this can occur in getLinkType in which case we'll conclude that the resource isn't a link
      return this.resource.isCollection();
    }
  }

  /**
   * Returns the content access object for this resource.
   * @return The reference to the content object
   * @exception AccessDeniedException If the resource is locked, readonly or authorization failed
   */
  public IContent getContent(boolean unfiltered) throws ResourceException, AccessDeniedException {
    if (this.content == null) {
      // FIXME workaround: claim there's no content for LINK resources (otherwise
      // the repository will resolve it
      // JRE: not anymore. 2001-09-18
      if (isRealCollection() || getLinkType().equals (LinkType.EXTERNAL) || getLinkType().equals(LinkType.INTERNAL))
        return null;
      else {
        this.content = unfiltered?
          this.resource.getUnfilteredContent() :  this.resource.getContent();
      }
    }
    return this.content;
  }

  /**
   * Returns a map with <code>IProperty</code> references for the resource.
   * @return The property map
   * @exception AccessDeniedException If the resource is locked, readonly or authorization failed
   */
  public IPropertyMap getProperties() throws ResourceException {
    if (this.properties == null) {
      try {
        this.properties = this.resource.getProperties();
      }
      catch (ResourceException ex) {
        this.properties = ex;
        throw ex;
      }
    }
    
    if (this.properties instanceof IPropertyMap) {
      return (IPropertyMap)this.properties;
    }
    else {
      throw (ResourceException)this.properties;
    }
  }

  public IProperty getProperty(IPropertyName pname) throws ResourceException {
    return this.getProperties().get(pname);
  }

  /**
   * Get the target URL/URI of a link. Returns null if the resource is not a link.
   * @return The URI
   */
  public String getTargetURL() throws ResourceException {
    if (this.targetURL == null) {
      this.targetURL = this.resource.getTargetURL();
    }
    return (this.targetURL != null)? this.targetURL.toString() : null;
  }

  /**
   * Returns the link type (NONE if the resource is not a link)
   * @return Boolean value
   */
  public LinkType getLinkType() throws ResourceException {
    if (this.linkType == null) {
      this.linkType = this.resource.getLinkType();
      if (this.linkType == null) {
        this.linkType = LinkType.NONE;
      }
    }
    return this.linkType;
  }

  /**
   * Returns true if the resource supports versioning. This is the case if the
   * repository manager implements the <code>IVersioningManager</code> interface.
   * @return Boolean value
   */
  public boolean isVersioned() throws ResourceException {
    if (this.isVersioned == null) {
      this.isVersioned = new Boolean(this.resource.isVersioned());
    }
    return this.isVersioned.booleanValue();
  }

  /**
   * Returns true if this resource instance represents a revision.
   * @return Boolean value
   */
  public boolean isRevision() {
    if (this.isRevision == null) {
      this.isRevision = new Boolean(this.resource.isRevision());
    }
    return this.isRevision.booleanValue();
  }

  public IResource getResource() {
    return this.resource;
  }

  /**
   * Returns a identifier for a revision. In most repositories this will be
   * a version number.
   * @return The revision identifier
   */
  public String getRevisionID() {
    if (this.revisionID == null) {
      this.revisionID = this.resource.getRevisionID();
    }
    return this.revisionID;
  }

  /**
   * Checks whether a resource is currently checked out.
   * @return Boolean value
   */
  public boolean isCheckedOut() throws ResourceException {
    if (this.isCheckedOut == null) {
      this.isCheckedOut = new Boolean(this.resource.isCheckedOut());
    }
    return this.isCheckedOut.booleanValue();
  }

  /**
   * Get the version history for the resource.
   * @return the version history
   */
  public IResourceList getVersionResources() throws ResourceException, NotSupportedException {
    if (this.versionResources == null) {

      IVersionControlledResource vcr = asVersionControlledResource();
      IVersionHistoryResource vhr = vcr == null ? null : (IVersionHistoryResource)vcr.getVersionHistoryResource().as(IVersionHistoryResource.class);

      // try again using IVersionResource interface
      if (vhr == null) {
        IVersionResource vr = asVersionResource();
        if (vr != null) {
          vhr = (IVersionHistoryResource)vr.getVersionHistoryResource().as(IVersionHistoryResource.class);
        }
      }

      if (vhr == null) {
        // give up -- use old style interface
        this.versionResources = this.resource.getVersionHistory();
      }
      else {
        this.versionResources = vhr.getVersionResourceSet();
      }
    }
    return this.versionResources;
  }

  public boolean hasAclSupport() throws ResourceException {
    return getAcl().hasManager();
  }

  public IPrivilege getSupportedPrivileges() throws ResourceException {
    if (this.supportedPrivilege == null) {
      this.supportedPrivilege = getAcl().getSupported();
    }
    return this.supportedPrivilege;
  }

  public Set getCurrentUserPrivileges() throws ResourceException {
    if (this.currentUserPrivileges == null) {
      this.currentUserPrivileges = getAcl().getCurrentUserSet();
    }
    return this.currentUserPrivileges;
  }


  public boolean isPrincipal() throws ResourceException {

    return getResource().isA(IUMPrincipal.class);
  }

  public IUMPrincipal asPrincipal() throws ResourceException {

    return (IUMPrincipal)getResource().as(IUMPrincipal.class);
  }

  public IVersionControlledResource asVersionControlledResource()
    throws ResourceException {

    return (IVersionControlledResource)getResource().as(
      IVersionControlledResource.class);
  }

  public IVersionHistoryResource asVersionHistoryResource()
    throws ResourceException {

    return (IVersionHistoryResource)getResource().as(
      IVersionHistoryResource.class);
  }

  public IVersionResource asVersionResource() throws ResourceException {

    return (IVersionResource)getResource().as(IVersionResource.class);
  }

  public IBindableResource asBindableResource() throws ResourceException {

    return (IBindableResource)getResource().as(IBindableResource.class);
  }
  
  public ICollectionVersionResource asCollectionVersionResource() throws ResourceException {

    return (ICollectionVersionResource)getResource().as(ICollectionVersionResource.class);
  }

  public boolean isVhr() throws ResourceException {

    return getResource().isA(
      com.sapportals.wcm.repository.IVersionHistoryResource.class);
  }


  // ---------------- protected / private ----------------------------

  private boolean getBoolean(IPropertyName pname, boolean defVal) throws ResourceException {
    IPropertyMap pm = getProperties();
    IProperty prop = pm.get(pname);
    return prop != null? prop.getBooleanValue() : defVal;
  }

  private String getString(IPropertyName pname, String defVal) throws ResourceException {
    IPropertyMap pm = getProperties();
    IProperty prop = pm.get(pname);
    return prop != null? prop.getValueAsString() : defVal;
  }

  private Date getDate(IPropertyName pname, Date defVal) throws ResourceException {
    IPropertyMap pm = getProperties();
    IProperty prop = pm.get(pname);
    return prop != null? prop.getDateValue() : defVal;
  }

}