package com.sapportals.wcm.protocol.webdav.repository.acl;

import java.util.Set;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.sapportals.portal.security.usermanagement.IUMPrincipal;
import com.sapportals.wcm.util.uri.IHierarchicalUri;

/**
 * Represents a single grant or deny of a set of {@link com.sapportals.wcm.protocol.webdav.repository.acl.IPrivilege}s,
 * a Access Control Entry (part of an Access Control List, see
 * {@link com.sapportals.wcm.protocol.webdav.repository.acl.WDAcl}).
 * <p>
 * A IAce has the following properties:
 * <ul>
 * <li>a principal: for whom privileges are granted or denied.</li>
 * <li>a type: GRANT or DENY</li>
 * <li>readonly: indicates that this ACE is not changeable for a resource</li>
 * <li>inherited: indicates that this ACE was inherited from a parent resource</li>
 * <li>privileges: a list of {@link com.sapportals.wcm.protocol.webdav.repository.acl.IPrivilege}s</li>
 * </ul>
 * <p>
 * Copyright (c) SAP AG 2001-2003
 * @author stefan.eissing@greenbytes.de
 * @version $Revision: 1.1.1.1 $
 */
public interface IAce {

  /** Type of ACE is granting privileges to principal */
  public static final int GRANT = 1;
  /** Type of ACE is denying privileges to principal */
  public static final int DENY = 0;

  /** Returns the principal whose access is controlled
   * @return the principal in this ACE
   */
  public IUMPrincipal getPrincipal();

  /** Returns the List of privileges which are granted or denied
   * @return the list of IPrivileges in this ACE
   */
  public Set getPrivileges();

  /** Returns if this entry is GRANT or DENY
   * @return the type of this ACE
   */
  public int getType();

  /** Determine if privilege is granted by this ace
   * @param privilege to check for
   * @return if privileges is granted
   */
  public boolean grants(IPrivilege privilege);

  /** Returns if this entry is GRANT
   * @return if this ACE is granting privileges
   */
  public boolean isGrant();

  /** Returns if this entry was inherited from another resource
   * @return if this ACE was inherited
   */
  public boolean isInherited();

  /** Returns if this entry is protected, e.g. cannot be changed.
   * @return if this ACE is read only
   */
  public boolean isReadOnly();

  /**
   * Return a DOM Element representing this ACE as defined in the
   * WebDAV Accesss Control Protocol Specification
   * @param doc the DOM Document to create the Element in
   * @param servletUri base URI of servlet
   * @return the created Element
   */
  public Element toElement(Document doc, IHierarchicalUri servletUri);
}