package com.sapportals.wcm.protocol.webdav.repository.acl;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.sapportals.wcm.protocol.webdav.Consts;
import com.sapportals.wcm.protocol.webdav.IInternationalString;
import com.sapportals.wcm.protocol.webdav.Name;
import com.sapportals.wcm.util.acl.IAclPermission;


/**
 * Represents a DAV:privilege as defined in
 * <a href="http://greenbytes.de/tech/webdav/rfc3744.html">WebDAV Access Control Protocol</a>.
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author stefan.eissing@greenbytes.de
 * @version $Revision: 1.3 $
 */
public interface IPrivilege {

  public static final Name N_ALL              = Consts.N_all;
  public static final Name N_READ             = Consts.N_read;
  public static final Name N_READ_ACL         = Consts.N_read_acl;
  public static final Name N_WRITE            = Consts.N_write;
  public static final Name N_WRITE_PROPERTIES = Consts.N_write_properties;
  public static final Name N_WRITE_CONTENT    = Consts.N_write_content;
  public static final Name N_WRITE_ACL        = Consts.N_write_acl;
  public static final Name N_READ_CUPRIVSET   = Consts.N_read_cuprivset;
  public static final Name N_UNLOCK           = Consts.N_unlock;
  public static final Name N_BIND             = Consts.N_bind;
  public static final Name N_UNBIND           = Consts.N_unbind;

  /** Returns the description of this privilege
   * @return description of this privilege
   */
  public IInternationalString getDescription();

  /** Returns the Name of this privilege
   * @return Name of this privilege
   */
  public Name getName();

  /** Returns if this privilege is abstract
   * @return if this privilege is abstract
   */
  public boolean isAbstract();

  /** Return IPermission object represented by this Privilege
   * @return the IPermission or <code>null</code> if there is no mapping
   */
  public IAclPermission getPermission();

  /** Construct DOM representation of this privilege
   * including child privileges
   * @return DOM representation of this privilege
   */
  public Element toElement(Document document);

  /** Find the privilege with name in subtree of this privilege
   * @param name of the privilege to look for
   * @return found privilege or <code>null</code>
   */
  public IPrivilege find(Name name);

  /** Return the parent in the tree of supported privileges
   * @return parent in supported privilege tree
   */
  public IPrivilege getParent();

  /** Return next sibling in the tree of supported privileges
   * @return next sibling in supported privilege tree
   */
  public IPrivilege getNextSibling();

  /** Return previous sibling in the tree of supported privileges
   * @return previous sibling in supported privilege tree
   */
  public IPrivilege getPrevSibling();

  /** Return first child in the tree of supported privileges
   * @return first child in supported privilege tree
   */
  public IPrivilege getFirstChild();

  /** Return last child in the tree of supported privileges
   * @return last child in supported privilege tree
   */
  public IPrivilege getLastChild();

  /** Return of privilege is contained in subtree-or-self of
   * this privilege
   * @param privilege to check for
   * @return if privilege is contained in subtree of this privilege (including self)
   */
  public boolean isContained(IPrivilege privilege);

  /** Return if this privilege is equal to the other
   * @return if this privilege is equal to the other
   */
  public boolean equals(IPrivilege other);

  /**
   * Return if this is ancestor of other
   * @retrun if this is ancestor of other
   */
  public boolean isAncestor(IPrivilege other);
}