package com.sapportals.wcm.protocol.webdav.repository.acl;

import java.util.List;
import java.util.Locale;
import java.util.Set;

import com.sapportals.wcm.protocol.webdav.Name;
import com.sapportals.wcm.repository.ResourceException;
import com.sapportals.wcm.util.acl.IAclPermission;
import com.sapportals.wcm.util.uri.RID;

/**
 * Represents a tree of DAV:privilege as defined in
 * <a href="http://greenbytes.de/tech/webdav/rfc3744.html">WebDAV Access Control Protocol</a>.
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author stefan.eissing@greenbytes.de
 * @version $Revision: 1.3 $
 */
interface IPrivilegeTree {

  /** Returns the root {@link com.sapportals.wcm.protocol.webdav.repository.acl.IPrivilege}
   * @return the root {@link com.sapportals.wcm.protocol.webdav.repository.acl.IPrivilege} of the tree
   */
  public IPrivilege getRoot();

  /**
   * Return the locale used for generating the privileges
   * @return the locale used for generating the privileges
   */
  public Locale getLocale();

  /** Find the {@link com.sapportals.wcm.protocol.webdav.repository.acl.IPrivilege}
   * with the given name
   * @param rid of the resource
   * @param name to look up
   * @return {@link com.sapportals.wcm.protocol.webdav.repository.acl.IPrivilege} with given name
   */
  public IPrivilege find(RID rid, Name name) throws ResourceException;

  /** Find the {@link com.sapportals.wcm.protocol.webdav.repository.acl.IPrivilege} which
   * represents the given {@link com.sapportals.wcm.util.acl.IAclPermission}
   * @param rid of the resource
   * @param permission to find matching privilege for
   * @return the {@link com.sapportals.wcm.protocol.webdav.repository.acl.IPrivilege} which
   * represents the given {@link com.sapportals.wcm.util.acl.IAclPermission}
   * @throws ResourceException on internal errors
   */
  public IPrivilege find(RID rid, IAclPermission permission) throws ResourceException;

  /**
   * Map a list of {@link com.sapportals.wcm.util.acl.IAclPermission}s to the
   * set of {@link com.sapportals.wcm.protocol.webdav.repository.acl.IPrivilege}s
   * with minimum redundancy.
   * @param rid of resource where permissions belong to
   * @param permissions list of permissions to map
   * @param canWriteAcl if the current user is allowed to modify the acl
   * @return set of matching {@link com.sapportals.wcm.protocol.webdav.repository.acl.IPrivilege}s
   */
  public Set mapPermissions(RID rid, List permissions, boolean canWriteAcl) throws ResourceException;

  /**
   * Map a IPrivilege to the list of {@link com.sapportals.wcm.util.acl.IAclPermission}
   * it represents. The list will be empty in case of an abstract IPrivilege
   * @param privilege the IPrivilege to retrieve the IPermissions for
   * @return list of {@link com.sapportals.wcm.util.acl.IAclPermission}s
   */
  public List mapPrivilege(IPrivilege privilege) throws ResourceException;

  /**
   * Remove from the set of IPrivileges all redundant privileges, e.g.
   * sub privileges of already contained ones.
   * @param privileges the set of IPrivileges
   */
  public void compress(Set privileges);
}