package com.sapportals.wcm.protocol.webdav.repository.acl;

import java.util.*;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.sap.tc.logging.Location;
import com.sapportals.portal.security.usermanagement.IUMPrincipal;
import com.sapportals.wcm.protocol.webdav.Consts;
import com.sapportals.wcm.protocol.webdav.DomElement;
import com.sapportals.wcm.protocol.webdav.repository.WDPrincipals;
import com.sapportals.wcm.repository.ResourceException;
import com.sapportals.wcm.util.uri.IHierarchicalUri;
import com.sapportals.wcm.util.uri.RID;

/**
 * Represents a single grant or deny of a set of {@link com.sapportals.wcm.protocol.webdav.repository.acl.IPrivilege}s,
 * a Access Control Entry (part of an Access Control List, see
 * {@link com.sapportals.wcm.protocol.webdav.repository.acl.WDAcl}).
 * <p>
 * A WDAce has the following properties:
 * <ul>
 * <li>a principal: for whom privileges are granted or denied.</li>
 * <li>a type: GRANT or DENY</li>
 * <li>readonly: indicates that this ACE is not changeable for a resource</li>
 * <li>inherited: indicates that this ACE was inherited from a parent resource</li>
 * <li>privileges: a list of {@link com.sapportals.wcm.protocol.webdav.repository.acl.IPrivilege}s</li>
 * </ul>
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author stefan.eissing@greenbytes.de
 * @version $Revision: 1.3 $
 */
class WDAce implements IAce {

  /** Definition of an ACE which grants everything to everyone */
  public static final WDAce ALL_FOR_ALL;
  /** Definition of an ACE which grants DAV:write-acl to everyone */
  public static final WDAce WRITE_ACL_FOR_ALL;
  /** Definition of an ACE which grants everything to everyone and
   * cannot be changed */
  public static final WDAce ALL_FOR_ALL_PROTECTED;
  /** Definition of an ACE which grants DAV:write-acl to everyone readonly*/
  public static final WDAce WRITE_ACL_FOR_ALL_PROTECTED;

  private static final Location log = Location.getLocation(WDAce.class);

  static {
    Set all = new HashSet(1);
    all.add(WDPrivilege.P_ALL);
    ALL_FOR_ALL = new WDAce(WDPrincipals.getEverybody(), GRANT, null, false, all, null);
    ALL_FOR_ALL_PROTECTED = new WDAce(WDPrincipals.getEverybody(), GRANT, null, true, all, null);
    all.clear();
    all.add(WDPrivilege.P_WRITE_ACL);
    WRITE_ACL_FOR_ALL = new WDAce(WDPrincipals.getEverybody(), GRANT, null, false, all, null);
    WRITE_ACL_FOR_ALL_PROTECTED = new WDAce(WDPrincipals.getEverybody(), GRANT, null, true, all, null);
  }

  public static WDAce getDefault(RID rid, IPrivilegeTree permissions, boolean readonly)
  throws ResourceException {
    IPrivilege full = permissions.find(rid, WDPrivilegeTree.N_SAP_FULL_CONTROL);
    if (full != null) {
      Set set = new HashSet(1);
      set.add(full);
      return new WDAce(WDPrincipals.getEverybody(), GRANT, null, readonly, set, null);
    }
    else {
      if (readonly) {
        return WDAce.ALL_FOR_ALL_PROTECTED;
      }
      else {
        return WDAce.ALL_FOR_ALL;
      }
    }
  }

  private final IUMPrincipal principal;
  private final int type;
  private final RID inheritedFrom;
  private final boolean isProtected;
  private final Set privileges;
  private final List raes;
  private final int hashCode;

  /** Create a new ACE with empty privileges
   * @param principal the principal to grant/deny privileges
   * @param type GRANT or DENY
   * @param inheritedFrom RID of resource ACE was inherited from or null
   * @param readonly this ACE cannot be changed
   */
  public WDAce(IUMPrincipal principal, int type, RID inheritedFrom, boolean readonly) {
    this(principal, type, inheritedFrom, readonly, null, null);
  }

  /** Create a new ACE.
   * @param principal the principal to grant/deny privileges
   * @param type GRANT or DENY
   * @param inheritedFrom the RID this ace was inherited from
   * @param readonly this ACE cannot be changed
   * @param privileges List of IPrivileges to grant/deny
   * @param raes List of IResourceAclEntry this WDAce is based on
   */
  protected WDAce(IUMPrincipal principal, int type, RID inheritedFrom,
                  boolean readonly, Set privileges, List raes) {
    this.principal = principal;
    this.type = type;
    this.inheritedFrom = inheritedFrom;
    this.isProtected = readonly;
    this.privileges = (privileges != null)? new HashSet(privileges) : new HashSet(17);
    this.raes = (raes != null)? new ArrayList(raes) : null;

    if (this.principal != null) {
      this.hashCode = this.principal.hashCode()*31 + this.privileges.hashCode();
    }
    else {
      this.hashCode = this.privileges.hashCode();
    }

    if (log.beDebug()) {
      log.debugT("new "+this.toString());
    }
  }

  public void addPrivilege(IPrivilege privilege) {
    this.privileges.add(privilege);
  }

  /**
   * Defines equality for WDAce objects. Two WDAces are defined equal if
   * <ul>
   * <li>they have the same principal</li>
   * <li>they have the same type</li>
   * <li>the contain the same privileges</li>
   * </ul>
   * inherited and readonly status play no role in equality.
   * @param o the object to compare with
   * @return WDAce equality if Object is a WDAce, false otherwise
   */
  public boolean equals(Object o) {
    if (o != null && o instanceof WDAce) {
      WDAce other = (WDAce)o;
      return ((this.principal.equals(other.principal))
               && (this.type == other.type)
               && (this.privileges.equals(other.privileges)));
    }
    return false;
  }

  public int hashCode() {
    return this.hashCode;
  }

  /**
   * @return the principal in this ACE
   */
  public IUMPrincipal getPrincipal() {
    return this.principal;
  }

  /**
   * @return the list of IPrivileges in this ACE
   */
  public Set getPrivileges() {
    return this.privileges;
  }

  protected List getResourceAclEntries() {
    return this.raes;
  }

  /**
   * @return the type of this ACE
   */
  public int getType() {
    return this.type;
  }

  public boolean grants(IPrivilege privilege) {

    if (isGrant()) {
      Iterator iter = this.privileges.iterator();
      while (iter.hasNext()) {
        IPrivilege priv = (IPrivilege)iter.next();
        if (priv.isContained(privilege)) {
          return true;
        }
      }
    }
    return false;
  }

  /**
   * @return if this ACE is granting privileges
   */
  public boolean isGrant() {
    return this.type == GRANT;
  }

  /**
   * @return if this ACE was inherited
   */
  public boolean isInherited() {
    return (this.inheritedFrom != null);
  }

  /**
   * @return if this ACE is read only
   */
  public boolean isReadOnly() {
    return this.isProtected;
  }

  /**
   * Return a DOM Element representing this ACE as defined in the
   * WebDAV Accesss Control Protocol Specification
   * @param doc the DOM Document to create the Element in
   * @return the created Element
   */
  public Element toElement(Document doc, IHierarchicalUri servletUri) {
    // ace (principal(href (url)), (grant|deny)(privilege+), protected?, inherited?)
    //
    Element ace = DomElement.makeNew(doc, Consts.N_ace);

    Element prince = DomElement.makeNew(doc, Consts.N_principal);
    if (this.principal.equals(WDPrincipals.getEverybody())) {
      prince.appendChild(DomElement.makeNew(doc, Consts.N_all));
    }
    else if (this.principal.equals(WDPrincipals.getAuthenticated())) {
      prince.appendChild(DomElement.makeNew(doc, Consts.N_authenticated));
    }
    else if (this.principal.equals(WDPrincipals.getAnonymous())) {
      prince.appendChild(DomElement.makeNew(doc, Consts.N_unauthenticated));
    }
    else {
      RID rid = WDPrincipals.getRid(this.principal);
      prince.appendChild(DomElement.makeNew(doc, Consts.N_href,
        servletUri.mapToAbsolutePath(rid)));
    }
    ace.appendChild(prince);

    Element grant = DomElement.makeNew(doc, (this.type == GRANT)? Consts.N_grant : Consts.N_deny);
    if (this.privileges != null) {
      Iterator iter = this.privileges.iterator();
      while (iter.hasNext()) {
        IPrivilege p = (IPrivilege)iter.next();
        Element priv = DomElement.makeNew(doc, Consts.N_privilege);
        priv.appendChild(DomElement.makeNew(doc, p.getName()));
        grant.appendChild(priv);
      }
    }
    ace.appendChild(grant);

    if (this.isProtected) ace.appendChild(DomElement.makeNew(doc, Consts.N_protected));
    if (this.inheritedFrom != null) {
      Element ein = DomElement.makeNew(doc, Consts.N_inherited);
      ein.appendChild(DomElement.makeNew(doc, Consts.N_href,
        servletUri.mapToAbsolutePath(this.inheritedFrom)));
      ace.appendChild(ein);
    }

    return ace;
  }

  public String toString() {
    return "WDAce["+this.principal.getId()+
      ((this.type == GRANT)? " grant" : " deny")+
      ((this.inheritedFrom != null)? " inherited("+this.inheritedFrom+")" : "")+
      (this.isProtected? " protected" : "")+
      this.privileges+"]";
  }
}