package com.sapportals.wcm.protocol.webdav.repository.acl;

import java.net.MalformedURLException;
import java.util.*;

import org.w3c.dom.Element;
import org.w3c.dom.Node;

import com.sapportals.portal.security.usermanagement.IUMPrincipal;
import com.sapportals.wcm.protocol.webdav.*;
import com.sapportals.wcm.protocol.webdav.dompath.DomPath;
import com.sapportals.wcm.protocol.webdav.repository.WDPrincipals;
import com.sapportals.wcm.repository.IResourceContext;
import com.sapportals.wcm.repository.ResourceException;
import com.sapportals.wcm.repository.security.IResourceAcl;
import com.sapportals.wcm.repository.security.IResourceAclEntry;
import com.sapportals.wcm.repository.security.IResourceAclEntryListIterator;
import com.sapportals.wcm.util.acl.AclException;
import com.sapportals.wcm.util.acl.IUMPrincipalListIterator;
import com.sapportals.wcm.util.logging.LoggingFormatter;
import com.sapportals.wcm.util.uri.IHierarchicalUri;
import com.sapportals.wcm.util.uri.IUriReference;
import com.sapportals.wcm.util.uri.RID;
import com.sapportals.wcm.util.uri.UriFactory;

/**
 *
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author stefan.eissing@greenbytes.de
 * @version $Revision: 1.3 $
 */
class WDAceList {

  private static final com.sap.tc.logging.Location log = com.sap.tc.logging.Location.getLocation(WDAceList.class.getName());

  private static final Name N_Principal = new Name(Consts.NS_WEBDAV, Consts.EN_principal);
  private static final Name N_Href      = new Name(Consts.NS_WEBDAV, Consts.EN_href);
  private static final Name N_Grant     = new Name(Consts.NS_WEBDAV, Consts.EN_grant);
  private static final Name N_Deny      = new Name(Consts.NS_WEBDAV, Consts.EN_deny);
  private static final Name N_Protected = new Name(Consts.NS_WEBDAV, Consts.EN_protected);
  private static final Name N_Inherited = new Name(Consts.NS_WEBDAV, Consts.EN_inherited);

  private static final DomPath DP_ace = new DomPath(
    "D:ace", DomPath.DEFAULT_NAMESPACE_BINDINGS);

  private static final DomPath DP_ace_set = new DomPath(
    "D:principal|D:principal/D:href|D:grant|D:deny|D:protected|D:inherited", DomPath.DEFAULT_NAMESPACE_BINDINGS);

  private static final DomPath DP_privileges = new DomPath(
    "D:privilege/*", DomPath.DEFAULT_NAMESPACE_BINDINGS);

  private static final DomPath DP_principal_all = new DomPath(
    "D:all", DomPath.DEFAULT_NAMESPACE_BINDINGS);

  private static final DomPath DP_principal_authenticated = new DomPath(
    "D:authenticated", DomPath.DEFAULT_NAMESPACE_BINDINGS);

  private static final DomPath DP_principal_unauthenticated = new DomPath(
    "D:unauthenticated", DomPath.DEFAULT_NAMESPACE_BINDINGS);

  private static final DomPath DP_principal_property = new DomPath(
    "D:property", DomPath.DEFAULT_NAMESPACE_BINDINGS);

  private static final DomPath DP_principal_self = new DomPath(
    "D:self", DomPath.DEFAULT_NAMESPACE_BINDINGS);

  private static final DomPath DP_href = new DomPath(
    "D:href", DomPath.DEFAULT_NAMESPACE_BINDINGS);

  private static final Comparator compareEntries;

  static {
    compareEntries = new Comparator() {
      public int compare(java.lang.Object A, java.lang.Object B) {
        IResourceAclEntry entryA = (IResourceAclEntry)A;
        IResourceAclEntry entryB = (IResourceAclEntry)B;
        try {
          int result = entryA.getSortIndex() - entryB.getSortIndex();
          if (result == 0) {
            result = entryA.getPrincipal().getId().compareTo(entryB.getPrincipal().getId());
            if (result == 0) {
              return entryA.isNegative()?
                (entryB.isNegative()? 0 : 1)
              : (entryB.isNegative()? -1 : 0);
            }
          }
          return result;
        }
        catch (AclException ex) {
          log.errorT("while comparing IResourceAclEntry - " + LoggingFormatter.extractCallstack(ex));
          return -1;
        }
      }
    };
  }

  protected static List getList(IResourceAcl acl, RID inheritedFrom,
                                IPrivilegeTree tree, IUMPrincipal user)
  throws ResourceException, AclException {
    // We retrieve the list of ResourceAclEntries, which we might
    // need to sort according to index, principal, and grant|deny.
    // Then we collect all entries for a specific principal with
    // the same grant|deny setting. This gives a list of IPermissions.
    //
    // The list of IPermissions is translated, using the IPrivilegeTree,
    // to a list of IPrivileges.
    //
    // Finally, iterator over all owners of the ResourceAcl and determine
    // those, which are not granted DAV:write_acl already. Generate a new
    // WDAce, granting DAV:write_acl, for that owner.
    //
    if (log.beDebug()) {
      log.debugT("build AceList for acl, inherited: "+inheritedFrom);
    }

    RID rid = acl.getResource().getRID();
    List aces = new ArrayList(17);
    List entries = new ArrayList(17);

    IResourceAclEntryListIterator raeIter = acl.getEntries().iterator();
    while (raeIter.hasNext()) {
      entries.add(raeIter.next());
    }
    Collections.sort(entries, compareEntries);

    List raes = new ArrayList(17);
    List permissions = new ArrayList(17);
    Set implicitOwners = new HashSet(17);
    IPrivilege write_acl = tree.find(rid, IPrivilege.N_WRITE_ACL);
    IUMPrincipal lastPrincipal = null;
    boolean lastGrant = true;
    boolean lastReadonly = false;

    for (Iterator iter = entries.iterator(); iter.hasNext(); /**/) {
      IResourceAclEntry rae = (IResourceAclEntry)iter.next();

      if (rae.getPrincipal().equals(lastPrincipal)
          && lastGrant == !rae.isNegative()
          && lastReadonly == rae.isReadOnly()) {
        raes.add(rae);
        permissions.add(rae.getPermission());
      }
      else {
        if (lastPrincipal != null) {
          Set privileges = tree.mapPermissions(rid, permissions, acl.isOwner(lastPrincipal));
          aces.add(new WDAce(lastPrincipal, lastGrant? WDAce.GRANT : WDAce.DENY,
            inheritedFrom, lastReadonly, privileges, raes));
          if (lastGrant && (privileges.contains(WDPrivilege.P_ALL)
              || (write_acl != null && privileges.contains(write_acl)))) {
            implicitOwners.add(lastPrincipal.getId());
          }

          raes.clear();
          permissions.clear();
        }

        lastPrincipal = rae.getPrincipal();
        lastGrant = !rae.isNegative();
        lastReadonly = rae.isReadOnly();
        raes.add(rae);
        permissions.add(rae.getPermission());
      }
    }

    if (lastPrincipal != null) {
      Set privileges = tree.mapPermissions(rid, permissions, acl.isOwner(lastPrincipal));
      aces.add(new WDAce(lastPrincipal, lastGrant? WDAce.GRANT : WDAce.DENY,
        inheritedFrom, lastReadonly, privileges, raes));
      if (lastGrant && (privileges.contains(WDPrivilege.P_ALL)
          || (write_acl != null && privileges.contains(write_acl)))) {
        implicitOwners.add(lastPrincipal.getId());
      }
    }

    if (write_acl != null) {
      IUMPrincipalListIterator pi = acl.getOwners().iterator();
      while (pi.hasNext()) {
        IUMPrincipal owner = pi.next();
        if (!implicitOwners.contains(owner.getId())) {
          if (log.beDebug()) {
            log.debugT("inventing DAV:write-acl ace for "+owner.getId()+" implicit owners "+implicitOwners);
          }
          WDAce ace = new WDAce(owner, WDAce.GRANT, inheritedFrom, false);
          ace.addPrivilege(write_acl);
          if (!aces.contains(ace)) {
            aces.add(ace);
          }
        }
      }
    }

    return aces;
  }

  protected static List getList(RID rid, Element acl, IPrivilegeTree tree, IHierarchicalUri baseUrl, IResourceContext context)
  throws ResourceException {
    List result = new ArrayList(17);
    Node[] nodes = DP_ace.matchNodes(acl);

    for (int i = 0; i < nodes.length; ++i) {
      WDAce ace = getAce(rid, nodes[i], tree, baseUrl, context);
      if (!result.contains(ace)) {
        result.add(ace);
      }
    }

    if (log.beDebug()) {
      log.debugT("Element "+DomSerializer.toString(acl, false) +" mapped to "+result);
    }
    return result;
  }

  private static WDAce getAce(RID rid, Node ace, IPrivilegeTree tree, IHierarchicalUri baseUrl, IResourceContext context)
  throws ResourceException {
    Map nodes = DP_ace_set.matchFirstNames(ace);

    assertName(nodes, N_Principal, "DAV:principal missing in DAV:ace");
    assertEitherOr(nodes, N_Grant, N_Deny, "DAV:grant/DAV:deny missing in DAV:ace");

    Element eprinc = (Element)nodes.get(N_Principal);
    IUMPrincipal principal = null;
    Element href = (Element)nodes.get(N_Href);
    if (href != null) {
      try {
        IUriReference ref = UriFactory.parseUriReference(DomElement.getText(href));
        RID principalRID = baseUrl.mapToResourceID(ref);
        principal = WDPrincipals.getPrincipal(principalRID, context);
        if (log.beInfo()) {
          if (principal != null) {
            log.infoT("mapped principal "+DomElement.getText(href)+" to "+principal.getId());
          }
          else {
            log.infoT("principal unknown: "+DomElement.getText(href));
          }
        }
      }
      catch (MalformedURLException ex) {
        log.debugT(" - " + LoggingFormatter.extractCallstack(ex));
      }
    }
    else {
      if (DP_principal_all.matchFirst(eprinc) != null) {
        principal = WDPrincipals.getEverybody();
      }
      else if (DP_principal_authenticated.matchFirst(eprinc) != null) {
        principal = WDPrincipals.getAuthenticated();
      }
      else if (DP_principal_unauthenticated.matchFirst(eprinc) != null) {
        principal = WDPrincipals.getAnonymous();
      }
      else if (DP_principal_property.matchFirst(eprinc) != null) {
        throw new AclPreconditionException("unsupported principal: "+DomSerializer.toString(eprinc, false), DavConditions.N_allowed_principal);
      }
      else if (DP_principal_self.matchFirst(eprinc) != null) {
        principal = context.getUser();
      }
      else {
        throw new AclPreconditionException("unknown principal: "+DomSerializer.toString(eprinc, false), DavConditions.N_recognized_principal);
      }
    }

    if (principal == null) {
      throw new AclPreconditionException("unknown principal: "+DomSerializer.toString(eprinc, false), DavConditions.N_recognized_principal);
    }

    boolean isGrant = true;
    Element egrant  = (Element)nodes.get(N_Grant);
    if (egrant == null) {
      isGrant = false;
      egrant = (Element)nodes.get(N_Deny);
    }

    RID inheritedFrom = null;
    Element inherit = (Element)nodes.get(N_Inherited);
    if (inherit != null) {
        Node ihref = DP_href.matchFirst(inherit);
      if (ihref != null) {
        try {
          IUriReference ref = UriFactory.parseUriReference(DomElement.getText(ihref));
          inheritedFrom = baseUrl.mapToResourceID(ref);
        }
        catch (MalformedURLException ex) {
          log.debugT(" - " + LoggingFormatter.extractCallstack(ex));
        }
      }
    }

    WDAce wdace = new WDAce(principal, isGrant? WDAce.GRANT : WDAce.DENY,
      inheritedFrom, nodes.containsKey(N_Protected));

    // jre: this may be wrong -- shouldn't the presence of any unknown
    // privilege cause a failure???
    Iterator eprivs = DP_privileges.match(egrant).iterator();
    while (eprivs.hasNext()) {
      Name name = new Name((Node)eprivs.next());
      IPrivilege privilege = tree.find(rid, name);
      if (privilege == null) {
        throw new AclPreconditionException("unknown privilege in DAV:ace "+name, DavConditions.N_supported_privilege);
      }
      else {
        if (privilege.isAbstract()) {
          throw new AclPreconditionException("abstract privilege in DAV:ace"+name, DavConditions.N_no_abstract);
        }
        wdace.addPrivilege(privilege);
      }
    }

    if (wdace.getPrivileges().isEmpty()) {
      throw new AclPreconditionException("no known privileges in DAV:ace", DavConditions.N_supported_privilege);
    }

    return wdace;
  }

  private static void assertName(Map map, Name name, String message)
  throws ResourceException {
    if (!map.containsKey(name)) {
      throw new IllegalArgumentException(message);
    }
  }

  private static void assertEitherOr(Map map, Name name1, Name name2, String message)
  throws ResourceException {
    if (!map.containsKey(name1) && !map.containsKey(name2)) {
      throw new IllegalArgumentException(message);
    }
  }
}