package com.sapportals.wcm.protocol.webdav.repository.acl;

import java.util.*;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.sap.tc.logging.Location;
import com.sapportals.portal.security.usermanagement.IUMPrincipal;
import com.sapportals.portal.security.usermanagement.IUser;
import com.sapportals.wcm.protocol.webdav.Consts;
import com.sapportals.wcm.protocol.webdav.DomElement;
import com.sapportals.wcm.protocol.webdav.Name;
import com.sapportals.wcm.protocol.webdav.repository.WDPrincipals;
import com.sapportals.wcm.repository.*;
import com.sapportals.wcm.repository.manager.IAclSecurityManager;
import com.sapportals.wcm.repository.manager.IRepositoryManager;
import com.sapportals.wcm.repository.manager.ISecurityManager;
import com.sapportals.wcm.repository.security.IResourceAcl;
import com.sapportals.wcm.repository.security.IResourceAclEntry;
import com.sapportals.wcm.repository.security.IResourceAclManager;
import com.sapportals.wcm.util.acl.*;
import com.sapportals.wcm.util.uri.IHierarchicalUri;
import com.sapportals.wcm.util.uri.RID;

/** Handles all information related to WebDAV ACL handling for a resource.
 * <p>
 * WDAcl handles all DAV: properties defined in WebDAV ACL for a
 * particular {@link com.sapportals.wcm.repository.IResource}.
 * <p>
 * In particular each WDAcl has the following properties:
 * <ul>
 * <li>a resource: to which the ACL belongs</li>
 * <li>an internal manager: (<code>hasManager()</code>). Without a manager,
 * the ACL really is a fake and cannot be changed. Especially without
 * a manager, an application can expect that all operations are allowed.</li>
 * <li>readonly: a WDAcl can be read only, so all changes can be expected
 * to fail.</li>
 * <li>supported privilege tree: a tree of {@link com.sapportals.wcm.protocol.webdav.repository.acl.IPrivilege}s
 * which can be used for named lookup of privileges and containement of privileges. All
 * WebDAV defined privileges can be expected to be present in the tree, however some
 * might be abstract. Abstract privileges cannot be set for a resource.</li>
 * <li>ace list: a list of {@link com.sapportals.wcm.protocol.webdav.repository.acl.IAce}s which
 * define which privilege is granted or denied for which principal.</li>
 * </ul>
 *
 * <p>
 * Copyright (c) SAP AG 2001-2003
 * @author stefan.eissing@greenbytes.de
 * @version $Revision: 1.5 $
 */
public final class WDAcl {

  /** Entries evaluated in the order specified */
  public static final int EVAL_FIRST_MATCH = 1;
  /** Entries evaluated by matching GRANTs before DENY */
  public static final int EVAL_ALL_GRANT_FIRST = 2;
  /** Entries evaluated by matching DENY for Principal first */
  public static final int EVAL_SPECIFIC_DENY_OVERRIDES = 3;

  private static final Location log = Location.getLocation(WDAcl.class);

  private final RID rid;
  private final IResource resource;
  private final IResourceAclManager manager;
  private final boolean readonly;
  private IUser user;
  private IResourceAcl resourceAcl;
  private RID inheritedFrom;
  private IPrivilegeTree supported;
  private Set cuprivset;
  private List aces;

  /** Map of IResourceAcl to AclLock objects for active locks */
  private final Map aclLocks = Collections.synchronizedMap(new HashMap(10));

  /**
   * Construct a WDAcl for a certain resource.
   * @param resource to retrieve ACL information for
   * @throws ResourceException on internal failures
   */
  public WDAcl(IResource resource) throws ResourceException {
    this.rid = resource.getRID();
    this.resource = resource;
    this.manager = getResourceAclManager(this.resource);
    boolean readonly = false;
    try {
      readonly = (this.manager != null) ? this.manager.isReadOnly() : true;
    }
    catch (AclException ex) {
      // $JL-EXC$
      throw map(ex, this.rid);
    }
    this.readonly = readonly;
  }

  /** Can only GRANT, but not DENY privileges
   * @return if ACL can only grant, but not deny privileges
   */
  public boolean allowsGrantOnly() throws ResourceException {
    return (this.manager != null) ? !this.manager.areNegativeAclEntriesSupported() : true;
  }

  /** Can have multiple entries for one Principal
   * @return if ACL can have multiple entries per Principal
   */
  public boolean allowsPrincipalsOnce() throws ResourceException {
    return false;
  }

  /** DENY ACL entries can follow GRANT entries
   * @return if DENY entries can follow GRANT entries
   */
  public boolean allowsDenyAfterGrant() throws ResourceException {
    return true;
  }

  /** Return the order in which entries are evaluated, when
   * multiple entries match a certain Principal.
   * @return evaluation type (EVAL_*)
   */
  public int getEvaluationOrder() throws ResourceException {
    return WDAcl.EVAL_FIRST_MATCH;
  }

  /**
   * Returns if there is a Manager handling ACLs
   * @return if there is a Manager handling ACLs
   */
  public boolean hasManager() {
    return this.manager != null;
  }

  /**
   * Return if ACLs for this resource are generally read only.
   * @return if ACLs are generally readonly
   */
  public boolean isReadOnly() {
    return this.readonly;
  }

  /** Return the List of {@link com.sapportals.wcm.protocol.webdav.repository.acl.IAce}
   * objects for this Access Control List
   * @return List of {@link com.sapportals.wcm.protocol.webdav.repository.acl.IAce}
   */
  public synchronized List getAces() throws ResourceException {
    if (this.aces == null) {
      IResourceAcl acl = getResourceAcl();
      if (acl == null) {
        // everything is allowed
        //
        if (log.beDebug()) {
          log.debugT("no IResourceAcl for " + this.resource + ", synthesizing default");
        }
        checkUser();
        getSupported();
        this.aces = new ArrayList(2);
        this.aces.add(WDAce.getDefault(this.rid, this.supported, isReadOnly()));
      }
      else {
        if (log.beDebug()) {
          log.debugT("found IResourceAcl for "+ this.resource + ": "+acl);
        }
        checkUser();
        getSupported();
        try {
          this.aces = Collections.unmodifiableList(WDAceList.getList(
            acl, this.inheritedFrom, this.supported, this.user));
        }
        catch (AclException ex) {
          // $JL-EXC$
          log.debugT("getting ace list" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(ex));
          throw map(ex, this.rid);
        }
      }
    }
    return this.aces;
  }

  /** Return the List of {@link com.sapportals.wcm.protocol.webdav.repository.acl.IAce}
   * objects, representing the given DAV:ace property
   * @param acl DAV:acl property as in ACL Method
   * @return List of {@link com.sapportals.wcm.protocol.webdav.repository.acl.IAce}
   */
  public List getAces(Element acl, IHierarchicalUri baseUrl, IResourceContext context) throws ResourceException {
    getSupported();
    return WDAceList.getList(this.rid, acl, this.supported, baseUrl, context);
  }

  /**
   * Retrieves the DAV:current-user-privilege-set as Set of
   * {@link com.sapportals.wcm.protocol.webdav.repository.acl.IPrivilege}s.
   * A privilege is part of the list, if the current user was
   * granted that privilege (implicit or explicit).
   *
   * @return set of {@link com.sapportals.wcm.protocol.webdav.repository.acl.IPrivilege}s
   * @throws ResourceException
   */
  public synchronized Set getCurrentUserSet() throws ResourceException {
    if (this.cuprivset == null) {
      IResourceAcl acl = getResourceAcl();
      if (acl == null) {
        // empty ACL means everything is allowed
        //
        this.cuprivset = new HashSet(17);
        collectNonAbstract(getSupported(), this.cuprivset);
      }
      else {
        checkUser();
        if (this.user == null) {
          this.cuprivset = Collections.EMPTY_SET;
        }
        else {
          this.cuprivset = new HashSet(17);

          AclLock lock = null;
          try {
            lock = new AclLock(acl);
            if (acl.isOwner(this.user)) {
              this.cuprivset.add(WDPrivilege.P_WRITE_ACL);
            }

            collectGranted(getSupported(), acl, this.user, this.cuprivset);
          }
          catch (AclException ex) {
            // $JL-EXC$
            throw map(ex, this.rid);
          }
          finally {
            if (lock != null) lock.release();
          }
        }
      }

      this.supported.compress(this.cuprivset);
    }

    return this.cuprivset;
  }

  /** Return the principal owning the ACL.
   * @return the owner of the ACL
   */
  public IUMPrincipal getOwner() throws ResourceException {
    // Return the first non-group principal in the owner list
    // of the ResourceAcl
    //
    IResourceAcl acl = getResourceAcl();
    if (acl != null) {
      try {
        IUMPrincipal owner = null;
        IUMPrincipalListIterator iter = acl.getOwners().iterator();
        while (iter.hasNext()) {
          IUMPrincipal p = iter.next();
          if (p.getType() == IUMPrincipal.IUSER) return p;
          else owner = p;
        }
        if (owner != null) return owner;
      }
      catch (AclException ex) {
        // $JL-EXC$
        throw map(ex, this.rid);
      }
    }

    return WDPrincipals.getEverybody();
  }

  /** Returns the resource to which this ACL belongs.
   * @return the resource to which this ACL belongs
   */
  public IResource getResource() {
    return this.resource;
  }

  /**
   * Retrieve the tree's root of supported IPrivileges. The tree can be
   * traversed using methods in {@link com.sapportals.wcm.protocol.webdav.repository.acl.IPrivilege}.
   *
   * @return root of privilege tree
   */
  public synchronized IPrivilege getSupported() throws ResourceException {
    if (this.supported == null && hasManager()) {
      this.supported = WDPrivilegeTree.getInstance(this.manager, this.resource);
    }
    return this.supported.getRoot();
  }

  /**
   * Append a WebDAV ACL conform representation of the Access Control
   * List to the given Element. The representation is a list of DAV:ace
   * Elements.
   *
   * @param element where DAV:ace shall be appended
   */
  public void appendTo(Element element, IHierarchicalUri servletUri) throws ResourceException {
    Document doc = element.getOwnerDocument();
    Iterator iter = getAces().iterator();
    while (iter.hasNext()) {
      element.appendChild(((IAce)iter.next()).toElement(doc, servletUri));
    }
  }

  /**
   * Return an Element representing the DAV:acl property for the
   * access control list of the resource.
   * @param document the Document where the Element is created in
   * @return the created DAV:acl Element
   */
  public Element toElement(Document document, IHierarchicalUri servletUri) throws ResourceException {
    Element acl = DomElement.makeNew(document, Consts.N_acl);
    appendTo(acl, servletUri);
    return acl;
  }


  /** Sets the list of given IAce objects as new access control list.
   * Any error in installing the new list will result in changing the
   * ACL back to its value before the call.
   *
   * @param newList the list of {@link com.sapportals.wcm.protocol.webdav.repository.acl.IAce}s
   * @throws ResourceException if ACL could not be changed
   */
  public synchronized void setAces(List newList, boolean deep) throws ResourceException {
    if (log.beDebug()) {
      log.debugT("setAces("+newList+", "+deep+")");
    }
    if (isReadOnly()) {
      throw new AccessDeniedException("acl is readonly");
    }

    AclLock lock = null;
    try {
      if (this.resourceAcl != null) {
        checkUser();
        if (!this.resourceAcl.isAuthorized(this.user)) {
          throw new AccessDeniedException("user " + this.user.getId() + " not authoized for acl changes");
        }
      }

      boolean isRemove = newList.isEmpty();
      IResourceAcl oldAcl = (this.inheritedFrom != null)? null: this.resourceAcl;
      getAces();

      if (log.beDebug()) {
        log.debugT("setAces oldAcl is "+oldAcl);
      }
      List oldAces = null;
      if (oldAcl != null && !isRemove) {
        lock = new AclLock(oldAcl);
        oldAces = new ArrayList(this.aces);
      }

      try {
        if (isRemove) {
          removeAces(deep);
        }
        else {
          setAcesInternal(newList);
        }
      }
      catch (ResourceException ex) {
        // Something went wrong, try to restore old ACL state
        //
        log.debugT("setting ACL failed" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(ex));
        aceListChanged();
        if (oldAces != null) {
          try {
            setAcesInternal(oldAces);
          }
          catch (ResourceException ex2) {
            log.warningT("trying to restore old ACL" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(ex2));
          }
        }
        else {
          try {
            this.manager.removeAcl(this.resource);
          }
          catch (Exception ex2) {
            log.warningT("trying to restore old ACL" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(ex2));
          }
        }
        throw ex;
      }
    }
    catch (AclException ex) {
      // $JL-EXC$
      if (log.beInfo()) {
        log.infoT("while setting ace list" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(ex));
      }
      throw map(ex, this.rid);
    }
    finally {
      aceListChanged();
      if (lock != null) lock.release();
    }
  }

  // ------------------------ protected / private ----------------------------

  /**
   * Retrieve the IResourceAclManager responsible for the resource.
   * @return the manager if present, or <code>null</code>
   */
  private static IResourceAclManager getResourceAclManager(IResource resource)
  throws ResourceException {
    IRepositoryManager rmgr = resource.getRepositoryManager();
    if (rmgr == null) {
      return null;
    }

    ISecurityManager secmgr = rmgr.getSecurityManager(resource);
    if (secmgr != null && secmgr instanceof IAclSecurityManager) {
      IAclSecurityManager aclsecmgr = (IAclSecurityManager)secmgr;
      return aclsecmgr.getAclManager();
    }
    return null;
  }

  /** Handles locks on IResourceAcl objects. These locks can be nested
   * and only the first AclLock (which really obtained the lock on the
   * IResourceAcl) will actually release it again.
   */
  private final class AclLock {

    private final IResourceAcl m_acl;
    private final AclLock m_master;

    /** Obtain lock on the IResourceAcl
     * @param acl the IResourceAcl to lock
     * @throws ResourceException (LOCKED) when lock could not be obtained
     */
    public AclLock(IResourceAcl acl) throws ResourceException, AclException {
      m_acl = acl;
      if (m_acl != null) {
        m_master = (AclLock)WDAcl.this.aclLocks.get(m_acl);
        if (m_master == null) {
          obtain();
          WDAcl.this.aclLocks.put(m_acl, this);
        }
      }
      else {
        m_master = null;
      }
    }

    /** Release lock on the IResourceAcl
     */
    public void release() throws ResourceException {
      if (m_acl != null && m_master == null) {
        WDAcl.this.aclLocks.remove(m_acl);
        try {
          m_acl.unlock();
        }
        catch (AclException ex) {
          // $JL-EXC$
          log.errorT("while releasing ACL lock" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(ex));
          throw map(ex, WDAcl.this.rid);
        }
      }
    }

    private void obtain() throws ResourceException, AclException {
      try {
        final long[] timeouts = { 200L, 500L, 1000L, 2000L };
        for (int i = 0; i < timeouts.length; ++i) {
          if (m_acl.lock()) {
            return;
          }

          try {
            Thread.sleep(timeouts[i]);
          }
          catch (InterruptedException ex) {
            log.warningT("interrupted while waiting for ACL lock on "+WDAcl.this.rid + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(ex));
            break;
          }
        }

        log.warningT("cannot obtain lock on ACL of resource "+WDAcl.this.rid);
        throw new LockedException("resource ACL is locked", WDAcl.this.rid, null);
      }
      catch (NotAuthorizedException ex) {
        throw new AuthorizationRequiredException("user is not authorized to obtain lock on ACL");
      }
    }
  }

  private void aceListChanged() {
    this.aces = null;
    this.cuprivset = null;
  }

  private void checkUser() {
    if (this.user == null) {
      this.user = this.resource.getContext().getUser();
    }
  }

  private IResourceAcl getResourceAcl() throws ResourceException {
    if (this.resourceAcl == null && hasManager()) {
      this.inheritedFrom = null;
      try {
        this.resourceAcl = this.manager.getAcl(this.resource);
        if (this.resourceAcl == null) {
          this.resourceAcl = this.manager.getInheritedAcl(this.resource);
          if (this.resourceAcl != null) {
            this.inheritedFrom = this.resourceAcl.getResource().getRID();
          }
        }
      }
      catch (AclException ex) {
        // $JL-EXC$
        throw map(ex, this.rid);
      }
    }
    return this.resourceAcl;
  }

  private static boolean collectGranted(IPrivilege current, IResourceAcl acl,
                                        IUMPrincipal principal, Set granted)
  throws AclException {
    boolean mayBeGranted = true;
    for (IPrivilege contained = current.getFirstChild();
         contained != null;
         contained = contained.getNextSibling()) {
      if (!collectGranted(contained, acl, principal, granted)) {
        mayBeGranted = false;
      }
    }

    if (current.isAbstract()) {
      // cannot be granted, but parent privilege is still possible
      //
      return true;
    }
    else {
      IAclPermission permission = current.getPermission();
      if (permission != null) {
        if (acl.checkPermission((IUser)principal, permission)) {
          granted.add(current);
          return true;
        }
        return false;
      }
      else {
        if (mayBeGranted) {
          granted.add(current);
          return true;
        }
        return false;
      }
    }
  }

  private static void collectNonAbstract(IPrivilege current, Set granted)
  throws ResourceException {

    for (IPrivilege contained = current.getFirstChild();
         contained != null;
         contained = contained.getNextSibling()) {
      collectNonAbstract(contained, granted);
    }

    if (!current.isAbstract()) {
      granted.add(current);
    }
  }

  private void removeAces(boolean deep) throws ResourceException {
    if (log.beDebug()) {
      log.debugT("removeAces("+deep+")");
    }
    try {
      this.manager.removeAcl(this.resource);
      if (deep && (this.resource instanceof ICollection)) {
        this.manager.propagateAcl_Remove(this.resource);
      }
    }
    catch (AclException ex) {
      // $JL-EXC$
      log.debugT("removing acl" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(ex));
      throw map(ex, this.rid);
    }
  }

  private void setAcesInternal(List newList) throws ResourceException {
    if (log.beDebug()) {
      log.debugT("setAcesInternal("+newList+")");
    }
    try {
      if ((this.inheritedFrom != null) || this.resourceAcl == null) {
        // We have to create a completely new IResourceAcl and then
        // parse the AceList again, since a new IResourceAcl will normally
        // have default aces.
        //
        if (log.beDebug()) {
          log.debugT("creating new Acl (" + this.inheritedFrom + "): "+newList);
        }
        this.resourceAcl = this.manager.createAcl(this.resource);
        this.inheritedFrom = null;
        this.aces = null;
      }

      // We have to modify an existing IResourceAcl:
      // - all non-inherited, non-readonly acl entries, which are not
      //   mentioned in the new list, need to be removed
      // - all other entries need to be kept
      // - existing readonly entries must be part of the new list
      // - all entries in the new list, which are not already
      //   part of the existing list, need to be set.
      // - the order in the new list need to be preserved
      // - as a last step, we need to assign/revoke DAV:write-acl
      //   privileges
      //
      List existingList = getAces();
      if (log.beDebug()) {
        log.debugT("modifying ace list from "+existingList+" to "+newList);
      }

      List removeList = new ArrayList(existingList.size());
      List keepList = new ArrayList(existingList.size());

      Iterator iter = existingList.iterator();
      while (iter.hasNext()) {
        IAce ace = (IAce)iter.next();

        if (ace.isInherited()) {
          keepList.add(ace);
        }
        else if (ace.isReadOnly()) {
          keepList.add(ace);
        }
        else {
          if (newList.contains(ace)) keepList.add(ace);
          else removeList.add(ace);
        }
      }

      removeAceListInternal(this.resourceAcl, removeList);
      int entries = insertAceListInternal(this.resourceAcl, newList, keepList);
      modifyOwnership(this.resourceAcl, newList, removeList);

      if (entries == 0) {
        if (false && !grantsRead(keepList)) {
          // Well, well, we have no read access for anyone any longer.
          // We better grant DAV:read to all owners
          //
          IPrivilege read = this.supported.find(this.rid, Consts.N_read);
          if (read != null) {
            int index = 1;
            IUMPrincipalListIterator oiter = this.resourceAcl.getOwners().iterator();
            while (oiter.hasNext()) {
              IUMPrincipal owner = oiter.next();
              WDAce ace = new WDAce(owner, WDAce.GRANT, null, false);
              ace.addPrivilege(read);
              addAceInternal(this.resourceAcl, ace, index++);

              if (log.beDebug()) {
                log.debugT("added failsafe DAV:read access for owner "+owner.getId());
              }
            }
          }
        }
      }
    }
    catch (AclException ex) {
      // $JL-EXC$
      log.debugT("setting acl" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(ex));
      throw map(ex, this.rid);
    }
  }

//  private int addAceListInternal(IResourceAcl resourceAcl, List aces, int index)
//  throws ResourceException, AclException {
//
//    if (log.beDebug()) {
//      log.debugT("setting ace list: "+aces);
//    }
//
//    int entries = 0;
//    getSupported();
//    Iterator iter = aces.iterator();
//    while (iter.hasNext()) {
//
//      entries += addAceInternal(resourceAcl, (IAce)iter.next(), index);
//      ++index;
//    }
//
//    return entries;
//  }

  private int addAceInternal(IResourceAcl resourceAcl, IAce ace, int index)
  throws ResourceException, AclException {

    int setEntries = 0;
    Iterator piter = ace.getPrivileges().iterator();
    while (piter.hasNext()) {

      IPrivilege privilege = (IPrivilege)piter.next();
      List permissions = this.supported.mapPrivilege(privilege);
      if (log.beDebug()) {
        log.debugT("addAceInternal: privilege "+privilege+" mapped to "+permissions);
      }

      for (Iterator permiter = permissions.iterator(); permiter.hasNext(); ) {
        IAclPermission permission = (IAclPermission)permiter.next();
        if (log.beDebug()) {
          log.debugT("setting permission: "+permission.getName());
        }
        IResourceAclEntry entry = this.manager.createAclEntry(
          ace.getPrincipal(), !ace.isGrant(), permission, index);
        resourceAcl.addEntry(entry);
        ++setEntries;
      }
    }

    return setEntries;
  }

  private void removeAceListInternal(IResourceAcl resourceAcl, List aces)
  throws ResourceException, AclException {
    getSupported();

    Iterator iter = aces.iterator();
    while (iter.hasNext()) {
      WDAce ace = (WDAce)iter.next();

      List raes = ace.getResourceAclEntries();
      if (raes != null) {
        Iterator riter = raes.iterator();
        while (riter.hasNext()) {
          IResourceAclEntry rae = (IResourceAclEntry)riter.next();
          if (log.beDebug()) {
            log.debugT("removing permission: "+rae.getPermission().getName());
          }
          resourceAcl.removeEntry(rae);
        }
      }
    }
  }

  private int insertAceListInternal(IResourceAcl resourceAcl,
                                    List newList, List existing)
  throws ResourceException, AclException {
    getSupported();

    int entries = 0;
    int index = 1;
    Iterator iter = newList.iterator();
    while (iter.hasNext()) {
      IAce ace = (IAce)iter.next();

      int found;
      if ((found = existing.indexOf(ace)) >= 0) {
        // this entry is already there, we only might have to
        // adjust the sort index
        //
        WDAce exist = (WDAce)existing.get(found);
        List raes = exist.getResourceAclEntries();
        if (raes != null && !raes.isEmpty()) {

          if (index != ((IResourceAclEntry)raes.get(0)).getSortIndex()) {
            Iterator riter = raes.iterator();
            while (riter.hasNext()) {
              IResourceAclEntry rae = (IResourceAclEntry)riter.next();
              if (log.beDebug()) {
                log.debugT("resort: "+rae.getPermission().getName()+" to index "+index);
              }
              rae.setSortIndex(index);
              ++entries;
            }
          }
        }

      }
      else {
        // this entry is new
        //
        entries += addAceInternal(resourceAcl, ace, index);
      }

      ++index;
    }
    return entries;
  }

  /** DAV:write-acl privileges are set/removed by modifying the ownership
   * of the IResourceAcl object of a resource. This is done in three steps:
   * <ol>
   * <li> Look for DAV:grant entries in the newList, which grant DAV:write-acl.
   *      Add the given principal to the owners of the IResourceAcl.
   * </li>
   * <li> Look for DAV:deny entries in the newList, which deny DAV:write-acl.
   *      Remove the given principal from the owners of the IResourceAcl.
   * </li>
   * <li> Look for DAV:grant entries in the removeList, which grant DAV:write-acl.
   *      Remove the given principal from the owners of the IResourceAcl.
   * </li>
   * </ol>
   */
  private void modifyOwnership(IResourceAcl resourceAcl, List newList, List removeList)
  throws ResourceException, AclException {

    List newOwners = new ArrayList(17);

    Iterator iter = newList.iterator();
    while (iter.hasNext()) {
      IAce ace = (IAce)iter.next();
      if (ace.isGrant()) {
        Set privileges = ace.getPrivileges();

        if (privileges.contains(WDPrivilege.P_WRITE_ACL)
            || privileges.contains(WDPrivilege.P_ALL)) {

          IUMPrincipal principal = ace.getPrincipal();
          if (!resourceAcl.isOwner(principal)) {
            resourceAcl.addOwner(principal);
          }
          if (!newOwners.contains(principal.getId())) {
            newOwners.add(principal.getId());
          }
        }
      }
    }

    iter = newList.iterator();
    while (iter.hasNext()) {
      IAce ace = (IAce)iter.next();
      if (!ace.isGrant()) {
        Set privileges = ace.getPrivileges();

        if (privileges.contains(WDPrivilege.P_WRITE_ACL)) {

          IUMPrincipal principal = ace.getPrincipal();
          if (!newOwners.contains(principal.getId())
              && resourceAcl.isOwner(principal)) {
            resourceAcl.removeOwner(principal);
          }
        }
      }
    }

    if (log.beDebug()) {
      log.debugT("modify Owner, removeList "+removeList+" newList "+newList);
    }

    iter = removeList.iterator();
    while (iter.hasNext()) {
      IAce ace = (IAce)iter.next();
      if (ace.isGrant()) {
        Set privileges = ace.getPrivileges();

        if (privileges.contains(WDPrivilege.P_WRITE_ACL)
            || privileges.contains(WDPrivilege.P_ALL)) {

          IUMPrincipal principal = ace.getPrincipal();
          if (!newOwners.contains(principal.getId())
               && resourceAcl.isOwner(principal)) {
            if (log.beDebug()) {
              log.debugT("removing owner: "+principal.getId());
            }
            resourceAcl.removeOwner(principal);
          }
        }
      }
    }
  }

  protected static ResourceException map(AclException ex, RID rid) {
    if (ex instanceof NotAuthorizedException) {
      return new AccessDeniedException("changing acl", rid, null, null);
    }
    else if (ex instanceof AclPersistenceException) {
      Throwable t = ex.getContainedException();
      if (t instanceof ResourceException) {
        return (ResourceException)t;
      }
      return new IOErrorException("acl persistence", ex, rid);
    }
    else {
      if (log.beDebug()) {
        log.debugT("unexpected exception:" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(ex));
      }
      return new InvalidArgumentException(ex.getMessage());
    }
  }

  private boolean grantsRead(List aceList) throws ResourceException {
    Name sapRead = new Name(Consts.NS_SAP, IAclPermission.ACL_PERMISSION_READ);
    IPrivilege pread = this.supported.find(this.rid, sapRead);
    if (pread == null) {
      pread = this.supported.find(this.rid, Consts.N_read);
    }

    Iterator kiter = aceList.iterator();
    while (kiter.hasNext()) {
      IAce ace = (IAce)kiter.next();
      if (ace.grants(pread)) {
        return true;
      }
    }

    return false;
  }

}