package com.sapportals.wcm.protocol.webdav.repository.acl;

import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.ResourceBundle;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.sap.tc.logging.Location;
import com.sapportals.wcm.protocol.webdav.*;
import com.sapportals.wcm.repository.ResourceException;
import com.sapportals.wcm.util.acl.AclException;
import com.sapportals.wcm.util.acl.IAclPermission;

/** Default implementation of {@link com.sapportals.wcm.protocol.webdav.repository.acl.IPrivilege}.
 *
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author stefan.eissing@greenbytes.de
 * @version $Revision: 1.4 $
 */
class WDPrivilege implements IPrivilege {

  public static final IPrivilege P_ALL;

  public static final IPrivilege P_READ;

  public static final IPrivilege P_WRITE;

  public static final IPrivilege P_WRITE_PROPERTIES;

  public static final IPrivilege P_WRITE_CONTENT;

  public static final IPrivilege P_READ_ACL;

  public static final IPrivilege P_WRITE_ACL;

  public static final IPrivilege P_READ_CUPRIVSET;

  public static final IPrivilege P_UNLOCK;

  public static final IPrivilege P_BIND;

  public static final IPrivilege P_UNBIND;

  private static final Location log = Location.getLocation(WDPrivilege.class);

  private static final Map DAV_NAMES;

  static {
    WDPrivilege wp = new WDPrivilege(N_ALL, new InternationalString("en", "all supported privileges"));
    wp.isAbstract(true);
    P_ALL = wp;

    P_READ = new WDPrivilege(N_READ, new InternationalString("en", "read properties and content"));

    P_WRITE = new WDPrivilege(N_WRITE, new InternationalString("en", "write properties and content"));

    wp = new WDPrivilege(N_READ_ACL, new InternationalString("en", "read access control list"));
    wp.isAbstract(true);
    P_READ_ACL = wp;

    wp = new WDPrivilege(N_WRITE_ACL, new InternationalString("en", "change access control list"));
    wp.isAbstract(false);
    P_WRITE_ACL = wp;

    wp = new WDPrivilege(N_READ_CUPRIVSET, new InternationalString("en", "read current user privilege set"));
    wp.isAbstract(true);
    P_READ_CUPRIVSET = wp;

    wp = new WDPrivilege(N_WRITE_PROPERTIES, new InternationalString("en", "write properties"));
    wp.isAbstract(true);
    P_WRITE_PROPERTIES = wp;

    wp = new WDPrivilege(N_WRITE_CONTENT, new InternationalString("en", "write content"));
    wp.isAbstract(true);
    P_WRITE_CONTENT = wp;

    wp = new WDPrivilege(N_UNLOCK, new InternationalString("en", "unlock resource"));
    wp.isAbstract(true);
    P_UNLOCK = wp;

    wp = new WDPrivilege(N_BIND, new InternationalString("en", "create child resource"));
    wp.isAbstract(true);
    P_BIND = wp;

    wp = new WDPrivilege(N_UNBIND, new InternationalString("en", "remove child resource"));
    wp.isAbstract(true);
    P_UNBIND = wp;

    DAV_NAMES = new HashMap();
    DAV_NAMES.put(IAclPermission.ACL_PERMISSION_READ, N_READ);
    DAV_NAMES.put(IAclPermission.ACL_PERMISSION_WRITE, N_WRITE);
  }

  private final Name name;

  private IInternationalString description;

  private IAclPermission permission;

  private boolean isAbstract;

  private WDPrivilege parent;

  private WDPrivilege firstChild;

  private WDPrivilege lastChild;

  private WDPrivilege nextSibling;

  private WDPrivilege prevSibling;

  public WDPrivilege(Name name, IInternationalString description) {
    this.name = name;
    this.description = description;
    this.permission = null;
  }

  public WDPrivilege(IPrivilege wp, ResourceBundle cmBundle, ResourceBundle davBundle) {
    this.name = wp.getName();
    String key = this.name.toString();
    if (key.length() > 32) {
      key = key.substring(0, 32);
    }
    
    this.description = new InternationalString(davBundle, cmBundle, key);
    this.permission = null;
    this.isAbstract = wp.isAbstract();
  }

  public static WDPrivilege getPrivAll(ResourceBundle cmBundle, ResourceBundle davBundle) {
    return new WDPrivilege(P_ALL, cmBundle, davBundle);
  }

  public static WDPrivilege getPrivRead(ResourceBundle cmBundle, ResourceBundle davBundle) {
    return new WDPrivilege(P_READ, cmBundle, davBundle);
  }

  public static WDPrivilege getPrivWrite(ResourceBundle cmBundle, ResourceBundle davBundle) {
    return new WDPrivilege(P_WRITE, cmBundle, davBundle);
  }

  public static WDPrivilege getPrivReadAcl(ResourceBundle cmBundle, ResourceBundle davBundle) {
    return new WDPrivilege(P_READ_ACL,cmBundle, davBundle);
  }

  public static WDPrivilege getPrivWriteAcl(ResourceBundle cmBundle, ResourceBundle davBundle) {
    return new WDPrivilege(P_WRITE_ACL, cmBundle, davBundle);
  }

  public static WDPrivilege getPrivReadCuprivset(ResourceBundle cmBundle, ResourceBundle davBundle) {
    return new WDPrivilege(P_READ_CUPRIVSET, cmBundle, davBundle);
  }

  public static WDPrivilege getPrivWriteProperties(ResourceBundle cmBundle, ResourceBundle davBundle) {
    return new WDPrivilege(P_WRITE_PROPERTIES, cmBundle, davBundle);
  }

  public static WDPrivilege getPrivWriteContent(ResourceBundle cmBundle, ResourceBundle davBundle) {
    return new WDPrivilege(P_WRITE_CONTENT, cmBundle, davBundle);
  }

  public static WDPrivilege getPrivUnlock(ResourceBundle cmBundle, ResourceBundle davBundle) {
    return new WDPrivilege(P_UNLOCK, cmBundle, davBundle);
  }

  public static WDPrivilege getPrivBind(ResourceBundle cmBundle, ResourceBundle davBundle) {
    return new WDPrivilege(P_BIND, cmBundle, davBundle);
  }

  public static WDPrivilege getPrivUnbind(ResourceBundle cmBundle, ResourceBundle davBundle) {
    return new WDPrivilege(P_UNBIND, cmBundle, davBundle);
  }

  public static WDPrivilege getInstance(IAclPermission permission, Locale locale) throws ResourceException {
    try {
      String pname = permission.getName();
      Name name = null;
      if (pname.startsWith("{")) {
        try {
          name = QualifiedName.getInstance(pname);
        }
        catch (java.text.ParseException ex) {
          // $JL-EXC$
          // ignore
        }
      }

      if (name == null) {
        name = (Name) DAV_NAMES.get(permission.getName());
        if (name == null) {
          name = new Name(Consts.NS_SAP, permission.getName());
        }
      }

      return new WDPrivilege(name, permission, locale);
    }
    catch (AclException ex) {
      // $JL-EXC$
      throw WDAcl.map(ex, null);
    }
  }

  /**
   * Create WDPrivilege from {@link IAclPermission}..
   * @param name of privilege
   * @param permission WCM permission
   * @param locale the language the permission description is using (currently
   * needed because IAclPermission doesn't provide an {@link IInternationalString}).
   */
  private WDPrivilege(Name name, IAclPermission permission, Locale locale) throws AclException {
    this.name = name;
    this.description = new InternationalString(locale.getLanguage(), permission.getDescription(locale));
    this.permission = permission;
  }

  public WDPrivilege(IPrivilege other) {
    this.name = other.getName();
    this.description = other.getDescription();
    this.permission = other.getPermission();
    this.isAbstract = other.isAbstract();
  }

  public int hashCode() {
    return this.name.hashCode();
  }

  public boolean equals(Object o) {
    if (o != null && o instanceof IPrivilege) { return equals((IPrivilege) o); }
    return false;
  }

  public boolean equals(IPrivilege other) {
    return this.name.equals(other.getName());
  }

  public IInternationalString getDescription() {
    return this.description;
  }

  public Name getName() {
    return this.name;
  }

  public boolean isAbstract() {
    return this.isAbstract;
  }

  public Element toElement(Document document) {
    Element element = DomElement.makeNew(document, Consts.N_supported_privilege);
    Element p = DomElement.makeNew(document, Consts.N_privilege);
    element.appendChild(p);
    p.appendChild(DomElement.makeNew(document, this.name));

    if (this.isAbstract) {
      element.appendChild(DomElement.makeNew(document, Consts.N_abstract));
    }

    Element desc = DomElement.makeNew(document, Consts.N_description, this.description);
    element.appendChild(desc);

    if (this.firstChild != null) {
      for (IPrivilege child = this.firstChild; child != null; child = child.getNextSibling()) {
        element.appendChild(child.toElement(document));
      }
    }
    return element;
  }

  public boolean isAncestor(IPrivilege other) {
    if (other == null) { return false; }
    IPrivilege oparent = other.getParent();
    if (oparent != null) {
      if (this.equals(oparent)) { return true; }
      return isAncestor(oparent);
    }
    return false;
  }

  public void isAbstract(boolean is_abstract) {
    this.isAbstract = is_abstract;
  }

  public boolean isContained(IPrivilege privilege) {
    if (this.equals(privilege)) { return true; }

    IPrivilege priv = this.firstChild;
    while (priv != null && priv != this) {

      if (log.beDebug()) {
        log.debugT(this.toString() + ": checking privilege " + priv + " against " + privilege);
      }

      if (priv.equals(privilege)) { return true; }
      IPrivilege next = priv.getFirstChild();
      if (next == null) {
        next = priv.getNextSibling();
        while (next == null && priv != this) {
          priv = priv.getParent();
          next = priv.getNextSibling();
        }
      }
      priv = (priv != this) ? next : null;
    }
    return false;
  }

  /** Return IPermission object represented by this Privilege
   * @return the IPermission or <code>null</code> if there is no mapping
   */
  public IAclPermission getPermission() {
    return this.permission;
  }

  public void setPermission(IAclPermission permission) {
    this.permission = permission;
  }

  public IPrivilege find(Name name) {
    return find(this, name);
  }

  public IPrivilege getParent() {
    return this.parent;
  }

  public IPrivilege getNextSibling() {
    return this.nextSibling;
  }

  public IPrivilege getFirstChild() {
    return this.firstChild;
  }

  public IPrivilege getPrevSibling() {
    return this.prevSibling;
  }

  public IPrivilege getLastChild() {
    return this.lastChild;
  }

  public void appendChild(WDPrivilege privilege) {
    if (privilege.parent != null) {
      privilege.parent.removeChild(privilege);
    }

    privilege.parent = this;
    if (this.lastChild != null) {
      this.lastChild.nextSibling = privilege;
    }
    else {
      this.firstChild = privilege;
    }
    privilege.prevSibling = this.lastChild;
    privilege.nextSibling = null;
    this.lastChild = privilege;
  }

  public void removeChild(WDPrivilege privilege) {
    boolean touched = false;
    if (this.firstChild == privilege) {
      touched = true;
      this.firstChild = privilege.nextSibling;
      if (this.firstChild != null) {
        this.firstChild.prevSibling = null;
      }
    }

    if (this.lastChild == privilege) {
      touched = true;
      this.lastChild = privilege.prevSibling;
      if (this.lastChild != null) {
        this.lastChild.nextSibling = null;
      }
    }

    if (!touched) {
      // Neither head nor tail of child list
      privilege.nextSibling.prevSibling = privilege.prevSibling;
      privilege.prevSibling.nextSibling = privilege.nextSibling;
    }

    privilege.parent = null;
    privilege.nextSibling = privilege.nextSibling = null;
  }

  public String toString() {
    StringBuffer sb = new StringBuffer();
    sb.append("[WDPrivilege ").append(this.name.toString());
    if (this.permission != null) {
      sb.append("(").append(this.permission).append(")");
    }
    sb.append("]");
    return sb.toString();
  }

  // ---------------------------- private ---------------------------------

  private static IPrivilege find(IPrivilege priv, Name name) {
    if (name.equals(priv.getName())) { return priv; }

    for (IPrivilege child = priv.getFirstChild(); child != null; child = child.getNextSibling()) {
      IPrivilege found = find(child, name);
      if (found != null) { return found; }
    }
    return null;
  }
}