package com.sapportals.wcm.protocol.webdav.repository.acl;

import java.util.*;

import com.sap.tc.logging.Location;
import com.sapportals.wcm.protocol.webdav.Consts;
import com.sapportals.wcm.protocol.webdav.Name;
import com.sapportals.wcm.repository.IResource;
import com.sapportals.wcm.repository.ResourceException;
import com.sapportals.wcm.repository.security.IResourceAclManager;
import com.sapportals.wcm.util.acl.AclException;
import com.sapportals.wcm.util.acl.IAclPermission;
import com.sapportals.wcm.util.acl.IAclPermissionList;
import com.sapportals.wcm.util.acl.IAclPermissionListIterator;
import com.sapportals.wcm.util.uri.RID;

/**
 * Represents a tree of DAV:privilege as defined in
 * <a href="http://greenbytes.de/tech/webdav/rfc3744.html#PROPERTY_acl">WebDAV Access Control Protocol</a>.
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author stefan.eissing@greenbytes.de
 * @version $Revision: 1.5 $
 */
final class WDPrivilegeTree implements IPrivilegeTree {

  private static final Location log = Location.getLocation(WDPrivilegeTree.class);

  public static final Name N_SAP_FULL_CONTROL = new Name(Consts.NS_SAP, IAclPermission.ACL_PERMISSION_FULL_CONTROL);
  public static final Name N_SAP_READ = new Name(Consts.NS_SAP, IAclPermission.ACL_PERMISSION_READ);
  public static final Name N_SAP_WRITE = new Name(Consts.NS_SAP, IAclPermission.ACL_PERMISSION_WRITE);
  public static final Name N_SAP_DELETE = new Name(Consts.NS_SAP, IAclPermission.ACL_PERMISSION_DELETE);
  public static final Name N_SAP_READWRITE = new Name(Consts.NS_SAP, IAclPermission.ACL_PERMISSION_READWRITE);

  private static final Map Managers;
  static {
    Managers = Collections.synchronizedMap(new WeakHashMap(100));
  }

  private static class TreeInfo {
    public final IAclPermissionList list;
    public final WDPrivilegeTree tree;

    public TreeInfo(IAclPermissionList list, WDPrivilegeTree tree) {
      this.list = list;
      this.tree = tree;
    }
  }

  protected static IPrivilegeTree getInstance(IResourceAclManager aclmgr, IResource resource)
    throws ResourceException {
    if (log.beDebug()) {
      log.debugT("Building PrivilegeTree for "+resource);
    }
    // Building the tree of WDPrivileges. The root of the tree
    // is the DAV:all privilege.
    // All supported WCM permissions are mapped to either
    // - DAV: privileges (example FULL_CONTROL -> DAV:all)
    // - NS_SAP: privileges, where the Map permissionParent defines
    //   the parent of a NS_SAP privilege. If no parent is defined
    //   the new privilege becomes child of DAV:all
    //
    Locale locale = resource.getContext().getLocale();
    try {
      IAclPermissionList permissions = aclmgr.getSupportedPermissions(resource);

      Map byLocale = (Map)Managers.get(aclmgr);
      if (byLocale != null) {
        TreeInfo ti = (TreeInfo)byLocale.get(locale);
        if (ti != null && ti.list.equals(permissions)) {
          if (log.beDebug()) {
            log.debugT("reusing existing tree");
          }
          return ti.tree;
        }
      }

      ResourceBundle cmBundle = ResourceBundle.getBundle(IAclPermission.class.getName(), locale);
      ResourceBundle davBundle = ResourceBundle.getBundle("com.sapportals.wcm.protocol.webdav.repository.acl.WDDavPrivilege", locale);
      
      List atomics = new ArrayList();
      List meta = new ArrayList();

      // Sort all permissions into atomic and non-atomic which contain other permissions
      //
      for (IAclPermissionListIterator iter = permissions.iterator(); iter.hasNext(); /**/) {
        IAclPermission permission = iter.next();
        if (log.beDebug()) {
          log.debugT("investigating acl permission: "+permission);
        }
        if (permission.isAtomic()) {
          atomics.add(permission);
        }
        else {
          meta.add(permission);
        }
      }

      List treeTops = Collections.EMPTY_LIST;
      if (!meta.isEmpty()) {
        // Remove all contained permissions from the treetop/atomic lists.
        // Make sure we only handle each permission once, in order to avoid
        // cyclic removal of all meta nodes.
        //
        treeTops = new ArrayList(meta);
        List visited = new ArrayList();
        for (Iterator i = meta.iterator(); i.hasNext(); /**/) {
          IAclPermission permission = (IAclPermission)i.next();
          if (!visited.contains(permission)) {
            visited.add(permission);
            for (IAclPermissionListIterator iter = permission.getMembers().iterator(); iter.hasNext(); /**/) {
              IAclPermission contained = iter.next();
              if (contained.isAtomic()) {
                atomics.remove(contained);
              }
              else {
                treeTops.remove(contained);
              }
            }
          }
        }
      }

      // Now, we can have
      // - atomic permissions left over, which are not contained anywhere else. The
      //   permissions are not in a tree and need an artificial root privilege
      // - one treeTop permission left. All non-atomic permissions are in a single tree.
      // - several treeTop permissions left. Use one articial root privilege to combine
      //   the several branches into one tree
      // - there are no meta permissions, all permissions are atomic. Use one artifical
      //   root privilege to contain all atomic permissions
      //
      PrivilegeMapper mapper = PrivilegeMapper.getInstance(locale);
      WDPrivilege root = null;
      if (atomics.isEmpty() && treeTops.size() == 1) {
        // permissions form a single tree
        //
        IAclPermission permission = (IAclPermission)treeTops.get(0);
        root = mapper.getPrivilege(permission,  null);
        // add members of permissions as children of root
        addMembers(permission, root, mapper);
        if (log.beDebug()) {
          log.debugT("single permission tree with top "+permission);
        }
      }
      else {
        // no single tree, we need a new root permission
        //
        if (log.beDebug()) {
          log.debugT("multiple permission tree with tops "+treeTops);
        }
        root = WDPrivilege.getPrivAll(cmBundle, davBundle);
        for (int i = 0, n = treeTops.size(); i < n; ++i) {
          IAclPermission permission = (IAclPermission)treeTops.get(i);
          WDPrivilege priv = mapper.getPrivilege(permission, root);
          // add members of permissions as children of priv
          addMembers(permission, priv, mapper);
        }

        for (int i = 0, n = atomics.size(); i < n; ++i) {
          IAclPermission permission = (IAclPermission)atomics.get(i);
          mapper.getPrivilege(permission, root);
        }
      }
      
      // Now, we have a single privilege tree. Make sure that all WebDAV ACL required
      // privileges are present in the tree. Otherwise they need to be added in the
      // "right" locations as abstract privileges.
      //

      boolean specialAllHandling = false;
      WDPrivilege p_all = (WDPrivilege)root.find(WDPrivilege.N_ALL);
      if (p_all == null) {
        // No DAV:all privilege there, add it at top of tree
        p_all = WDPrivilege.getPrivAll(cmBundle, davBundle);
        // we want to avoid an abstract DAV:all privilege:
        // 1. users like to be able to set DAV:all
        // 2. if the previous root is not abstract, we can map
        // 3. the only other child next to previous root will be
        //    DAV:write-acl which we map to acl owner. This is a special
        //    case anyway.
        if (!root.isAbstract()) {
          p_all.isAbstract(false);
          specialAllHandling = true;
        }
        p_all.appendChild(root);
        root = p_all;
      }

      WDPrivilege p_read = (WDPrivilege)root.find(WDPrivilege.N_READ);
      if (p_read == null) {
        p_read = WDPrivilege.getPrivRead(cmBundle, davBundle);
        p_all.appendChild(p_read);
      }

      WDPrivilege p_read_acl = (WDPrivilege)root.find(WDPrivilege.N_READ_ACL);
      if (p_read_acl == null) {
        p_read_acl = WDPrivilege.getPrivReadAcl(cmBundle, davBundle);
        p_read.appendChild(p_read_acl);
      }

      WDPrivilege p_read_cuprivset = (WDPrivilege)root.find(WDPrivilege.N_READ_CUPRIVSET);
      if (p_read_cuprivset == null) {
        p_read_cuprivset = WDPrivilege.getPrivReadCuprivset(cmBundle, davBundle);
        p_read.appendChild(p_read_cuprivset);
      }

      WDPrivilege p_unlock = (WDPrivilege)root.find(WDPrivilege.N_UNLOCK);
      if (p_unlock == null) {
        p_unlock = WDPrivilege.getPrivUnlock(cmBundle, davBundle);
        p_read.appendChild(p_unlock);
      }

      WDPrivilege p_write = (WDPrivilege)root.find(WDPrivilege.N_WRITE);
      if (p_write == null) {
        p_write = WDPrivilege.getPrivWrite(cmBundle, davBundle);
        p_all.appendChild(p_write);
      }

      WDPrivilege p_write_properties = (WDPrivilege)root.find(WDPrivilege.N_WRITE_PROPERTIES);
      if (p_write_properties == null) {
        p_write_properties = WDPrivilege.getPrivWriteProperties(cmBundle, davBundle);
        p_write.appendChild(p_write_properties);
      }

      WDPrivilege p_write_content = (WDPrivilege)root.find(WDPrivilege.N_WRITE_CONTENT);
      if (p_write_content == null) {
        p_write_content = WDPrivilege.getPrivWriteContent(cmBundle, davBundle);
        p_write.appendChild(p_write_content);
      }

      WDPrivilege p_write_acl = (WDPrivilege)root.find(WDPrivilege.N_WRITE_ACL);
      if (p_write_acl == null) {
        p_write_acl = WDPrivilege.getPrivWriteAcl(cmBundle, davBundle);
        p_all.appendChild(p_write_acl);
      }

      WDPrivilege p_bind = (WDPrivilege)root.find(WDPrivilege.N_BIND);
      if (p_bind == null) {
        p_bind = WDPrivilege.getPrivBind(cmBundle, davBundle);
        p_write_properties.appendChild(p_bind);
      }

      WDPrivilege p_unbind = (WDPrivilege)root.find(WDPrivilege.N_UNBIND);
      if (p_unbind == null) {
        p_unbind = WDPrivilege.getPrivUnbind(cmBundle, davBundle);
        p_write_properties.appendChild(p_unbind);
      }

      WDPrivilegeTree result = new WDPrivilegeTree(p_all, locale, specialAllHandling);
      if (byLocale == null) {
        byLocale = Collections.synchronizedMap(new HashMap(17));
        Managers.put(aclmgr, byLocale);
      }
      TreeInfo ti = new TreeInfo(permissions, result);
      byLocale.put(locale, ti);
      if (log.beDebug()) {
        log.debugT("PrivilegeTree for "+resource+" is "+result);
      }
      return result;
    }
    catch (AclException ex) {
      // $JL-EXC$
      throw WDAcl.map(ex, resource.getRID());
    }
  }

  private final WDPrivilege root;
  private final Locale locale;
  private final boolean specialAllHandling;
  private Map perm2priv;
  private Map name2priv;

  private WDPrivilegeTree(WDPrivilege root, Locale locale, boolean specialAllHandling) {
    this.root = root;
    this.locale = locale;
    this.specialAllHandling = specialAllHandling;

    if (log.beDebug()) {
      StringBuffer sb = new StringBuffer();
      sb.append("constructed tree: ");
      dump(sb, this.root);
      log.debugT(sb.toString());
    }
  }

  public IPrivilege getRoot() {
    return this.root;
  }

  public Locale getLocale() {
    return this.locale;
  }

  public IPrivilege find(RID rid, IAclPermission permission) throws ResourceException {
    try {
      if (this.perm2priv == null) {
        this.perm2priv = new HashMap(17);
        mapPermToPrivilege(this.root);
      }
      return (IPrivilege)this.perm2priv.get(permission.getName());
    }
    catch (AclException ex) {
      // $JL-EXC$
      throw WDAcl.map(ex, rid);
    }
  }

  public IPrivilege find(RID rid, Name name) throws ResourceException {
    try {
      if (this.name2priv == null) {
        this.name2priv = new HashMap(17);
        mapNames(this.root);
      }
      return (IPrivilege)this.name2priv.get(name);
    }
    catch (AclException ex) {
      // $JL-EXC$
      throw WDAcl.map(ex, rid);
    }
  }

  public Set mapPermissions(RID rid, List permissions, boolean canWriteAcl)
    throws ResourceException {
    // 1. Translate all IPermissions to IPrivileges
    // 2. If user can write the ACL, add IPrivilege for DAV:write-acl
    // 3. include all privileges which have all children included
    // 4. remove then from the list all IPrivileges whose ancestor
    //    is already contained in the list
    //
    int len = permissions.size();
    Set privileges = new HashSet(len+2);
    for (Iterator iter = permissions.iterator(); iter.hasNext(); /**/) {
      IAclPermission permission = (IAclPermission)iter.next();
      IPrivilege p = find(rid, permission);
      if (p != null) {
        privileges.add(p);
      }
    }

    IPrivilege write_acl = null;
    if (canWriteAcl) {
      write_acl = find(rid, WDPrivilege.N_WRITE_ACL);
      if (write_acl != null) {
        privileges.add(write_acl);
      }
    }

    compress(privileges);

    if (write_acl != null) {
      privileges.remove(write_acl);
    }

    if (log.beDebug()) {
      log.debugT("mapped acl permissions "+permissions+" to privileges "+privileges);
    }
    return privileges;
  }

  public void compress(Set privileges) {
    if (log.beDebug()) {
      log.debugT("privileges before closure "+privileges);
    }

    includeInClosure(privileges, this.root);
    if (log.beDebug()) {
      log.debugT("privileges after include "+privileges);
    }
    prune(privileges, this.root);

    if (log.beDebug()) {
      log.debugT("privileges after closure "+privileges);
    }
  }

  /**
   * Map a IPrivilege to the list of {@link com.sapportals.wcm.util.acl.IAclPermission}
   * it represents. The list will be empty in case of an abstract IPrivilege
   * @param privilege the IPrivilege to retrieve the IPermissions for
   * @return list of {@link com.sapportals.wcm.util.acl.IAclPermission}s
   */
  public List mapPrivilege(IPrivilege privilege) {
    List result = new ArrayList(17);
    collectPermissions(privilege, result);
    return result;
  }


  // -------------- private -------------------------------

  private void mapPermToPrivilege(IPrivilege p) throws AclException {

    for (IPrivilege child = p.getFirstChild(); child != null; child = child.getNextSibling()) {
      mapPermToPrivilege(child);
    }

    IAclPermission permission = p.getPermission();
    if (permission != null) {
      this.perm2priv.put(permission.getName(), p);
    }
  }

  private void mapNames(IPrivilege p) throws AclException {

    for (IPrivilege child = p.getFirstChild(); child != null; child = child.getNextSibling()) {
      mapNames(child);
    }

    this.name2priv.put(p.getName(), p);
  }

  private boolean includeInClosure(Set privileges, IPrivilege p) {

    if (privileges.contains(p)) {
      // explicitly in closure
      //
      return true;
    }
    else if (p.isAbstract() || (this.specialAllHandling && p.getName().equals(WDPrivilege.N_ALL))) {
      // See, if all children are implicitly or explicitly part of
      // the closure. Special handling for the DAV:all privilege.
      //
      IPrivilege child = p.getFirstChild();
      if (child != null) {
        boolean childrenInClosure = true;
        for (; child != null; child = child.getNextSibling()) {
          if (!includeInClosure(privileges, child)) {
            childrenInClosure = false;
          }
        }
        if (childrenInClosure) {// && p.getPermission() == null) {
          privileges.add(p);
          return true;
        }
      }
      else {
        return true;
      }
    }

    return false;
  }

  private void prune(Set privileges, IPrivilege p) {
    if (privileges.contains(p)) {
      removeDescendants(privileges, p);
    }
    else {
      for (IPrivilege child = p.getFirstChild(); child != null; child = child.getNextSibling()) {
        prune(privileges, child);
      }
    }
  }

  private void removeDescendants(Set privileges, IPrivilege p) {
    for (IPrivilege child = p.getFirstChild(); child != null; child = child.getNextSibling()) {
      removeDescendants(privileges, child);
      privileges.remove(child);
    }
  }

  private void collectPermissions(IPrivilege privilege, List permissions) {
    IAclPermission permission = privilege.getPermission();
    if (permission != null) {
      permissions.add(permission);
    }
    else {
      for (IPrivilege child = privilege.getFirstChild(); child != null; child = child.getNextSibling()) {
        collectPermissions(child, permissions);
      }
    }
  }

  private void dump(StringBuffer sb, IPrivilege priv) {
    sb.append(priv);
    if (priv.getFirstChild() != null) {
      sb.append(" [");
      for (IPrivilege child = priv.getFirstChild(); child != null; child = child.getNextSibling()) {
        dump(sb, child);
        sb.append(", ");
      }
      sb.append("] ");
    }
  }

  private static void addMembers(IAclPermission permission, WDPrivilege priv, PrivilegeMapper mapper)
     throws ResourceException, AclException {
    for (IAclPermissionListIterator i = permission.getMembers().iterator(); i.hasNext(); /**/) {
      IAclPermission child = i.next();
      WDPrivilege childPriv = mapper.getPrivilege(child, priv);
      if (!child.isAtomic()) {
        addMembers(child, childPriv, mapper);
      }
    }
  }

  private static class PrivilegeMapper {

    public static PrivilegeMapper getInstance(Locale locale) {
      return new PrivilegeMapper(locale);
    }

    private final Map byPermission;
    private final Locale locale;

    private PrivilegeMapper(Locale locale) {
      this.byPermission = new HashMap();
      this.locale = locale;
    }

    public WDPrivilege getPrivilege(IAclPermission permission, WDPrivilege parent)
      throws ResourceException {
      WDPrivilege privilege = (WDPrivilege)this.byPermission.get(permission);
      if (privilege != null) {
        // known privilege
        if ((privilege.getParent() == null) || (privilege.getParent().isAncestor(parent))) {
          parent.appendChild(privilege);
        }
        else {
          // do not touch privilege's parent
        }
      }
      else {
        privilege = WDPrivilege.getInstance(permission, this.locale);
        this.byPermission.put(permission, privilege);
        if (parent != null) {
          parent.appendChild(privilege);
        }
      }
      return privilege;
    }
  }
}