package com.sapportals.wcm.protocol.webdav.repository.landscape;

import java.net.MalformedURLException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.List;
import java.util.Properties;
import java.util.StringTokenizer;

import com.sap.tc.logging.Location;
import com.sapportals.portal.security.usermanagement.ILogonDataBroker;
import com.sapportals.portal.security.usermanagement.IUser;

import com.sapportals.wcm.service.landscape.*;
import com.sapportals.wcm.util.http.UserInfo;
import com.sapportals.wcm.util.logging.LoggingFormatter;
import com.sapportals.wcm.util.uri.IHierarchicalUri;
import com.sapportals.wcm.util.uri.UriFactory;

/**
 * Informs about landscape systems.
 * <p>
 * Copyright (c) SAP AG 2002-2003
 * @author stefan.eissing@greenbytes.de
 * @version $Id: ServerInfo.java,v 1.4 2004/06/22 14:21:58 sei Exp $
 */
final class ServerInfo implements IServerInfo {

  private static final Location log = Location.getLocation(ServerInfo.class);

  private static final String PROP_SERVER = "server";
  private static final String PROP_HTTP_MAX = "http.maxConnections";
  private static final String PROP_SAME_USER_DOMAIN = "sameUserDomain";
  private static final String PROP_SERVER_ALIASES = "server.aliases";
  private static final String PROP_APPLY_USER_MAPPING = "applyUserMapping";

  private static final int DEFAULT_HTTP_MAX = Integer.MAX_VALUE;

  private final String systemID;
  private final IHierarchicalUri uri;
  private final UserInfo defaultUserInfo;
  private final BackendConnection connection;
  private final int maxHttpConnections;
  private final boolean isSameUserDomain;
  private final boolean applyUserMapping;
  private final List aliases;

  ServerInfo(ISystem system, ISystemFactory factory) throws MalformedURLException {
    this.systemID = system.getId();
    this.uri = UriFactory.parseHierarchicalUri(system.getProperty(PROP_SERVER));
    this.defaultUserInfo = getDefaultUserInfo(system);
    // Feat for config disabled. Play safe since defaults in config may be wrong...
    this.applyUserMapping = true; //getBoolean(system, PROP_APPLY_USER_MAPPING, true);
    this.isSameUserDomain = getBoolean(system, PROP_SAME_USER_DOMAIN, false);
    this.maxHttpConnections = getInt(system, PROP_HTTP_MAX, DEFAULT_HTTP_MAX);
    this.aliases = getUris(system, PROP_SERVER_ALIASES);
    this.connection = this.applyUserMapping? getBackendConnection(system, factory) : null;
    if (log.beDebug()) {
      log.debugT("new Info about: "+this.systemID+" applyusermapping: "+this.applyUserMapping+" connection: "+this.connection);
    }
  }

  public String toString() {
    return "ServerInfo["+this.systemID+"]";
  }

  //----------- IServerInfo ---------------------------

  public IHierarchicalUri getUri() {
    return this.uri;
  }

  public UserInfo getDefaultUserMapping() {
    return this.defaultUserInfo;
  }

  public UserInfo getUserMappingFor(IUser user) {
    if (this.connection != null) {
      boolean debug = log.beDebug();
      synchronized (this) {
        this.connection.reset();
        ILogonDataBroker broker = user.getMappingInfo();
        if (broker != null) {
          if (debug) log.debugT("got logon broker");
          try {
            broker.enrich(this.connection);
            if (debug) log.debugT("enriched connnection");
            return UserInfo.valueOf(connection.getUser(), connection.getPassword());
          }
          catch (Exception ex) {
            log.infoT("unable to get usermapping of "+user.getId()+" for system "+this.systemID + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(ex));
          }
        }
        else {
          if (debug) log.debugT("no logon broker for system " + this.systemID);
        }
      }
    }

    return UserInfo.ANONYMOUS;
  }

  public int getMaxHttpConnections() {
    return this.maxHttpConnections;
  }

  public boolean isSameUserDomain() {
    return this.isSameUserDomain;
  }

  public List getAliasUris() {
    return this.aliases;
  }

  //----------- private ------------------------------

  private static UserInfo getDefaultUserInfo(ISystem system) {
    return UserInfo.valueOf(system.getProperty("user"), system.getProperty("password"));
  }

  private static BackendConnection getBackendConnection(ISystem system, ISystemFactory factory) {
    Properties credentialsProps = new Properties();
    try {
      if (factory != null) {
        ICredentialsFactory credentialsFactory = factory.getCredentialsFactory();

        if (credentialsFactory != null) {
          ICredentialList credentials = credentialsFactory.getCredentials(system.getType());

          if (credentials != null) {
            ICredentialListIterator iterator = credentials.iterator();

            while (iterator.hasNext()) {
              ICredential credential = iterator.next();
              Enumeration names = credential.getPropertyNames();

              while (names.hasMoreElements()) {

                String name = (String)names.nextElement();
                String value = (String)credential.getProperty(name);

                if (value != null && value.length() > 0)
                  credentialsProps.put(credential.getName()+'.'+name, value);
              }
            }
          }
        }
      }
    }
    catch (Exception ex) {
      log.errorT("retrieving credentials for system "+system.getId() + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(ex));
    }

    if (credentialsProps.isEmpty()) {
      log.warningT("no credentials for system "+system.getId());
      return null;
    }

    return new BackendConnection(system, credentialsProps);
  }

  private int getInt(ISystem system, String propname, int defValue) {
    String value = system.getProperty(propname);
    if (value != null && value.length() > 0) {
      try {
        return Integer.parseInt(value);
      }
      catch (NumberFormatException e) {
        log.warningT("not a number: " + value + "- " + LoggingFormatter.extractCallstack(e));
        // ignore
      }
    }
    return defValue;
  }

  private boolean getBoolean(ISystem system, String propname, boolean defValue) {
    String value = system.getProperty(propname);
    if (value != null && value.length() > 0) {
      return "true".equalsIgnoreCase(value.trim());
    }
    return defValue;
  }

  private List getUris(ISystem system, String propname) throws MalformedURLException {
    String value = system.getProperty(propname);
    if (value != null && value.length() > 0) {
      List result = new ArrayList();
      StringTokenizer tok = new StringTokenizer(value, " \t\r\n");
      while (tok.hasMoreTokens()) {
        result.add(UriFactory.parseUri(tok.nextToken()));
      }
      return result;
    }
    return Collections.EMPTY_LIST;
  }
}
