package com.sapportals.wcm.protocol.webdav.repository.monitor;

import java.util.Date;
import java.util.Locale;
import java.util.Properties;

import com.sap.tc.logging.Location;
import com.sapportals.wcm.crt.component.ComponentState;
import com.sapportals.wcm.crt.component.ComponentStateHandler;
import com.sapportals.wcm.crt.component.ContextException;
import com.sapportals.wcm.crt.component.IComponent;
import com.sapportals.wcm.crt.component.IComponentInfo;
import com.sapportals.wcm.crt.component.IComponentManager;
import com.sapportals.wcm.crt.component.IContainer;
import com.sapportals.wcm.crt.component.IContainerManager;
import com.sapportals.wcm.crt.component.IContext;
import com.sapportals.wcm.crt.component.IContextualizable;
import com.sapportals.wcm.crt.component.ILifecycleInfo;
import com.sapportals.wcm.crt.component.IThreadSafe;
import com.sapportals.wcm.crt.component.StartupException;
import com.sapportals.wcm.crt.configuration.ConfigurationException;

/**
 * Holds runtime (crt) objects for each registered IServerComponent.
 * <p>
 * Copyright (c) SAP AG 2002-2003
 * @author stefan.eissing@greenbytes.de
 * @version $Id: ServerCollection.java,v 1.4 2004/07/15 13:06:59 sei Exp $
 */
public final class ServerCollection implements IComponent, IContainer, 
  IContextualizable, IThreadSafe, ILifecycleInfo, IComponentInfo, IServerRegistry {

  private static final Location log = Location.getLocation(ServerCollection.class);

  public static final String NAME = "servers";
  
  private String name;
  private final ComponentStateHandler stateHandler;
  private IContainerManager containerManager;
  private ConfigurationException dummyException;

  public ServerCollection() {
    this.name = NAME;
    this.stateHandler = new ComponentStateHandler(this);
  }
  
  // ---------- IServerRegistry ------------------------------------------------
  
  public void register(IServerComponent server) throws Exception {
    String key = ServerProxy.getKey(server, false);
    if (this.containerManager.hasComponent(key)) {
      // name clash, try longer version
      key = ServerProxy.getKey(server, true);
    }
    if ( ! this.containerManager.hasComponent(key)) {
      server.setID(key);
      this.containerManager.addChild(key, ServerProxy.class.getName(), null, null, server);
    }
  }
  
  public void unregister(IServerComponent server) {
    this.containerManager.removeChild(server.getID());
  }
  
  // ---------- IContainer------------------------------------------------------

  public IComponentManager getComponentManager() {
    return this.containerManager;
  }
  
  //----------- IContextualizable ----------------------------------------------
  
	public void contextualize(IContext context) throws ContextException {
    this.stateHandler.preContextualize();
    this.containerManager = context.getContainerManager();
    Object parm = context.getParameters();
    if (parm instanceof String) {
      this.name = (String)parm;
    }
    this.stateHandler.postContextualize();
	}

  //----------- ILifecycleInfo -------------------------------------------------

  public Date getCreationDate() {
    return this.stateHandler.getCreationDate();
  }

  public Date getLastReconfigurationDate() {
    return this.stateHandler.getLastReconfigurationDate();
  }

  public Date getNextAutoRestartDate() {
    return this.stateHandler.getNextAutoRestartDate();
  }

  public StartupException getStartupException() {
    return this.stateHandler.getStartupException();
  }

  public ComponentState getState() {
    // Combine state of all child components:
    // If not all components are running and config is OK create a fake state and exception so
    // that the CRT repository will return a WARN or ERROR mimetype (yellow or red bullet!).
    String[] childKeys = this.getComponentManager().listComponentKeys();
    int countNotRunning = 0;
    int countNotInit = 0;
    int countConfigException = 0;
    IComponent comp = null;
    for (int i = 0; i < childKeys.length; i++) {
      try {
        comp = this.getComponentManager().lookupComponent(childKeys[i]);
        ComponentState state = ((ILifecycleInfo)comp).getState();
        if (state != null) {
          if (ComponentState.NOT_INIT.equals(state)) {
            countNotInit++;
          }
          else if ( ! ComponentState.RUNNING.equals(state)) {
            countNotRunning++;
          }
        }
        if (((ILifecycleInfo)comp).getLastConfigurationException() != null) {
          countConfigException++;
        }
      }
      catch (Exception ex) {
        com.sap.tc.logging.Location.getLocation(this.getClass()).debugT(
          com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(ex));
      }
      finally {
        if (comp != null) {
          this.getComponentManager().releaseComponent(comp);
        }
      }
    }
    this.dummyException = null;
    if (countNotInit > 0 && countNotInit == childKeys.length) {
      return ComponentState.NOT_INIT;
    }
    else if (countNotRunning == childKeys.length && childKeys.length > 0) {
      return ComponentState.INIT_ERROR;
    }
    else if (countNotRunning > 0 || countConfigException > 0) {
      this.dummyException = new ConfigurationException("");
    }
    return ComponentState.RUNNING;
  }

  public ConfigurationException getLastConfigurationException() {
    return this.dummyException;
  }

  //----------- IComponentInfo -------------------------------------------------

	public String getDescription(Locale locale) {
		return "Lists all HTTP Servers used in this Web Repository";
	}

	public String getName() {
		return this.name;
	}

	public Properties getProperties() {
		// nothing to report yet
		return null;
	}

  //----------- private --------------------------------------------------------
  
}
