package com.sapportals.wcm.protocol.webdav.repository.monitor;

import java.io.IOException;
import java.io.InputStream;
import java.io.InterruptedIOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.Iterator;
import java.util.List;

import org.w3c.dom.Document;

import com.sapportals.wcm.WcmException;
import com.sapportals.wcm.repository.ServiceUnavailableException;
import com.sapportals.wcm.util.http.ConnectException;
import com.sapportals.wcm.util.http.DateFormat;
import com.sapportals.wcm.util.http.IRequest;
import com.sapportals.wcm.util.http.IRequester;
import com.sapportals.wcm.util.http.IResponse;
import com.sapportals.wcm.util.uri.IHierarchicalUri;

/**
 * Hands out status information for a particular http server.
 * <p>
 * Copyright (c) SAP AG 2002-2003
 * @author stefan.eissing@greenbytes.de
 * @version $Id: ServerComponent.java,v 1.2 2004/06/25 11:54:43 sei Exp $
 */
public final class ServerComponent implements IServerComponent {
  
  private static final int MAX_WEIGHT = 100;
  private static final int MAX_FAILURES = 10;
  private static final int MAX_ERRORS = 10;
  
  
  public static ServerComponent getInstance(IHierarchicalUri uri) {
    return new ServerComponent(uri);
  }
  
  private String id;
  private final IHierarchicalUri serverUri;
  
  private int requestCount;
  private int failureCount;
  private int errorCount;
  private int successCount;
  private long responseAvgMS;
  private int responseWeight;
  private int bytesPerSec;
  private int bytesWeight;
  private final List errors;
  private final List failures;
  private String ceaseMessage;
  private long ceaseOperationUntil;
  
  private ServerComponent(IHierarchicalUri uri) {
    this.serverUri = uri;
    this.failures = new ArrayList(MAX_FAILURES);
    this.errors = new ArrayList(MAX_ERRORS);
    this.responseWeight = 1;
    this.bytesWeight = 1;
  }
  
  //----------- IServerComponent -----------------------------------------------
  
  public synchronized String getID() {
    return this.id;
  }
  
  public synchronized void setID(String id) {
    this.id = id;
  }
  
  public IHierarchicalUri getServerUri() {
    return this.serverUri;
  }

  public synchronized IStatistics getSnapshot() {
    return Statistics.getInstance(this.requestCount, this.failureCount, this.errorCount,
      this.successCount, this.responseAvgMS, this.bytesPerSec);
  }
  
  public synchronized List getLastErrors() {
    if (this.errors.isEmpty()) {
      return Collections.EMPTY_LIST;
    }
    return new ArrayList(this.errors);
  }
  
  public synchronized List getLastFailures() {
    if (this.failures.isEmpty()) {
      return Collections.EMPTY_LIST;
    }
    return new ArrayList(this.failures);
  }
  
  //----------- ServerComponent ------------------------------------------------

  public IResponse perform(IRequester requester, IRequest request) throws WcmException {
    return perform(requester, request, TimeGuard.getDummy());
  }
                           
  public IResponse perform(IRequester requester, IRequest request, ITimeGuard guard) 
  throws WcmException {
    synchronized(this) {
      if (this.ceaseOperationUntil != 0) {
        if (System.currentTimeMillis() < this.ceaseOperationUntil) {
          Date until = new Date(this.ceaseOperationUntil);
          throw new ServiceUnavailableException(
            this.ceaseMessage+", ceasing operation until "+until, until, null);
        }
        this.ceaseOperationUntil = 0;
        this.ceaseMessage = null;
      }
      ++this.requestCount;
    }
    long duration = System.currentTimeMillis();
    try {
      guard.check(duration);
      guard.adjustSOTimeout(requester);
      IResponse response = requester.perform(request);
    
      duration = System.currentTimeMillis() - duration;
      this.gotResponse(response, request, duration);
      
      return MonitorResponse.getInstance(request, response, guard, this);    
    }
    catch (ConnectException ex) {
      this.gotException(ex, request);
      // Problems on connect. The exception gives a recommended retry interval which
      // we use to fail early on future operations
      synchronized (this) {
        this.ceaseMessage = ex.getMessage();
        this.ceaseOperationUntil = System.currentTimeMillis() + (ex.getRetryInSec() * 1000);
      }
      throw ex;
    }
    catch (WcmException ex) {
      this.gotException(ex, request);
      throw ex;
    }
    catch (RuntimeException ex) {
      this.gotException(ex, request);
      throw ex;
    }
    catch (Error ex) {
      this.gotException(ex, request);
      throw ex;
    }
  }
  
  public synchronized void gotException(Throwable ex, IRequest request) {
    ++this.failureCount;
    DateFormat df = new DateFormat();
    StringBuffer sb = new StringBuffer();
    sb.append(df.formatISO8601(new Date()));
    sb.append(": ");
    sb.append(request.getMethod());
    sb.append(" ");
    sb.append(request.getURI());
    sb.append(": ");
    sb.append(ex.getClass().getName());
    sb.append(": ");
    sb.append(ex.getMessage());
    if (ex instanceof WcmException) {
      WcmException wex = (WcmException)ex;
      Throwable t = wex.getContainedException();
      if (t != null) {
        sb.append(" (");
        sb.append(t.getClass().getName());
        sb.append(": ");
        sb.append(t.getMessage());
        sb.append(")");
      }
    }
    add(sb.toString(), this.failures, MAX_FAILURES);
  }

  //----------- private --------------------------------------------------------

  private void gotThroughput(int bytes, long duration) {
    int measurement = (int)((bytes * 1000L) / duration);
    this.bytesPerSec = 
      (int)((((long)this.bytesPerSec * (this.bytesWeight-1)) + measurement + (this.bytesWeight / 2))
        / this.bytesWeight); 
    if (this.bytesWeight < MAX_WEIGHT) {
      ++this.bytesWeight;
    }
  }
  
  private synchronized void gotResponse(IResponse response, IRequest request,  long duration) {
    int status = response.getStatus();
    if (status < 400) {
      ++this.successCount;
    }
    else {
      DateFormat df = new DateFormat();
      StringBuffer sb = new StringBuffer();
      sb.append(df.formatISO8601(new Date()));
      sb.append(": ");
      sb.append(request.getMethod());
      sb.append(" ");
      sb.append(request.getURI());
      sb.append(": ");
      sb.append(response.getStatus());
      sb.append(" ");
      sb.append(response.getStatusDescription());
      
      if (status >= 400 && status < 500) {
        ++this.errorCount;
        add(sb.toString(), this.errors, MAX_ERRORS);
      }
      else {
        ++this.failureCount;
        add(sb.toString(), this.failures, MAX_FAILURES);
      }      
    }
    
    this.responseAvgMS = ((this.responseAvgMS * (this.responseWeight-1)) + duration + (this.responseWeight / 2)) / (this.responseWeight);
    if (this.responseWeight < MAX_WEIGHT) {
      ++this.responseWeight;
    }
  }
  
  private void add(String s, List items, int max)  {
    if (items.size() >= max) {
      items.remove(0);
    }
    items.add(s);
  }

  private static class Statistics implements IStatistics {
    
    public static IStatistics getInstance(int requests, int failures, int errors, 
      int successes, long responseAvgMS, int bytesPerSecond) {
      return new Statistics(requests, failures, errors, successes, responseAvgMS, bytesPerSecond); 
    }
    
    private final int requests;
    private final int failures;
    private final int errors;
    private final int successes;
    private final long responseAvgMS;
    private final int bytesPerSecond;
      
    private Statistics(int requests, int failures, int errors, 
      int successes, long responseAvgMS, int bytesPerSecond) {
      this.requests = requests;
      this.failures = failures;
      this.errors = errors;
      this.successes = successes;
      this.responseAvgMS = responseAvgMS;
      this.bytesPerSecond = bytesPerSecond;
    }
    
    public int getErrorCount() {
      return this.errors;
    }

    public int getFailureCount() {
      return this.failures;
    }

    public int getRequestCount() {
      return this.requests;
    }

    public long getResponseTimeAverageMS() {
      return this.responseAvgMS;
    }

    public int getSuccessCount() {
      return this.successes;
    }

    public int getBytesPerSecond() {
      return this.bytesPerSecond;
    }
  }
  
  private static class MonitorResponse implements IResponse {
    public static IResponse getInstance(IRequest request, IResponse response, 
      ITimeGuard guard, ServerComponent server) {
      if (response.hasStream()) {
        return new MonitorResponse(request, response, guard, server);
      }
      return response;
    }
    
    private final IRequest request;
    private final IResponse response;
    private final ITimeGuard guard;
    private final ServerComponent server;
    private InputStream monitorStream;
    
    private MonitorResponse(IRequest request, IResponse response, 
      ITimeGuard guard, ServerComponent server) {
      this.request = request;
      this.response = response;
      this.guard = guard;
      this.server = server;
    }
    
    
    public long getContentLength() {
      return this.response.getContentLength();
    }

    public String getContentType() {
      return this.response.getContentType();
    }

    public Document getDocument() {
      return this.response.getDocument();
    }

    public String getHeader(String name) {
      return this.response.getHeader(name);
    }

    public Iterator getHeaderNames() {
      return this.response.getHeaderNames();
    }

    public String getHttpVersion() {
      return this.response.getHttpVersion();
    }

    public int getStatus() {
      return this.response.getStatus();
    }

    public String getStatusDescription() {
      return this.response.getStatusDescription();
    }

    public InputStream getStream() {
      if (this.monitorStream == null) {
        this.monitorStream = MonitorStream.getInstance(
          this.response.getStream(), this.server, this.request, this.guard);
      }
      return this.monitorStream;
    }

    public boolean hasDocument() {
      return this.response.hasDocument();
    }

    public boolean hasStream() {
      return this.response.hasStream();
    }

  }
  
  private static final class MonitorStream extends InputStream {
    public static InputStream getInstance(
      InputStream in, ServerComponent server, IRequest request, ITimeGuard guard) {
      return new MonitorStream(in, server, request, guard);
    }
    
    private final InputStream in;
    private final ITimeGuard guard;
    private final ServerComponent server;
    private final IRequest request;
    private long start;
    private int bytesRead;
    private int bytesMarked;
    private boolean started;
    private boolean reported;
    
    private MonitorStream(InputStream in, ServerComponent server, 
      IRequest request, ITimeGuard guard) {
      this.in = in;
      this.server = server;
      this.request = request;
      this.guard = guard;
    }
    
    public int available() throws IOException {
      return this.in.available();
    }

    public void close() throws IOException {
      this.in.close();
      report();
    }

    public void mark(int readlimit) {
      this.in.mark(readlimit);
      // remember current position
      this.bytesMarked = this.bytesRead;
    }

    public void reset() throws IOException {
      this.in.reset();
      this.bytesRead = this.bytesMarked;
    }

    public boolean markSupported() {
      return this.in.markSupported();
    }

    public int read() throws IOException {
      start();
      int n = this.in.read();
      if (n != -1) {
        ++this.bytesRead;
      }
      else {
        report();
      }
      return n;
    }

    public int read(byte[] b, int off, int len) throws IOException {
      start();
      try {
        int bytes = this.in.read(b, off, len);
        if (bytes == -1) {
          report();
        }
        else {
          this.bytesRead += bytes;
        }
        return bytes;
      }
      catch (IOException ex) {
        this.reported = true;
        this.server.gotException(ex, this.request);
        throw ex;
      }
    }

    public int read(byte[] b) throws IOException {
      return this.read(b, 0, b.length);
    }

    public long skip(long n) throws IOException {
      return this.in.skip(n);
    }

    private void start() throws IOException {
      try {
        if ( ! this.started) {
          this.started = true;
          this.start = System.currentTimeMillis();
          this.guard.check(this.start);
        }
        else {
          this.guard.check();
        }
      }
      catch (TimeExceededException ex) {
        this.reported = true;
        this.server.gotException(ex, this.request);
        throw new InterruptedIOException(ex.getMessage());        
      }
    }

    private void report() {
      if ( ! this.reported) {
        this.reported = true;
        if (this.bytesRead > 0) {
          long duration = System.currentTimeMillis() - this.start;
          if (duration > 0) {
            this.server.gotThroughput(this.bytesRead, duration);
          }
        }
      }
    }
  }
}
