package com.sapportals.wcm.protocol.webdav.repository.monitor;

import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.Properties;

import com.sapportals.wcm.crt.component.ComponentState;
import com.sapportals.wcm.crt.component.ComponentStateHandler;
import com.sapportals.wcm.crt.component.ContextException;
import com.sapportals.wcm.crt.component.IComponent;
import com.sapportals.wcm.crt.component.IComponentInfo;
import com.sapportals.wcm.crt.component.IContext;
import com.sapportals.wcm.crt.component.IContextualizable;
import com.sapportals.wcm.crt.component.ILifecycleInfo;
import com.sapportals.wcm.crt.component.IThreadSafe;
import com.sapportals.wcm.crt.component.StartupException;
import com.sapportals.wcm.crt.configuration.ConfigurationException;
import com.sapportals.wcm.util.uri.HttpUrl;
import com.sapportals.wcm.util.uri.IHierarchicalUri;

public class ServerProxy implements IComponent, IContextualizable, IThreadSafe, ILifecycleInfo, IComponentInfo {

  public static String getKey(IServerComponent server, boolean longVersion) {
    IHierarchicalUri uri = server.getServerUri();
    if (longVersion) {
      StringBuffer key = new StringBuffer();
      key.append(uri.getAuthority());
      String path = uri.getPath();
      if (path != null && path.length() > 0 && ! "/".equals(path)) {
        key.append(" (");
        String[] segments = uri.getPathSegments();
        for (int i = 0; i < segments.length; ++i) {
          if (i != 0) key.append(' ');
          key.append(segments[i]);
        }
        key.append(")");
      }
      return key.toString();
    }
    else {
      if (uri instanceof HttpUrl) {
        return ((HttpUrl)uri).getHost();
      }
      return uri.getAuthority();
    }
  }

  private final ComponentStateHandler stateHandler;
  private IServerComponent server;
  private String name;
  
  public ServerProxy() {
    this.stateHandler = new ComponentStateHandler(this);
  }
  
  //----------- IContextualizable ----------------------------------------------
  
  public void contextualize(IContext context) throws ContextException {
    this.stateHandler.preContextualize();
    this.server = (IServerComponent)context.getParameters();
    this.name = this.server.getID();
    this.stateHandler.postContextualize();
  }

  //----------- ILifecycleInfo -------------------------------------------------

  public Date getCreationDate() {
    return this.stateHandler.getCreationDate();
  }

  public ConfigurationException getLastConfigurationException() {
    return this.stateHandler.getLastConfigurationException();
  }

  public Date getLastReconfigurationDate() {
    return this.stateHandler.getLastReconfigurationDate();
  }

  public Date getNextAutoRestartDate() {
    return this.stateHandler.getNextAutoRestartDate();
  }

  public StartupException getStartupException() {
    return this.stateHandler.getStartupException();
  }

  public ComponentState getState() {
    IServerComponent.IStatistics stat = this.server.getSnapshot();
    if (stat.getRequestCount() > 0) {
      if (stat.getSuccessCount() > 0) {
        if ((stat.getFailureCount() + stat.getErrorCount()) >= stat.getSuccessCount()) {
          return ComponentState.INIT_ERROR;
        }
        else {
          return ComponentState.RUNNING;
        }
      }
      else {
        // no successes, did we have failures or errors?
        if (stat.getFailureCount() > 0 || stat.getErrorCount() > 0) {
          return ComponentState.INIT_ERROR;
        }
        else {
          return ComponentState.NOT_INIT;
        }
      }
    }
    else {
     return ComponentState.NOT_INIT;
    }
  }

  //----------- IComponentInfo -------------------------------------------------

  public String getDescription(Locale locale) {
    return this.server.getServerUri().toString();
  }

  public String getName() {
    return this.name;
  }

  public Properties getProperties() {
    IServerComponent.IStatistics stat = this.server.getSnapshot();
    Properties props = new Properties();
    props.setProperty("requests", Integer.toString(stat.getRequestCount()));
    props.setProperty("errors", Integer.toString(stat.getErrorCount()));
    props.setProperty("failures", Integer.toString(stat.getFailureCount()));
    props.setProperty("successes", Integer.toString(stat.getSuccessCount()));
    props.setProperty("response-average", Long.toString(stat.getResponseTimeAverageMS()));
    props.setProperty("bytes-per-second", Integer.toString(stat.getBytesPerSecond()));
    addPropList("last-error", this.server.getLastErrors(), props);
    addPropList("last-failure", this.server.getLastFailures(), props);
    return props;
  }

  //----------- private --------------------------------------------------------
  
  private void addPropList(String prefix, List items, Properties props) {
    for (int i = 0, n = items.size(); i < n; ++i) {
      props.setProperty(prefix+"-"+i, items.get(i).toString());
    }
  }
}
