package com.sapportals.wcm.protocol.webdav.repository.monitor;

import com.sap.tc.logging.Location;
import com.sapportals.wcm.IWcmConst;
import com.sapportals.wcm.repository.IResource;
import com.sapportals.wcm.repository.IResourceContext;
import com.sapportals.wcm.repository.ResourceException;
import com.sapportals.wcm.util.http.IRequester;
import com.sapportals.wcm.util.uri.RID;

/**
 * Guards time usage.
 * <p>
 * Copyright (c) SAP AG 2002-2004
 * @author stefan.eissing@greenbytes.de
 * @version $Id: TimeGuard.java,v 1.2 2004/04/19 12:48:06 jre Exp $
 */
public final class TimeGuard {

  private static final Location log = Location.getLocation(TimeGuard.class);
  
  public static final String GUARD_KEY = 
    IWcmConst.SAP_WCM_NAMESPACE + IWcmConst.NAMESPACE_SEPARATOR
     + "repository" + IWcmConst.NAMESPACE_SEPARATOR
     + "web" + IWcmConst.NAMESPACE_SEPARATOR
     + "guard";
    
  // Change to this, once IWcmConsts crawls again over the wall 
  //private static final String RC_LOOKUP_TIME_MS = IWcmConst.CACHE_MAX_LOOKUP_TIME_MS;
  private static final String RC_LOOKUP_TIME_MS = 
    IWcmConst.SAP_WCM_NAMESPACE + IWcmConst.NAMESPACE_SEPARATOR + "cache_max_lookup_time_ms";
  
  /**
   * Install a new {@link ITimeGuard} in the given context
   * @param context to install {@link ITimeGuard} for
   * @param operation name of operation
   * @param rid on which operation takes place
   * @return installed {@link ITimeGuard} instance
   */
  public static ITimeGuard install(IResourceContext context, String operation, RID rid) {
    Object o = context.getObjectValue(RC_LOOKUP_TIME_MS);
    if (o instanceof Long) {
      long timeout = ((Long)o).longValue();
      if (timeout > 0) {
        if (log.beDebug()) {
          log.debugT("Installing ITimeGuard for "+operation+" on "+rid
            + " with timeout of "+timeout+" ms");
        }
        ActiveGuard guard = new ActiveGuard(context, timeout, operation, rid);
        guard.activate();
        return guard;       
      }
    }
    if (log.beDebug()) {
      log.debugT("No timeout set for "+operation+" on "+rid);
    }
    return InactiveGuard.INSTANCE;
  }
  
  /**
   * Install a new {@link ITimeGuard} in the given context
   * @param operation name of operation
   * @param resource on which operation takes place
   * @return installed {@link ITimeGuard} instance
   * @throws ResourceException on general failure
   */
  public static ITimeGuard install(String operation, IResource resource) 
  throws ResourceException {
    return install(resource.getContext(), operation, resource.getRID());
  }
  
  /**
   * Perform a check if a installed time limit has been exceeded.
   * @param context to check in
   * @throws TimeExceededException if time limit has been exceeded.
   */
  public static void check(IResourceContext context) throws TimeExceededException {
    Object o = context.getObjectValue(GUARD_KEY);
    if (o instanceof ITimeGuard) {
      ((ITimeGuard)o).check();
    }
  }
  
  /**
   * Retrieve the installed {@link ITimeGuard} from the given context.
   * The method never returns <code>null</code>. If not guard is installed,
   * a guard instance which never fails a check is returned.
   * @param context to get guard for
   * @return installed {@link ITimeGuard} or dummy instance
   */
  public static ITimeGuard getInstalled(IResourceContext context) {
    Object o = context.getObjectValue(GUARD_KEY);
    if (o instanceof ITimeGuard) {
      return (ITimeGuard)o;
    }
    return InactiveGuard.INSTANCE;
  }
  
  /**
   * Returns a dummy {@link ITimeGuard} instance which never fails any check
   * @return dummy instance which never fails
   */
  public static ITimeGuard getDummy() {
    return InactiveGuard.INSTANCE;  
  }
  
  private static final class ActiveGuard implements ITimeGuard {
    
    private final long timeout;
    private final long limit;
    private final IResourceContext context;
    private final String operation;
    private final RID rid;
    
    public ActiveGuard(IResourceContext context, long timeout, String op, RID rid) {
      this.context = context;
      this.timeout = timeout;
      this.limit = System.currentTimeMillis() + this.timeout;
      this.operation = op;
      this.rid = rid;
    }
    
		public void check() throws TimeExceededException {
      check(System.currentTimeMillis());
		}

    public void check(long time) throws TimeExceededException {
      if (time > this.limit) {
        StringBuffer sb = new StringBuffer(128);
        sb.append("time limit of ").append(this.timeout);
        sb.append(" ms has been exceeded for operation ");
        sb.append(this.operation);
        if (this.rid != null) {
          sb.append(" on ").append(this.rid);
        }
        throw new TimeExceededException(sb.toString());
      }
    }

		public void deactivate() {
      try {
        this.context.setObjectValue(GUARD_KEY, null);
      }
      catch (ResourceException ex) {
        // $JL-EXC$
        // ignore
      }
		}

    public void activate() {
      try {
        this.context.setObjectValue(GUARD_KEY, this);
      }
      catch (ResourceException ex) {
        // $JL-EXC$
        // ignore
      }
    }
    
    public long getRemainingMS() {
      return this.limit - System.currentTimeMillis();
    }

    public void adjustSOTimeout(IRequester requester) {
      long remaining = this.getRemainingMS();
      if (remaining > 0 && remaining != Long.MAX_VALUE) {
        // Make sure that the timeout for http socket operations does not exceed
        // the time left in the timeguard
        //
        int httpTimeout = requester.getSoTimeout();
        if ((httpTimeout == 0) || (httpTimeout > remaining)) {
          int newTimeout = Integer.MAX_VALUE;
          if (remaining < Integer.MAX_VALUE) {
            newTimeout = (int)remaining;
          }
          if (log.beDebug()) {
            log.debugT("Adjusting HTTP Socket Operation timeout from "+httpTimeout+" to "+newTimeout+" ms");
          }
          requester.setSoTimeout(newTimeout);
        }
      }
    }
  }
  
  private static final class InactiveGuard implements ITimeGuard {
    
    public static final InactiveGuard INSTANCE = new InactiveGuard();
    
    private InactiveGuard() {
    }
    
		public void check() throws TimeExceededException {
      // there is not time guard set, nothing to do
		}

    public void check(long time) throws TimeExceededException {
      // there is not time guard set, nothing to do
    }

		public void deactivate() {
      // we did not modify the context, nothing to do
		}
    
    public long getRemainingMS() {
      return Long.MAX_VALUE;
    }
    
		public void adjustSOTimeout(IRequester requester) {
      // nothing to do
		}

  }
}
