package com.sapportals.wcm.protocol.webdav.repository.um;

import java.util.*;

import com.sap.tc.logging.Location;
import com.sapportals.portal.security.usermanagement.*;
import com.sapportals.wcm.IWcmConst;
import com.sapportals.wcm.protocol.webdav.Consts;
import com.sapportals.wcm.protocol.webdav.PropertyDescription;
import com.sapportals.wcm.protocol.webdav.repository.WDPrincipals;
import com.sapportals.wcm.repository.IPropertyName;
import com.sapportals.wcm.repository.IncompleteOperationException;
import com.sapportals.wcm.repository.PropertyName;
import com.sapportals.wcm.repository.ResourceException;
import com.sapportals.wcm.util.name.IName;
import com.sapportals.wcm.util.uri.IRidSet;
import com.sapportals.wcm.util.uri.RID;
import com.sapportals.wcm.util.uri.RidSet;
import com.sapportals.wcm.util.usermanagement.WPUMFactory;


/**
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author julian.reschke@greenbytes.de
 * @version $Revision: 1.1.1.1 $
 */

public class WDPrincipalSearch  {

  private static final Location log = Location.getLocation(WDPrincipalSearch.class);

  private static final String NS_UM_BASE =
    IWcmConst.SAP_WCM_NAMESPACE + IWcmConst.NAMESPACE_SEPARATOR +
    "um" + IWcmConst.NAMESPACE_SEPARATOR;

  private static final String NS_UM_USERS = NS_UM_BASE + "corporateLDAP";

  private static final String NS_UM_COLLECTIONS = NS_UM_BASE;

  private static final String[] SA = new String[0];
  private static final PropertyDescription DISPLAYNAME;

  static {
    try {
      DISPLAYNAME = new PropertyDescription(new PropertyName(Consts.N_displayname), "en", "display name");
    }
    catch (ResourceException ex) {
      throw new RuntimeException(ex.getMessage());
    }
  }

  public static Set getSearchableProperties() throws UserManagementException, ResourceException {
    Set result = new HashSet();
    result.addAll(getSearchableUserProperties());
    result.addAll(getSearchableRoleProperties());
    result.addAll(getSearchableGroupProperties());
    return result;
  }

  private static Set getSearchableUserProperties() throws UserManagementException, ResourceException {

    Set result = new HashSet();
    ISearchUser su = WPUMFactory.getUserFactory().getSearchUser();

    String keys[] = su.getSearchKeys();

    if (keys != null) {
      for (int i = 0; i < keys.length; i++) {
        // uses key name as description for now
        result.add(new PropertyDescription (new PropertyName(NS_UM_USERS, keys[i]), "en", keys[i]));
        if (log.beDebug()) {
          log.debugT("searchable user key: " + new PropertyName(NS_UM_USERS, keys[i]));
        }
      }
    }

    // displayname is always searchable
    result.add(DISPLAYNAME);

    return result;
  }

  private static Set getSearchableGroupProperties() throws UserManagementException, ResourceException {

    Set result = new HashSet();
    ISearchGroup sg = WPUMFactory.getGroupFactory().getSearchGroup();

    String keys[] = sg.getSearchKeys();

    if (keys != null) {
      for (int i = 0; i < keys.length; i++) {
        // uses key name as description for now
        result.add (new PropertyDescription (new PropertyName(NS_UM_COLLECTIONS, keys[i]), "en", keys[i] + " *"));
        if (log.beDebug()) {
          log.debugT("searchable group key: " + new PropertyName(NS_UM_COLLECTIONS, keys[i]));
        }
      }
    }

    // displayname is always searchable
    result.add(DISPLAYNAME);

    return result;
  }

  private static Set getSearchableRoleProperties() throws UserManagementException, ResourceException {

    Set result = new HashSet();
    ISearchRole sr = WPUMFactory.getRoleFactory().getSearchRole();

    String keys[] = sr.getSearchKeys();

    if (keys != null) {
      for (int i = 0; i < keys.length; i++) {
        // uses key name as description for now
        result.add (new PropertyDescription(new PropertyName(NS_UM_COLLECTIONS, keys[i]), "en", keys[i]  + " *"));
        if (log.beDebug()) {
          log.debugT("searchable role key: " + new PropertyName(NS_UM_COLLECTIONS, keys[i]));
        }
      }
    }

    // displayname is always searchable
    result.add(DISPLAYNAME);

    return result;
  }

  public static IRidSet searchPrincipals(Map search)
    throws UserManagementException, ResourceException {

    Set searchableUserProperties = getSearchableUserProperties();
    Set searchableGroupProperties = getSearchableGroupProperties();
    Set searchableRoleProperties = getSearchableRoleProperties();

    Set searchableProperties = new HashSet();
    searchableProperties.addAll(searchableUserProperties);
    searchableProperties.addAll(searchableGroupProperties);
    searchableProperties.addAll(searchableRoleProperties);

    // set unsupportedProperties = Collections.EMPTY_SET;
    IRidSet rl = new RidSet();

    IUserFactory uf = WPUMFactory.getUserFactory();
    IGroupFactory gf = WPUMFactory.getGroupFactory();
    IRoleFactory rf = WPUMFactory.getRoleFactory();

    ISearchUser su = uf.getSearchUser();
    ISearchGroup sg = gf.getSearchGroup();
    ISearchRole sr = rf.getSearchRole();

    su.setSearchMethod(ISearchUMPrincipal.SEARCHMETHOD_AND);
    sg.setSearchMethod(ISearchUMPrincipal.SEARCHMETHOD_AND);
    sr.setSearchMethod(ISearchUMPrincipal.SEARCHMETHOD_AND);

    boolean searchUsers = true;
    boolean searchGroups = true;
    boolean searchRoles = true;

    for (Iterator it = search.keySet().iterator(); it.hasNext(); /**/) {
      IName tmp = (IName)it.next();
      IPropertyName name = new PropertyName(tmp.getNamespace(), tmp.getName());
      List values = (List)search.get(name);
      String key = null;
      PropertyDescription dummy = new PropertyDescription(name, null);
      String namespace = name.getNamespace();

      // map WebDAV property namespaces to Usermanagement namespaces
      if (searchableProperties.contains(dummy)) {
        if (Consts.N_displayname.equals(name)) {
          // map it
          namespace = null;
          key = IUMPrincipal.DISPLAYNAME;
        }
        else {
          namespace = namespace.substring(NS_UM_BASE.length());
          key = name.getName();
        }
      }

      if (key != null) {
        log.debugT("adding search conditions for UM property: '{" + namespace + "}" + key + "', values: " + values);

        if (searchableUserProperties.contains(dummy)) {
          log.debugT("added as user condition");
          
          if (namespace == null) {
            su.setValuesForKey(key, (String[])values.toArray(SA));
          }
          else {
            su.setValuesForKey(namespace, key, (String[])values.toArray(SA));
          }
        }
        else {
          searchUsers = false;
        }

        if (searchableGroupProperties.contains(dummy)) {
          log.debugT("added as group condition");

          sg.setValuesForKey(key, (String[])values.toArray(SA));
        }
        else {
          searchGroups = false;
        }

        if (searchableRoleProperties.contains(dummy)) {
          log.debugT("added as role condition");

          sr.setValuesForKey(key, (String[])values.toArray(SA));
        }
        else {
          searchRoles = false;
        }
      }
//      else {
//        if (unsupportedProperties.isEmpty()) {
//          unsupportedProperties = new HashSet();
//        }
//        log.debugT("'" + name + "' is not supported as search condition");
//        unsupportedProperties.add(name);
//      }
    }

    String results[] = searchUsers ? uf.findUsers(su, 0) : null;
    if (searchUsers && log.beDebug()) {
      log.debugT("user search result list: " + dump(su.getSearchResult(), results));
    }
    
    if (searchUsers && su.getSearchResult() == ISearchUMPrincipal.SEARCH_RESULT_INCOMPLETE) {
      throw new IncompleteOperationException(null, RID.getRID(null));
    }

    if (results != null) {
      for (int i = 0; i < results.length; i++) {
        rl.add(WDPrincipals.getUserRid(results[i]));
      }
    }

    results = searchGroups ? gf.findGroups(sg, 0) : null;
    if (searchGroups && log.beDebug()) {
      log.debugT("group search result list: " + dump(sg.getSearchResult(), results));
    }
    
    if (searchGroups && sg.getSearchResult() == ISearchUMPrincipal.SEARCH_RESULT_INCOMPLETE) {
      throw new IncompleteOperationException(null, RID.getRID(null));
    }

    if (results != null) {
      for (int i = 0; i < results.length; i++) {
        rl.add (WDPrincipals.getGroupRid(results[i]));
      }
    }

    results = searchRoles ? rf.findRoles(sr, 0) : null;
    if (searchRoles && log.beDebug()) {
      log.debugT("group search result list: " + dump(sr.getSearchResult(), results));
    }

    if (searchRoles && sr.getSearchResult() == ISearchUMPrincipal.SEARCH_RESULT_INCOMPLETE) {
      throw new IncompleteOperationException(null, RID.getRID(null));
    }

    if (results != null) {
      for (int i = 0; i < results.length; i++) {
        rl.add(WDPrincipals.getRoleRid(results[i]));
      }
    }

    return rl;
  }
  
  private static String dump(int code, String results[]) {
    StringBuffer sb = new StringBuffer("code: " + code);
    if (results != null) {
      sb.append(", list: " + Arrays.asList(results));
    }
    
    return sb.toString();
  }

}