package com.sapportals.wcm.protocol.webdav.repository.util;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import com.sap.tc.logging.Location;
import com.sapportals.wcm.repository.Content;
import com.sapportals.wcm.repository.ICollection;
import com.sapportals.wcm.repository.IMutablePropertyMap;
import com.sapportals.wcm.repository.IProperty;
import com.sapportals.wcm.repository.IResource;
import com.sapportals.wcm.repository.IResourceContext;
import com.sapportals.wcm.repository.IResourceFactory;
import com.sapportals.wcm.repository.InvalidTargetException;
import com.sapportals.wcm.repository.MutablePropertyMap;
import com.sapportals.wcm.repository.Property;
import com.sapportals.wcm.repository.ResourceException;
import com.sapportals.wcm.repository.ResourceFactory;
import com.sapportals.wcm.repository.ResourceNotFoundException;
import com.sapportals.wcm.util.logging.LoggingFormatter;
import com.sapportals.wcm.util.uri.RID;

/**
 * <p>
 * Copyright (c) SAP AG 2005
 * @author stefan.eissing@greenbytes.de
 * @version $Id: ResourceInitializer.java,v 1.1 2005/03/17 12:48:32 sei Exp $
 */
public class ResourceInitializer {

  private static final Location log = Location.getLocation(ResourceInitializer.class);

  public static ResourceInitializer getInstance(IResourceContext context) throws ResourceException {
    return new ResourceInitializer(ResourceFactory.getInstance(), context);
  }
  
  private final IResourceFactory factory;
  private final IResourceContext context;
  private final Map errors;
  
  private ResourceInitializer(IResourceFactory factory, IResourceContext context) {
    this.factory = factory;
    this.context = context;
    this.errors = new HashMap();
  }
  
  public ICollection mkcol(RID rid) throws ResourceException {
    return mkcol(rid, null);
  }
  
  public ICollection mkcol(RID rid, IProperty[] props) throws ResourceException {
    IResource res = this.factory.getResource(rid, this.context);
    if (res == null) {
      res = this.factory.getResource(rid, this.context, true);
      if (props != null && props.length > 0) {
        res.setProperties(Arrays.asList(props)); 
      }
    }
    if (!res.isCollection()) {
      throw new InvalidTargetException("resource exists, but is not a collection", rid); 
    }
    return (ICollection)res;
  }
  
	public void copy(RID targetCollection, boolean createTarget, ClassResource[] resources) 
    throws ResourceException {
		IResource res = null;
    if (createTarget) {
      res = mkcol(targetCollection); 
    }
    else {
      res = this.factory.getResource(targetCollection, this.context);
    }
    
    if (res == null) {
      throw new ResourceNotFoundException(targetCollection); 
    }
    else if ( ! res.isCollection()) {
      throw new InvalidTargetException("target collection exists, but is not a collection", targetCollection); 
    }
    
    ICollection tcol = (ICollection)res;
    for (int i = 0; i < resources.length; ++i) {
      ClassResource cr = resources[i];
      try {
        IResource target = this.factory.getResource(targetCollection.add(cr.targetName), this.context);
        IMutablePropertyMap pmap = null;
        if (cr.hidden) {
          pmap = new MutablePropertyMap();
          pmap.put(Property.createHiddenProp(true));
        }
        if (target == null) {
          target = tcol.createResource(cr.targetName, pmap, 
              new Content(cr.base.getResourceAsStream(cr.localName), cr.contentType, -1));
        }
        else {
          target.update(new Content(cr.base.getResourceAsStream(cr.localName), 
              cr.contentType, -1), pmap);
        }
      }
      catch (ResourceException ex) {
        log.errorT("exception creating resource " + ex.getMessage() + " - " + LoggingFormatter.extractCallstack(ex));
        this.errors.put(ex.getRID(), ex);
      }
    }    
  }
  
  public boolean hadErrors() {
    return !this.errors.isEmpty(); 
  }
  
  public Map getErrors() {
    return Collections.unmodifiableMap(this.errors); 
  }
}
