package com.sapportals.wcm.protocol.webdav.server.loadtests;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.*;

import org.w3c.dom.Document;
import org.w3c.dom.Node;

import com.sap.tc.logging.Location;
import com.sapportals.wcm.WcmException;
import com.sapportals.wcm.protocol.webdav.Consts;
import com.sapportals.wcm.protocol.webdav.DomElement;
import com.sapportals.wcm.protocol.webdav.dompath.DomPath;
import com.sapportals.wcm.util.http.HttpRequest;
import com.sapportals.wcm.util.http.HttpStatus;
import com.sapportals.wcm.util.http.IRequester;
import com.sapportals.wcm.util.http.IResponse;
import com.sapportals.wcm.util.uri.*;

/**
 * Various tests for GET method.
 * <p>
 * Copyright (c) SAP AG 2001-2005
 * @author stefan.eissing@greenbytes.de
 */

public class BatchReportTest extends TestMember
{
  private static final Location log = Location.getLocation(BatchReportTest.class);

  private static final String[] desc=new String[] {
    "00 CRAWL REPORT resource tree",
  };

  public static int NrOfTests = desc.length;

  private static final DomPath DP_href;
  private static final DomPath DP_hrefstatus;
  private static final DomPath DP_locationhref;
  private static final DomPath DP_uris;

  static {
    Map nsmap = new HashMap(DomPath.DEFAULT_NAMESPACE_BINDINGS);
    nsmap.put("sap", Consts.NS_SAP);
    DP_href = new DomPath("/D:multistatus/D:response/D:href", nsmap);
    DP_hrefstatus = new DomPath("../D:status", nsmap);
    DP_locationhref = new DomPath("../D:location/D:href", nsmap);
    DP_uris = new DomPath("../D:propstat/D:prop/sap:embedded-links/D:href", nsmap);
  }

  private static class Resource {
    private final IHierarchicalUri uri;
    private int depth;

    public Resource(IHierarchicalUri uri) {
      this.uri = uri;
      this.depth = 0;
    }

    public Resource(IHierarchicalUri uri, int depth) {
      this.uri = uri;
      this.depth = depth;
    }

    public IHierarchicalUri getURI() {
      return this.uri;
    }

    public int getDepth() {
      return this.depth;
    }

    public void setDepth(int depth) {
      this.depth = depth;
    }

    public String toString() {
      StringBuffer sb = new StringBuffer(128);
      sb.append("Resource[").append(this.uri).append(", ").append(this.depth).append("]");
      return sb.toString();
    }
  }

  private static Set s_uris;

  private static int threads = 1;
  private static int chunkSize = 10;

  // private long maxPropfind;
  // private RID maxUri;
  private long duration;

  public BatchReportTest(Integer nr)
  {
    super(nr, desc[nr.intValue()]);
  }

  protected void runTest() {
    try {
      // IResponse response;
      clear(request);
      switch (nr) {
        case 0: {
          Set visited = Collections.synchronizedSet(new HashSet(101));
          String startpath = System.getProperty("crawl", "/web/www.spiegel.de/");
          int depth = Integer.parseInt(System.getProperty("depth", "2"));
          duration = System.currentTimeMillis();
          HttpUrl base = m_factory.getBase();
          HttpUrl start = (HttpUrl)base.appendPath(localpart).appendPath(startpath);
          String tmp = System.getProperty("threads", "1");
          if (tmp != null && tmp.length() > 0) {
            try {
              threads = Integer.parseInt(tmp);
            }
            catch (NumberFormatException e) { /* ignore */ }
          }
          tmp = System.getProperty("chunks", "10");
          if (tmp != null && tmp.length() > 0) {
            try {
              chunkSize = Integer.parseInt(tmp);
            }
            catch (NumberFormatException e) { /* ignore */ }
          }

          walkParallel(start, depth, visited, threads);

          duration = System.currentTimeMillis() - duration;
          int count = visited.size();
          float seconds = ((float)duration)/1000;
          float sd = seconds/count;
          if (log.beDebug()) {
            log.debugT("REPORT ("+threads+" Threads) "+count+" documents in "+seconds+" seconds ("+sd+" secs/doc)");
          }
          else {
            System.out.print("["+count+" docs]");
          }
          s_uris = visited;
          break;
        }

      }
    }
    catch (Exception ex) {
      StringWriter sw = new StringWriter();
      ex.printStackTrace(new PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }

  private void walkParallel(IHierarchicalUri start, int maxDepth, Set visited, int threads) throws Exception {
    List tasks = new ArrayList(threads);
    ResourceQueue queue = new ResourceQueue(start, threads, visited);
    for (int i = 0; i < threads; ++i) {
      Walker walker = new Walker(m_factory.getRequester(m_context), queue, start, maxDepth);
      walker.start();
      tasks.add(walker);
    }

    Iterator iter = tasks.iterator();
    while (iter.hasNext()) {
      Walker w = (Walker)iter.next();
      w.join();
    }
  }

  private class Walker extends Thread {

    private IHierarchicalUri base;
    private IRequester requester;
    private ResourceQueue queue;
    private int maxDepth;

    public Walker(IRequester requester, ResourceQueue queue, IHierarchicalUri base, int maxDepth) {
      this.base = base;
      this.requester = requester;
      this.queue = queue;
      this.maxDepth = maxDepth;
    }

    public void run() {
      //System.out.println("starting walker: "+this);
      try {
        for (List r = queue.next(chunkSize); r != null; r = queue.next(chunkSize)) {
          //System.out.println("next walk on "+r);
          if (r.isEmpty()) {
            break;
          }
          walk(r, this.base);
        }
      }
      catch (Exception ex) {
        StringWriter sw = new StringWriter();
        ex.printStackTrace(new PrintWriter(sw));
        fail(sw.getBuffer().toString());
      }
      finally {
        this.queue.workerTerminated();
      }
      //System.out.println("finished walker: "+this+" at "+(System.currentTimeMillis()/1000));
    }

    private void walk(List rs, IHierarchicalUri base)
    throws WcmException, IOException {

      for (ListIterator i = rs.listIterator(); i.hasNext(); ) {
        Resource r = (Resource)i.next();
        if (r.getDepth() >= this.maxDepth) {
          i.remove();
        }
      }

      //System.out.println("reporting on: "+rs);
      HttpRequest request = new HttpRequest();
      List resources = find(rs, this.requester, request, base);

      //System.out.println("result "+uri+" has "+resources.size()+" links");
      for (ListIterator i = resources.listIterator(); i.hasNext(); ) {
        Resource r = (Resource)i.next();

        if (!this.base.isAncestorOf(r.getURI()) || r.getDepth() >= this.maxDepth) {
          i.remove();
        }
      }

      //System.out.println("adding to queue("+queue.size()+"): "+resources);
      this.queue.add(resources);
    }

    private List find(List rs, IRequester requester, HttpRequest request, IHierarchicalUri baseURI) throws WcmException {
      clear(request);
      request.setMethod("REPORT");
      IUriReference baseref = new UriReference(baseURI, null);
      request.setURI(baseref);
      request.setHeader("Depth", "infinity");
      request.setHeader("Content-Type", "text/xml");

      Map depths = new HashMap(rs.size());
      StringBuffer sb = new StringBuffer(1024);
      sb.append("<wd:expand-property xmlns='DAV:' xmlns:wd='"+Consts.NS_WDSERVLET+"' xmlns:sap='"+Consts.NS_SAP+"'>");
      sb.append("<wd:restrict-to>");
      for (int i = 0, n = rs.size(); i < n; ++i) {
        Resource r = (Resource)rs.get(i);
        sb.append("<href>");
        IHierarchicalUri uri = r.getURI();
        Integer d = new Integer(r.getDepth());
        String path = uri.getPath();
        depths.put(path, d);
        if (path.endsWith("/")) {
          depths.put(path.substring(0, path.length()-1), d);
        }
        //System.out.println("REPORT on "+uriref);
        sb.append(uri);
        sb.append("</href>");
      }
      sb.append("</wd:restrict-to>");
      sb.append("<property name='embedded-links' namespace='"+Consts.NS_SAP+"'/>");
      sb.append("</wd:expand-property>");
      request.setBody(sb.toString());
      request.expectsResponseDocument(true);

      IResponse response = requester.perform(request);
      //System.out.println("REPORT find response: "+response.getStatus());

      Document doc = response.getDocument();
      assertTrue("profind result doc is null for "+rs,
                 response.getStatus() != 207 ||  doc != null);

      List result = new ArrayList();
      HttpStatus stat = new HttpStatus();
      if (doc != null) {
        //System.out.println("REPORT Response: "+DomSerializer.toString(doc, true));
        Node[] hrefs = DP_href.matchNodes(doc);
        for (int i = 0; i < hrefs.length; ++i) {
          Node href = hrefs[i];
          String uriref = DomElement.getText(href);
          int depth = this.maxDepth;
          Integer n = (Integer)depths.get(uriref);
          if (n != null) {
            depth = n.intValue()+1;
          }

          int status = 200;
          Node nstat = DP_hrefstatus.matchFirst(href);
          if (nstat != null) {
            stat.parse(DomElement.getText(nstat));
            status = stat.getCode();
          }

          //System.out.println("REPORT answer for "+uriref+", status "+status);
          switch (status) {
            case 200: {
              Node[] nuris = DP_uris.matchNodes(href);
              for (int j = 0; j < nuris.length; ++j) {
                IHierarchicalUri ruri = resolveUri(DomElement.getText(nuris[j]), baseURI);
                //System.out.println("REPORT found "+ruri+" in depth "+depth);
                result.add(new Resource(ruri, depth));
              }
              break;
            }

            case 302: {
              Node nlhref = DP_locationhref.matchFirst(href);
              if (nlhref != null) {
                String location = DomElement.getText(nlhref);
                try {
                  IHierarchicalUri ruri = resolveUri(location, baseURI);
                  if (ruri != null && baseURI.isAncestorOf(ruri)) {
                    result.add(new Resource(ruri, depth));
                  }
                }
                catch (Exception e) {
                  // ignore
                }

              }
            }

            default:
              break;
          }
        }
      }

      return result;
    }
  }

  private IHierarchicalUri resolveUri(String uriref, IHierarchicalUri base) {
    try {
      IUriReference href = UriFactory.parseUriReference(uriref);
      IUri uri = base.resolve(href);
      if (uri instanceof IHierarchicalUri) {
        return (IHierarchicalUri)uri;
      }
    }
    catch (Exception ex) {
      log.errorT("parsing uri reference: "+uriref);
    }
    return null;
  }
  
  private static class ResourceQueue {

    private int workers;
    private Set visited;
    private Set seen;
    private List resources;
    private int waiting;
    private boolean eof;

    public ResourceQueue(IHierarchicalUri start, int workers, Set visited) {
      this.workers = workers;
      this.visited = visited;
      this.seen = new HashSet(1023);
      this.resources = new ArrayList(1023);
      this.resources.add(new Resource(start));
      this.waiting = 0;
      this.eof = false;
    }

    public synchronized Resource next() {
      if (this.eof) {
        return null;
      }

      ++this.waiting;
      if (this.waiting >= this.workers && this.resources.isEmpty()) {
        this.eof = true;
        this.notifyAll();
      }
      //System.out.println("waiting on queue: "+this.waiting);
      while (this.resources.isEmpty() && !this.eof) {
        try {
          this.wait();
        }
        catch (InterruptedException e) {
          //ignore
        }
      }

      --this.waiting;

      if (this.eof) {
        return null;
      }

      Resource r = (Resource)this.resources.remove(0);
      this.visited.add(r.getURI());
      //System.out.println("handing out: "+r.getRID());
      return r;
    }

    public synchronized List next(int count) {
      List result = new ArrayList(count);
      for (int i = 0; i < count && (i == 0 || !this.resources.isEmpty()); ++i) {
        Resource r = next();
        if (r != null) {
          result.add(r);
        }
      }

      return result;
    }

    public synchronized void add(Resource r) {
      //System.out.println("adding to queue: "+r.getRID());
      if (!this.seen.contains(r.getURI())) {
        this.seen.add(r.getURI());
        this.resources.add(r);
      }

      this.notifyAll();
    }

    public synchronized void add(List resources) {
      //System.out.println("adding to queue: "+r.getRID());
      for (int i = 0, n = resources.size(); i < n; ++i) {
        Resource r = (Resource)resources.get(i);
        if (!this.seen.contains(r.getURI())) {
          this.seen.add(r.getURI());
          this.resources.add(r);
        }
      }

      this.notifyAll();
    }

    public synchronized void workerTerminated() {
      --this.workers;
      if (this.waiting >= this.workers) {
        this.eof = true;
        this.notifyAll();
      }
    }

    public synchronized int size() {
      return this.resources.size();
    }
  }
}
