package com.sapportals.wcm.protocol.webdav.server.loadtests;

import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.*;

import org.w3c.dom.Document;
import org.w3c.dom.Node;

import com.sapportals.wcm.WcmException;
import com.sapportals.wcm.protocol.webdav.Consts;
import com.sapportals.wcm.protocol.webdav.DomElement;
import com.sapportals.wcm.protocol.webdav.dompath.DomPath;
import com.sapportals.wcm.util.http.HttpRequest;
import com.sapportals.wcm.util.http.IRequester;
import com.sapportals.wcm.util.http.IResponse;
import com.sapportals.wcm.util.uri.*;

/**
 * Report on result of CrawlCopy.
 * <p>
 * Copyright (c) SAP AG 2001-2005
 * @author stefan.eissing@greenbytes.de
 */

public class CopyReport extends TestMember
{
  private static final String[] desc=new String[] {
    "00 CPREPORT PROPFIND copied resources",
    "01 CPREPORT GET copied resources",
  };

  public static int NrOfTests = desc.length;

  private static final DomPath DP_href;
  private static final DomPath DP_uris;

  static {
    Map nsmap = new HashMap(DomPath.DEFAULT_NAMESPACE_BINDINGS);
    nsmap.put("sap", Consts.NS_SAP);
    DP_href = new DomPath("/D:multistatus/D:response/D:href", nsmap);
    DP_uris = new DomPath("../D:propstat/D:prop/sap:embedded-links/xs:string", nsmap);
  }

  private static class Resource {
    private final RID uri;
    private int depth;

    public Resource(RID uri) {
      this.uri = uri;
      this.depth = 0;
    }

    public Resource(RID uri, int depth) {
      this.uri = uri;
      this.depth = depth;
    }

    public RID getRID() {
      return this.uri;
    }

    public int getDepth() {
      return this.depth;
    }

    public void setDepth(int depth) {
      this.depth = depth;
    }
  }

  private static Set s_uris;

  private static int threads = 1;

  private long maxPropfind;
  private RID maxUri;
  private long duration;

  public CopyReport(Integer nr)
  {
    super(nr, desc[nr.intValue()]);
  }

  protected void runTest() {
    try {
      clear(request);
      switch (nr) {
        case 0: {
          Set visited = Collections.synchronizedSet(new HashSet(101));
          int max = Integer.parseInt(System.getProperty("max", "2000"));
          duration = System.currentTimeMillis();
          RID start = RID.getRID(System.getProperty("crawl", "/cm/sei-docs"));
          String tmp = System.getProperty("threads", "1");
          if (tmp != null && tmp.length() > 0) {
            try {
              threads = Integer.parseInt(tmp);
            }
            catch (NumberFormatException e) {
              // ignore
            }
          }
          if (threads <= 1) {
            walk(start, max, request, visited, start);
          }
          else {
            walkParallel(start, max, visited, threads);
          }
          duration = System.currentTimeMillis() - duration;
          int count = visited.size();
          float seconds = ((float)duration)/1000;
          float sd = seconds/count;
          float av = ((float)maxPropfind)/1000;
          System.out.println("PROPFIND ("+threads+" Threads) "+count+" documents in "+seconds+" seconds ("
                             +sd+" secs/doc, max "+av+" secs on "+maxUri+") ");
          s_uris = visited;
          break;
        }

        case 1: {
          duration = System.currentTimeMillis();
          int count = s_uris.size();
          long bytes = 0L;
          if (threads > 1) {
            bytes = getAllParallel(s_uris);
          }
          else {
            bytes = getAll(s_uris);
          }
          duration = System.currentTimeMillis() - duration;
          float seconds = ((float)duration)/1000;
          float sd = seconds/count;
          float av = (((float)bytes)/(1024))/seconds;
          System.out.println("GET ("+threads+" Threads) "+count+" documents in "+seconds+" seconds ("
                             +sd+" secs/doc, throughput "+av+" KB/s) ");
          break;
        }

      }
    }
    catch (Exception ex) {
      StringWriter sw = new StringWriter();
      ex.printStackTrace(new PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }

  private void walk(RID start, int max, HttpRequest request, Set visited, RID base)
    throws WcmException {
    
    for (int i = 0; i < max; ++i) {
      RID rid = start.add(String.valueOf(i));
      //System.out.println("visiting "+rid);
      // List resources =
      find(rid, requester, request, base);
      visited.add(rid);
    }
  }

  private List find(RID rid, IRequester requester, HttpRequest request, RID baseRID) throws WcmException {

    clear(request);
    request.setMethod("PROPFIND");
    request.setURI(URICodec.Encode(localpart + rid.toString()));
    request.setHeader("Depth", "0");
    request.setHeader("Content-Type", "text/xml");
    request.setBody("<propfind xmlns='DAV:' xmlns:sap='"+Consts.NS_SAP+"'>"
                    +"<prop><sap:embedded-links/></prop></propfind>");
    request.expectsResponseDocument(true);

    long duration = System.currentTimeMillis();
    IResponse response = requester.perform(request);
    duration = System.currentTimeMillis() - duration;
    if (duration > maxPropfind) {
      maxPropfind = duration;
      maxUri = rid;
    }

    Document doc = response.getDocument();
    assertTrue("profind result doc is null for "+rid,
               response.getStatus() != 207 ||  doc != null);

    List result = new ArrayList();
    if (doc != null) {
      Node href = DP_href.matchFirst(doc);
      if (href != null) {
        Node[] nuris = DP_uris.matchNodes(href);
        for (int i = 0; i < nuris.length; ++i) {
          RID ruri = RID.getRID(DomElement.getText(nuris[i]));
//          if (ruri.toString().indexOf('{') >= 0) {
//            System.out.println("resource "+uri+" has suspicious link: "+ruri);
//          }
          result.add(new Resource(ruri));
        }
      }
    }
    else if (response.getStatus() == 302) {
      String location = response.getHeader("Location");
      if (location != null) {
        try {
          IHierarchicalUri href = UriFactory.parseHierarchicalUri(location);
          HttpUrl base = this.m_factory.getBase();
          RID ref = base.mapToResourceID(new UriReference(href, null));
          if (ref != null && baseRID.isAncestorOf(ref)) {
            result.add(new Resource(ref));
          }
        }
        catch (Exception e) {
          // ignore
        }
      }
    }

    return result;
  }

  private long getAll(Set uris) throws Exception {
    long bytes = 0L;
    Iterator iter = uris.iterator();
    while (iter.hasNext()) {
      RID uri = (RID)iter.next();
      clear(request);
      request.setMethod("GET");
      request.setURI(URICodec.Encode(localpart + uri.toString()));
      request.setHeader("Translate", "F");
      request.expectsResponseStream(true);
      IResponse response = requester.perform(request);
      checkResponseCode(response, 200, 204, 302);
      bytes += readContent(response.getStream());
    }
    return bytes;
  }

  private long getAllParallel(Set uris) throws Exception {
    int max = threads;
    List uriPool = new ArrayList(uris);
    List tasks = new ArrayList(max);
    for (int i = 0; i < max; ++i) {
      Getter getter = new Getter(m_factory.getRequester(m_context), uriPool);
      getter.start();
      tasks.add(getter);
    }

    long bytes = 0L;

    Iterator iter = tasks.iterator();
    while (iter.hasNext()) {
      Getter g = (Getter)iter.next();
      while (!g.isFinished()) {
        try {
          Thread.sleep(200L);
        }
        catch (InterruptedException e) {
          // ignore
        }
      }
      bytes += g.getBytesRead();
    }

    return bytes;
  }

  private long readContent(InputStream is) throws IOException {
    if (is == null)
      return 0L;

    long bytes = 0L;
    byte[] buffer = new byte[64 * 1024];
    int len;
    while ((len = is.read(buffer)) != -1) {
      bytes += len;
    }
    is.close();
    return bytes;
  }

  private class Getter extends Thread {

    private IRequester m_requester;
    private List uris;
    private long m_bytes = 0L;
    private boolean finished;

    public Getter(IRequester requester, List uris) {
      m_requester = requester;
      this.uris = uris;
    }

    public void run() {
      try {
        RID rid;
        HttpRequest request = new HttpRequest();
        while ((rid = next()) != null) {
          request.clear();
          request.setHeader(Consts.SERVLET_TRACE_HEADER, "GET [cT: "+this.getName()+"] - "+rid);
          request.setMethod("GET");
          request.setURI(URICodec.Encode(localpart + rid.toString()));
          //request.setHeader("Translate", "F");
          request.expectsResponseStream(true);
          IResponse response = m_requester.perform(request);
          switch (response.getStatus()) {
            case 200:
            case 204:
              break;

            case 302:
              // needs handling?
              break;

            default:
              System.out.println("GET "+rid+" returned: "+response.getStatus());
              break;
          }
          m_bytes += readContent(response.getStream());
        }
      }
      catch (Exception ex) {
        System.out.println("exception exit");
        ex.printStackTrace(System.out);
        StringWriter sw = new StringWriter();
        ex.printStackTrace(new PrintWriter(sw));
        fail(sw.getBuffer().toString());
      }
      finally {
        this.finished = true;
      }
    }

    public boolean isFinished() {
      return this.finished;
    }

    public long getBytesRead() {
      return m_bytes;
    }

    private RID next() {
      RID rid = null;
      synchronized (this.uris) {
        if (!this.uris.isEmpty()) {
          rid = (RID)this.uris.remove(this.uris.size()-1);
        }
      }
      return rid;
    }
  }

  private void walkParallel(RID base, int max, Set visited, int threads) throws Exception {
    List tasks = new ArrayList(threads);
    ResourceQueue queue = new ResourceQueue(base.add("0"), threads, visited);
    for (int i = 1; i < max; ++i) {
      queue.add(new Resource(base.add(String.valueOf(i)), 1));
    }

    for (int i = 0; i < threads; ++i) {
      Walker walker = new Walker(m_factory.getRequester(m_context), queue, base);
      walker.start();
      tasks.add(walker);
    }

    Iterator iter = tasks.iterator();
    while (iter.hasNext()) {
      Walker w = (Walker)iter.next();
      w.join();
    }
  }

  private class Walker extends Thread {

    private RID base;
    private IRequester requester;
    private ResourceQueue queue;

    public Walker(IRequester requester, ResourceQueue queue, RID base) {
      this.base = base;
      this.requester = requester;
      this.queue = queue;
    }

    public void run() {
      //System.out.println("starting walker: "+this);
      try {
        for (Resource r = queue.next(); r != null; r = queue.next()) {
          walk(r.getRID(), r.getDepth(), this.base);
        }
      }
      catch (Exception ex) {
        StringWriter sw = new StringWriter();
        ex.printStackTrace(new PrintWriter(sw));
        fail(sw.getBuffer().toString());
      }
      finally {
        this.queue.workerTerminated();
      }
      //System.out.println("finished walker: "+this+" at "+(System.currentTimeMillis()/1000));
    }

    private void walk(RID uri, int depth, RID base)
    throws WcmException, IOException {

      HttpRequest request = new HttpRequest();
      // List resources =
      find(uri, this.requester, request, base);
    }
  }

  private static class ResourceQueue {

    private int workers;
    private Set visited;
    private Set seen;
    private List resources;
    private int waiting;
    private boolean eof;

    public ResourceQueue(RID start, int workers, Set visited) {
      this.workers = workers;
      this.visited = visited;
      this.seen = new HashSet(1023);
      this.resources = new ArrayList(1023);
      this.resources.add(new Resource(start));
      this.waiting = 0;
      this.eof = false;
    }

    public synchronized Resource next() {
      if (this.eof) {
        return null;
      }

      ++this.waiting;
      if (this.waiting >= this.workers && this.resources.isEmpty()) {
        this.eof = true;
        this.notifyAll();
      }
      //System.out.println("waiting on queue: "+this.waiting);
      while (this.resources.isEmpty() && !this.eof) {
        try {
          this.wait();
        }
        catch (InterruptedException e) {
          //ignore
        }
      }

      --this.waiting;

      if (this.eof) {
        return null;
      }

      Resource r = (Resource)this.resources.remove(0);
      this.visited.add(r.getRID());
      //System.out.println("handing out: "+r.getRID());
      return r;
    }

    public synchronized void add(Resource r) {
      //System.out.println("adding to queue: "+r.getRID());
      if (!this.seen.contains(r.getRID())) {
        this.seen.add(r.getRID());
        this.resources.add(r);
      }

      this.notifyAll();
    }

    public synchronized void workerTerminated() {
      --this.workers;
      if (this.waiting >= this.workers) {
        this.eof = true;
        this.notifyAll();
      }
    }
  }
}
