package com.sapportals.wcm.protocol.webdav.server.loadtests;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.*;

import org.w3c.dom.Document;
import org.w3c.dom.Node;

import com.sapportals.wcm.WcmException;
import com.sapportals.wcm.protocol.webdav.Consts;
import com.sapportals.wcm.protocol.webdav.DomElement;
import com.sapportals.wcm.protocol.webdav.dompath.DomPath;
import com.sapportals.wcm.util.http.HttpRequest;
import com.sapportals.wcm.util.http.IRequester;
import com.sapportals.wcm.util.http.IResponse;
import com.sapportals.wcm.util.uri.*;

/**
 * WebDAV copy crawler.
 * <p>
 * Copyright (c) SAP AG 2001-2005
 * @author stefan.eissing@greenbytes.de
 */

public class CrawlCopyTest extends TestMember
{
  private static final String[] desc=new String[] {
    "00 CRAWL PROPFIND resource tree",
    "01 CRAWL COPY",
  };

  public static int NrOfTests = desc.length;

  private static final DomPath DP_href;
  private static final DomPath DP_uris;

  static {
    Map nsmap = new HashMap(DomPath.DEFAULT_NAMESPACE_BINDINGS);
    nsmap.put("sap", Consts.NS_SAP);
    DP_href = new DomPath("/D:multistatus/D:response/D:href", nsmap);
    DP_uris = new DomPath("../D:propstat/D:prop/sap:embedded-links/xs:string", nsmap);
  }

  private static class Resource {
    private final RID uri;
    private int depth;

    public Resource(RID uri) {
      this.uri = uri;
      this.depth = 0;
    }

    public Resource(RID uri, int depth) {
      this.uri = uri;
      this.depth = depth;
    }

    public RID getRID() {
      return this.uri;
    }

    public int getDepth() {
      return this.depth;
    }

    public void setDepth(int depth) {
      this.depth = depth;
    }
  }

  private static Set s_uris;

  private static int threads = 1;

  private long maxPropfind;
  private RID maxUri;
  private long duration;

  public CrawlCopyTest(Integer nr)
  {
    super(nr, desc[nr.intValue()]);
  }

  protected void runTest() {
    try {
      // IResponse response;
      clear(request);
      switch (nr) {
        case 0: {
          Set visited = Collections.synchronizedSet(new HashSet(101));
          String startpath = System.getProperty("crawl", "/web/www.spiegel.de/");
          int depth = Integer.parseInt(System.getProperty("depth", "2"));
          duration = System.currentTimeMillis();
          RID start = RID.getRID(startpath);
          String tmp = System.getProperty("threads", "1");
          if (tmp != null && tmp.length() > 0) {
            try {
              threads = Integer.parseInt(tmp);
            }
            catch (NumberFormatException e) {
              // ignore
            }
          }
          if (threads <= 1) {
            walk(start, request, visited, depth, start);
          }
          else {
            walkParallel(start, depth, visited, threads);
          }
          duration = System.currentTimeMillis() - duration;
          int count = visited.size();
          float seconds = ((float)duration)/1000;
          float sd = seconds/count;
          float av = ((float)maxPropfind)/1000;
          System.out.println("PROPFIND ("+threads+" Threads) "+count+" documents in "+seconds+" seconds ("
                             +sd+" secs/doc, max "+av+" secs on "+maxUri+") ");
          s_uris = visited;
          break;
        }

        case 1: {
          duration = System.currentTimeMillis();
          RID destination = RID.getRID(System.getProperty("dest", "/cm/sei"));
          int copiedDocs = copyAll(s_uris, destination);
          duration = System.currentTimeMillis() - duration;
          float seconds = ((float)duration)/1000;
          float sd = seconds/copiedDocs;
          System.out.println("COPY: "+copiedDocs+" documents in "+seconds+" seconds ("
                             +sd+" secs/doc)");
          break;
        }

      }
    }
    catch (Exception ex) {
      StringWriter sw = new StringWriter();
      ex.printStackTrace(new PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }

  private void walk(RID uri, HttpRequest request, Set visited, int depth, RID base)
  throws WcmException, IOException {

    if (depth <= 0) {
      return;
    }
    --depth;

    //System.out.println("visiting "+uri);
    List resources = find(uri, requester, request, base);
    visited.add(uri);

    Iterator iter = resources.iterator();
    while (iter.hasNext()) {
      Resource r = (Resource)iter.next();

      if (!visited.contains(r.getRID()) && base.isAncestorOf(r.getRID())) {
        walk(r.getRID(), request, visited, depth, base);
      }
    }
  }

  private List find(RID uri, IRequester requester, HttpRequest request, RID baseRID) throws WcmException {

    clear(request);
    request.setMethod("PROPFIND");
    request.setURI(URICodec.Encode(localpart + uri.toString()));
    request.setHeader("Depth", "0");
    request.setHeader("Content-Type", "text/xml");
    request.setBody("<propfind xmlns='DAV:' xmlns:sap='"+Consts.NS_SAP+"'>"
                    +"<prop><sap:embedded-links/></prop></propfind>");
    request.expectsResponseDocument(true);

    long duration = System.currentTimeMillis();
    IResponse response = requester.perform(request);
    duration = System.currentTimeMillis() - duration;
    if (duration > maxPropfind) {
      maxPropfind = duration;
      maxUri = uri;
    }

    Document doc = response.getDocument();
    assertTrue("profind result doc is null for "+uri,
               response.getStatus() != 207 ||  doc != null);

    List result = new ArrayList();
    if (doc != null) {
      Node href = DP_href.matchFirst(doc);
      if (href != null) {
        Node[] nuris = DP_uris.matchNodes(href);
        for (int i = 0; i < nuris.length; ++i) {
          RID ruri = RID.getRID(DomElement.getText(nuris[i]));
//          if (ruri.toString().indexOf('{') >= 0) {
//            System.out.println("resource "+uri+" has suspicious link: "+ruri);
//          }
          result.add(new Resource(ruri));
        }
      }
    }
    else if (response.getStatus() == 302) {
      String location = response.getHeader("Location");
      if (location != null) {
        try {
          IHierarchicalUri href = UriFactory.parseHierarchicalUri(location);
          HttpUrl base = this.m_factory.getBase();
          RID ref = base.mapToResourceID(new UriReference(href, null));
          if (ref != null && baseRID.isAncestorOf(ref)) {
            result.add(new Resource(ref));
          }
        }
        catch (Exception e) {
          // ignore
        }
      }
    }

    return result;
  }

//  private long getAllPropfind(Set uris) throws Exception {
//    long bytes = 0L;
//    Iterator iter = uris.iterator();
//    while (iter.hasNext()) {
//      URI uri = (URI)iter.next();
//      clear(m_request);
//      m_request.setMethod("PROPFIND");
//      m_request.setURI(URICodec.Encode(m_localpart + uri.toString()));
//      m_request.setHeader("Translate", "F");
//      m_request.expectsResponseStream(false);
//      m_request.setHeader("Depth", "0");
//      m_request.setHeader("Content-Type", "text/xml");
//      m_request.setBody("<propfind xmlns='DAV:' xmlns:sap='"+Consts.NS_SAP+"'>"
//                      +"<prop><displayname/><getcontentlength/><getcontenttype/></prop></propfind>");
//      IResponse response = m_requester.perform(m_request);
//      checkResponseCode(response, 207, 302);
//      bytes += readContent(response.getStream());
//    }
//    return bytes;
//  }

  private int copyAll(Set uris, RID destDir) throws Exception {
    // long bytes = 0L;
    int i = 0;
    int copies = 0;
    Iterator iter = uris.iterator();
    while (iter.hasNext()) {
      RID rid = (RID)iter.next();
      clear(request);
      request.setMethod("COPY");
      request.setURI(URICodec.Encode(localpart + rid.toString()));
      request.setHeader("Destination", URICodec.Encode(localpart + destDir.add(String.valueOf(i))));
      request.setHeader("Overwrite", "T");
      request.setHeader("Translate", "F");
      IResponse response = requester.perform(request);
      switch (response.getStatus()) {
        case 200:
        case 201:
        case 204:
          ++i;
          ++copies;
          break;

        default:
          System.out.println("COPY of "+rid+" returned: "+response.getStatus());
          break;
      }
    }
    return copies;
  }

//  private long readContent(InputStream is) throws IOException {
//    if (is == null)
//      return 0L;
//
//    long bytes = 0L;
//    byte[] buffer = new byte[64 * 1024];
//    int len;
//    while ((len = is.read(buffer)) != -1) {
//      bytes += len;
//    }
//    is.close();
//    return bytes;
//  }

  private void walkParallel(RID start, int maxDepth, Set visited, int threads) throws Exception {
    List tasks = new ArrayList(threads);
    ResourceQueue queue = new ResourceQueue(start, threads, visited);
    for (int i = 0; i < threads; ++i) {
      Walker walker = new Walker(m_factory.getRequester(m_context), queue, start, maxDepth);
      walker.start();
      tasks.add(walker);
    }

    Iterator iter = tasks.iterator();
    while (iter.hasNext()) {
      Walker w = (Walker)iter.next();
      w.join();
    }
  }

  private class Walker extends Thread {

    private RID base;
    private IRequester requester;
    private ResourceQueue queue;
    private int maxDepth;

    public Walker(IRequester requester, ResourceQueue queue, RID base, int maxDepth) {
      this.base = base;
      this.requester = requester;
      this.queue = queue;
      this.maxDepth = maxDepth;
    }

    public void run() {
      //System.out.println("starting walker: "+this);
      try {
        for (Resource r = queue.next(); r != null; r = queue.next()) {
          walk(r.getRID(), r.getDepth(), this.base);
        }
      }
      catch (Exception ex) {
        StringWriter sw = new StringWriter();
        ex.printStackTrace(new PrintWriter(sw));
        fail(sw.getBuffer().toString());
      }
      finally {
        this.queue.workerTerminated();
      }
      //System.out.println("finished walker: "+this+" at "+(System.currentTimeMillis()/1000));
    }

    private void walk(RID uri, int depth, RID base)
    throws WcmException, IOException {

      if (depth >= this.maxDepth) {
        return;
      }
      ++depth;

      HttpRequest request = new HttpRequest();
      List resources = find(uri, this.requester, request, base);

      //System.out.println("result "+uri+" has "+resources.size()+" links");
      if (depth < this.maxDepth) {
        for (Iterator iter = resources.iterator(); iter.hasNext(); ) {
          Resource r = (Resource)iter.next();

          if (this.base.isAncestorOf(r.getRID())) {
            r.setDepth(depth);
            this.queue.add(r);
          }
        }
      }
    }
  }

  private static class ResourceQueue {

    private int workers;
    private Set visited;
    private Set seen;
    private List resources;
    private int waiting;
    private boolean eof;

    public ResourceQueue(RID start, int workers, Set visited) {
      this.workers = workers;
      this.visited = visited;
      this.seen = new HashSet(1023);
      this.resources = new ArrayList(1023);
      this.resources.add(new Resource(start));
      this.waiting = 0;
      this.eof = false;
    }

    public synchronized Resource next() {
      if (this.eof) {
        return null;
      }

      ++this.waiting;
      if (this.waiting >= this.workers && this.resources.isEmpty()) {
        this.eof = true;
        this.notifyAll();
      }
      //System.out.println("waiting on queue: "+this.waiting);
      while (this.resources.isEmpty() && !this.eof) {
        try {
          this.wait();
        }
        catch (InterruptedException e) {
          //ignore
        }
      }

      --this.waiting;

      if (this.eof) {
        return null;
      }

      Resource r = (Resource)this.resources.remove(0);
      this.visited.add(r.getRID());
      //System.out.println("handing out: "+r.getRID());
      return r;
    }

    public synchronized void add(Resource r) {
      //System.out.println("adding to queue: "+r.getRID());
      if (!this.seen.contains(r.getRID())) {
        this.seen.add(r.getRID());
        this.resources.add(r);
      }

      this.notifyAll();
    }

    public synchronized void workerTerminated() {
      --this.workers;
      if (this.waiting >= this.workers) {
        this.eof = true;
        this.notifyAll();
      }
    }
  }
}
