package com.sapportals.wcm.protocol.webdav.server.loadtests;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import com.sap.tc.logging.Location;
import com.sapportals.wcm.protocol.webdav.Consts;
import com.sapportals.wcm.util.http.HttpRequest;
import com.sapportals.wcm.util.http.IRequester;
import com.sapportals.wcm.util.http.IResponse;
import com.sapportals.wcm.util.uri.HttpUrl;
import com.sapportals.wcm.util.uri.IHierarchicalUri;
import com.sapportals.wcm.util.uri.UriReference;

/**
 * Various tests for crawling.
 * <p>
 * Copyright (c) SAP AG 2001-2005
 * @author stefan.eissing@greenbytes.de
 */

public class CrawlTest extends TestMember
{
  private static final Location log = Location.getLocation(CrawlTest.class);

  private static final String[] desc=new String[] {
    "00 CRAWL PROPFIND resource tree",
    "01 CRAWL GET in parallel",
  };

  public static int NrOfTests = desc.length;

  private static final boolean debug = true;
  
  private static Set s_uris;

  private static int threads = 1;

  public CrawlTest(Integer nr)
  {
    super(nr, desc[nr.intValue()]);
  }

  protected void runTest() {
    try {
      clear(request);
      switch (nr) {
        case 0: {
          String startpath = System.getProperty("crawl", "/web/www.spiegel.de/");
          int depth = Integer.parseInt(System.getProperty("depth", "2"));
          IHierarchicalUri base = m_factory.getBase().appendPath(localpart);
          IHierarchicalUri start = (HttpUrl)base.appendPath(startpath);
          String tmp = System.getProperty("threads", "1");
          if (tmp != null && tmp.length() > 0) {
            try {
              threads = Integer.parseInt(tmp);
            }
            catch (NumberFormatException e) {
              // ignore
            }
          }
          
          SummaryObserver observer = SummaryObserver.getInstance(start, depth, threads, debug);
          WebDAVCrawler crawler = WebDAVCrawler.getInstance(m_factory, m_context, base);
          
          observer.start();
          crawler.walk(start, depth, threads, observer);
          observer.finish();
          
          observer.printSummary(System.out);
          s_uris = observer.getVisited();
          break;
        }

        case 1: {
          long duration = System.currentTimeMillis();
          int count = s_uris.size();
          
          long bytes = getAllParallel(s_uris);
          
          duration = System.currentTimeMillis() - duration;
          float seconds = ((float)duration)/1000;
          float sd = seconds/count;
          float av = (((float)bytes)/(1024))/seconds;
          if (debug) {
            System.out.println("GET ("+threads+" Threads) "+count+" documents in "+seconds+" seconds ("
                      +sd+" secs/doc, throughput "+av+" KB/s) ");
          }
          else {
            System.out.print("["+count+" docs]");
          }
          break;
        }

      }
    }
    catch (Exception ex) {
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }

  private long getAllParallel(Set uris) throws Exception {
    int max = threads;
    List uriPool = new ArrayList(uris);
    List tasks = new ArrayList(max);
    for (int i = 0; i < max; ++i) {
      Getter getter = new Getter(m_factory.getRequester(m_context), uriPool);
      getter.start();
      tasks.add(getter);
    }

    long bytes = 0L;

    Iterator iter = tasks.iterator();
    while (iter.hasNext()) {
      Getter g = (Getter)iter.next();
      while (!g.isFinished()) {
        try {
          Thread.sleep(200L);
        }
        catch (InterruptedException e) {
          // ignore
        }
      }
      bytes += g.getBytesRead();
    }

    return bytes;
  }

  private long readContent(InputStream is) throws IOException {
    if (is == null)
      return 0L;

    long bytes = 0L;
    byte[] buffer = new byte[64 * 1024];
    int len;
    while ((len = is.read(buffer)) != -1) {
      bytes += len;
    }
    is.close();
    return bytes;
  }

  private class Getter extends Thread {

    private IRequester m_requester;
    private List uris;
    private long m_bytes = 0L;
    private boolean finished;

    public Getter(IRequester requester, List uris) {
      m_requester = requester;
      this.uris = uris;
    }

    public void run() {
      try {
        IHierarchicalUri uri;
        HttpRequest request = new HttpRequest();
        while ((uri = next()) != null) {
          request.clear();
          request.setHeader(Consts.SERVLET_TRACE_HEADER, "GET [cT: "+this.getName()+"] - "+uri);
          request.setHeader("User-Agent", "greenbytes crawler 1.1");
          request.setMethod("GET");
          request.setURI(new UriReference(uri, null));
          //request.setHeader("Translate", "F");
          request.expectsResponseStream(true);
          IResponse response = m_requester.perform(request);
          //System.out.println("GET on "+request.getURI()+" returned "+response.getStatus());
          switch (response.getStatus()) {
            case 200:
            case 204:
              break;

            case 302:
              // needs handling?
              break;

            default:
              if (log.beDebug()) {
                log.debugT("GET "+uri+" returned: "+response.getStatus());
              }
              break;
          }
          m_bytes += readContent(response.getStream());
        }
      }
      catch (Exception ex) {
        System.out.println("exception exit");
        ex.printStackTrace(System.out);
        java.io.StringWriter sw = new java.io.StringWriter();
        ex.printStackTrace(new java.io.PrintWriter(sw));
        fail(sw.getBuffer().toString());
      }
      finally {
        this.finished = true;
      }
    }

    public boolean isFinished() {
      return this.finished;
    }

    public long getBytesRead() {
      return m_bytes;
    }

    private IHierarchicalUri next() {
      IHierarchicalUri rid = null;
      synchronized (this.uris) {
        if (!this.uris.isEmpty()) {
          rid = (IHierarchicalUri)this.uris.remove(this.uris.size()-1);
        }
      }
      return rid;
    }
  }

}
