package com.sapportals.wcm.protocol.webdav.server.loadtests;

import com.sapportals.wcm.util.http.IRequester;
import com.sapportals.wcm.util.http.IResponse;
import com.sapportals.wcm.util.http.HttpRequest;
import com.sapportals.wcm.protocol.webdav.*;
import com.sapportals.wcm.protocol.webdav.dompath.DomPath;

import java.io.*;

import org.w3c.dom.Node;
import org.w3c.dom.Document;

public class DownloadTest extends TestMember {

  private static final String[] desc = new String[]{
    "00 measuring download performance: 1 file",
    "01 measuring download performance: 10 files parallel",
    "02 measuring download performance: 1 file on busy file system",
    "03 measuring download performance: 10 files parallel on busy file system",
  };

  public static int NrOfTests = desc.length;

  private File tmpdir;

  public DownloadTest(Integer nr) {
    super(nr, desc[nr.intValue()]);
    String tmp = System.getProperty("dltmp");
    if (tmp == null) {
      fail("missing system property: dltmp");
    }
    tmpdir = new File(tmp);
    if (!tmpdir.exists() && !tmpdir.mkdirs()) {
      fail("could not create directory: " + tmpdir);
    }
  }

  protected void runTest() {
    try {
      clear(request);
      switch (nr) {
        case 0:
          request.setMethod("MKCOL");
          request.setURI(localpart);
          requester.perform(request);

          for (int i = 0; i < Resource.RESOURCES.length; i++) {
            request.clear();
            request.setMethod("HEAD");
            request.setURI(localpart + Resource.RESOURCES[i].getRID());
            IResponse response = requester.perform(request);
            int status = response.getStatus();

            request.setMethod("PUT");
            request.setContentType("application/octet-stream");
            request.setContentLength(Resource.RESOURCES[i].getContent().getLength());
            request.setBody(Resource.RESOURCES[i].getContent().getStream());
            requester.perform(request);
          }

          for (int i = 0; i < Resource.RESOURCES.length; i++) {
            execute(1, 10, Resource.RESOURCES[i], 0);
          }
          break;

        case 1:

          for (int i = 0; i < Resource.RESOURCES.length; i++) {
            execute(10, 5, Resource.RESOURCES[i], 0);
          }
          break;

        case 2:

          for (int i = 0; i < Resource.RESOURCES.length; i++) {
            execute(1, 10, Resource.RESOURCES[i], 50);
          }
          break;

        case 3:

          for (int i = 0; i < Resource.RESOURCES.length; i++) {
            execute(10, 5, Resource.RESOURCES[i], 50);
          }
          break;
      }
    }
    catch (Exception ex) {
      StringWriter sw = new StringWriter();
      ex.printStackTrace(new PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }


  //---------------- private ---------------------------------------

  private void execute(int workerCount, int repeat, Resource res, int load)
      throws Exception {

    FileSystemLoad fsLoad = new FileSystemLoad(load, 0);
    fsLoad.start();

    Worker[] workers = new Worker[workerCount];
    for (int i = 0; i < workers.length; ++i) {
      workers[i] = new Worker(m_factory.getRequester(m_context), m_base, localpart, res, repeat);
    }

    for (int i = 0; i < workers.length; ++i) {
      workers[i].start();
    }

    for (int i = 0; i < workers.length; ++i) {
      workers[i].join();
    }

    fsLoad.abort();

    long rate = 0;
    for (int i = 0; i < workers.length; ++i) {
      GetOp.GetStatistics stat = workers[i].getStatistics();
      if (stat == null) {
        System.out.println("no statistic available for " + res.getRID());
        return;
      }
      rate += stat.getRatio();
    }
    long len = res.getContent().getLength() / 1024;
    System.out.println(
        "parallel download of " + workerCount + " files of size " + len + " kB: " + rate
        + "kB/sec (file sys load = " + load + ")");
  }

  private static final DomPath DP_status = new DomPath("/D:multistatus/D:response/D:status",
                                                       DomPath.DEFAULT_NAMESPACE_BINDINGS);
  private static final DomPath DP_href = new DomPath("../D:href",
                                                     DomPath.DEFAULT_NAMESPACE_BINDINGS);

  private static Name getCondition(IResponse response) {
    Document doc = response.getDocument();
    if (doc == null || !Consts.N_error.equalsNode(doc.getDocumentElement())) {
      return null;
    }
    else {
      for (Node c = doc.getDocumentElement().getFirstChild(); c != null; c = c.getNextSibling()) {
        if (c.getNodeType() == Node.ELEMENT_NODE) return new Name(c);
      }
    }
    return null;
  }

  private static String getInvocationCounter(IResponse response) {
    String invoc = response.getHeader(Consts.SERVLET_TRACE_HEADER);
    return invoc == null ? "" : invoc;
  }

  private static void dumpResponse(String method, IResponse response) {

    Name cond = getCondition(response);
    String invoc = getInvocationCounter(response);

    System.out.println(method + " returned: " + response.getStatus() +
                       " " + response.getStatusDescription() +
                       (cond == null ? "" : ", condition: " + cond.toString()) +
                       (invoc == null ? "" : ", invocation: " + invoc));

    if (response.hasDocument()) {
      System.out.println("response body is: " + DomSerializer.toString(response.getDocument(), true));
    }
  }


  private static final class Resource {

    private static final Resource[] RESOURCES = new Resource[]{
      new Resource("/file1", 1024 * 1024 * 1),
      new Resource("/file2", 1024 * 1024 * 20),
    };

    private final String rid;
    private final Content content;

    public Resource(String rid, int length) {
      this.rid = rid;
      this.content = new Content(length);
    }

    public String getRID() {
      return this.rid;
    }

    public Content getContent() {
      return this.content;
    }
  }

  private static final class Content {

    private int length;

    private InputStream in = new InputStream() {

      boolean eof = false;
      int cnt = 0;

      public int read() throws IOException {
        if (!this.eof) {
          this.cnt++;
          this.eof = this.cnt > Content.this.getLength();
          if (!this.eof) {
            return (this.cnt % 256);
          }
        }
        return -1;
      }
    };

    public Content(int length) {
      this.length = length;
    }

    public int getLength() {
      return this.length;
    }

    public InputStream getStream() {
      return in;
    }
  }

  private final class FileSystemLoad {

    private final class LoadThread extends Thread{

      private long until;
      private File dir;

      public LoadThread(File dir, int duration) {
        super.setDaemon(true);
        this.until = (duration == 0)? Long.MAX_VALUE : System.currentTimeMillis() + duration;
        this.dir = dir;
        this.dir.mkdirs();
      }

      public void run() {
        try {
          while (!FileSystemLoad.this.aborted && System.currentTimeMillis() < until) {
            this.dir.lastModified();
            this.dir.listFiles();
            File file = new File(this.dir, "tmp");
            file.createNewFile();
            file.setLastModified(System.currentTimeMillis());
            file.delete();
            try {
              Thread.sleep(100);
            }
            catch (InterruptedException e) {
            }
          }
        }
        catch (IOException e) {
          e.printStackTrace();
        }
      }
    }

    private int threadCount;
    private int duration;

    private LoadThread[] threads;

    private boolean aborted = false;

    public FileSystemLoad(int threadCount, int duration) {
      this.threadCount = threadCount;
      this.duration = duration;

      this.threads = new LoadThread[this.threadCount];
      for (int k = 0; k < this.threadCount; k++) {
        this.threads[k] = new LoadThread(new File(tmpdir, "dir" + k), this.duration);
      }
    }

    public void start() {
      for (int k = 0; k < this.threadCount; k++) {
        this.threads[k].start();
      }
    }

    public void abort() {
      this.aborted = true;
    }

    public boolean isAborted() {
      return this.aborted;
    }
  }

  private static final class GetOp {

    public final class GetStatistics {

      private int repeat;
      private long length;
      private long time;
      private long ratio;

      public GetStatistics(int repeat, long fileLength, long totalTime) {

        this.repeat = repeat;
        this.length = fileLength;
        this.time = totalTime;
        this.ratio = (long) (((double) this.length / 1024) / ((double) this.time / (1000 * this.repeat)));
      }

      public int getRepeat() {
        return this.repeat;
      }

      public long getLength() {
        return this.length;
      }

      public long getTime() {
        return this.time;
      }

      public long getRatio() {
        return this.ratio;
      }
    }

    private static final GetOp INSTANCE = new GetOp();

    public static GetOp getInstance() {
      return INSTANCE;
    }

    private GetOp() {
    }

    public GetStatistics perform(Resource resource, IRequester requester, String base, String uriPrefix, int repeat)
        throws Exception {

      HttpRequest request = new HttpRequest(uriPrefix + resource.getRID());
      request.setHeader(Consts.SERVLET_TRACE_HEADER, "DOWNLOAD");
      request.setMethod("GET");
      request.setHeader("Translate", "F");
      request.expectsResponseStream(true);

      long tSum = 0;
      long lSum = 0;
      IResponse response = null;
      for (int k = 0; k < repeat; k++) {
        response = requester.perform(request);
        InputStream stream = response.getStream();
        if (response.getStatus() == 200) {

          boolean ok = false;
          long length = response.getContentLength();

          long len = 0;
          byte[] buf = new byte[1024 * 1024];

          long time = System.currentTimeMillis();
          for (long r = 0; (r = stream.read(buf)) > 0; len += r) {}
          time = System.currentTimeMillis() - time;
          tSum += time;

          if (len != resource.getContent().getLength()) {
            System.out.println(
                "wrong content length: " + len + ", expected was "
                + resource.getContent().getLength() + " on " + resource.getRID());
            return null;
          }
          lSum += len;
        }
        else {
          if (stream != null) {
            stream.close();
          }
          dumpResponse(request.getMethod(), response);
          return null;
        }
      }

      return new GetStatistics(repeat, resource.getContent().getLength(), tSum);
    }
  }

  private static final class Worker extends Thread {

    private final IRequester requester;
    private final String prefix;
    private final String base;
    private final GetOp op = GetOp.getInstance();
    private Resource resource;
    private int repeat;

    private GetOp.GetStatistics stat = null;

    public Worker(IRequester requester, String base, String prefix, Resource resource, int repeat) {
      this.requester = requester;
      this.base = base;
      this.prefix = prefix;
      this.resource = resource;
      this.repeat = repeat;
    }

    public void run() {
      try {
        this.stat = op.perform(this.resource, this.requester, this.base, this.prefix, repeat);
      }
      catch (Exception e) {
        e.printStackTrace(System.out);
      }
    }

    public GetOp.GetStatistics getStatistics() {
      return this.stat;
    }
  }
}
