package com.sapportals.wcm.protocol.webdav.server.loadtests;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.*;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import com.sapportals.wcm.WcmException;
import com.sapportals.wcm.protocol.webdav.Consts;
import com.sapportals.wcm.protocol.webdav.DomElement;
import com.sapportals.wcm.protocol.webdav.DomSerializer;
import com.sapportals.wcm.protocol.webdav.dompath.DomPath;
import com.sapportals.wcm.util.http.*;
import com.sapportals.wcm.util.uri.*;

/**
 * Various tests for concurrent GET invocations.
 * <p>
 * Copyright (c) SAP AG 2001-2005
 * @author stefan.eissing@greenbytes.de
 */

public class GetTest extends TestMember {
  
  private static final String[] desc=new String[] {
    "00 1 resource, 1 reader",
    "01 1 resource, 2 reader",
    "02 1 resource, 4 reader",
    "03 1 resource, 8 reader",
    "04 1 resource, 16 reader",
    "05 1 resource, 32 reader",
    "06 1 resource, 64 reader",
    "07 1 resource, 128 reader",
    "08 2 resource, 4 reader",
    "09 2 resource, 16 reader",
    "10 2 resource, 64 reader",
    "11 2 resource, 128 reader",
    "12 4 resource, 16 reader",
    "13 4 resource, 64 reader",
    "14 4 resource, 128 reader",
  };

  public static int NrOfTests = desc.length;

  public GetTest(Integer nr) {
    super(nr, desc[nr.intValue()]);
  }

  protected void runTest() {

    boolean showStats = "yes".equals(System.getProperty("statistics", "yes"));
    
    try {
      // IResponse response;
      clear(request);
      switch (nr) {
        case 0: {
          HttpRequest request = new HttpRequest();
          clear(request);
          request.setMethod("GET");
          request.setURI(localpart + "/");
          request.expectsResponseDocument(true);
          m_factory.getRequester(m_context).perform(request);
          
          Clerk clerk = Clerk.getInstance(requester, m_base, localpart);
          clerk.clean();
          for (int i = 0; i < ResourceBag.R.length; ++i) {
            clerk.copy(ResourceBag.R[i]);
          }
          perform(1, 1, showStats);
          break;
        }

        case 1: {
          perform(1, 2, showStats);
          break;
        }

        case 2: {
          perform(1, 4, showStats);
          break;
        }

        case 3: {
          perform(1, 8, showStats);
          break;
        }

        case 4: {
          perform(1, 16, showStats);
          break;
        }

        case 5: {
          perform(1, 32, showStats);
          break;
        }

        case 6: {
          perform(1, 64, showStats);
          break;
        }

        case 7: {
          perform(1, 128, showStats);
          break;
        }

        case 8: {
          perform(2, 4, showStats);
          break;
        }

        case 9: {
          perform(2, 16, showStats);
          break;
        }

        case 10: {
          perform(2, 64, showStats);
          break;
        }

        case 11: {
          perform(2, 128, showStats);
          break;
        }

        case 12: {
          perform(4, 16, showStats);
          break;
        }

        case 13: {
          perform(4, 64, showStats);
          break;
        }

        case 14: {
          perform(4, 128, showStats);
          break;
        }
      }
    }
    catch (Exception ex) {
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }

  private void perform(int resources, int getter, boolean showStats) throws Exception {
    ResourceBag bag = ResourceBag.getInstance(resources);
    List workers = new ArrayList();
    long now = System.currentTimeMillis();

    for (int i = 0; i < getter; ++i) {
      Getter g = new Getter(m_factory.getRequester(m_context), localpart, 
          bag, 256 / getter, super.m_allowRedirects);
      workers.add(g);
      g.start();
    }
    long bytes = 0;
    for (int i = 0, n = workers.size(); i < n; ++i) {
      Getter g = (Getter)workers.get(i); 
      g.join();
      bytes += g.getBytesRead();
    }

    long duration = System.currentTimeMillis() - now;
    StringBuffer sb = new StringBuffer();
    sb.append("("+resources+"/"+getter+")got " 
        + (bytes/1000) + "KB in " + (((double)duration)/1000) + " sec ("
        + bytes/(1024*1024*(((double)duration)/1000)) + " MB/s)");
    if (super.m_allowRedirects && workers.size() == 1) {
      long redirectTime = 0;
      for (int i = 0, n = workers.size(); i < n; ++i) {
        Getter g = (Getter)workers.get(i);
        redirectTime += g.getRedirectTime();
      }
      sb.append(" (redirect time: " + ((double)redirectTime/1000) + ")");
    }
    if (showStats) {
      System.out.println(sb.toString());
    }
  }
  
  private static class Resource {
  
    public static Resource getInstance(String name, ContentType ctype, byte[] content) {
      return new Resource(name, ctype, content);
    }
    
    private final String name;
    private final ContentType ctype;
    private final byte[] content;
    
    private Resource(String name, ContentType ctype, byte[] content) {
      this.name = name;
      this.ctype = ctype;
      this.content = content;
    }
    
    public String getName() {
      return this.name;
    }
    
    public ContentType getContentType() {
      return this.ctype;
    }
    
    public byte[] getContent() {
      return this.content;
    }
    
    public void checkContent(InputStream is) throws Exception {
      byte[] buffer = new byte[64 * 1024];
      int read = 0;
      int offset = 0;
      if (Math.random() > 0.8) {
        while ((read = is.read(buffer)) >= 0) {
          for (int i = 0; i < read; ++i) {
            int pos = offset + i;
            if (pos < this.content.length &&  buffer[i] != this.content[pos]) {
              throw new IOException("content bytes for " + this.name + " differs: expected="
                  + this.content[pos] + ", got=" + buffer[i] + " at offset "
                  + (pos));
            }
          }
          offset += read;
        }
      }
      else {
        while ((read = is.read(buffer)) >= 0) {
          offset += read;
        }
      }
        
      if (offset != this.content.length) {
        throw new IOException("content length for " + this.name + " differs: expected="
            + this.content.length + ", got=" + offset + " bytes");
      }
    }
  }
  
  private static class ResourceBag {
    
    public static final Resource[] R = {
        Resource.getInstance("r1", ContentType.APPLICATION_OCTET_STREAM, 
            makeContent(500000, new byte[] { 0x00, -1 })),
        Resource.getInstance("r2", ContentType.APPLICATION_OCTET_STREAM, 
            makeContent(256000, new byte[] { -1, 0x00 })),
        Resource.getInstance("r3", ContentType.TEXT_PLAIN, 
            makeContent(25000, new byte[] { 0x41, 0x42, 0x43, 0x44 })),            
        Resource.getInstance("r4", ContentType.TEXT_HTML, 
            makeContent(125000, new byte[] { 0x41, 0x42, 0x43, 0x44 })),            
    };
    
    public static ResourceBag getInstance(int size) {
      if (size > R.length) {
        throw new IndexOutOfBoundsException("index out of range: " + size); 
      }
      return new ResourceBag(size);
    }
    
    private static byte[] makeContent(int times, byte[] sample) {
      byte[] content = new byte[times * sample.length];
      for (int i = 0; i < content.length; i += sample.length) {
        System.arraycopy(sample, 0, content, i, sample.length);
      }
      return content;
    }
    
    private final int size;
    private final Random r;
    
    private ResourceBag(int size)  {
      this.size = size;
      this.r = new Random();
    }
    
    public int size() {
      return this.size;
    }
    
    public Resource getResource(int index) {
      if (index < 0 || index >= this.size) {
        throw new IndexOutOfBoundsException("index out of range: " + index); 
      }
      return R[index];
    }
    
    public Resource getRandom() {
      return getResource(this.r.nextInt(this.size));
    }
  }
  
  private static class Pool {
    
    private static final Map map = new HashMap();
    
    public synchronized static IRequester getRequester(HttpUrl url) throws Exception {
      IRequesterFactory factory = (IRequesterFactory)map.get(url);
      if (factory == null) {
        factory = RequesterFactory.newInstance();
        factory.setBaseURL(url.toExternalForm());
        factory.setSoTimeout(30000);
        map.put(url, factory);
      }
      return factory.getRequester(null);
    }
  }
  
  private static class Getter extends Thread {

    private final String id;
    private final IRequester requester;
    private final String localpart;
    private final ResourceBag bag;
    private final int requests;
    private final boolean allowCacheRedirects;
    private long bytes;
    private long redirectTime;

    public Getter(IRequester requester, String localpart, ResourceBag bag, 
        int requests, boolean allowCacheRedirects) {
      this.id = super.getName();
      this.requester = requester;
      this.localpart = localpart;
      this.bag = bag;
      this.requests = requests;
      this.allowCacheRedirects = allowCacheRedirects;
    }

    public void run() {
      try {
        HttpRequest request = new HttpRequest();
        for (int i = 0; i < this.requests; ++i) {
          try {
            Resource r = this.bag.getRandom();
            request.clear();
            request.setHeader(Consts.SERVLET_TRACE_HEADER, "GET [cT:"+this.id+"] - "+r.getName());
            request.setHeader("User-Agent", "greenbytes GET stress 1.0");
            if ( ! this.allowCacheRedirects ) {
              request.setHeader(HEADER_OPT, '"' + HTTPEXT_N_NO_CACHE_REDIRECT.getNamespaceUri() + '"' + " ;ns=00");
              request.setHeader("00-" + HTTPEXT_N_NO_CACHE_REDIRECT.getLocalName(), "");
            }
            request.setMethod("GET");
            request.setURI(this.localpart + "/" + r.getName());
            //request.setHeader("Translate", "F");
            request.expectsResponseStream(true);
            long start = System.currentTimeMillis();
            IResponse response = requester.perform(request);
            InputStream is = response.getStream();
            try {
              switch (response.getStatus()) {
                case 200:
                  if (this.allowCacheRedirects) {
                    System.out.println("redirect not happening on " + r.getName());
                  }
                  r.checkContent(is);
                  this.bytes += r.getContent().length;
                  break;

                case 301:
                case 302:
                case 307:
                  this.redirectTime += (System.currentTimeMillis() - start);
                  if ( ! this.allowCacheRedirects) {
                    System.out.println("unexpected redirect " + response.getStatus() + " on " + r.getName()
                        + " to " + response.getHeader("location"));
                  }
                  follow(response.getHeader("location"), r, true);
                  break;

                default:
                  System.out.println("unexpected answer on " + r.getName() + ": " + response.getStatus());
                  break;
              }
            }
            finally {
              if (is != null) {
                is.close();
              }
            }
          }
          catch (Exception ex) {
            ex.printStackTrace(System.out);
            java.io.StringWriter sw = new java.io.StringWriter();
            ex.printStackTrace(new java.io.PrintWriter(sw));
            System.out.println(sw.getBuffer().toString());
          }
        }
      }
      finally {
        this.requester.release();
      }
    }

    public long getBytesRead() {
      return this.bytes;
    }
    
    public long getRedirectTime() {
      return this.redirectTime;
    }
    
    private void follow(String location, Resource r, boolean follow) throws Exception {
      IHierarchicalUri locuri = UriFactory.parseHierarchicalUri(location);
      if ( ! (locuri instanceof HttpUrl)) {
        throw new Exception("redirect location is not a http url: " + location);
      }
      HttpUrl uri = (HttpUrl)locuri;
      HttpUrl root = (HttpUrl)uri.getRoot();
      IRequester requester = Pool.getRequester(root);
      String redirect = null;
      try {
        IUriReference href = root.unresolve(uri, null);
        HttpRequest request = new HttpRequest();

        request.setURI(href);
        request.setMethod("GET");
        request.setHeader(Consts.SERVLET_TRACE_HEADER, "GET [cT:"+this.id+"] - "+r.getName());
        request.setHeader("User-Agent", "greenbytes GET stress 1.0");
        request.expectsResponseStream(true);
        
        IResponse response = requester.perform(request);
        
        InputStream is = response.getStream();
        try {
          switch (response.getStatus()) {
            case 200:
              r.checkContent(is);
              this.bytes += r.getContent().length;
              break;

            case 301:
            case 302:
            case 307:
              if (follow) {
                redirect = response.getHeader("location");
                break;
              }
              // fall through
            default:
              System.out.println("unexpected answer on " + r.getName() + ": " + response.getStatus());
              break;
          }
        }
        finally {
          if (is != null) {
            is.close();
          }
        }
        
      }
      finally {
        requester.release();
      }

      if (follow && redirect != null) {
        System.out.println("unexpected redirect (again) to " + redirect);
        follow(redirect, r, false);
      }
    }
  }
  
  private static class Clerk {
    
    private static Clerk INSTANCE;
    
    public synchronized static Clerk getInstance(IRequester requester, String host, String localpart) {
      if (INSTANCE != null) {
        if (INSTANCE.requester == requester 
            && INSTANCE.host == host
            && INSTANCE.localpart == localpart) {
          return INSTANCE;
        }
        INSTANCE = null;
      }

      if (INSTANCE == null) {
        INSTANCE = new Clerk(requester, host, localpart);
      }
      return INSTANCE;
    }
    
    private static final DomPath DP_hrefs = new DomPath(
        "/D:multistatus/D:response/D:href", DomPath.DEFAULT_NAMESPACE_BINDINGS);
      private static final DomPath DP_locktoken_href = new DomPath(
        "../D:propstat/D:prop/D:lockdiscovery/D:activelock/D:locktoken/D:href", DomPath.DEFAULT_NAMESPACE_BINDINGS);

      private final IRequester requester;
      private final HttpRequest request;
      private final URICodec codec;
      private final String host;
      private final String localpart;

      public Clerk(IRequester requester, String host, String localpart) {
        this.requester = requester;
        this.request = new HttpRequest();
        this.codec = new URICodec();
        this.host = host;
        if (localpart != null && localpart.endsWith("\\"))
          localpart = localpart.substring(0, localpart.length()-1 );
        this.localpart = localpart;
      }

      public void clean() throws WcmException
      {
        String ua = System.getProperty("userAgent", null);
        
        // create target if not present
        int status = 503;
        int retries = 25;
        IResponse response;
        
        while (status == 503 && retries-- > 0) {
          request.clear();
          request.setMethod("MKCOL");
          request.setHeader("x-greenbytes", "create target collection");
          if (ua != null) request.setHeader("user-agent", ua);
          request.setURI(localpart);
          response = requester.perform(request);
          status = response.getStatus();
          
          // System.out.println("status: " + status);

          if (status == 503) {
            try {
              // service not available
              Thread.sleep(20000);
            }
            catch (InterruptedException e) {
              // don't care
            }
          }
        }
        
        request.clear();
        request.setMethod("PROPFIND");
        request.setURI(localpart);
        request.setHeader("Depth", "1");
        request.expectsResponseDocument(true);
        response = requester.perform(request);

        if (response.getStatus() != 207) {
          throw new WcmException ("PROPFIND returned " + response.getStatus() +
          " instead of 207");
        }

        Document doc = response.getDocument();
        if (doc == null) {
          throw new WcmException ("PROPFIND did not return (well-formed) response body");
        }

        List hrefs = new ArrayList();
        List locks = new ArrayList();
        getURLs(doc, "/", hrefs, locks);
        for (int i = 0; i < hrefs.size(); ++i) {
          String uri = (String)hrefs.get(i);
          String lock = (String)locks.get(i);
          if (lock.length() > 0) {
            request.clear();
            request.setMethod("UNLOCK");
            request.setURI(localpart + uri);
            request.setHeader("Lock-Token", "<"+lock+">");
            response = requester.perform(request);
          }
          request.clear();
          request.setMethod("DELETE");
          request.setURI(localpart + uri);
          request.setHeader("Depth", "Infinity");
          request.setHeader("Apply-To-Redirect-Ref", "");
          response = requester.perform(request);
          if (response.getStatus() != 204 && response.getStatus() != 404) {
            Document ms = response.getDocument();
            throw new WcmException("removing "+uri
                                   +" got status "+response.getStatus()
                                   + ((ms != null)? DomSerializer.toString(ms, true) : ""));
          }
        }
      }

      public void copy (Resource r) throws Exception {
        copy(r.getName(), r.getContentType(), r.getContent());
      }
      
      private void copy(String destURI, ContentType ctype, byte[] content) throws Exception {
        InputStream fis = new ByteArrayInputStream(content);
        try {
          request.clear();
          request.setMethod("PUT");
          request.setURI(codec.encode(localpart + "/" + destURI));
          if (ctype != null) {
            request.setHeader ("Content-Type", ctype.getContentType());
          }
          request.setHeader("Content-Length", String.valueOf(content.length));
          request.setBody(fis);
          IResponse response = requester.perform(request);
          int status = response.getStatus();
          if (status != 201 && status != 200) {
            throw new IOException("put on "+destURI+" returned "+status);
          }
        }
        finally {
          fis.close();
        }
      }

      private void getURLs(Document doc, String ignore, List result, List resultTokens) {
        Node[] hrefs = DP_hrefs.matchNodes(doc);
        for (int i = 0; i < hrefs.length; ++i) {
          Element href = (Element)hrefs[i];
          String url = DomElement.getText(href);
          if (url.startsWith(host)) {
            url = url.substring(host.length());
          }
          if (url.startsWith(localpart)) {
            url = url.substring(localpart.length());
          }
          if (url.length() > 0 && !url.equals(ignore)) {
            result.add(url);
            String tokenString = "";

            Node[] tokens = DP_locktoken_href.matchNodes(href);
            for (int j = 0; j < tokens.length; ++j) {
              Element locktoken = (Element)tokens[j];
              tokenString = DomElement.getText(locktoken);
            }

            resultTokens.add(tokenString);
          }
        }
        //System.err.println("hrefs+tokens: "+result+" "+resultTokens);
      }
    
  }
}
