package com.sapportals.wcm.protocol.webdav.server.loadtests;

import java.io.InputStream;

import com.sapportals.wcm.WcmException;
import com.sapportals.wcm.util.http.IResponse;


/**
 * Test maximum resource sizes
 * <p>
 * Copyright (c) SAP AG 2005
 * @author julian.reschke@greenbytes.de
 */

public class PutSizeTest extends TestMember {
  
  private static long KB = 1024;
  private static long MB = KB * KB;
  private static long sizes[] = {
     64 * KB,
    128 * KB,
    256 * KB,
    512 * KB,
    1 * MB,
    2 * MB,
    4 * MB,
    8 * MB,
    16 * MB,
    32 * MB,
    64 * MB,
    128 * MB,
    256 * MB,
    512 * MB,
    1024 * MB,
    2048 * MB,
    4192 * MB};
  
  private static final boolean BMODE = true;
  
  private static final String[] desc;
  
  static {
    if (BMODE) {
    	desc = new String[] {
          "0 PS search maximum puttable content length",
          "1 PS search maximum puttable content length (chunked)"
      };
    }
    else {
      String r[] = new String[sizes.length * 2];
      for (int i = 0; i < sizes.length * 2; i++) {
        String s = Long.toString(i);
        r[i] = (s.length() == 1 ? "0" : "") + s + " PS test resource size " + pretty(sizes[i % sizes.length]) + ((i > sizes.length) ? " (chunked)" : "");
      }
      desc = r;
    }
  }

  public static int NrOfTests = desc.length;

  public PutSizeTest(Integer nr) {
    super(nr, desc[nr.intValue()]);
  }

  protected void runTest() {
    boolean showStats = "yes".equals(System.getProperty("statistics", "yes"));
    try {
      boolean chunked = (this.nr == 1);
      if (BMODE) {
        int workingIndex = -1;
        int nonWorkingIndex = sizes.length;
        
        while ((workingIndex + 1) < nonWorkingIndex) {
          int index = workingIndex + ((nonWorkingIndex - workingIndex) / 2);
          //logComment("index="+index+", working="+workingIndex+", nonwork="+nonWorkingIndex);
          try {
            putSizeTest(sizes[index], chunked, showStats);
            workingIndex = index;
          }
          catch (Throwable ex) {
            ex.printStackTrace(System.out);
            nonWorkingIndex = index;
          }
        }

        String result = (workingIndex >= 0)? pretty(sizes[workingIndex]) : "NONE";
        super.logComment("largest working PUT size(chunked="+chunked+"): " + result);
      }
      else {      
        putSizeTest(sizes[nr % sizes.length], nr >= sizes.length, showStats);
      }
    }
    catch (Exception ex) {
      ex.printStackTrace(System.err);
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }

  private static String pretty(long size) {
    return ((size >= MB) 
      ? (size/MB) + "MB"
      : (size/KB) + "KB");
  }
  
  private void putSizeTest(long size, boolean chunked, boolean showStats) throws Exception {
    
    IResponse response;
    String uri = localpart + "/sizetest-" + pretty(size);

    try {
      // make sure we're authenticated
      clear(request);
      request.setMethod("DELETE");
      request.setURI(uri);
      response = requester.perform(request);
      checkResponseCode(response, 204, 404);

      long ts = System.currentTimeMillis();
      
      clear(request);
      request.setMethod("PUT");
      request.setURI(uri);
      if (!chunked) request.setHeader("Content-Length", Long.toString(size));
      request.setBody(new DataGenerator(0, size));
      response = requester.perform(request);
      checkResponseCode(response, 201);
      
      long duration = System.currentTimeMillis() - ts;
      if (showStats) {
        System.out.println("PUT: " +
            (size >= MB
              ? (size/MB) + " MB"
              : (size/KB) + " KB") + " in " + (((float)duration)/1000) + "s (" + ((1000L*size)/KB)/duration + " KB/s)");
      }

      ts = System.currentTimeMillis();
      clear(request);
      request.setMethod("GET");
      request.setURI(uri);
      request.expectsResponseStream(true);
      response = requester.perform(request);
      checkResponseCode(response, 200);
      checkContentLength(response, size);
      duration = System.currentTimeMillis() - ts;
      if (showStats) {
        System.out.println("GET: " + 
            (size >= MB
              ? (size/MB) + " MB"
              : (size/KB) + " KB") + " in " + (((float)duration)/1000) + "s (" + ((1000L*size)/KB)/duration + " KB/s)");
      }
      
      clear(request);
      request.setMethod("HEAD");
      request.setURI(uri);
      response = requester.perform(request);
      checkResponseCode(response, 200);
      assertEquals(Long.toString(size), response.getHeader("content-length"));
    }
    finally {
      clear(request);
      request.setMethod("DELETE");
      request.setURI(uri);
      response = requester.perform(request);
    }
  }
  
  private void checkContentLength(IResponse response, long expectedSize) throws Exception {
  	InputStream is = response.getStream();
    assertTrue("input stream missing", is != null);
    long length = 0;
    int read = 0;
    byte[] buffer = new byte[8 * 8 * 1024];
    while ((read = is.read(buffer)) >= 0) {
      length += read; 
    }
    is.close();
    assertEquals(expectedSize, length);
  }
  
  private class DataGenerator extends InputStream {

    private int pattern;
    private long length;
    private long position;

    public DataGenerator (int pattern, long length) {
      this.pattern = pattern;
      this.length = length;
      this.position = 0;
    }

    public int read() {
      if (this.position++ < this.length) {
        return pattern;
      }
      else {
        return -1;
      }
    }
    
    public int read(byte b[]) {
      return read(b, 0, b.length);
    }

    public int read(byte b[], int off, int bufLen) {
      if (this.position >= this.length) {
        return -1;
      }
      else if (bufLen + this.position > this.length) {
        this.position = this.length;
        return (int)(this.length - this.position);
      }
      else {
        this.position += bufLen;
        return bufLen;
      }
    }
  }
}
