package com.sapportals.wcm.protocol.webdav.server.loadtests;

import java.io.PrintStream;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import com.sapportals.wcm.util.http.IResponse;
import com.sapportals.wcm.util.uri.IHierarchicalUri;

final class SummaryObserver implements IObserver {
  private final IHierarchicalUri startUri;
  private final int depth;
  private final int workerCount;
  private final boolean verbose;
  private final Set inTransit;
  private final Set visited;
  private long start;
  private long end;
  private boolean finished;
  private int discovered;
  private int todo;
  private long maxDuration;
  private IHierarchicalUri maxUri;
  private long interval;
  private PrintStream intermediateOS;
  private PrintStream log;
  private IntervalDumper dumper;
    
  public static SummaryObserver getInstance(IHierarchicalUri start, int depth, 
                                            int workerCount, boolean verbose) {
    return new SummaryObserver(start, depth, workerCount, verbose);
  }
    
  private SummaryObserver(IHierarchicalUri start, int depth, 
                          int workerCount, boolean verbose) {
    this.startUri = start;
    this.depth = depth;
    this.visited = new HashSet();
    this.inTransit = new HashSet();
    this.workerCount = workerCount;
    this.verbose = verbose;
  }

  public void setLog(PrintStream log) {
    this.log = log;
  }
  
  public synchronized void outputIntermediateTo(PrintStream os, long interval) {
    this.interval = interval;
    this.intermediateOS = os;
  }
  
  public synchronized void start() {
    this.start = System.currentTimeMillis();
    this.visited.clear();
    this.inTransit.clear();    
    this.discovered = 0;
    this.todo = 0;
    this.maxDuration = 0;
    this.maxUri = null;
    this.finished = false;
    if (this.interval > 0) {
      this.dumper = IntervalDumper.getInstance(this, this.intermediateOS, this.interval);
      this.dumper.start();
    }
  }
    
  public synchronized void finish() {
    this.end = System.currentTimeMillis();
    this.finished = true;
    if (this.dumper != null) {
      this.dumper.finish();
      this.dumper = null;
    }
  }
    
  public synchronized void dequeued(IHierarchicalUri uri) {
    this.inTransit.add(uri);
  }
    
  public synchronized void discovered(IHierarchicalUri uri, long duration, IResponse response) {
    ++this.discovered;
    if ( ! this.visited.add(uri)) {
      System.err.println("discovered AGAIN "+uri);
    }
    this.inTransit.remove(uri);
    if (duration > this.maxDuration) {
      this.maxDuration = duration;
      this.maxUri = uri;
    }
    
    if (this.log != null) {
      this.log.println(uri+" "+response.getStatus()+" "+response.getStatusDescription()
          +" ("+duration+" ms)");
    }
    else if (this.verbose && response.getStatus() >= 400) {
      System.out.println("Lookup returned "+response.getStatus()
        +" "+response.getStatusDescription()+ " on "+uri);
    }
  }

  public synchronized void error(IHierarchicalUri uri, Throwable ex) {
    if (this.log != null) {
      this.log.println(uri+" Exception: " + ex.getMessage());
      ex.printStackTrace(this.log);
    }
    else {
      System.out.println("Lookup ERROR on "+uri);
      ex.printStackTrace(System.out);
    }
  }
    
  public synchronized boolean hasSeen(IHierarchicalUri uri) {
    return this.visited.contains(uri) || this.inTransit.contains(uri);
  }
    
  public void todo(int count) {
    this.todo = count;
  }
  
  public long getMaxDuration() {
    return this.maxDuration;
  }
    
  public IHierarchicalUri getMaxUri() {
    return this.maxUri;
  }
    
  public Set getVisited() {
    if (this.visited.size() != this.discovered) {
      System.err.println("Observer: visited size "+this.visited.size()+" != discovered "+this.discovered);
    }
    return Collections.unmodifiableSet(this.visited);
  }

  public synchronized void printIntermediate(PrintStream os) {
    os.print("\r("+this.discovered+"/"+this.todo+"/"+this.inTransit.size()
      +") (found/todo/in progress)");
  }
  
  public synchronized void printSummary(PrintStream os) {
    long duration = this.end - this.start;
    int count = visited.size();
    float seconds = ((float)duration)/1000;
    float sd = seconds/count;
    float av = ((float)this.getMaxDuration())/1000;
    if (this.verbose) {
      os.println("Crawled:     " + this.startUri);          
      os.println("MaxDepth:    " + this.depth); 
      os.println("Workers:     " + this.workerCount); 
      os.println("Documents:   " + count); 
      os.println("Duration:    " + seconds + " sec"); 
      os.println("Seconds/Doc: " + sd + " ms"); 
      os.println("Maximum:     " + av + " sec on " + this.getMaxUri()); 
    }
    else {
      os.print("["+count+" docs]");
    }
  }
  
  private static class IntervalDumper extends Thread {
    public static IntervalDumper getInstance(SummaryObserver observer, PrintStream os, long interval) {
      return new IntervalDumper(observer, os, interval);
    }

    private final SummaryObserver observer;
    private final PrintStream os;
    private final long interval;
    private boolean finished;
    
    private IntervalDumper(SummaryObserver observer, PrintStream os, long interval) {
      this.observer = observer;
      this.os = os;
      this.interval = interval;
    }
    
    public synchronized void finish() {
      this.finished = true;
    }
    
    public synchronized boolean isFinished() {
      return this.finished;
    }
    
		public void run() {
      while (!isFinished()) {
        try {
          Thread.sleep(this.interval);
        }
        catch (InterruptedException ex) {
        }
        
        if (!isFinished()) {
          this.observer.printIntermediate(this.os);
        }
      }
		}
  }
}
