package com.sapportals.wcm.protocol.webdav.server.loadtests;

import java.io.IOException;
import java.util.*;

import org.w3c.dom.Document;
import org.w3c.dom.Node;

import com.sap.tc.logging.Location;
import com.sapportals.wcm.WcmException;
import com.sapportals.wcm.protocol.webdav.Consts;
import com.sapportals.wcm.protocol.webdav.DomElement;
import com.sapportals.wcm.protocol.webdav.dompath.DomPath;
import com.sapportals.wcm.util.http.*;
import com.sapportals.wcm.util.uri.*;

public class WebDAVCrawler {

  private static final Location log = Location.getLocation(WebDAVCrawler.class);
  
  public static WebDAVCrawler getInstance(IRequesterFactory factory, IContext context, IHierarchicalUri base) {
    return new WebDAVCrawler(factory, context, base);
  }
  
  private final IRequesterFactory factory;
  private final IContext context;
  private final IHierarchicalUri base;
  
  private WebDAVCrawler(IRequesterFactory factory, IContext context, IHierarchicalUri base) {
    this.factory = factory;
    this.context = context;
    this.base = base;
  }
  
  public void walk(IHierarchicalUri start, int maxDepth, int threads, IObserver observer) throws Exception {
    List tasks = new ArrayList(threads);
    LinkExtractor extractor = LinkExtractor.getInstance(this.base);
    ResourceQueue queue = new ResourceQueue(observer);
    queue.add(new Resource(start));

    // build the walkers
    for (int i = 0; i < threads; ++i) {
      Walker walker = new Walker("CrawlTestWalker "+i, observer, extractor, this.factory.getRequester(this.context), 
        queue, start, maxDepth);
      tasks.add(walker);
    }

    // start them
    for (int i = 0, n = tasks.size(); i < n; ++i) {
      Walker walker = (Walker)tasks.get(i);
      walker.start();
    }
    
     // wait until done
    Iterator iter = tasks.iterator();
    while (iter.hasNext()) {
      Walker w = (Walker)iter.next();
      w.join();
    }
  }

  private static class Resource {
    private final IHierarchicalUri uri;
    private int depth;

    public Resource(IHierarchicalUri uri) {
      this.uri = uri;
      this.depth = 0;
    }

    public Resource(IHierarchicalUri uri, int depth) {
      this.uri = uri;
      this.depth = depth;
    }

    public IHierarchicalUri getURI() {
      return this.uri;
    }

    public int getDepth() {
      return this.depth;
    }

    public void setDepth(int depth) {
      this.depth = depth;
    }
    
    public boolean equals(Object o) {
      if (o instanceof Resource) {
        return this.uri.equals(((Resource)o).uri);
      }
      return false;
    }
    
    public int hashCode() {
      return this.uri.hashCode();
    }
  }

  private static class Walker extends Thread {
    private final String name;
    private final IObserver observer;
    private final LinkExtractor extractor;
    private final IHierarchicalUri scope;
    private final IRequester requester;
    private final ResourceQueue queue;
    private final int maxDepth;

    public Walker(String name, IObserver observer, LinkExtractor extractor, 
                  IRequester requester, ResourceQueue queue, 
                  IHierarchicalUri scope, int maxDepth) {
      this.name = name;
      this.observer = observer;
      this.extractor = extractor;
      this.scope = scope;
      this.maxDepth = maxDepth;
      this.requester = requester;
      this.queue = queue;
      this.queue.registerWalker(this);
    }

    public void run() {
      //System.out.println("starting walker: "+this);
      try {
        for (Resource r = queue.next(); r != null; r = queue.next()) {
          walk(r.getURI(), r.getDepth());
        }
      }
      catch (Exception ex) {
        java.io.StringWriter sw = new java.io.StringWriter();
        ex.printStackTrace(new java.io.PrintWriter(sw));
      }
      finally {
        this.queue.unregisterWalker(this);
      }
      //System.out.println("finished walker: "+this+" at "+(System.currentTimeMillis()/1000));
    }

    private void walk(IHierarchicalUri uri, int depth)
    throws WcmException, IOException {

      if (depth >= this.maxDepth) {
        return;
      }
      ++depth;

      HttpRequest request = new HttpRequest();
      List resources = find(uri, this.requester, request);

      //System.out.println("result "+uri+" has "+resources.size()+" links");
      if (depth < this.maxDepth) {
        for (Iterator iter = resources.iterator(); iter.hasNext(); ) {
          Resource r = (Resource)iter.next();

          if (this.scope.isAncestorOf(r.getURI())) {
            r.setDepth(depth);
            this.queue.add(r);
          }
        }
      }
      this.observer.todo(this.queue.size());
    }

    private List find(IHierarchicalUri uri, IRequester requester, HttpRequest request) throws WcmException {
      try {
        request.clear();
        request.setHeader(Consts.SERVLET_TRACE_HEADER, this.name);
        request.setMethod("PROPFIND");
        request.setURI(new UriReference(uri, null));
        request.setHeader("Depth", "0");
        request.setHeader("Cache-Control", "no-cache");
        request.setHeader("User-Agent", "greenbytes crawler 1.1");
        request.setHeader("Content-Type", "text/xml");
        
        request.setBody("<propfind xmlns='DAV:' xmlns:sap='"+Consts.NS_SAP+"'>"
                        +"<prop><sap:embedded-links/><sap:restricted-links/></prop></propfind>");
        request.expectsResponseDocument(true);

        long duration = System.currentTimeMillis();
        IResponse response = requester.perform(request);
        duration = System.currentTimeMillis() - duration;
        this.observer.discovered(uri, duration, response);

        return extractor.extract(uri, response);
      }
      catch (WcmException ex) {
        this.observer.error(uri, ex);
        return Collections.EMPTY_LIST;
      }
    }

  }

  private static class ResourceQueue {
    
    private final IObserver observer;
    private List todo;
    private final Set walkers;
    private int waitingWorkers;
    private boolean eof;

    public ResourceQueue(IObserver observer) {
      this.observer = observer;
      this.walkers = new HashSet();
      this.todo = new ArrayList();
      this.waitingWorkers = 0;
      this.eof = false;
    }

    public synchronized Resource next() {
      if (this.eof) {
        return null;
      }

      ++this.waitingWorkers;
      if (this.waitingWorkers >= this.walkers.size() && this.todo.isEmpty()) {
        this.eof = true;
        this.notifyAll();
      }
      //System.out.println("waiting on queue: "+this.waiting);
      while (this.todo.isEmpty() && !this.eof) {
        try {
          this.wait();
        }
        catch (InterruptedException e) {
          //ignore
        }
      }

      --this.waitingWorkers;

      if (this.eof) {
        return null;
      }

      Resource r = (Resource)this.todo.remove(0);
      this.observer.dequeued(r.getURI());
      //System.out.println("handing out: "+r.getRID());
      return r;
    }

    public synchronized void add(Resource r) {
      //System.out.println("adding to queue: "+r.getRID());
      if ( ! this.observer.hasSeen(r.getURI()) && ! this.todo.contains(r)) {
        this.todo.add(r);
      }

      this.notifyAll();
    }

    public synchronized int size() {
      return this.todo.size();
    }
    
    public synchronized void registerWalker(Walker walker) {
      this.walkers.add(walker);
    }

    public synchronized void unregisterWalker(Walker walker) {
      this.walkers.remove(walker);
      if (this.waitingWorkers >= this.walkers.size()) {
        this.eof = true;
        this.notifyAll();
      }
    }
    
  }
  
  private static class LinkExtractor {
    
    private static final DomPath DP_href;
    private static final DomPath DP_embedded;
    private static final DomPath DP_restricted;
    private static final DomPath DP_old_embedded;
    private static final DomPath DP_old_restricted;

    static {
      Map nsmap = new HashMap(DomPath.DEFAULT_NAMESPACE_BINDINGS);
      nsmap.put("sap", Consts.NS_SAP);
      DP_href = new DomPath("/D:multistatus/D:response/D:href", nsmap);
      DP_embedded = new DomPath("../D:propstat/D:prop/sap:embedded-links/D:href", nsmap);
      DP_restricted = new DomPath("../D:propstat/D:prop/sap:restricted-links/D:href", nsmap);
      DP_old_embedded = new DomPath("../D:propstat/D:prop/sap:embedded-links/xs:string", nsmap);
      DP_old_restricted = new DomPath("../D:propstat/D:prop/sap:restricted-links/xs:string", nsmap);
    }

    public static LinkExtractor getInstance(IHierarchicalUri base) {
      return new LinkExtractor(base);
    }
    
    private final IHierarchicalUri base;
    
    private LinkExtractor(IHierarchicalUri base) {
      this.base = base;
    }
    
    public List extract(IHierarchicalUri uri, IResponse response) {
      Document doc = response.getDocument();
      if (doc != null) {
        List result = Collections.EMPTY_LIST;
        Node href = DP_href.matchFirst(doc);
        if (href != null) {
          Set restricted = getRestricted(href, uri);
          result = getEmbedded(href, uri, restricted);
        }
        return result;
      }
      else {
        switch (response.getStatus()) {
          case 301:
          case 302: 
          case 307: {
            String location = response.getHeader("Location");
            if (location != null) {
              try {
                IHierarchicalUri href = UriFactory.parseHierarchicalUri(location);
                if (href != null) {
                  List result = new ArrayList(1);
                  result.add(new Resource(href));
                  return result;
                }
              }
              catch (Exception e) {
                // ignore
              }
            }
            break;
          }
          
          default:
            break;
        }
      }
      
      return Collections.EMPTY_LIST;
    }

    private Set getRestricted(Node href, IHierarchicalUri uri) {
      Node[] nrestricted = DP_restricted.matchNodes(href);
      if (nrestricted.length > 0) {
        Set restricted = new HashSet();
        for (int i = 0; i < nrestricted.length; ++i) {
          restricted.add(resolveUri(DomElement.getText(nrestricted[i]), uri));
        }
        return restricted;
      }
      else {
        // try old style (xs:string) with RIDs
        nrestricted = DP_old_restricted.matchNodes(href);
        if (nrestricted.length > 0) {
          Set restricted = new HashSet();
          for (int i = 0; i < nrestricted.length; ++i) {
            RID rid = RID.getRID(DomElement.getText(nrestricted[i]));
            IHierarchicalUri ruri = this.base.appendPath(URICodec.EncodePath(rid.getPath()));
            if (rid.getQuery() != null) {
              ruri = ruri.setQuery(rid.getQuery());
            }
            restricted.add(ruri);
          }
          return restricted;
        }
      }
      return Collections.EMPTY_SET;
    }

    private List getEmbedded(Node href, IHierarchicalUri uri, Set restricted) {    
      Node[] nuris = DP_embedded.matchNodes(href);
      if (nuris.length > 0) {
        List result = new ArrayList(nuris.length);
        for (int i = 0; i < nuris.length; ++i) {
          IHierarchicalUri ruri = resolveUri(DomElement.getText(nuris[i]), uri);
          if (!restricted.contains(ruri)) {
            result.add(new Resource(ruri));
          }
        }
        return result;
      }
      else {
        // try old style (xs:string) with RIDs
        nuris = DP_old_embedded.matchNodes(href);
        if (nuris.length > 0) {
          List result = new ArrayList(nuris.length);
          for (int i = 0; i < nuris.length; ++i) {
            RID rid = RID.getRID(DomElement.getText(nuris[i]));
            IHierarchicalUri ruri = this.base.appendPath(URICodec.EncodePath(rid.getPath()));
            if (rid.getQuery() != null) {
              ruri = ruri.setQuery(rid.getQuery());
            }
            if (!restricted.contains(ruri)) {
              result.add(new Resource(ruri));
            }
          }
          return result;
        }
      }
      return Collections.EMPTY_LIST;
    }
    
    private IHierarchicalUri resolveUri(String uriref, IHierarchicalUri lbase) {
      try {
        IUriReference href = UriFactory.parseUriReference(uriref);
        IUri uri = lbase.resolve(href);
        if (uri instanceof IHierarchicalUri) {
          return (IHierarchicalUri)uri;
        }
      }
      catch (Exception ex) {
        log.errorT("parsing uri reference: "+uriref);
      }
      return null;
    }
  }
  
}
