package com.sapportals.wcm.protocol.webdav.server.loadtests;

import java.io.*;
import java.util.*;

import junit.framework.AssertionFailedError;

import org.w3c.dom.Document;
import org.w3c.dom.Node;

import com.sapportals.wcm.protocol.webdav.*;
import com.sapportals.wcm.protocol.webdav.dompath.DomPath;
import com.sapportals.wcm.util.http.HttpRequest;
import com.sapportals.wcm.util.http.HttpStatus;
import com.sapportals.wcm.util.http.IRequester;
import com.sapportals.wcm.util.http.IResponse;

/**
 * Various tests for GET/PUT/MOVE/COPY methods.
 * <p>
 * Copyright (c) SAP AG 2002-2005
 * @author stefan.eissing@greenbytes.de
 */

public class WildTest extends TestMember {
  private static final String[] desc = new String[]{
    "00 WILD testing DELETE",
    "01 WILD testing PUT/MKCOL",
    "02 WILD testing GET",
    "03 WILD testing PUT/MKCOL/GET",
    "04 WILD testing DELETE/PUT/MKCOL",
    "05 WILD testing DELETE/PUT/MKCOL/MOVE",
    "06 WILD testing LOCK/UNLOCK",
    "07 WILD testing DELETE with autoversioncontrol",
    "08 WILD testing PUT/MKCOL with autoversioncontrol",
    "09 WILD testing GET with autoversioncontrol",
    "10 WILD testing PUT/MKCOL/GET with autoversioncontrol",
    "11 WILD testing DELETE/PUT/MKCOL with autoversioncontrol",
    "12 WILD testing DELETE/PUT/MKCOL/MOVE with autoversioncontrol",
    "13 WILD testing LOCK/UNLOCK with autoversioncontrol",
  };

  public static int NrOfTests = desc.length;

  public WildTest(Integer nr) {
    super(nr, desc[nr.intValue()]);
  }

  protected void runTest() throws Exception {
      //logComment(String.valueOf(nr));
      boolean autoversioncontrol = (nr > 6);
      clear(request);
      switch (nr % 7) {
        case 0:
          if (autoversioncontrol) {
            //wait three seconds so the locks will vanish...
            Thread.sleep(3000);
          }

          // make sure the context is authorized
          request.setMethod("HEAD");
          request.setURI(localpart);
          requester.perform(request);
          
          request.setMethod("MKCOL");
          request.setURI(localpart);
          requester.perform(request);

          execute(4, Operations.Delete, 5000L, autoversioncontrol);
          request.setMethod("DELETE");
          for (int i = 0; i < Resource.RESOURCES.length; i++) {
            request.setURI(localpart + Resource.RESOURCES[i].getRID());
            // response =
            requester.perform(request);
          }
          break;

        case 1:
          execute(8, Operations.Create, 10000L, autoversioncontrol);
          break;

        case 2:
          execute(8, Operations.Get, 5000L, autoversioncontrol);
          break;

        case 3:
          execute(8, Operations.CreateAndGet, 5000L, autoversioncontrol);
          break;

        case 4:
          execute(8, Operations.CreateAndDelete, 20000L, autoversioncontrol);
          break;

        case 5:
          execute(8, Operations.CreateDeleteMove, 20000L, autoversioncontrol);
          break;

        case 6:
          execute(8, Operations.Lock, 5000L, autoversioncontrol);
          break;
      }
  }


  //---------------- private ---------------------------------------

  private void execute(int workerCount, OperationGenerator generator, long duration, boolean enhanced)
      throws Exception {
    Worker[] workers = new Worker[workerCount];
    for (int i = 0; i < workers.length; ++i) {
      workers[i] = new Worker(m_factory.getRequester(m_context), m_base, localpart, generator, enhanced);
    }

    for (int i = 0; i < workers.length; ++i) {
      workers[i].start();
    }

    Thread.sleep(duration);

    for (int i = 0; i < workers.length; ++i) {
      workers[i].abort();
    }

    for (int i = 0; i < workers.length; ++i) {
      while (!workers[i].aborted()) {
        Thread.sleep(250L);
      }
    }
    
    List errors = new LinkedList();
    for (int i = 0; i < workers.length; ++i) {
      errors.addAll(workers[i].getErrors());
    }
    if (errors.size() > 0) {
      throw new MultipleException("Exception(s) in Worker Thread(s) caught.", errors);
    }
    
  }

  private static final Random RANDOM = new Random(System.currentTimeMillis());

  private static final DomPath DP_status = new DomPath("/D:multistatus/D:response/D:status",
                                                       DomPath.DEFAULT_NAMESPACE_BINDINGS);
  private static final DomPath DP_href = new DomPath("../D:href",
                                                     DomPath.DEFAULT_NAMESPACE_BINDINGS);

  private static void assertMultiStatus(Document doc, int[] okCodes, String method, String invocation) throws Exception {
    Node[] nodes = DP_status.matchNodes(doc);
    for (int i = 0, n = nodes.length; i < n; ++i) {
      int status = new HttpStatus(DomElement.getText(nodes[i])).getCode();
      boolean ok = false;
      if (status == 424) {
        ok = true;
      }
      else {
        for (int j = 0; j < okCodes.length; j++) {
          ok = ok || (status == okCodes[j]);
        }
      }
      if (!ok) {
        Node href = DP_href.matchFirst(nodes[i]);
        if (href != null) {
          fail(method + " multistatus unexpected: " + status + " on " + DomElement.getText(href) + ", invocation: " + invocation);
        }
        else {
          fail(method + " multistatus unexpected: " + DomSerializer.toString(doc, true));
        }
      }
    }
  }
  
  private static Name getCondition(IResponse response) {
    Document doc = response.getDocument();
    if (doc == null || !Consts.N_error.equalsNode(doc.getDocumentElement())) {
      return null;
    }
    else {
      for (Node c = doc.getDocumentElement().getFirstChild(); c != null; c = c.getNextSibling()) {
        if (c.getNodeType() == Node.ELEMENT_NODE) return new Name(c);
      }
    }
    return null;
  }

  private static String getInvocationCounter(IResponse response) {
    String invoc = response.getHeader(Consts.SERVLET_TRACE_HEADER);
    return invoc == null ? "" : invoc;
  }

  private static void dumpResponse(HttpRequest request, IResponse response) {

    Name cond = getCondition(response);
    String invoc = getInvocationCounter(response);

    StringBuffer buf = new StringBuffer();
    buf.append(request.getMethod()).append(" on ");
    buf.append(request.getURI()).append(" returned: ").append(response.getStatus());
    buf.append(" ");
    buf.append(response.getStatusDescription());
    buf.append((cond == null ? "" : ", condition: " + cond.toString()));
    buf.append((invoc == null ? "" : ", invocation: " + invoc));

    if (response.hasDocument()) {
      buf.append("\n");
      buf.append("response body is: ");
      buf.append(DomSerializer.toString(response.getDocument(), true));
    }
    fail(buf.toString());
  }


  private static final class Resource {

    private static final Resource[] RESOURCES = new Resource[]{
      new Resource("/file1", Resource.PLAIN),
      new Resource("/file2", Resource.PLAIN),
      new Resource("/file3", Resource.PLAIN),
      new Resource("/file4", Resource.PLAIN),
      new Resource("/file5", Resource.PLAIN),
      new Resource("/file6", Resource.PLAIN),
      new Resource("/file7", Resource.PLAIN),
      new Resource("/file8", Resource.PLAIN),
      new Resource("/file9", Resource.PLAIN),
      new Resource("/dir1", Resource.COLLECTION),
      new Resource("/dir2", Resource.COLLECTION),
      new Resource("/dir3", Resource.COLLECTION),
      new Resource("/dir4", Resource.COLLECTION),
      new Resource("/dir1/file11", Resource.PLAIN),
      new Resource("/dir1/file12", Resource.PLAIN),
      new Resource("/dir1/file13", Resource.PLAIN),
      new Resource("/dir1/file14", Resource.PLAIN),
      new Resource("/dir2/file21", Resource.PLAIN),
      new Resource("/dir2/file22", Resource.PLAIN),
      new Resource("/dir2/file23", Resource.PLAIN),
      new Resource("/dir2/file24", Resource.PLAIN),
      new Resource("/dir3/file31", Resource.PLAIN),
      new Resource("/dir3/file32", Resource.PLAIN),
      new Resource("/dir3/file33", Resource.PLAIN),
      new Resource("/dir3/file34", Resource.PLAIN),
      new Resource("/dir4/file41", Resource.PLAIN),
      new Resource("/dir4/file42", Resource.PLAIN),
      new Resource("/dir4/file43", Resource.PLAIN),
      new Resource("/dir4/file44", Resource.PLAIN),
    };

    public static final Resource getRandom() {
      int i = RANDOM.nextInt(RESOURCES.length);
      return RESOURCES[i];
    }

    public static final Resource getRandom(String excludeNamespace) {
      Resource r = getRandom();
      if (excludeNamespace != null) {
        while (r != null && r.getRID().startsWith(excludeNamespace)) {
          r = getRandom();
        }
      }
      return r;
    }

    public static final int PLAIN = 0;
    public static final int COLLECTION = 1;

    private final String rid;
    private final int type;

    public Resource(String rid, int type) {
      this.rid = rid;
      this.type = type;
    }

    public String getRID() {
      return this.rid;
    }

    public int getType() {
      return this.type;
    }
  }

  private static final class Content {

    private static final Content[] CONTENTS = new Content[]{
      new Content("application/octet-stream", 32890),
      new Content("text/plain", 13709),
      new Content("text/plain;charset=UTF-8", 8734),
      new Content("application/octet-stream", 280987),
    };

    public static final Content getRandom() {
      int i = RANDOM.nextInt(CONTENTS.length);
      return CONTENTS[i];
    }

    public static boolean check(int length, InputStream stream, String invocation)
        throws IOException {
      Content c = null;
      for (int i = 0; i < CONTENTS.length; ++i) {
        if (length == CONTENTS[i].getLength()) {
          c = CONTENTS[i];
          break;
        }
      }

      boolean result = false;
      if (c != null) {
        result = true;
        for (int i = 0; i < length; ++i) {
          int b = stream.read();
          if (b != ((c.bytes[i]) & 0xff)) {
            if (b == -1) {
              fail("unexpected EOF at offset " + i + ", invocation: " + invocation);
            }
            else {
              fail("unexpected bytes at offset " + i + ": " + b + " instead of: " + ((c.bytes[i]) & 0xff) + ", invocation: " + invocation);
            }
            result = false;
            break;
          }
        }

        try {
          stream.close();
        }
        catch (IOException e) {
          // ignore
        }
      }
      else {
        fail("unexpected content length: " + length);
      }
      return result;
    }

    private final String type;
    private final byte[] bytes;

    public Content(String type, int length) {
      this.type = type;
      this.bytes = new byte[length];
      for (int i = 0; i < length; ++i) {
        this.bytes[i] = (byte) RANDOM.nextInt();
      }
    }

    public String getType() {
      return this.type;
    }

    public int getLength() {
      return this.bytes.length;
    }

    public InputStream getStream() {
      return new ByteArrayInputStream(this.bytes);
    }
  }

  private static interface Operation {

    public void perform(Resource resource, IRequester requester, String base,
                        String uriPrefix, boolean enhanced)
        throws Exception;
  }

  private static final class CreateOp implements Operation {

    private static final CreateOp INSTANCE = new CreateOp();

    public static CreateOp getInstance() {
      return INSTANCE;
    }

    private CreateOp() {
      //
    }

    public void perform(Resource resource, IRequester requester, String base,
                        String uriPrefix, boolean enhanced)
        throws Exception {
      HttpRequest request = new HttpRequest(uriPrefix + resource.getRID());
      request.setHeader(Consts.SERVLET_TRACE_HEADER, "WILD-CREATE");
      // make sure the context is authorized
      request.setMethod("HEAD");
      requester.perform(request);
      
      switch (resource.getType()) {
        case Resource.PLAIN:
          request.setMethod("PUT");
          Content c = Content.getRandom();
          request.setContentType(c.getType());
          request.setContentLength(c.getLength());
          request.setBody(c.getStream());
          break;

        case Resource.COLLECTION:
          request.setMethod("MKCOL");
          break;

        default:
          throw new IllegalArgumentException("CREATEOP: unknown resource type: " + resource.getType());
      }

      IResponse response = requester.perform(request);
      Name cond = getCondition(response);
      boolean isok = false;

      switch (response.getStatus()) {
        case 200: // overwritten
        case 201: // created
        case 204: // Apache/moddav response to not created, e.g. parent missing
//        case 503: // deadlock
          isok = true;
          break;

        case 403:
          isok = DavConditions.N_checkout_of_checked_out_version_is_discouraged.equals(cond)
              || DavConditions.N_checkout_of_version_with_descendant_is_discouraged.equals(cond);
          break;

        case 405: // MKCOL not allowed
//          isok = resource.getType() == Resource.COLLECTION;
          isok = true;  // ResourceType may change during operation
          break;

        case 409:
          isok = DavConditions.N_SAP_parent_collection_must_exist.equals(cond)
              || DavConditions.N_SAP_resource_supports_entity_body.equals(cond)
               //because autoversioning is not atomic at the moment, we have to accept DavConditions.N_must_be_checked_out
              || DavConditions.N_must_be_checked_out.equals(cond);
          break;

        case 207:
          WildTest.assertMultiStatus(response.getDocument(), new int[] {423}, request.getMethod(), getInvocationCounter(response));
          isok = true;
          break;
      }

      if (!isok) {
        dumpResponse(request, response);
      }
      else {
        String etag = response.getHeader(Consts.HEADER_ETAG);
        if (etag != null) {
          if (response.getStatus() >= 400) {
            fail("got unexpected ETag on response status " + response.getStatus() + " " + getInvocationCounter(response));
          }
          else {
            recordETag(resource.getRID(), etag);
          }
        }
        if (enhanced && resource.getType() == Resource.COLLECTION) {
          request = new HttpRequest(uriPrefix + resource.getRID());
          request.setHeader(Consts.SERVLET_TRACE_HEADER, "WILD-CREATE-SETENHANCED");
          request.setMethod("PROPPATCH");
          request.setBody(getEcProppatchBody(true));
          response = requester.perform(request);
          if (response.getStatus() != 404
            && response.getStatus() != 207
//            && response.getStatus() != 503
            && response.getStatus() != 501)
            dumpResponse(request, response);
        }
      }
    }
  }

  private static String getEcProppatchBody(boolean enhanced) {

    String s = "<propertyupdate xmlns='DAV:'>";
    s += "<set><prop><enhancedcollection xmlns='http://sapportals.com/xmlns/cm' "
        + "xmlns:xsi='http://www.w3.org/2001/XMLSchema-instance' xmlns:xsd='http://www.w3.org/2001/XMLSchema' "
        + "xsi:type='xsd:boolean'>" + new Boolean(enhanced) + "</enhancedcollection></prop></set>";

    s += "</propertyupdate>";

    return s;
  }

  private static final class GetOp implements Operation {

    private static final GetOp INSTANCE = new GetOp();

    public static GetOp getInstance() {
      return INSTANCE;
    }

    private GetOp() {
      //
    }

    public void perform(Resource resource, IRequester requester, String base, String uriPrefix, boolean enhanced)
        throws Exception {
      if (resource.getType() != Resource.PLAIN) {
        return;
      }
      HttpRequest request = new HttpRequest(uriPrefix + resource.getRID());
      request.setHeader(HEADER_MAN, '"' + HTTPEXT_N_NO_CACHE_REDIRECT.getNamespaceUri() + '"' + " ;ns=01");
      request.setHeader("01-" + HTTPEXT_N_NO_CACHE_REDIRECT.getLocalName(), "");
      request.setHeader(Consts.SERVLET_TRACE_HEADER, "WILD-GET");
      request.setMethod("GET");
      request.setHeader("Translate", "F");
      request.expectsResponseStream(true);

      IResponse response = requester.perform(request);
      InputStream stream = response.getStream();
      switch (response.getStatus()) {
        case 200: // ok
          boolean ok = false;
          long length = response.getContentLength();
          if (length < 0) {
            stream.close();
          }
          else if (length < Integer.MAX_VALUE) {
            try {
              ok = Content.check((int) length, stream, getInvocationCounter(response));
            }
            catch (Exception e) {
              fail("GET content check failed: " + e.getMessage());
            }
          }
          else {
            stream.close();
          }
          if (!ok) {
            fail("GET unexpected/incomplete content on " + resource.getRID() + " length " + length + " invocation: " +
                               getInvocationCounter(response));
          }
          break;
        case 404: // not found
        case 405: // not a resource?
        case 409: // parent does not exist?
//        case 503: // deadlock
          if (stream != null) {
            stream.close();
          }
          break;
        default:
          if (stream != null) {
            stream.close();
          }
          dumpResponse(request, response);
          break;
      }
    }
  }

  private static final class LockOp implements Operation {

    private static final LockOp INSTANCE = new LockOp();

    public static LockOp getInstance() {
      return INSTANCE;
    }

    private LockOp() {
      //
    }

    public void perform(Resource resource, IRequester requester, String base, String uriPrefix, boolean enhanced)
        throws Exception {
      HttpRequest request = new HttpRequest(uriPrefix + resource.getRID());
      request.setMethod("LOCK");
      request.setHeader(Consts.SERVLET_TRACE_HEADER, "WILD-LOCK");
      request.setHeader("Depth", "0");
      request.setHeader("Timeout", "Second-2");
      request.setContentType("text/xml");
      request.setBody(makeLockInfo("WILD-" + Thread.currentThread().getName()));

      IResponse response = requester.perform(request);
      String locktoken = null;

      switch (response.getStatus()) {

        case 200: // ok
        case 201: // resource created with lock
          locktoken = response.getHeader("Lock-Token");
          break;

        case 207: {
            Document doc = response.getDocument();
            WildTest.assertMultiStatus(doc, new int[] {423}, "LOCK", getInvocationCounter(response));
          }
          break;

        case 404: // not found
        case 405: // not a resource?
        case 409: // parent does not exist?
        case 423: // locked
        case 501: // repository does not support locks
//        case 503: // deadlock
          break;

        default:
          dumpResponse(request, response);
          break;
      }

      if (locktoken != null) {
        Thread.sleep(500L);
        request.clear();
        request.setURI(uriPrefix + resource.getRID());
        request.setMethod("UNLOCK");
        request.setHeader(Consts.SERVLET_TRACE_HEADER, "WILD-UNLOCK");
        request.setHeader("Lock-Token", locktoken);

        response = requester.perform(request);
        switch (response.getStatus()) {
          case 204: // UNLOCKED
          case 403: // lock token does not match
            break;
          default:
            dumpResponse(request, response);
            break;
        }
      }
    }

    private String makeLockInfo(String owner) {
      String res = "<lockinfo xmlns='DAV:'>";

      if (owner != null) {
        res += "<owner>" + owner + "</owner>";
      }

      res += "<lockscope><exclusive /></lockscope><locktype><write /></locktype></lockinfo>";
      return res;
    }
  }

  private static final class DeleteOp implements Operation {

    private static final DeleteOp INSTANCE = new DeleteOp();

    public static DeleteOp getInstance() {
      return INSTANCE;
    }

    private DeleteOp() {
      //
    }

    public void perform(Resource resource, IRequester requester, String base, String uriPrefix, boolean enhanced)
        throws Exception {
      HttpRequest request = new HttpRequest(uriPrefix + resource.getRID());
      request.setHeader(Consts.SERVLET_TRACE_HEADER, "WILD-DELETE");
      request.setMethod("DELETE");
      request.setHeader("Depth", "infinity");

      IResponse response = requester.perform(request);
      switch (response.getStatus()) {
        case 204: // deleted
        case 404: // not found
        case 409: // parent does not exist
//        case 503: // deadlock
          break;
        case 207:
//          WildTest.assertMultiStatus(response.getDocument(), new int[] {404, 503}, "DELETE", getInvocationCounter(response));
          WildTest.assertMultiStatus(response.getDocument(), new int[] {404}, "DELETE", getInvocationCounter(response));
          break;
        default:
          dumpResponse(request, response);
          break;
      }
    }
  }

  private static final class MoveOp implements Operation {

    private static final MoveOp INSTANCE = new MoveOp();

    public static MoveOp getInstance() {
      return INSTANCE;
    }

    private MoveOp() {
      //
    }

    public void perform(Resource resource, IRequester requester, String base, String uriPrefix, boolean enhanced)
        throws Exception {
      Resource destination = Resource.getRandom(resource.getRID());
      while (destination == null) {
        destination = Resource.getRandom(resource.getRID());
      }

      HttpRequest request = new HttpRequest(uriPrefix + resource.getRID());
      request.setHeader(Consts.SERVLET_TRACE_HEADER, "WILD-MOVE to " + destination.getRID());
      request.setMethod("MOVE");
      request.setHeader("Depth", "infinity");
      request.setHeader("Destination", base + uriPrefix + destination.getRID());

      IResponse response = requester.perform(request);
      Name cond = getCondition(response);
      boolean isok = false;

      switch (response.getStatus()) {
        case 201: // created
        case 204: // moved
        case 404: // not found
        case 409: // at destination is no collection
//        case 503: // deadlock
          isok = true;
          break;

        case 403:
          isok = DavConditions.N_SAP_destination_may_not_be_ancestor_of_source.equals(cond)
              || DavConditions.N_SAP_destination_must_be_in_collection.equals(cond);
          break;
      }

      if (!isok) {
        dumpResponse(request, response);
      }
    }
  }

  private static interface OperationGenerator {
    Operation next();
  }

  private static class Operations implements OperationGenerator {

    public static final Operations Delete = new Operations(new Operation[]{
      DeleteOp.getInstance(),
    });
    public static final Operations Create = new Operations(new Operation[]{
      CreateOp.getInstance(),
    });
    public static final Operations Get = new Operations(new Operation[]{
      GetOp.getInstance(),
    });
    public static final Operations Lock = new Operations(new Operation[]{
      LockOp.getInstance(),
    });
    public static final Operations CreateAndGet = new Operations(new Operation[]{
      CreateOp.getInstance(),
      GetOp.getInstance(),
    });
    public static final Operations CreateAndDelete = new Operations(new Operation[]{
      CreateOp.getInstance(),
      DeleteOp.getInstance(),
    });
    public static final Operations CreateDeleteMove = new Operations(new Operation[]{
      CreateOp.getInstance(),
      DeleteOp.getInstance(),
      MoveOp.getInstance(),
    });

    private final Operation[] ops;
    private final Random r;

    private Operations(Operation[] ops) {
      this.ops = ops;
      r = new Random(System.currentTimeMillis());
    }

    public Operation next() {
      int index = r.nextInt(this.ops.length);
      return this.ops[index];
    }
  }

  private static final class Worker extends Thread {

    private final OperationGenerator gen;
    private final IRequester requester;
    private final String prefix;
    private final String base;
    private boolean abort = false;
    private boolean aborted = false;
    private boolean enhanced = false;
    private List errors = new LinkedList();

    public Worker(IRequester requester, String base, String prefix, OperationGenerator gen, boolean enhanced) {
      this.requester = requester;
      this.base = base;
      this.prefix = prefix;
      this.gen = gen;
      this.enhanced = enhanced;
    }

    public void run() {
      while (!this.abort) {
        try {
          Operation op = gen.next();
          op.perform(Resource.getRandom(), this.requester, this.base, this.prefix, this.enhanced);
        }
        catch (Throwable t) {
          this.errors.add(t);
        }
      }
      this.aborted = true;
    }

    public void abort() {
      this.abort = true;
    }

    public boolean aborted() {
      return this.aborted;
    }
    
    public List getErrors() {
      return this.errors;
    }
  }

  private static class MultipleException extends AssertionFailedError {
    
    private List errors;
    
    public MultipleException(String message, List errors) {
      super(message);
      this.errors = errors;
    }

    public String getLocalizedMessage() {
      return this.getMessage();
    }

    public void printStackTrace() {
      this.printStackTrace(System.err);
    }

    public void printStackTrace(PrintStream s) {
      super.printStackTrace(s);
      s.println("Contained exceptions are:");
      int i = 0;
      for (Iterator it = this.errors.iterator(); it.hasNext(); ) {
        s.print("#" + (i++) + " ");
        Throwable t = (Throwable)it.next();
        t.printStackTrace(s);
      }
    }

    public void printStackTrace(PrintWriter s) {
      super.printStackTrace(s);
      s.println("Contained exceptions are:");
      int i = 0;
      for (Iterator it = this.errors.iterator(); it.hasNext(); ) {
        s.print("#" + (i++) + " ");
        Throwable t = (Throwable)it.next();
        t.printStackTrace(s);
      }
    }

  }
  
  // tracking ETags

  private static Map etagmap = new HashMap();

  private static synchronized void recordETag(String rid, String etag) {

    Set etags = (Set) etagmap.get(rid);

    if (etags == null) {
      etags = new HashSet();
      etags.add(etag);
      etagmap.put(rid, etags);
    }
    else {
      if (!etags.add(etag)) {
        fail("duplicate ETag " + etag + " on RID " + rid);
      }
    }
  }
}
