package com.sapportals.wcm.protocol.webdav.server.tests;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.StringTokenizer;

import junit.framework.TestCase;
import junit.framework.TestSuite;

import com.sap.tc.logging.Location;

import com.sapportals.wcm.WcmException;
import com.sapportals.wcm.util.http.HttpRequest;
import com.sapportals.wcm.util.http.IRequester;
import com.sapportals.wcm.util.http.IResponse;

/*
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author martin.boettcher@greenbytes.de
 * @author stefan.eissing@greenbytes.de
 * @author julian.reschke@greenbytes.de
 * @version $Id: AllTests.java,v 1.13 2005/10/11 16:03:01 mko Exp $
 */

public class AllTests extends TestCase {

  private static final Location log = Location.getLocation(AllTests.class);

  public static final String DEFAULT_SUITE = "WDServlet";

  private static class OptionsSet {
    private final String features;
    private final String methods;

    public OptionsSet(String methods, String features) {
      this.methods = methods.toUpperCase();
      this.features = features;
    }

    public boolean supportsFeature(String feature) {
      return (this.features != null && (this.features.indexOf(feature) >= 0));
    }

    public boolean supportsMethod(String method) {
      return (this.methods != null && (this.methods.indexOf(method.toUpperCase()) >= 0));
    }

    public String toString() {
      return "OptionSet["+this.features+", "+this.methods+"]";
    }
  }

  public AllTests(String name) {
    super(name);
  }

  private static Set readCases() {
    String cases = System.getProperty("cases", null);
    if (cases != null && cases.length() > 0) {
      StringTokenizer tok = new StringTokenizer(cases, ", \t\n\r");
      Set s = new HashSet();
      while (tok.hasMoreTokens()) {
        s.add(tok.nextToken().toUpperCase()); 
      }
      return Collections.unmodifiableSet(s);
    }
    else {
      return null; 
    }
  }
  
  private static void addTest(TestSuite suite, String name, Class c, FSTestSuite factory) {
    Set cases = readCases();
    if (cases == null || cases.contains(name.toUpperCase())) {
      suite.addTest(factory.suite(c, name));
    }
  }
  
  public static TestSuite suite() throws Exception {
    try {
      log.debugT("building test suite");
      FSTestSuite factory = new FSTestSuite();
      String base = factory.getBaseURI();
      String host = factory.getHost();

      String flavour = System.getProperty("suite", DEFAULT_SUITE);
      String testContent = System.getProperty("testContent", "no");
      String servletPath = System.getProperty("servletPath", null);

      if (!flavour.equals(DEFAULT_SUITE)) {
        System.setProperty("statusFlavour", flavour);
      }

      TestSuite suite = new TestSuite("DAV-compliance " + flavour + " at '"
                                      + host + base + "'");

      factory.setUp(true);

      OptionsSet baseSet = getFeatures(base, factory);

      addTest(suite, "RS", RepositoryManagerStatusTest.class, factory);

      try {
        addTest(suite, "OPTIONS", OptionsTest.class, factory);
        addTest(suite, "TR", TraceTest.class, factory);
      }
      catch (Exception ex) {
        ex.printStackTrace(System.err);
      }
      catch (Throwable t) {
        t.printStackTrace(System.err);
      }

      OptionsSet fileSet = getFeatures(base+"/ANSI.TXT", factory);
      if (fileSet == null) {
        fileSet = baseSet;
      }

      if (base.startsWith("/wcm/docs/dav/") && !fileSet.supportsFeature("version-history")) {
        System.out.println("base options: "+baseSet);
        System.out.println("file options: "+fileSet);
      }

      if (baseSet != null) {
        addTest(suite, "PROPFIND", PropfindTest.class, factory);
        if (baseSet.supportsMethod("PROPPATCH") || fileSet.supportsMethod("PROPPATCH")) {
          addTest(suite, "PROPPATCH", ProppatchTest.class, factory);
        }
        addTest(suite, "MKCOL", MkColTest.class, factory);
        addTest(suite, "HEAD", HeadTest.class, factory);
        addTest(suite, "POST", PostTest.class, factory);
        addTest(suite, "GET", GetTest.class, factory);
        addTest(suite, "PUT", PutTest.class, factory);
        addTest(suite, "COPY", CopyTest.class, factory);
        addTest(suite, "MOVE", MoveTest.class, factory);

        if (baseSet.supportsMethod("LOCK") || fileSet.supportsMethod("LOCK")) {
          addTest(suite, "LO", LockTest.class, factory);
          addTest(suite, "SL", SharedLockTest.class, factory);
          addTest(suite, "DL", DeepLockTest.class, factory);
        }
        addTest(suite, "DEL", DeleteTest.class, factory);

        if (baseSet.supportsFeature("ordered-collections") ||
          baseSet.supportsFeature("ordering")) {
          addTest(suite, "OC", OrderedCollectionTest.class, factory);
          addTest(suite, "OP", OrderpatchTest.class, factory);
        }

        if (baseSet.supportsFeature("redirectrefs")) {
          addTest(suite, "RF", RedirectrefTest.class, factory);
        }

        if ((baseSet.supportsFeature("ordering") || baseSet.supportsFeature("ordered-collections"))
          && baseSet.supportsFeature("redirectrefs")) {
          addTest(suite, "IL", InternalLinksTest.class, factory);
        }

        if (baseSet.supportsFeature("version-control") || fileSet.supportsFeature("version-control")) {
          addTest(suite, "VS", VersionTest.class, factory);
          addTest(suite, "EC", EnhancedCollectionTest.class, factory);
          if (baseSet.supportsFeature("redirectrefs")) {
            addTest(suite, "VL", VersionedLinksTest.class, factory);
          }
        }

        if (fileSet.supportsFeature("version-history")) {
          addTest(suite, "VH", VersionHistoryResourceTest.class, factory);
        }

        if (fileSet.supportsFeature("version-controlled-collection")) {
          addTest(suite, "VC", VersionedCollectionTest.class, factory);
        }

        if (fileSet.supportsFeature("update")) {
          addTest(suite, "UP", UpdateTest.class, factory);
        }

        if (baseSet.supportsFeature("DASL")) {
          addTest(suite, "DA", DaslTest.class, factory);
        }

        if (baseSet.supportsFeature("access-control")) {
          addTest(suite, "ACL", AclTest.class, factory);
          if (fileSet.supportsFeature("version-history")){
            addTest(suite, "AI", AclInheritanceTest.class, factory);
          }
          addTest(suite, "UM", UmRepTest.class, factory);
        }

        if (servletPath != null) {
          addTest(suite, "SA", ServletActionsTest.class, factory);
        }

        if (baseSet.supportsMethod("LOCK") || fileSet.supportsMethod("LOCK")) {
          addTest(suite, "OL", MsOfficeLockTest.class, factory);
        }

        if (testContent.equals("yes")) {
          addTest(suite, "CT", ContentTest.class, factory);
        }

        if (fileSet.supportsFeature("working-resource")) {
          addTest(suite, "WR", WorkingResourceTest.class, factory);
        }

        if (baseSet.supportsMethod("BIND") || fileSet.supportsMethod("BIND")) {
          addTest(suite, "BIND", BindTest.class, factory);
        }

        if (baseSet.supportsFeature("workspace")) {
          addTest(suite, "WS", WorkspaceTest.class, factory);
        }

        addTest(suite, "URM", UriToResourceMappingTest.class, factory);

        addTest(suite, "ETAG", EntityTagTest.class, factory);

        addTest(suite, "REPORTING", ReportingTest.class, factory);
      
        if (FSInteractionTest.hasCollectionsInFS()) {
          addTest(suite, "FSI", FSInteractionTest.class, factory);
        }
      
        addTest(suite, "SE", ServletEngineTest.class, factory);
        addTest(suite, "XY", XythosClientTest.class, factory);
      }
      return suite;
    }
    catch (Exception ex) {
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
      return null;
    }
  }

  private static OptionsSet getFeatures(String uri, FSTestSuite factory) {
    try {
      IRequester requester = factory.getRequester();
      HttpRequest request = new HttpRequest(uri);
      request.setMethod("OPTIONS");
      request.setHeader("x-greenbytes-resource-context", "allow-bind-method=yes");
      IResponse response = requester.perform(request);
      if (response.getStatus() == 200) {
        String protocolFeatures = response.getHeader("DAV");

        if (protocolFeatures != null) {
          String methods = response.getHeader("Allow");
          if (methods != null && methods.indexOf("SEARCH") >= 0) {
            protocolFeatures += ", DASL";
          }

          return new OptionsSet(methods, protocolFeatures);
        }
      }
    }
    catch (WcmException ex) {
      ex.printStackTrace(System.err);
    }

    return null;
  }
}