package com.sapportals.wcm.protocol.webdav.server.tests;

import java.util.List;
import java.util.Set;

import org.w3c.dom.Element;

import com.sapportals.wcm.protocol.webdav.*;
import com.sapportals.wcm.util.http.IResponse;

/**
 * Various tests for BIND method.
 * <p>
 * Copyright (c) SAP AG 2002-2004
 * @author julian.reschke@greenbytes.de
 * @version $Id: BindTest.java,v 1.4 2004/05/25 16:29:22 jre Exp $
 */

public class BindTest extends TestMember {

  private static final String[] desc = new String[] {
    "00 BIND (setup)",
    "01 BIND check allow header",
    "02 BIND create second bind",
    "03 BIND check number of childs in collection",
    "04 BIND check resource-ids of both binds",
    "05 BIND check parent-sets of both binds",
    "06 BIND PUT to one bind, check both etags",
    "07 BIND PROPPATCH to one bind, PROPFIND on the other",
    "08 BIND LOCK on one bind, discover lock on other",
    "09 BIND make sure that second bind can be renamed",
    "10 BIND cleanup lock",
    "11 BIND copy collection containing two binds",
    "12 BIND UNBIND original bind, check remaining bind",
    "13 BIND bind unmapped resource",
    "14 BIND cyclic bind",
    "15 BIND to collection, try to create loop using MOVE",
    "   BIND cleanup",
  };

  public static int NrOfTests = desc.length;

  public BindTest(Integer nr) {
    super(nr, desc[nr.intValue()]);
  }

  private static String locktoken = null;

  protected void runTest() {
    if (nr == 0) super.logIdentifier("BIND");

    try {
      String respfile = "BIND-" + TestMember.THREEDIGITS.format(nr) + ".txt";
      IResponse response;

      switch (nr) {

        case 0:
          clear(request);
          request.setMethod("MKCOL");
          request.setURI(getLocalpart() + "/binds");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/binds/file");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;

        case 1:
          assertMethodAllowed(getLocalpart() + "/binds/", "BIND");
          break;

        case 2:
          {
            clear(request);
            request.setMethod("PROPFIND");
            request.setHeader("Depth", "1");
            request.setURI(getLocalpart() + "/binds/");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            List responses = DomElement.getChildren(response.getDocument().getDocumentElement(), Consts.N_response);
            assertTrue(responses.size() == 2);

            clear(request);
            request.setMethod("BIND");
            request.setHeader("x-greenbytes-resource-context", "allow-bind-method=yes");
            request.setURI(getLocalpart() + "/binds/");
            request.setBody("<bind xmlns='DAV:'>"
              + "<segment>second</segment>"
              + "<href>"+getLocalpart() + "/binds/file"+"</href></bind>");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("BIND");
            request.setHeader("overwrite", "F");
            request.setHeader("x-greenbytes-resource-context", "allow-bind-method=yes");
            request.setURI(getLocalpart() + "/binds/");
            request.setBody("<bind xmlns='DAV:'>"
              + "<segment>second</segment>"
              + "<href>"+getLocalpart() + "/binds/file"+"</href></bind>");
            response = requester.perform(request);
            checkStatusAndErrorElement(response, 409, DavConditions.N_can_overwrite);

            clear(request);
            request.setMethod("BIND");
            request.setHeader("x-greenbytes-resource-context", "allow-bind-method=yes");
            request.setURI(getLocalpart() + "/binds/");
            request.setBody("<bind xmlns='DAV:'>"
              + "<segment>second</segment>"
              + "<href>"+getLocalpart() + "/binds/file"+"</href></bind>");
            response = requester.perform(request);
            checkResponseCode(response, 204);
          }
          break;

        case 3:
          {
            // this tests that the child cache was updated properly after creating the new child binding
            clear(request);
            request.setMethod("PROPFIND");
            request.setHeader("Depth", "1");
            request.setURI(getLocalpart() + "/binds/");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            List responses = DomElement.getChildren(response.getDocument().getDocumentElement(), Consts.N_response);
            assertTrue("should have 2 childs, but has: " + (responses.size() - 1), responses.size() == 3);
          }
          break;

        case 4:
          {
            Element r1 = getProperty(getLocalpart() + "/binds/file", Consts.N_resource_id);
            Element r2 = getProperty(getLocalpart() + "/binds/second", Consts.N_resource_id);
            assertTrue(DomSerializer.toString(r1, false).equals(DomSerializer.toString(r2, false)));
          }
          break;

        case 5:
          {
            Element r1 = getProperty(getLocalpart() + "/binds/file", Consts.N_parent_set);
            Element r2 = getProperty(getLocalpart() + "/binds/second", Consts.N_parent_set);
            assertTrue(DomSerializer.toString(r1, false).equals(DomSerializer.toString(r2, false)));
          }
          break;

        case 6:
          {
            clear(request);
            request.setMethod("PUT");
            request.setBody("bind content");
            request.setURI(getLocalpart() + "/binds/file");
            response = requester.perform(request);
            checkResponseCode(response, 200);

            Element r1 = getProperty(getLocalpart() + "/binds/file", Consts.N_getetag);
            Element r2 = getProperty(getLocalpart() + "/binds/second", Consts.N_getetag);
            assertTrue(DomSerializer.toString(r1, false).equals(DomSerializer.toString(r2, false)));
          }
          break;

        case 7:
          {
            setProperty(getLocalpart() + "/binds/file", Consts.N_comment, "xyz", false, 200);

            Element r1 = getProperty(getLocalpart() + "/binds/second", Consts.N_comment);
            assertTrue(DomElement.getText(r1).equals("xyz"));
          }
          break;

        case 8:
          {
            clear(request);
            request.setMethod("LOCK");
            request.setHeader("Timeout", "Infinite");
            request.setBody("<lockinfo xmlns='DAV:'>"+
              "<lockscope><exclusive/></lockscope><locktype><write/></locktype>" +
              "<owner>foobar</owner>" +
              "</lockinfo>");
            request.setURI(getLocalpart() + "/binds/file");
            response = requester.perform(request);
            checkResponseCode(response, 200);
            locktoken = response.getHeader("lock-token");

            Element r1 = getProperty(getLocalpart() + "/binds/second", Consts.N_lockdiscovery);
            List al = DomElement.getChildren(r1, Consts.N_activelock);
            assertTrue(al.size() == 1);
            List lt = DomElement.getChildren((Element)al.get(0), Consts.N_locktoken);
            assertTrue(lt.size() == 1);
            Set lthrefs = DomElement.getHrefs((Element) lt.get(0));
            assertTrue(lthrefs.size() == 1);
            assertTrue(("<" + (String)lthrefs.iterator().next() + ">").equals(locktoken));

            // todo: check lockroot
          }
          break;

        case 9:
          {
            clear(request);
            request.setMethod("MOVE");
            request.setURI(getLocalpart() + "/binds/second");
            request.setHeader("Destination", getLocalpart() + "/binds/second2");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("MOVE");
            request.setURI(getLocalpart() + "/binds/second2");
            request.setHeader("Destination", getLocalpart() + "/binds/second");
            response = requester.perform(request);
            checkResponseCode(response, 201);
          }
          break;

        case 10:
          if (locktoken != null) {
            clear(request);
            request.setMethod("UNLOCK");
            request.setHeader("lock-token", locktoken);
            request.setURI(getLocalpart() + "/binds/file");
            response = requester.perform(request);
            checkResponseCode(response, 204);
          }
          break;

        case 11:
          {
            clear(request);
            request.setMethod("COPY");
            request.setURI(getLocalpart() + "/binds");
            request.setHeader("Depth", "Infinity");
            request.setHeader("Destination", getLocalpart() + "/binds2");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("PROPFIND");
            request.setURI(getLocalpart() + "/binds2");
            request.setHeader("Depth", "1");
            request.setBody("<propfind xmlns='DAV:'><prop><parent-set/></prop></propfind>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            checkResponseDocument(response, respfile, null);
          }
          break;

        case 12:
          clear(request);
          request.setMethod("UNBIND");
          request.setURI(getLocalpart() + "/binds/");
          request.setBody("<unbind xmlns='DAV:'><segment>file</segment></unbind>");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          clear(request);
          request.setMethod("HEAD");
          request.setURI(getLocalpart() + "/binds/second");
          response = requester.perform(request);
          checkResponseCode(response, 200);
          break;

        case 13:
          clear(request);
          request.setMethod("BIND");
          request.setHeader("x-greenbytes-resource-context", "allow-bind-method=yes");
          request.setURI(getLocalpart() + "/binds/");
          request.setBody("<bind xmlns='DAV:'>"
            + "<segment>teststetststs</segment>"
            + "<href>"+getLocalpart() + "/binds/doesnotexist"+"</href></bind>");
          response = requester.perform(request);
          checkResponseCode(response, 409);
          checkErrorElement(response, DavConditions.N_bind_source_exists);
          break;

        case 14:
          clear(request);
          request.setMethod("BIND");
          request.setHeader("x-greenbytes-resource-context", "allow-bind-method=yes");
          request.setURI(getLocalpart() + "/binds/");
          request.setBody("<bind xmlns='DAV:'>"
            + "<segment>loop</segment>"
            + "<href>"+getLocalpart() + "/binds/"+"</href></bind>");
          response = requester.perform(request);
          checkStatusAndErrorElement(response, 403, DavConditions.N_cycle_allowed);
          break;

        case 15:
          clear(request);
          request.setMethod("BIND");
          request.setHeader("x-greenbytes-resource-context", "allow-bind-method=yes");
          request.setURI(getLocalpart() + "/");
          request.setBody("<bind xmlns='DAV:'>"
            + "<segment>binds2</segment>"
            + "<href>"+getLocalpart() + "/binds/"+"</href></bind>");
          response = requester.perform(request);
          checkResponseCode(response, 201, 204);

          clear(request);
          request.setMethod("MOVE");
          request.setURI(getLocalpart() + "/binds2");
          request.setHeader("Destination", getLocalpart() + "/binds/binds2");
          response = requester.perform(request);
          checkStatusAndErrorElement(response, 403, DavConditions.N_cycle_allowed);

          clear(request);
          request.setMethod("UNBIND");
          request.setURI(getLocalpart() + "/");
          request.setBody("<unbind xmlns='DAV:'><segment>binds2</segment></unbind>");
          response = requester.perform(request);
          checkResponseCode(response, 200, 403);
          if (response.getStatus() == 403) {
            super.checkErrorElement(response, DavConditions.N_unbind_source_exists);
          }
          break;

        default:
          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/binds");
          response = requester.perform(request);
          checkResponseCode(response, 204, 404);

          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/binds2");
          response = requester.perform(request);
          checkResponseCode(response, 204, 404);
          break;

      }
    }
    catch (Exception ex) {
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }

}
