package com.sapportals.wcm.protocol.webdav.server.tests;

import com.sapportals.wcm.WcmException;
import com.sapportals.wcm.protocol.webdav.DavConditions;
import com.sapportals.wcm.util.http.IResponse;

/**
 * Various tests for COPY method.
 *
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author martin.boettcher@greenbytes.de
 * @author stefan.eissing@greenbytes.de
 * @author julian.reschke@greenbytes.de
 * @version $Id: CopyTest.java,v 1.4 2004/09/13 12:54:45 jre Exp $
 */

public class CopyTest extends TestMember
{
  private static final String[] desc = new String[]
  {
    "00 COPY with improper URI no dest",
    "01 COPY with unknown file no dest",
    "02 COPY with known file no dest",
    "03 COPY with improper URI",
    "04 COPY with unknown file",
    "05 COPY with unknown collection",
    "06 COPY with known file",
    "07 COPY with known file 2",
    "08 COPY with known file overwrite=F",
    "09 COPY with known file overwrite=T and absent overwrite header",
    "10 COPY with known file 3",
    "11 COPY removing copied file 1",
    "12 COPY removing copied file 2",
    "13 COPY with wrong destination host",
    "14 COPY with wrong destination port",
    "15 COPY Overwrite=T on collection",
    "16 COPY Overwrite=F on collection",
    "17 COPY Overwrite=T of non-collection on collection",
    "18 COPY source -> source/a (namespace violation)",
    "19 COPY source/a -> source",
    "20 COPY to destination with trailing blanks",
    "21 COPY to target in non-existing parent",

    "   COPY cleanup",
  };

  public static int NrOfTests = desc.length;

  public CopyTest(Integer nr)
  {
    super(nr, desc[nr.intValue()]);
  }

  protected void runTest() {

    if (nr == 0) super.logIdentifier("COPY");

    String respfile = "COPY-" + THREEDIGITS.format (nr) + ".txt";

    try
    {
      IResponse response;
      request.setMethod("COPY");
      request.removeHeader("Overwrite");

      switch (nr)
      {
        case 0:
          request.setURI(getLocalpart());
          response = requester.perform(request);
          checkResponseCode(response, 409, 400);
          break;

        case 1:
          request.setURI(getLocalpart() + "/abc");
          response = requester.perform(request);
          checkResponseCode(response, 404);
          break;

        case 2:
          request.setURI(getLocalpart() + "/Worddatei.doc");
          response = requester.perform(request);
          checkResponseCode(response, 409, 400);
          break;

        case 3:
          request.setHeader("Destination", this.getBaseAndLocalpart() + "/leer/zzz.yyy");
          request.setURI(getLocalpart());
          response = requester.perform(request);
          checkResponseCode(response, 403);
          break;

        case 4:
          request.setHeader("Destination", this.getBaseAndLocalpart() + "/leer/zzz.yyy");
          request.setURI(getLocalpart() + "/abc");
          response = requester.perform(request);
          checkResponseCode(response, 404);
          break;

        case 5:
          request.setHeader("Destination", this.getBaseAndLocalpart() + "/leer/zzz.yyy");
          request.setURI(getLocalpart() + "/a/b/c.xyz");
          response = requester.perform(request);
          checkResponseCode(response, 404);
          break;

        case 6:
          request.setHeader("Destination", this.getBaseAndLocalpart() + "/leer/zzz.yyy");
          request.setURI(getLocalpart() + "/Worddatei.doc");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;

        case 7:
          request.setHeader("Destination", this.getBaseAndLocalpart() + "/leer/zzz.yyy");
          request.setHeader("Overwrite", "T");
          request.setURI(getLocalpart() + "/Worddatei.doc");
          response = requester.perform(request);
          checkResponseCode(response, 204);
          break;

        case 8:
          request.setHeader("Destination", this.getBaseAndLocalpart() + "/leer/zzz.yyy");
          request.setURI(getLocalpart() + "/Worddatei.doc");
          request.setHeader("Overwrite", "F");
          response = requester.perform(request);
          checkResponseCode(response, 412);
          break;

        case 9:
          request.setHeader("Destination", this.getBaseAndLocalpart() + "/leer/zzz.yyy");
          request.setURI(getLocalpart() + "/Worddatei.doc");
          request.setHeader("Overwrite", "T");
          response = requester.perform(request);
          checkResponseCode(response, 204);

          // same thing without overwrite header (which should default to 'T')
          request.clear();
          request.setMethod("COPY");
          request.setHeader("Destination", this.getBaseAndLocalpart() + "/leer/zzz.yyy");
          request.setURI(getLocalpart() + "/Worddatei.doc");
          response = requester.perform(request);
          checkResponseCode(response, 204);
          break;

        case 10:
          request.setURI(getLocalpart() + "/Worddatei.doc");
          request.setHeader("Destination", this.getBaseAndLocalpart() + "/abcd.zzzz");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;

        case 11:
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/leer/zzz.yyy");
          response = requester.perform(request);
          checkResponseCode(response, 204);
          break;

        case 12:
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/abcd.zzzz");
          response = requester.perform(request);
          checkResponseCode(response, 204);
          break;

        case 13:
          request.setMethod("COPY");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          request.setHeader("Destination", "http://greenbytes.de:81" + "/leer/zzz.yyy");
          response = requester.perform(request);
          checkResponseCode(response, 502);
          break;

        case 14:
          String wrongBase;
          int index = this.getBaseAndLocalpart().lastIndexOf(':');
          if (index > 0
              && index+1 < this.getBase().length()
              && Character.isDigit(this.getBase().charAt(index+1))) {
            wrongBase = this.getBase().substring(0, index) + ":4711";
          }
          else {
            wrongBase = this.getBase() + ":4711";
          }
          request.setMethod("COPY");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          request.setHeader("Destination", wrongBase + "/leer/zzz.yyy");
          response = requester.perform(request);
          checkResponseCode(response, 502);
          break;

        case 15:
          setup2testcollections();

          clear(request);
          request.setMethod("COPY");
          request.setHeader("Overwrite", "T");
          request.setURI(getLocalpart() + "/source");
          request.setHeader("Destination", this.getBaseAndLocalpart() + "/destination");
          response = requester.perform(request);
          checkResponseCode(response, 204);

          clear(request);
          request.setMethod("PROPFIND");
          request.setURI(getLocalpart() + "/destination");
          request.setHeader("Depth", "1");
          super.checkMultistatus("<propfind xmlns='DAV:'><prop><resourcetype/></prop></propfind>", respfile);

          break;

        case 16:
          delete2testcollections();
          setup2testcollections();

          clear(request);
          request.setMethod("COPY");
          request.setHeader("Overwrite", "F");
          request.setURI(getLocalpart() + "/source");
          request.setHeader("Destination", this.getBaseAndLocalpart() + "/destination");
          response = requester.perform(request);
          checkResponseCode(response, 412);

          break;

        case 17:
          delete2testcollections();
          setup2testcollections();

          clear(request);
          request.setMethod("COPY");
          request.setHeader("Overwrite", "T");
          request.setURI(getLocalpart() + "/source/a");
          request.setHeader("Destination", this.getBaseAndLocalpart() + "/destination");
          response = requester.perform(request);
          checkResponseCode(response, 204);

          clear(request);
          request.setMethod("PROPFIND");
          request.setURI(getLocalpart() + "/destination");
          request.setHeader("Depth", "1");
          super.checkMultistatus("<propfind xmlns='DAV:'><prop><resourcetype/></prop></propfind>", respfile);

          // namespace sanity check
          clear(request);
          request.setMethod("HEAD");
          request.setURI(getLocalpart() + "/destination/b");
          response = requester.perform(request);
          checkResponseCode(response, 404);

          delete2testcollections();
          break;

        case 18:
          delete2testcollections();
          setup2testcollections();

          clear(request);
          request.setMethod("COPY");
          request.setHeader("Overwrite", "T");
          request.setURI(getLocalpart() + "/source");
          request.setHeader("Destination", this.getBaseAndLocalpart() + "/source/a");
          response = requester.perform(request);
          checkResponseCode(response, 403);
          checkErrorElement(response, DavConditions.N_SAP_destination_may_not_be_ancestor_of_source);

          delete2testcollections();
          break;

        case 19:
          delete2testcollections();
          setup2testcollections();

          clear(request);
          request.setMethod("COPY");
          request.setHeader("Overwrite", "T");
          request.setURI(getLocalpart() + "/source/a");
          request.setHeader("Destination", this.getBaseAndLocalpart() + "/source");
          response = requester.perform(request);
          checkResponseCode(response, 403);
          checkErrorElement(response, DavConditions.N_SAP_source_may_not_be_ancestor_of_destination);

          delete2testcollections();
          break;

        case 20:
          delete2testcollections();
          setup2testcollections();

          clear(request);
          request.setMethod("COPY");
          request.setURI(getLocalpart() + "/source/a");
          request.setHeader("Destination", this.getBaseAndLocalpart() + "/source/b%20");
          response = requester.perform(request);
          checkResponseCode(response, 403);

          delete2testcollections();
          break;
          
        case 21:
          request.setURI(getLocalpart() + "/Worddatei.doc");
          request.setHeader("Destination", this.getBaseAndLocalpart() + "/xyz/destination");
          response = requester.perform(request);
          checkResponseCode(response, 409);
          checkErrorElement(response, DavConditions.N_SAP_destination_must_be_in_collection);
          break;

        default:
          delete2testcollections();
          break;
      }
    }
    catch (Exception ex) {
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }

  private void setup2testcollections () throws WcmException {
    IResponse response;

    clear(request);
    request.setMethod("MKCOL");
    request.setURI(getLocalpart() + "/source");
    response = requester.perform(request);
    checkResponseCode(response, 201);

    clear(request);
    request.setMethod("PUT");
    request.setURI(getLocalpart() + "/source/a");
    response = requester.perform(request);
    checkResponseCode(response, 201);

    clear(request);
    request.setMethod("MKCOL");
    request.setURI(getLocalpart() + "/destination");
    response = requester.perform(request);
    checkResponseCode(response, 201);

    clear(request);
    request.setMethod("PUT");
    request.setURI(getLocalpart() + "/destination/b");
    response = requester.perform(request);
    checkResponseCode(response, 201);
  }

  private void delete2testcollections ()  throws WcmException {
    IResponse response;

    clear(request);
    request.setMethod("DELETE");
    request.setURI(getLocalpart() + "/source");
    response = requester.perform(request);
    checkResponseCode(response, 204, 404);

    clear(request);
    request.setMethod("DELETE");
    request.setURI(getLocalpart() + "/destination");
    response = requester.perform(request);
    checkResponseCode(response, 204, 404);
  }

}
