package com.sapportals.wcm.protocol.webdav.server.tests;

import java.util.List;

import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import com.sapportals.wcm.protocol.webdav.Consts;
import com.sapportals.wcm.protocol.webdav.DomElement;
import com.sapportals.wcm.protocol.webdav.dompath.DomPath;
import com.sapportals.wcm.util.http.IResponse;

/**
 * Various tests for namespace operations between repositories.
 *
 * <p>
 * Copyright (c) SAP AG 2002-2004
 * @author julian.reschke@greenbytes.de
 * @version $Id: CrossRepositoryTest.java,v 1.4 2004/05/25 16:29:22 jre Exp $
 */

public class CrossRepositoryTest extends TestMember
{
  private static final String[] desc = new String[]
  {
    "00 CR setup",
    "01 CR COPY complete test tree to target",
    "02 CR COPY/overwrite complete test tree to target",
    "03 CR MOVE tree including versioned resource",
    "04 CR PROPFIND on source namespace after move",
    "05 CR attempt BIND to other namespace",
    "06 CR move resource with ACL",
    "   CR cleanup",
  };

  public static int NrOfTests = desc.length;

  public CrossRepositoryTest(Integer nr)
  {
    super(nr, desc[nr.intValue()]);
  }

  static String targetCollection = null;

  static {
    targetCollection = System.getProperty("base2", null);
  }

  protected void runTest()
  {
    if (nr == 0) super.logIdentifier("CR");

    String respfile = "CR-" + TestMember.THREEDIGITS.format(nr) + ".txt";

    try
    {
      IResponse response;

      switch (nr)
      {
        case 00:
          break;

        case 01:
          clear(request);
          request.setMethod("DELETE");
          request.setURI(targetCollection);
          response = requester.perform(request);
          checkResponseCode(response, 204, 404);

          clear(request);
          request.setMethod("COPY");
          request.setURI(getLocalpart());
          request.setHeader("Destination", this.getBase() + targetCollection);
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;

        case 02:
          clear(request);
          request.setMethod("COPY");
          request.setURI(getLocalpart());
          request.setHeader("Overwrite", "T");
          request.setHeader("Destination", this.getBase() + targetCollection);
          response = requester.perform(request);
          checkResponseCode(response, 204);
          break;

        case 03:
          clear(request);
          request.setMethod("VERSION-CONTROL");
          request.setURI(getLocalpart() + "/Worddatei.doc");
          response = requester.perform(request);
          checkResponseCode(response, 200, 201);

          clear(request);
          request.setMethod("MOVE");
          request.setURI(getLocalpart());
          request.setHeader("Overwrite", "T");
          request.setHeader("Destination", this.getBase() + targetCollection);
          response = requester.perform(request);
          checkResponseCode(response, 207);
          checkResponseDocument(response, respfile, null);
          break;

        case 04:
          clear(request);
          request.setMethod("PROPFIND");
          request.setURI(getLocalpart());
          request.setHeader("Depth", "1");
          request.setBody("<propfind xmlns='DAV:'><prop/></propfind>");
          response = requester.perform(request);
          checkResponseCode(response, 207);
          checkResponseDocument(response, respfile, null);
          break;

        case 05:
          {
            clear(request);
            request.setMethod("BIND");
            request.setHeader("x-greenbytes-resource-context", "allow-bind-method=yes");
            request.setURI(getLocalpart() + "/Worddatei.doc");
            request.setBody("<bind xmlns='DAV:'><segment>testbind</segment>" +
              "<href>"+this.getBase() + targetCollection+"</href></bind>");
            response = requester.perform(request);

            if (response.getStatus() < 300 && response.getStatus() >= 200) {

              Element pset = getProperty(getLocalpart() + "/Worddatei.doc", Consts.N_parent_set);
              List parents = DomElement.getChildren(pset, Consts.N_parent);
              assertTrue(parents.size() == 2);

              // ok, try to delete the new bind again
              clear(request);
              request.setMethod("DELETE");
              request.setURI(this.getBase() + targetCollection + "/testbind");
              response = requester.perform(request);
              checkResponseCode(response, 204);

              // re-check parent set
              pset = getProperty(getLocalpart() + "/Worddatei.doc", Consts.N_parent_set);
              parents = DomElement.getChildren(pset, Consts.N_parent);
              assertTrue(parents.size() == 1);
            }
          }
          break;

        case 06:
          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/acltest");
          request.setBody("testcontent");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          request.setMethod("ACL");
          request.setURI(getLocalpart() + "/acltest");
          request.setBody("<acl xmlns='DAV:'>"
                            +"<ace><principal><all/></principal>"
                            +"  <grant><privilege><read/></privilege></grant>"
                            +"</ace>"
                            +"<ace><principal><all/></principal>"
                            +"  <grant><privilege><write/></privilege></grant>"
                            +"</ace>"
                            +"<ace><principal><all/></principal>"
                            +"  <grant><privilege><write-acl/></privilege></grant>"
                            +"</ace>"
                            +"</acl>");
          response = requester.perform(request);
          checkResponseCode(response, 200, 403, 501);

          if (response.getStatus() == 403) {
            NodeList elements
                = DomElement.getElements(
                    response.getDocument().getDocumentElement(),
                    Consts.N_protected_ace_conflict);
            assertTrue(elements.getLength() > 0);
          }
          else if (response.getStatus() == 200) {

            clear(request);
            request.setMethod("PROPFIND");
            request.setURI(this.getLocalpart() + "/acltest");
            request.setHeader("Depth", "0");
            response = requester.perform(request);
            this.checkMultistatus(
              "<propfind xmlns='DAV:'><prop><acl/></prop></propfind>",
              respfile,
              new DomPath("/D:multistatus/D:response/D:href", DomPath.DEFAULT_NAMESPACE_BINDINGS));

            clear(request);
            request.setMethod("MOVE");
            request.setURI(getLocalpart() + "/acltest");
            request.setHeader("Overwrite", "T");
            request.setHeader("Destination", this.getBaseAndLocalpart() + "/acltest2");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("PROPFIND");
            request.setURI(this.getLocalpart() + "/acltest2");
            request.setHeader("Depth", "0");
            response = requester.perform(request);
            this.checkMultistatus(
              "<propfind xmlns='DAV:'><prop><acl/></prop></propfind>",
              respfile,
              new DomPath("/D:multistatus/D:response/D:href", DomPath.DEFAULT_NAMESPACE_BINDINGS));

            clear(request);
            request.setMethod("MOVE");
            request.setURI(getLocalpart() + "/acltest2");
            request.setHeader("Overwrite", "T");
            request.setHeader("Destination", this.getBase() + targetCollection + "/acltest");
            response = requester.perform(request);
            checkResponseCode(response, 201, 403);

            if (response.getStatus() == 201) {

              clear(request);
              request.setMethod("PROPFIND");
              request.setURI(targetCollection + "/acltest");
              request.setHeader("Depth", "0");
              response = requester.perform(request);
              this.checkMultistatus(
                "<propfind xmlns='DAV:'><prop><acl/></prop></propfind>",
                respfile,
                new DomPath("/D:multistatus/D:response/D:href", DomPath.DEFAULT_NAMESPACE_BINDINGS));
            }
          }
          break;

        default:
          break;
      }
    }
    catch (Exception ex) {
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }
}
