package com.sapportals.wcm.protocol.webdav.server.tests;

import java.io.ByteArrayInputStream;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.List;

import org.w3c.dom.Element;

import com.sapportals.wcm.WcmException;
import com.sapportals.wcm.protocol.webdav.*;
import com.sapportals.wcm.protocol.webdav.dasl.BasicSearch;
import com.sapportals.wcm.protocol.webdav.dompath.DomPath;
import com.sapportals.wcm.repository.PropertyName;
import com.sapportals.wcm.repository.XMLMarkup;
import com.sapportals.wcm.util.http.DateFormat;
import com.sapportals.wcm.util.http.IResponse;


/**
 * Various tests for SEARCH method.
 * <p>
 * Copyright (c) SAP AG 2002-2004
 * @author julian.reschke@greenbytes.de
 * @version $Id: DaslTest.java,v 1.9 2004/07/06 14:42:47 jre Exp $
 */

public class DaslTest extends TestMember {

  private static final String[] desc = new String[]{
    "00 DA Setting up Searching",
    "01 DA make dir structure",
    "02 DA create plain resources",
    "03 DA set dead properties",
    "04 DA set typed properties",
    "05 DA set large properties",
    "06 DA set multi-valued properties",
    "07 DA set SAP-live properties",
    "08 DA QSD",
    "09 DA supported-query-grammar-set",
    "10 DA empty where clause",
    "11 DA someEq mv success",
    "12 DA eq mv fail",
    "13 DA someLikeEq mv success",
    "14 DA like mv fail",
    "15 DA not(someEq)",
    "16 DA eq on long property values",
    "17 DA ge on long property values",
    "18 DA eq on internal description property",
    "19 DA like on internal description property",
    "20 DA gte on internal description property",
    "21 DA isdefined dead string",
    "22 DA eq dead string",
    "23 DA like dead string",
    "24 DA eq dead long",
    "25 DA eq dead long typed",
    "26 DA eq dead date",
    "27 DA eq dead date typed",
    "28 DA mixed dead types",
    "29 DA search on iscollection pseudo-prop",
    "30 DA is-collection operator",
    "31 DA eq DAV:getcontenttype",
    "32 DA like DAV:getcontenttype",
    "33 DA gte DAV:getcontenttype",
    "34 DA like DAV:getcontenttype when server persists encoding",
    "35 DA comparisons on DAV:getcontentlength",
    "36 DA sorted on DAV:getcontentlength",
    "37 DA getlastmodified </> yesterday",
    "38 DA creationdate </> yesterday",
    "39 DA contenttype not eq application/msword",
    "40 DA contenttype eq application/msword or text/plain",
    "41 DA contenttype eq application/msword and text/plain / eq application/vnd.excel",
    "42 DA query with wrong datatype for internal SAP property",
    "43 DA with quotes and problematic Unicode in literals",
    "44 DA getlastmodified > just before PUT",
    "45 DA scope != arbiter",
    "46 DA scope invalid",
    "47 DA find all versions of resources in scope",
    "48 DA find all VCRs and versions with specific length",
    "49 DA find dead property with trailing blank",
    "50 DA find description with trailing blank",
    "51 DA find application/mswword from root collection",
    "52 DA eq dead boolean typed",
    "53 DA is-defined on XML typed property",
    "54 DA like on XML typed property",
    "55 DA is-defined on multivalued string property",
    "56 DA sanity checks on and/or",
    "57 DA not(is-defined) or equals (plus sub-expressions)",
    "   DA cleanup",
  };

  public static final int NrOfTests = desc.length;

  private static final DateFormat DF = new DateFormat();

  private static boolean supportsDead = false;
  private static boolean supportsCreationDate = false;
  private static boolean supportsTypedDead = false;
  private static boolean supportsLargeString = false;
  private static boolean supportsVersioning = false;
  private static boolean supportsMv = false;
  private static boolean supportsDesc = false;
  private static boolean persistsEncodingWithType = false;

  private static int DEEP_INCLUDING_VERSIONS = -1;

  private static Name pDesc = new Name(PropertyName.createDescription().getNamespace(), PropertyName.createDescription().getName());
  private static Name pString = new Name("tc:", "string");
  private static Name pVersString = new Name("tc:", "vers-string");
  private static Name pDate = new Name("tc:", "date");
  private static Name pLong = new Name("tc:", "long");
  private static Name pBoolean = new Name("tc:", "boolean");
  private static Name pLongString = new Name("tc:", "long-string");
  private static Name pMixed = new Name("tc:", "mixed");
  private static Name pMvLong = new Name("tc:", "mv-long");
  private static Name pMvString = new Name("tc:", "mv-string");
  private static Name pXml = new Name("tc:", "xml");

  private static String longStringVal;

  private static int totalResources;

  static {
    StringBuffer sb = new StringBuffer();
    for (int i = 0; i < 1000; i++) {
      sb.append("1234567890");
    }
    longStringVal = sb.toString();
  }

  private static String[] empty = new String[]{};

  public DaslTest(Integer nr) {
    super(nr, desc[nr.intValue()]);
  }

  protected void runTest() {
    if (nr == 0) logIdentifier("DA");

    try {
      String respfile = "DASL-" + TestMember.THREEDIGITS.format(nr) + ".txt";
      IResponse response;

      switch (nr) {

        case 0:
          assertMethodAllowed(getLocalpart(), "SEARCH");
          break;

        case 1:
          {
            for (int i = 0; i < collections.length; i++) {
              clear(request);
              request.setMethod("MKCOL");
              request.setURI(getLocalpart() + collections[i]);
              response = requester.perform(request);
              checkResponseCode(response, 201);
              totalResources += 1;
            }
          }
          break;

        case 2:
          {
            putBody("/dasl/1/word.doc", "application/msword", 123456);
            totalResources += 1;
            putBody("/dasl/1/excel.xls", "application/vnd.ms-excel", 0);
            totalResources += 1;
            putBody("/dasl/1/plain.txt", "text/plain", 8323);
            totalResources += 1;
            putBody("/dasl/1/plain-u.txt", "text/plain; charset=UTF-8", 8324);
            totalResources += 1;
            putBody("/dasl/1/no-ext", null, 434);
            totalResources += 1;
            putBody("/dasl/1/mv-long", null, 0);
            totalResources += 1;
            putBody("/dasl/1/mv-string", null, 0);
            totalResources += 1;
            putBody("/dasl/1/mv-string-empty", null, 0);
            totalResources += 1;
            putBody("/dasl/2/long-string-1", null, 0);
            totalResources += 1;
            putBody("/dasl/2/long-string-2", null, 0);
            totalResources += 1;
            putBody("/dasl/3/sapdesc1", null, 0);
            totalResources += 1;
            putBody("/dasl/3/sapdesc2", null, 0);
            totalResources += 1;
            putBody("/dasl/4/boolean", null, 0);
            totalResources += 1;
            putBody("/dasl/4/string", null, 0);
            totalResources += 1;
            putBody("/dasl/4/date", null, 0);
            totalResources += 1;
            putBody("/dasl/4/long", null, 0);
            totalResources += 1;
            putBody("/dasl/4/xml", null, 0);
            totalResources += 1;
            putBody("/dasl/5/mixed-integer", null, 0);
            totalResources += 1;
            putBody("/dasl/5/mixed-string1", null, 0);
            totalResources += 1;
            putBody("/dasl/5/mixed-string2", null, 0);
            totalResources += 1;
            putBody("/dasl/6/versioned", null, 0);
            totalResources += 1;

            Element ct = getProperty(getLocalpart() + "/dasl/1/plain-u.txt", Consts.N_getcontenttype);
            persistsEncodingWithType = DomElement.getText(ct).indexOf(";") >= 0;

            ct = getProperty(getLocalpart() + "/dasl/1/plain-u.txt", Consts.N_creationdate);
            supportsCreationDate = ct != null && DomElement.getText(ct).length() > 0;
          }
          break;

        case 3:
          {
            DavStatus stat = setProperty(getLocalpart() + "/dasl/4/string", pString, "property _%", false);

            if (stat.getStatus() == 200) {
              supportsDead = true;
              super.logComment("dead");
            }

            if (supportsDead) {
              setProperty(getLocalpart() + "/dasl/6/versioned", pVersString, "foobar", false);
              supportsVersioning = versionControl("/dasl/6/versioned");
            }
          }
          break;

        case 4:
          if (supportsDead) {
            DavStatus stat = setProperty(getLocalpart() + "/dasl/4/date", pDate,
                                         DF.parseISO8601("2002-12-17T12:00:00Z"), true);

            if (stat.getStatus() == 200 && stat.getCondition() == null) {
              supportsTypedDead = true;
              super.logComment("typed");

            }

            stat = setProperty(getLocalpart() + "/dasl/4/long", pLong, new Long(-123456789), true);
            assertTrue(stat.getStatus() == 200);

            stat = setProperty(getLocalpart() + "/dasl/4/boolean", pBoolean, Boolean.TRUE, true);
            assertTrue(stat.getStatus() == 200);

            stat = setProperty(getLocalpart() + "/dasl/4/xml", pXml, new XMLMarkup("<xml xmlns='tc:'><foobar>foobar</foobar></xml>"), true);
            assertTrue(stat.getStatus() == 200);

            stat = setProperty(getLocalpart() + "/dasl/5/mixed-integer", pMixed, new Integer(2), false);
            assertTrue(stat.getStatus() == 200);

            stat = setProperty(getLocalpart() + "/dasl/5/mixed-string1", pMixed, "1", false);
            assertTrue(stat.getStatus() == 200);

            stat = setProperty(getLocalpart() + "/dasl/5/mixed-string2", pMixed, "A", false);
            assertTrue(stat.getStatus() == 200);

          }
          break;

        case 5:
          {
            setProperty(getLocalpart() + "/dasl/2/long-string-1", pLongString, longStringVal, false);
            setProperty(getLocalpart() + "/dasl/2/long-string-2", pLongString, "1234567890", false);

            Element p = getProperty(getLocalpart() + "/dasl/2/long-string-1", pLongString);
            if (p != null && DomElement.getText(p).equals(longStringVal)) {
              supportsLargeString = true;
              super.logComment("ls");
            }
          }
          break;

        case 6:
          {
            List vals = new ArrayList();
            vals.add(new Long(-1));
            vals.add(new Long(42));

            DavStatus stat = setProperty(getLocalpart() + "/dasl/1/mv-long", pMvLong, vals, true);
            if (stat.getStatus() == 200 && stat.getCondition() == null) {
              supportsMv = true;
              super.logComment("mv");
            }

            vals = new ArrayList();
            vals.add("abcde");
            vals.add("fghij");
            stat = setProperty(getLocalpart() + "/dasl/1/mv-string", pMvString, vals, true);

            vals = new ArrayList();
            stat = setProperty(getLocalpart() + "/dasl/1/mv-string-empty", pMvString, vals, true);
          }
          break;

        case 7:
          {
            DavStatus stat = setProperty(getLocalpart() + "/dasl/3/sapdesc1",
                                         pDesc, "There is a risk in doing this", false);
            setProperty(getLocalpart() + "/dasl/3/sapdesc2",
                        pDesc, "RISK", false);

            if (stat.getStatus() == 200) {
              supportsDesc = true;
            }
          }
          break;

        case 8:
          clear(request);
          request.setMethod("SEARCH");
          request.setURI(getLocalpart());
          checkMultistatus("<query-schema-discovery xmlns='DAV:'>" +
                           "<basicsearch><from><scope><href>" + getLocalpart() + "/" + "</href><depth>infinity</depth></scope></from></basicsearch>" +
                           "</query-schema-discovery>", respfile);
          break;

        case 9:
          Element sqgs = getProperty(getLocalpart(), Consts.N_supported_query_grammar_set);

          DomPath bs = new DomPath("D:supported-query-grammar/D:grammar/D:basicsearch", DomPath.DEFAULT_NAMESPACE_BINDINGS);
          assertTrue(bs.matchNodes(sqgs).length == 1);

          clear(request);
          request.setMethod("OPTIONS");
          request.setURI(getLocalpart() + "/dasl");
          response = requester.perform(request);
          checkResponseCode(response, 200);
          String dasl = response.getHeader("dasl");
          assertTrue(dasl.indexOf("DAV:basicsearch") >= 0);
          break;

        case 10: {
            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     null,
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            List hrefs = getHrefsFromMultistatus(response.getDocument());
            assertTrue(hrefs.size() == totalResources);
          }
          break;

        case 11:
          if (supportsMv) {

            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     nestString(BasicSearch.N_SAP_someEq,
                                                prop(pMvLong) + literal("42")),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/1/mv-long");
          }
          break;

        case 12:
          if (supportsMv) {

            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprEq(pMvLong, "42"),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, empty);
          }
          break;

        case 13:
          if (supportsMv) {

            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     nestString(BasicSearch.N_SAP_someLike,
                                                prop(pMvString) + literal("%bc%%")),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/1/mv-string");
          }
          break;

        case 14:
          if (supportsMv) {

            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     nestString(BasicSearch.N_like,
                                                prop(pMvString) + literal("%bc%%")),
                                     null, 0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, empty);
          }
          break;

        case 15:
          if (supportsMv) {

            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprNot(
                                         nestString(BasicSearch.N_SAP_someEq,
                                                    prop(pMvString) + literal("abcde"))),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/1/mv-string-empty");
          }
          break;

        case 16:
          if (supportsLargeString) {

            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprEq(pLongString, longStringVal),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/2/long-string-1");
          }
          break;

        case 17:
          if (supportsLargeString) {

            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     nestString(BasicSearch.N_gte,
                                                prop(pLongString) + literal("1234")),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, new String[]{
              getLocalpart() + "/dasl/2/long-string-1",
              getLocalpart() + "/dasl/2/long-string-2"
            });
          }
          break;

        case 18:
          if (supportsDesc) {

            String[] expect = new String[]{
              getLocalpart() + "/dasl/3/sapdesc2",
            };

            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprEq(pDesc, "RISK"),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, expect);

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprEqIgnoreCase(pDesc, "RISK"),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, expect);

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprEq(pDesc, "risk"),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, empty);

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprEqIgnoreCase(pDesc, "risk"),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, expect);
          }
          break;

        case 19:
          if (supportsDesc) {

            String[] expectBoth = new String[]{
              getLocalpart() + "/dasl/3/sapdesc1",
              getLocalpart() + "/dasl/3/sapdesc2",
            };

            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprLike(pDesc, "%is%"),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/3/sapdesc1");

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprLikeIgnoreCase(pDesc, "%is%"),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, expectBoth);

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprLike(pDesc, "%IS%"),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/3/sapdesc2");

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprLikeIgnoreCase(pDesc, "%IS%"),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, expectBoth);
          }
          break;

        case 20:
          if (supportsDesc) {

            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprGreaterEq(pDesc, "T"),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/3/sapdesc1");
          }
          break;

        case 21:
          if (supportsDead) {

            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprIsDefinedOldSyntax(pString),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/4/string");

            body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprIsDefined(pString),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/4/string");
            
            body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprNot(exprIsDefined(pString)),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            List hrefs = getHrefsFromMultistatus(response.getDocument());
            assertTrue("should return all resources in scope but one, but didn't", hrefs.size() == totalResources - 1);
          }
          break;

        case 22:
          if (supportsDead) {

            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprEq(pString, "property _%"),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/4/string");

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprEq(pString, "pRoPeRty _%"),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, empty);

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprEqIgnoreCase(pString, "pRoPeRty _%"),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/4/string");
          }
          break;

        case 23:
          if (supportsDead) {

            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprLike(pString, "p_ope_ty \\_\\%"),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/4/string");

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprLikeIgnoreCase(pString, "p_ope_ty \\_\\%"),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/4/string");

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprLikeIgnoreCase(pString, "%er%"),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/4/string");

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprLikeIgnoreCase(pString, "_er%"),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, empty);

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprLike(pString, "%ER%"),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, empty);

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprLike(pString, "%er%"),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/4/string");

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprLikeIgnoreCase(pString, "%ER%"),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/4/string");

            // malformed like clause: "\" must escape "\", "_" or "%"
            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprLikeIgnoreCase(pString, "\\ "),
                              null,
                              0);
            clear(request);
            request.setMethod("SEARCH");
            request.setURI(getLocalpart());
            request.setBody(body);
            response = requester.perform(request);
            checkResponseCode(response, 422);
          }
          break;

        case 24:
          if (supportsDead) {

            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprEq(pLong, "-123456789"),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/4/long");

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprNot(exprEq(pLong, "-123456789")),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, empty);
          }
          break;

        case 25:
          if (supportsTypedDead) {

            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprLessEq(pLong, "-123456788"),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/4/long");
          }
          break;

        case 26:
          if (supportsDead) {

            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprEq(pDate, "2002-12-17T12:00:00Z"),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/4/date");

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprNot(exprEq(pDate, "2002-12-17T12:00:00Z")),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, empty);
          }
          break;

        case 27:
          if (supportsTypedDead) {

            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprLessEq(pDate, "2002-12-17T13:00:00Z"),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/4/date");
          }
          break;

        case 28:
          if (supportsDead) {

            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprGreaterEq(pMixed, "1"),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, new String[]{
              getLocalpart() + "/dasl/5/mixed-integer",
              getLocalpart() + "/dasl/5/mixed-string1",
              getLocalpart() + "/dasl/5/mixed-string2",
            });
          }
          break;

        case 29:
          {

            // old notation
            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprEq(BasicSearch.N_iscollection, "1"),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, new String[]{
              getLocalpart() + "/dasl/",
              getLocalpart() + "/dasl/1/",
              getLocalpart() + "/dasl/2/",
              getLocalpart() + "/dasl/3/",
              getLocalpart() + "/dasl/4/",
              getLocalpart() + "/dasl/5/",
              getLocalpart() + "/dasl/6/"
            });

            // old notation
            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprEq(BasicSearch.N_iscollection, "0"),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            List resultlist = getHrefsFromMultistatus(response.getDocument());
            for (Iterator it = resultlist.iterator(); it.hasNext(); /**/) {
              String uri = it.next().toString();
              assertTrue("result set contains collection: " + uri,
                         !uri.endsWith("/"));
            }

            // old notation with invalid operand
            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprEq(BasicSearch.N_iscollection, "2"),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, empty);
          }
          break;

        case 30:
          {
            // new notation
            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     "<is-collection xmlns='DAV:'/>",
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, new String[]{
              getLocalpart() + "/dasl/",
              getLocalpart() + "/dasl/1/",
              getLocalpart() + "/dasl/2/",
              getLocalpart() + "/dasl/3/",
              getLocalpart() + "/dasl/4/",
              getLocalpart() + "/dasl/5/",
              getLocalpart() + "/dasl/6/"
            });

            // new notation
            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprNot("<is-collection xmlns='DAV:'/>"),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            List resultlist = getHrefsFromMultistatus(response.getDocument());
            for (Iterator it = resultlist.iterator(); it.hasNext(); /**/) {
              String uri = it.next().toString();
              assertTrue("result set contains collection: " + uri,
                         !uri.endsWith("/"));
            }
          }
          break;

        case 31:
          {
            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprEq(Consts.N_getcontenttype, "application/msword"),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/1/word.doc");

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprEq(Consts.N_getcontenttype, "application/MSWord"),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, empty);

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprEqIgnoreCase(Consts.N_getcontenttype, "application/MSWord"),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/1/word.doc");
          }
          break;

        case 32:
          {
            String body = buildBasic("<getcontenttype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprLike(Consts.N_getcontenttype, "application/ms____"),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/1/word.doc");
          }
          break;

        case 33:
          {
            String body = buildBasic("<getcontenttype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprAnd(
                                         exprGreaterEq(Consts.N_getcontenttype, "a"),
                                         exprLessEq(Consts.N_getcontenttype, "b"),
                                         exprNot(exprLike(Consts.N_getcontenttype, "%/octet-stream%"))
                                     ),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, new String[]{getLocalpart() + "/dasl/1/word.doc",
                                                  getLocalpart() + "/dasl/1/excel.xls"});
          }
          break;

        case 34:
          if (persistsEncodingWithType) {
            String[] single = new String[]{
              getLocalpart() + "/dasl/1/plain-u.txt",
            };
            String[] both = new String[]{
              getLocalpart() + "/dasl/1/plain.txt",
              getLocalpart() + "/dasl/1/plain-u.txt",
            };

            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprLike(Consts.N_getcontenttype, "text/plain%"),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, both);

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprLike(Consts.N_getcontenttype, "text/plain%;%"),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, single);

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprLikeIgnoreCase(Consts.N_getcontenttype, "%;%charset=UTF-8%"),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, single);
          }
          break;

        case 35:
          {
            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprEq(Consts.N_getcontentlength, "8323"),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/1/plain.txt");

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprGreaterEq(Consts.N_getcontentlength, "8324"),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, new String[]{
              getLocalpart() + "/dasl/1/plain-u.txt",
              getLocalpart() + "/dasl/1/word.doc"
            });

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprAnd(
                                  exprLessEq(Consts.N_getcontentlength, "10000"),
                                  exprGreaterEq(Consts.N_getcontentlength, "1")
                              ),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, new String[]{
              getLocalpart() + "/dasl/1/plain-u.txt",
              getLocalpart() + "/dasl/1/plain.txt",
              getLocalpart() + "/dasl/1/no-ext"
            });
          }
          break;

        case 36:
          {
            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprGreaterEq(Consts.N_getcontentlength, "1"),
                                     "<order><prop><getcontentlength/></prop><descending/></order>",
                                     4);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkSortedResultSet(response, new String[]{
              getLocalpart() + "/dasl/1/word.doc",
              getLocalpart() + "/dasl/1/plain-u.txt",
              getLocalpart() + "/dasl/1/plain.txt",
              getLocalpart() + "/dasl/1/no-ext"
            });

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprGreaterEq(Consts.N_getcontentlength, "1"),
                              "<order><prop><getcontentlength/></prop><ascending/></order>",
                              3);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkSortedResultSet(response, new String[]{
              getLocalpart() + "/dasl/1/no-ext",
              getLocalpart() + "/dasl/1/plain.txt",
              getLocalpart() + "/dasl/1/plain-u.txt"
            });
          }
          break;

        case 37:
          {
            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprGreaterEq(Consts.N_getlastmodified, yesterday()),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            List hrefs = getHrefsFromMultistatus(response.getDocument());
            assertTrue("expected "+totalResources+" resources, got: "+hrefs.size(), hrefs.size() == totalResources);

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprLessEq(Consts.N_getlastmodified, yesterday()),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, empty);
          }
          break;

        case 38:
          if (supportsCreationDate) {
            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprGreaterEq(Consts.N_creationdate, yesterdayISO()),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            List hrefs = getHrefsFromMultistatus(response.getDocument());
            assertTrue("expected "+totalResources+" resources, got: "+hrefs.size(), hrefs.size() == totalResources);

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprLessEq(Consts.N_creationdate, yesterday()),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, empty);
          }
          break;

        case 39:
          {
            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprNot(exprEq(Consts.N_getcontenttype, "application/msword")),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);

            checkResultSet(response, new String[]{
              getLocalpart() + "/dasl/1/excel.xls",
              getLocalpart() + "/dasl/1/plain.txt",
              getLocalpart() + "/dasl/1/plain-u.txt",
              getLocalpart() + "/dasl/1/no-ext",
              getLocalpart() + "/dasl/1/mv-long",
              getLocalpart() + "/dasl/1/mv-string",
              getLocalpart() + "/dasl/1/mv-string-empty",
              getLocalpart() + "/dasl/2/long-string-1",
              getLocalpart() + "/dasl/2/long-string-2",
              getLocalpart() + "/dasl/3/sapdesc1",
              getLocalpart() + "/dasl/3/sapdesc2",
              getLocalpart() + "/dasl/4/string",
              getLocalpart() + "/dasl/4/date",
              getLocalpart() + "/dasl/4/long",
              getLocalpart() + "/dasl/4/boolean",
              getLocalpart() + "/dasl/4/xml",
              getLocalpart() + "/dasl/5/mixed-integer",
              getLocalpart() + "/dasl/5/mixed-string1",
              getLocalpart() + "/dasl/5/mixed-string2",
              getLocalpart() + "/dasl/6/versioned"
            });
          }
          break;

        case 40:
          {
            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprOr(exprEq(Consts.N_getcontenttype, "application/msword"),
                                            exprEq(Consts.N_getcontenttype, "application/vnd.ms-excel")),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, new String[]{getLocalpart() + "/dasl/1/word.doc",
                                                  getLocalpart() + "/dasl/1/excel.xls"});
          }
          break;

        case 41:
          {
            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprAnd(exprEq(Consts.N_getcontenttype, "application/msword"),
                                             exprEq(Consts.N_getcontenttype, "application/vnd.ms-excel")),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, empty);

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprEq(Consts.N_getcontenttype, "wrxwrx"),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, empty);
          }
          break;

        case 42: {
            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprGreaterEq(Consts.N_getlastmodified, "abc"),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResponseCode(response, 207); // any non error is ok
          }
          break;

        case 43: {
            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprEq(Consts.N_getcontenttype, "'\"\u00c4\u00d6\u00dc\u00e4\u00f6\u00fc-\u00df\u20ac"),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, empty);

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprEq(new Name("foobar", "dead"), "'\"\u00c4\u00d6\u00dc\u00e4\u00f6\u00fc-\u00df\u20ac"),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, empty);
          }
          break;

        case 44: {
            Element lm = getProperty(getLocalpart() + "/dasl/1/plain.txt", Consts.N_getlastmodified);
            assertTrue("DAV:getlastmodified present", lm != null);
            String lastmodOfPlainText = DomElement.getText(lm);

            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl/1/plain.txt",
                                     Integer.MAX_VALUE,
                                     exprGreaterEq(Consts.N_getlastmodified, lastmodOfPlainText),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() +  "/dasl/1/plain.txt");
          }
          break;

        case 45: {
            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     null,
                                     null,
                                     0);
            response = executeSearch(getLocalpart(), body);
            List hrefs = getHrefsFromMultistatus(response.getDocument());
            assertTrue(hrefs.size() == totalResources);
          }
          break;

        case 46: {
            String body = buildBasic("<resourcetype/>",
                                     "DAV:invalidscope",
                                     Integer.MAX_VALUE,
                                     null,
                                     null,
                                     0);
            clear(request);
            request.setMethod("SEARCH");
            request.setURI(getLocalpart());
            request.setBody(body);
            response = requester.perform(request);
            checkResponseCode(response, 403);
            checkErrorElement(response, DavConditions.N_search_scope_valid);
          }
          break;

        case 47:
          if (supportsVersioning) {
            String body = buildBasic("<version-name/>",
                                     getLocalpart() + "/dasl",
                                     DEEP_INCLUDING_VERSIONS,
                                     exprIsDefined(Consts.N_version_name),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            List hrefs = getHrefsFromMultistatus(response.getDocument());
            assertTrue("expected "+1+" resources, got: "+hrefs.size(), hrefs.size() == 1);
          }
          break;

        case 48:
          if (supportsVersioning) {
            String body = buildBasic(pVersString.getEmptyTag(),
                                     getLocalpart() + "/dasl",
                                     DEEP_INCLUDING_VERSIONS,
                                     exprIsDefined(pVersString),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            List hrefs = getHrefsFromMultistatus(response.getDocument());
            assertTrue("expected "+2+" resources, got: "+hrefs.size(), hrefs.size() == 2);
          }
          break;

        case 49:
          if (supportsDead) {
            DavStatus stat = setProperty(getLocalpart() + "/dasl/5/mixed-string1", pString, "Test", false);
            assertTrue(stat.getStatus() == 200);
            stat = setProperty(getLocalpart() + "/dasl/5/mixed-string2", pString, "Test ", false);
            assertTrue(stat.getStatus() == 200);

            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprEq(pString, "Test"),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/5/mixed-string1");

            body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprEq(pString, "Test "),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/5/mixed-string2");
          }
          break;

        case 50:
          if (supportsDesc) {
            DavStatus stat = setProperty(getLocalpart() + "/dasl/5/mixed-string1", pDesc, "bar", false);
            assertTrue(stat.getStatus() == 200);
            stat = setProperty(getLocalpart() + "/dasl/5/mixed-string2", pDesc, "bar ", false);
            assertTrue(stat.getStatus() == 200);

            String body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprEq(pDesc, "bar"),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/5/mixed-string1");

            body = buildBasic("<resourcetype/>",
                                     getLocalpart() + "/dasl",
                                     Integer.MAX_VALUE,
                                     exprEq(pDesc, "bar "),
                                     null,
                                     0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/5/mixed-string2");

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprNot(exprIsDefined(pDesc)),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            List hrefs = getHrefsFromMultistatus(response.getDocument());
            assertTrue("expected " + (totalResources - 4) + " resources, got: " + hrefs.size(), hrefs.size() == (totalResources - 4));

            body = buildBasic("<resourcetype/>",
                              getLocalpart() + "/dasl",
                              Integer.MAX_VALUE,
                              exprIsDefined(pDesc),
                              null,
                              0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            hrefs = getHrefsFromMultistatus(response.getDocument());
            assertTrue("expected "+4+" resources, got: "+hrefs.size(), hrefs.size() == 4);
          }
          break;

        case 51:
        {
          String body = buildBasic("<resourcetype/>",
          getLocalpart() + "/..",
          Integer.MAX_VALUE,
          //don't change the propertyname here.
          //this is used to perform a search with
          //the builtin PropertySearchManager of the
          //RM.
          exprEq(new Name("http://sapportals.com/xmlns/cm", "contenttype"), "application/msword"), 
            null, 0);
          response = executeSearch(getLocalpart() + "/..", body);
          List hrefs = getHrefsFromMultistatus(response.getDocument());
          assertTrue("expected at least "+2+" resources, got: "+hrefs, hrefs.size() >= 2);
          Iterator it = hrefs.iterator();
          String lp = getLocalpart();
          String root = lp.substring(0, lp.lastIndexOf("/"));
          while(it.hasNext()) {
            String s = (String)it.next();
            if (!s.startsWith(lp)) {
              //found entries belonging to other parent folder
              if (!s.startsWith(root)) fail(s + " doesn't belong to repository " + root);
            }
          }
        }
        break;
        
      case 52: if (supportsTypedDead) {

          String body = buildBasic("<resourcetype/>", getLocalpart() + "/dasl", Integer.MAX_VALUE, exprEq(pBoolean, "1"), null, 0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/4/boolean");

// TODO: re-add
//            body = buildBasic("<resourcetype/>", getLocalpart() + "/dasl", Integer.MAX_VALUE, exprEq(pBoolean, "true"), null, 0);
//            response = executeSearch(getLocalpart() + "/dasl", body);
//            checkResultSet(response, getLocalpart() + "/dasl/4/boolean");
          } 
        break;
      
      case 53: if (supportsDead) {
            String body = buildBasic("<resourcetype/>", getLocalpart() + "/dasl", Integer.MAX_VALUE, exprIsDefined(pXml), null, 0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/4/xml");

            body = buildBasic("<resourcetype/>", getLocalpart() + "/dasl", Integer.MAX_VALUE, exprNot(exprIsDefined(pXml)),
                null, 0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            List hrefs = getHrefsFromMultistatus(response.getDocument());
            assertTrue("should return all resources in scope but one, but didn't", hrefs.size() == totalResources - 1);
          }
          break;

      case 54:
          if (supportsDead) {
            String body = buildBasic("<resourcetype/>", getLocalpart() + "/dasl", Integer.MAX_VALUE, exprLike(pXml, "%oo%"), null, 0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            List hrefs = getHrefsFromMultistatus(response.getDocument());
            assertTrue("expected 0 resources, got: " + hrefs.size(), hrefs.size() == 0);
          }
          break;

      case 55:
          if (supportsDead && supportsMv) {
            String body = buildBasic("<resourcetype/>", getLocalpart() + "/dasl", Integer.MAX_VALUE, exprIsDefined(pMvString), null, 0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            checkResultSet(response, getLocalpart() + "/dasl/1/mv-string", getLocalpart() + "/dasl/1/mv-string-empty");

            body = buildBasic("<resourcetype/>", getLocalpart() + "/dasl", Integer.MAX_VALUE, exprNot(exprIsDefined(pMvString)), null, 0);
            response = executeSearch(getLocalpart() + "/dasl", body);
            List hrefs = getHrefsFromMultistatus(response.getDocument());
            assertTrue("should return all resources in scope but one, but didn't", hrefs.size() == totalResources - 2);
          }
          break;

      case 56: {
          String body = buildBasic("<resourcetype/>", getLocalpart() + "/dasl", Integer.MAX_VALUE, exprAnd(exprIsCollection(), exprNot(exprIsCollection())), null, 0);
          response = executeSearch(getLocalpart() + "/dasl", body);
          List hrefs = getHrefsFromMultistatus(response.getDocument());
          assertTrue("should return no resources", hrefs.size() == 0);

          body = buildBasic("<resourcetype/>", getLocalpart() + "/dasl", Integer.MAX_VALUE, exprOr(exprIsCollection(), exprNot(exprIsCollection())), null, 0);
          response = executeSearch(getLocalpart() + "/dasl", body);
          hrefs = getHrefsFromMultistatus(response.getDocument());
          assertTrue("should return all resources", hrefs.size() == totalResources);
        }
        break;

      case 57:
        if (supportsTypedDead) {
          String body = buildBasic("<resourcetype/>", getLocalpart() + "/dasl", Integer.MAX_VALUE, exprOr(exprNot(exprIsDefined(pBoolean)), exprEq(pBoolean, "1")), null, 0);
          response = executeSearch(getLocalpart() + "/dasl", body);
          List hrefs = getHrefsFromMultistatus(response.getDocument());
          assertTrue("should return all " + totalResources + " resources, but returned " + hrefs.size(), hrefs.size() == totalResources);

          body = buildBasic("<resourcetype/>", getLocalpart() + "/dasl", Integer.MAX_VALUE, exprOr(exprNot(exprIsDefined(pBoolean)), exprEq(pBoolean, "0")), null, 0);
          response = executeSearch(getLocalpart() + "/dasl", body);
          hrefs = getHrefsFromMultistatus(response.getDocument());
          assertTrue("should return all resources but one", hrefs.size() == totalResources - 1);

          body = buildBasic("<resourcetype/>", getLocalpart() + "/dasl", Integer.MAX_VALUE, exprEq(pBoolean, "1"), null, 0);
          response = executeSearch(getLocalpart() + "/dasl", body);
          checkResultSet(response, getLocalpart() + "/dasl/4/boolean");

          body = buildBasic("<resourcetype/>", getLocalpart() + "/dasl", Integer.MAX_VALUE, exprNot(exprEq(pBoolean, "1")), null, 0);
          response = executeSearch(getLocalpart() + "/dasl", body);
          hrefs = getHrefsFromMultistatus(response.getDocument());
          assertTrue("should return empty result", hrefs.size() == 0);
        }
        break;

      default: {
          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/dasl");
          response = requester.perform(request);
          checkResponseCode(response, 204);
        }
        break;
      }
    }
    catch (Exception ex) {
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }

  private static String collections[] = {
    "/dasl/",
    "/dasl/1/", "/dasl/2/", "/dasl/3/", "/dasl/4/", "/dasl/5/", "/dasl/6/"
  };

  private void putBody(String name, String type, int length) throws WcmException {
    clear(request);
    request.setMethod("PUT");
    request.setURI(getLocalpart() + name);
    request.setBody(new ByteArrayInputStream(new byte[length]));
    if (type != null) request.setHeader("content-type", type);
    IResponse response = requester.perform(request);
    checkResponseCode(response, 201);
  }

  private boolean versionControl(String name) throws WcmException {
    clear(request);
    request.setMethod("VERSION-CONTROL");
    request.setURI(getLocalpart() + name);
    IResponse response = requester.perform(request);
    return response.getStatus() == 200 || response.getStatus() == 201;
  }

  private static String buildBasic(String select, String href, int depth,
                                   String where, String orderby, int limit) {
    StringBuffer sb = new StringBuffer();
    String depthString = (depth == 1? "1" : (depth == 0? "0" : "infinity"));

    sb.append("<searchrequest xmlns='DAV:'>");
    sb.append("  <basicsearch>");
    sb.append("    <select>");
    if (select != null) {
      sb.append("      <prop>");
      sb.append("        ").append(select);
      sb.append("      </prop>");
    }
    else {
      sb.append("<allprop/>");
    }
    sb.append("    </select>");
    sb.append("    <from>");
    sb.append("      <scope>");
    sb.append("        <href>").append(href).append("</href>");
    sb.append("        <depth>").append(depthString).append("</depth>");
    if (depth == DEEP_INCLUDING_VERSIONS) sb.append("<include-versions/>");
    sb.append("      </scope>");
    sb.append("    </from>");

    if (where != null) {
      sb.append("    <where>");
      sb.append("      ").append(where);
      sb.append("    </where>");
    }

    if (orderby != null) {
      sb.append("    <orderby>");
      sb.append("      ").append(orderby);
      sb.append("    </orderby>");
    }

    if (limit > 0) {
      sb.append("    <limit>");
      sb.append("      <nresults>").append(limit).append("</nresults>");
      sb.append("    </limit>");
    }

    sb.append("  </basicsearch>");
    sb.append("</searchrequest>");

    return sb.toString();
  }

  private String nestString(Name name) {
    return name.getEmptyTag();
  }

  private String nestString(Name name, String contents) {
    return name.getStartTag() + contents + name.getEndTag();
  }

  private String exprAnd(String cont1, String cont2) {
    return nestString(BasicSearch.N_and, cont1 + cont2);
  }

  private String exprAnd(String cont1, String cont2, String cont3) {
    return nestString(BasicSearch.N_and, cont1 + cont2 + cont3);
  }

  private String exprEq(Name propname, String lit) {
    return nestString(BasicSearch.N_eq, prop(propname) + literal(lit));
  }

  private String exprGreaterEq(Name propname, String lit) {
    return nestString(BasicSearch.N_gte, prop(propname) + literal(lit));
  }

  private String exprEqIgnoreCase(Name propname, String lit) {
    return "<eq xmlns='DAV:' casesensitive='0'>" + prop(propname) + literal(lit) + "</eq>";
  }

  private String exprIsCollection() {
    return nestString(BasicSearch.N_is_collection);
  }

  private String exprIsDefined(Name propname) {
    return nestString(BasicSearch.N_is_defined, prop(propname));
  }

  private String exprIsDefinedOldSyntax(Name propname) {
    return nestString(BasicSearch.N_isdefined, prop(propname));
  }

  private String exprLessEq(Name propname, String lit) {
    return nestString(BasicSearch.N_lte, prop(propname) + literal(lit));
  }

  private String exprLike(Name propname, String lit) {
    return nestString(BasicSearch.N_like, prop(propname) + literal(lit));
  }

  private String exprLikeIgnoreCase(Name propname, String lit) {
    return "<like xmlns='DAV:' caseless='yes'>" + prop(propname) + literal(lit) + "</like>";
  }

  private String exprNot(String cont) {
    return nestString(BasicSearch.N_not, cont);
  }

  private String exprOr(String cont1, String cont2) {
    return nestString(BasicSearch.N_or, cont1 + cont2);
  }

  private String literal(String lit) {
    return nestString(BasicSearch.N_literal, lit);
  }

  private String prop(Name propname) {
    return nestString(Consts.N_prop, propname.getEmptyTag());
  }



  private IResponse executeSearch(String requesturi, String body) throws Exception {

    clear(request);
    request.setMethod("SEARCH");
    request.setURI(requesturi);
    request.setBody(body);
    IResponse response = requester.perform(request);
    checkResponseCode(response, 207);
    return response;
  }

  private void checkResultSet(IResponse resp, String uri) throws Exception {
    String[] uris = new String[1];
    uris[0] = uri;
    checkResultSet(resp, uris);
  }

  private void checkResultSet(IResponse resp, String uri1, String uri2) throws Exception {
    String[] uris = new String[2];
    uris[0] = uri1;
    uris[1] = uri2;
    checkResultSet(resp, uris);
  }

  private static String yesterday() {
    return DF.formatHTTP(new Date(System.currentTimeMillis() - (1000 * 60 * 60 * 24)));
  }

  private static String yesterdayISO() {
    return DF.formatISO8601(new Date(System.currentTimeMillis() - (1000 * 60 * 60 * 24)));
  }

}
