package com.sapportals.wcm.protocol.webdav.server.tests;

import java.util.List;
import java.util.Set;

import org.w3c.dom.Element;

import com.sapportals.wcm.protocol.webdav.Consts;
import com.sapportals.wcm.protocol.webdav.DavConditions;
import com.sapportals.wcm.protocol.webdav.DomElement;
import com.sapportals.wcm.protocol.webdav.dompath.DomPath;
import com.sapportals.wcm.util.http.IResponse;

/**
 * Various tests for WEBDAV <a href="http://greenbytes.de/tech/webdav/rfc2518.html#METHOD_LOCK"
 * >LOCK</a>  method (special case for deep locks).
 *
 * <p>
 * Copyright (c) SAP AG 2002-2004
 * @author julian.reschke@greenbytes.de
 * @version $Id: DeepLockTest.java,v 1.5 2004/07/22 16:16:01 jre Exp $
 */

public class DeepLockTest extends TestMember {

  private static final String[] desc = new String[]
  {
    "00 DL Set up directory and test resource",
    "01 DL make sure deep locks on collections are supported",
    "02 DL attempt a deep lock on a collection with locked member",
    "03 DL add newly created resource to lock and use it for unlock",
    "04 DL move child out of lock scope",
    "05 DL move child within lock scope",
    "06 DL attempt lock on child in deep-locked collection",
    "07 DL move collection",
    "08 DL create shared lock",
    "09 DL try to create exclusive locks",
    "10 DL create additional shared locks",
    "   DL cleanup",
  };

  public static int NrOfTests = desc.length;

  public DeepLockTest(Integer nr) {
    super(nr, desc[nr.intValue()]);
  }

  private static String lockTokenColl = null, lockTokenColl2 = null;
  private static String lockTokenFile = null;

  private static boolean deepLockSupported = true;
  private static boolean returnsLockRoot = false;

  protected void runTest() {
    if (nr == 0) logIdentifier("DL");

    String respfile = "DEEPLOCK-" + TestMember.THREEDIGITS.format(nr) + ".txt";

    String container = "/deep-locks";
    String container2 = "/deep-locks2";
    String testfile = container + "/testfile";
    String testfile2 = container2 + "/testfile";

    try {
      IResponse response;
      clear(request);

      switch (nr) {

        case 0:
          clear(request);
          request.setMethod("MKCOL");
          request.setURI(getLocalpart() + container);
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + testfile);
          request.setBody("foo");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;

        case 1:
          clear(request);
          request.setMethod("LOCK");
          request.setURI(getLocalpart() + container);
          request.setHeader("Depth", "infinity");
          request.setBody(makeLockInfo(null, false));
          request.expectsResponseDocument(true);
          response = requester.perform(request);
          if (response.getStatus() >= 400) {
            deepLockSupported = false;
          }
          else {
            checkResponseCode(response, 200);
            checkResponseDocument(response, respfile, null);
            lockTokenColl = response.getHeader("lock-token");
            assertTrue(lockTokenColl != null);

            DomPath dp = new DomPath("/D:prop/D:lockdiscovery/D:activelock/D:lockroot", DomPath.DEFAULT_NAMESPACE_BINDINGS);
            List lrs = dp.matchOrDie(response.getDocument(), 0, 1, "must have at most one D:lockroot element");

            if (lrs.size() == 1) {
              Element lr = (Element) lrs.get(0);
              Set hrefs = DomElement.getHrefs(lr);
              assertTrue(hrefs.size() == 1);
              String s = (String) hrefs.iterator().next();
              if (s.endsWith("/")) s = s.substring(0, s.length()-1);
              assertTrue(s.endsWith(container));
              returnsLockRoot = true;
            }

            clear(request);
            request.setMethod("UNLOCK");
            request.setURI(getLocalpart() + container);
            request.setHeader("lock-token", lockTokenColl);
            response = requester.perform(request);
            checkResponseCode(response, 204);
            lockTokenColl = null;
          }
          logComment(deepLockSupported ? "deep-lock:yes" : "deep-lock:no");
          if (deepLockSupported) logComment(returnsLockRoot ? "lr:yes" : "lr:no");
          break;

        case 2:
          if (deepLockSupported) {
            clear(request);
            request.setMethod("LOCK");
            request.setURI(getLocalpart() + testfile);
            request.setHeader("Depth", "0");
            request.setBody(makeLockInfo(null, false));
            response = requester.perform(request);
            checkResponseCode(response, 200);
            lockTokenFile = response.getHeader("lock-token");
            assertTrue(lockTokenFile != null);

            clear(request);
            request.setMethod("LOCK");
            request.setURI(getLocalpart() + container);
            request.setHeader("Depth", "infinity");
            request.setBody(makeLockInfo(null, false));
            response = requester.perform(request);
            checkResponseCode(response, 207);
            checkResponseDocument(response, respfile, null);

            clear(request);
            request.setMethod("UNLOCK");
            request.setURI(getLocalpart() + testfile);
            request.setHeader("lock-token", lockTokenFile);
            response = requester.perform(request);
            checkResponseCode(response, 204);
            lockTokenFile = null;
          }
          break;

        case 3:
          if (deepLockSupported) {
            clear(request);
            request.setMethod("LOCK");
            request.setURI(getLocalpart() + container);
            request.setHeader("Depth", "infinity");
            request.setBody(makeLockInfo(null, false));
            response = requester.perform(request);
            checkResponseCode(response, 200);
            lockTokenColl = response.getHeader("lock-token");
            assertTrue(lockTokenColl != null);

            // PUT into collection without lock token in if header
            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + testfile2);
            request.setBody("foobar2");
            response = requester.perform(request);
            assertTrue(response.getStatus() != 201);

            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + testfile + "2");
            request.setHeader("If", "(" + lockTokenColl + ")");
            request.setBody("foobar2");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            Element activelock = getProperty(getLocalpart() + testfile + "2", Consts.N_activelock);
            assertTrue(activelock == null || activelock.hasChildNodes() == false);

            clear(request);
            request.setMethod("UNLOCK");
            request.setURI(getLocalpart() + testfile + "2");
            request.setHeader("lock-token", lockTokenColl);
            response = requester.perform(request);
            checkResponseCode(response, 204);
            lockTokenColl = null;

            activelock = getProperty(getLocalpart() + container, Consts.N_activelock);
            assertTrue(activelock == null || activelock.hasChildNodes() == false);
          }
          break;

        case 4:
          if (deepLockSupported) {
            clear(request);
            request.setMethod("LOCK");
            request.setURI(getLocalpart() + container);
            request.setHeader("Depth", "infinity");
            request.setBody(makeLockInfo(null, false));
            response = requester.perform(request);
            checkResponseCode(response, 200);
            lockTokenColl = response.getHeader("lock-token");
            assertTrue(lockTokenColl != null);

            // MOVE out of collection without lock token in if header
            clear(request);
            request.setMethod("MOVE");
            request.setURI(getLocalpart() + testfile);
            request.setHeader("destination", getLocalpart() + "/dlfile");
            response = requester.perform(request);
            assertTrue(response.getStatus() >= 400);

            clear(request);
            request.setMethod("MOVE");
            request.setURI(getLocalpart() + testfile);
            request.setHeader("destination", getLocalpart() + "/dlfile");
            request.setHeader("If", "<" + getBase() + getLocalpart() + testfile + ">" + "(" + lockTokenColl + ")"
                                    + " <" + getBase() + getLocalpart() + container + ">" + "(" + lockTokenColl + ")");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            Element activelock = getProperty(getLocalpart() + "/dlfile", Consts.N_activelock);
            assertTrue(activelock == null || activelock.hasChildNodes() == false);

            clear(request);
            request.setMethod("UNLOCK");
            request.setURI(getLocalpart() + container);
            request.setHeader("lock-token", lockTokenColl);
            response = requester.perform(request);
            checkResponseCode(response, 204);
            lockTokenColl = null;

            activelock = getProperty(getLocalpart() + container, Consts.N_activelock);
            assertTrue(activelock == null || activelock.hasChildNodes() == false);
          }
          break;

        case 5:
          if (deepLockSupported) {
            if (lockTokenColl == null) {
              clear(request);
              request.setMethod("LOCK");
              request.setURI(getLocalpart() + container);
              request.setHeader("Depth", "infinity");
              request.setBody(makeLockInfo(null, false));
              response = requester.perform(request);
              checkResponseCode(response, 200);
              lockTokenColl = response.getHeader("lock-token");
              assertTrue(lockTokenColl != null);
            }

            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + testfile);
            request.setHeader("If", "(" + lockTokenColl + ")");
            request.setBody("foobar2");
            response = requester.perform(request);
            checkResponseCode(response, 200, 201);

            // MOVE within collection without lock token in if header
            clear(request);
            request.setMethod("MOVE");
            request.setURI(getLocalpart() + testfile);
            request.setHeader("destination", getLocalpart() + testfile + "2");
            response = requester.perform(request);
            assertTrue(response.getStatus() >= 400);

            // remove target if already present
            clear(request);
            request.setMethod("DELETE");
            request.setURI(getLocalpart() + testfile + "2");
            request.setHeader("If", "(" + lockTokenColl + ")");
            response = requester.perform(request);
            // don't care, just make sure it's gone

            clear(request);
            request.setMethod("MOVE");
            request.setURI(getLocalpart() + testfile);
            request.setHeader("destination", getLocalpart() + testfile + "2");
            request.setHeader("If", "<" + getBase() + getLocalpart() + testfile + ">" + "(" + lockTokenColl + ")"
                                    + " <" + getBase() + getLocalpart() + container + ">" + "(" + lockTokenColl + ")");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("MOVE");
            request.setURI(getLocalpart() + testfile + "2");
            request.setHeader("destination", getLocalpart() + testfile);
            request.setHeader("If", "<" + getBase() + getLocalpart() + testfile + "2" + ">" + "(" + lockTokenColl + ")"
                                    + " <" + getBase() + getLocalpart() + container + ">" + "(" + lockTokenColl + ")");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("UNLOCK");
            request.setURI(getLocalpart() + container);
            request.setHeader("lock-token", lockTokenColl);
            response = requester.perform(request);
            checkResponseCode(response, 204);
            lockTokenColl = null;

            Element activelock = getProperty(getLocalpart() + container, Consts.N_activelock);
            assertTrue(activelock == null || activelock.hasChildNodes() == false);
          }
          break;

        case 6:
          if (deepLockSupported) {
            if (lockTokenColl == null) {
              clear(request);
              request.setMethod("LOCK");
              request.setURI(getLocalpart() + container);
              request.setHeader("Depth", "infinity");
              request.setBody(makeLockInfo(null, false));
              response = requester.perform(request);
              checkResponseCode(response, 200);
              lockTokenColl = response.getHeader("lock-token");
              assertTrue(lockTokenColl != null);
            }

            clear(request);
            request.setMethod("LOCK");
            request.setURI(getLocalpart() + testfile);
            request.setBody(makeLockInfo(null, false));
            response = requester.perform(request);
            assertTrue(response.getStatus() != 200);

            clear(request);
            request.setMethod("UNLOCK");
            request.setURI(getLocalpart() + container);
            request.setHeader("lock-token", lockTokenColl);
            response = requester.perform(request);
            checkResponseCode(response, 204);
            lockTokenColl = null;

            Element activelock = getProperty(getLocalpart() + container, Consts.N_activelock);
            assertTrue(activelock == null || activelock.hasChildNodes() == false);
          }
          break;

        case 7:
          if (deepLockSupported) {
            if (lockTokenColl == null) {
              clear(request);
              request.setMethod("LOCK");
              request.setURI(getLocalpart() + container);
              request.setHeader("Depth", "infinity");
              request.setBody(makeLockInfo(null, false));
              response = requester.perform(request);
              checkResponseCode(response, 200);
              lockTokenColl = response.getHeader("lock-token");
              assertTrue(lockTokenColl != null);
            }

            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + testfile2);
            request.setBody("foobar2");
            response = requester.perform(request);
            assertTrue(response.getStatus() != 201);

            // remove target if already present
            clear(request);
            request.setMethod("DELETE");
            request.setURI(getLocalpart() + container2);
            response = requester.perform(request);
            // don't care, just make sure it's gone

            clear(request);
            request.setMethod("MOVE");
            request.setURI(getLocalpart() + container);
            request.setHeader("destination", getLocalpart() + container2);
            request.setHeader("If", "<" + getBase() + getLocalpart() + container + ">" + "(" + lockTokenColl + ")" +
                                    "<" + getBase() + getLocalpart() + testfile + ">" + "(" + lockTokenColl + ")");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + testfile2);
            request.setBody("foobar2");
            response = requester.perform(request);
            checkResponseCode(response, 200);

            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + testfile2 + ".new");
            request.setBody("foobar2");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            //move to old place
            clear(request);
            request.setMethod("MOVE");
            request.setURI(getLocalpart() + container2);
            request.setHeader("destination", getLocalpart() + container);
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + testfile);
            request.setBody("foobar2");
            response = requester.perform(request);
            assertTrue(response.getStatus() != 201);

            clear(request);
            request.setMethod("UNLOCK");
            request.setURI(getLocalpart() + container);
            request.setHeader("lock-token", lockTokenColl);
            response = requester.perform(request);
            checkResponseCode(response, 403);
            checkErrorElement(response, DavConditions.N_SAP_lock_token_matches);
            lockTokenColl = null;

            Element activelock = getProperty(getLocalpart() + container, Consts.N_activelock);
            assertTrue(activelock == null || activelock.hasChildNodes() == false);
          }
          break;

        case 8:
          if (deepLockSupported) {
            clear(request);
            request.setMethod("LOCK");
            request.setHeader("Timeout", "Second-60");
            request.setHeader("Depth", "infinity");
            request.setURI(getLocalpart() + container);
            request.setHeader("Content-Type", "text/xml");
            request.setBody(makeLockInfo(null, true));
            request.expectsResponseDocument(true);
            response = requester.perform(request);
            if (response.getStatus() == 200) {
              lockTokenColl = response.getHeader("lock-token");
              assertTrue("locktoken null", lockTokenColl != null);

              logComment("deep-shared-lock:yes");

              clear(request);
              request.setMethod("PROPFIND");
              request.setHeader("Depth", "1");
              request.setURI(getLocalpart() + container);
              request.setHeader("Content-Type", "text/xml");
              checkStatusAndBody(207, "<propfind xmlns='DAV:'><prop><lockdiscovery/></prop></propfind>", respfile);
            }
            else {
              logComment("deep-shared-lock:no");
            }

          }
          break;

        case 9:
          if (deepLockSupported && lockTokenColl != null) {
            clear(request);
            request.setMethod("LOCK");
            request.setURI(getLocalpart() + container);
            request.setHeader("Depth", "infinity");
            request.setBody(makeLockInfo(null, false));
            response = requester.perform(request);
            assertTrue("exclusive lock on collection not allowed", response.getStatus() >= 400);

            clear(request);
            request.setMethod("LOCK");
            request.setURI(getLocalpart() + testfile);
            request.setHeader("Depth", "0");
            request.setBody(makeLockInfo(null, false));
            response = requester.perform(request);
            assertTrue("exclusive lock on resource not allowed", response.getStatus() >= 400);
          }
          break;

        case 10:
          if (deepLockSupported && lockTokenColl != null) {
            clear(request);
            request.setMethod("LOCK");
            request.setHeader("Timeout", "Second-60");
            request.setHeader("Depth", "infinity");
            request.setURI(getLocalpart() + container);
            request.setHeader("Content-Type", "text/xml");
            request.setBody(makeLockInfo(null, true));
            request.expectsResponseDocument(true);
            response = requester.perform(request);
            if (response.getStatus() == 200) {
              lockTokenColl2 = response.getHeader("lock-token");
            }
            assertTrue("can't create another shared lock on collection", response.getStatus()==200);

            clear(request);
            request.setMethod("LOCK");
            request.setHeader("Timeout", "Second-60");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + testfile);
            request.setHeader("Content-Type", "text/xml");
            request.setBody(makeLockInfo(null, true));
            request.expectsResponseDocument(true);
            response = requester.perform(request);
            if (response.getStatus() == 200) {
              lockTokenFile = response.getHeader("lock-token");
            }
            assertTrue("can't create another shared lock on file", response.getStatus()==200);

          }
          break;

        default:
          if (lockTokenFile != null) {
            clear(request);
            request.setMethod("UNLOCK");
            request.setURI(getLocalpart() + testfile);
            request.setHeader("lock-token", lockTokenFile);
            response = requester.perform(request);
            checkResponseCode(response, 204, 404);
            lockTokenFile = null;
          }

          if (lockTokenColl != null) {
            clear(request);
            request.setMethod("UNLOCK");
            request.setHeader("lock-token", lockTokenColl);
            request.setURI(getLocalpart() + container);
            response = requester.perform(request);

            clear(request);
            request.setMethod("UNLOCK");
            request.setHeader("lock-token", lockTokenColl);
            request.setURI(getLocalpart() + container2);
            response = requester.perform(request);
          }

          if (lockTokenColl2 != null) {
            clear(request);
            request.setMethod("UNLOCK");
            request.setHeader("lock-token", lockTokenColl2);
            request.setURI(getLocalpart() + container);
            response = requester.perform(request);
          }

          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + container);
          response = requester.perform(request);
          checkResponseCode(response, 204, 404);

          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + container2);
          response = requester.perform(request);
          checkResponseCode(response, 204, 404);

          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/dlfile");
          response = requester.perform(request);
          checkResponseCode(response, 204, 404);
          break;
      }
    }
    catch (Exception ex) {
      ex.printStackTrace(System.err);
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }


  private String makeLockInfo(String owner, boolean shared) {
    String res = "<lockinfo xmlns='DAV:'>";

    if (owner != null) {
      res += "<owner>" + owner + "</owner>";
    }

    res += "<lockscope>" + (shared ? "<shared/>" : "<exclusive/>") + "</lockscope>";
    res += "<locktype><write /></locktype></lockinfo>";
    return res;
  }

}
