package com.sapportals.wcm.protocol.webdav.server.tests;

import java.io.InputStream;

import com.sapportals.wcm.util.http.IRequester;
import com.sapportals.wcm.util.http.IResponse;

/**
 * Various tests for DELETE method.
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author stefan.eissing@greenbytes.de
 * @author martin.boettcher@greenbytes.de
 * @author julian.reschke@greenbytes.de
 * @version $Id: DeleteTest.java,v 1.5 2004/07/23 15:42:07 sei Exp $
 */

public class DeleteTest extends TestMember
{
  private static final String[] desc = new String[] {
    "00 DEL setting up duplicate resources",
    "01 DEL DELETE readonly resource",
    "02 DEL DELETE unknown resource",
    "03 DEL DELETE root resource",
    "04 DEL DELETE resource in collection",
    "05 DEL DELETE deleted root resource",
    "06 DEL DELETE deleted resource in collection",
    "07 DEL DELETE empty collection",
    "08 DEL DELETE deleted empty collection",
    "09 DEL DELETE full collection",
    "10 DEL DELETE deleted full collection",
    "11 DEL OPTIONS repository root should not allow DELETE",
    "12 DEL DELETE full collection including one 'open' resource",
    "13 DEL visibility check for another user after DELETE",
    "14 DEL very deep delete",
    "15 DEL remove FS content, then delete",

    "   DEL cleanup",
  };

  public static int NrOfTests = desc.length;

  public DeleteTest(Integer nr)
  {
    super(nr, desc[nr.intValue()]);
  }

  private static InputStream str = null;
  private static boolean wasDeleted = false;

  protected void runTest()
  {
    if (nr == 0) super.logIdentifier("DEL");

    try {
      IResponse response;
      clear(request);
      request.setMethod("DELETE");

      switch (nr) {

        case 0:
          request.setMethod("COPY");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          request.setHeader("Destination", this.getBaseAndLocalpart() + "/ANSI3.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("COPY");
          request.setURI(getLocalpart() + "/Worddatei.doc");
          request.setHeader("Destination", this.getBaseAndLocalpart() + "/xxx.doc");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("COPY");
          request.setURI(getLocalpart() + "/voll");
          request.setHeader("Destination", this.getBaseAndLocalpart() + "/voll2");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("COPY");
          request.setURI(getLocalpart() + "/leer");
          request.setHeader("Destination", this.getBaseAndLocalpart() + "/leer2");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;

        case 1:
          request.setURI(getLocalpart() + "/xxx.doc");
          response = requester.perform(request);
          checkResponseCode(response, 204);
          break;

        case 2:
          request.setURI(getLocalpart() + "/xxx.doc");
          response = requester.perform(request);
          checkResponseCode(response, 404);
          break;

        case 3:
          request.setURI(getLocalpart() + "/ANSI3.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 204);
          break;

        case 4:
          request.setURI(getLocalpart() + "/voll2/ist%20was%20drin/office.zip");
          response = requester.perform(request);
          checkResponseCode(response, 204);
          break;

        case 5:
          request.setURI(getLocalpart() + "/ANSI3.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 404);
          break;

        case 6:
          request.setURI(getLocalpart() + "/voll2/ist%20was%20drin/office.zip");
          response = requester.perform(request);
          checkResponseCode(response, 404);
          break;

        case 7:
          request.setURI(getLocalpart() + "/leer2");
          response = requester.perform(request);
          checkResponseCode(response, 204);
          break;

        case 8:
          request.setURI(getLocalpart() + "/leer2");
          response = requester.perform(request);
          checkResponseCode(response, 404);
          break;

        case 9:
          request.setURI(getLocalpart() + "/voll2");
          response = requester.perform(request);
          checkResponseCode(response, 204);
          break;

        case 10:
          request.setURI(getLocalpart() + "/voll2");
          response = requester.perform(request);
          checkResponseCode(response, 404);
          break;

        case 11: {
          // make sure that OPTIONS doesn't advertise DELETE on the repository root
          int index = getLocalpart().lastIndexOf('/');
          assertTrue("strange uri: "+getLocalpart(),  index > 0);
          String repositoryRoot = getLocalpart().substring(0, index);
          request.setMethod("OPTIONS");
          request.setURI(repositoryRoot);
          response = requester.perform(request);
          checkResponseCode(response, 200);
          String allow = response.getHeader("Allow");
          assertTrue(allow != null);
          assertTrue(allow.indexOf("DELETE") < 0);
          break;
        }

        case 12:
          clear(request);
          request.setMethod("MKCOL");
          request.setURI(getLocalpart() + "/delete-test");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/delete-test/testresource");
          request.setBody(new java.io.BufferedInputStream(new DataGenerator(0, 1024 * 1024)));
          response = requester.perform(request);
          checkResponseCode(response, 201);

          try {
            clear(request);
            request.setMethod("GET");
            request.expectsResponseStream(true);
            request.setURI(getLocalpart() + "/delete-test/testresource");
            response = requester.perform(request);
            checkResponseCode(response, 200);
            str = response.getStream();

            IRequester requester2 = getSecondRequester();

            clear(request);
            request.setMethod("DELETE");
            request.setURI(getLocalpart() + "/delete-test");
            requester2.setSoTimeout(10000);
            response = requester2.perform(request);
            checkResponseCode(response, 204);
            wasDeleted = true;
          }
          finally {
            readEmptyAndClose(str);
            str = null;
          }
          break;

        case 13: {
          if (!wasDeleted) {
            clear(request);
            request.setMethod("DELETE");
            request.setURI(getLocalpart() + "/delete-test");
            response = requester.perform(request);
            checkResponseCode(response, 204);
          }

          clear(request);
          request.setMethod("MKCOL");
          request.setURI(getLocalpart() + "/delete-test");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          wasDeleted = false;

          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/delete-test/testresource");
          request.setBody(new java.io.BufferedInputStream(new DataGenerator(0, 1024)));
          response = requester.perform(request);
          checkResponseCode(response, 201);

          // Make sure that 2nd user has seen the resource
          clear(request);
          request.setMethod("HEAD");
          request.setURI(getLocalpart() + "/delete-test/testresource");
          response = this.getSecondRequester().perform(request);
          checkResponseCode(response, 200);

          // let 1st user remoove resource          
          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/delete-test/testresource");
          response = requester.perform(request);
          checkResponseCode(response, 204);

          // Check that 2nd user does not longer see the resource
          clear(request);
          request.setMethod("HEAD");
          request.setURI(getLocalpart() + "/delete-test/testresource");
          response = this.getSecondRequester().perform(request);
          checkResponseCode(response, 404);
          break;
        }
          
        case 14: {
          if (!wasDeleted) {
            clear(request);
            request.setMethod("DELETE");
            request.setURI(getLocalpart() + "/delete-test");
            response = requester.perform(request);
            checkResponseCode(response, 204);
          }

          clear(request);
          request.setMethod("MKCOL");
          request.setURI(getLocalpart() + "/delete-test");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          wasDeleted = false;
          
          String uri = getLocalpart() + "/delete-test";
          for (int k = 0; k < 25; k++) {
            uri += ("/" + k);

            clear(request);
            request.setMethod("MKCOL");
            request.setURI(uri);
            response = requester.perform(request);
            checkResponseCode(response, 201);
          }

          clear(request);
          request.setMethod("PUT");
          request.setURI(uri + "/testresource");
          request.setBody(new java.io.BufferedInputStream(new DataGenerator(0, 1024)));
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/delete-test");
          response = requester.perform(request);
          checkResponseCode(response, 204);
          wasDeleted = true;
          break;
        }


        case 15: { 
          if (FSInteractionTest.storesContentInFS()) {
            if (!wasDeleted) {
              clear(request);
              request.setMethod("DELETE");
              request.setURI(getLocalpart() + "/delete-test");
              response = requester.perform(request);
              checkResponseCode(response, 204);
            }

            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + "/delete-test");
            request.setBody("this is a test content to be removed");
            response = requester.perform(request);
            checkResponseCode(response, 201);
            wasDeleted = false;

            assertTrue(
              "unable to remove in file system: " + "delete-test", 
              FSInteractionTest.removeFSContent("delete-test"));

            clear(request);
            request.setMethod("DELETE");
            request.setURI(getLocalpart() + "/delete-test");
            response = requester.perform(request);
            checkResponseCode(response, 204, 404);
            wasDeleted = true;
          }
          break;
        }


        default:
          if (str != null) {
            str.close();
            Thread.sleep(1000); // give the server time to release the stream
          }

          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/delete-test");
          response = requester.perform(request);
          break;
      }
    }
    catch (Exception ex) {
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }

  private void readEmptyAndClose(InputStream in) {
    try { 
      byte[] buffer = new byte[16 * 1024];
      while (in.read(buffer) >= 0) {
        // read empty
      }
      in.close();
    }
    catch (Exception ex) {
      //ignore
    }
  }
  
  private class DataGenerator extends java.io.InputStream {

    private int pattern;
    private int length;
    private int position;

    public DataGenerator (int pattern, int length) {
      this.pattern = pattern;
      this.length = length;
      this.position = 0;
    }

    public int read() {
      if (this.position++ < this.length) {
        return pattern;
      }
      else {
        return -1;
      }
    }
  }

}
