package com.sapportals.wcm.protocol.webdav.server.tests;

import com.sapportals.wcm.WcmException;
import com.sapportals.wcm.protocol.webdav.Consts;
import com.sapportals.wcm.protocol.webdav.DomElement;
import com.sapportals.wcm.util.http.IResponse;

/**
 * Various tests for entity tags.
 * <p>
 * Copyright (c) SAP AG 2002-2004
 * @author julian.reschke@greenbytes.de
 * @version $Id: EntityTagTest.java,v 1.4 2004/06/17 13:10:13 jre Exp $
 */

public class EntityTagTest extends TestMember
{
  private static final String[] desc = new String[]
  {
    "00 ETAG (setup)",
    "01 ETAG compare tags from PUT / HEAD / PROPFIND",
    "02 ETAG subsequent PUTs must create distinct etags",
    "03 ETAG unique by URI after MOVE",
    "04 ETAG create dir1/a dir2/a, then MOVE and compare etags",
    "   ETAG cleanup of created resources",
  };

  public static int NrOfTests = desc.length;

  public EntityTagTest(Integer nr)
  {
    super(nr, desc[nr.intValue()]);
  }

  protected void runTest()
  {
    if (nr == 0) super.logIdentifier("ETAG");

    try
    {
      IResponse response;

      switch (nr)
      {
        case 0: {
            clear(request);
            request.setMethod("MKCOL");
            request.setURI(getLocalpart() + "/etags");
            response = requester.perform(request);
            checkResponseCode(response, 201);
          }
          break;

        case 1: {
            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + "/etags/file");
            request.setBody("foobar");
            response = requester.perform(request);
            checkResponseCode(response, 201);
            String etagFromPut = response.getHeader("etag");
            checkEntityTagSyntax(etagFromPut);

            clear(request);
            request.setMethod("HEAD");
            request.setURI(getLocalpart() + "/etags/file");
            request.setHeader("Translate", "f");
            response = requester.perform(request);
            checkResponseCode(response, 200);
            String etagFromHead = response.getHeader("etag");
            checkEntityTagSyntax(etagFromHead);

            String etagFromPropfind = DomElement.getText(
              getProperty(getLocalpart() + "/etags/file", Consts.N_getetag));
            checkEntityTagSyntax(etagFromPropfind);

            assertTrue("must return ETag from HEAD", etagFromHead != null);
            assertTrue("must return DAV:getetag", etagFromPropfind != null);

            assertTrue("HEAD/PROPFIND must agree on ETag", etagFromHead.equals(etagFromPropfind));

            if (etagFromPut != null) {
              assertTrue("PUT (" + etagFromPut + ")/PROPFIND(" + etagFromPropfind +") must agree on ETag when it was returned upon PUT", etagFromPut.equals(etagFromPropfind));
            }
          }
          break;

        case 2: {
            clear(request);
            request.setMethod("HEAD");
            request.setURI(getLocalpart() + "/etags/file");
            request.setHeader("Translate", "f");
            response = requester.perform(request);
            checkResponseCode(response, 200);
            String etag1 = response.getHeader("etag");

            String etag2 = put(getLocalpart() + "/etags/file", "foobar2");
            String etag3 = put(getLocalpart() + "/etags/file", "foobar23");

            assertTrue("etag1 present", etag1 != null);
            assertTrue("etag2 present", etag2 != null);
            assertTrue("etag3 present", etag3 != null);

            assertTrue("etag1 != etag2: " + etag1 + ", " + etag2, ! etag1.equals(etag2));
            assertTrue("etag1 != etag3: " + etag1 + ", " + etag3, ! etag1.equals(etag3));
            assertTrue("etag2 != etag3: " + etag2 + ", " + etag3, ! etag2.equals(etag3));

            checkEntityTagSyntax(etag1);
            checkEntityTagSyntax(etag2);
            checkEntityTagSyntax(etag3);
           }
          break;

        case 3: {
            String etaga = put(getLocalpart() + "/etags/filea", "foobar234");
            String etagb = put(getLocalpart() + "/etags/fileb", "foobar235");

            clear(request);
            request.setMethod("MOVE");
            request.setURI(getLocalpart() + "/etags/filea");
            request.setHeader("destination", getLocalpart() + "/etags/fileb");
            response = requester.perform(request);
            checkResponseCode(response, 204);

            clear(request);
            request.setMethod("HEAD");
            request.setURI(getLocalpart() + "/etags/fileb");
            request.setHeader("Translate", "f");
            response = requester.perform(request);
            checkResponseCode(response, 200);
            String etagb2 = response.getHeader("etag");

            assertTrue("etaga present", etaga != null);
            assertTrue("etagb present", etagb != null);
            assertTrue("etagb2 present", etagb2 != null);
            assertTrue("etags on fileb not changed", ! etagb.equals(etagb2));

            checkEntityTagSyntax(etaga);
            checkEntityTagSyntax(etagb);
          }
          break;

        case 4: {
            clear(request);
            request.setMethod("MKCOL");
            request.setURI(getLocalpart() + "/etags/dir1");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("MKCOL");
            request.setURI(getLocalpart() + "/etags/dir2");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            // String etaga1 =
            put(getLocalpart() + "/etags/dir1/a", "foobar2346");
            String etaga2 = put(getLocalpart() + "/etags/dir2/a", "foobar2357");

            clear(request);
            request.setMethod("MOVE");
            request.setURI(getLocalpart() + "/etags/dir1");
            request.setHeader("destination", getLocalpart() + "/etags/dir2");
            response = requester.perform(request);
            checkResponseCode(response, 204);

            clear(request);
            request.setMethod("HEAD");
            request.setURI(getLocalpart() + "/etags/dir2/a");
            request.setHeader("Translate", "f");
            response = requester.perform(request);
            checkResponseCode(response, 200);
            String etaga2b = response.getHeader("etag");

            assertTrue("etags on dir2/a must have changed", ! etaga2.equals(etaga2b));
            checkEntityTagSyntax(etaga2);
            checkEntityTagSyntax(etaga2b);
          }
          break;

        default: {
            clear(request);
            request.setMethod("DELETE");
            request.setURI(getLocalpart() + "/etags");
            response = requester.perform(request);
            checkResponseCode(response, 204);
          }
          break;
      }
    }
    catch (Exception ex) {
      ex.printStackTrace(System.err);
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }

  private String put(String uri, String body) throws WcmException {

    clear(request);
    request.setMethod("PUT");
    request.setURI(uri);
    request.setBody(body);
    IResponse response = requester.perform(request);
    checkResponseCode(response, 200, 201);
    String etag = response.getHeader("etag");

    // get  etag using HEAD iff necessary
    if (etag == null) {
      clear(request);
      request.setMethod("HEAD");
      request.setURI(uri);
      request.setHeader("Translate", "f");
      response = requester.perform(request);
      checkResponseCode(response, 200);
      etag = response.getHeader("etag");
    }

    return etag;
  }

  public void checkEntityTagSyntax(String etag) {
    
    // allow empty
    if (etag == null) return;
    
    // weak?
    if (etag.startsWith("W/")) {
      String t = etag.substring(2);
      assertTrue("W/ must be followed by double quote", t.startsWith("\""));
      assertTrue("W/ must be end in double quote", t.endsWith("\""));
      assertTrue("quoted section may not be empty", t.length() > 2);
    }
    else{
      assertTrue("must start with double quote", etag.startsWith("\""));
      assertTrue("be end in double quote", etag.endsWith("\""));
      assertTrue("quoted section may not be empty", etag.length() > 2);      
    }
  }
  
}
