package com.sapportals.wcm.protocol.webdav.server.tests;

import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.OutputStream;
import java.io.Writer;
import java.net.UnknownHostException;
import java.util.List;

import org.w3c.dom.Element;

import com.sapportals.wcm.IWcmConst;
import com.sapportals.wcm.protocol.webdav.Consts;
import com.sapportals.wcm.protocol.webdav.DavStatus;
import com.sapportals.wcm.protocol.webdav.DomElement;
import com.sapportals.wcm.protocol.webdav.Name;
import com.sapportals.wcm.repository.util.FileUtils;
import com.sapportals.wcm.util.http.IResponse;

/**
 * Tests for checking interactions between direct file system access and Repository Managers
 * that access those.
 * <p>
 * <em>IMPORTANT NOTE:</em> the runtime behaviour of these tests depends on the VM's behaviour,
 * which, in JDK 1.4, is influenced by the "CanonicalPath" cache. Test cases will only run
 * properly when it's explicitly turned off (use <code>-Dsun.io.useCanonCaches=false</code>.
 * <p>
 * <em>IMPORTANT NOTE:</em> running these tests inside the portal's VM (in the test framework)
 * may skew results as both the test cases and the KM will use the same (VM-based) caches.
 * <p>
 * Copyright (c) SAP AG 2003-2005
 * @author stefan.eissing@greenbytes.de
 * @author julian.reschke@greenbytes.de
 */
public class FSInteractionTest extends TestMember {

  private static final String[] desc = new String[] {
    "00 FSI (setup)",
    "01 FSI create initial test coll, check using HEAD",
    "02 FSI create file, check presence, sap:owner and dav:getlastmodified",
    "03 FSI create file, check presence after lookup on parent coll, check sap:owner and dav:getlastmodified",
    "04 FSI create file, check presence after GET then PROPFIND lookup on parent coll",
    "05 FSI update file through fs, check using HEAD",
    "06 FSI update file through HTTP, check DAV:lastmodifiedby",
    "07 FSI create collection, check presence and owner",
    "08 FSI create collection, update through PROPFIND on parent coll, check presence and owner",
    "09 FSI PROPPATCH displayname, touch file, check metadata kept",
    "10 FSI create VCR, touch file, compare base revision and VCR",
    "11 FSI create file, HEAD, rename file to upper case, change content, HEAD, check existence and content length",
    "12 FSI create VCC, delete non-VCR child, HEAD, check that VCC is checked out.",
    "13 FSI create file with leading blank, check with HEAD",
    "14 FSI create collection, modify child resource, set collection enhanced, modify child resource, set collection unenhanced",
    "15 FSI create file, GET, modify file, GET",
    "16 FSI create file, HEAD, set dead prop, modify file, check dead prop",
    "17 FSI create file, HEAD, delete file, HEAD",
    "18 FSI create file, HEAD, set dead prop, move file away, HEAD, move file back, check dead prop",
    "19 FSI create file, GET, delete file, create folder with same name, GET",
    "20 FSI create folder, GET, delete folder, create file with same name, GET",
    "21 FSI create folder a, create subfolder b, create resource c, GET on a, check PROPFIND on b",
    "22 FSI create file, HEAD, set dead prop, move file away, HEAD, create new res with same uri, check dead prop",
    "23 FSI unused (rel-backup)",
    "24 FSI unused (rel-backup)",
    "25 FSI create hierarchy, sync root, sync leaf, check intermediate",
    "26 FSI create collection with kids, HEAD, delete collection, HEAD",
    "27 FSI create link, place file at same location into FS",
    "28 FSI create link, place folder at same location into FS",
     "   FSI (cleanup)"
  };

  public static final int NrOfTests = desc.length;

  public static final String NS_CRT = "http://sapportals.com/xmlns/cm/crt";
   
  public static final Name PN_modifiedby = new Name(IWcmConst.SAP_WCM_NAMESPACE, "modifiedby");
  public static final Name PN_CRT_cfgClass = new Name(NS_CRT, "config-class");
  public static final Name PN_CRT_cfgLocalRoot = new Name(NS_CRT, "config-localroot");
  public static final Name PN_CRT_cfgPersistenceMode = new Name(NS_CRT, "config-persistence.mode");
  public static final Name PN_CRT_cfgVersionRoot = new Name(NS_CRT, "config-versioning.localroot");
  public static final Name PN_CRT_cfgNoSync = new Name(NS_CRT, "config-fsnosync");

  private static FSBasedRepository FS;
  
  private static synchronized FSBasedRepository initFS() throws Exception {
    FSInteractionTest test = new FSInteractionTest(new Integer(0));
    new FSTestSuite().init(test);
    return test.discoverFileSystemRepository();
  }

  public static boolean hasCollectionsInFS() throws Exception {
    FSBasedRepository fs = FS;
    if (fs == null) {
      fs = initFS();
    }
    return (fs != null && fs.keepsCollections);
  }

  public static boolean storesContentInFS() throws Exception {
    FSBasedRepository fs = FS;
    if (fs == null) {
      fs = initFS();
    }
    return (fs != null && fs.basecoll != null);
  }

  public static boolean IsCMRepository() throws Exception {
    FSBasedRepository fs = FS;
    if (fs == null) {
      fs = initFS();
    }
    return (fs != null && fs.isCMRepository());
  }
  
  /**
   * Create a file instead of a workspace collection in the file system of the given name.
   * @param wsName name of workspace
   * @return true if collection is there, false if creation was not done
   */
  public static boolean createWorkspaceDir(String wsName) throws Exception {
    FSBasedRepository fs = FS;
    if (fs == null) {
      fs = initFS();
    }
    assertTrue("test repository is not filesystem based", fs != null);
    if (fs.wscoll != null) {
      File workspaceDir = new File(fs.wscoll, wsName);
      workspaceDir.mkdir();
      return true;
    }    
    return false; 
  }
  
  /**
   * Remove the content of a resource in the file system
   * @param path below localpart of resource
   * @return true if content could be removed
   */
  public static boolean removeFSContent(String path) throws Exception {
    FSBasedRepository fs = FS;
    if (fs == null) {
      fs = initFS();
    }
    assertTrue("test repository is not filesystem based", fs != null);
    if (fs.keepsCollections) {
      File resfile = new File(fs.basecoll, path);
      return resfile.delete();
    }
    else if ("cmdbfs".equals(fs.mode)) {
      FSInteractionTest test = new FSInteractionTest(new Integer(0));
      new FSTestSuite().init(test);
      return test.removeDBFSContent(path);
    }
    return false;
  }

  /**
   * Create an empty content collection in old structure
   * @return path relative of base collection of content resources of dir
   */
  public static String assureEmptyContentCollection() throws Exception {
    FSBasedRepository fs = FS;
    if (fs == null) {
      fs = initFS();
    }
    assertTrue("test repository is not filesystem based", 
      fs != null && "cmdbfs".equals(fs.mode));

    FSInteractionTest test = new FSInteractionTest(new Integer(0));
    new FSTestSuite().init(test);
    return test.assureEmptyDBFSContentCollection();
  }

  /**
   * Check that an empty content collection is no longer there
   */
  public static void assureContentCollectionGone(String path) throws Exception {
    FSBasedRepository fs = FS;
    if (fs == null) {
      fs = initFS();
    }
    assertTrue("test repository is not filesystem based", 
      fs != null && "cmdbfs".equals(fs.mode));

    FSInteractionTest test = new FSInteractionTest(new Integer(0));
    new FSTestSuite().init(test);
    test.assureDBFSContentCollectionGone(path);
  }

  /**
   * Check that a tmp file is no longer there
   */
  public static void AssureTmpFileGone(String path) throws Exception {
    FSBasedRepository fs = FS;
    if (fs == null) {
      fs = initFS();
    }
    assertTrue("test repository is not CM repository based", 
      fs != null && fs.isCMRepository());

    FSInteractionTest test = new FSInteractionTest(new Integer(0));
    new FSTestSuite().init(test);
    test.assureTmpFileGone(path);
  }

  /**
   * Create an seemingly left over tmp file
   * @return path relative of base collection of tmp file
   */
  public static String AssureLeftoverTmpFile() throws Exception {
    FSBasedRepository fs = FS;
    if (fs == null) {
      fs = initFS();
    }
    assertTrue("test repository is not CM repository based", 
      fs != null && fs.isCMRepository());

    FSInteractionTest test = new FSInteractionTest(new Integer(0));
    new FSTestSuite().init(test);
    return test.assureLeftoverTmpFile();
  }
  
  /**
   * Create an seemingly left overv version file
   * @return path relative of base collection of version file
   */
  public static String AssureLeftoverVersion() throws Exception {
    FSBasedRepository fs = FS;
    if (fs == null) {
      fs = initFS();
    }
    assertTrue("test repository is not CM repository based", 
      fs != null && fs.isCMRepository() && fs.versioncoll != null);

    FSInteractionTest test = new FSInteractionTest(new Integer(0));
    new FSTestSuite().init(test);
    return test.assureLeftoverVersion();
  }
  
  /**
   * Remove a leftover version file created before
   * @param path relative path of version file
   */
  public static void RemoveLeftoverVersion(String path) throws Exception {
    FSBasedRepository fs = FS;
    if (fs == null) {
      fs = initFS();
    }
    assertTrue("test repository is not CM repository based", 
      fs != null && fs.isCMRepository() && fs.versioncoll != null);

    FSInteractionTest test = new FSInteractionTest(new Integer(0));
    new FSTestSuite().init(test);
    test.removeLeftoverVersion(path);
  }
  
  private static final String REP_CLASS_CM =  "com.sapportals.wcm.repository.manager.cm.CmRepositoryManager";
  private static final String REP_CLASS_SFS = "com.sapportals.wcm.repository.manager.sfs.FSRepositoryManager";

  private static String etag1;
  private static String etag2;

  public static final Name N_SAP_link_type = new Name (Consts.NS_WDSERVLET, "link-type");
  public static final String ELEMENT_type_internal_link = N_SAP_link_type.getStartTag()
  + Consts.N_SAP_internal_link.getEmptyTag() + N_SAP_link_type.getEndTag();

  private static boolean supportsInternalLinks = true;

  public static class FSBasedRepository {

    public static FSBasedRepository getInstance(String base, String mode, String path, String versionpath, boolean nosync) 
      throws UnknownHostException {
      path = FileUtils.convertHostnameToIPaddr(path);
      if (versionpath != null) {
        versionpath = FileUtils.convertHostnameToIPaddr(versionpath);
      }
      return new FSBasedRepository(base, mode, path, versionpath, nosync);    
    }
    
    public final String mode;
    public final boolean keepsCollections;
    public final boolean nosync;
    public final File basecoll;
    public final File testcoll;
    public final File plain;
    public final File plain_leadingblank;
    public final File plain2;
    public final File plain3;
    public final File plain4;
    public File plain5;
    public final File childcoll;
    public final File childcoll2;
    public final File childcoll3;
    public final File childcollEnhanced;
    public final File plain6;
    public final File plain7;
    public final File versioncoll;
    public final File wscoll;
    public final File wrcoll;
    
    public final File childcoll16;
    public final File plain16;
    public final File childcoll17;
    public final File plain17;
    public final File childcoll18;
    public final File plain18;
    public final File childcoll19;
    public final File plain19;
    public final File plain20;
    public final File plain21;
    public final File plain22;
    public final File childcoll22;
    public final File childcoll23a;
    public final File childcoll23b;
    public final File destcoll23;
    public final File childcoll24a;
    public final File childcoll24b;
    public final File destcoll24;
    public final File test25;
    public final File test25b;

    private FSBasedRepository(String base, String mode, String path, String versionpath, boolean nosync) {
      this.mode = mode;
      this.nosync = nosync;
      this.keepsCollections = ("sfs".equals(mode) || "cmfsdb".equals(mode));
      if (this.keepsCollections) {
        this.basecoll = new File(path, base);
      }
      else {
        this.basecoll = new File(path);
      }
        
      this.testcoll = new File(basecoll, "fstest");
      this.plain = new File(testcoll, "testfile.txt");
      this.plain2 = new File(testcoll, "testfile2.txt");
      this.plain3 = new File(testcoll, "testfile3.txt");
      this.plain4 = new File(testcoll, "testfile4.txt");
      this.plain5 = new File(testcoll, "testfile5.txt");
      this.plain_leadingblank = new File(testcoll, " leadingblank");
      this.childcoll = new File(testcoll, "childcoll");
      this.childcoll2 = new File(testcoll, "childcoll2");
      this.childcoll3 = new File(testcoll, "childcoll3");
      this.plain6 = new File(childcoll3, "testfile6.txt");
      if (versionpath != null) {
        this.versioncoll = new File(versionpath);
        this.wscoll = new File(this.versioncoll, "workspaces");
        this.wrcoll = new File(this.versioncoll, "workingresources");
      }
      else {
        this.versioncoll = null;
        this.wscoll = null;
        this.wrcoll = null;
      }
      this.childcollEnhanced = new File(this.testcoll, "childcollEnhanced");
      this.plain7 = new File(this.childcollEnhanced, "testfile7.txt");
      this.childcoll16 = new File(this.testcoll, "childcoll16");
      this.plain16 = new File(this.childcoll16, "testfile16.txt");
      this.childcoll17 = new File(this.testcoll, "childcoll17");
      this.plain17 = new File(this.childcoll17, "testfile17.txt");
      this.childcoll18 = new File(this.testcoll, "childcoll18");
      this.plain18 = new File(this.testcoll, "testfile18.txt");
      this.childcoll19 = new File(this.testcoll, "childcoll19");
      this.plain19 = new File(this.childcoll19, "testfile19.txt");
      this.plain20 = new File(this.testcoll, "testfile20.txt");
      this.plain21 = new File(this.testcoll, "testfile21.txt");
      this.plain22 = new File(this.testcoll, "testfile22.txt");
      this.childcoll22 = new File(this.testcoll, "childcoll22");
      this.childcoll23a = new File(this.testcoll, "childcoll23a");
      this.childcoll23b = new File(new File(new File(this.childcoll23a, "sub1"), "sub2"), "childcoll23b");
      this.destcoll23 = new File((versionpath == null)? this.basecoll : this.versioncoll, "destcoll23");
      this.childcoll24a = new File(this.testcoll, "childcoll24a");
      this.childcoll24b = new File(new File(new File(this.childcoll24a, "sub1"), "sub2"), "childcoll24b");
      this.destcoll24 = new File((versionpath == null)? this.basecoll : this.versioncoll, "destcoll24");
      this.test25 = new File(new File(new File(this.testcoll, "childcoll25"), "sub1"), "sub2");
      this.test25b = new File(new File(new File(this.testcoll, "childcoll25"), "sub1"), "sub2b");
    }
    
    public boolean isCMRepository() {
      return (mode != null && mode.startsWith("cm"));
    }
  }

  public FSInteractionTest(Integer nr) {
    super(nr, desc[nr.intValue()]);
  }

  protected void runTest() {
    final String prop_ResType = "<propfind xmlns='DAV:'><prop><resourcetype /></prop></propfind>";
    
    if (nr == 0) logIdentifier("FSI");

    try {
      synchronized (this.getClass()) {
        if (FS == null) {
          FS = discoverFileSystemRepository();
        }
      }

      String respfile = "FSI-" + TestMember.THREEDIGITS.format(nr) + ".txt";
      IResponse response;

      switch (nr) {

        case 0: {
            assertTrue("base coll does not exist or isn't a directory", FS.basecoll.exists() && FS.basecoll.isDirectory());
          }
          break;

        case 1: {
            if (FS.testcoll.exists()) FS.testcoll.delete(); 
              
            assertTrue("can't create test directory: " + FS.testcoll, FS.testcoll.mkdirs());

            clear(request);
            request.setMethod("HEAD");
            request.setHeader("cache-control", "no-cache");
            request.setURI(getLocalpart() + "/fstest");
            response = requester.perform(request);
            checkResponseCode(response, FS.nosync ? 404 : 200);
            
            if (FS.nosync) {
              request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
              response = requester.perform(request);
              checkResponseCode(response, 200);
            }
            
            request.clear();
            request.setMethod("MKREDIRECTREF");
            request.setURI(getLocalpart() + "/fstest/il");
            request.setBody("<mkredirectref xmlns='DAV:'>" + ELEMENT_type_internal_link + "<reftarget><href>" + getLocalpart()
                + "/fstest/foo</href></reftarget>" + "</mkredirectref>");
            request.expectsResponseDocument(true);
            response = requester.perform(request);
            checkResponseCode(response, 201, 501);
            if (response.getStatus() == 501) {
              supportsInternalLinks = false;
            }
          }
          break;

        case 2: {
            // Timestamp in a non-DST period
            long ts = 86400L * 1000 * 365 * 10L;
            String tss = "Sun, 30 Dec 1979 00:00:00 GMT";
            
            assertTrue("test file not created", FS.plain.createNewFile());
            assertTrue("last mod not set", FS.plain.setLastModified(ts));
            
            clear(request);
            request.setMethod("HEAD");
            request.setURI(getLocalpart() + "/fstest/testfile.txt");
            request.setHeader("Translate", "f");
            request.setHeader("cache-control", "no-cache");
            response = requester.perform(request);
            checkResponseCode(response, FS.nosync ? 404 : 200);
            
            if (FS.nosync) {
              request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
              response = requester.perform(request);
              checkResponseCode(response, 200);
            }
            assertTrue("content length header missing or not '0'", "0".equals(response.getHeader("content-length")));
            etag1 = response.getHeader("etag");

            Element owner = getProperty(getLocalpart() + "/fstest/testfile.txt", PN_modifiedby);
            assertTrue("sap:owner must be null", owner == null);

            assertTrue("HEAD: last-modified header wrong, was '" + response.getHeader("last-modified") + 
                "', expected: '" + tss, tss.equals(response.getHeader("last-modified")));
            
            // also do PROPFIND (which gets the system property instead of asking IContent
            Element lastmod = super.getProperty(getLocalpart() + "/fstest/testfile.txt", Consts.N_getlastmodified);
            assertTrue("PROPFIND: DAV:getlastmodified date wrong", tss.equals(DomElement.getText(lastmod)));
          }
          break;

        case 3: {
            long parentmod = FS.testcoll.lastModified();
            assertTrue("test file not created", FS.plain2.createNewFile());
            if (parentmod == FS.testcoll.lastModified()) {
              try{
                Thread.sleep(10000);
              }
              catch (Exception ignore) {
                // ignore
              }
            }                
            
            // timestamp in DST period
            long ts = 1065633600000L;
            String tss = "Wed, 08 Oct 2003 17:20:00 GMT";
            
            assertTrue("last mod not set", FS.plain2.setLastModified(ts));
            assertTrue("lastmod of parent did not change", parentmod != FS.testcoll.lastModified());
            
            clear(request);
            request.setMethod("PROPFIND");
            request.setURI(getLocalpart() + "/fstest/");
            request.setHeader("Depth", "1");
            request.setHeader("cache-control", "no-cache");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            List hrefs = getHrefsFromMultistatus(response.getDocument());
            if (FS.nosync) {
              assertFalse("testfile2.txt present in PROPFIND response: " + hrefs, hrefs.contains(getLocalpart() + "/fstest/testfile2.txt"));
              request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
              response = requester.perform(request);
              checkResponseCode(response, 207);
              hrefs = getHrefsFromMultistatus(response.getDocument());
            } 
            assertTrue("testfile2.txt not present in PROPFIND response: " + hrefs, hrefs.contains(getLocalpart() + "/fstest/testfile2.txt"));
            
            clear(request);
            request.setMethod("HEAD");
            request.setURI(getLocalpart() + "/fstest/testfile2.txt");
            request.setHeader("Translate", "f");
            request.setHeader("cache-control", "no-cache");
            response = requester.perform(request);
            checkResponseCode(response, 200);
            assertTrue("content length header missing or not '0'", "0".equals(response.getHeader("content-length")));
            etag1 = response.getHeader("etag");

            Element owner = getProperty(getLocalpart() + "/fstest/testfile2.txt", PN_modifiedby);
            assertTrue("sap:owner must be null", owner == null);

            assertTrue("HEAD: last-modified header wrong", tss.equals(response.getHeader("last-modified")));

            // also do PROPFIND (which gets the system property instead of asking IContent
            Element lastmod = super.getProperty(getLocalpart() + "/fstest/testfile2.txt", Consts.N_getlastmodified);
            assertTrue("PROPFIND: DAV:getlastmodified date wrong", tss.equals(DomElement.getText(lastmod)));
          }
          break;

        case 4: {
          long parentmod = FS.testcoll.lastModified();
          assertTrue("test file not created", FS.plain3.createNewFile());
          if (parentmod == FS.testcoll.lastModified()) {
            try{
              Thread.sleep(10000);
            }
            catch (Exception ignore) {
              // ignore
            }
          }                                              
          
          assertTrue("lastmod of parent did not change", parentmod != FS.testcoll.lastModified());
          assertTrue("lastmod not set", FS.plain3.setLastModified(1065633600000L));  
          
          clear(request);
          request.setMethod("GET");
          request.setURI(getLocalpart() + "/fstest/");
          request.setHeader("Depth", "1");
          request.setHeader("cache-control", "no-cache");
          response = requester.perform(request);
          checkResponseCode(response, 200);
          
          clear(request);
          request.setMethod("PROPFIND");
          request.setURI(getLocalpart() + "/fstest/");
          request.setHeader("Depth", "1");
          request.setHeader("cache-control", "no-cache");
          response = requester.perform(request);
          checkResponseCode(response, 207);
          List hrefs = getHrefsFromMultistatus(response.getDocument());
          if (FS.nosync) {
            assertFalse("testfile3.txt present in PROPFIND response: " + hrefs, hrefs.contains(getLocalpart() + "/fstest/testfile3.txt"));
            request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
            response = requester.perform(request);
            hrefs = getHrefsFromMultistatus(response.getDocument());
          }
          assertTrue("testfile3.txt not present in PROPFIND response: " + hrefs, hrefs.contains(getLocalpart() + "/fstest/testfile3.txt"));
        }
        break;

        case 5: {
            OutputStream os = new FileOutputStream(FS.plain);
            os.write("Dies ist ein Test".getBytes());
            os.close();

            clear(request);
            request.setMethod("GET");
            request.setURI(getLocalpart() + "/fstest/testfile.txt");
            request.setHeader("Translate", "f");
            request.setHeader("cache-control", "no-cache");
            if (FS.nosync) request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
            response = requester.perform(request);
            checkResponseCode(response, 200);

            assertTrue("content length header missing or not '17'", "17".equals(response.getHeader("content-length")));

            etag2 = response.getHeader("etag");
            assertTrue("new etag", ! etag1.equals(etag2));
          }
          break;

        case 6: {
            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + "/fstest/testfile.txt");
            request.setBody("foobar");
            response = requester.perform(request);
            checkResponseCode(response, 200);

            Element owner = getProperty(getLocalpart() + "/fstest/testfile.txt", PN_modifiedby);
            if (owner == null) {
              super.logComment("modifiedby: not set");
            }
            else {
              assertTrue("SAP:modified-by must be '" + m_userid + "' when present", m_userid.equalsIgnoreCase(DomElement.getText(owner)));
              super.logComment("modifiedby: set");
            }
          }
          break;

        case 7: {
            assertTrue("child collection not created", FS.childcoll.mkdir());
            assertTrue("last mod not set", FS.childcoll.setLastModified(1065633600000L));

            clear(request);
            request.setMethod("HEAD");
            request.setURI(getLocalpart() + "/fstest/childcoll");
            request.setHeader("Translate", "f");
            request.setHeader("cache-control", "no-cache");
            if (FS.nosync) request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
            response = requester.perform(request);
            checkResponseCode(response, 200);

            Element owner = getProperty(getLocalpart() + "/fstest/childcoll", PN_modifiedby);
            assertTrue("sap:owner must be null", owner == null);

            Element lastmod = getProperty(getLocalpart() + "/fstest/childcoll", Consts.N_getlastmodified);
            assertTrue("PROPFIND: DAV:getlastmodified date '" + DomElement.getText(lastmod) + "' wrong", "Wed, 08 Oct 2003 17:20:00 GMT".equals(DomElement.getText(lastmod)));
          }
          break;

        case 8: {
            assertTrue("child collection not created", FS.childcoll2.mkdir());
            assertTrue("last mod not set", FS.childcoll2.setLastModified(1065633600000L));

            clear(request);
            request.setMethod("PROPFIND");
            request.setURI(getLocalpart() + "/fstest/");
            request.setHeader("Depth", "1");
            request.setHeader("cache-control", "no-cache");
            if (FS.nosync) request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
            response = requester.perform(request);
            checkResponseCode(response, 207);

            clear(request);
            request.setMethod("HEAD");
            request.setURI(getLocalpart() + "/fstest/childcoll2");
            request.setHeader("Translate", "f");
            request.setHeader("cache-control", "no-cache");
            if (FS.nosync) request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
            response = requester.perform(request);
            checkResponseCode(response, 200);

            Element owner = getProperty(getLocalpart() + "/fstest/childcoll2", PN_modifiedby);
            assertTrue("sap:owner must be null", owner == null);

            Element lastmod = getProperty(getLocalpart() + "/fstest/childcoll2", Consts.N_getlastmodified);
            assertTrue("PROPFIND: DAV:getlastmodified date '" + DomElement.getText(lastmod) + "' wrong", "Wed, 08 Oct 2003 17:20:00 GMT".equals(DomElement.getText(lastmod)));
          }
          break;
          
        case 9: {
            DavStatus s = setProperty(getLocalpart() + "/fstest/testfile.txt", Consts.N_displayname, "foobar", false);
            if (s.getStatus() == 200) {
              
              // PROPPATCH successful?
              FS.plain.setLastModified(System.currentTimeMillis() + 100000L);
              
              // force refresh
              clear(request);
              request.setMethod("GET");
              request.setURI(getLocalpart() + "/fstest/testfile.txt");
              response = requester.perform(request);
              checkResponseCode(response, 200);
              
              // check displayname
              Element disp = getProperty(getLocalpart() + "/fstest/testfile.txt", Consts.N_displayname);
              assertTrue("DAV:displayname present", disp != null);
              assertTrue("value for DAV:displayname is as aset previously", DomElement.getText(disp).equals("foobar"));
            }
          }
          break;

        case 10: {
            // create a new file
            assertTrue("test file not created", FS.plain4.createNewFile());
            clear(request);
            request.setMethod("HEAD");
            request.setURI(getLocalpart() + "/fstest/testfile4.txt");
            request.setHeader("Translate", "f");
            request.setHeader("cache-control", "no-cache");
            response = requester.perform(request);
            assertTrue("content length header missing or not '0'", "0".equals(response.getHeader("content-length")));
  
            // make it a VCR
            clear(request);
            request.setMethod("VERSION-CONTROL");
            request.setURI(getLocalpart() + "/fstest/testfile4.txt");
            response = requester.perform(request);
            int responseCode = response.getStatus();
            if (responseCode >= 200 && responseCode < 300) {
              String versionUri = getRelativeUriFromLocationHeader(response);
  
              // modify the file
              FS.plain4.setLastModified(System.currentTimeMillis() + 100000L);
              
              // force synchronize
              request.setMethod("HEAD");
              request.setURI(getLocalpart() + "/fstest/testfile4.txt");
              request.setHeader("Translate", "f");
              request.setHeader("cache-control", "no-cache");
              response = requester.perform(request);
              assertTrue("content length header missing or not '0'", "0".equals(response.getHeader("content-length")));
              
              // test: no new version was created but the resource was checked out implicitly
              
              Element prop = super.getProperty(getLocalpart() + "/fstest/testfile4.txt", Consts.N_checked_out);
              assertNotNull("resource is not checked out", prop);
              String newVersionURI = super.getRelativeUriFromPropertyHrefs(prop);
              assertTrue("a new version created", versionUri.equals(newVersionURI));
            }
          }
          break;

        case 11: {
            // create a new file
            assertTrue("test file not created", FS.plain5.createNewFile());
              
            clear(request);
            request.setMethod("HEAD");
            request.setURI(getLocalpart() + "/fstest/testfile5.txt");
            request.setHeader("Translate", "f");
            request.setHeader("cache-control", "no-cache");
            response = requester.perform(request);
            assertTrue("content length header missing or not '0'", "0".equals(response.getHeader("content-length")));
    
            // rename it to upper case
            // change content length
            File temp = new File(FS.testcoll, "testfile5_xxx.txt");
            FS.plain5.renameTo(temp);
            FS.plain5 = temp;
              
            temp = new File(FS.testcoll, "TESTFILE5.txt");
            FS.plain5.renameTo(temp);
            FS.plain5 = temp;
            
            Writer out = new FileWriter(FS.plain5);
            out.write("Hello World!");
            out.flush(); out.close();
            
            // check existence and content length on correct URI
            clear(request);
            request.setMethod("HEAD");
            request.setURI(getLocalpart() + "/fstest/TESTFILE5.txt");
            request.setHeader("Translate", "f");
            request.setHeader("cache-control", "no-cache");
            if (FS.nosync) request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
            response = requester.perform(request);
            checkResponseCode(response, 200);
            assertTrue("content length header missing or not '12'", "12".equals(response.getHeader("content-length")));
            
            // This test case will fail if the jdk caches the results of getCanonicalPath. This was introduced
            // as part of jdk 1.4.2; and can be disabled by setting "sun.io.useCanonCaches=false".
  
            // check existence and content length on "old" URI
            clear(request);
            request.setMethod("HEAD");
            request.setURI(getLocalpart() + "/fstest/testfile5.txt");
            request.setHeader("Translate", "f");
            request.setHeader("cache-control", "no-cache");
            response = requester.perform(request);
            checkResponseCode(response, 200, 404);
            // allow 200 if content-location header points to correct one
            if (response.getStatus() == 200) {
              String cl = getRelativeUriFromContentLocationHeader(response);
              assertTrue("when matching caseless, correct URI must be returned in content-location header", cl.endsWith("/fstest/TESTFILE5.txt"));
              assertTrue("content length header missing or not '12'", "12".equals(response.getHeader("content-length")));
            }

            // get content of parent collection
            clear(request);
            request.setMethod("PROPFIND");
            request.setURI(getLocalpart() + "/fstest/");
            request.setHeader("Depth", "1");
            request.setHeader("cache-control", "no-cache");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            List hrefs = getHrefsFromMultistatus(response.getDocument());
            assertTrue("TESTFILE.txt not present in PROPFIND response: " + hrefs, hrefs.contains(getLocalpart() + "/fstest/TESTFILE5.txt"));
          }
          break;
          

          case 12: {
              // create a new VCC
              assertTrue("test collection not created", FS.childcoll3.mkdir());
              assertTrue("test file not created", FS.plain6.createNewFile());
  
              request.setMethod("HEAD");
              request.setURI(getLocalpart() + "/fstest/childcoll3");
              request.setHeader("Translate", "f");
              request.setHeader("cache-control", "no-cache");
              if (FS.nosync) request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
              response = requester.perform(request);              
              checkResponseCode(response, 200);

              // make it a VCR
              clear(request);
              request.setMethod("VERSION-CONTROL");
              request.setURI(getLocalpart() + "/fstest/childcoll3");
              response = requester.perform(request);
              int responseCode = response.getStatus();

              if (responseCode == 201) {

                clear(request);
                request.setMethod("CHECKOUT");
                request.setURI(getLocalpart() + "/fstest/childcoll3");
                response = requester.perform(request);
                checkResponseCode(response, 200);

                clear(request);
                request.setMethod("VERSION-CONTROL");
                request.setURI(getLocalpart() + "/fstest/childcoll3/testfile6.txt");
                response = requester.perform(request);
                checkResponseCode(response, 201);

                clear(request);
                request.setMethod("CHECKIN");
                request.setURI(getLocalpart() + "/fstest/childcoll3");
                response = requester.perform(request);
                checkResponseCode(response, 201);

                String versionUri = getRelativeUriFromLocationHeader(response);

                // test: VCC is checked in
                Element prop = super.getProperty(getLocalpart() + "/fstest/childcoll3", Consts.N_checked_in);
                assertNotNull("resource is not checked in", prop);

                // delete child resource
                assertTrue("test file not deleted", FS.plain6.delete());

                // force synchronize
                request.setMethod("GET");
                request.setURI(getLocalpart() + "/fstest/childcoll3");
                request.setHeader("Translate", "f");
                request.setHeader("cache-control", "no-cache");
                if (FS.nosync) request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
                response = requester.perform(request);
                checkResponseCode(response, 200);

                // test: no new version was created but the resource was checked out implicitly
                prop = super.getProperty(getLocalpart() + "/fstest/childcoll3", Consts.N_checked_out);
                assertNotNull("resource is not checked out", prop);
                String newVersionURI = super.getRelativeUriFromPropertyHrefs(prop);
                assertTrue("a new version was created", versionUri.equals(newVersionURI));
              }
            }
            break;
        
        case 13: {
            assertTrue("test file not created", FS.plain_leadingblank.createNewFile());
            
            clear(request);
            request.setMethod("HEAD");
            request.setURI(getLocalpart() + "/fstest/%20leadingblank");
            request.setHeader("Translate", "f");
            request.setHeader("cache-control", "no-cache");
            if (FS.nosync) request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
            response = requester.perform(request);
            checkResponseCode(response, 200);
          }
          break;

       case 14: {

          assertTrue("test collection not created", FS.childcollEnhanced.mkdir());
          assertTrue("test file not created", FS.plain7.createNewFile());
        
          String rootRef = "/" + FS.testcoll.getName();
          String colRef = rootRef + "/" + FS.childcollEnhanced.getName();
          String plainRef = colRef + "/" + FS.plain7.getName();
        
          clear(request);
          request.setMethod("HEAD");
          request.setURI(getLocalpart() + plainRef);
          request.setHeader("Translate", "f");
          request.setHeader("cache-control", "no-cache");
          if (FS.nosync) request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
          response = requester.perform(request);
          checkResponseCode(response, 200);
        
          Name pec = new Name("http://sapportals.com/xmlns/cm", "enhancedcollection");
          Element prop = getProperty(getLocalpart() + colRef, pec);
          boolean ecSupport = (prop != null);
        
          OutputStream out = new FileOutputStream(FS.plain7);
          out.write(0);
          out.close();

          if (ecSupport && !FS.nosync) {
            setProperty(getLocalpart() + colRef, pec, Boolean.TRUE, true);
          
            prop = getProperty(getLocalpart() + plainRef, Consts.N_getcontentlength);
            assertNotNull("content length is unknown", prop);
            String contLen = DomElement.getText(prop);
            assertEquals("wrong content length", "1", contLen);

            out = new FileOutputStream(FS.plain7);
            out.write(0);
            out.write(0);
            out.close();

            setProperty(getLocalpart() + colRef, pec, Boolean.FALSE, true);

            prop = getProperty(getLocalpart() + plainRef, Consts.N_getcontentlength);
            assertNotNull("content length is unknown", prop);
            contLen = DomElement.getText(prop);
            assertEquals("wrong content length", "2", contLen);
          }
        }
        break;

       case 15: {
          clear(request);
          request.setMethod("GET");
          request.setURI(getLocalpart() + "/fstest/testfile.txt");
          request.setHeader("Translate", "f");
          request.setHeader("cache-control", "no-cache");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          Writer out = new FileWriter(FS.plain4);
          String s = "15 FSI new content sdksad sadkajshd sad haskjdh ksajd hajksd haskdh asjkdh aksjdh asjkhdads!";
          out.write(s);
          out.flush();
          out.close();

          clear(request);
          request.setMethod("GET");
          request.setURI(getLocalpart() + "/fstest/testfile4.txt");
          request.setHeader("Translate", "f");
          // request.setHeader("cache-control", "no-cache");
          if (FS.nosync) request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
          response = requester.perform(request);
          checkResponseCode(response, 200);
          assertTrue("content length header is: '"+response.getHeader("content-length")+"', should have been: '" + s.length() + "'", Integer.toString(s.length()).equals(response.getHeader("content-length")));
        }
        break;

       case 16: {

         assertTrue("test file not created", FS.plain20.createNewFile());
         
         String rootRef = "/" + FS.testcoll.getName();
         String plainRef = rootRef + "/" + FS.plain20.getName();
         
         clear(request);
         request.setMethod("HEAD");
         request.setURI(getLocalpart() + plainRef);
         request.setHeader("Translate", "f");
         request.setHeader("cache-control", "no-cache");
         if (FS.nosync) request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
         response = requester.perform(request);
         checkResponseCode(response, 200);

         clear(request);
         request.setMethod("PROPPATCH");
         request.setURI(getLocalpart() + plainRef);
         request.setBody("<propertyupdate xmlns='DAV:'><set><prop><fsitest xmlns='fsitest:'>fsitest</fsitest></prop></set></propertyupdate>");
         response = requester.perform(request);
         checkResponseCode(response, 207, 501);

         if (response.getStatus() == 207) {
 
           final Name propName = new Name("fsitest:", "fsitest");
 
           Element fsitest = super.getProperty(getLocalpart() + plainRef, propName);
           assertNotNull("missing property 'fsitest'", fsitest);
           String propVal = DomElement.getText(fsitest);
           assertEquals("wrong value of property 'fsitest', expected: fsitest, was: " + fsitest, "fsitest", propVal);
 
           OutputStream os = new FileOutputStream(FS.plain20);
           os.write("Dies ist ein Test".getBytes());
           os.close();
           
           fsitest = super.getProperty(getLocalpart() + plainRef, propName);
           assertNotNull("missing property 'fsitest'", fsitest);
           propVal = DomElement.getText(fsitest);
           assertEquals("wrong value of property 'fsitest', expected: fsitest, was: " + fsitest, "fsitest", propVal);
         }
        }
        break;

       case 17: {

         assertTrue("test file not created", FS.plain21.createNewFile());
         
         String rootRef = "/" + FS.testcoll.getName();
         String plainRef = rootRef + "/" + FS.plain21.getName();
         
         clear(request);
         request.setMethod("HEAD");
         request.setURI(getLocalpart() + plainRef);
         request.setHeader("Translate", "f");
         request.setHeader("cache-control", "no-cache");
         if (FS.nosync) request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
         response = requester.perform(request);
         checkResponseCode(response, 200);

         assertTrue("test file not deleted", FS.plain21.delete());
           
         clear(request);
         request.setMethod("HEAD");
         request.setURI(getLocalpart() + plainRef);
         request.setHeader("Translate", "f");
         request.setHeader("cache-control", "no-cache");
         if (FS.nosync) request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
         response = requester.perform(request);
         checkResponseCode(response, 404);
        }
        break;

       case 18: {

         assertTrue("test file not created", FS.plain18.createNewFile());
         assertTrue("test coll not created", FS.childcoll18.mkdir());
         final File destFile = new File(FS.childcoll18, FS.plain18.getName());
         
         String rootRef = "/" + FS.testcoll.getName();
         String plainRef = rootRef + "/" + FS.plain18.getName();
//         String collRef = rootRef + "/" + FS.childcoll18.getName();
//         String destRef = collRef + "/" + FS.plain18.getName();
         
         clear(request);
         request.setMethod("HEAD");
         request.setURI(getLocalpart() + plainRef);
         request.setHeader("Translate", "f");
         request.setHeader("cache-control", "no-cache");
         if (FS.nosync) request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
         response = requester.perform(request);
         checkResponseCode(response, 200);

         clear(request);
         request.setMethod("PROPPATCH");
         request.setURI(getLocalpart() + plainRef);
         request.setBody("<propertyupdate xmlns='DAV:'><set><prop><fsitest xmlns='fsitest:'>fsitest</fsitest></prop></set></propertyupdate>");
         response = requester.perform(request);
         checkResponseCode(response, 207, 501);

         boolean propSet = (response.getStatus() == 207); 

         assertTrue("test file not moved", FS.plain18.renameTo(destFile));

         clear(request);
         request.setMethod("HEAD");
         request.setURI(getLocalpart() + plainRef);
         request.setHeader("Translate", "f");
         request.setHeader("cache-control", "no-cache");
         response = requester.perform(request);
         checkResponseCode(response, 404);

         assertTrue("test file not moved", destFile.renameTo(FS.plain18));

         clear(request);
         request.setMethod("HEAD");
         request.setURI(getLocalpart() + plainRef);
         request.setHeader("Translate", "f");
         request.setHeader("cache-control", "no-cache");
         response = requester.perform(request);
         checkResponseCode(response, 200);
         
         if (propSet) { 
           final Name propName = new Name("fsitest:", "fsitest");
           
           Element fsitest = super.getProperty(getLocalpart() + plainRef, propName);
           assertNull("found unexpected property 'fsitest'", fsitest);
         }
       }
        break;

       case 19: {
         // create test file
         String name = "FSI_16";
         File temp = new File(FS.testcoll, name);
         temp.createNewFile();
         assertFalse("test file is directory in local file system: " + temp.getPath(), temp.isDirectory());
         
         // GET on test file
         clear(request);
         request.setMethod("PROPFIND");
         request.setURI(getLocalpart() + "/fstest/" + name);
         if (FS.nosync) {
           request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
         }
         request.setHeader("Depth", "0");
         request.setBody(prop_ResType);
         response = requester.perform(request);
         checkResponseCode(response, 207);
         checkResponseDocument(response, respfile + ".a.txt", null);
         
         // delete test file
         temp.delete();
         assertFalse("temp file still exists: " + temp.getPath(), temp.exists());
         
         // create test collection with same name
         assertTrue("test directory was not created: " + temp.getPath(), temp.mkdir());
         
         // GET on test collection
         request.setMethod("GET");
         request.setURI(getLocalpart() + "/fstest/" + name);
         request.setHeader("Translate", "f");
         request.setHeader("cache-control", "no-cache");
         response = requester.perform(request);
         checkResponseCode(response, 200);
         
         clear(request);
         request.setMethod("PROPFIND");
         request.setURI(getLocalpart() + "/fstest/" + name);
         if (FS.nosync) {
           request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
         }
         request.setHeader("Depth", "0");
         request.setBody(prop_ResType);
         response = requester.perform(request);
         checkResponseCode(response, 207);
         checkResponseDocument(response, respfile + ".b.txt", null);
         
       }
       break;
       
       case 20: {
         // create test collection
         String name = "FSI_17";
         File temp = new File(FS.testcoll, name);
         assertTrue("test directory was not created: " + temp.getPath(), temp.mkdir());
         
         // GET on test file
         clear(request);
         request.setMethod("PROPFIND");
         request.setURI(getLocalpart() + "/fstest/" + name);
         request.setHeader("Depth", "0");
         if (FS.nosync) {
           request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
         }
         request.setBody(prop_ResType);
         response = requester.perform(request);
         checkResponseCode(response, 207);
         checkResponseDocument(response, respfile + ".a.txt", null);
         
         // delete test collection
         temp.delete();
         assertFalse("test collection still exists: " + temp.getPath(), temp.exists());
         
         // create test file with same name
         temp.createNewFile();
         assertTrue("test file does not exist: " + temp.getPath(), temp.exists());
         assertFalse("test file is directory in local file system: " + temp.getPath(), temp.isDirectory());
         
         // GET on test collection
         request.setMethod("GET");
         request.setURI(getLocalpart() + "/fstest/" + name);
         request.setHeader("Translate", "f");
         request.setHeader("cache-control", "no-cache");
         if (FS.nosync) {
           request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
         }
         response = requester.perform(request);
         checkResponseCode(response, 200);
         
         clear(request);
         request.setMethod("PROPFIND");
         request.setURI(getLocalpart() + "/fstest/" + name);
         request.setHeader("Depth", "0");
         request.setBody(prop_ResType);
         response = requester.perform(request);
         checkResponseCode(response, 207);
         checkResponseDocument(response, respfile + ".b.txt", null);
         
       }
       break;
        
       case 21: {
         String nameA = "FSI_21";
         File temp = new File(FS.testcoll, nameA);
         assertTrue("test directory was not created: " + temp.getPath(), temp.mkdir());
         
         String nameB = "b";
         temp = new File(temp, nameB);
         assertTrue("test directory was not created: " + temp.getPath(), temp.mkdir());
         
         String nameC = "c";
         temp = new File(temp, nameC);
         assertTrue("test directory was not created: " + temp.getPath(), temp.createNewFile());

         // GET on test collection
         request.setMethod("GET");
         request.setURI(getLocalpart() + "/fstest/" + nameA);
         request.setHeader("Translate", "f");
         request.setHeader("cache-control", "no-cache");
         if (FS.nosync) {
           request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
         }
         response = requester.perform(request);
         checkResponseCode(response, 200);
         
         // GET on test file
         clear(request);
         request.setMethod("PROPFIND");
         request.setURI(getLocalpart() + "/fstest/" + nameA + "/" + nameB);
         request.setHeader("Depth", "1");
         if (FS.nosync) {
           request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
         }
         request.setBody(prop_ResType);
         response = requester.perform(request);
         checkResponseCode(response, 207);
         checkResponseDocument(response, respfile, null);
       }
       break;
        
       case 22: {

         assertTrue("test file not created", FS.plain22.createNewFile());
         assertTrue("test coll not created", FS.childcoll22.mkdir());
         final File destFile = new File(FS.childcoll22, FS.plain22.getName());
         
         String rootRef = "/" + FS.testcoll.getName();
         String plainRef = rootRef + "/" + FS.plain22.getName();
//         String collRef = rootRef + "/" + FS.childcoll22.getName();
//         String destRef = collRef + "/" + FS.plain22.getName();
         
         clear(request);
         request.setMethod("HEAD");
         request.setURI(getLocalpart() + plainRef);
         request.setHeader("Translate", "f");
         request.setHeader("cache-control", "no-cache");
         if (FS.nosync) {
           request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
         }
         response = requester.perform(request);
         checkResponseCode(response, 200);

         clear(request);
         request.setMethod("PROPPATCH");
         request.setURI(getLocalpart() + plainRef);
         request.setBody("<propertyupdate xmlns='DAV:'><set><prop><fsitest xmlns='fsitest:'>fsitest</fsitest></prop></set></propertyupdate>");
         response = requester.perform(request);
         checkResponseCode(response, 207, 501);

         boolean propSet = (response.getStatus() == 207); 

         assertTrue("test file not moved", FS.plain22.renameTo(destFile));

         clear(request);
         request.setMethod("PUT");
         request.setURI(getLocalpart() + plainRef);
         request.setHeader("Translate", "f");
         request.setHeader("cache-control", "no-cache");
         request.setBody("body");
         response = requester.perform(request);
         checkResponseCode(response, 201);

         if (propSet) { 

           final Name propName = new Name("fsitest:", "fsitest");
           
           Element fsitest = super.getProperty(getLocalpart() + plainRef, propName);
           assertNull("found unexpected property 'fsitest'", fsitest);
         }
       }
       break;

       case 23: {
          // unused (rel-backup)
        }
        break;

       case 24: {
          // unused (rel-backup)
        }
        break;

       case 25: {

         assertTrue("test coll not created", FS.test25.mkdirs());
         assertTrue("test coll not created", FS.test25b.mkdirs());
         
         String rootRef = "/" + FS.testcoll.getName();
         String collRef = rootRef + "/childcoll25";
         String subRef1 = collRef + "/sub1";
         String subRef2 = subRef1 + "/sub2";
         
         clear(request);
         request.setMethod("HEAD");
         request.setURI(getLocalpart() + collRef);
         request.setHeader("Translate", "f");
         request.setHeader("cache-control", "no-cache");
         if (FS.nosync) {
           request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
         }
         response = requester.perform(request);
         checkResponseCode(response, 200);

         clear(request);
         request.setMethod("HEAD");
         request.setURI(getLocalpart() + subRef2);
         request.setHeader("Translate", "f");
         request.setHeader("cache-control", "no-cache");
         if (FS.nosync) {
           request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
         }
         response = requester.perform(request);
         checkResponseCode(response, 200);

         clear(request);
         request.setMethod("PROPFIND");
         request.setURI(getLocalpart() + subRef1);
         request.setHeader("Depth", "1");
         if (FS.nosync) {
           request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
         }
         request.setBody("<propfind xmlns='DAV:'><allprop/></propfind>");
         response = requester.perform(request);
         checkResponseCode(response, 207);
         List hrefs = super.getHrefsFromMultistatus(response.getDocument());
         
         assertTrue("missing child resource", hrefs != null && hrefs.size() > 2);
        }
        break;

       case 26: {
         String name = "FSI_26";
         String nameA = "a";

         File temp = new File(FS.testcoll, name);
         assertTrue("test directory was not created: " + temp.getPath(), temp.mkdir());
         File tempA = new File(temp, nameA);
         assertTrue("test file was not created: " + temp.getPath(), tempA.createNewFile());
         
         clear(request);
         request.setMethod("HEAD");
         request.setURI(getLocalpart() + "/fstest/" + name);
         request.setHeader("Translate", "f");
         request.setHeader("cache-control", "no-cache");
         if (FS.nosync) {
           request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
         }
         response = requester.perform(request);
         checkResponseCode(response, 200);
         
         request.setURI(getLocalpart() + "/fstest/" + name + "/" + nameA);
         response = requester.perform(request);
         checkResponseCode(response, 200);

         assertTrue("test file not deleted", tempA.delete());
         assertTrue("test directory not deleted", temp.delete());
           
         clear(request);
         request.setMethod("HEAD");
         request.setURI(getLocalpart() + "/fstest/" + name);
         request.setHeader("Translate", "f");
         request.setHeader("cache-control", "no-cache");
         if (FS.nosync) {
           request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
         }
         response = requester.perform(request);
         checkResponseCode(response, 404);
       }
       break;

       case 27: if (supportsInternalLinks) {
         request.clear();
         request.setMethod("MKREDIRECTREF");
         request.setURI(getLocalpart() + "/fstest/il_res");
         request.setBody("<mkredirectref xmlns='DAV:'>" + ELEMENT_type_internal_link + "<reftarget><href>" + getLocalpart()
             + "/fstest/foo</href></reftarget>" + "</mkredirectref>");
         request.expectsResponseDocument(true);
         response = requester.perform(request);
         checkResponseCode(response, 201);

         File file = new File(FS.testcoll, "il_res");
         assertTrue("Couldn't create test file:" + file.getPath(), file.createNewFile());

         clear(request);
         request.setMethod("PROPFIND");
         request.setURI(getLocalpart() + "/fstest/il_res");
         request.setHeader("Depth", "1");
         request.setHeader("cache-control", "no-cache");
         request.setHeader("Apply-To-Redirect-Ref", "T");
         if (FS.nosync) request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
         response = this.checkMultistatus(
           "<propfind xmlns='DAV:'><prop><resourcetype/></prop></propfind>",
           respfile);
         
       }
       break;
       
       case 28: if (supportsInternalLinks) {
         request.clear();
         request.setMethod("MKREDIRECTREF");
         request.setURI(getLocalpart() + "/fstest/il_coll");
         request.setBody("<mkredirectref xmlns='DAV:'>" + ELEMENT_type_internal_link + "<reftarget><href>" + getLocalpart()
             + "/fstest/foo</href></reftarget>" + "</mkredirectref>");
         request.expectsResponseDocument(true);
         response = requester.perform(request);
         checkResponseCode(response, 201);

         File file = new File(FS.testcoll, "il_coll");
         assertTrue("Couldn't create test file:" + file.getPath(), file.mkdir());
         
         clear(request);
         request.setMethod("PROPFIND");
         request.setURI(getLocalpart() + "/fstest/il_coll");
         request.setHeader("Depth", "1");
         request.setHeader("cache-control", "no-cache");
         request.setHeader("Apply-To-Redirect-Ref", "T");
         if (FS.nosync) request.setHeader("x-greenbytes-resource-context", "PersistenceFSDB.SYNC=true");
         response = this.checkMultistatus(
           "<propfind xmlns='DAV:'><prop><resourcetype/></prop></propfind>",
           respfile);
       }
       break;
       
       default: {
         supportsInternalLinks = true;
         
          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/fstest");
          response = requester.perform(request);
          checkResponseCode(response, 204);
          
          // reset static members
          FS = null;
        }
        break;
      }
    }
    catch (Exception ex) {
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }

  public FSBasedRepository discoverFileSystemRepository() throws Exception {
    Element prop = this.getProperty(getLocalpart(), Consts.N_SAP_repository);
    assertTrue("repository property not defined for: "+getLocalpart(), prop != null); 
    String repository = this.getRelativeUriFromPropertyHrefs(prop);
    
    prop = this.getProperty(repository, PN_CRT_cfgClass);
    if (prop != null) {
      String configClass = DomElement.getText(prop);
      String localroot = null;
      String versionroot = null;
      String mode = null;
      boolean nosync = false;
      if (REP_CLASS_CM.equals(configClass)) {
        prop = this.getProperty(repository, PN_CRT_cfgPersistenceMode);
        if (prop != null) {
          mode = "cm" + DomElement.getText(prop);
        }
        prop = this.getProperty(repository, PN_CRT_cfgLocalRoot);
        if (prop != null) {
          localroot = DomElement.getText(prop);
        }
        prop = this.getProperty(repository, PN_CRT_cfgVersionRoot);
        if (prop != null) {
          versionroot = DomElement.getText(prop);
        }
        prop = this.getProperty(repository, PN_CRT_cfgNoSync);
        if (prop != null) {
          nosync = "true".equals(DomElement.getText(prop).toString());
        }
      }
      else if (REP_CLASS_SFS.equals(configClass)) {
        mode = "sfs";
        prop = this.getProperty(repository, PN_CRT_cfgLocalRoot);
        if (prop != null) {
          localroot = DomElement.getText(prop);
        }
      }
      
      if (localroot != null && localroot.length() > 0) {
        // get the basename of the localpart as that should be the directory to create
        // all rest resources in
        File tmp = new File(getLocalpart());
        return FSBasedRepository.getInstance(tmp.getName(), mode, localroot, versionroot, nosync);
      }
    }
    return null;
  }
  
  private boolean removeDBFSContent(String path) throws Exception {
    FSBasedRepository fs = FS;
    if (fs == null) {
      fs = initFS();
    }
    String uri = getLocalpart() + "/" + path;
    Element prop = this.getProperty(uri, Consts.N_getetag);
    assertTrue("etag property not found for: " + uri, prop != null);
    String etag = DomElement.getText(prop);
    int len = etag.length();
    if (len < 3 || '\"' != etag.charAt(0) || '\"' != etag.charAt(len-1)) {
      fail("unexpected etag format: " + etag);
    }
    etag = etag.substring(1, len - 1);
    long contentid = Long.parseLong(etag);
    File content = createDBFSName(fs.basecoll, contentid);
    assertTrue("content does not exist: uri="+uri+" path="+content, content.exists()); 
    return content.delete();    
  }

  // Dir structure used by DBFS
  private final static int[] DIR_STRUCT = {100, 100};
  
  private File createDBFSName(File parent, long id) {
    StringBuffer sb = new StringBuffer(128);
    sb.append(parent.getPath());
    int len = sb.length();
    if (len > 0 && sb.charAt(len - 1) == '/') {
      sb.delete(len - 1, len);
    }
    sb.append("/content");

    long tmpId = id;
    for (int i = 0, n = DIR_STRUCT.length; i < n; i++) {
      int radix = DIR_STRUCT[i];
      sb.append('/').append(tmpId % radix);
      tmpId = tmpId / radix;
    }

    sb.append('/').append(id).append(".CONTENT");
    return new File(sb.toString());
  }

  private String assureEmptyDBFSContentCollection() throws Exception {
    FSBasedRepository fs = FS;
    if (fs == null) {
      fs = initFS();
    }
    String candidate = "0";
    File candir = new File(fs.basecoll, candidate);
    if (candir.exists()) {
      if (isAlmostEmpty(candir) && hasNumericalNames(candir, 0, 100)) {
        return candidate;
      }
      
      // OK, top level 0 collection is not empty. Try find a suitable subdir
      for (int i = 0; i < 100; ++i) {
        candidate = "0/" + i;
        candir = new File(fs.basecoll, candidate);
        if (candir.exists()) {
          if (candir.list().length == 0) {
            // should also be reported
            return candidate;
          }
        }
        else {
          // create it and return
          assertTrue("unable to create empty dir: " +candir, candir.mkdirs());
          return candidate;        
        }
      }
    }
    else {
      assertTrue("unable to create empty dir: " +candir, candir.mkdirs());
      return candidate;        
    }
    
    assertTrue("unable to find/create empty dir in dbfs content: " + fs.basecoll, false);
    return null;
  }

  /**
   * Determine if filename, interpreted as number, sits in interval [lower, upper]
   * @param file file's name to treat as number
   * @param lower bound of interval (inclusive)
   * @param upper bound of interval (exclusive)
   * @return if text as number is in interval, false otherwise
   */
  private boolean hasNumericalNames(File file, int lower, int upper) {
    try {
      int i = Integer.parseInt(file.getName());
      if ((i < lower) || (i >= upper))
        return false;
      else if (file.isDirectory()) {
        String[] kids = file.list(); 
        for(int j = 0; j < kids.length; j++) {
          int k = Integer.parseInt(kids[j]);
          if ((k < lower) || (k >= upper))
            return false;
        }
      }
      return true;
    }
    catch (NumberFormatException ex) {
      return false;
    }
  }
  
  private boolean isAlmostEmpty(File dir) {
    File[] kids = dir.listFiles();
    if (kids != null && kids.length > 0) {
      //Hmm, it is not empty...
      for (int i = 0; i < kids.length; ++i) {
        if ((!kids[i].isDirectory()) || (!isAlmostEmpty(kids[i]))) {
          return false;
        }
      }
    }
    return true;
  }

  private void assureDBFSContentCollectionGone(String path) throws Exception {
    FSBasedRepository fs = FS;
    if (fs == null) {
      fs = initFS();
    }
    File candir = new File(fs.basecoll, path);
    assertTrue("content collection still exists: " +candir, !candir.exists());
  }

  private String assureLeftoverTmpFile() throws Exception {
    FSBasedRepository fs = FS;
    if (fs == null) {
      fs = initFS();
    }
    String tmppath = "tmp/";
    File tmp = null;
    if ("cmdbfs".equals(fs.mode)) {
      tmppath += "dbfs0000TEST.tmp";
      tmp = new File(fs.basecoll, tmppath);
    }
    else if ("cmfsdb".equals(fs.mode)) {
      tmppath = "cmfs0000TEST.tmp";
      tmp = new File(fs.versioncoll, tmppath);
    }
    else {
      assertTrue("unknown CM repository mode: " + fs.mode, false);
    }
    
    if ( ! tmp.exists()) {
      tmp.createNewFile();
    }
    tmp.setLastModified(System.currentTimeMillis() - (7 * 24 * 60 * 60 * 1000L));
    
    return tmppath;
  }

  private void assureTmpFileGone(String path) throws Exception {
    FSBasedRepository fs = FS;
    if (fs == null) {
      fs = initFS();
    }
    String tmppath = "tmp/";
    File tmp = null;
    if ("cmdbfs".equals(fs.mode)) {
      tmppath += path;
      tmp = new File(fs.basecoll, tmppath);
    }
    else if ("cmfsdb".equals(fs.mode)) {
      tmppath = path;
      tmp = new File(fs.versioncoll, tmppath);
    }
    else {
      assertTrue("unknown CM repository mode: " + fs.mode, false);
    }
    
    assertTrue("tmp file still exists: "+tmp, !tmp.exists());
  }

  private String assureLeftoverVersion() throws Exception {
    FSBasedRepository fs = FS;
    if (fs == null) {
      fs = initFS();
    }
    String tmppath = "versions/0/0/00000.noversion";
    File tmp = null;
    if ("cmfsdb".equals(fs.mode)) {
      tmp = new File(fs.versioncoll, tmppath);
    }
    else {
      assertTrue("unsuuported CM repository mode: " + fs.mode, false);
    }
    
    if ( ! tmp.exists()) {
      File pdir = tmp.getParentFile();
      if ( ! pdir.exists() ) {
        pdir.mkdirs();
      }
      tmp.createNewFile();
    }
    tmp.setLastModified(System.currentTimeMillis() - (7 * 24 * 60 * 60 * 1000L));
    
    return tmppath;
  }

  private void removeLeftoverVersion(String path) throws Exception {
    FSBasedRepository fs = FS;
    if (fs == null) {
      fs = initFS();
    }
    File tmp = null;
    if ("cmfsdb".equals(fs.mode)) {
      tmp = new File(fs.versioncoll, path);
    }
    else {
      assertTrue("unsupported CM repository mode: " + fs.mode, false);
    }
    
    if ( tmp.exists() ) {
      tmp.delete();
    }
  }
}
