package com.sapportals.wcm.protocol.webdav.server.tests;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;

import com.sapportals.wcm.util.http.IResponse;

/**
 * Various tests for GET method.
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author martin.boettcher@greenbytes.de
 * @author julian.reschke@greenbytes.de
 * @version $Id: GetTest.java,v 1.8 2005/01/24 12:49:45 jre Exp $
 */

public class GetTest extends TestMember {

  private static final String[] desc = new String[] {
    "00 (unused)",
    "01 GET with collection",
    "02 GET with unknown file",
    "03 GET for Worddatei.doc",
    "04 GET HEAD / test for vary header",
    "05 GET Range bytes=0-19 from ANSI.TXT",
    "06 GET Range bytes=10- from ANSI.TXT",
    "07 GET Range bytes=-10 from ANSI.TXT",
    "08 GET if-none-match etag Worddatei.doc",
    "09 GET if-modified-since Worddatei.doc",
    "10 GET Range bytes=0-65535 from ANSI.TXT",
    "11 GET Range bytes=100-65535 from ANSI.TXT",
    "12 GET with Expect 100-continue",
    "13 GET with Expect must-fail (disabled)",
    "14 GET if-none-match etag without Translate Worddatei.doc",
    "15 GET with trailing blanks",
    "16 GET UTF8.TXT - ensure charset parameter present on content-type",
    "17 GET if-unmodified-since 1900 Worddatei.doc",
    "18 GET if-modified-since 2100 Worddatei.doc",
  };

  public static int NrOfTests = desc.length;
  
  public GetTest(Integer nr) {
    super(nr, desc[nr.intValue()]);
  }

  private static String ETagForWorddatei;
  private static String LastModifiedForWorddatei;

  protected void runTest()
  {
    if (nr == 0) super.logIdentifier("GET");

    try {
      IResponse response;
      String content = null;

      switch (nr) {
        
        case 0:
          break;

        case 1:
          clear(request);
          request.setMethod("GET");
          request.setURI(getLocalpart() + "/voll");
          response = requester.perform(request);
          checkResponseCode(response, 200);
          break;

        case 2:
          clear(request);
          request.setMethod("GET");
          request.setURI(getLocalpart() + "/abc.xyz");
          response = requester.perform(request);
          checkResponseCode(response, 404);
          break;

        case 3:
          clear(request);
          request.setMethod("GET");
          request.setURI(getLocalpart() + "/Worddatei.doc");
          response = requester.perform(request);
          checkResponseCode(response, 200);
          break;

        case 4:
          clear(request);
          request.setMethod("HEAD");
          request.setURI(getLocalpart() + "/Worddatei.doc");
          request.setHeader("Translate", "f");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          ETagForWorddatei = response.getHeader("ETag");
          LastModifiedForWorddatei = response.getHeader("Last-Modified");

          // only test if response isn't HTTP/1.0 (in which the servlet
          // won't send the Vary header
          if(! "HTTP/1.0".equals(response.getHttpVersion())) {
            String varyHeader = response.getHeader("Vary");
            assertTrue("vary header not present", varyHeader != null);
            java.util.StringTokenizer st = new java.util.StringTokenizer(varyHeader, " ,");
            int found = 0;
            while (st.hasMoreTokens()) {
              String t = st.nextToken();
              if ("cookie".equalsIgnoreCase(t) ||
                  "translate".equalsIgnoreCase(t) ||
                  "accept-encoding".equalsIgnoreCase(t) ||
                  "accept-language".equalsIgnoreCase(t) ||
                  "user-agent".equalsIgnoreCase(t)) {
                found++;
              }
            }
            assertTrue("header missing from: " + varyHeader, found == 5);
          }
          break;

        case 5:
          clear(request);
          request.setMethod("GET");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          request.setHeader("Range", "bytes=0-19");
          request.expectsResponseStream(true);
          response = requester.perform(request);
          checkResponseCode(response, 206);
          content = getContent(response.getStream());
          assertTrue ("unexpected content: "+content, content.equals("Dies ist eine ANSI-T"));
          checkRangeHeader(response, "bytes 0-19/29");
          break;

        case 6:
          clear(request);
          request.setMethod("GET");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          request.setHeader("Range", "bytes=10-");
          request.expectsResponseStream(true);
          response = requester.perform(request);
          checkResponseCode(response, 206);
          content = getContent(response.getStream());
          assertTrue ("unexpected content: "+content, content.equals("ine ANSI-Textdatei."));
          checkRangeHeader(response, "bytes 10-28/29");
          break;

        case 7:
          clear(request);
          request.setMethod("GET");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          request.setHeader("Range", "bytes=-10");
          request.setHeader("Translate", "F");
          request.expectsResponseStream(true);
          response = requester.perform(request);
          checkResponseCode(response, 206);
          content = getContent(response.getStream());
          assertTrue ("unexpected content: "+content, content.equals("Textdatei."));
          checkRangeHeader(response, "bytes 19-28/29");
          break;

        case 8:
          clear(request);
          request.setMethod("GET");
          request.setURI(getLocalpart() + "/Worddatei.doc");
          request.setHeader("Translate", "F");
          request.setHeader("if-none-match", ETagForWorddatei);
          request.expectsResponseStream(true);
          response = requester.perform(request);
          checkResponseCode(response, 304);
          content = getContent(response.getStream());
          assertTrue(content.length() == 0);
          break;

        case 9:
          clear(request);
          request.setMethod("GET");
          request.setURI(getLocalpart() + "/Worddatei.doc");
          request.setHeader("Translate", "F");
          request.setHeader("if-modified-since", LastModifiedForWorddatei);
          request.expectsResponseStream(true);
          response = requester.perform(request);
          
          checkResponseCode(response, 304);
          content = getContent(response.getStream());
          assertTrue(content.length() == 0);
          break;

        case 10:
          clear(request);
          request.setMethod("GET");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          request.setHeader("Range", "bytes=0-65535");
          request.setHeader("Translate", "F");
          request.expectsResponseStream(true);
          response = requester.perform(request);
          checkResponseCode(response, 206);
          content = getContent(response.getStream());
          assertTrue ("unexpected content: "+content, content.equals("Dies ist eine ANSI-Textdatei."));
          checkRangeHeader(response, "bytes 0-28/29");
          break;

        case 11:
          clear(request);
          request.setMethod("GET");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          request.setHeader("Range", "bytes=100-65535");
          request.setHeader("Translate", "F");
          response = requester.perform(request);
          checkResponseCode(response, 416);
          break;

        case 12:
          clear(request);
          request.setMethod("GET");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          request.setHeader("Expect", "100-continue");
          request.setHeader("Translate", "F");
          request.setBody("this is a test body which should only be send after an 100");
          request.setContentType("text/plain");
          request.expectsResponseStream(false);
          response = requester.perform(request);
          checkResponseCode(response, 200); // J2EE handles 100-continue
          break;

        case 13:
          // this test fails with the J2EE because it doesn't reject unknown
          // Expect headers. Moved to SE tests.
//          clear(request);
//          request.setMethod("GET");
//          request.setURI(getLocalpart() + "/ANSI.TXT");
//          request.setHeader("Expect", "unknown-greenbytes-feature");
//          request.setHeader("Translate", "F");
//          request.expectsResponseStream(false);
//          response = requester.perform(request);
//          checkResponseCode(response, 417);
          break;

        case 14:
          clear(request);
          request.setMethod("GET");
          request.setURI(getLocalpart() + "/Worddatei.doc");
          request.setHeader("if-none-match", ETagForWorddatei);
          request.expectsResponseStream(true);
          response = requester.perform(request);
          // we allow 200 here because Translate t/f may make a difference for the etag (which we obtained with Translate:t
          checkResponseCode(response, 200, 304);
          if (response.getStatus() == 304) {
            //if status is "not modified", ensure we didn't get any content
            content = getContent(response.getStream());
            assertTrue(content.length() == 0);
          }
          break;

        case 15:
          String uri = getLocalpart() + "/ANSI.TXT";
          clear(request);
          request.setMethod("GET");
          request.setURI(uri + "%20");
          request.setHeader("Translate", "f");
          request.expectsResponseStream(true);
          response = requester.perform(request);
          checkResponseCode(response, 404, 200);

          if (response.getStatus() == 200) {
            String contLocation = response.getHeader("Content-Location");
            assertTrue("Content-Location header is missing", contLocation != null);
            assertTrue(
                "Content-Location header has wrong value: " + contLocation + ", should be: " + this.getBase() + uri,
                contLocation.equals(this.getBase() + uri));
          }

          break;
          
        case 16:
          clear(request);
          request.setMethod("GET");
          request.setURI(getLocalpart() + "/UTF8.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 200);
          String ctype = response.getHeader("content-type");
          assertTrue(ctype.toLowerCase().indexOf("utf-8") >= 0);
          break;

        case 17:
          clear(request);
          request.setMethod("GET");
          request.setURI(getLocalpart() + "/Worddatei.doc");
          request.setHeader("Translate", "F");
          request.setHeader("if-unmodified-since", "Mon, 01 Jan 1900 01:23:45 GMT");
          request.expectsResponseStream(true);
          response = requester.perform(request);
          
          checkResponseCode(response, 412);
          content = getContent(response.getStream());
          break;

        case 18:
          clear(request);
          request.setMethod("GET");
          request.setURI(getLocalpart() + "/Worddatei.doc");
          request.setHeader("Translate", "F");
          request.setHeader("if-modified-since", "Fri, 01 Jan 2100 01:23:45 GMT");
          request.expectsResponseStream(true);
          response = requester.perform(request);
          
          checkResponseCode(response, 304);
          content = getContent(response.getStream());
          assertTrue(content.length() == 0);
          break;
      }
    }
    catch (Exception ex) {
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }

  private void checkRangeHeader(IResponse response, String expected)
  throws Exception {
    String header = response.getHeader("Content-Range");
    assertTrue("Content-Range unexpected: "+header+", expected "+expected,
               expected.equals(header));

  }

  private String getContent(InputStream is) throws IOException {
    if (is == null)
      return null;

    byte[] buffer = new byte[64 * 1024];
    ByteArrayOutputStream bos = new ByteArrayOutputStream();
    int len;
    while ((len = is.read(buffer)) != -1) {
      bos.write(buffer, 0, len);
    }
    is.close();

    return bos.toString();
  }
}
