package com.sapportals.wcm.protocol.webdav.server.tests;

import java.io.ByteArrayInputStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.w3c.dom.Document;

import com.sapportals.wcm.WcmException;
import com.sapportals.wcm.protocol.webdav.DavConditions;
import com.sapportals.wcm.protocol.webdav.DomSerializer;
import com.sapportals.wcm.util.http.IRequester;
import com.sapportals.wcm.util.http.IResponse;


/**
 * Various tests for WEBDAV <a href="http://greenbytes.de/tech/webdav/rfc2518.html#METHOD_LOCK"
 * >LOCK</a>  method.
 *
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author martin.boettcher@greenbytes.de
 * @author stefan.eissing@greenbytes.de
 * @author julian.reschke@greenbytes.de
 * @version $Id: LockTest.java,v 1.9 2004/09/13 12:54:45 jre Exp $
 */

public class LockTest extends TestMember {

  private static final String[] desc = new String[]
  {
    "00 LO LOCK with empty Timeout no body",
    "01 LO LOCK with empty Timeout depth='1'",
    "02 LO LOCK with empty Timeout depth='infinity'",
    "03 LO UNLOCK deep Lock",
    "04 LO LOCK non existing resource lock.null",
    "05 LO PROPFIND null-resource lock.null",
    "06 LO PROPPATCH null resource no if lock.null",
    "07 LO PROPPATCH null resource wrong if header lock.null",
    "08 LO LOCK refresh null-resource lock.null",
    "09 LO DELETE null-resource lock.null",
    "10 LO LOCK with empty Timeout depth='0'",
    "11 LO LOCK again",
    "12 LO LOCK already locked collection-entry",
    "13 LO DELETE locked file",
    "14 LO PUT a locked file",
    "15 LO PUT a locked file with if header",
    "16 LO PUT a locked file with wrong header",
    "17 LO UNLOCK just locked file",
    "18 LO DELETE locked collection member",
    "19 LO PROPPATCH file of locked collection no if",
    "20 LO PROPPATCH locked collection with if header",
    "21 LO PROPPATCH locked collection no if",
    "22 LO PROPPATCH locked collection with wrong if header",
    "23 LO LOCK with 3 seconds Timeout copied ANSIX.TXT",
    "24 LO LOCK again on ANSIX.TXT",
    "25 LO LOCK refresh ANSIX.TXT (co content-length header)",
    "26 LO DELETE locked file ANSIX.TXT",
    "27 LO DELETE locked file ANSIX.TXT",
    "28 LO DELETE expired null-resource lock.null",
    "29 LO LOCK with infinite Timeout",
    "30 LO UNLOCK unlocked file no locktoken",
    "31 LO UNLOCK locked file no locktoken",
    "32 LO UNLOCK locked collection no locktoken",
    "33 LO UNLOCK unlocked file wrong token",
    "34 LO UNLOCK locked file wrong token",
    "35 LO UNLOCK locked collection wrong token",
    "36 LO UNLOCK unlocked file",
    "37 LO UNLOCK locked file",
    "38 LO UNLOCK locked file 2",
    "39 LO UNLOCK locked collection",
    "40 LO UNLOCK locked collection 2",
    "41 LO UNLOCK expired ANSIX.TXT",
    "42 LO UNLOCK expired null-resource lock.null",
    "43 LO DELETE expired null-resource with If-Header lock.null",
    "44 LO DELETE expired null-resource lock.null",
    "45 LO PUT test resource",
    "46 LO LOCK test resource",
    "47 LO DELETE locked test resource",
    "48 LO LOCK/DELETE where DAV:owner is logged in user",
    "49 LO LOCK/DELETE where DAV:owner contains arbitrary XML",
    "50 LO MOVE test resource",
    "51 LO LOCK/UNLOCK different users",
    "52 LO rename collection, lock on child should go away",
    "53 LO check lock cache invalidation",
    "   LO cleanup",
  };

  public static int NrOfTests = desc.length;

  public LockTest(Integer nr) {
    super(nr, desc[nr.intValue()]);
  }

  private static String lockTokenFromTest2;
  private static String lockNullToken;
  private static String lockTokenFromTest10;
  private static String lockTokenFromTest12;
  private static String ansixToken;
  private static String unicodeLock;
  private static IRequester requester2 = null;

  private static final String TIMEOUT = "Second-2";

  private static List cleanupList = new ArrayList();

  private void lockNull(String uri, String file) throws WcmException {
    final String LOCKIT = "<lockinfo xmlns=\"DAV:\"><lockscope>"
        + "<exclusive /></lockscope><locktype><write /></locktype></lockinfo>";

    clear(request);
    request.setMethod("LOCK");
    request.setURI(getLocalpart() + uri);
    request.setHeader("Depth", "0");
    request.setHeader("Timeout", TIMEOUT);
    request.setHeader("Content-Type", "text/xml");
    request.setBody(LOCKIT);
    request.expectsResponseDocument(true);
    IResponse response = requester.perform(request);
    checkResponseCode(response, 200, 201);
    checkResponseDocument(response, file, null);
    lockNullToken = getLockToken(response);
  }

  public static String makeLockInfo(String owner) {
    String res = "<lockinfo xmlns='DAV:'>";

    if (owner != null) {
      res += "<owner>" + owner + "</owner>";
    }

    res += "<lockscope><exclusive /></lockscope><locktype><write /></locktype></lockinfo>";
    return res;
  }

  protected void runTest() {
    if (nr == 0) logIdentifier("LO");

    String respfile = "LOCK-" + TestMember.THREEDIGITS.format(nr) + ".txt";

    final String WRONGPROPERTYUPDATE = "<propertyupdate xmlns='DAV:'><set>"
        + "<prop><yy xmlns='xx:'>Not allowed!</yy></prop></set></propertyupdate>";

    try {
      IResponse response;
      clear(request);
      request.setMethod("LOCK");

      if (requester2 == null) {
        requester2 = getSecondRequester();
      }

      switch (nr) {
        case 0:
          // setup
          clear(request);
          request.setMethod("COPY");
          request.setURI(getLocalpart() + "/voll/ist%20was%20drin/office.zip");
          request.setHeader("Destination", getBaseAndLocalpart() + "/office.backup");
          response = requester.perform(request);
          checkResponseCode(response, 201, 200);

          clear(request);
          request.setMethod("LOCK");
          request.setURI(getLocalpart() + "/leer");
          response = requester.perform(request);
          checkResponseCode(response, 400);
          break;

        case 1:
          request.setHeader("Depth", "1");
          request.setURI(getLocalpart() + "/leer");
          request.setHeader("Content-Type", "text/xml");
          request.setBody(makeLockInfo(null));
          response = requester.perform(request);
          checkResponseCode(response, 400);
          break;

        case 2:
          request.setHeader("Depth", "infinity");
          request.setURI(getLocalpart() + "/leer");
          request.setHeader("Content-Type", "text/xml");
          request.setBody(makeLockInfo(null));
          request.expectsResponseDocument(true);
          response = requester.perform(request);
          checkResponseCode(response, 501, 200);
          if (response.getStatus() == 200) {
            lockTokenFromTest2 = getLockToken(response);
          }
          else {
            lockTokenFromTest2 = null;
          }
          break;

        case 3:
          if (lockTokenFromTest2 != null) {
            request.setMethod("UNLOCK");
            request.setURI(getLocalpart() + "/leer");
            request.setHeader("Lock-Token", "<" + lockTokenFromTest2 + ">");
            response = requester.perform(request);
            checkResponseCode(response, 204);
          }
          break;

        case 4:
          lockNull("/leer/lock.null", respfile);
          assertTrue("locktoken null", lockNullToken != null);
          break;

        case 5:
          request.setMethod("PROPFIND");
          request.setHeader("Depth", "0");
          request.setURI(getLocalpart() + "/leer/lock.null");
          checkMultistatus("<propfind xmlns='DAV:'><prop><resourcetype/><lockdiscovery/></prop></propfind>", respfile);
          break;

        case 6:
          request.setMethod("PROPPATCH");
          request.setHeader("Content-Type", "text/xml");
          request.setBody(WRONGPROPERTYUPDATE);
          request.setURI(getLocalpart() + "/leer/lock.null");
          response = requester.perform(request);
          checkResponseCode(response, 423);
          break;

        case 7:
          request.setMethod("PROPPATCH");
          request.setHeader("Content-Type", "text/xml");
          request.setBody(WRONGPROPERTYUPDATE);
          request.setHeader("If", "(<" + lockTokenFromTest12 + ">)");
          request.setURI(getLocalpart() + "/leer/lock.null");
          response = requester.perform(request);
          checkResponseCode(response, 412);
          break;

        case 8:
          request.setURI(getLocalpart() + "/leer/lock.null");
          request.setMethod("PROPFIND");
          request.expectsResponseDocument(true);
          response = requester.perform(request);
          Document doc = response.getDocument();
          request.setMethod("LOCK");
          request.setHeader("If", "(<" + lockNullToken + ">)");
          request.setHeader("Timeout", TIMEOUT);
          response = requester.perform(request);
          int status = response.getStatus();
          if (status != 200) {
            assertTrue("response code: " + status + " token "
                       + lockNullToken + "\n PROPFIND was " +
                       ((doc != null)? DomSerializer.toString(doc, true) : "null"),
                       false);
          }
          break;

        case 9:
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/leer/lock.null");
          response = requester.perform(request);
          checkResponseCode(response, 423);
          break;

        case 10:
          request.setHeader("Depth", "0");
          request.setURI(getLocalpart() + "/voll/ist%20was%20drin");
          request.setHeader("Content-Type", "text/xml");
          request.setBody(makeLockInfo(null));
          request.expectsResponseDocument(true);
          response = requester.perform(request);
          checkResponseCode(response, 200);
          lockTokenFromTest10 = getLockToken(response);
          assertTrue("Lock-Token header doesn't match token from response body",
                     ("<" + lockTokenFromTest10 + ">").equals(response.getHeader("Lock-Token")));
          break;

        case 11:
          request.setHeader("Depth", "0");
          request.setURI(getLocalpart() + "/voll/ist%20was%20drin");
          request.setHeader("Content-Type", "text/xml");
          request.setBody(makeLockInfo(null));
          response = requester.perform(request);
          checkResponseCode(response, 423);
          break;

        case 12:
          request.setHeader("Depth", "0");
          request.setURI(getLocalpart() + "/voll/ist%20was%20drin/office.zip");
          request.setHeader("Content-Type", "text/xml");
          request.setBody(makeLockInfo(null));
          request.expectsResponseDocument(true);
          response = requester.perform(request);
          checkResponseCode(response, 200);
          lockTokenFromTest12 = getLockToken(response);
          break;

        case 13:
          assertTrue("lock not set in previous test", lockTokenFromTest12 != null);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/voll/ist%20was%20drin/office.zip");
          response = requester.perform(request);
          checkResponseCode(response, 423);
          break;

        case 14:
          assertTrue("lock not set in previous test", lockTokenFromTest12 != null);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/voll/ist%20was%20drin/office.zip");
          response = requester.perform(request);
          checkResponseCode(response, 423);
          break;

        case 15: {
            assertTrue("lock not set in previous test", lockTokenFromTest12 != null);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + "/voll/ist%20was%20drin/office.zip");
            request.setHeader("If", "(<" + lockTokenFromTest12 + ">)");
            response = requester.perform(request);
            checkResponseCode(response, 200);
            String etag = response.getHeader("etag");

            if (etag == null) {
              // get the etag, then try another additional PUT
              clear(request);
              request.setMethod("HEAD");
              request.setHeader("Translate", "f");
              request.setURI(getLocalpart() + "/voll/ist%20was%20drin/office.zip");
              response = requester.perform(request);
              checkResponseCode(response, 200);
              etag = response.getHeader("etag");
            }

            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + "/voll/ist%20was%20drin/office.zip");
            request.setHeader("If", "(<" + lockTokenFromTest12 + "> ["+etag+"])");
            request.setHeader("Translate", "f");
            response = requester.perform(request);
            checkResponseCode(response, 200);
          }
          break;

        case 16: {
            assertTrue("lock not set in previous test", lockTokenFromTest12 != null);

            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + "/voll/ist%20was%20drin/office.zip");
            request.setHeader("If", "(<" + lockTokenFromTest10 + ">)");
            response = requester.perform(request);
            checkResponseCode(response, 412);

            // add "or" with lock token test that evaluates to true
            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + "/voll/ist%20was%20drin/office.zip");
            request.setHeader("If", "(<" + lockTokenFromTest10 + ">) (Not <DAV:no-lock>)");
            response = requester.perform(request);
            checkResponseCode(response, 423);
          }
          break;

        case 17:
          assertTrue("lock not set in previous test", lockTokenFromTest12 != null);
          request.setMethod("UNLOCK");
          request.setURI(getLocalpart() + "/voll/ist%20was%20drin/office.zip");
          request.setHeader("Lock-Token", "<" + lockTokenFromTest12 + ">");
          response = requester.perform(request);
          checkResponseCode(response, 204);
          break;

        case 18:
          if (lockTokenFromTest10 != null) {
            request.setMethod("DELETE");
            request.setURI(getLocalpart() + "/voll/ist%20was%20drin/office.zip");
            checkMultistatus(null, "LOCK-014.txt");
          }
          break;

        case 19:
          request.setMethod("PROPPATCH");
          request.setURI(getLocalpart() + "/voll/ist%20was%20drin/office.zip");
          checkMultistatus(WRONGPROPERTYUPDATE, "LOCK-015.txt");
          break;

        case 20:
          // PROPPATCH on locked collection with lock token
          if (lockTokenFromTest10 != null) {
            request.setMethod("PROPPATCH");
            request.setURI(getLocalpart() + "/voll/ist%20was%20drin");
            request.setHeader("If", "(<" + lockTokenFromTest10 + ">)");
            checkMultistatus(WRONGPROPERTYUPDATE, respfile);
          }
          break;

        case 21:
          // PROPPATCH on locked collection without lock token
          if (lockTokenFromTest10 != null) {
            request.setMethod("PROPPATCH");
            request.setHeader("Content-Type", "text/xml");
            request.setBody(WRONGPROPERTYUPDATE);
            request.setURI(getLocalpart() + "/voll/ist%20was%20drin");
            response = requester.perform(request);
            checkResponseCode(response, 423);
          }
          break;

        case 22:
          request.setMethod("PROPPATCH");
          request.setHeader("Content-Type", "text/xml");
          request.setBody(WRONGPROPERTYUPDATE);
          request.setHeader("If", "(<" + lockTokenFromTest12 + ">)");
          request.setURI(getLocalpart() + "/voll/ist%20was%20drin");
          response = requester.perform(request);
          checkResponseCode(response, 412);
          break;

        case 23:
          request.setMethod("COPY");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          request.setHeader("Destination", this.getBaseAndLocalpart() + "/ANSIX.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("LOCK");
          request.setHeader("Timeout", TIMEOUT);
          request.setURI(getLocalpart() + "/ANSIX.TXT");
          request.setHeader("Content-Type", "text/xml");
          request.setBody(makeLockInfo(null));
          request.expectsResponseDocument(true);
          response = requester.perform(request);
          checkResponseCode(response, 200);
          ansixToken = getLockToken(response);
          break;

        case 24:
          request.setMethod("LOCK");
          request.setHeader("Timeout", TIMEOUT);
          request.setURI(getLocalpart() + "/ANSIX.TXT");
          request.setHeader("Content-Type", "text/xml");
          request.setBody(makeLockInfo(null));
          request.expectsResponseDocument(true);
          response = requester.perform(request);
          checkResponseCode(response, 423);
          break;

        case 25:
          request.setMethod("LOCK");
          request.setHeader("If", "(<" + ansixToken + ">)");
          request.setURI(getLocalpart() + "/ANSIX.TXT");
          request.setHeader("Timeout", TIMEOUT);
          request.setBody(new ByteArrayInputStream(new byte[0])); // no content-length header, please
          response = requester.perform(request);
          checkResponseCode(response, 200);
          break;

        case 26:
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/ANSIX.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 423);
          break;

        case 27:
          java.lang.Thread.sleep(2100);  // wait for lock to time out
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/ANSIX.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 204);
          break;

        case 28:
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/leer/lock.null");
          response = requester.perform(request);
          checkResponseCode(response, 404, 204, 423);
          break;

        case 29:
          request.setHeader("Timeout", "Infinite, Second-12345");
          request.setHeader("Depth", "0");
          request.setURI(getLocalpart() + "/Unicode.TXT");
          request.setHeader("Content-Type", "text/xml");
          request.setBody(makeLockInfo(null));
          request.expectsResponseDocument(true);
          response = requester.perform(request);
          checkResponseCode(response, 200);
          unicodeLock = getLockToken(response);
          break;

        case 30:
          clear(request);
          request.setMethod("UNLOCK");
          request.setURI(getLocalpart() + "/Worddatei.doc");
          response = requester.perform(request);
          checkResponseCode(response, 400);
          checkErrorElement(response, DavConditions.N_SAP_lock_token_present);
          break;

        case 31:
          clear(request);
          request.setMethod("UNLOCK");
          request.setURI(getLocalpart() + "/Unicode.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 400);
          checkErrorElement(response, DavConditions.N_SAP_lock_token_present);
          break;

        case 32:
          clear(request);
          request.setMethod("UNLOCK");
          request.setURI(getLocalpart() + "/voll/ist%20was%20drin");
          response = requester.perform(request);
          checkResponseCode(response, 400);
          checkErrorElement(response, DavConditions.N_SAP_lock_token_present);
          break;

        case 33:
          if (lockTokenFromTest10 != null) {
            clear(request);
            request.setMethod("UNLOCK");
            request.setURI(getLocalpart() + "/Worddatei.doc");
            request.setHeader("Lock-Token", "<" + lockTokenFromTest10 + ">");
            response = requester.perform(request);
            checkResponseCode(response, 403);
            checkErrorElement(response, DavConditions.N_SAP_lock_token_matches);
          }
          break;

        case 34:
          if (lockTokenFromTest10 != null) {
            clear(request);
            request.setMethod("UNLOCK");
            request.setURI(getLocalpart() + "/Unicode.TXT");
            request.setHeader("Lock-Token", "<" + lockTokenFromTest10 + ">");
            response = requester.perform(request);
            checkResponseCode(response, 403);
            checkErrorElement(response, DavConditions.N_SAP_lock_token_matches);
          }
          break;

        case 35: {
            clear(request);
            request.setMethod("UNLOCK");
            request.setURI(getLocalpart() + "/voll/ist%20was%20drin");
            request.setHeader("Lock-Token", "<" + unicodeLock + ">");
            response = requester.perform(request);
            checkResponseCode(response, 403);
            checkErrorElement(response, DavConditions.N_SAP_lock_token_matches);
          }
          break;

        case 36: {
            clear(request);
            request.setMethod("UNLOCK");
            request.setURI(getLocalpart() + "/Worddatei.doc");
            request.setHeader("Lock-Token", "<" + unicodeLock + ">");
            response = requester.perform(request);
            checkResponseCode(response, 403);
            checkErrorElement(response, DavConditions.N_SAP_lock_token_matches);
          }
          break;

        case 37: {
            clear(request);
            request.setMethod("UNLOCK");
            request.setURI(getLocalpart() + "/Unicode.TXT");
            request.setHeader("Lock-Token", "<" + unicodeLock + ">");
            response = requester.perform(request);
            checkResponseCode(response, 204);
          }
          break;

        case 38: {
            clear(request);
            request.setMethod("UNLOCK");
            request.setURI(getLocalpart() + "/Unicode.TXT");
            request.setHeader("Lock-Token", "<" + unicodeLock + ">");
            response = requester.perform(request);
            checkResponseCode(response, 403);
            checkErrorElement(response, DavConditions.N_SAP_lock_token_matches);
          }
          break;

        case 39:
          if (lockTokenFromTest10 != null) {
            clear(request);
            request.setMethod("UNLOCK");
            request.setURI(getLocalpart() + "/voll/ist%20was%20drin");
            request.setHeader("Lock-Token", "<" + lockTokenFromTest10 + ">");
            response = requester.perform(request);
            checkResponseCode(response, 204);
          }
          break;

        case 40:
          if (lockTokenFromTest10 != null) {
            clear(request);
            request.setMethod("UNLOCK");
            request.setURI(getLocalpart() + "/voll/ist%20was%20drin");
            request.setHeader("Lock-Token", "<" + lockTokenFromTest10 + ">");
            request.setContentLength(0);
            response = requester.perform(request);
            checkResponseCode(response, 403);
            checkErrorElement(response, DavConditions.N_SAP_lock_token_matches);
          }
          break;

        case 41:
          request.setMethod("UNLOCK");
          request.setURI(getLocalpart() + "/ANSIX.TXT");
          request.setHeader("Lock-Token", "<" + ansixToken + ">");
          request.setContentLength(0);
          response = requester.perform(request);
          checkResponseCode(response, 404, 403);

          if (response.getStatus() == 403) {
            clear(request);
            request.setMethod("DELETE");
            request.setURI(getLocalpart() + "/ANSIX.TXT");
            response = requester.perform(request);
            checkResponseCode(response, 204);
          }
          break;

        case 42:
          // Check if we can unlock the expired null-resource. This
          // should either result in a 404, since the resource should
          // have vanished, or in a 204 (successful unlock), since
          // some servers keep expired locks around for some time (-> Office).
          //
          request.setMethod("UNLOCK");
          request.setURI(getLocalpart() + "/leer/lock.null");
          request.setHeader("Lock-Token", "<" + lockNullToken + ">");
          request.setContentLength(0);
          response = requester.perform(request);
          checkResponseCode(response, 404, 204, 403);
          break;

        case 43:
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/leer/lock.null");
          request.setHeader("If", "<" + getBaseAndLocalpart() + "/leer/lock.null"
                                  + ">(<" + lockNullToken + ">)");
          response = requester.perform(request);
          checkResponseCode(response, 404, 412);
          break;

        case 44:
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/leer/lock.null");
          response = requester.perform(request);
          checkResponseCode(response, 404, 412);
          break;

        case 45:
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/voll/ist%20was%20drin/unwichtig");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;

        case 46:
          request.setMethod("LOCK");
          request.setHeader("Timeout", TIMEOUT);
          request.setURI(getLocalpart() + "/voll/ist%20was%20drin/unwichtig");
          request.setHeader("Content-Type", "text/xml");
          request.setHeader("Depth", "0");
          request.setBody(makeLockInfo(null));
          request.expectsResponseDocument(true);
          response = requester.perform(request);
          checkResponseCode(response, 200);
          ansixToken = getLockToken(response);
          break;

        case 47:
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/voll/ist%20was%20drin/unwichtig");
          request.setHeader("If", "<" + getLocalpart() + "/voll/ist%20was%20drin/unwichtig"
                                  + ">(<" + ansixToken + ">)");
          response = requester.perform(request);
          checkResponseCode(response, 204);
          break;

          // lock owner specified, identical to loggon user
        case 48:
          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/voll/ist%20was%20drin/unwichtig");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("LOCK");
          request.setHeader("Depth", "0");
          request.setHeader("Timeout", "Second-10");
          request.setURI(getLocalpart() + "/voll/ist%20was%20drin/unwichtig");
          response = checkStatusAndBody(200, makeLockInfo(m_userid), respfile);
          ansixToken = getLockToken(response);

          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/voll/ist%20was%20drin/unwichtig");
          request.setHeader("If", "<" + getLocalpart() + "/voll/ist%20was%20drin/unwichtig"
                                  + ">(<" + ansixToken + ">)");
          response = requester.perform(request);
          checkResponseCode(response, 204);
          break;


          // lock owner specified, containing "arbitrary" XML
        case 49:
          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/voll/ist%20was%20drin/unwichtig");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("LOCK");
          request.setHeader("Depth", "0");
          request.setHeader("Timeout", "Second-10");
          request.setURI(getLocalpart() + "/voll/ist%20was%20drin/unwichtig");
          response = checkStatusAndBody(200, makeLockInfo("foo<embedded xmlns='xml'/>bar"), respfile);
          ansixToken = getLockToken(response);

          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/voll/ist%20was%20drin/unwichtig");
          request.setHeader("If", "<" + getLocalpart() + "/voll/ist%20was%20drin/unwichtig"
                                  + ">(<" + ansixToken + ">)");
          response = requester.perform(request);
          checkResponseCode(response, 204);
          break;

        case 50:
          {
            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + "/voll/ist%20was%20drin/unwichtig");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("LOCK");
            request.setHeader("Depth", "0");
            request.setHeader("Timeout", "Second-10");
            request.setURI(getLocalpart() + "/voll/ist%20was%20drin/unwichtig");
            request.setBody(makeLockInfo("foo<embedded xmlns='xml'/>bar"));
            response = requester.perform(request);
            checkResponseCode(response, 200);
            String lt = response.getHeader("lock-token");
            assertTrue("locktoken null", lt != null);

            clear(request);
            request.setMethod("MOVE");
            request.setURI(getLocalpart() + "/voll/ist%20was%20drin/unwichtig");
            request.setHeader("Destination", getBaseAndLocalpart() + "/123");
            request.setHeader("If", "<" + getLocalpart() + "/voll/ist%20was%20drin/unwichtig"
                                    + ">(" + lt + ")");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + "/123");
            response = requester.perform(request);
            checkResponseCode(response, 200);

            clear(request);
            request.setMethod("DELETE");
            request.setURI(getLocalpart() + "/123");
            response = requester.perform(request);
            checkResponseCode(response, 204);
          }
          break;

          // unlock with different user
        case 51:
          {
            String res = getLocalpart() + "/voll/ist%20was%20drin/unwichtig";

            clear(request);
            request.setMethod("LOCK");
            request.setHeader("Depth", "0");
            request.setHeader("Timeout", "Second-10");
            request.setURI(res);
            request.setBody(makeLockInfo(null));
            response = requester.perform(request);
            checkResponseCode(response, 201);
            String lt = getLockToken(response);

            cleanupList.add(new DelItem(res, lt));

            clear(request);
            request.setMethod("UNLOCK");
            request.setURI(res);
            request.setHeader("Lock-Token", "<"+lt+">");
            response = requester2.perform(request);
            checkResponseCode(response, 403);

            clear(request);
            request.setMethod("DELETE");
            request.setHeader("If",
              //"<" + res + ">" +
              "(<"+lt+">)");
            request.setURI(res);
            response = requester.perform(request);
            checkResponseCode(response, 204);
          }
          break;

        case 52: {
            clear(request);
            request.setMethod("MKCOL");
            request.setURI(getLocalpart() + "/lockmove");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("LOCK");
            request.setURI(getLocalpart() + "/lockmove/x");
            request.setBody(makeLockInfo(null));
            response = requester.perform(request);
            checkResponseCode(response, 201);
            String lt = getLockToken(response);

            cleanupList.add(new DelItem(getLocalpart() + "/lockmove/x", lt));
            cleanupList.add(new DelItem(getLocalpart() + "/lockmove", null));
            cleanupList.add(new DelItem(getLocalpart() + "/lockmovex/x", lt));
            cleanupList.add(new DelItem(getLocalpart() + "/lockmovex", null));

            clear(request);
            request.setMethod("MOVE");
            request.setURI(getLocalpart() + "/lockmove");
            request.setHeader("destination", getBaseAndLocalpart() + "/lockmovex");
            request.setHeader("If", "<" + getBase() + getLocalpart() + "/lockmove/x" + ">" + "(<" + lt + ">)");
            response = requester.perform(request);
            checkResponseCode(response, 201, 204);

            // make sure child resource not locked anymore
            clear(request);
            request.setMethod("DELETE");
            request.setURI(getLocalpart() + "/lockmovex/x");
            response = requester.perform(request);
            checkResponseCode(response, 201, 204);
          }
          break;

        case 53: {
          // (tries to check) that cache invalidation doesn't affect other
          // resources where the URIs prefixes match.
          
          clear(request);
          request.setMethod("MKCOL");
          request.setURI(getLocalpart() + "/lock1");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/lock1/bar");
          response = requester.perform(request);
          checkResponseCode(response, 201);
 
          clear(request);
          request.setMethod("LOCK");
          request.setURI(getLocalpart() + "/lock1");
          request.setBody(makeLockInfo(null));
          response = requester.perform(request);
          checkResponseCode(response, 200, 501);
          
          if (response.getStatus() == 501) {
            // no deep locks available
            cleanupList.add(new DelItem(getLocalpart() + "/lock1", null));
          }
          else {
            String lt = getLockToken(response);

            clear(request);
            request.setMethod("MKCOL");
            request.setURI(getLocalpart() + "/lock12");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + "/lock12/bar");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("LOCK");
            request.setURI(getLocalpart() + "/lock12/bar");
            request.setBody(makeLockInfo(null));
            response = requester.perform(request);
            checkResponseCode(response, 200);
            String lt2 = getLockToken(response);

            cleanupList.add(new DelItem(getLocalpart() + "/lock1", lt));
            cleanupList.add(new DelItem(getLocalpart() + "/lock12/bar", lt2));
            cleanupList.add(new DelItem(getLocalpart() + "/lock12", null));

            clear(request);
            request.setMethod("DELETE");
            request.setURI(getLocalpart() + "/lock1");
            request.setHeader("If", "(<" + lt + ">)");
            response = requester.perform(request);
            checkResponseCode(response, 204);
            
            // if the previous DELETE invalidated more locks than it should, the next
            // request will fail
            
            clear(request);
            request.setMethod("HEAD");
            request.setURI(getLocalpart() + "/lock12/bar");
            request.setHeader("If", "(<" + lt2 + ">)");
            response = requester.perform(request);
            checkResponseCode(response, 200);

            clear(request);
            request.setMethod("UNLOCK");
            request.setURI(getLocalpart() + "/lock12/bar");
            request.setHeader("Lock-Token", "<" + lt2 + ">");
            response = requester.perform(request);
            checkResponseCode(response, 204);

            clear(request);
            request.setMethod("DELETE");
            request.setURI(getLocalpart() + "/lock12");
            response = requester.perform(request);
            checkResponseCode(response, 204);
          }
        }
        break;

        default:
          {
            for (Iterator it = cleanupList.iterator(); it.hasNext(); /**/) {
              DelItem di = (DelItem)it.next();
              safeDelete(di.uri, di.locktoken);
            }

            clear(request);
            request.setMethod("MOVE");
            request.setURI(getLocalpart() + "/office.backup");
            request.setHeader("Destination", getBaseAndLocalpart() + "/voll/ist%20was%20drin/office.zip");
            request.setHeader("Overwrite", "T");
            response = requester.perform(request);
            checkResponseCode(response, 204);
          }
          break;
      }
    }
    catch (Exception ex) {
      ex.printStackTrace(System.err);
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }

  public class DelItem {
    String uri;
    String locktoken;

    public DelItem(String uri, String locktoken) {
      this.uri = uri;
      this.locktoken = locktoken;
    }


  }

}
