package com.sapportals.wcm.protocol.webdav.server.tests;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import com.sapportals.wcm.protocol.webdav.*;
import com.sapportals.wcm.protocol.webdav.dompath.DomPath;
import com.sapportals.wcm.util.http.IResponse;

/**
 * Various tests for MOVE method.
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author julian.reschke@greenbytes.de
 * @author martin.boettcher@greenbytes.de
 * @version $Id: MoveTest.java,v 1.6 2004/07/15 18:26:03 jre Exp $
 */

public class MoveTest extends TestMember {

  private static final String[] desc = new String[]
  {
    "00 MOVE with improper URI",
    "01 MOVE with unknown file",
    "02 MOVE with known file",
    "03 MOVE with known file to root",
    "04 MOVE with known file in root",
    "05 MOVE with known file from root",
    "06 MOVE with collection",
    "07 MOVE collection back",
    "08 MOVE file back",
    "09 MOVE do HEAD on moved resource",
    "10 MOVE rename file (in same collection)",
    "11 MOVE undo rename",
    "12 MOVE to external destination",
    "13 MOVE onto existing file",
    "14 MOVE onto existing collection",
    "15 MOVE collection onto existing file",
    "16 MOVE resource into parent collection",
    "17 MOVE collection into parent collection",
    "18 MOVE resource onto itself",
    "19 MOVE rename to almost same name A->a",
    "20 MOVE rename and validate child list",
    "21 MOVE to destination with trailing blanks",
    "   MOVE cleanup",
  };

  public static int NrOfTests = desc.length;

  public MoveTest(Integer nr) {
    super(nr, desc[nr.intValue()]);
  }

  protected void runTest() {
    if (nr == 0) super.logIdentifier("MOVE");

    try {
      String respfile = "MOVE-" + TestMember.THREEDIGITS.format(nr) + ".txt";
      IResponse response;
      request.setMethod("MOVE");
      request.removeHeader("Overwrite");
      request.removeHeader("Depth");

      switch (nr) {

        case 0:
          request.setHeader("Destination", getBaseAndLocalpart() + "/123");
          request.setURI(getLocalpart());
          response = requester.perform(request);
          checkResponseCode(response, 403);
          break;

        case 1:
          request.setHeader("Destination", getBaseAndLocalpart() + "/cba");
          request.setURI(getLocalpart() + "/zyx.cba");
          response = requester.perform(request);
          checkResponseCode(response, 404);
          break;

        case 2:
          request.setHeader("Destination", getBaseAndLocalpart() + "/leer/abc");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;

        case 3:
          request.setHeader("Destination", getBaseAndLocalpart() + "/cba");
          request.setURI(getLocalpart() + "/leer/abc");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;

        case 4:
          request.setHeader("Destination", getBaseAndLocalpart() + "/def");
          request.setURI(getLocalpart() + "/cba");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;

        case 5:
          request.setHeader("Destination", getBaseAndLocalpart() + "/leer/abc");
          request.setURI(getLocalpart() + "/def");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;

        case 6:
          // make sure that collection child is cached
          clear(request);
          request.setMethod("HEAD");
          request.setURI(getLocalpart() + "/leer/abc");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          clear(request);
          request.setMethod("MOVE");
          request.setHeader("overwrite", "f");
          request.setHeader("Destination", getBaseAndLocalpart() + "/testcol");
          request.setURI(getLocalpart() + "/leer");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          // check that caches were indeed deep invalidated
          request.clear();
          request.setMethod("HEAD");
          request.setURI(getLocalpart() + "/leer/abc");
          response = requester.perform(request);
          checkResponseCode(response, 404);

          request.clear();
          request.setMethod("PROPFIND");
          request.setHeader("Depth", "1");
          request.setURI(getLocalpart() + "/testcol");
          request.setBody("<propfind xmlns='DAV:'><prop/></propfind>");
          response = requester.perform(request);
          checkResponseCode(response, 207);
          checkResultSet(response,  new String[]{
            getLocalpart() + "/testcol/",
            getLocalpart() + "/testcol/abc"});
          break;

        case 7:
          request.setHeader("Destination", getBaseAndLocalpart() + "/leer");
          request.setURI(getLocalpart() + "/testcol");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;

        case 8:
          request.setHeader("Destination", getBaseAndLocalpart() + "/ANSI.TXT");
          request.setURI(getLocalpart() + "/leer/abc");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;

        case 9:
          request.setMethod("HEAD");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 200);
          break;

        case 10:
          request.setHeader("Destination", getBaseAndLocalpart() + "/umbenannte%20Datei");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          request.setHeader("Overwrite", "F");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;

        case 11:
          request.setHeader("Destination", getBaseAndLocalpart() + "/ANSI.TXT");
          request.setURI(getLocalpart() + "/umbenannte%20Datei");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;

        case 12:
          request.setHeader("Destination", "http://www.greenbytes.de/foo/bar/ANSI.TXT");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 502);
          break;

        case 13:
          request.setMethod("COPY");
          request.setHeader("Destination", getBaseAndLocalpart() + "/xxx");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("MOVE");
          request.setHeader("Destination", getBaseAndLocalpart() + "/ANSI.TXT");
          request.setHeader("Overwrite", "F");
          request.setURI(getLocalpart() + "/xxx");
          response = requester.perform(request);
          checkResponseCode(response, 412);

          clear(request);
          request.setMethod("MOVE");
          request.setHeader("Destination", getBaseAndLocalpart() + "/ANSI.TXT");
          // request.setHeader("Overwrite", "T"); // it's the default anyway
          request.setURI(getLocalpart() + "/xxx");
          response = requester.perform(request);
          checkResponseCode(response, 204);
          break;

        case 14:
          request.setMethod("COPY");
          request.setHeader("Destination", getBaseAndLocalpart() + "/xxx");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("MKCOL");
          request.setURI(getLocalpart() + "/xxxdir");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("MOVE");
          request.setHeader("Destination", getBaseAndLocalpart() + "/xxxdir");
          request.setHeader("Overwrite", "F");
          request.setURI(getLocalpart() + "/xxx");
          response = requester.perform(request);
          checkResponseCode(response, 412);

          clear(request);
          request.setMethod("MOVE");
          request.setHeader("Destination", getBaseAndLocalpart() + "/xxxdir");
          request.setHeader("Overwrite", "T");
          request.setURI(getLocalpart() + "/xxx");
          response = requester.perform(request);
          checkResponseCode(response, 204);

          clear(request);
          request.setMethod("PROPFIND");
          request.setHeader("Depth", "0");
          request.setURI(getLocalpart() + "/xxxdir");
          checkMultistatus("<propfind xmlns='DAV:'><prop><resourcetype/></prop></propfind>",
                           respfile);

          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/xxxdir");
          response = requester.perform(request);
          checkResponseCode(response, 204);
          break;

        case 15:
          request.setMethod("COPY");
          request.setHeader("Destination", getBaseAndLocalpart() + "/xxx");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("MKCOL");
          request.setURI(getLocalpart() + "/xxxdir");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("MOVE");
          request.setHeader("Destination", getBaseAndLocalpart() + "/xxx");
          request.setURI(getLocalpart() + "/xxxdir");
          request.setHeader("Overwrite", "F");
          response = requester.perform(request);
          checkResponseCode(response, 412);

          clear(request);
          request.setMethod("MOVE");
          request.setHeader("Destination", getBaseAndLocalpart() + "/xxx");
          request.setHeader("Overwrite", "T");
          request.setURI(getLocalpart() + "/xxxdir");
          response = requester.perform(request);
          checkResponseCode(response, 204); // FIXME

// ADD ME
          clear(request);
          request.setMethod("PROPFIND");
          request.setHeader("Depth", "0");
          request.setURI(getLocalpart() + "/xxx");
          checkMultistatus("<propfind xmlns='DAV:'><prop><resourcetype/></prop></propfind>",
                           respfile);

          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/xxx");
          response = requester.perform(request);
          checkResponseCode(response, 204, 404);

          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/xxxdir");
          response = requester.perform(request);
          checkResponseCode(response, 204, 404);
          break;


        case 16:
          clear(request);
          request.setMethod("MKCOL");
          request.setURI(getLocalpart() + "/xxxdir");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("MKCOL");
          request.setURI(getLocalpart() + "/xxxdir/subdir");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("COPY");
          request.setHeader("Destination", getBaseAndLocalpart() + "/xxxdir/ANSI.TXT");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("COPY");
          request.setHeader("Destination", getBaseAndLocalpart() + "/xxxdir/ANSI2.TXT");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("COPY");
          request.setHeader("Destination", getBaseAndLocalpart() + "/xxxdir/subdir/ANSI3.TXT");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("MOVE");
          request.setHeader("Destination", getBaseAndLocalpart() + "/xxxdir");
          request.setHeader("Overwrite", "F");
          request.setURI(getLocalpart() + "/xxxdir/ANSI.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 412);

          clear(request);
          request.setMethod("MOVE");
          request.setHeader("Destination", getBaseAndLocalpart() + "/xxxdir");
          request.setHeader("Overwrite", "T");
          request.setURI(getLocalpart() + "/xxxdir/ANSI.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 204, 403);

          if (response.getStatus() == 204) {
            // changed to plain resource
            Element resourcetype = getProperty(getLocalpart() + "/xxxdir", Consts.N_resourcetype);
            assertTrue(DomElement.getChildren(resourcetype, Consts.N_collection).isEmpty());
          }
          else {
            // both collection and plain resource are still there
            Element resourcetype = super.getProperty(getLocalpart() + "/xxxdir", Consts.N_resourcetype);
            assertTrue(! DomElement.getChildren(resourcetype, Consts.N_collection).isEmpty());
            resourcetype = super.getProperty(getLocalpart() + "/xxxdir/ANSI.TXT", Consts.N_resourcetype);
            assertTrue(DomElement.getChildren(resourcetype, Consts.N_collection).isEmpty());
          }
          break;


        case 17:
          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/xxxdir");
          response = requester.perform(request);
          checkResponseCode(response, 204, 404);

          clear(request);
          request.setMethod("MKCOL");
          request.setURI(getLocalpart() + "/xxxdir");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("MKCOL");
          request.setURI(getLocalpart() + "/xxxdir/subdir");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("COPY");
          request.setHeader("Destination", getBaseAndLocalpart() + "/xxxdir/ANSI.TXT");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("COPY");
          request.setHeader("Destination", getBaseAndLocalpart() + "/xxxdir/ANSI2.TXT");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("COPY");
          request.setHeader("Destination", getBaseAndLocalpart() + "/xxxdir/subdir/ANSI3.TXT");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("MOVE");
          request.setHeader("Destination", getBaseAndLocalpart() + "/xxxdir");
          request.setHeader("Overwrite", "F");
          request.setURI(getLocalpart() + "/xxxdir/subdir");
          response = requester.perform(request);
          checkResponseCode(response, 412);

          clear(request);
          request.setMethod("MOVE");
          request.setHeader("Destination", getBaseAndLocalpart() + "/xxxdir");
          request.setHeader("Overwrite", "T");
          request.setURI(getLocalpart() + "/xxxdir/subdir");
          response = requester.perform(request);
          checkResponseCode(response, 204, 403);

          break;

        case 18:
          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/xxxdir");
          response = requester.perform(request);
          checkResponseCode(response, 204, 404);

          clear(request);
          request.setMethod("MKCOL");
          request.setURI(getLocalpart() + "/xxxdir");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("COPY");
          request.setHeader("Destination", getBaseAndLocalpart() + "/xxxdir/ANSI.TXT");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("MOVE");
          request.setHeader("Destination", getBaseAndLocalpart() + "/xxxdir/ANSI.TXT");
          request.setURI(getLocalpart() + "/xxxdir/ANSI.TXT");
          request.setHeader("Overwrite", "F");
          response = requester.perform(request);
          checkResponseCode(response, 403);
          checkErrorElement(response, DavConditions.N_SAP_source_and_destination_must_differ);

          clear(request);
          request.setMethod("MOVE");
          request.setHeader("Destination", getBaseAndLocalpart() + "/xxxdir/ANSI.TXT");
          request.setURI(getLocalpart() + "/xxxdir/ANSI.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 403);
          checkErrorElement(response, DavConditions.N_SAP_source_and_destination_must_differ);
          break;

        case 19:
          // This test case may fail if the jdk caches the results of getCanonicalPath and the
          // repository manager is based on java.io.File. This was introduced
          // as part of jdk 1.4.2; and can be disabled by setting "sun.io.useCanonCaches=false".

          clear(request);
          request.setMethod("MKCOL");
          request.setURI(getLocalpart() + "/xxxdir");
          response = requester.perform(request);
          checkResponseCode(response, 201, 405);

          clear(request);
          request.setMethod("COPY");
          request.setHeader("Destination", getBaseAndLocalpart() + "/xxxdir/ANSI.TXT");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 200, 201, 204);

          clear(request);
          request.setMethod("MOVE");
          request.setURI(getLocalpart() + "/xxxdir/ANSI.TXT");
          request.setHeader("Destination", getBaseAndLocalpart() + "/xxxdir/ansi.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 201, 204);

          // make sure remote side is not lying
          clear(request);
          request.setMethod("HEAD");
          request.setURI(getLocalpart() + "/xxxdir/ansi.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 200);
          assertTrue(response.getHeader("content-location") == null); // it's really there
          break;

        case 20:
          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/xxxdir");
          response = requester.perform(request);
          checkResponseCode(response, 204, 404);

          clear(request);
          request.setMethod("MKCOL");
          request.setURI(getLocalpart() + "/xxxdir");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("COPY");
          request.setHeader("Destination", getBaseAndLocalpart() + "/xxxdir/ANSI.TXT");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("PROPFIND");
          request.setURI(getLocalpart() + "/xxxdir");
          request.setHeader("Depth", "1");
          request.setBody("<propfind xmlns='DAV:'><prop><resourcetype/></prop></propfind>");
          request.expectsResponseDocument(true);
          response = requester.perform(request);
          checkResponseCode(response, 207);
          Document doc = response.getDocument();
          assertTrue("response document is null", doc != null);
          assertHrefEnding("/ANSI.TXT", doc);

          clear(request);
          request.setMethod("MOVE");
          request.setHeader("Destination", getBaseAndLocalpart() + "/xxxdir/ANSI2.TXT");
          request.setURI(getLocalpart() + "/xxxdir/ANSI.TXT");
          request.setHeader("Overwrite", "F");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("PROPFIND");
          request.setURI(getLocalpart() + "/xxxdir");
          request.setHeader("Depth", "1");
          request.setBody("<propfind xmlns='DAV:'><prop><resourcetype/></prop></propfind>");
          request.expectsResponseDocument(true);
          response = requester.perform(request);
          checkResponseCode(response, 207);
          doc = response.getDocument();
          assertTrue("response document is null", doc != null);
          assertHrefEnding("/ANSI2.TXT", doc);

          break;

        case 21:
          clear(request);
          request.setMethod("MOVE");
          request.setHeader("Destination", getBaseAndLocalpart() + "/xxxdir/ANSI.TXT%20");
          request.setURI(getLocalpart() + "/xxxdir/ANSI2.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 403);

          clear(request);
          request.setMethod("MKCOL");
          request.setURI(getLocalpart() + "/xxxdir/subdir");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("MOVE");
          request.setHeader("Destination", getBaseAndLocalpart() + "/xxxdir/subdir/ANSI.TXT%20");
          request.setURI(getLocalpart() + "/xxxdir/ANSI2.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 403);

          break;

        default:
          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/xxxdir");
          response = requester.perform(request);
          checkResponseCode(response, 204, 404);
          break;

      }
    }
    catch (Exception ex) {
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }


  private static final DomPath DP_href = new DomPath("//D:href", DomPath.DEFAULT_NAMESPACE_BINDINGS);

  private void assertHrefEnding(String ref, Document doc) throws Exception {
    Node[] nodes = DP_href.matchNodes(doc);
    boolean found = false;
    for (int i = 0; (i < nodes.length) && !found; ++i) {
      String text = DomElement.getText(nodes[i]);
      if (text.endsWith(ref)) {
        found = true;
      }
    }
    if (!found) {
      assertTrue("href ending in "+ref+" not found in response: "+DomSerializer.toString(doc, true), false);
    }
  }

}
