package com.sapportals.wcm.protocol.webdav.server.tests;

import org.w3c.dom.Element;

import com.sapportals.wcm.protocol.webdav.Consts;
import com.sapportals.wcm.protocol.webdav.dompath.DomPath;
import com.sapportals.wcm.util.http.IResponse;

/**
 * Test cases simulating the typical locking behaviour of MS Office.
 * <p>
 * Copyright (c) SAP AG 2001-2003
 * @author julian.reschke@greenbytes.de
 * @version $Id: MsOfficeLockTest.java,v 1.6 2004/08/16 13:22:16 jre Exp $
 */

public class MsOfficeLockTest extends TestMember
{

  private static java.text.DecimalFormat m_threeDigits = new
    java.text.DecimalFormat ("000");

  private static final String[] desc = new String[]
  {
    "00 OL create null-lock (save as)",
    "01 OL PROPFIND on null-lock",
    "02 OL put content",
    "03 OL LOCK refresh",
    "04 OL UNLOCK",
    "05 OL switch on version control",
    "06 OL LOCK it",
    "07 OL update content",
    "08 OL update properties",
    "09 OL UNLOCK",
    "10 OL version history incl. comments",
    "11 OL delete testdoc",
    "12 OL create null-lock (save as)",
    "13 OL CHECKOUT null-lock",
    "14 OL PUT on checked-out null-lock",
    "15 OL PROPPATCH on checked-out",
    "16 OL CHECKIN",
    "17 OL UNLOCK",
    "18 OL version history incl. comments",
    "   OL cleanup",
  };

  public static int NrOfTests = desc.length;

  public MsOfficeLockTest(Integer nr)
  {
    super(nr, desc[nr.intValue()]);
  }

  private static final DomPath DP_auto_version = new DomPath(
    "/D:multistatus/D:response/D:propstat/D:prop/D:auto-version/D:checkout-unlocked-checkin", DomPath.DEFAULT_NAMESPACE_BINDINGS);
  private static final DomPath DP_versioned = new DomPath(
    "/D:multistatus/D:response/D:propstat/D:prop/D:checked-in", DomPath.DEFAULT_NAMESPACE_BINDINGS);

  private static String locktoken = null;
  private static boolean isautoversioned = false;
  private static boolean isversioned = false;
  private static String testdocvhr = null;
  private static String testdocvhr2 = null;

  protected void runTest()
  {
    if (nr == 0) super.logIdentifier("OL");

    String respfile = "MSOFFICELOCK-" + m_threeDigits.format (nr) + ".txt";
    String testdoc = "/officetest.doc";
    String testdoc2 = "/officetest2.doc";

    try {
      IResponse response;
      clear(request);

      switch (nr) {
        case 0:
          request.setMethod("LOCK");
          setRelativeURI(testdoc);
          request.setHeader("Depth", "0");
          request.setHeader("Timeout", "Second-180");
          request.setHeader("Content-Type", "text/xml");
          request.setBody("<lockinfo xmlns='DAV:'><lockscope><exclusive/></lockscope><locktype><write/></locktype><owner>whatever-office-likes-here</owner></lockinfo>");
          request.expectsResponseDocument(true);
          response = requester.perform(request);
          checkResponseCode(response, 200, 201);
          checkResponseDocument(response, respfile, null);
          locktoken = getLockToken(response);
          break;

        case 1:
          request.setMethod("PROPFIND");
          request.setHeader("Depth", "0");
          setRelativeURI(testdoc);
          response = requester.perform(request);
          response = checkStatusAndBody(207,
            "<propfind xmlns='DAV:'><prop><lockdiscovery/></prop></propfind>",
            respfile);
          break;

        case 2:
          request.setMethod("PUT");
          setRelativeURI(testdoc);
          request.setHeader("If", "(<" + locktoken + ">)");
          request.setBody("Do not read this");
          response = requester.perform(request);
          checkResponseCode(response, 200);
          break;

        case 3:
          request.setMethod("LOCK");
          setRelativeURI(testdoc);
          request.setHeader("If", "(<" + locktoken + ">)");
          response = requester.perform(request);
          checkResponseCode(response, 200);
          break;

        case 4:
          request.setMethod("UNLOCK");
          setRelativeURI(testdoc);
          request.setHeader("Lock-Token", "<" + locktoken + ">");
          response = requester.perform(request);
          checkResponseCode(response, 204);
          locktoken = null;
          break;

        case 5:
          request.setMethod("VERSION-CONTROL");
          setRelativeURI(testdoc);
          response = requester.perform(request);
          if (response.getStatus() == 200 || response.getStatus() == 201) {

            // also get auto-versioning property
            clear(request);
            request.setMethod("PROPFIND");
            request.setHeader("Depth", "0");
            setRelativeURI(testdoc);
            request.setBody("<propfind xmlns='DAV:'><prop><auto-version/><checked-in/><checked-out/></prop></propfind>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            isautoversioned = DP_auto_version.matchNodes(response.getDocument()).length > 0;
            isversioned = DP_versioned.matchNodes(response.getDocument()).length > 0;
            
            Element vhr = getProperty(getLocalpart() + testdoc, Consts.N_version_history);
            if (vhr != null) {
              testdocvhr = getRelativeUriFromPropertyHrefs(vhr);
            }
          }
          break;

        case 6:
          if (isautoversioned) {
            request.setMethod("LOCK");
            setRelativeURI(testdoc);
            request.setHeader("Depth", "0");
            request.setHeader("Timeout", "Second-180");
            response = checkStatusAndBody(200,
              "<lockinfo xmlns='DAV:'><lockscope><exclusive/></lockscope><locktype><write/></locktype><owner>whatever-office-likes-here</owner></lockinfo>",
              respfile);
            locktoken = getLockToken(response);
          }
          break;

        case 7:
          if (isautoversioned) {
            request.setMethod("PUT");
            setRelativeURI(testdoc);
            request.setHeader("Content-Language", "en-setintestcase7");
            request.setHeader("If", "(<" + locktoken + ">)");
            request.setBody("Do not read this (but this is longer)");
            response = requester.perform(request);
            checkResponseCode(response, 200);
          }
          break;

        case 8:
          if (isautoversioned) {
            request.setMethod("PROPPATCH");
            setRelativeURI(testdoc);
            request.setHeader("If", "(<" + locktoken + ">)");
            response = checkStatusAndBody(207,
              "<propertyupdate xmlns='DAV:'><set><prop>" +
                "<comment>set during test case 8 (should not be overwritten by automatically generated comment)</comment>" +
                "</prop></set></propertyupdate>",
              respfile);
          }
          break;

        case 9:
          if (isautoversioned) {
            request.setMethod("UNLOCK");
            setRelativeURI(testdoc);
            request.setHeader("Lock-Token", "<" + locktoken + ">");
            response = requester.perform(request);
            checkResponseCode(response, 204);
            locktoken = null;
          }
          break;

        case 10:
          if (isautoversioned) {
            request.setMethod("REPORT");
            setRelativeURI(testdoc);
            checkMultistatus("<version-tree xmlns='DAV:'><prop><comment/><version-name/><getcontentlanguage/><predecessor-set/><successor-set/><checkout-set/></prop></version-tree>",
              respfile);
          }
          break;

        case 11:
          if (isautoversioned) {
            request.setMethod("DELETE");
            setRelativeURI(testdoc);
            response = requester.perform(request);
            checkResponseCode(response, 204);
          }
          break;

        case 12:
          if (isversioned) {
            request.setMethod("LOCK");
            setRelativeURI(testdoc2);
            request.setHeader("Depth", "0");
            request.setHeader("Timeout", "Second-180");
            request.setHeader("Content-Type", "text/xml");
            request.setBody("<lockinfo xmlns='DAV:'><lockscope><exclusive/></lockscope><locktype><write/></locktype><owner>whatever-office-likes-here</owner></lockinfo>");
            request.expectsResponseDocument(true);
            response = requester.perform(request);
            checkResponseCode(response, 200, 201);
            checkResponseDocument(response, respfile, null);
            locktoken = getLockToken(response);
          }
          break;

        case 13:
          if (isversioned) {
            request.setMethod("VERSION-CONTROL");
            setRelativeURI(testdoc2);
            request.setHeader("If", "(<" + locktoken + ">)");
            response = requester.perform(request);
            checkResponseCode(response, 200, 201);
            Element vhr = getProperty(getLocalpart() + testdoc2, Consts.N_version_history);
            if (vhr != null) {
              testdocvhr2 = getRelativeUriFromPropertyHrefs(vhr);
            }

            clear(request);
            request.setMethod("CHECKOUT");
            setRelativeURI(testdoc2);
            request.setHeader("If", "(<" + locktoken + ">)");
            response = requester.perform(request);
            checkResponseCode(response, 200);
          }
          break;

        case 14:
          if (isversioned) {
            request.setMethod("PUT");
            setRelativeURI(testdoc2);
            request.setHeader("If", "(<" + locktoken + ">)");
            request.setBody("Do not read this (but this is longer)");
            response = requester.perform(request);
            checkResponseCode(response, 200);
          }
          break;

        case 15:
          if (isversioned) {
            request.setMethod("PROPPATCH");
            setRelativeURI(testdoc2);
            request.setHeader("If", "(<" + locktoken + ">)");
            response = checkStatusAndBody(207,
              "<propertyupdate xmlns='DAV:'><set><prop><comment>set during test case 15</comment></prop></set></propertyupdate>",
              respfile);
          }
          break;

        case 16:
          if (isversioned) {
            request.setMethod("CHECKIN");
            setRelativeURI(testdoc2);
            request.setHeader("If", "(<" + locktoken + ">)");
            response = requester.perform(request);
            checkResponseCode(response, 201);
          }
          break;

        case 17:
          if (isversioned) {
            request.setMethod("UNLOCK");
            setRelativeURI(testdoc2);
            request.setHeader("Lock-Token", "<" + locktoken + ">");
            response = requester.perform(request);
            checkResponseCode(response, 204);
            locktoken = null;
          }
          break;

        case 18:
          if (isversioned) {
            request.setMethod("REPORT");
            setRelativeURI(testdoc2);
            checkMultistatus("<version-tree xmlns='DAV:'><prop><comment/><version-name/><predecessor-set/><successor-set/></prop></version-tree>",
              respfile);
          }
          break;

        default:
          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + testdoc);
          if (locktoken != null) request.setHeader("If", "(<" + locktoken + ">)");
          response = requester.perform(request);
          checkResponseCode(response, 204, 404);

          if (testdocvhr != null && vhrPreserved()) {
            clear(request);
            request.setMethod("DELETE");
            request.setURI(testdocvhr);
            response = requester.perform(request);
            checkResponseCode(response, 204);
            
            // make sure it's gone
            clear(request);
            request.setMethod("HEAD");
            request.setURI(testdocvhr);
            response = requester.perform(request);
            checkResponseCode(response, 404);
          }
          
          if (isversioned) {
            clear(request);
            request.setMethod("DELETE");
            request.setURI(getLocalpart() + testdoc2);
            if (locktoken != null) request.setHeader("If", "(<" + locktoken + ">)");
            response = requester.perform(request);
            checkResponseCode(response, 204);

            if (testdocvhr2 != null && vhrPreserved()) {
              clear(request);
              request.setMethod("DELETE");
              request.setURI(testdocvhr2);
              response = requester.perform(request);
              checkResponseCode(response, 204);
            }
}
          break;
      }
    }
    catch (Exception ex) {
      ex.printStackTrace(System.err);
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }
}
