package com.sapportals.wcm.protocol.webdav.server.tests;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.w3c.dom.Element;
import org.w3c.dom.Node;

import com.sapportals.wcm.protocol.webdav.Consts;
import com.sapportals.wcm.protocol.webdav.DavConditions;
import com.sapportals.wcm.protocol.webdav.DomElement;
import com.sapportals.wcm.protocol.webdav.dompath.DomPath;
import com.sapportals.wcm.util.http.IResponse;
import com.sapportals.wcm.util.uri.IUri;
import com.sapportals.wcm.util.uri.UriFactory;

/**
 * Various tests for ordered collections.
 * <p>
 * Copyright (c) SAP AG 2001-2005
 * @author martin.boettcher@greenbytes.de
 * @author julian.reschke@greenbytes.de
 * @author manfred.baedke@greenbytes.de
 * @version $Id: OrderedCollectionTest.java,v 1.6 2005/01/28 11:50:47 jre Exp $
 */

public class OrderedCollectionTest extends TestMember {

  private static final String[] desc = new String[]  {
   "00 OC MKCOL ordered",
   "01 OC PUT into ordered(first)",
   "02 OC PUT into ordered(last)",
   "03 OC PUT into ordered(before)",
   "04 OC PUT into ordered(after)",
   "05 OC PROPFIND and depth 1 on ordered",
   "06 OC DELETE ordered/three.txt",
   "07 OC PROPFIND and depth 1 on ordered after first delete",
   "08 OC MKCOL ordered/three",
   "09 OC PROPFIND and depth 1 on ordered after MKCOL",
   "10 OC COPY into ordered",
   "11 OC PROPFIND and depth 1 on ordered after COPY",
   "12 OC take ordered collection under version control, setup ordered collection version",
   "13 OC setup new collection version with different ordering",
   "14 OC update from ordered versions, check ordering",
   "15 OC check version controlled binding sets",
   "16 OC check ordering types of collection and checked-in-version",
   "17 OC checkout, set ordering type, checkin, check ordering types",
   "18 OC checkout, set ordering type, check ordering type, undocheckout, check ordering type",
   "19 OC update, check ordering type",
   "20 OC check ordering in working collection",
   "21 OC check ordering in workspace controlled collection",
   "22 OC PUT/MKCOL/COPY with position header into non-ordered collections",
   "   OC cleanup",
   };

  private static IUri specOrderType = null;
  private static IUri tmpOrderType = null;

  private static String orderedCollectionVersionHistory;
  
  private static String secondVersionOfCollection;
  private static String lastVersionOfCollection;

  private static String wspUri = null;

  private static boolean supportsRFC = false;
//  private static boolean supportsOldSapImpl = false;
  private static boolean supportsWorkingCollections = false;

  private static IUri CUSTOM, UNORDERED;

  static {
    try {
      CUSTOM = UriFactory.parseUri("DAV:custom");
      UNORDERED = UriFactory.parseUri("DAV:unordered");
      specOrderType = UriFactory.parseUri("DAV:specifiedorderingtype");
      tmpOrderType = UriFactory.parseUri("DAV:temporaryorderingtype");
    }
    catch (Exception ex) {
      throw new java.lang.RuntimeException(ex.getMessage());
    }
  }
  public static int NrOfTests = desc.length;

  
  public OrderedCollectionTest(Integer nr) {
    super(nr, desc[nr.intValue()]);
  }


  protected void runTest() {
    if (nr == 0) super.logIdentifier("OC");

    try {
      
      IResponse response;

      switch (nr) {
        
        case 0:
          request.setMethod("MKCOL");
          request.setHeader("Ordering-Type", CUSTOM.toExternalForm());
          request.setURI(getLocalpart() + "/ordered");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("OPTIONS");
          request.setURI(getLocalpart() + "/ordered");
          response = requester.perform(request);
          checkResponseCode(response, 200);
          supportsRFC = response.getHeader("DAV").indexOf("ordered-collections") >= 0;
//          supportsOldSapImpl = response.getHeader("DAV").indexOf("ordering") >= 0;
          break;

        case 1:
          doPut("/ordered/one.txt", "first");
          break;

        case 2:
          doPut("/ordered/four.txt", "last");
          break;

        case 3:
          doPut("/ordered/three.txt", "before four.txt");
          break;

        case 4:
          doPut("/ordered/two.txt", "after one.txt");
          break;

        case 5:
          checkOrdering(getLocalpart() + "/ordered",
            new String[] {
                getLocalpart() + "/ordered/one.txt",
                getLocalpart() + "/ordered/two.txt",
                getLocalpart() + "/ordered/three.txt",
                getLocalpart() + "/ordered/four.txt"},
            null,
            CUSTOM);
          break;

        case 6:
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/ordered/three.txt");
          response = requester.perform(request);
          checkResponseCode(response, 204);
          break;

        case 7:
          checkOrdering(getLocalpart() + "/ordered",
            new String[] {
                getLocalpart() + "/ordered/one.txt",
                getLocalpart() + "/ordered/two.txt",
                getLocalpart() + "/ordered/four.txt"},
            null,
            CUSTOM);
          break;

        case 8:
          request.setMethod("MKCOL");
          request.setHeader("Position", "after two.txt");
          request.setURI(getLocalpart() + "/ordered/three");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;

        case 9:
          checkOrdering(getLocalpart() + "/ordered",
            new String[] {
                getLocalpart() + "/ordered/one.txt",
                getLocalpart() + "/ordered/two.txt",
                getLocalpart() + "/ordered/three/",
               getLocalpart() + "/ordered/four.txt"},
            null,
            CUSTOM);
          break;

        case 10:
          request.setMethod("COPY");
          request.setHeader("Destination", this.getBaseAndLocalpart() + "/ordered/Mappe1.xls");
          request.setHeader("Position", "after two.txt");
          request.setURI(getLocalpart() + "/Mappe1.xls");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;

        case 11:
          checkOrdering(getLocalpart() + "/ordered",
            new String[] {
                getLocalpart() + "/ordered/one.txt",
                getLocalpart() + "/ordered/two.txt",
                getLocalpart() + "/ordered/Mappe1.xls",
                getLocalpart() + "/ordered/three/",
                getLocalpart() + "/ordered/four.txt"},
            null,
            CUSTOM);
         break;

        case 12:
          clear(request);
          request.setMethod("VERSION-CONTROL");
          request.setURI(getLocalpart() + "/ordered");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          Element vhr = getProperty(getLocalpart() + "/ordered", Consts.N_version_history);
          orderedCollectionVersionHistory = getRelativeUriFromPropertyHrefs(vhr);
          
          clear(request);
          request.setMethod("CHECKOUT");
          request.setURI(getLocalpart() + "/ordered");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          clear(request);
          request.setMethod("VERSION-CONTROL");
          request.setURI(getLocalpart() + "/ordered/one.txt");
          response = requester.perform(request);
          checkResponseCode(response, 200, 201);

          clear(request);
          request.setMethod("VERSION-CONTROL");
          request.setURI(getLocalpart() + "/ordered/two.txt");
          response = requester.perform(request);
          checkResponseCode(response, 200, 201);

          clear(request);
          request.setMethod("VERSION-CONTROL");
          request.setURI(getLocalpart() + "/ordered/three");
          response = requester.perform(request);
          checkResponseCode(response, 200, 201);

          clear(request);
          request.setMethod("VERSION-CONTROL");
          request.setURI(getLocalpart() + "/ordered/four.txt");
          response = requester.perform(request);
          checkResponseCode(response, 200, 201);

          clear(request);
          request.setMethod("CHECKIN");
          request.setURI(getLocalpart() + "/ordered");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          secondVersionOfCollection = getRelativeUriFromLocationHeader(response);
          break;

        case 13:

          clear(request);
          request.setMethod("CHECKOUT");
          request.setURI(getLocalpart() + "/ordered");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          request.setMethod("ORDERPATCH");
          request.setURI(getLocalpart() + "/ordered");
          request.setBody(
              "<orderpatch xmlns='DAV:'>"
              + "<ordering-type><href>" + CUSTOM.toExternalForm() + "</href></ordering-type>"
              + "<order-member>"
              + "<segment>one.txt</segment><position><first/></position>"
              + "</order-member>"
              + "<order-member>"
              + "<segment>four.txt</segment><position><first/></position>"
              + "</order-member>"
              + "<order-member>"
              + "<segment>three</segment><position><first/></position>"
              + "</order-member>"
              + "<order-member>"
              + "<segment>two.txt</segment><position><first/></position>"
              + "</order-member>"
              + "</orderpatch>");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          clear(request);
          request.setMethod("CHECKIN");
          request.setURI(getLocalpart() + "/ordered");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          lastVersionOfCollection = getRelativeUriFromLocationHeader(response);
          break;

        case 14:
          String[] hrefs = new String[] {
              getLocalpart() + "/ordered/two.txt",
              getLocalpart() + "/ordered/three/",
              getLocalpart() + "/ordered/four.txt",
              getLocalpart() + "/ordered/one.txt"};
          HashSet ignore = new HashSet(
              Arrays.asList(new String[] {getLocalpart() + "/ordered/", getLocalpart() + "/ordered/Mappe1.xls"}));
          checkOrdering(getLocalpart() + "/ordered", hrefs, ignore, CUSTOM);

          clear(request);
          request.setMethod("UPDATE");
          request.setURI(getLocalpart() + "/ordered");
          request.setBody("<update xmlns='DAV:'><version><href>"+
                          secondVersionOfCollection + "</href></version></update>");
          response = requester.perform(request);
          checkResponseCode(response, 207, 501);

          if (response.getStatus() == 207) {

            clear(request);
            request.setMethod("PROPFIND");
            request.setURI(getLocalpart() + "/ordered");
            request.setHeader("Depth", "1");
            request.setBody("<propfind xmlns='DAV:'><prop/></propfind>");
            request.expectsResponseDocument(true);
            response = requester.perform(request);
            checkResponseCode(response, 207);

            hrefs = new String[] {
                getLocalpart() + "/ordered/one.txt",
                getLocalpart() + "/ordered/two.txt",
                getLocalpart() + "/ordered/three/",
                getLocalpart() + "/ordered/four.txt"};
            ignore = new HashSet(
                Arrays.asList(new String[] {getLocalpart() + "/ordered/", getLocalpart() + "/ordered/Mappe1.xls"}));
            checkSortedResultSet(response, hrefs, ignore);
          }

          break;

        case 15:

          Element vcbs
              = getProperty(lastVersionOfCollection, Consts.N_version_controlled_binding_set);
          assertTrue(vcbs != null);
          assertTrue("expected 4 bindings, found " + vcbs.getChildNodes().getLength(),
                     vcbs.getChildNodes().getLength() == 4);
          //List bindings = DomElement.getChildren(vcbs, Consts.N_version_controlled_binding);

          Element binding
              = (Element) DomElement.getChildren(vcbs, Consts.N_version_controlled_binding).get(0);
          String bindingName
              = DomElement.getText((Node) DomElement.getChildren(binding, Consts.N_binding_name).get(0));
          assertTrue("two.txt".equals(bindingName));

          binding
              = (Element) DomElement.getChildren(vcbs, Consts.N_version_controlled_binding).get(1);
          bindingName
              = DomElement.getText((Node) DomElement.getChildren(binding, Consts.N_binding_name).get(0));
          assertTrue("three".equals(bindingName));

          binding
              = (Element) DomElement.getChildren(vcbs, Consts.N_version_controlled_binding).get(2);
          bindingName
              = DomElement.getText((Node) DomElement.getChildren(binding, Consts.N_binding_name).get(0));
          assertTrue("four.txt".equals(bindingName));

          binding
              = (Element) DomElement.getChildren(vcbs, Consts.N_version_controlled_binding).get(3);
          bindingName
              = DomElement.getText((Node) DomElement.getChildren(binding, Consts.N_binding_name).get(0));
          assertTrue("one.txt".equals(bindingName));

          vcbs =  getProperty(secondVersionOfCollection, Consts.N_version_controlled_binding_set);
          assertTrue(vcbs != null);
          assertTrue("expected 4 bindings, found " + vcbs.getChildNodes().getLength(),
                     vcbs.getChildNodes().getLength() == 4);
          // List bindings = DomElement.getChildren(vcbs, Consts.N_version_controlled_binding);

          binding
              = (Element) DomElement.getChildren(vcbs, Consts.N_version_controlled_binding).get(0);
          bindingName
              = DomElement.getText((Node) DomElement.getChildren(binding, Consts.N_binding_name).get(0));
          assertTrue("one.txt".equals(bindingName));

          binding
              = (Element) DomElement.getChildren(vcbs, Consts.N_version_controlled_binding).get(1);
          bindingName
              = DomElement.getText((Node) DomElement.getChildren(binding, Consts.N_binding_name).get(0));
          assertTrue("two.txt".equals(bindingName));

          binding
              = (Element) DomElement.getChildren(vcbs, Consts.N_version_controlled_binding).get(2);
          bindingName
              = DomElement.getText((Node) DomElement.getChildren(binding, Consts.N_binding_name).get(0));
          assertTrue("three".equals(bindingName));

          binding
              = (Element) DomElement.getChildren(vcbs, Consts.N_version_controlled_binding).get(3);
          bindingName
              = DomElement.getText((Node) DomElement.getChildren(binding, Consts.N_binding_name).get(0));
          assertTrue("four.txt".equals(bindingName));

          break;

        case 16:

          clear(request);
          request.setMethod("UPDATE");
          request.setURI(getLocalpart() + "/ordered");
          request.setBody("<update xmlns='DAV:'><version><href>"+
                          lastVersionOfCollection + "</href></version></update>");
          response = requester.perform(request);
          checkResponseCode(response, 207, 501);

          Element orderingType = this.getProperty(getLocalpart() + "/ordered", Consts.N_ordering_type);
          String uri = DomElement.getSingleHref(orderingType);
          assertTrue(
              "Orderingtype should be " + CUSTOM.toExternalForm() + ", but is " + uri,
              CUSTOM.toExternalForm().equals(uri));

          orderingType = this.getProperty(lastVersionOfCollection, Consts.N_ordering_type);
          uri = DomElement.getSingleHref(orderingType);
          assertTrue(
              "Orderingtype should be " + CUSTOM.toExternalForm() + ", but is " + uri,
              CUSTOM.toExternalForm().equals(uri));

          break;

        case 17:

          clear(request);
          request.setMethod("CHECKOUT");
          request.setURI(getLocalpart() + "/ordered");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          request.setMethod("ORDERPATCH");
          request.setURI(getLocalpart() + "/ordered");
          request.setBody(
              "<orderpatch xmlns='DAV:'>"
              + "<ordering-type><href>" + specOrderType.toExternalForm() + "</href></ordering-type>"
              + "</orderpatch>");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          clear(request);
          request.setMethod("CHECKIN");
          request.setURI(getLocalpart() + "/ordered");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          lastVersionOfCollection = getRelativeUriFromLocationHeader(response);

          orderingType = this.getProperty(getLocalpart() + "/ordered", Consts.N_ordering_type);
          uri = DomElement.getSingleHref(orderingType);
          assertTrue(
              "Orderingtype should be " + specOrderType.toExternalForm() + ", but is " + uri,
              specOrderType.toExternalForm().equals(uri));

          orderingType = this.getProperty(lastVersionOfCollection, Consts.N_ordering_type);
          uri = DomElement.getSingleHref(orderingType);
          assertTrue(
              "Orderingtype should be " + specOrderType.toExternalForm() + ", but is " + uri,
              specOrderType.toExternalForm().equals(uri));

          break;

        case 18:

          clear(request);
          request.setMethod("CHECKOUT");
          request.setURI(getLocalpart() + "/ordered");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          request.setMethod("ORDERPATCH");
          request.setURI(getLocalpart() + "/ordered");
          request.setBody(
              "<orderpatch xmlns='DAV:'>"
              + "<ordering-type><href>" + tmpOrderType.toExternalForm() + "</href></ordering-type>"
              + "</orderpatch>");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          orderingType = this.getProperty(getLocalpart() + "/ordered", Consts.N_ordering_type);
          uri = DomElement.getSingleHref(orderingType);
          assertTrue(
              "Orderingtype should be " + tmpOrderType.toExternalForm() + ", but is " + uri,
              tmpOrderType.toExternalForm().equals(uri));

          clear(request);
          request.setMethod("UNCHECKOUT");
          request.setURI(getLocalpart() + "/ordered");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          orderingType = this.getProperty(getLocalpart() + "/ordered", Consts.N_ordering_type);
          uri = DomElement.getSingleHref(orderingType);
          assertTrue(
              "Orderingtype should be " + specOrderType.toExternalForm() + ", but is " + uri,
              specOrderType.toExternalForm().equals(uri));

          break;

        case 19:

          clear(request);
          request.setMethod("UPDATE");
          request.setURI(getLocalpart() + "/ordered");
          request.setBody("<update xmlns='DAV:'><version><href>"+
                          secondVersionOfCollection + "</href></version></update>");
          response = requester.perform(request);
          checkResponseCode(response, 207, 501);

          if (response.getStatus() == 207) {
            orderingType = this.getProperty(getLocalpart() + "/ordered", Consts.N_ordering_type);
            uri = DomElement.getSingleHref(orderingType);
            assertTrue(
                "Orderingtype should be " + CUSTOM.toExternalForm() + ", but is " + uri,
                CUSTOM.toExternalForm().equals(uri));
          }

          break;

        case 20:

          clear(request);
          request.setMethod("CHECKOUT");
          request.setBody("<checkout xmlns='DAV:'><fork-ok/></checkout>");
          request.setURI(lastVersionOfCollection);
          response = requester.perform(request);
          checkResponseCode(response, 201, 403, 501);
          supportsWorkingCollections = (response.getStatus() == 201);

          if (supportsWorkingCollections) {
            String wcUri = getRelativeUriFromLocationHeader(response);
            // all uris end with slashes because they are version history bindings
            hrefs = new String[] {
                wcUri + "/two.txt/",
                wcUri + "/three/",
                wcUri + "/four.txt/",
                wcUri + "/one.txt/"};
            this.checkOrdering(wcUri, hrefs, null, specOrderType);

            clear(request);
            request.setMethod("UNCHECKOUT");
            request.setURI(wcUri);
            response = requester.perform(request);
            checkResponseCode(response, 200);

            // working collection should be gone
            clear(request);
            request.setMethod("HEAD");
            request.setURI(wcUri);
            response = requester.perform(request);
            checkResponseCode(response, 404);
           }

          break;

        case 21:

          //get workspace collection  set
          clear(request);
          request.setMethod("OPTIONS");
          request.setURI(getLocalpart());
          request.expectsResponseDocument(true);
          request.setBody("<options xmlns='DAV:'><workspace-collection-set/></options>");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          //get workspace collection set as live property
          String wspColUri = null;
          Element wspColElement = this.getProperty(getLocalpart(), Consts.N_workspace_collection_set);
          if (wspColElement != null) {
            Set hrefSet = DomElement.getHrefs(wspColElement);
            Iterator hrefIter = hrefSet.iterator();
            if (hrefIter.hasNext()) {
              wspColUri = (String) hrefIter.next();
            }
          }

          if (wspColUri != null) {

            wspUri = wspColUri + getLocalpart().substring(getLocalpart().lastIndexOf('/')) + "wsp";

            //create workspace
            clear(request);
            request.setMethod("MKWORKSPACE");
            request.setURI(wspUri);
            response = requester.perform(request);
            checkResponseCode(response, 201);

            if (response.getStatus() == 201) {

              //create WS-controlled version controlled collection
              clear(request);
              request.setMethod("VERSION-CONTROL");
              request.setURI(wspUri + "/oct21");
              request.setBody("<version-control xmlns='DAV:'><version>"
                              + "<href>" + lastVersionOfCollection + "</href></version></version-control>");
              response = requester.perform(request);
              checkResponseCode(response, 201, 501);

              if (response.getStatus() == 201) {
                hrefs = new String[] {
                    wspUri + "/oct21/two.txt",
                    wspUri + "/oct21/three/",
                    wspUri + "/oct21/four.txt",
                    wspUri + "/oct21/one.txt"};
                this.checkOrdering(wspUri + "/oct21", hrefs, null, specOrderType);
              }
            }
          }
          break;

        case 22: {
            clear(request);
            request.setMethod("MKCOL");
            request.setHeader("position", "first");
            request.setURI(getLocalpart() + "/orderedcolltest");
            response = requester.perform(request);
            checkResponseCode(response, 409);
            checkErrorElement(response, DavConditions.N_collection_must_be_ordered);
            
            clear(request);
            request.setMethod("PUT");
            request.setHeader("position", "first");
            request.setURI(getLocalpart() + "/orderedcolltest");
            response = requester.perform(request);
            checkResponseCode(response, 409);
            checkErrorElement(response, DavConditions.N_collection_must_be_ordered);

            clear(request);
            request.setURI(getLocalpart() + "/Worddatei.doc");
            request.setMethod("COPY");
            request.setHeader("position", "first");
            request.setHeader("destination", getLocalpart() + "/orderedcolltest");
            response = requester.perform(request);
            checkResponseCode(response, 409);
            checkErrorElement(response, DavConditions.N_collection_must_be_ordered);
          }
          break;

        default: {
            if (wspUri != null) {
              clear(request);
              request.setMethod("DELETE");
              request.setURI(wspUri);
              response = requester.perform(request);
              checkResponseCode(response, 204);
            }
  
            clear(request);
            request.setMethod("DELETE");
            request.setURI(getLocalpart() + "/orderedcolltest");
            response = requester.perform(request);
            checkResponseCode(response, 204, 404);
  
            clear(request);
            request.setMethod("DELETE");
            request.setURI(getLocalpart() + "/ordered");
            response = requester.perform(request);
            checkResponseCode(response, 204);
  
            clear(request);
            request.setMethod("DELETE");
            request.setURI(orderedCollectionVersionHistory);
            response = requester.perform(request);
            checkResponseCode(response, vhrPreserved() ? 204 : 404);
          }
        break;
      }
    }
    catch (Exception ex) {
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }

  private void doPut(String name, String position) throws com.sapportals.wcm.WcmException {
    IResponse response;

    clear(request);
    request.setMethod("PUT");
    request.setURI(getLocalpart() + name);
    request.setHeader("Position", position);
    response = requester.perform(request);
    checkResponseCode(response, 201);
  }

  private static DomPath DP_ordering_type = new DomPath
    ("/D:multistatus/D:response/D:propstat[contains(D:status,' 200 ')]/D:prop/D:ordering-type|"+
    "/D:multistatus/D:response/D:propstat[contains(D:status,' 200 ')]/D:prop/D:orderingtype",
    DomPath.DEFAULT_NAMESPACE_BINDINGS);

  private void checkOrdering(String uri, String hrefs[], Set ignore, IUri orderingtype) throws java.lang.Exception {

    if (ignore == null) {
      ignore = new HashSet();
      ignore.add(uri.endsWith("/") ? uri : uri + "/");
    }

    clear(request);
    request.setMethod("PROPFIND");
    request.setURI(uri);
    request.setHeader("Depth", "1");
    request.setBody("<propfind xmlns='DAV:'><prop>"
      + (supportsRFC ? "<ordering-type/>" : "<orderingtype/>")
      + "</prop></propfind>");
    IResponse response = requester.perform(request);
    checkResponseCode(response, 207);

    Node ot = DP_ordering_type.matchFirst(response.getDocument());
    IUri oturi = null;
    if (supportsRFC) {
      String othref = DomElement.getSingleHref(ot);
      oturi = UriFactory.parseUri(othref);
    }
    else {
      if (DomElement.hasChild(ot, Consts.N_custom)) {
        oturi = CUSTOM;
      }
      else if (DomElement.hasChild(ot, Consts.N_unordered)) {
        oturi = UNORDERED;
      }
      else {
        String othref = DomElement.getSingleHref(ot);
        oturi = UriFactory.parseUri(othref);
      }
    }
    assertTrue("ordering type expected: " + orderingtype + ", was: " + oturi, oturi.equals(orderingtype));

    super.checkSortedResultSet(response, hrefs, ignore);
  }


}
