package com.sapportals.wcm.protocol.webdav.server.tests;

import org.w3c.dom.Element;

import com.sapportals.wcm.protocol.webdav.*;
import com.sapportals.wcm.util.http.IResponse;
import com.sapportals.wcm.util.uri.IUri;
import com.sapportals.wcm.util.uri.UriFactory;

/**
 * Various tests for <a href="http://greenbytes.de/tech/webdav/rfc3648.html#METHOD_ORDERPATCH">ORDERPATCH</a> method.
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author julian.reschke@greenbytes.de
 * @version $Id: OrderpatchTest.java,v 1.4 2004/05/25 16:29:22 jre Exp $
 */


public class OrderpatchTest extends TestMember
{
  private static final String collectionName = "orderpatch";

  private static final String[] desc = new String[]
  {
    "00 OP create unordered collection",
    "01 OP PROPFIND on unordered collection",
    "02 OP set ordering to DAV:custom",
    "03 OP PROPFIND on ordered (DAV:custom) collection",
    "04 OP try removing DAV:orderingtype using PROPPATCH",
    "05 OP set DAV:orderingtype to a DAV:href",
    "06 OP PROPFIND on ordered (DAV:href) collection",
    "07 OP PUT four members",
    "08 OP PROPFIND on created collection with members",
    "09 OP move one element to start",
    "10 OP PROPFIND after moving 'four' to the beginning",
    "11 OP complete reorder",
    "12 OP PROPFIND with expected order four/three/two/one",
    "13 OP flip four/three",
    "14 OP PROPFIND with expected order three/four/two/one",
    "15 OP flip four/two",
    "16 OP PROPFIND with expected order three/two/four/one",
    "17 OP move non-existing resource",
    "18 OP move non-existing resource with absolute URI",
    "19 OP explicit reorder",
    "20 OP move second entry down",
    "21 OP check order after move down",
    "22 OP move third entry up",
    "23 OP check order after move up",
    "24 OP illegal reorder request, href to exist. resource in different collection",
    "25 OP illegal reorder request, reference to exist. resource in different collection",
    "   OP (cleanup)",
  };

  public static int NrOfTests = desc.length;

  public OrderpatchTest(Integer nr)
  {
    super(nr, desc[nr.intValue()]);
  }

  private static boolean supportsRFC = false;
  private static boolean supportsOldSapImpl = false;

  protected void runTest()
  {
    if (nr == 0) logIdentifier("OP");

    String collname = "/" + collectionName;
    String respfile = "ORDERPATCH-" + TestMember.THREEDIGITS.format(nr) + ".txt";

    try {

      IUri UNORDERED = UriFactory.parseUri("dav:unordered");
      IUri CUSTOM = UriFactory.parseUri("dav:custom");

      IResponse response;

      switch (nr) {

        case 0:
          clear(request);
          request.setMethod("MKCOL");
          request.setURI(getLocalpart() + collname);
          response = requester.perform(request);
          checkResponseCode(response, 201);

          assertMethodAllowed(getLocalpart() + collname, "ORDERPATCH");

          clear(request);
          request.setMethod("OPTIONS");
          request.setURI(getLocalpart() + collname);
          response = requester.perform(request);
          checkResponseCode(response, 200);
          supportsRFC = response.getHeader("DAV").indexOf("ordered-collections") >= 0;
          supportsOldSapImpl = response.getHeader("DAV").indexOf("ordering") >= 0;
          break;

        case 1: {
            if (supportsOldSapImpl) {
              Element ot = getProperty(getLocalpart() + collname, Consts.N_orderingtype);
              assertTrue(ot != null && DomElement.hasChild(ot, Consts.N_unordered));
            }

            if (supportsRFC) {
              Element ot = getProperty(getLocalpart() + collname, Consts.N_ordering_type);
              assertTrue(UNORDERED.equals(UriFactory.parseUri(DomElement.getSingleHref(ot))));
            }
          }
          break;

        case 2: {
            if (supportsOldSapImpl) {
              clear(request);
              request.setMethod("ORDERPATCH");
              request.setURI(getLocalpart() + collname);
              request.setBody("<order xmlns='DAV:'><orderingtype><custom /></orderingtype></order>");
              response = requester.perform(request);
              checkResponseCode(response, 200);
            }

            if (supportsRFC) {
              clear(request);
              request.setMethod("ORDERPATCH");
              request.setURI(getLocalpart() + collname);
              request.setBody("<orderpatch xmlns='DAV:'><ordering-type><href>DAV:custom</href></ordering-type></orderpatch>");
              response = requester.perform(request);
              checkResponseCode(response, 200);
            }
          }
          break;

        case 3: {
            if (supportsOldSapImpl) {
              Element ot = getProperty(getLocalpart() + collname, Consts.N_orderingtype);
              assertTrue(ot != null && DomElement.hasChild(ot, Consts.N_custom));
            }

            if (supportsRFC) {
              Element ot = getProperty(getLocalpart() + collname, Consts.N_ordering_type);
              assertTrue(CUSTOM.equals(UriFactory.parseUri(DomElement.getSingleHref(ot))));
            }
          }
          break;

        case 4: {
            if (supportsOldSapImpl) {
              DavStatus s = removeProperty(getLocalpart() + collname, Consts.N_orderingtype);
              assertTrue(s.getStatus() == 403 && DavConditions.N_cannot_modify_protected_property.equals(s.getCondition()));
            }

            if (supportsRFC) {
              DavStatus s = removeProperty(getLocalpart() + collname, Consts.N_ordering_type);
              assertTrue(s.getStatus() == 403 && DavConditions.N_cannot_modify_protected_property.equals(s.getCondition()));
            }
          }
          break;

        case 5: {
            if (supportsOldSapImpl) {
              clear(request);
              request.setMethod("ORDERPATCH");
              request.setURI(getLocalpart() + collname);
              request.setBody("<order xmlns='DAV:'><orderingtype><href>mailto:julian.reschke@greenbytes.de?subject=orderpatch</href></orderingtype></order>");
              response = requester.perform(request);
              checkResponseCode(response, 200);
            }

            if (supportsRFC) {
              clear(request);
              request.setMethod("ORDERPATCH");
              request.setURI(getLocalpart() + collname);
              request.setBody("<orderpatch xmlns='DAV:'><ordering-type><href>mailto:julian.reschke@greenbytes.de?subject=orderpatch</href></ordering-type></orderpatch>");
              response = requester.perform(request);
              checkResponseCode(response, 200);
            }
          }
          break;

        case 6: {
            if (supportsOldSapImpl) {
              Element ot = getProperty(getLocalpart() + collname, Consts.N_orderingtype);
              assertTrue(ot != null);
              String href = DomElement.getSingleHref(ot);
              assertTrue("mailto:julian.reschke@greenbytes.de?subject=orderpatch".equals(href));
            }

            if (supportsRFC) {
              Element ot = getProperty(getLocalpart() + collname, Consts.N_ordering_type);
              assertTrue(ot != null);
              String href = DomElement.getSingleHref(ot);
              assertTrue("mailto:julian.reschke@greenbytes.de?subject=orderpatch".equals(href));
            }
          }
          break;

        case 7:
          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + collname + "/one");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("MKRESOURCE");
          request.setURI(getLocalpart() + collname + "/two");
          request.setBody("<propertyupdate xmlns='DAV:'>" +
                                    "<set>" +
                                    "<prop>" +
                                    "<resourcetype><redirectref/></resourcetype>" +
                                    "<reftarget><href>DAV:foobar</href></reftarget>" +
                                    "</prop>" +
                                    "</set>" +
                                    "</propertyupdate>");

          response = requester.perform(request);
          if (response.getStatus() >= 300) {
            // no links
            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + collname + "/two");
            checkResponseCode(response, 201);
          }


          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + collname + "/three");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + collname + "/four");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;

        case 8:
          checkOrdering(getLocalpart() + collname, new String[]{
            "one", "two", "three", "four"});
          break;

        case 9: {
            if (supportsOldSapImpl) {
              clear(request);
              request.setMethod("ORDERPATCH");
              request.setURI(getLocalpart() + collname);
              request.setBody("<order xmlns='DAV:'>" +
                  "<ordermember>" +
                    "<href>four</href><position><first/></position>" +
                  "</ordermember>" +
                "</order>");
              response = requester.perform(request);
              checkResponseCode(response, 200);
            }

            if (supportsRFC) {
              clear(request);
              request.setMethod("ORDERPATCH");
              request.setURI(getLocalpart() + collname);
              request.setBody("<orderpatch xmlns='DAV:'>" +
                  "<order-member>" +
                    "<segment>four</segment><position><first/></position>" +
                  "</order-member>" +
                "</orderpatch>");
              response = requester.perform(request);
              checkResponseCode(response, 200);
            }
          }
          break;

        case 10:
          checkOrdering(getLocalpart() + collname, new String[]{
            "four", "one", "two", "three"});
          break;

        case 11: {
          clear(request);
          request.setMethod("ORDERPATCH");
          request.setURI(getLocalpart() + collname);
          if (supportsRFC) {
            request.setBody("<orderpatch xmlns='DAV:'>" +
                "<order-member>" +
                  "<segment>one</segment><position><last/></position>" +
                "</order-member>" +
                "<order-member>" +
                  "<segment>two</segment><position><first/></position>" +
                "</order-member>" +
                "<order-member>" +
                  "<segment>four</segment><position><before><segment>two</segment></before></position>" +
                "</order-member>" +
                "<order-member>" +
                  "<segment>three</segment><position><after><segment>four</segment></after></position>" +
                "</order-member>" +
              "</orderpatch>");
          }
          else {
            request.setBody("<order xmlns='DAV:'>" +
                "<ordermember>" +
                  "<href>one</href><position><last/></position>" +
                "</ordermember>" +
                "<ordermember>" +
                  "<href>two</href><position><first/></position>" +
                "</ordermember>" +
                "<ordermember>" +
                  "<href>four</href><position><before><segment>two</segment></before></position>" +
                "</ordermember>" +
                "<ordermember>" +
                  "<href>three</href><position><after><segment>four</segment></after></position>" +
                "</ordermember>" +
              "</order>");
          }
          response = requester.perform(request);
          checkResponseCode(response, 200);
          }
          break;

        case 12:
          checkOrdering(getLocalpart() + collname, new String[]{
            "four", "three", "two", "one"});
          break;

        case 13: {
          clear(request);
          request.setMethod("ORDERPATCH");
          request.setURI(getLocalpart() + collname);
          if (supportsRFC) {
            request.setBody("<orderpatch xmlns='DAV:'>" +
                "<order-member>" +
                  "<segment>three</segment><position><before><segment>four</segment></before></position>" +
                "</order-member>" +
              "</orderpatch>");
          }
          else {
            request.setBody("<order xmlns='DAV:'>" +
                "<ordermember>" +
                  "<href>three</href><position><before><segment>four</segment></before></position>" +
                "</ordermember>" +
              "</order>");
          }
          response = requester.perform(request);
          checkResponseCode(response, 200);
          }
          break;

        case 14:
          checkOrdering(getLocalpart() + collname, new String[]{
            "three", "four", "two", "one"});
          break;

        case 15: {
            clear(request);
            request.setMethod("ORDERPATCH");
            request.setURI(getLocalpart() + collname);
            if (supportsRFC) {
              request.setBody("<orderpatch xmlns='DAV:'>" +
                  "<order-member>" +
                    "<segment>four</segment><position><after><segment>two</segment></after></position>" +
                  "</order-member>" +
                "</orderpatch>");
            }
            else {
              request.setBody("<order xmlns='DAV:'>" +
                  "<ordermember>" +
                    "<href>four</href><position><after><segment>two</segment></after></position>" +
                  "</ordermember>" +
                "</order>");
            }
            response = requester.perform(request);
            checkResponseCode(response, 200);
          }
          break;

        case 16:
          checkOrdering(getLocalpart() + collname, new String[]{
            "three", "two", "four", "one"});
          break;

        case 17: {
            clear(request);
            request.setMethod("ORDERPATCH");
            request.setURI(getLocalpart() + collname);
            String body =
              supportsRFC
              ?
              "<orderpatch xmlns='DAV:'>" +
                "<order-member>" +
                  "<segment>four</segment><position><before><segment>two</segment></before></position>" +
                "</order-member>" +
                "<order-member>" +
                  "<segment>five</segment><position><after><segment>four</segment></after></position>" +
                "</order-member>" +
                "<order-member>" +
                  "<segment>two</segment><position><after><segment>five</segment></after></position>" +
                "</order-member>" +
              "</orderpatch>"
              :
              "<order xmlns='DAV:'>" +
                "<ordermember>" +
                  "<href>four</href><position><before><segment>two</segment></before></position>" +
                "</ordermember>" +
                "<ordermember>" +
                  "<href>five</href><position><after><segment>four</segment></after></position>" +
                "</ordermember>" +
                "<ordermember>" +
                  "<href>two</href><position><after><segment>five</segment></after></position>" +
                "</ordermember>" +
              "</order>";
            checkMultistatus(body, respfile);
          }
          break;

        case 18: {
            clear(request);
            request.setMethod("ORDERPATCH");
            request.setURI(getLocalpart() + collname);
            String body = supportsRFC
              ?
                "<orderpatch xmlns='DAV:'>" +
                  "<order-member>" +
                    "<segment>somewhere_else/four</segment><position><first /></position>" +
                  "</order-member>" +
                  "<order-member>" +
                    "<segment>/absolute/somewhere_else/four</segment><position><first /></position>" +
                  "</order-member>" +
                "</orderpatch>"
              :
                "<order xmlns='DAV:'>" +
                  "<ordermember>" +
                    "<href>somewhere_else/four</href><position><first /></position>" +
                  "</ordermember>" +
                  "<ordermember>" +
                    "<href>/absolute/somewhere_else/four</href><position><first /></position>" +
                  "</ordermember>" +
                "</order>";
            checkMultistatus(body, respfile);
          }
          break;

        case 19: {
            clear(request);
            request.setMethod("ORDERPATCH");
            request.setURI(getLocalpart() + collname);
            request.setBody(
              supportsRFC ? "<orderpatch xmlns='DAV:'>" +
                "<order-member>" +
                  "<segment>one</segment><position><first/></position>" +
                "</order-member>" +
                "<order-member>" +
                  "<segment>four</segment><position><last/></position>" +
                "</order-member>" +
                "<order-member>" +
                  "<segment>three</segment><position><before><segment>four</segment></before></position>" +
                "</order-member>" +
                "<order-member>" +
                  "<segment>two</segment><position><after><segment>one</segment></after></position>" +
                "</order-member>" +
              "</orderpatch>" :
              "<order xmlns='DAV:'>" +
                "<ordermember>" +
                  "<href>one</href><position><first/></position>" +
                "</ordermember>" +
                "<ordermember>" +
                  "<href>four</href><position><last/></position>" +
                "</ordermember>" +
                "<ordermember>" +
                  "<href>three</href><position><before><segment>four</segment></before></position>" +
                "</ordermember>" +
                "<ordermember>" +
                  "<href>two</href><position><after><segment>one</segment></after></position>" +
                "</ordermember>" +
              "</order>");
            response = requester.perform(request);
            checkResponseCode(response, 200);
          }
          break;

        case 20: {
            clear(request);
            request.setMethod("ORDERPATCH");
            request.setURI(getLocalpart() + collname);
            request.setBody(supportsRFC ?
              "<orderpatch xmlns='DAV:'>" +
                  "<order-member>" +
                    "<segment>two</segment><position><after><segment>three</segment></after></position>" +
                  "</order-member>" +
                "</orderpatch>" : "<order xmlns='DAV:'>" +
                "<ordermember>" +
                  "<href>two</href><position><after><segment>three</segment></after></position>" +
                "</ordermember>" +
              "</order>");
            response = requester.perform(request);
            checkResponseCode(response, 200);
          }
          break;

        case 21:
          checkOrdering(getLocalpart() + collname, new String[]{
            "one", "three", "two", "four"});
          break;

        case 22: {
            clear(request);
            request.setMethod("ORDERPATCH");
            request.setURI(getLocalpart() + collname);
            request.setBody(supportsRFC ? "<orderpatch xmlns='DAV:'>" +
                "<order-member>" +
                  "<segment>two</segment><position><before><segment>three</segment></before></position>" +
                "</order-member>" +
              "</orderpatch>" : "<order xmlns='DAV:'>" +
                "<ordermember>" +
                  "<href>two</href><position><before><segment>three</segment></before></position>" +
                "</ordermember>" +
              "</order>");
            response = requester.perform(request);
            checkResponseCode(response, 200);
          }
          break;

        case 23:
          checkOrdering(getLocalpart() + collname, new String[]{
            "one", "two", "three", "four"});
          break;

        case 24: {
            clear(request);
            request.setMethod("ORDERPATCH");
            request.setURI(getLocalpart() + collname);
            String body = supportsRFC ? "<orderpatch xmlns='DAV:'>" +
                "<order-member>" +
                  "<segment>../Worddatei.doc</segment><position><first /></position>" +
                "</order-member>" +
              "</orderpatch>" : "<order xmlns='DAV:'>" +
                "<ordermember>" +
                  "<href>../Worddatei.doc</href><position><first /></position>" +
                "</ordermember>" +
              "</order>";
            checkMultistatus(body, respfile);
          }
          break;

        case 25: {
            clear(request);
            request.setMethod("ORDERPATCH");
            request.setURI(getLocalpart() + collname);
            String body = supportsRFC ? "<orderpatch xmlns='DAV:'>" +
              "<order-member>" +
                "<segment>four</segment><position><after><segment>"+getLocalpart() + "/Worddatei.doc"+"</segment></after></position>" +
              "</order-member>" +
            "</orderpatch>" : "<order xmlns='DAV:'>" +
              "<ordermember>" +
                "<href>four</href><position><after><segment>"+getLocalpart() + "/Worddatei.doc"+"</segment></after></position>" +
              "</ordermember>" +
            "</order>";
            checkMultistatus(body, respfile);
          }
          break;

        default:
          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + collname);
          response = requester.perform(request);
          checkResponseCode(response, 200, 204);
          break;
      }
    }
    catch (Exception ex) {
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }

  private void checkOrdering(String collection, String names[]) throws Exception {
    request.setMethod("PROPFIND");
    request.setURI(collection);
    request.setHeader("Depth", "1");
    request.setBody("<propfind xmlns='DAV:'><prop><resourcetype/></prop></propfind>");
    IResponse response = requester.perform(request);
    checkResponseCode(response, 207);
    String uris[] = new String[names.length];
    for (int i = 0; i < names.length; i++) uris[i] = collection + "/" + names[i];
    checkSortedResultSet(response, uris, collection + "/");
  }
}
