package com.sapportals.wcm.protocol.webdav.server.tests;

import java.util.List;

import org.w3c.dom.Element;

import com.sapportals.wcm.protocol.webdav.Consts;
import com.sapportals.wcm.protocol.webdav.DavConditions;
import com.sapportals.wcm.protocol.webdav.DavStatus;
import com.sapportals.wcm.protocol.webdav.DomElement;
import com.sapportals.wcm.protocol.webdav.dompath.DomPath;
import com.sapportals.wcm.repository.Content;
import com.sapportals.wcm.util.http.IResponse;

/**
 * Various tests for PROPFIND method.
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author martin.boettcher@greenbytes.de
 * @author julian.reschke@greenbytes.de
 * @version $Id: PropfindTest.java,v 1.10 2004/09/02 09:29:35 jre Exp $
 */

public class PropfindTest extends TestMember {
  
  private static final String[] desc = new String[] {
    "00 PROPFIND with allprop",
    "01 PROPFIND with allprop and depth 0",
    "02 PROPFIND with allprop and depth 1",
    "03 PROPFIND with allprop and depth infinity",
    "04 PROPFIND with propname",
    "05 PROPFIND with propname and depth 0",
    "06 PROPFIND with propname and depth 1",
    "07 PROPFIND with propname and depth infinity",
    "08 PROPFIND with prop",
    "09 PROPFIND with prop and depth 0",
    "10 PROPFIND with prop and depth 1",
    "11 PROPFIND with prop and depth infinity",
    "12 PROPFIND with legal but unknown extension elements",
    "13 PROPFIND supportedlock",
    "14 PROPFIND getcontentlanguage",
    "15 PROPFIND/HEAD getcontenttype",
    "16 PROPFIND parent WebDAV namespace?",
    "17 PROPFIND non-wellformed request body",
    "18 PROPFIND invalid request body",
    "   PROPFIND clean up",
  };

  public static int NrOfTests = desc.length;

  private static final DomPath DP_unwanted = new DomPath(
    "/*/*/D:propstat/D:prop/D:acl | /*/*/D:propstat/D:prop/D:acl-semantics | /*/*/D:propstat/D:prop/D:acl-restrictions | /*/*/D:propstat/D:prop/D:supported-privilege-set "
    +"| /*/*/D:propstat/D:prop/D:current-user-privilege-set | /*/*/D:propstat/D:prop/D:principal-collection-set"
    +"| /*/*/D:propstat/D:prop/D:owner | /*/*/D:propstat/D:prop/D:displayname | /*/*/D:propstat/D:prop/D:getcontentlanguage"
    +"| /*/*/D:propstat/D:prop/D:supportedlock | /*/*/D:propstat/D:prop/D:inherited-acl-set | /*/*/D:propstat/D:prop/*[local-name()='property-displayname-set']"
    +"| /*/*/D:propstat/D:prop/D:version-history-collection-set | /*/*/D:propstat/D:prop/D:workspace-collection-set"
    +"| /*/*/D:propstat/D:prop/D:workspace | /*/*/D:propstat/D:prop/D:resource-id | /*/*/D:propstat/D:prop/D:parent-set"
    +"| /*/*/D:propstat/D:prop/D:orderingtype | /*/*/D:propstat/D:prop/D:ordering-type | /*/*/D:propstat/D:prop/D:getetag"
    +"| /*/*/D:propstat/D:prop/D:group | /*/*/D:propstat/D:prop/D:creationdate | /*/*/D:propstat/D:prop/*[local-name()='appproperties']"
    +"| /*/*/D:propstat/D:prop/*[local-name()='personalized-appproperties']| /*/*/D:propstat/D:prop/*[local-name()='repository']"
    +"| /*/*/D:propstat/D:prop/D:source",
    DomPath.DEFAULT_NAMESPACE_BINDINGS);

  public PropfindTest(Integer nr) {
    super(nr, desc[nr.intValue()]);
  }

  protected void runTest() {

    if (nr == 0) logIdentifier("PROPFIND");

    final String allprops = "<propfind xmlns='DAV:'><allprop /></propfind>";
    final String propname = "<propfind xmlns='DAV:'><propname /></propfind>";
    final String prop = "<propfind xmlns='DAV:'><prop><resourcetype /><wrxrxrxyzr xmlns='wlwmx:'/></prop></propfind>";

    try {
      
      String respfile = "PROPFIND-" + TestMember.THREEDIGITS.format(nr) + ".txt";
      IResponse response;

      switch (nr) {
        
        case 0:
          clear(request);
          request.setMethod("PROPFIND");
          request.setHeader("translate", "f");
          request.setURI(getLocalpart());
          request.setBody(allprops);
          response = requester.perform(request);
          checkResponseCode(response, 207, 403);
          if (response.getStatus() == 207) {
            checkResponseDocument(response, respfile, DP_unwanted);
          }
          else {
            checkErrorElement(response, DavConditions.N_SAP_depth_infinity_allowed);
          }
          break;

        case 1:
          clear(request);
          request.setMethod("PROPFIND");
          request.setHeader("translate", "f");
          request.setURI(getLocalpart());
          request.setHeader("Depth", "0");
          checkMultistatus(allprops, respfile, DP_unwanted);
          break;

        case 2:
          clear(request);
          request.setMethod("PROPFIND");
          request.setHeader("translate", "f");
          request.setURI(getLocalpart());
          request.setHeader("Depth", "1");
          checkMultistatus(allprops, respfile,  DP_unwanted);
          break;

        case 3:
          clear(request);
          request.setMethod("PROPFIND");
          request.setHeader("translate", "f");
          request.setURI(getLocalpart());
          request.setHeader("Depth", "infinity");
          request.setBody(allprops);
          response = requester.perform(request);
          checkResponseCode(response, 207, 403);
          if (response.getStatus() == 207) {
            checkResponseDocument(response, respfile, DP_unwanted);
          }
          else {
            checkErrorElement(response, DavConditions.N_SAP_depth_infinity_allowed);
          }
          break;

        case 4:
          clear(request);
          request.setMethod("PROPFIND");
          request.setHeader("translate", "f");
          request.setURI(getLocalpart());
          request.removeHeader("Depth");
          request.setBody(propname);
          response = requester.perform(request);
          checkResponseCode(response, 207, 403);
          if (response.getStatus() == 207) {
            checkResponseDocument(response, respfile, DP_unwanted);
          }
          else {
            checkErrorElement(response, DavConditions.N_SAP_depth_infinity_allowed);
          }
          break;

        case 5:
          clear(request);
          request.setMethod("PROPFIND");
          request.setHeader("translate", "f");
          request.setURI(getLocalpart());
          request.setHeader("Depth", "0");
          checkMultistatus(propname, respfile, DP_unwanted);
          break;

        case 6:
          clear(request);
          request.setMethod("PROPFIND");
          request.setHeader("translate", "f");
          request.setURI(getLocalpart());
          request.setHeader("Depth", "1");
          checkMultistatus(propname, respfile, DP_unwanted);
          break;

        case 7:
          clear(request);
          request.setMethod("PROPFIND");
          request.setHeader("translate", "f");
          request.setURI(getLocalpart());
          request.setHeader("Depth", "infinity");
          request.setBody(propname);
          response = requester.perform(request);
          checkResponseCode(response, 207, 403);
          if (response.getStatus() == 207) {
            checkResponseDocument(response, respfile, DP_unwanted);
          }
          else {
            checkErrorElement(response, DavConditions.N_SAP_depth_infinity_allowed);
          }
          break;

        case 8:
          clear(request);
          request.setMethod("PROPFIND");
          request.setURI(getLocalpart());
          request.setBody(prop);
          response = requester.perform(request);
          checkResponseCode(response, 207, 403);
          if (response.getStatus() == 207) {
            checkResponseDocument(response, respfile, null);
          }
          else {
            checkErrorElement(response, DavConditions.N_SAP_depth_infinity_allowed);
          }
          break;

        case 9:
          clear(request);
          request.setMethod("PROPFIND");
          request.setURI(getLocalpart());
          request.setHeader("Depth", "0");
          checkMultistatus(prop, respfile);
          break;

        case 10:
          clear(request);
          request.setMethod("PROPFIND");
          request.setURI(getLocalpart());
          request.setHeader("Depth", "1");
          checkMultistatus(prop, respfile);
          break;

        case 11:
          clear(request);
          request.setMethod("PROPFIND");
          request.setURI(getLocalpart());
          request.setHeader("Depth", "infinity");
          request.setBody(prop);
          response = requester.perform(request);
          checkResponseCode(response, 207, 403);
          if (response.getStatus() == 207) {
            checkResponseDocument(response, respfile, DP_unwanted);
          }
          else {
            checkErrorElement(response, DavConditions.N_SAP_depth_infinity_allowed);
          }
          break;

        case 12:
          clear(request);
          request.setMethod("PROPFIND");
          request.setURI(getLocalpart());
          request.setHeader("Depth", "1");
          request.setBody("<propfind xmlns='DAV:'><whoAmI /><prop><resourcetype/></prop></propfind>");
          response = requester.perform(request);
          assertTrue(response.getStatus() == 207);

          clear(request);
          request.setMethod("PROPFIND");
          request.setURI(getLocalpart());
          request.setHeader("Depth", "1");
          request.setBody("<propfind xmlns='DAV:'><whoAmI xmlns='xyz:' /><prop><resourcetype/></prop></propfind>");
          response = requester.perform(request);
          assertTrue(response.getStatus() == 207);
          break;

        case 13:
          clear(request);
          request.setMethod("PROPFIND");
          request.setURI(getLocalpart());
          request.setHeader("Depth", "1");
          checkMultistatus("<propfind xmlns='DAV:'><prop><supportedlock /></prop></propfind>",
            respfile, null);
          break;

        case 14:
          clear(request);
          request.setMethod("PROPFIND");
          request.setURI(getLocalpart());
          request.setHeader("Depth", "1");
          checkMultistatus("<propfind xmlns='DAV:'><prop><getcontentlanguage /></prop></propfind>",
            respfile, null);
          break;

        case 15:
          {
            Element getcontenttype = getProperty(getLocalpart() + "/UTF8.TXT", Consts.N_getcontenttype);
            assertTrue(getcontenttype != null);
            String propfindContentType = DomElement.getText(getcontenttype);

            clear(request);
            request.setMethod("HEAD");
            request.setURI(getLocalpart() + "/UTF8.TXT");
            response = requester.perform(request);
            checkResponseCode(response, 200);

            String headContentType = response.getHeader("content-type");

            boolean encodingPreserved = propfindContentType.indexOf("charset=") >= 0;
            logComment("putPreservesCharset:" + encodingPreserved);
            
            Content pct = new Content(null, propfindContentType, 0);
            Content hct = new Content(null, headContentType, 0);
            
            String expected = "text/plain";
            assertTrue("was: '" + hct.getContentType() + "', should be '" +  expected + "'", expected.equals(hct.getContentType()));
            assertTrue("was: '" + pct.getContentType() + "', should be '" +  expected + "'", expected.equals(pct.getContentType()));
            
            if (encodingPreserved) {
              expected = "UTF-8";
              assertTrue("was: '" + hct.getEncoding() + "', should be '" +  expected + "'", expected.equals(hct.getEncoding()));
              assertTrue("was: '" + pct.getEncoding() + "', should be '" +  expected + "'", expected.equals(pct.getEncoding()));
            }
          }
          break;

        // check that the base collection appears as member of it's parent collection
        case 16: {
          clear(request);
          request.setMethod("PROPFIND");
          request.setURI(getLocalpart() + "/..");
          request.setHeader("Depth", "1");
          request.setBody("<propfind xmlns='DAV:'><prop><resourcetype /></prop></propfind>");
          response = requester.perform(request);
          if (response.getStatus() == 207) {
            List hrefs = getHrefsFromMultistatus(response.getDocument());
            assertTrue("if parent collection is a WebDAV collection, ensure it contains the test collection", hrefs.contains(getLocalpart() + "/"));
          }
          else {
            logComment("dav-parent:false");  
          }
          break;
        }
          
        case 17 :
					clear(request);
					request.setMethod("PROPFIND");
					request.setURI(getLocalpart());
					request.setHeader("Depth", "0");
					request.setBody("<propfind foo xmlns='DAV:'></propfind>");
					response = requester.perform(request);
					checkStatusAndErrorElement(response,
							DavStatus.SC_BAD_REQUEST,
							DavConditions.N_SAP_request_body_wellformed);
					break;

				case 18 :
					clear(request);
					request.setMethod("PROPFIND");
					request.setURI(getLocalpart());
					request.setHeader("Depth", "0");
					request.setBody("<propfind xmlns='DAV:'></propfind>");
					response = requester.perform(request);
					checkStatusAndErrorElement(response,
							DavStatus.SC_UNPROCESSABLE_ENTITY,
							DavConditions.N_SAP_request_body_valid);
					break;

        default:
        	break;
      }
      
    }
    catch(Exception ex) {
      ex.printStackTrace(System.err);
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }

}
