package com.sapportals.wcm.protocol.webdav.server.tests;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;

import com.sapportals.wcm.WcmException;
import com.sapportals.wcm.protocol.webdav.DavConditions;
import com.sapportals.wcm.util.http.IRequester;
import com.sapportals.wcm.util.http.IResponse;
import com.sapportals.wcm.util.uri.URICodec;

/**
 * Various tests for PUT method.
 * <p>
 * Copyright (c) SAP AG 2001-2005
 * @author martin.boettcher@greenbytes.de
 * @author stefan.eissing@greenbytes.de
 * @author julian.reschke@greenbytes.de
 * @version $Id: PutTest.java,v 1.14 2005/01/21 17:15:37 jre Exp $
 */

public class PutTest extends TestMember
{
  private static final String[] desc = new String[]
  {
    "00 PUT with improper URI no data",
    "01 PUT into new URI no data",
    "02 PUT into existing URI no data",
    "03 PUT with unknown collection no data",
    "04 PUT with improper URI with data",
    "05 PUT into new URI with data",
    "06 PUT into existing URI with data",
    "07 PUT with unknown collection with data",
    "08 PUT make new temporary collection",
    "09 PUT with special chars in name",
    "10 PUT propfind on result collection",
    "11 PUT with badly-encoded Euro sign (webfolder hack)",
    "12 PUT check change of ETAG",
    "13 PUT with \\uE8 in uri",
    "14 PUT overwrite with no content length header",
    "15 PUT overwrite/truncate",
    "16 PUT truncated",
    "17 PUT with uri containing trailing blanks",
		"18 PUT and GET with japanese characters in filename",
    "19 PUT conditional (when not present)",
    "20 PUT conditional (when present)",
    "21 PUT with '+' in uri",
    "22 PUT with if (tagged) condition on non-existing resource",
    "23 PUT with malformed If header",
    "24 PUT with leading space in name",
    "25 PUT with only 3 spaces in name",
    "26 PUT with Expect 100-continue header",
    "   PUT cleanup of created resources",
  };

  public static int NrOfTests = desc.length;

  public PutTest(Integer nr)
  {
    super(nr, desc[nr.intValue()]);
  }

  static private String eTag = null;
  static private long timer = 0;

  private static boolean createdSpaceResource;

  protected void runTest()
  {
    if (nr == 0) super.logIdentifier("PUT");

    String respfile = "PUT-" + TestMember.THREEDIGITS.format(nr) + ".txt";

    try
    {
      IResponse response;
      clear(request);
      request.setMethod("PUT");

      switch (nr)
      {
        case 0:
          // prepare resource for later etag comparison
          timer = System.currentTimeMillis();
          request.clear();
          request.setMethod("PUT");
          request.setBody("Dies ist der alte Body!");
          request.setURI(getLocalpart() + "/etagtest");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          request.clear();
          request.setMethod("HEAD");
          request.setHeader("Translate", "f");
          request.setURI(getLocalpart() + "/etagtest");
          response = requester.perform(request);
          eTag = response.getHeader("Etag");
          assertTrue(eTag != null);

          request.clear();
          request.setMethod("PUT");
          request.setURI(getLocalpart());
          response = requester.perform(request);
          checkResponseCode(response, 409, 405);
          break;

        case 1:
          request.setURI(getLocalpart() + "/abc.xyz");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;

        case 2:
          request.setURI(getLocalpart() + "/abc.xyz");
          request.setContentType("text/plain");
          response = requester.perform(request);
          checkResponseCode(response, 200, 204);
          break;

        case 3:
          request.setURI(getLocalpart() + "/a/b/c.xyz");
          response = requester.perform(request);
          checkResponseCode(response, 404, 409, 412);
          break;

        case 4:
          requester.close();
          request.setBody("Dies ist der Body!");
          request.setURI(getLocalpart());
          response = requester.perform(request);
          checkResponseCode(response, 409, 405);
          break;

        case 5:
          request.setBody("Dies ist der Body!");
          request.setURI(getLocalpart() + "/abcd.xyz");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;

        case 6:
          request.setURI(getLocalpart() + "/abcd.xyz");
          request.setBody("Dies ist der Body!");
          request.setContentType("text/plain");
          response = requester.perform(request);
          checkResponseCode(response, 200);
          break;

        case 7:
          request.setBody("Dies ist der Body!");
          request.setURI(getLocalpart() + "/a/b/c.xyz");
          response = requester.perform(request);
          checkResponseCode(response, 404, 409, 412);
          break;

        case 8:
          request.setMethod("MKCOL");
          request.setURI(getLocalpart() + "/tmp");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;

        case 9: {
          String[] names = new String[] {
            "$test.txt", "~test.txt", ".test.txt", /*"<test>.txt", */
            "test%.txt", "test%20.txt",
          };
          for (int i = 0; i < names.length; ++i) {
            String name = URICodec.Encode(names[i]);
            clear(request);
            request.setMethod("PUT");
            request.setBody("Dies ist der Body!");
            request.setURI(getLocalpart() + "/tmp/"+name);
            response = requester.perform(request);
            //assertTrue("could not put "+name+": "+response.getStatus(),
            //           response.getStatus() == 201);
          }
          break;
        }

        case 10:
          request.setMethod("PROPFIND");
          request.setURI(getLocalpart() + "/tmp");
          request.setHeader("Depth", "1");
          response = this.checkMultistatus(
            "<propfind xmlns='DAV:'><prop><resourcetype/></prop></propfind>",
            respfile);
          break;

        // TEST HACK for old Microsoft webfolder client
        case 11:
          request.setBody("Dies ist der Body!");
          request.setURI(getLocalpart() + "/Euro-\u0080");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("HEAD");
          request.setURI(getLocalpart() + URICodec.Encode("/Euro-\u20ac"));
          response = requester.perform(request);
          assertTrue(response.getStatus() == 200);
          break;

        case 12:

          if (System.currentTimeMillis() < timer + 1000) {
            Thread.sleep(1000);
          }

          clear(request);
          request.setMethod("PUT");
          request.setBody("Dies ist der neue Body!");
          request.setURI(getLocalpart() + "/etagtest");
          response = requester.perform(request);
          checkResponseCode(response, 200, 201);

          clear(request);
          request.setMethod("HEAD");
          request.setHeader("Translate", "f");
          request.setURI(getLocalpart() + "/etagtest");
          response = requester.perform(request);
          assertTrue ("ETag was: " + eTag + " and did not change",
            ! eTag.equalsIgnoreCase(response.getHeader("Etag")));
          break;

        case 13:
          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/uri-\u00E8test");
          request.setBody("Dies ist der Body!");
          response = requester.perform(request);
          checkResponseCode(response, 200, 201);

          clear(request);
          request.setMethod("HEAD");
          request.setHeader("Translate", "f");
          request.setURI(getLocalpart() + URICodec.EncodePath("/uri-\u00E8test"));
          response = requester.perform(request);
          checkResponseCode(response, 200);
          break;

        case 14:
          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/abc.xyz");
          byte byt[] = new byte[]{1,2,3,4,5,6,7};
          request.setBody(new java.io.ByteArrayInputStream(byt));
          response = requester.perform(request);
          checkResponseCode(response, 200, 201);

          clear(request);
          request.setMethod("HEAD");
          request.setURI(getLocalpart() + "/abc.xyz");
          request.setHeader("Translate", "f");
          request.setHeader("Accept-Encoding", "");
          response = requester.perform(request);
          checkResponseCode(response, 200);
          assertTrue("content-length missing", response.getHeader("Content-Length") != null);
          assertTrue("7".equals(response.getHeader("Content-Length").trim()));
          break;

        case 15:
          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/abc.xyz");
          request.setBody("abc");
          response = requester.perform(request);
          checkResponseCode(response, 200, 201);

          clear(request);
          request.setMethod("GET");
          request.setURI(getLocalpart() + "/abc.xyz");
          request.setHeader("Translate", "f");
          request.expectsResponseStream(true);
          response = requester.perform(request);
          checkResponseCode(response, 200);
          assertTrue("content-length missing", response.getHeader("Content-Length") != null);
          assertTrue("3".equals(response.getHeader("Content-Length").trim()));
          String res = getContent(response.getStream());
          assertTrue("abc".equals(res));
          break;

        case 16: {
            // claim that the request body has len bytes, but truncate earlier 
            int len        = 1000000;
            int truncateAt = 100000;
            boolean gotExc = false;
            
            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + "/truncated");
            request.setHeader("content-length", "" + len);
            request.setBody(new DataGenerator(0, len, requester, truncateAt));
            try {
              response = requester.perform(request);
            }
            catch (WcmException ex) {
              gotExc = true;
            }
            
            assertTrue("expected exception not thrown", gotExc);
            
            // make sure resource not created
            clear(request);
            request.setMethod("HEAD");
            request.setURI(getLocalpart() + "/truncated");
            response = requester.perform(request);
            checkResponseCode(response, 404);
          }
          break;

        case 17:
          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/abcde.xyz%20");
          request.setBody("abcde");
          response = requester.perform(request);
          checkResponseCode(response, 403, 409);
          checkErrorElement(response, DavConditions.N_name_allowed);

          break;
          
        case 18:
        	clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/2-1_CRM%e3%83%9e%e3%82%b9%e3%82%bf%e2%91%a2%e8%a3%bd%e5%93%81.txt");
          request.setBody("abc");
          response = requester.perform(request);
          checkResponseCode(response, 200, 201);

          clear(request);
          request.setMethod("GET");
          request.setURI(getLocalpart() + "/2-1_CRM%e3%83%9e%e3%82%b9%e3%82%bf%e2%91%a2%e8%a3%bd%e5%93%81.txt");
          request.setHeader("Translate", "f");
          request.expectsResponseStream(true);
          response = requester.perform(request);
          checkResponseCode(response, 200);
          assertTrue("content-length missing", response.getHeader("Content-Length") != null);
          assertTrue("3".equals(response.getHeader("Content-Length").trim()));
          res = getContent(response.getStream());
          assertTrue("abc".equals(res));
          break;                  

        case 19:
          // <http://greenbytes.de/tech/webdav/rfc2616.html#rfc.section.14.26.p.2>
          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/when-not-present");
          request.setHeader("if-none-match", "*");
          request.setBody("abc");
          response = requester.perform(request);
          checkResponseCode(response, 201, 200);

          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/when-not-present");
          request.setHeader("if-none-match", "*");
          request.setBody("def");
          response = requester.perform(request);
          checkResponseCode(response, 412);
          break;                  

        case 20:
          // <http://greenbytes.de/tech/webdav/rfc2616.html#rfc.section.14.24.p.7>
          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/when-present");
          request.setHeader("if-match", "*");
          request.setBody("abc");
          response = requester.perform(request);
          checkResponseCode(response, 412);

          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/when-present");
          request.setBody("abc");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/when-present");
          request.setHeader("if-match", "*");
          request.setBody("def");
          response = requester.perform(request);
          checkResponseCode(response, 200);
          break;                  

        case 21:
          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/uri+test");
          request.setBody("Dies ist der Body!");
          response = requester.perform(request);
          checkResponseCode(response, 200, 201);

          clear(request);
          request.setMethod("HEAD");
          request.setHeader("Translate", "f");
          request.setURI(getLocalpart() + URICodec.EncodePath("/uri+test"));
          response = requester.perform(request);
          checkResponseCode(response, 200);
          break;
          
        case 22:
          // see cvstrac #34
          // 
          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/foobar");
          request.setHeader("Translate", "F");
          request.setHeader("if", "<" + getLocalpart() + "/gibtesnicht> ([\"some-etag\"])");
          response = requester.perform(request);
          checkResponseCode(response, 207);
          break;

        case 23:
          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/Worddatei.doc");
          request.setHeader("Translate", "F");
          request.setHeader("if", "<" + getLocalpart() + "/gibtesnicht> [\"some-etag\"])");
          response = requester.perform(request);
          checkResponseCode(response, 400);
          checkErrorElement(response, DavConditions.N_SAP_legal_if_header_value);
          break;

        case 24:
          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/%20uritest");
          request.setBody("Dies ist der Body!");
          response = requester.perform(request);
          checkResponseCode(response, 403, 201);

          boolean created = (response.getStatus() == 201);
          clear(request);
          request.setMethod("HEAD");
          request.setHeader("Translate", "f");
          request.setURI(getLocalpart() + "/%20uritest");
          response = requester.perform(request);
          checkResponseCode(response, (created? 200 : 404));
          break;
          
        case 25:
          // This test case may fail if the jdk caches the results of getCanonicalPath and the
          // repository manager is based on java.io.File. This was introduced
          // as part of jdk 1.4.2; and can be disabled by setting "sun.io.useCanonCaches=false".

          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/%20%20%20");
          request.setBody("Dies ist der Body!");
          response = requester.perform(request);
          checkResponseCode(response, 403);

          createdSpaceResource = true;

          clear(request);
          request.setMethod("HEAD");
          request.setHeader("Translate", "f");
          request.setURI(getLocalpart() + "/%20%20%20");
          response = requester.perform(request);
          checkResponseCode(response, 404);
          break;

        case 26:
          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/Expect.txt");
          request.setHeader("Expect", "100-continue");
          request.setBody("this is a test body which should only be send after an 100");
          request.setContentType("text/plain");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;


        default:
          String[] cleanup = new String[] {
            "/tmp",
            "/abc.xyz",
            "/abcd.xyz",
            "/abcde.xyz",
            "/Euro-\u0080",
            "/etagtest",
            "/uri-%c3%a8test",
            "/when-present",
            "/when-not-present",
            "/uri+test",
            "/%20uritest",
            "/Expect.txt",
          };
          
          for (int i = 0; i < cleanup.length; ++i) {
            
            clear(request);
            request.setMethod("DELETE");
            request.setURI(getLocalpart() + cleanup[i]);
            response = requester.perform(request);
            
            checkResponseCode(response, 204, 404);
          }
          
          {
            clear(request);
            request.setMethod("DELETE");
            request.setURI(getLocalpart() + "/truncated");
            response = requester.perform(request);
            if (response.getStatus() == 500) {
              // this is what happens if the requests are overlapping and the system is still trying to write to the file
              checkErrorElement(response, DavConditions.N_SAP_io_error);
            }
            else {
              checkResponseCode(response, 204, 404);
            }
          }
          
          if (createdSpaceResource) {
            clear(request);
            request.setMethod("DELETE");
            request.setURI(getLocalpart() + "/%20%20%20");
            response = requester.perform(request);
            checkResponseCode(response, 204, 200, 404);
          }
          break;
      }
    }
    catch (Exception ex) {
      ex.printStackTrace(System.err);
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }

  private String getContent(InputStream is) throws IOException {
    if (is == null)
      return null;

    byte[] buffer = new byte[64 * 1024];
    ByteArrayOutputStream bos = new ByteArrayOutputStream();
    int len;
    while ((len = is.read(buffer)) != -1) {
      bos.write(buffer, 0, len);
    }
    is.close();

    return bos.toString();
  }
  
  private class DataGenerator extends InputStream {

    private int pattern;
    private int length;
    private int position;
    private IRequester req;
    private int truncateAt;

    public DataGenerator (int pattern, int length, IRequester req, int truncateAt) {
      this.pattern = pattern;
      this.length = length;
      this.position = 0;
      this.req = req;
      this.truncateAt = truncateAt;
    }

    public int read() {

      if (this.position >= this.truncateAt) {
        req.close();
      }

      if (this.position++ < this.length) {
        return pattern;
      }
      else {
        return -1;
      }
    }
  }
}
