package com.sapportals.wcm.protocol.webdav.server.tests;

import java.io.File;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import org.w3c.dom.Element;

import com.sapportals.wcm.IWcmConst;
import com.sapportals.wcm.protocol.webdav.*;
import com.sapportals.wcm.protocol.webdav.dompath.DomPath;
import com.sapportals.wcm.util.http.IRequester;
import com.sapportals.wcm.util.http.IResponse;

/**
 * Various tests for reporting service.
 * <p>
 * Copyright (c) SAP AG 2003-2004
 * @author stefan.eissing@greenbytes.de
 * @author julian.reschke@greenbytes.de
 * @version $Id: ReportingTest.java,v 1.13 2005/11/28 11:46:48 mko Exp $
 */

public class ReportingTest extends TestMember {
  
  private static final String REP_NS = "http://sapportals.com/xmlns/cm/reporting";

  private static final Name PN_modifiedby = new Name(IWcmConst.SAP_WCM_NAMESPACE, "modifiedby");
  private static final Name PN_owner = new Name(IWcmConst.SAP_WCM_NAMESPACE, "owner");
  private static final Name PN_lock_owner = new Name(REP_NS, "lock-owner");

  private static final String[] desc = new String[] {
    "00 REPORTING get supported reports",
    "01 REPORTING set up test resources",
    "02 REPORTING get report schema for link-consistency",
    "03 REPORTING run link-consistency report",
    "04 REPORTING (REPORT) run link-consistency report, command delete",
    "05 REPORTING (POST) run link-consistency report, command delete",
    "06 REPORTING run link-consistency report, check that link has been deleted",
    "07 REPORTING get report schema for user-resources",
    "08 REPORTING run user-resources report (properties, owning user)",
    "09 REPORTING run user-resources report (properties, alternate user)",
    "10 REPORTING run user-resources report (locks)",
    "11 REPORTING run user-resources report (privileges)",
    "12 REPORTING user-resources set privileges, report again 1st user",
    "13 REPORTING user-resources report again 2nd user (privileges)",
    "14 REPORTING user-resources transfer privileges from user1 to user2",
    "15 REPORTING user-resources report user1 (privileges)",
    "16 REPORTING user-resources report user2 (privileges)",
    "17 REPORTING delete acl resource",
    "18 REPORTING check maxresults (parse error)",
    "19 REPORTING check maxresults (0)",
    "20 REPORTING get report schema for principal-consistency",
    "21 REPORTING principal-consistency (properties)",
    "22 REPORTING try to update DAV:owner property",
    "23 REPORTING get report schema for resource-statistics",
    "24 REPORTING get report schema for statemanagement",
    "25 REPORTING check apply-to marshalling (bad href)",
    "26 REPORTING LinkConsistency apply-to",
    "27 REPORTING UserResources apply-to",
    "28 REPORTING attempt user-in-property replacement in locked resource",
    "29 REPORTING attempt user-in-property replacement in checked-in version-controlled resource",
    "30 REPORTING run FSDB synchronization report",
    "31 REPORTING create new file/folder in FS, run FSDB synchronization report",
    "32 REPORTING remove file/folder in FS, run FSDB synchronization report",
    "33 REPORTING replace file/folder by each other in FS, run FSDB synchronization report",
    "   REPORTING cleanup"
    };

  public static int NrOfTests = desc.length;

  private static final DomPath DP_cm_synchronization_report = new DomPath("D:supported-report/D:report/*[local-name()='cm.synchronize']", 
      DomPath.DEFAULT_NAMESPACE_BINDINGS);
      
  private static final DomPath DP_link_consistency_report = new DomPath("D:supported-report/D:report/*[local-name()='link-consistency']", 
    DomPath.DEFAULT_NAMESPACE_BINDINGS);
    
  private static final DomPath DP_user_resources_report = new DomPath("D:supported-report/D:report/*[local-name()='user-resources']", 
    DomPath.DEFAULT_NAMESPACE_BINDINGS);

  private static final DomPath DP_principal_consistency_report = new DomPath("D:supported-report/D:report/*[local-name()='principal-consistency']", 
    DomPath.DEFAULT_NAMESPACE_BINDINGS);

  private static final DomPath DP_resource_statistics_report = new DomPath("D:supported-report/D:report/*[local-name()='resource-statistics']", 
    DomPath.DEFAULT_NAMESPACE_BINDINGS);

  private static final DomPath DP_statemanagement_report = new DomPath("D:supported-report/D:report/*[local-name()='statemanagement']", 
    DomPath.DEFAULT_NAMESPACE_BINDINGS);

  private static final DomPath DP_displaynames = new DomPath("//*[local-name()='displayname']", 
    DomPath.DEFAULT_NAMESPACE_BINDINGS);

  private static boolean hasCmSynchronizationReport = false;
  private static boolean hasUserResourcesReport = false;
  private static boolean hasLinkConsistencyReport = false;
  private static boolean hasPrincipalConsistencyReport = false;
  private static boolean hasResourceStatisticsReport = false;
  private static boolean hasStatemanagementReport = false;
  private static boolean hasInternalLinks = false;
  private static boolean hasLocks = false;
  private static boolean hasModifiedByProps = false;
  private static boolean hasOwnerProps = false;
  private static boolean hasCmRepInFsdbNosyncMode = false;

  private static IRequester requester2 = null;

  private static FSInteractionTest.FSBasedRepository FS;
  
  public ReportingTest(Integer nr) {
    super(nr, desc[nr.intValue()]);
  }

  protected void runTest() {
    if (nr == 0)
      logIdentifier("REPORTING");

    try {
      String respfile = "REPORTING-" + TestMember.THREEDIGITS.format(nr) + ".txt";
      IResponse response;

      if (requester2 == null) {
        requester2 = getSecondRequester();
      }

      switch (nr) {

        case 0: {
            Element reports = getProperty(getLocalpart(), Consts.N_supported_report_set);
            assertTrue(reports != null);
            ReportingTest.hasLinkConsistencyReport = ReportingTest.DP_link_consistency_report.matches(reports);
            ReportingTest.hasUserResourcesReport = ReportingTest.DP_user_resources_report.matches(reports);
            ReportingTest.hasPrincipalConsistencyReport = ReportingTest.DP_principal_consistency_report.matches(reports);
            ReportingTest.hasResourceStatisticsReport = ReportingTest.DP_resource_statistics_report.matches(reports);
            ReportingTest.hasStatemanagementReport = ReportingTest.DP_statemanagement_report.matches(reports);
            ReportingTest.hasCmSynchronizationReport = ReportingTest.DP_cm_synchronization_report.matches(reports);
          }
          break;
          
        case 1: {
          clear(request);
          request.setMethod("MKCOL");
          request.setURI(getLocalpart() + "/reporting/");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/reporting/plain");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          
          DavStatus res = setProperty(getLocalpart() + "/reporting/", PN_owner, "id-of-unknown-user", false);
          hasOwnerProps = res.getStatus() == 200;        
  
          clear(request);
          request.setMethod("MKRESOURCE");
          request.setURI(getLocalpart() + "/reporting/link");
          request.setBody("<propertyupdate xmlns='DAV:'>" +
                          "<set>" +
                          "<prop>" +
                          "<resourcetype>" + InternalLinksTest.ELEMENT_internal_link + "</resourcetype>" +
                          "<reftarget><href>" + getLocalpart() + "/reporting/dangling" + "</href></reftarget>" +
                          "</prop>" +
                          "</set>" +
                          "</propertyupdate>");
          response = requester.perform(request);
          checkResponseCode(response, 201, 501);
          
          ReportingTest.hasInternalLinks = response.getStatus() == 201;
          logComment("internal-links:" + ReportingTest.hasInternalLinks);
          
          if (ReportingTest.hasInternalLinks) {
            // create link to unknown repository
            clear(request);
            request.setMethod("MKRESOURCE");
            request.setURI(getLocalpart() + "/reporting/link2");
            request.setBody("<propertyupdate xmlns='DAV:'>" +
                            "<set>" +
                            "<prop>" +
                            "<resourcetype>" + InternalLinksTest.ELEMENT_internal_link + "</resourcetype>" +
                            "<reftarget><href>../../../unknown/broken</href></reftarget>" +
                            "</prop>" +
                            "</set>" +
                            "</propertyupdate>");
            response = requester.perform(request);
            checkResponseCode(response, 201, 501);
          }
        }
        break;

        case 2: if (ReportingTest.hasLinkConsistencyReport) {
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setBody("<link-consistency xmlns='http://sapportals.com/xmlns/cm/reporting'><report-schema xmlns='http://sapportals.com/xmlns/cm/webdav'/></link-consistency>");
            response = requester.perform(request);
            checkResponseCode(response, 200);
            checkResponseDocument(response, respfile, null);
          }
          break;

        case 3: if (ReportingTest.hasInternalLinks) {
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setBody("<link-consistency xmlns='http://sapportals.com/xmlns/cm/reporting'/>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            checkResponseDocument(response, respfile, null);
          }
          break;

        case 4: if (ReportingTest.hasInternalLinks) {
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(false);
            request.setBody("<link-consistency xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'><wd:commands><delete/></wd:commands></link-consistency>");
            response = requester.perform(request);
            checkResponseCode(response, 400);
          }
          break;

        case 5: if (ReportingTest.hasInternalLinks) {
            clear(request);
            request.setMethod("POST");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setContentType("text/xml; charset=\"UTF-8\"");
            request.setBody("<link-consistency xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'><wd:commands><delete/></wd:commands></link-consistency>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            checkResponseDocument(response, respfile, null);
          }
          break;

        case 6: if (ReportingTest.hasInternalLinks) {
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setBody("<link-consistency xmlns='http://sapportals.com/xmlns/cm/reporting'></link-consistency>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            checkResponseDocument(response, respfile, null);
          }
          break;

        case 7: if (ReportingTest.hasUserResourcesReport) {
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setBody("<user-resources xmlns='http://sapportals.com/xmlns/cm/reporting'><report-schema xmlns='http://sapportals.com/xmlns/cm/webdav'/></user-resources>");
            response = requester.perform(request);
            checkResponseCode(response, 200);
            checkResponseDocument(response, respfile, null);
          }
          break;

        case 8: {
            ReportingTest.hasModifiedByProps = getProperty(getLocalpart()+"/reporting/", PN_modifiedby) != null;
            if (ReportingTest.hasUserResourcesReport && ReportingTest.hasModifiedByProps) {
              clear(request);
              request.setMethod("REPORT");
              request.setHeader("Depth", "0");
              request.setURI(getLocalpart() + "/reporting/");
              request.expectsResponseDocument(true);
              request.setBody("<user-resources xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'><wd:params><user>"
                + this.m_userid
                + "</user><search-properties>true</search-properties></wd:params></user-resources>");
              response = requester.perform(request);
              checkResponseCode(response, 207);
              checkResponseDocument(response, respfile, null);
            }
          }
          break;

        case 9: if (ReportingTest.hasUserResourcesReport && ReportingTest.hasModifiedByProps) {
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setBody("<user-resources xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'><wd:params><user>"
              + this.m_userid2
              + "</user><search-properties>true</search-properties></wd:params></user-resources>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            checkResponseDocument(response, respfile, null);
          } 
          break;

        case 10: if (ReportingTest.hasUserResourcesReport) {
            clear(request);
            request.setMethod("LOCK");
            request.setURI(getLocalpart() + "/reporting/plain");
            request.setBody(LockTest.makeLockInfo(m_userid));
            request.expectsResponseDocument(true);
            response = requester.perform(request);
            checkResponseCode(response, 200, 501);
            ReportingTest.hasLocks = (response.getStatus() == 200);
            
            if (ReportingTest.hasLocks) {
              String lockToken = TestMember.getLockToken(response);           
              clear(request);
              request.setMethod("REPORT");
              request.setHeader("Depth", "0");
              request.setURI(getLocalpart() + "/reporting/");
              request.expectsResponseDocument(true);
              request.setBody("<user-resources xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'>"
                + "<wd:scope><href xmlns='DAV:'>" + getLocalpart() + "/reporting/" + "</href></wd:scope>"
                + "<wd:params><user>"
                + this.m_userid
                + "</user><search-locks>true</search-locks><search-properties>false</search-properties></wd:params></user-resources>");
              response = requester.perform(request);
 
              // unlock before assertions for robustness
              clear(request);
              request.setMethod("UNLOCK");
              request.setURI(getLocalpart() + "/reporting/plain");
              request.setHeader("Lock-Token", "<" + lockToken + ">");
              IResponse response2 = requester.perform(request);
              checkResponseCode(response2, 204);
  
              checkResponseCode(response, 207);
              checkResponseDocument(response, respfile, DP_displaynames);
            }
          } 
          break;

        case 11: if (ReportingTest.hasUserResourcesReport) {
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setBody("<user-resources xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'><wd:params><user>"
              + this.m_userid
              + "</user><search-privileges>true</search-privileges><search-properties>false</search-properties></wd:params></user-resources>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            checkResponseDocument(response, respfile, null);
          } 
          break;

        case 12: if (ReportingTest.hasUserResourcesReport && AclTest.getSelfURI() != null) {

            AclTest.setAllOnlyMyself("/reporting/plain", this);
            
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setBody("<user-resources xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'><wd:params><user>"
              + this.m_userid
              + "</user><search-privileges>true</search-privileges><search-properties>false</search-properties></wd:params></user-resources>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            checkResponseDocument(response, respfile, null);
            
            // redo with multiple users
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setBody("<user-resources xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'><wd:params><user>"
              + "foobar," + this.m_userid
              + "</user><search-privileges>true</search-privileges><search-properties>false</search-properties></wd:params></user-resources>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            checkResponseDocument(response, respfile, null);
          } 
          break;

        case 13: if (ReportingTest.hasUserResourcesReport && AclTest.getSelfURI() != null) {
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setBody("<user-resources xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'><wd:params><user>"
              + this.m_userid2
              + "</user><search-privileges>true</search-privileges></wd:params></user-resources>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            checkResponseDocument(response, respfile, null);
          } 
          break;

        case 14: if (ReportingTest.hasUserResourcesReport && AclTest.getSelfURI() != null) {
            clear(request);
            request.setMethod("POST");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setContentType("text/xml");
            request.setBody("<user-resources xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'>"
              + "<wd:params>"
              + "<search-privileges>true</search-privileges><search-properties>false</search-properties>"
              + "<user>"+this.m_userid+"</user>"
              + "</wd:params>" 
              + "<wd:commands><transfer><wd:params><replacement>"+this.m_userid2+"</replacement></wd:params></transfer></wd:commands></user-resources>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            checkResponseDocument(response, respfile, null);
          }
          break;

        case 15: if (ReportingTest.hasUserResourcesReport && AclTest.getSelfURI() != null) {
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setBody("<user-resources xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'><wd:params><user>"
              + this.m_userid
              + "</user><search-privileges>true</search-privileges><search-properties>false</search-properties></wd:params></user-resources>");
            response = requester2.perform(request);
            checkResponseCode(response, 207);
            checkResponseDocument(response, respfile, null);
          } 
          break;

        case 16: if (ReportingTest.hasUserResourcesReport && AclTest.getSelfURI() != null) {
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setBody("<user-resources xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'><wd:params><user>"
              + this.m_userid2
              + "</user><search-privileges>true</search-privileges><search-properties>false</search-properties></wd:params></user-resources>");
            response = requester2.perform(request);
            checkResponseCode(response, 207);
            checkResponseDocument(response, respfile, null);
          } 
          break;
        
        case 17: if (ReportingTest.hasUserResourcesReport && AclTest.getSelfURI() != null) {
            clear(request);
            request.setMethod("DELETE");
            request.setURI(getLocalpart() + "/reporting/plain");
            response = this.getSecondRequester().perform(request);
            checkResponseCode(response, 204);
          } 
          break;
        
        case 18: if (ReportingTest.hasLinkConsistencyReport) {
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setBody("<link-consistency xmlns='http://sapportals.com/xmlns/cm/reporting'><max-results xmlns='http://sapportals.com/xmlns/cm/webdav'>sasas</max-results></link-consistency>");
            response = requester.perform(request);
            checkResponseCode(response, 422);
          } 
          break;

        case 19: if (ReportingTest.hasLinkConsistencyReport) {
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setBody("<link-consistency xmlns='http://sapportals.com/xmlns/cm/reporting'><max-results xmlns='http://sapportals.com/xmlns/cm/webdav'>0</max-results></link-consistency>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            List responses = DomElement.getChildren(response.getDocument().getDocumentElement(), Consts.N_response);
            assertTrue(responses.size() == 0);
          } 
          break;

        case 20: if (ReportingTest.hasPrincipalConsistencyReport) {
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setBody("<principal-consistency xmlns='http://sapportals.com/xmlns/cm/reporting'><report-schema xmlns='http://sapportals.com/xmlns/cm/webdav'/></principal-consistency>");
            response = requester.perform(request);
            checkResponseCode(response, 200);
            checkResponseDocument(response, respfile, null);
          }
          break;

        case 21: {
            if (ReportingTest.hasPrincipalConsistencyReport && ReportingTest.hasOwnerProps) {
              clear(request);
              request.setMethod("REPORT");
              request.setHeader("Depth", "0");
              request.setURI(getLocalpart() + "/reporting/");
              request.expectsResponseDocument(true);
              request.setBody("<principal-consistency xmlns='http://sapportals.com/xmlns/cm/reporting'/>");
              response = requester.perform(request);
              checkResponseCode(response, 207);
              List responses = DomElement.getChildren(response.getDocument().getDocumentElement(), Consts.N_response);
              assertTrue(responses.size() == 1); // expect exactly one missing principal
            }
          }
          break;

        case 22: {
            if (ReportingTest.hasUserResourcesReport && ReportingTest.hasOwnerProps) {
              DavStatus res = setProperty(getLocalpart() + "/reporting/", PN_lock_owner,
                Arrays.asList(new String[] {"foo", "bar", "id-of-unknown-user"}), true);
              boolean canSetLockOwnerProp = res.getStatus() == 200 && res.getCondition() == null;     
  
              clear(request);
              request.setMethod("POST");
              request.setHeader("Depth", "0");
              request.setURI(getLocalpart() + "/reporting/");
              request.expectsResponseDocument(true);
              request.setContentType("text/xml");
              request.setBody(
                "<user-resources xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'>"
                  + "<wd:params><user>id-of-unknown-user</user><search-properties>true</search-properties></wd:params>"
                  + "<wd:commands><transfer><wd:params><replacement>"+super.m_userid+"</replacement></wd:params></transfer></wd:commands>"
                  + "</user-resources>");
              response = requester.perform(request);
              checkResponseCode(response, 207);
              List responses = DomElement.getChildren(response.getDocument().getDocumentElement(), Consts.N_response);
              assertTrue(responses.size() == 1); // expect exactly one found resource
              
              // check that the value really changed
              Element owner = getProperty(getLocalpart() + "/reporting/", PN_owner);
              assertTrue("DAV:owner did not change", super.m_userid.equals(DomElement.getText(owner)));
              
              if (canSetLockOwnerProp) {
                Element lockowner = getProperty(getLocalpart() + "/reporting/", PN_lock_owner);
                assertTrue("rpt:lock-owner property lost", lockowner != null);
                String ser = DomSerializer.toString(lockowner, false);
                assertTrue("user id wasn' replaced: " + ser + "\n" +
                    "Make sure that rpt:lock-owner is configured as multivalued property of 'user' type.", ser.indexOf("id-of-unknown-user") < 0);
              }  
            }
          }
          break;

        case 23: if (ReportingTest.hasResourceStatisticsReport) {
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setBody("<resource-statistics xmlns='http://sapportals.com/xmlns/cm/reporting'><report-schema xmlns='http://sapportals.com/xmlns/cm/webdav'/></resource-statistics>");
            response = requester.perform(request);
            checkResponseCode(response, 200);
            checkResponseDocument(response, respfile, null);
          }
          break;

        case 24: if (ReportingTest.hasStatemanagementReport) {
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setBody("<statemanagement xmlns='http://sapportals.com/xmlns/cm/reporting'><report-schema xmlns='http://sapportals.com/xmlns/cm/webdav'/></statemanagement>");
            response = requester.perform(request);
            checkResponseCode(response, 200);
            checkResponseDocument(response, respfile, null);
          }
          break;

        case 25: if (ReportingTest.hasLinkConsistencyReport) {
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setBody("<link-consistency xmlns:D='DAV:' xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'>" 
              + "<wd:apply-to><D:href>/gibtsnich</D:href><D:href>ftp://blafasel</D:href></wd:apply-to>"
              + "</link-consistency>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            checkResponseDocument(response, respfile, null);
          }
          break;

        case 26: if (ReportingTest.hasLinkConsistencyReport && ReportingTest.hasInternalLinks) {
            clear(request);
            request.setMethod("MKRESOURCE");
            request.setURI(getLocalpart() + "/reporting/brokenlink");
            request.expectsResponseDocument(true);
            request.setContentType("text/xml");
            request.setBody("<propertyupdate xmlns='DAV:'>" +
                            "<set>" +
                            "<prop>" +
                            "<resourcetype>" + InternalLinksTest.ELEMENT_internal_link + "</resourcetype>" +
                            "<reftarget><href>" + getLocalpart() + "/reporting/dangling" + "</href></reftarget>" +
                            "</prop>" +
                            "</set>" +
                            "</propertyupdate>");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + "/reporting/plaintxt");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("POST");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.setContentType("text/xml");
            request.expectsResponseDocument(true);
            request.setBody("<link-consistency xmlns:D='DAV:' xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'>" 
              + "<wd:apply-to><D:href>plaintxt</D:href><D:href>brokenlink</D:href></wd:apply-to>"
              + "<wd:commands><delete/></wd:commands>"
              + "</link-consistency>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            checkResponseDocument(response, respfile, null);

            clear(request);
            request.setMethod("HEAD");
            request.setURI(getLocalpart() + "/reporting/brokenlink");
            response = requester.perform(request);
            checkResponseCode(response, 404);
          }
          break;

        case 27: {
            if (ReportingTest.hasUserResourcesReport && ReportingTest.hasOwnerProps) {
              DavStatus res = setProperty(getLocalpart() + "/reporting/", PN_lock_owner,
                Arrays.asList(new String[] {"foo", "bar", "id-of-unknown-user"}), true);
              boolean canSetLockOwnerProp = res.getStatus() == 200 && res.getCondition() == null;     
  
              clear(request);
              request.setMethod("POST");
              request.setHeader("Depth", "0");
              request.setURI(getLocalpart() + "/reporting/");
              request.expectsResponseDocument(true);
              request.setContentType("text/xml");
              request.setBody(
                "<user-resources xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav' xmlns:D='DAV:'>"
                  + "<wd:apply-to><D:href>" + getLocalpart() + "/reporting/" + "</D:href></wd:apply-to>"
                  + "<wd:params><user>id-of-unknown-user</user><search-properties>true</search-properties></wd:params>"
                  + "<wd:commands><transfer><wd:params><replacement>"+super.m_userid+"</replacement></wd:params></transfer></wd:commands>"
                  + "</user-resources>");
              response = requester.perform(request);
              checkResponseCode(response, 207);
              List responses = DomElement.getChildren(response.getDocument().getDocumentElement(), Consts.N_response);
              assertTrue(responses.size() == 1); // expect exactly one found resource
              
              // check that the value really changed
              Element owner = getProperty(getLocalpart() + "/reporting/", PN_owner);
              assertTrue("DAV:owner did not change", super.m_userid.equals(DomElement.getText(owner)));
              
              if (canSetLockOwnerProp) {
                Element lockowner = getProperty(getLocalpart() + "/reporting/", PN_lock_owner);
                assertTrue("rpt:lock-owner property lost", lockowner != null);
                String ser = DomSerializer.toString(lockowner, false);
                assertTrue("user id wasn't replaced: " + ser + "\n" +
                    "Make sure that rpt:lock-owner is configured as multivalued property of 'user' type.", ser.indexOf("id-of-unknown-user") < 0);
              }  
            }
          }
          break;
        
        case 28: if (ReportingTest.hasUserResourcesReport && ReportingTest.hasLocks && ReportingTest.hasOwnerProps) {
          
            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + "/reporting/plain-locked");
            response = requester.perform(request);
            checkResponseCode(response, 201);
          
            DavStatus res = setProperty(getLocalpart() + "/reporting/plain-locked", PN_owner, "replace-me-locked", false);
            assertTrue(res.getStatus() == 200);

            clear(request);
            request.setMethod("LOCK");
            request.setURI(getLocalpart() + "/reporting/plain-locked");
            request.setBody(LockTest.makeLockInfo(m_userid));
            request.expectsResponseDocument(true);
            response = requester.perform(request);
            checkResponseCode(response, 200);
            String lockToken = TestMember.getLockToken(response);           

            clear(request);
            request.setMethod("POST");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setContentType("application/xml");
            request.setBody("<user-resources xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'>"
              + "<wd:params>"
              + "<search-privileges>false</search-privileges><search-properties>true</search-properties>"
              + "<user>replace-me-locked</user>"
              + "</wd:params>" 
              + "<wd:commands><transfer><wd:params><replacement>"+this.m_userid2+"</replacement></wd:params></transfer></wd:commands></user-resources>");
            response = requester.perform(request);

            // unlock before assertions for robustness
            clear(request);
            request.setMethod("UNLOCK");
            request.setURI(getLocalpart() + "/reporting/plain-locked");
            request.setHeader("Lock-Token", "<" + lockToken + ">");
            IResponse response2 = requester.perform(request);
            checkResponseCode(response2, 204);

            checkResponseCode(response, 207);
            checkResponseDocument(response, respfile, null);
          }
          break;

        case 29: if (ReportingTest.hasUserResourcesReport && ReportingTest.hasOwnerProps) {
          
            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + "/reporting/plain-vcr");
            response = requester.perform(request);
            checkResponseCode(response, 201);
          
            DavStatus res = setProperty(getLocalpart() + "/reporting/plain-vcr", PN_owner, "replace-me-vcr", false);
            assertTrue(res.getStatus() == 200);
            
            clear(request);
            request.setMethod("VERSION-CONTROL");
            request.setURI(getLocalpart() + "/reporting/plain-vcr");
            response = requester.perform(request);
            
            if (response.getStatus() == 200 || response.getStatus() == 201) {
              clear(request);
              request.setMethod("POST");
              request.setHeader("Depth", "0");
              request.setURI(getLocalpart() + "/reporting/");
              request.expectsResponseDocument(true);
              request.setContentType("application/xml");
              request.setBody("<user-resources xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'>"
                + "<wd:params>"
                + "<search-privileges>false</search-privileges><search-properties>true</search-properties>"
                + "<user>replace-me-vcr</user>"
                + "</wd:params>" 
                + "<wd:commands><transfer><wd:params><replacement>"+this.m_userid2+"</replacement></wd:params></transfer></wd:commands></user-resources>");
              response = requester.perform(request);
              checkResponseCode(response, 207);
              checkResponseDocument(response, respfile, null);

              Element checkedIn = getProperty(getLocalpart() + "/reporting/plain-vcr", Consts.N_checked_in);
              assertTrue("resource should be still checked-in after report execution", checkedIn != null);
            }
            else {
              // nothing to do if not versionable
            }
          }
          break;

        case 30: if (ReportingTest.hasCmSynchronizationReport) {
        
            FSInteractionTest test = new FSInteractionTest(new Integer(0));
            new FSTestSuite().init(test);
            FS = test.discoverFileSystemRepository();
            hasCmRepInFsdbNosyncMode = (FS != null && FS.keepsCollections && FS.nosync);
            
            if (hasCmRepInFsdbNosyncMode) {
              // run report
              clear(request);
              request.setMethod("REPORT");
              request.setHeader("Depth", "0");
              request.setURI(getLocalpart() + "/reporting/");
              request.expectsResponseDocument(true);
              request.setBody("<cm.synchronize xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'>"
                  + "<wd:scope><href xmlns='DAV:'>" + getLocalpart() + "</href></wd:scope>"
                  + "<wd:params>" 
                  + "<collect.differing>true</collect.differing>" 
                  + "<collect.new>true</collect.new>" 
                  + "<collect.lost>true</collect.lost>" 
                  + "</wd:params>" 
                  + "</cm.synchronize>");
              response = requester.perform(request);
              checkResponseCode(response, 207);
              List hrefs = getHrefsFromMultistatus(response.getDocument());

            if (hrefs.size() > 0) {
              // run report: synchronize prepared new resources
              clear(request);
              request.setMethod("POST");
              request.setHeader("Depth", "0");
              request.setURI(getLocalpart() + "/reporting/");
              request.setContentType("text/xml");
              request.expectsResponseDocument(true);
              String body = "<cm.synchronize xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'>"
                  + "<wd:apply-to>";
              for (Iterator it = hrefs.iterator(); it.hasNext(); ) {
                body += "<href xmlns='DAV:'>" + it.next() + "</href>";
              }
              body += "</wd:apply-to>"
                  + "<wd:commands><synchronize/></wd:commands>"
                  + "</cm.synchronize>";
              request.setBody(body);
              response = requester.perform(request);
              checkResponseCode(response, 207);
              hrefs = getHrefsFromMultistatus(response.getDocument());
            }
              assertEquals("FSDB nosync repository contains resources that are not in sync: " + hrefs, 0, hrefs.size());
            }
          }
          break;
          
        case 31: if (ReportingTest.hasCmSynchronizationReport && ReportingTest.hasCmRepInFsdbNosyncMode) {
            
            clear(request);
            request.setMethod("MKCOL");
            request.setURI(getLocalpart() + "/reporting/coll");
            response = requester.perform(request);
            checkResponseCode(response, 201);
            
            clear(request);
            request.setMethod("MKCOL");
            request.setURI(getLocalpart() + "/reporting/coll2");
            response = requester.perform(request);
            checkResponseCode(response, 201);
            
            try {
              Thread.sleep(1000);
            }
            catch (InterruptedException ex) {
              ex.getMessage();
            }
            
            // create new file/folder
            File tmp = new File(FS.basecoll.getPath() + "\\reporting\\coll\\reporting_31");
            assertTrue("Could not create test collection: " + tmp.getPath(), tmp.mkdir());
            tmp = new File(FS.basecoll.getPath() + "\\reporting\\coll2\\reporting_31");
            assertTrue("Could not create test file: " + tmp.getPath(), tmp.createNewFile());
            
            // run report: collect only lost resources
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setBody("<cm.synchronize xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'>"
                + "<wd:scope><href xmlns='DAV:'>" + getLocalpart() + "</href></wd:scope>"
                + "<wd:params>" 
                + "<collect.differing>false</collect.differing>" 
                + "<collect.new>false</collect.new>" 
                + "<collect.lost>true</collect.lost>" 
                + "</wd:params>" 
                + "</cm.synchronize>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            List hrefs = getHrefsFromMultistatus(response.getDocument());
            assertEquals("Unexpected resources found out of sync: " + hrefs, 0, hrefs.size());
            
            // run report: collect only new resources
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setBody("<cm.synchronize xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'>"
                + "<wd:scope><href xmlns='DAV:'>" + getLocalpart() + "</href></wd:scope>"
                + "<wd:params>" 
                + "<collect.differing>false</collect.differing>" 
                + "<collect.new>true</collect.new>" 
                + "<collect.lost>false</collect.lost>" 
                + "</wd:params>" 
                + "</cm.synchronize>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            hrefs = getHrefsFromMultistatus(response.getDocument());
            
            assertEquals("Unexpected resources found out of sync: " + hrefs, 2, hrefs.size());
            String expectedUri = getLocalpart() + "/reporting/coll/reporting_31";
            assertTrue("Prepared new file not reported as out of sync:" + expectedUri, hrefs.contains(expectedUri));
            expectedUri = getLocalpart() + "/reporting/coll2/reporting_31";
            assertTrue("Prepared new file not reported as out of sync:" + expectedUri, hrefs.contains(expectedUri));

            // run report: synchronize prepared new resources
            clear(request);
            request.setMethod("POST");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.setContentType("text/xml");
            request.expectsResponseDocument(true);
            String body = "<cm.synchronize xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'>"
                + "<wd:apply-to>";
            for (Iterator it = hrefs.iterator(); it.hasNext(); ) {
              body += "<href xmlns='DAV:'>" + it.next() + "</href>";
            }
            body += "</wd:apply-to>"
                + "<wd:commands><synchronize/></wd:commands>"
                + "</cm.synchronize>";
            request.setBody(body);
            response = requester.perform(request);
            checkResponseCode(response, 207);
            hrefs = getHrefsFromMultistatus(response.getDocument());
            
            assertEquals("Unexpected resources found out of sync: " + hrefs, 2, hrefs.size());
            expectedUri = getLocalpart() + "/reporting/coll/reporting_31";
            assertTrue("Prepared new file not reported as out of sync:" + expectedUri, hrefs.contains(expectedUri));
            expectedUri = getLocalpart() + "/reporting/coll2/reporting_31";
            assertTrue("Prepared new file not reported as out of sync:" + expectedUri, hrefs.contains(expectedUri));
            
            // run report: collect only differing resources
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setBody("<cm.synchronize xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'>"
                + "<wd:scope><href xmlns='DAV:'>" + getLocalpart() + "</href></wd:scope>"
                + "<wd:params>" 
                + "<collect.differing>true</collect.differing>" 
                + "<collect.new>false</collect.new>" 
                + "<collect.lost>false</collect.lost>" 
                + "</wd:params>" 
                + "</cm.synchronize>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            hrefs = getHrefsFromMultistatus(response.getDocument());
            
            assertEquals("Unexpected resources found out of sync: " + hrefs, 2, hrefs.size());
            expectedUri = getLocalpart() + "/reporting/coll";
            assertTrue("Prepared new file not reported as out of sync:" + expectedUri, hrefs.contains(expectedUri));
            expectedUri = getLocalpart() + "/reporting/coll2";
            assertTrue("Prepared new file not reported as out of sync:" + expectedUri, hrefs.contains(expectedUri));

            // run report: synchronize prepared differing resource
            clear(request);
            request.setMethod("POST");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.setContentType("text/xml");
            request.expectsResponseDocument(true);
            body = "<cm.synchronize xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'>"
                + "<wd:apply-to>";
            for (Iterator it = hrefs.iterator(); it.hasNext(); ) {
              body += "<href xmlns='DAV:'>" + it.next() + "</href>";
            }
            body += "</wd:apply-to>"
                + "<wd:commands><synchronize/></wd:commands>"
                + "</cm.synchronize>";
            request.setBody(body);
            response = requester.perform(request);
            checkResponseCode(response, 207);
            hrefs = getHrefsFromMultistatus(response.getDocument());
            
            assertEquals("Unexpected resources found out of sync: " + hrefs, 2, hrefs.size());
            expectedUri = getLocalpart() + "/reporting/coll";
            assertTrue("Prepared new file not reported as out of sync:" + expectedUri, hrefs.contains(expectedUri));
            expectedUri = getLocalpart() + "/reporting/coll2";
            assertTrue("Prepared new file not reported as out of sync:" + expectedUri, hrefs.contains(expectedUri));
          }
          break;
          
        case 32: if (ReportingTest.hasCmSynchronizationReport && ReportingTest.hasCmRepInFsdbNosyncMode) {
          
            // delete file/folder
            File tmp = new File(FS.basecoll.getPath() + "\\reporting\\coll\\reporting_31");
            assertTrue("Could not delete test collection: " + tmp.getPath(), tmp.delete());
            tmp = new File(FS.basecoll.getPath() + "\\reporting\\coll2\\reporting_31");
            assertTrue("Could not delete test file: " + tmp.getPath(), tmp.delete());
            
            // run report: collect only new resources
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setBody("<cm.synchronize xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'>"
                + "<wd:scope><href xmlns='DAV:'>" + getLocalpart() + "</href></wd:scope>"
                + "<wd:params>" 
                + "<collect.differing>false</collect.differing>" 
                + "<collect.new>true</collect.new>" 
                + "<collect.lost>false</collect.lost>" 
                + "</wd:params>" 
                + "</cm.synchronize>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            List hrefs = getHrefsFromMultistatus(response.getDocument());
            assertEquals("Unexpected resources found out of sync: " + hrefs, 0, hrefs.size());
            
            // run report: collect only lost resources
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setBody("<cm.synchronize xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'>"
                + "<wd:scope><href xmlns='DAV:'>" + getLocalpart() + "</href></wd:scope>"
                + "<wd:params>" 
                + "<collect.differing>false</collect.differing>" 
                + "<collect.new>false</collect.new>" 
                + "<collect.lost>true</collect.lost>" 
                + "</wd:params>" 
                + "</cm.synchronize>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            hrefs = getHrefsFromMultistatus(response.getDocument());
            
            assertEquals("Unexpected resources found out of sync: " + hrefs, 2, hrefs.size());
            String expectedUri = getLocalpart() + "/reporting/coll/reporting_31";
            assertTrue("Prepared new file not reported as out of sync:" + expectedUri, hrefs.contains(expectedUri));
            expectedUri = getLocalpart() + "/reporting/coll2/reporting_31";
            assertTrue("Prepared new file not reported as out of sync:" + expectedUri, hrefs.contains(expectedUri));
  
            // run report: synchronize prepared lost resources
            clear(request);
            request.setMethod("POST");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.setContentType("text/xml");
            request.expectsResponseDocument(true);
            String body = "<cm.synchronize xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'>"
                + "<wd:apply-to>";
            for (Iterator it = hrefs.iterator(); it.hasNext(); ) {
              body += "<href xmlns='DAV:'>" + it.next() + "</href>";
            }
            body += "</wd:apply-to>"
                + "<wd:commands><synchronize/></wd:commands>"
                + "</cm.synchronize>";
            request.setBody(body);
            response = requester.perform(request);
            checkResponseCode(response, 207);
            hrefs = getHrefsFromMultistatus(response.getDocument());
            
            assertEquals("Unexpected resources found out of sync: " + hrefs, 2, hrefs.size());
            expectedUri = getLocalpart() + "/reporting/coll/reporting_31";
            assertTrue("Prepared new file not reported as out of sync:" + expectedUri, hrefs.contains(expectedUri));
            expectedUri = getLocalpart() + "/reporting/coll2/reporting_31";
            assertTrue("Prepared new file not reported as out of sync:" + expectedUri, hrefs.contains(expectedUri));
            
            // run report: collect only differing resources
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setBody("<cm.synchronize xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'>"
                + "<wd:scope><href xmlns='DAV:'>" + getLocalpart() + "</href></wd:scope>"
                + "<wd:params>" 
                + "<collect.differing>true</collect.differing>" 
                + "<collect.new>false</collect.new>" 
                + "<collect.lost>false</collect.lost>" 
                + "</wd:params>" 
                + "</cm.synchronize>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            hrefs = getHrefsFromMultistatus(response.getDocument());
            
            assertEquals("Unexpected resources found out of sync: " + hrefs, 2, hrefs.size());
            expectedUri = getLocalpart() + "/reporting/coll";
            assertTrue("Prepared new file not reported as out of sync:" + expectedUri, hrefs.contains(expectedUri));
            expectedUri = getLocalpart() + "/reporting/coll2";
            assertTrue("Prepared new file not reported as out of sync:" + expectedUri, hrefs.contains(expectedUri));
  
            // run report: synchronize prepared differing resource
            clear(request);
            request.setMethod("POST");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.setContentType("text/xml");
            request.expectsResponseDocument(true);
            body = "<cm.synchronize xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'>"
                + "<wd:apply-to>";
            for (Iterator it = hrefs.iterator(); it.hasNext(); ) {
              body += "<href xmlns='DAV:'>" + it.next() + "</href>";
            }
            body += "</wd:apply-to>"
                + "<wd:commands><synchronize/></wd:commands>"
                + "</cm.synchronize>";
            request.setBody(body);
            response = requester.perform(request);
            checkResponseCode(response, 207);
            hrefs = getHrefsFromMultistatus(response.getDocument());
            
            assertEquals("Unexpected resources found out of sync: " + hrefs, 2, hrefs.size());
            expectedUri = getLocalpart() + "/reporting/coll";
            assertTrue("Prepared new file not reported as out of sync:" + expectedUri, hrefs.contains(expectedUri));
            expectedUri = getLocalpart() + "/reporting/coll2";
            assertTrue("Prepared new file not reported as out of sync:" + expectedUri, hrefs.contains(expectedUri));
          }
          break;
          
        case 33: if (ReportingTest.hasCmSynchronizationReport && ReportingTest.hasCmRepInFsdbNosyncMode) {
            
            //
            // replace test collection by test file
            //
            File tmp = new File(FS.basecoll.getPath() + "\\reporting\\coll");
            assertTrue("Could not delete test collection: " + tmp.getPath(), tmp.delete());
            assertTrue("Could not create test file: " + tmp.getPath(), tmp.createNewFile());

            // run report: collect only differing resources
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setBody("<cm.synchronize xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'>"
                + "<wd:scope><href xmlns='DAV:'>" + getLocalpart() + "</href></wd:scope>"
                + "<wd:params>" 
                + "<collect.differing>true</collect.differing>" 
                + "<collect.new>false</collect.new>" 
                + "<collect.lost>false</collect.lost>" 
                + "</wd:params>" 
                + "</cm.synchronize>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            List hrefs = getHrefsFromMultistatus(response.getDocument());

            assertEquals("Unexpected resources found out of sync: " + hrefs, 2, hrefs.size());
            String expectedUri = getLocalpart() + "/reporting";
            assertTrue("Prepared new file not reported as out of sync:" + expectedUri, hrefs.contains(expectedUri));
            expectedUri = getLocalpart() + "/reporting/coll";
            assertTrue("Prepared new file not reported as out of sync:" + expectedUri, hrefs.contains(expectedUri));
            
            // run report: synchronize prepared differing resource
            clear(request);
            request.setMethod("POST");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.setContentType("text/xml");
            request.expectsResponseDocument(true);
            String body = "<cm.synchronize xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'>"
                + "<wd:apply-to>";
            for (Iterator it = hrefs.iterator(); it.hasNext(); ) {
              body += "<href xmlns='DAV:'>" + it.next() + "</href>";
            }
            body += "</wd:apply-to>"
                + "<wd:commands><synchronize/></wd:commands>"
                + "</cm.synchronize>";
            request.setBody(body);
            response = requester.perform(request);
            checkResponseCode(response, 207);
            hrefs = getHrefsFromMultistatus(response.getDocument());

            assertEquals("Unexpected resources found out of sync: " + hrefs, 2, hrefs.size());
            expectedUri = getLocalpart() + "/reporting";
            assertTrue("Prepared new file not reported as out of sync:" + expectedUri, hrefs.contains(expectedUri));
            expectedUri = getLocalpart() + "/reporting/coll";
            assertTrue("Prepared new file not reported as out of sync:" + expectedUri, hrefs.contains(expectedUri));
            
            // run report: collect only differing resources
            // assert: everything's clean
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setBody("<cm.synchronize xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'>"
                + "<wd:scope><href xmlns='DAV:'>" + getLocalpart() + "</href></wd:scope>"
                + "<wd:params>" 
                + "<collect.differing>true</collect.differing>" 
                + "<collect.new>false</collect.new>" 
                + "<collect.lost>false</collect.lost>" 
                + "</wd:params>" 
                + "</cm.synchronize>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            hrefs = getHrefsFromMultistatus(response.getDocument());
            assertEquals("Unexpected resources found out of sync: " + hrefs, 0, hrefs.size());
            
            
            //
            // replace test file by test collection
            //
            assertTrue("Could not delete test file: " + tmp.getPath(), tmp.delete());
            assertTrue("Could not create test collection: " + tmp.getPath(), tmp.createNewFile());
            
            // run report: collect only differing resources
            clear(request);
            request.setMethod("REPORT");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.expectsResponseDocument(true);
            request.setBody("<cm.synchronize xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'>"
                + "<wd:scope><href xmlns='DAV:'>" + getLocalpart() + "</href></wd:scope>"
                + "<wd:params>" 
                + "<collect.differing>true</collect.differing>" 
                + "<collect.new>false</collect.new>" 
                + "<collect.lost>false</collect.lost>" 
                + "</wd:params>" 
                + "</cm.synchronize>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            hrefs = getHrefsFromMultistatus(response.getDocument());

            assertEquals("Unexpected resources found out of sync: " + hrefs, 2, hrefs.size());
            expectedUri = getLocalpart() + "/reporting";
            assertTrue("Prepared new file not reported as out of sync:" + expectedUri, hrefs.contains(expectedUri));
            expectedUri = getLocalpart() + "/reporting/coll";
            assertTrue("Prepared new file not reported as out of sync:" + expectedUri, hrefs.contains(expectedUri));
            
            // run report: synchronize prepared differing resource
            clear(request);
            request.setMethod("POST");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/reporting/");
            request.setContentType("text/xml");
            request.expectsResponseDocument(true);
            body = "<cm.synchronize xmlns='http://sapportals.com/xmlns/cm/reporting' xmlns:wd='http://sapportals.com/xmlns/cm/webdav'>"
                + "<wd:apply-to>";
            for (Iterator it = hrefs.iterator(); it.hasNext(); ) {
              body += "<href xmlns='DAV:'>" + it.next() + "</href>";
            }
            body += "</wd:apply-to>"
                + "<wd:commands><synchronize/></wd:commands>"
                + "</cm.synchronize>";
            request.setBody(body);
            response = requester.perform(request);
            checkResponseCode(response, 207);
            hrefs = getHrefsFromMultistatus(response.getDocument());
          }
          break;
          
        default: {
            clear(request);
            request.setMethod("DELETE");
            request.setURI(getLocalpart() + "/reporting/");
            response = requester.perform(request);
            checkResponseCode(response, 204, 404);
          }
          break;
          
      }
    }
    catch (Exception ex) {
      ex.printStackTrace(System.err);
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }
}
