package com.sapportals.wcm.protocol.webdav.server.tests;

import org.w3c.dom.Node;

import com.sapportals.wcm.protocol.webdav.DomElement;
import com.sapportals.wcm.protocol.webdav.dompath.DomPath;
import com.sapportals.wcm.util.http.IResponse;
import com.sapportals.wcm.util.uri.URICodec;
import com.sapportals.wcm.util.xml.SimpleSerializer;

/**
 * Various tests for WebDAV servlet special action requests.
 *
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author julian.reschke@greenbytes.de
 * @version $Id: ServletActionsTest.java,v 1.8 2004/07/13 11:03:48 jre Exp $
 */

public class ServletActionsTest extends TestMember
{
  private static final String[] desc = new String[]
  {
    "00 SA invalid action",
    "01 SA dowload non-existing resource",
    "02 SA dowload existing resource",
    "03 SA lock&dowload existing resource",
    "04 SA PROPFIND the lock token",
    "05 SA UNLOCK",
    "06 SA download existing resource with interesting name",
    "07 SA HEAD on download URI",
    "08 SA content-language detection",
    "09 SA action=PROPFIND",
  };

  public static int NrOfTests = desc.length;

  public ServletActionsTest(Integer nr)
  {
    super(nr, desc[nr.intValue()]);
  }

  private static final DomPath DP_locktoken_href = new DomPath(
    "/D:multistatus/D:response/D:propstat/D:prop/D:lockdiscovery/D:activelock/D:locktoken/D:href", DomPath.DEFAULT_NAMESPACE_BINDINGS);

  private static String locktoken = null;
  private static boolean waslocked = false;

  protected void runTest()
  {
    if (nr == 0) super.logIdentifier("SA");

    try {
      IResponse response;

      switch (nr) {
        
        case 0:
          clear(request);
          request.setMethod("GET");
          request.setURI(super.servletPath + "?action=foobar");
          response = requester.perform(request);
          checkResponseCode(response, 400);
          break;

        case 1:
          clear(request);
          request.setMethod("GET");
          request.setURI(super.servletPath + "?action=download&URI=xyz");
          response = requester.perform(request);
          checkResponseCode(response, 404);
          break;

        case 2:
          clear(request);
          request.setMethod("GET");
          request.setURI(super.servletPath + "?action=download&URI=" + getLocalpart() + "/ANSI.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 200);
          assertTrue("expected: 'attachment; filename=ANSI.TXT', got:'"+response.getHeader("content-disposition")+"'",
              "attachment; filename=ANSI.TXT".equals(response.getHeader("content-disposition")));
          break;

        case 3: {
            // try both POST (new) and GET (deprecated)
            clear(request);
            request.setMethod("POST");
            request.setContentType("application/x-www-form-urlencoded");
            request.setURI(getLocalpart() + "/ANSI.TXT");
            request.setBody("action=lockeddownload");
            response = requester.perform(request);
            checkResponseCode(response, 303);
            assertTrue((super.servletPath + "?action=download&URI=" + getLocalpart() + "/ANSI.TXT").equals(response.getHeader("location")));          
            waslocked = true;
          }
          break;

        case 4:
          if (waslocked) {
            clear(request);
            request.setURI(getLocalpart() + "/ANSI.TXT");
            request.setMethod("PROPFIND");
            request.setHeader("Depth", "0");
            request.setBody("<propfind xmlns='DAV:'><prop><lockdiscovery/></prop></propfind>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            Node n = DP_locktoken_href.matchFirst(response.getDocument());
            assertTrue("locktoken not present in PROPFIND response: " + SimpleSerializer.toString(response.getDocument(), true), n != null);
            locktoken = DomElement.getText(n);            
          }
          break;

        case 5:
          if (waslocked) {
            clear(request);
            request.setURI(getLocalpart() + "/ANSI.TXT");
            request.setMethod("UNLOCK");
            request.setHeader("lock-token", "<"+locktoken+">");
            response = requester.perform( request );
            checkResponseCode(response, 204);
          }
          break;

        case 6:
          clear(request);
          request.setMethod("GET");
          request.setURI(super.servletPath + "?action=download&URI=" + getLocalpart() + URICodec.Encode("/Euro\u20ac.txt"));
          request.setHeader("User-Agent", "Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0)");
          response = requester.perform(request);
          checkResponseCode(response, 200);
          String cd = response.getHeader("content-disposition");
          assertTrue("unexpected content-disposition header value: '" + cd + "'", "attachment; filename=Euro%e2%82%ac.txt".equals(cd));
          break;

        case 7:
          //HEAD vs GET and non-IE user agent header
          clear(request);
          request.setMethod("HEAD");
          request.setURI(super.servletPath + "?action=download&URI=" + getLocalpart() + URICodec.Encode("/Euro\u20ac.txt"));
          response = requester.perform(request);
          checkResponseCode(response, 200);
          cd = response.getHeader("content-disposition");
          assertTrue("unexpected content-disposition header value: '" + cd + "'", "attachment; filename*=utf-8''Euro%e2%82%ac.txt".equals(response.getHeader("content-disposition")));
          break;

        case 8:
          clear(request);
          request.setMethod("GET");
          request.setURI(getLocalpart());
          request.setHeader("Accept-Language", "de");
          response = requester.perform(request);
          checkResponseCode(response, 200);
          assertTrue("response content language should be 'de', but was: '" + response.getHeader("Content-Language")+ "'", "de".equalsIgnoreCase(response.getHeader("Content-Language")));
          break;

        case 9:
          clear(request);
          request.setMethod("GET");
          request.setURI(super.servletPath + "?action=PROPFIND&URI=" + getLocalpart() + "/ANSI.TXT");
          response = requester.perform(request);
          checkResponseCode(response, 200);
          break;
      }
    }
    catch (Exception ex) {
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }
}
