package com.sapportals.wcm.protocol.webdav.server.tests;

import com.sapportals.wcm.util.http.HttpRequest;
import com.sapportals.wcm.util.http.IRequester;
import com.sapportals.wcm.util.http.IResponse;

/**
 * Various tests for WEBDAV <a href="http://www.greenbytes.de/tech/webdav/rfc2518.html#METHOD_LOCK"
 * >LOCK</a>  method (special case for shared locks).
 *
 * <p>
 * Copyright (c) SAP AG 2001-2003
 * @author julian.reschke@greenbytes.de
 * @version $Id: SharedLockTest.java,v 1.3 2004/03/29 11:52:07 jre Exp $
 */

public class SharedLockTest extends TestMember
{

  private static java.text.DecimalFormat m_threeDigits = new
    java.text.DecimalFormat ("000");

  private static final String[] desc = new String[]
  {
    "00 SL Set up directory and test resource",
    "01 SL take out a shared lock",
    "02 SL take out another shared lock",
    "03 SL make sure no exclusive lock allowed",
    "04 SL let user2 take out a shared lock",
    "05 SL try modification without lock token",
    "06 SL make sure that file can be modified with any of the lock tokens",
    "   SL cleanup",
  };

  public static int NrOfTests = desc.length;

  private static IRequester requester2 = null;
  private static HttpRequest request2 = new HttpRequest();

  public SharedLockTest(Integer nr)
  {
    super(nr, desc[nr.intValue()]);
  }

  private String makeLockInfo (String owner, boolean shared) {
    String res = "<lockinfo xmlns='DAV:'>";

    if (owner != null) {
      res += "<owner>" + owner + "</owner>";
    }

    res += "<lockscope>" + (shared ? "<shared/>" : "<exclusive/>") + "</lockscope>";
    res += "<locktype><write /></locktype></lockinfo>";
    return res;
  }

  static String token1 = null;
  static String token2 = null;
  static String token3 = null;

  protected void runTest()
  {
    if (nr == 0) super.logIdentifier("SL");

    String respfile = "SHAREDLOCK-" + m_threeDigits.format (nr) + ".txt";
    String container = "/shared-locks";
    String member = "/shared-locked-resource";
    String target = container + member;

    clear(request);
    try {
      if (requester2 == null) {
        requester2 = getSecondRequester();
      }

      IResponse response;
      clear(request);

      switch (nr) {
        case 0:
          clear(request);
          request.setMethod("MKCOL");
          request.setURI(getLocalpart() + container);
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + target);
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;

        case 1:
          clear(request);
          request.setMethod("LOCK");
          request.setHeader("Timeout", "Second-60");
          request.setHeader("Depth", "0");
          request.setURI(getLocalpart() + target);
          request.setHeader("Content-Type", "text/xml");
          request.setBody(makeLockInfo (null, true));
          request.expectsResponseDocument(true);
          response = requester.perform(request);
          if (response.getStatus() == 200) {
            token1 = getLockToken(response);
            assertTrue ("locktoken null", token1 != null);

            super.logComment("shared-lock:yes");

            clear(request);
            request.setMethod("PROPFIND");
            request.setHeader("Depth", "1");
            request.setURI(getLocalpart() + target);
            request.setHeader("Content-Type", "text/xml");
            super.checkStatusAndBody(207, "<propfind xmlns='DAV:'><prop><lockdiscovery/></prop></propfind>", respfile);
          }
          else {
            super.logComment("shared-lock:no");
          }
          break;

        case 2:
          clear(request);
          request.setMethod("LOCK");
          request.setHeader("Timeout", "Second-60");
          request.setHeader("Depth", "0");
          request.setURI(getLocalpart() + target);
          request.setHeader("Content-Type", "text/xml");
          request.setBody(makeLockInfo (null, true));
          request.expectsResponseDocument(true);
          response = requester.perform(request);
          if (token1 != null) {
            assertTrue("couldn't get second shared lock", response.getStatus() == 200);
          }
          if (response.getStatus() == 200) {
            token2 = getLockToken(response);
            assertTrue("locktoken null", token2 != null);
          }
          break;

        case 3:
          if (token1 != null) {
            clear(request);
            request.setMethod("LOCK");
            request.setHeader("Depth", "0");
            request.setHeader("Timeout", "Second-60");
            request.setURI(getLocalpart() + target);
            request.setHeader("Content-Type", "text/xml");
            request.setBody(makeLockInfo (null, false));
            request.expectsResponseDocument(true);
            response = requester.perform(request);
            checkResponseCode(response, 423);
          }
          break;

        case 4:
          clear(request2);
          request2.setMethod("LOCK");
          request2.setHeader("Depth", "0");
          request2.setHeader("Timeout", "Second-60");
          request2.setURI(getLocalpart() + target);
          request2.setHeader("Content-Type", "text/xml");
          request2.setBody(makeLockInfo (null, true));
          request2.expectsResponseDocument(true);
          response = requester2.perform(request2);
          if (token1 != null) {
            assertTrue("couldn't get third shared lock", response.getStatus() == 200);
          }
          if (response.getStatus() == 200) {
            token3 = getLockToken(response);
            assertTrue ("locktoken null", token3 != null);
          }
          break;

        case 5:
          if (token1 != null) {
            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + target);
            request.setBody("first mod (should fail)");
            response = requester.perform(request);
            checkResponseCode(response, 412, 423);
          }
          break;

        case 6:
          if (token1 != null) {
            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + target);
            request.setHeader("If", "(<" + token1 + ">)");
            request.setBody("second mod (should succeed)");
            response = requester.perform(request);
            checkResponseCode(response, 200);
          }

          if (token2 != null) {
            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + target);
            request.setHeader("If", "(<" + token2 + ">)");
            request.setBody("third mod (should succeed)");
            response = requester.perform(request);
            checkResponseCode(response, 200);
          }

          if (token3 != null) {
            clear(request2);
            request2.setMethod("PUT");
            request2.setURI(getLocalpart() + target);
            request2.setHeader("If", "(<" + token3 + ">)");
            request2.setBody("4th mod (should succeed)");
            response = requester2.perform(request2);
            checkResponseCode(response, 200);
          }
          break;

        default:
          if (token3 != null) {
            clear(request2);
            request2.setMethod("UNLOCK");
            request2.setURI(getLocalpart() + target);
            request2.setHeader("Lock-Token", "<" + token3 + ">");
            response = requester2.perform(request2);
            checkResponseCode(response, 204);
          }

          if (token2 != null) {
            clear(request);
            request.setMethod("UNLOCK");
            request.setURI(getLocalpart() + target);
            request.setHeader("Lock-Token", "<" + token2 + ">");
            response = requester.perform(request);
            checkResponseCode(response, 204);
          }


          if (token1 != null) {
            clear(request);
            request.setMethod("UNLOCK");
            request.setURI(getLocalpart() + target);
            request.setHeader("Lock-Token", "<" + token1 + ">");
            response = requester.perform(request);
            checkResponseCode(response, 204);
          }

          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + container);
          response = requester.perform(request);
          checkResponseCode(response, 204);
          break;
      }
    }
    catch (Exception ex) {
      ex.printStackTrace(System.err);
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }
}
