package com.sapportals.wcm.protocol.webdav.server.tests;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.util.*;

import junit.framework.TestCase;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import com.sapportals.wcm.WcmException;
import com.sapportals.wcm.protocol.webdav.*;
import com.sapportals.wcm.protocol.webdav.dompath.DomPath;
import com.sapportals.wcm.protocol.webdav.repository.WDProperties;
import com.sapportals.wcm.repository.*;
import com.sapportals.wcm.repository.enum.PropertyType;
import com.sapportals.wcm.util.http.*;
import com.sapportals.wcm.util.name.IName;
import com.sapportals.wcm.util.uri.RID;

/**
 * Abstract class. All implementing classes have access to a check of a
 * normalized multistatus responsecode. To compare the results the path to
 * the response-files is needed. It is stored in the system property
 * 'statusPath' and has the default
 * "c:/home/projects/wcm/testcases/JUnitTestResults/".
 *
 * <p>
 * Copyright (c) SAP AG 2001-2005
 * @author martin.boettcher@greenbytes.de
 * @author julian.reschke@greenbytes.de
 * @version $Id: TestMember.java,v 1.13 2005/02/18 12:46:59 sei Exp $
 */
public abstract class TestMember extends TestCase
{
  public static final String FAILURE_EXT = ".failure";
  public static final String UNTRANSFORMED_PREF = "raw-";
  public static final String PROP_NORMALIZE_PATH = "normalizerPath";
  public static final String PROP_STATUS_PATH = "statusPath";
  public static final String PROP_STATUS_FLAVOUR = "statusFlavour";

  protected static Xslt m_xslt = null;

  private static IRequester requester2;
  private static IRequester adminRequester;
  
  private String testName;

  protected int nr=0;
  private String base;
  protected IRequesterFactory m_factory;
  protected IRequester requester;
  protected HttpRequest request;
  private String localPart;
  protected String servletPath;
  protected String statusPath;
  protected String statusFlavour;
  protected String m_userid;
  protected String m_userid2;
  protected String m_admin;

  protected static final java.text.DecimalFormat THREEDIGITS = new
    java.text.DecimalFormat ("000");
  protected static final java.text.DecimalFormat TWODIGITS = new
    java.text.DecimalFormat ("00");

  private static String UA = System.getProperty("userAgent", null);
  
  /**
   * @param nr the current number of the desired test in the subclass.
   * @param s the name of the testclass.
   */
  public TestMember(Integer nr, String s)
  {
    super(s);
    testName = s;
    this.nr = nr.intValue();
  }

  public void clear(HttpRequest req) {
    req.clear();
    req.setHeader(Consts.SERVLET_TRACE_HEADER, testName + " [" + this.getBase() + this.getLocalpart() + "]");
    if (TestMember.UA != null) request.setHeader("user-agent", TestMember.UA);
  }

  public String getBase() {
    return this.base;
  }

  public String getLocalpart() {
    return this.localPart;
  }

  public String getBaseAndLocalpart() {
    return this.base + this.localPart;
  }

  public boolean sameURI(String uri1, String uri2) {
    return uri1.equals(uri2);
  }

  public String removeProtocolAndBase(String uri) {
    return uri.startsWith(getBase()) ? uri.substring(getBase().length()) : uri;
  }

  public String getRelativeUriFromLocationHeader(IResponse response) {
    String loc = response.getHeader("Location");
    assertTrue("location header missing", loc != null);
    return removeProtocolAndBase(loc);
  }

  public String getRelativeUriFromContentLocationHeader(IResponse response) {
    String loc = response.getHeader("Content-Location");
    assertTrue("'Content-Location' header missing", loc != null);
    return removeProtocolAndBase(loc);
  }

  public String getRelativeUriFromPropertyHrefs(Element container) {
    assertTrue(container != null);
    Set hrefs = DomElement.getHrefs(container);
    assertTrue(new Name(container) + " must contain exactly one DAV:href", hrefs.size() == 1);
    return removeProtocolAndBase((String)hrefs.iterator().next());
  }

  public Set getRelativeUrisFromPropertyHrefs(Element container) {
    Set result = new HashSet();
    assertTrue(container != null);
    Set hrefs = DomElement.getHrefs(container);
    for (Iterator it = hrefs.iterator(); it.hasNext(); /**/) {
      result.add(removeProtocolAndBase((String)it.next()));
    }
    return result;
  }

  public void setRelativeURI(String uri) {
    this.request.setURI(getLocalpart() + uri);
  }

  public void init(IRequesterFactory factory, IRequester rqstr,
                   String bas, String localpart,
                   String svltpath, String userid) {
    m_factory = factory;
    this.base = bas;
    this.localPart = localpart;
    this.servletPath = svltpath;
    this.requester = rqstr;
    this.request = new HttpRequest(localpart);
    clear(this.request);
    request.setURI(localpart);

    m_userid = userid;
    m_userid2 = System.getProperty("user2");
    m_admin = System.getProperty("adminuser");
    
    statusPath = System.getProperty(PROP_STATUS_PATH,
      "c:/projects/WCM3/testcases/JUnitTestResults/");
    statusFlavour = System.getProperty(PROP_STATUS_FLAVOUR, null);

    if (!statusPath.endsWith("/"))
      statusPath += "/";

    if (statusFlavour != null && !statusFlavour.endsWith("/")) {
      statusFlavour += "/";
    }

    if (m_xslt == null) {
      RID umrep = (RID.getRID(localpart)).parent().parent().add("um/");
      m_xslt = new Xslt (localpart, umrep.add("users/").toString(),
          userid, System.getProperty("user2"),
          umrep.add("groups/").toString()
      );
    }
  }

  protected IRequester getSecondRequester() throws WcmException {
    synchronized (TestMember.class) {
      if (TestMember.adminRequester == null) {
        IContext context = m_factory.getContext("user2");
        if (context.isNew()) {
          context.setUserInfo(UserInfo.valueOf(m_userid2, System.getProperty("password2")));
        }
        TestMember.requester2 = m_factory.getRequester(context);
      }

      return TestMember.requester2;
    }
  }

  protected IRequester getAdminRequester() throws WcmException {
    synchronized (TestMember.class) {
      if (TestMember.adminRequester == null) {
        IContext context = m_factory.getContext("admin");
        if (context.isNew()) {
          context.setUserInfo(UserInfo.valueOf(m_admin, System.getProperty("adminpassword")));
        }
        TestMember.adminRequester = m_factory.getRequester(context);
      }

      return TestMember.adminRequester;
    }
  }

  private void writeToFile(String contents, String filename)
  {
    System.err.println("Creating response file: " + filename);
    try{
      File file = new File(filename);
      File parent = file.getParentFile();
      parent.mkdirs();
      java.io.FileOutputStream f = new java.io.FileOutputStream(filename);
      f.write( contents.getBytes() );
      f.close();
    }catch(Exception e){
      System.err.println("could not write response to file: " + e.toString() );
    }
  }

  private String readFromFile (java.io.InputStream f)
  {
    if (f == null) return null;

    try{
      ByteArrayOutputStream bos = new ByteArrayOutputStream();
      byte[] buffer = new byte[64 * 1024];
      int len = 0;
      while ((len = f.read(buffer)) != -1) {
        bos.write(buffer, 0, len);
      }
      f.close();
      return bos.toString();
    }
    catch(Exception e){
      System.err.println("could not read response from file: " + e.toString() );
      return null;
    }
  }

  /**
   * Performs the desired m_request and expects a '207' (Multistatus) return code.
   * The returned document will first have all nodes matching the given
   * DomPath removed, then it will be normalized and compared against the
   * content of a file.
   * @param body the desired body of them_request.
   * @param expected the filename (without path) of the "to-be-compared" contents.
   * @param remove the DomPath matching Nodes to be removed
   * @return the response of the request.
   */
  protected IResponse checkMultistatus (String body, String expected, DomPath remove)
    throws com.sapportals.wcm.WcmException
  {
    return checkStatusAndBody (207, body, expected, remove, true);
  }

  /**
   * Performs the desired m_request and expects a '207' (Multistatus) return code.
   * The return-value will be normalized and compared to the contents of a file.
   * @param body the desired body of them_request.
   * @param expected the filename (without path) of the "to-be-compared" contents.
   * @return the response of the request.
   */
  protected IResponse checkMultistatus (String body, String expected)
    throws com.sapportals.wcm.WcmException
  {
    return checkStatusAndBody (207, body, expected, null, true);
  }

  /**
   * Performs the desired m_request and expects a '207' (Multistatus) return code.
   * The return-value will be normalized and compared to the contents of a file.
   * @param body the desired body of them_request.
   * @param expected the filename (without path) of the "to-be-compared" contents.
   * @return the response of the request.
   */
  protected IResponse checkMultistatusUnsorted (String body, String expected)
    throws com.sapportals.wcm.WcmException
  {
    return checkStatusAndBody (207, body, expected, null, false);
  }

  private String normalizeLineEnds (String inp) {

    StringBuffer s = new StringBuffer("");
    char lastChar = 0;

    for (int i = 0; i < inp.length(); i++) {
      lastChar = inp.charAt(i);
      if (lastChar != '\r') {
        s.append(lastChar);
      }
    }

    if (lastChar != '\n') s.append('\n');
    return s.toString();
  }

  /**
   * Performs the desired m_request and expects a specific return code.
   * The return-value will be normalized and compared to the contents of a file.
   * @param body the desired body of them_request.
   * @param expected the filename (without path) of the "to-be-compared" contents.
   * @return the response of them_request.
   */
  protected IResponse checkStatusAndBody(int expectedStatus,
                                           String body,
                                           String expected)
    throws com.sapportals.wcm.WcmException
  {
    return checkStatusAndBody(expectedStatus, body, expected, null, true);
  }

  /**
   * Performs the desired m_request and expects a specific return code.
   * The return-value will be normalized and compared to the contents of a file.
   * @param body the desired body of them_request.
   * @param expected the filename (without path) of the "to-be-compared" contents.
   * @return the response of them_request.
   */
  protected IResponse checkStatusAndBody(int expectedStatus,
                                           String body,
                                           String expected,
                                           boolean sortResponses)
    throws com.sapportals.wcm.WcmException
  {
    return checkStatusAndBody(expectedStatus, body, expected, null, sortResponses);
  }

  /**
   * Performs the desired m_request and expects a specific return code.
   * The return-value will be normalized and compared to the contents of a file.
   * @param body the desired body of them_request.
   * @param expected the filename (without path) of the "to-be-compared" contents.
   * @param path DomPath matchind Nodes to be removed
   * @return the response of them_request.
   */
  protected IResponse checkStatusAndBody(int expectedStatus,
                                           String body,
                                           String expected,
                                           DomPath path,
                                           boolean sortResponses)
    throws com.sapportals.wcm.WcmException
  {
    this.request.setHeader("Content-Type", "text/xml");
    if (body != null) this.request.setBody(body);
    this.request.expectsResponseDocument(true);
    IResponse response = this.requester.perform(this.request);
    assertTrue ("expected " + expectedStatus + " return code, got " + response.getStatus(),
           response.getStatus() == expectedStatus);

    checkResponseDocument(response, expected, path, sortResponses);
    return response;
  }

  protected void checkResponseDocument(IResponse response,
                                       String expected,
                                       DomPath path)
  {
    checkResponseDocument(response, expected, path, true);
  }

  protected void checkResponseDocument(IResponse response,
                                       String expected,
                                       DomPath path,
                                       boolean sortResponses)
  {
    assertTrue ("expected result document, got null", response.hasDocument() );

//    System.err.println( com.sapportals.wcm.protocol.webdav.DomSerializer.toString(response.getDocument(), true) );

    Document doc = response.getDocument();
    if (path != null) {
      path.deleteNodes(doc);
    }

    String s = m_xslt.transform(doc, sortResponses);
    assertTrue ("failed to normalize result document", s != null);
    s = normalizeLineEnds (s);


    // Determine the response file we are comparing against
    // First look in the flavour directory, if defined
    // Then look in the status directory
    // Create the response file if it is missing
    // Create a response.failure file if it differs
    // delete old .failure files when everything is ok
    //
    String t = null;
    String filePath = null;
    java.io.InputStream referenceResult = null;

    if (statusFlavour != null) {
      referenceResult = this.getClass().getResourceAsStream("results/" + statusFlavour + expected);
      filePath = statusPath + statusFlavour + expected;
      t = readFromFile (referenceResult);
    }
    if (t == null) {
// System.err.println ("results/" + expected);
      referenceResult = this.getClass().getResourceAsStream("results/" + expected);
      filePath = statusPath + expected;
      t = readFromFile(referenceResult);
    }

    if (t == null)
    {
      logComment("(creating: " + expected + ")");
      writeToFile(s, filePath);
      try {
        java.io.File toDelete = new java.io.File (filePath + FAILURE_EXT);
        toDelete.delete();
      }
      catch (Exception e) {
        // ignored
      }
    }
    else if (s != null)
    {
      t = normalizeLineEnds (t);
      if (!s.equals(t)) {
        writeToFile(s, filePath + FAILURE_EXT);
        writeToFile(t, filePath);
        writeToFile(DomSerializer.toString(doc, true), statusPath + UNTRANSFORMED_PREF + expected);
        assertTrue ("multistatus response differs from "+expected+": "+s, false);
      }
    }
  }

  protected void checkResponseCode(IResponse response, int expected) {
    checkResponseCode(response, expected, -1, -1);
  }

  protected void checkResponseCode(IResponse response, int expected1, int expected2) {
    checkResponseCode(response, expected1, expected2, -1);
  }

  protected void checkResponseCode(IResponse response, int expected1,
                                   int expected2, int expected3) {
    int code = response.getStatus();

    if (code == 405) { // not allowed
      String allow = response.getHeader("Allow");
      assertTrue ("allow header must be present when status is NOT ALLOWED",
        allow != null && ! "".equals(allow));
    }

    if (code != expected1 && code != expected2 && code != expected3) {
      StringBuffer sb = new StringBuffer();
      sb.append("expected response codes (").append(expected1);
      if (expected2 >= 0) {
        sb.append(", ").append(expected2);
        if (expected3 >= 0) {
          sb.append(", ").append(expected3);
        }
      }
      sb.append(") got ").append(code);
      Document doc = response.getDocument();
      if (doc != null) {
        sb.append(" with body\n");
        sb.append(DomSerializer.toString(doc, true));
      }
      assertTrue (sb.toString(), false);
    }
  }

  protected void checkErrorElement(IResponse resp, IName errorName) {

    Document doc = resp.getDocument();
    assertTrue ("response document missing", doc != null);
    Element docelem = doc.getDocumentElement();
    assertTrue ("expected document element " + Consts.N_error, Consts.N_error.equalsNode(docelem));
    List found = DomElement.getChildren(docelem, errorName);
    assertTrue ("expected " + errorName + " missing in error element", found.size() == 1);
  }

  protected void checkStatusAndErrorElement(IResponse resp, int status, IName errorName) {

    assertTrue("expected status: " + status + ", got: " + resp.getStatus(), resp.getStatus() == status);
    checkErrorElement(resp, errorName);
  }

  static private boolean LOGIT = "true".equals(System.getProperty("logIdentifier", "false"));

  protected void logIdentifier(String identifier) {
    if (LOGIT) {
      System.out.print("[" + identifier + "]");
    }
  }

  protected void logComment(String comment) {
    if (LOGIT) {
      System.out.print("{" + comment + "}");
    }
  }

  private static String toXml (Name name) {
    return "<" + name.getLocalName() + " xmlns='" + name.getNamespaceUri() + "'/>";
  }

  private void assertMethodAllowed(String requestUri, String method, boolean assertion) throws com.sapportals.wcm.WcmException {

    clear(request);
    request.setMethod("OPTIONS");
    request.setHeader("x-greenbytes-resource-context", "allow-bind-method=yes");
    request.setURI(requestUri);
    IResponse response = requester.perform(request);
    checkResponseCode(response, 200);

    String allow = response.getHeader("allow");
    assertTrue("Allow header must be present in OPTIONS response", allow != null);

    StringTokenizer st = new StringTokenizer(allow, ", ");
    while (st.hasMoreElements()) {
      String s = st.nextToken();
      if (s.equals(method)) {
        if (assertion) {
          return;
        }
        else {
          fail("method " + method + " is allowed, but shouldn't be, header was: " + allow);
        }
      }
    }

    if (assertion) {
      fail("method " + method + " isn't allowed, but should be, header was: " + allow);
    }
  }

  protected void assertMethodAllowed(String requestUri, String method) throws com.sapportals.wcm.WcmException {
    assertMethodAllowed(requestUri, method, true);
  }

  protected void assertMethodNotAllowed(String requestUri, String method) throws com.sapportals.wcm.WcmException {
    assertMethodAllowed(requestUri, method, false);
  }

  protected List getHrefsFromMultistatus(Document doc) {

    List result = new ArrayList();
    List resp = DomElement.getChildren(doc.getDocumentElement(), Consts.N_response);

    for (Iterator it = resp.listIterator(); it.hasNext(); /**/) {
      String name = getRelativeUriFromPropertyHrefs((Element) it.next());
      result.add(name);
    }

    return result;
  }

  protected void checkSortedResultSet(IResponse resp, String uris[]) throws Exception {
    checkSortedResultSet(resp, uris, (String) null);
  }

  protected void checkSortedResultSet(IResponse resp, String uris[], String ignore) throws Exception {

    List hrefs = getHrefsFromMultistatus(resp.getDocument());

    if (ignore != null) {
      assertTrue("result set doesn't contain member to be ignored: " + ignore, hrefs.contains(ignore));
      hrefs.remove(ignore);
    }

    assertTrue("number of search results wrong, got: "
               + hrefs + ", expected: " + Arrays.asList(uris),
               hrefs.size() == uris.length);

    Iterator hr = hrefs.iterator();

    for (int i = 0; i < uris.length; i++) {
      String t = hr.next().toString();
      assertTrue("mismatch for entry " + i + ": " + uris[i] + " " + t,
                 t.equals(uris[i]));
    }
  }

  protected void checkSortedResultSet(IResponse resp, String uris[], Set ignore) throws Exception {

    List hrefs = getHrefsFromMultistatus(resp.getDocument());

    if (ignore != null) {
      Iterator iter = ignore.iterator();

      while (iter.hasNext()) {
        String href = (String) iter.next();
        assertTrue("result set doesn't contain member to be ignored: " + href, hrefs.contains(href));
        hrefs.remove(href);
      }
    }

    assertTrue("number of search results wrong, got: "
               + hrefs + ", expected: " + Arrays.asList(uris),
               hrefs.size() == uris.length);

    Iterator hr = hrefs.iterator();

    for (int i = 0; i < uris.length; i++) {
      String t = hr.next().toString();
      assertTrue("mismatch for entry " + i + ": " + uris[i] + " " + t,
                 t.equals(uris[i]));
    }
  }

  protected void checkResultSet(IResponse resp, String uris[]) throws Exception {

    List hrefs = getHrefsFromMultistatus(resp.getDocument());

    assertTrue("number of search results wrong, got: "
               + hrefs + ", expected: " + Arrays.asList(uris),
               hrefs.size() == uris.length);

    for (int i = 0; i < uris.length; i++) {
      assertTrue(uris[i] + " missing from result set", hrefs.contains(uris[i]));
    }
  }

  private static final DomPath DP_propsPath = 
    new DomPath("/D:multistatus/D:response/D:propstat[contains(D:status,'200')]/D:prop/*", DomPath.DEFAULT_NAMESPACE_BINDINGS);
  private static final DomPath DP_propsNfPath = 
    new DomPath("/D:multistatus/D:response/D:propstat[contains(D:status,'404')]/D:prop/*", DomPath.DEFAULT_NAMESPACE_BINDINGS);

  protected Element getProperty(String requestUri, Name property) throws com.sapportals.wcm.WcmException {

    clear(request);
    request.setMethod("PROPFIND");
    request.setURI(requestUri);
    request.setHeader("Depth", "0");
    request.setHeader("Cache-Control", "no-cache");
    request.setHeader("Translate", "f");
    request.setHeader("Apply-To-Redirect-Ref", "t");
    request.expectsResponseDocument(true);
    request.setBody("<propfind xmlns='DAV:'><prop>" +
      "<" + property.getLocalName() + " xmlns='" + property.getNamespaceUri() + "'/>" +
      "</prop>" + toXml(Consts.N_SAP_include_hidden_flag) + toXml(Consts.N_SAP_include_protected_flag) +
      "</propfind>");
    IResponse response = requester.perform(request);
    checkResponseCode(response, 207);

    Document doc = response.getDocument();
    if (doc == null) return null;

    List props = DP_propsPath.match(doc);
    for (Iterator it = props.iterator(); it.hasNext(); ) {
      Node n = (Node)it.next();
      if (property.equalsNode(n)) {
        return (Element) n;
      }
    }

    // 404ed?
    props = DP_propsNfPath.match(doc);
    for (Iterator it = props.iterator(); it.hasNext(); ) {
      Node n = (Node)it.next();
      if (property.equalsNode(n)) {
        return null;
      }
    }

    fail("property " + property + " not reported with status 200 or 404: " + DomSerializer.toString(response.getDocument(), true));
    return null;
  }

  protected DavStatus removeProperty(String requestUri, IName property) throws com.sapportals.wcm.WcmException {

    Document doc = DomBuilder.newDocument(Consts.N_propertyupdate);
    Element set = DomElement.makeNew(doc, Consts.N_remove);
    Element prop = DomElement.makeNew(doc, Consts.N_prop);

    doc.getDocumentElement().appendChild(set);
    set.appendChild(prop);

    prop.appendChild(DomElement.makeNew(doc, property));

    clear(request);
    request.setMethod("PROPPATCH");
    request.setURI(requestUri);
    request.setHeader(Consts.SERVLET_TRACE_HEADER, testName);
    request.setBody(DomSerializer.toString(doc, false));
    // System.out.println(DomSerializer.toString(doc, true));
    IResponse response = requester.perform(request);
    if (response.getStatus() != 207) {
      return new DavStatus(response.getStatus());
    }

    String resptxt = DomSerializer.toString(response.getDocument(), true);

    DomPath propsPath = new DomPath("/D:multistatus/D:response/D:propstat/D:prop/*", DomPath.DEFAULT_NAMESPACE_BINDINGS);
    List props = propsPath.match(response.getDocument());
    assertTrue("response body contains multiple properties: " + resptxt, props.size() == 1);

    Element propstat = (Element) ((Node)props.get(0)).getParentNode().getParentNode();
    // System.out.println(DomSerializer.toString(propstat, true));
    List statuss = DomElement.getChildren(propstat, Consts.N_status);
    assertTrue("propstat element should only have one status child element: " + resptxt, statuss.size() == 1);

    Element status = (Element)statuss.get(0);
    com.sapportals.wcm.util.http.HttpStatus wds = new com.sapportals.wcm.util.http.HttpStatus();
    wds.parse(DomElement.getText(status));

    DomPath DP_conditions = new DomPath("D:responsedescription/D:error/*", DomPath.DEFAULT_NAMESPACE_BINDINGS);
    List conditions = DP_conditions.match(propstat);
    Name condition = conditions.isEmpty() ? null : new Name((Element)conditions.get(0));

    return new DavStatus(wds.getCode(), condition);
  }

  protected DavStatus setProperty(String requestUri, IName property, Object value, boolean checkTypePreserved) throws com.sapportals.wcm.WcmException {

    Document doc = DomBuilder.newDocument(Consts.N_propertyupdate);
    Element set = DomElement.makeNew(doc, Consts.N_set);
    Element prop = DomElement.makeNew(doc, Consts.N_prop);

    doc.getDocumentElement().appendChild(set);
    set.appendChild(prop);

    boolean mv = value instanceof List;
    Object typ = mv
      ? (((List)value).size() == 0) ? "" : ((List)value).get(0)
      : value;

    PropertyType ptype;
    if (typ instanceof Integer) {
      ptype = PropertyType.INTEGER;
    }
    else if (typ instanceof Long) {
      ptype = PropertyType.LONG;
    }
    else if (typ instanceof Boolean) {
      ptype = PropertyType.BOOLEAN;
    }
    else if (typ instanceof Date) {
      ptype = PropertyType.DATE;
    }
    else if (typ instanceof XMLMarkup) {
      ptype = PropertyType.XML;
    }
    else {
      ptype = PropertyType.STRING;
    }

    IPropertyName pn = new PropertyName(property.getNamespace(), property.getName());
    IProperty p = mv
      ? new Property(pn, new PropertyDef(ptype, mv), (List)value)
      : new Property(pn, value);

    WDProperties.appendProperty(prop, pn, p, new com.sapportals.wcm.util.http.DateFormat(), null);

    clear(request);
    request.setMethod("PROPPATCH");
    request.setURI(requestUri);
    request.setHeader(Consts.SERVLET_TRACE_HEADER, testName);
    request.setBody(DomSerializer.toString(doc, false));
    // System.out.println(DomSerializer.toString(doc, true));
    IResponse response = requester.perform(request);
    if (response.getStatus() != 207) {
      return new DavStatus(response.getStatus());
    }

    String resptxt = DomSerializer.toString(response.getDocument(), true);

    DomPath propsPath = new DomPath("/D:multistatus/D:response/D:propstat/D:prop/*", DomPath.DEFAULT_NAMESPACE_BINDINGS);
    List props = propsPath.match(response.getDocument());
    assertTrue("response body contains multiple properties: " + resptxt, props.size() == 1);

    Name xsdt = XsdDataTypes.getXsdTypeInfo((Element)props.get(0));

    Element propstat = (Element) ((Node)props.get(0)).getParentNode().getParentNode();
    // System.out.println(DomSerializer.toString(propstat, true));
    List statuss = DomElement.getChildren(propstat, Consts.N_status);
    assertTrue("propstat element should only have one status child element: " + resptxt, statuss.size() == 1);

    Element status = (Element)statuss.get(0);
    HttpStatus wds = new HttpStatus();
    wds.parse(DomElement.getText(status));

    DomPath DP_conditions = new DomPath("D:responsedescription/D:error/*", DomPath.DEFAULT_NAMESPACE_BINDINGS);
    List conditions = DP_conditions.match(propstat);
    Name condition = conditions.isEmpty() ? null : new Name((Element)conditions.get(0));

    if (checkTypePreserved && wds.getCode() == 200 && condition == null) {

      condition = DavConditions.N_SAP_property_type_preserved;

      if (mv && XsdDataTypes.SOAP_ARRAY.equals(xsdt)) {
        condition = null;
      }
      else if (typ instanceof Date && XsdDataTypes.XSD_DATETIME.equals(xsdt)) {
        condition = null;
      }
      else if (typ instanceof Integer && XsdDataTypes.XSD_INT.equals(xsdt)) {
        condition = null;
      }
      else if (typ instanceof Long && XsdDataTypes.XSD_LONG.equals(xsdt)) {
        condition = null;
      }
      // add more positive checks
    }

    return new DavStatus(wds.getCode(), condition);
  }

  protected void setProperty(String requestUri, IName property, Object value, boolean checkTypePreserved, int expected) throws com.sapportals.wcm.WcmException {

    DavStatus ds = setProperty(requestUri, property, value, checkTypePreserved);
    assertTrue("wrong property status " + ds + ", expected: " + expected, ds.getStatus() == expected);
  }

  protected void setProperty(String requestUri, IName property, Object value, boolean checkTypePreserved, DavStatus expected) throws com.sapportals.wcm.WcmException {

    DavStatus ds = setProperty(requestUri, property, value, checkTypePreserved);
    assertTrue("wrong property status " + ds + ", expected: " + expected, ds.equals(expected));
  }

  protected void safeDelete(String requestUri, String locktoken) throws com.sapportals.wcm.WcmException {

    if (locktoken != null) {
      clear(request);
      request.setMethod("UNLOCK");
      request.setURI(requestUri);
      request.setHeader("Lock-Token", "<"+locktoken+">");
      requester.perform(request);
    }

    clear(request);
    request.setMethod("DELETE");
    request.setURI(requestUri);
    requester.perform(request);
  }

  /**
   * to be implemented by the subclasses.
   */
  protected abstract void runTest();

  public static String getLockToken(IResponse response) {
    String lt = response.getHeader("Lock-Token");
    assertTrue("lock token header missing", lt != null);
    assertTrue("lock token '" + lt + "' contains blank(s)", -1 == lt.indexOf(" "));
    return lt.substring(1, lt.length() - 1);
  }
  
  private static Boolean VHR_PRESERVED = null;
  
  /**
   * @return whether this repository preserves version histories
   */
  public boolean vhrPreserved() throws WcmException {

    if (VHR_PRESERVED == null) {

      String vhr = null;

      // create a test resource, VCR it, check VHR, Delete
      clear(request);
      request.setMethod("PUT");
      request.setURI(localPart + "/vcrtest");
      request.setBody("");
      IResponse response = requester.perform(request);
      checkResponseCode(response, 200, 201);

      clear(request);
      request.setMethod("VERSION-CONTROL");
      request.setURI(localPart + "/vcrtest");
      requester.perform(request);
      response = requester.perform(request);

      Element e = getProperty(localPart + "/vcrtest", Consts.N_version_history);
      if (e != null) {
        vhr = getRelativeUriFromPropertyHrefs(e);
      }

      clear(request);
      request.setMethod("DELETE");
      request.setURI(localPart + "/vcrtest");
      response = requester.perform(request);
      checkResponseCode(response, 204);

      if (vhr != null) {
        // is VHR still present?
        clear(request);
        request.setMethod("DELETE");
        request.setURI(vhr);
        response = requester.perform(request);
        VHR_PRESERVED = (response.getStatus() < 400) ? Boolean.TRUE : Boolean.FALSE;
      }
      else {
        // not versioning
        VHR_PRESERVED = Boolean.FALSE;
      }

    }

    return VHR_PRESERVED.booleanValue();
  }
  
  
}
