package com.sapportals.wcm.protocol.webdav.server.tests;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;

import org.w3c.dom.Element;

import com.sapportals.wcm.protocol.webdav.Consts;
import com.sapportals.wcm.protocol.webdav.DavConditions;
import com.sapportals.wcm.protocol.webdav.DomElement;
import com.sapportals.wcm.util.http.IResponse;

/**
 * Tests mappings of various URIs ro resources (verifies that there aren't
 * multiple valid RIDs mapped to the same resource, which would break ACL
 * handling).
 * <p>
 * Copyright (c) SAP AG 2002-2003
 * @author julian.reschke@greenbytes.de
 * @version $Id: UriToResourceMappingTest.java,v 1.3 2004/03/29 11:52:07 jre Exp $
 */

public class UriToResourceMappingTest extends TestMember {

  private static final String[] desc = new String[]{
    "00 URM setup",
    "01 URM check 'urm' and 'urm/'",
    "02 URM check 'urm.'",
    "03 URM check 'URM'",
    "04 URM create test file 'file'",
    "05 URM create test file 'FILE'",
    "06 URM create test file 'file.'",
    "07 URM check content lengths",
    "08 URM create dummy file",
    "09 URM copy to 'FILE'",
    "10 URM copy to 'file.'",
    "11 URM move to 'FILE'",
    "12 URM check 'prefix./urm'",
    "13 URM check 'prefix%20/urm'",
    "14 URM check 'urm#'",
    "15 URM check 'urm#123'",
    "   URM cleanup",
  };

  public static int NrOfTests = desc.length;

  public UriToResourceMappingTest(Integer nr) {
    super(nr, desc[nr.intValue()]);
  }

  private static boolean caseInsensitive = false;
  private static boolean trailingDots = false;
  private static boolean allowsBothUpperAndLower = false;

  private static String content1 = "1234";
  private static String content2 = "12345";
  private static String content3 = "123456";

  protected void runTest() {
    if (nr == 0) super.logIdentifier("URM");

    try {
      IResponse response;
      clear(request);

      switch (nr) {

        case 0:
          clear(request);
          request.setMethod("MKCOL");
          request.setURI(getLocalpart() + "/urm");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;

        case 1: {
            Element rt1 = getProperty(getLocalpart() + "/urm", Consts.N_resourcetype);
            assertTrue (rt1 != null && DomElement.hasChild(rt1, Consts.N_collection));
            Element rt2 = getProperty(getLocalpart() + "/urm/", Consts.N_resourcetype);
            assertTrue (rt2 != null && DomElement.hasChild(rt2, Consts.N_collection));
          }
          break;

        case 2:
          caseInsensitive = ! assertNonExistantOrEqual(getLocalpart() + "/urm.", getLocalpart() + "/urm");
          break;

        case 3:
          assertNonExistantOrEqual(getLocalpart() + "/URM", getLocalpart() + "/urm");
          break;

        case 4:
          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/urm/file");
          request.setBody(content1);
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;

        case 5:
          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/urm/FILE");
          request.setBody(content2);
          response = requester.perform(request);
          if (caseInsensitive) {
            checkResponseCode(response, 200);
            String contentloc = getRelativeUriFromContentLocationHeader(response);
            assertTrue((getLocalpart() + "/urm/file").equals(contentloc));
            content1 = content2;
          }
          else {
            checkResponseCode(response, 201, 409);
            if (response.getStatus() == 409) {
              content2 = null;  // wasn't written
              checkErrorElement(response, DavConditions.N_name_allowed);
            }
          }

          allowsBothUpperAndLower = response.getStatus() == 201;
          break;

        case 6:
          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/urm/file.");
          request.setBody(content3);
          response = requester.perform(request);
          if (caseInsensitive) {
            checkResponseCode(response, 200);
            String contentloc = getRelativeUriFromContentLocationHeader(response);
            assertTrue((getLocalpart() + "/urm/file").equals(contentloc));
            content1 = content3;
            content2 = content3;
          }
          else {
            checkResponseCode(response, 201, 409);
            if (response.getStatus() == 409) {
              checkErrorElement(response, DavConditions.N_name_allowed);
              content3 = null;  // not written
            }
          }

          trailingDots = response.getStatus() == 201;
          break;

        case 7:
          clear(request);
          request.setMethod("GET");
          request.setURI(getLocalpart() + "/urm/file");
          request.setHeader("Accept-Encoding", "");
          request.expectsResponseStream(true);
          response = requester.perform(request);
          checkResponseCode(response, 200);
          assertTrue(response.getHeader("content-length").equals(""+content1.length()));
          assertTrue(content1.equals(getContent(response.getStream())));

          clear(request);
          request.setMethod("GET");
          request.setURI(getLocalpart() + "/urm/FILE");
          request.setHeader("Accept-Encoding", "");
          request.expectsResponseStream(true);
          response = requester.perform(request);
          if (content2 == null) {
            checkResponseCode(response, 404);
          }
          else {
            checkResponseCode(response, 200);
            assertTrue(response.getHeader("content-length").equals(""+content2.length()));
            assertTrue(content2.equals(getContent(response.getStream())));
          }

          clear(request);
          request.setMethod("HEAD");
          request.setURI(getLocalpart() + "/urm/file.");
          request.setHeader("Accept-Encoding", "");
          response = requester.perform(request);
          checkResponseCode(response, 200, 404);
          if (response.getStatus() == 200) {
            assertTrue(response.getHeader("content-length").equals("6"));
          }
          break;

        case 8:
          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/urm/source");
          request.setBody("12345678");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;

        case 9:
          clear(request);
          request.setMethod("COPY");
          request.setURI(getLocalpart() + "/urm/source");
          request.setHeader("Destination", getLocalpart() + "/urm/FILE");
          response = requester.perform(request);
          if (allowsBothUpperAndLower) {
            checkResponseCode(response, 204);
          }
          else if (caseInsensitive) {
            checkResponseCode(response, 204);
          }
          else {
            checkResponseCode(response, 400, 409);
            if (response.getStatus() == 409) checkErrorElement(response, DavConditions.N_name_allowed);
          }
          break;

        case 10:
          clear(request);
          request.setMethod("COPY");
          request.setURI(getLocalpart() + "/urm/source");
          request.setHeader("Destination", getLocalpart() + "/urm/file.");
          response = requester.perform(request);
          if (trailingDots) {
            checkResponseCode(response, 204);
          }
          else if (caseInsensitive) {
            checkResponseCode(response, 204);
          }
          else {
            checkResponseCode(response, 400, 409);
            if (response.getStatus() == 409) checkErrorElement(response, DavConditions.N_name_allowed);
          }
          break;

        case 11:
          clear(request);
          request.setMethod("MOVE");
          request.setURI(getLocalpart() + "/urm/source");
          request.setHeader("Destination", getLocalpart() + "/urm/FILE");
          response = requester.perform(request);
          if (allowsBothUpperAndLower) {
            checkResponseCode(response, 204);
          }
          else {
            checkResponseCode(response, 400, 409);
            if (response.getStatus() == 409) checkErrorElement(response, DavConditions.N_name_allowed);
          }
          break;

        case 12:
          assertNonExistantOrEqual(getLocalpart() + "./urm", getLocalpart() + "/urm");
          break;

        case 13:
          assertNonExistantOrEqual(getLocalpart() + "%20/urm", getLocalpart() + "/urm");
          break;

        case 14:
          assertBadRequest(getLocalpart() + "urm#");
          break;

        case 15:
          assertBadRequest(getLocalpart() + "urm#123");
          break;

        default:
          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/urm");
          response = requester.perform(request);
          checkResponseCode(response, 200, 204);
          break;
      }
    }
    catch (Exception ex) {
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }

  public boolean assertNonExistantOrEqual(String uri, String tolerated) throws com.sapportals.wcm.WcmException {
    clear(request);
    request.setMethod("PROPFIND");
    request.setHeader("Depth", "0");
    request.setURI(uri);
    IResponse response = requester.perform(request);
    checkResponseCode(response, 207, 404);
    if (response.getStatus() == 404) {
      // all is well
      return true;
    }
    else {
      String contentlocation = response.getHeader("content-location");
      assertTrue(tolerated != null);
      assertTrue("should have sent content-location header", contentlocation != null);
      String rel = getRelativeUriFromContentLocationHeader(response);
      assertTrue(rel.equals(tolerated));
      return false;
    }
  }

  public void assertBadRequest(String uri) throws com.sapportals.wcm.WcmException {
    clear(request);
    request.setMethod("PROPFIND");
    request.setHeader("Depth", "0");
    request.setURI(uri);
    IResponse response = requester.perform(request);
    checkResponseCode(response, 400);
  }

  private String getContent(InputStream is) throws IOException {
    assertTrue(is != null);

    byte[] buffer = new byte[64 * 1024];
    ByteArrayOutputStream bos = new ByteArrayOutputStream();
    int len;
    while ((len = is.read(buffer)) != -1) {
      bos.write(buffer, 0, len);
    }
    is.close();

    return bos.toString();
  }


}
