package com.sapportals.wcm.protocol.webdav.server.tests;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.sapportals.wcm.protocol.webdav.*;
import com.sapportals.wcm.protocol.webdav.dompath.DomPath;
import com.sapportals.wcm.protocol.webdav.dompath.DomPathAssertionFailedException;
import com.sapportals.wcm.util.http.IResponse;
import com.sapportals.wcm.util.uri.IHierarchicalUri;
import com.sapportals.wcm.util.uri.UriFactory;

/**
 * Various tests for Versioning methods.
 *
 * <p>
 * Copyright (c) SAP AG 2001-2005
 * @author stefan.eissing@greenbytes.de
 * @author julian.reschke@greenbytes.de
 * @version $Id: VersionTest.java,v 1.9 2005/02/18 12:46:59 sei Exp $
 */
public class VersionTest extends TestMember
{
  private static final String[] desc = new String[]
  {
    "00 VS Setting up Versioning",
    "01 VS CHECKOUT resource",
    "02 VS PUT new content resource",
    "03 VS CHECKIN resource",
    "04 VS check content of resource",
    "05 VS CHECKOUT again resource",
    "06 VS PUT new content resource",
    "07 VS UNCHECKOUT resource",
    "08 VS check content of resource",
    "09 VS CHECKIN of checked in resource",
    "10 VS PUT without checkout in resource",
    "11 VS check content of resource",
    "12 VS CHECKOUT twice of resource",
    "13 VS REPORT version-tree of resource",
    "14 VS PROPFIND resource checkout-set",
    "15 VS PROPFIND resource-version checkout-set",
    "16 VS (unused)",
    "17 VS CHECKIN with wrong expected checkinUri",
    "18 VS CHECKIN with correct expected checkinUri",
    "19 VS CHECKOUT with compute-checkin-Uri",
    "20 VS checked-out property after checkout",
    "21 VS DELETE version while being checked out",
    "22 VS REPORT version tree of version resource",
    "23 VS retry DELETE after CHECKIN",
    "24 VS REPORT version-tree of versionable but non-versioned resource",
    "25 VS REPORT expand-property on VCR",
    "26 VS REPORT with wrong report name",
    "27 VS PUT/PROPPATCH on version resource",
    "28 VS COPY/Overwrite-T on version resource",
    "29 VS COPY/Overwrite-F on version resource",
    "30 VS COPY/Overwrite-F collection to collection",
    "31 VS COPY/Overwrite-T collection to collection",
    "32 VS LABEL - OPTIONS must not return label feature",
    "33 VS LABEL - PROPFIND DAV:supported-method-set",
    "34 VS LABEL non-versioned resource",
    "35 VS add one label on resource",
    "36 VS add same label again",
    "37 VS add same label on parent collection",
    "38 VS add new label on parent collection",
    "39 VS checkout one resource",
    "40 VS try to label checked-out version",
    "41 VS checkin one resource",
    "42 VS set existing label on parent collection",
    "43 VS remove a label on parent collection",
    "44 VS report expand-property on versioned resource",
    "45 VS test CHECKIN keep-checked-out",
    "46 VS REPORT labeled-version",
    "47 VS copy version",
    "48 VS PROPPATCH on auto-versioned resource",
    "49 VS create vcr with three versions, add labels only differing by leading/trailing blank on different versions",
    "50 VS add labels only differing by leading/trailing blank on the same version",
    "51 VS report labels only differing by leading/trailing blank on the same version",
    "52 VS DELETE versions descending",
    "53 VS DELETE versions ascending",
    "54 VS delete root version after deleting it's unique successor",
    "   VS cleanup",
  };

  public static final int NrOfTests = desc.length;

  // make sure all have different sizes
  private static final String CONTENT_1       = "this is the content of the first version (VS 00)   x";
  private static final String CONTENT_2       = "this is the content of the second version (VS 02)  xx";
  private static final String CONTENT_3       = "this is the content of the third version (VS 10)   xxxx";
  private static final String CONTENT_IGNORED = "this is the content of the ignored version (VS 06) xxxxx";

  private static DomPath DP_co_href = null;
  private static DomPath DP_checked_in = null;
  private static DomPath DP_s_m_s_label = null;

  static {
    DP_co_href = new DomPath ("/D:multistatus/D:response/D:propstat[contains(D:status,'200')]/D:prop/D:checked-out/D:href",
      DomPath.DEFAULT_NAMESPACE_BINDINGS);
    DP_checked_in = new DomPath ("/D:multistatus/D:response/D:propstat[contains(D:status,'200')]/D:prop/D:checked-in/D:href",
      DomPath.DEFAULT_NAMESPACE_BINDINGS);
    DP_s_m_s_label = new DomPath ("/D:multistatus/D:response/D:propstat/D:prop/D:supported-method-set/D:supported-method[@name = 'LABEL']",
      DomPath.DEFAULT_NAMESPACE_BINDINGS);
  }

  private static String rootVersion = null;
  private static String expectedUri = null;
  private static String checkedOutUri = null;
  private static String someVersionUri = null;
  private static boolean hasLabelMethod;
  
  private static Set vhrs = new HashSet();
 
  public VersionTest(Integer nr)
  {
    super(nr, desc[nr.intValue()]);
  }

  protected void runTest()
  {
    if (nr == 0) super.logIdentifier("VS");

    String respfile = "VERSIONING-" + TestMember.TWODIGITS.format(nr) + ".txt";

    try
    {
      IResponse response;
      Element href;
      String content;

      clear(request);
      switch (nr)
      {
        case 0:
          request.setMethod("MKCOL");
          request.setURI(getLocalpart() + "/versioning");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          request.setBody(CONTENT_1);
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("PROPPATCH");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          request.setBody("<propertyupdate xmlns='DAV:'><set><prop><comment xml:lang='en'>Initial version put under version control</comment></prop></set></propertyupdate>");
          response = requester.perform(request);
          checkResponseCode(response, 207);

          clear(request);
          request.setMethod("OPTIONS");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          response = requester.perform(request);
          checkResponseCode(response, 200);
          String allow = response.getHeader("Allow");
          assertTrue("VERSION-CONTROL missing in Allow Header: "+allow, allow.indexOf("VERSION-CONTROL") >= 0);

          clear(request);
          request.setMethod("VERSION-CONTROL");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          rootVersion = getRelativeUriFromLocationHeader(response);
          
          Element vhr = getProperty(getLocalpart() + "/versioning/resource.txt", Consts.N_version_history);
          if (vhr != null) {
            vhrs.add(getRelativeUriFromPropertyHrefs(vhr));
          }

          clear(request);
          request.setMethod("REPORT");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          checkMultistatus("<version-tree xmlns='DAV:'><prop><comment/><version-name/><successor-set/></prop></version-tree>",
            respfile);
          break;
          // revision 1 created

        // this test is going to fail if for some reason the new resource
        // is initially checked out (SharePoint). However, the remaining
        // tests should then work (working on a known state).
        case 1: {
            request.setMethod("CHECKOUT");
            request.setURI(getLocalpart() + "/versioning/resource.txt");
            response = requester.perform(request);
            checkResponseCode(response, 200);
            request.setMethod("REPORT");
            request.setURI(getLocalpart() + "/versioning/resource.txt");
            checkMultistatus("<version-tree xmlns='DAV:'><prop><comment/><version-name/><successor-set/></prop></version-tree>",
              respfile);

            // the root version should now be checked out
            Element cos = getProperty(rootVersion, Consts.N_checkout_set);
            String cor = getRelativeUriFromPropertyHrefs(cos);
            assertTrue(cor.equals(getLocalpart() + "/versioning/resource.txt"));
          }
          break;

        case 2:
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          request.setBody(CONTENT_2);
          response = requester.perform(request);
          checkResponseCode(response, 200);
          break;

        case 3:
          clear(request);
          request.setMethod("PROPPATCH");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          request.setBody("<propertyupdate xmlns='DAV:'><set><prop><comment xml:lang='de'>Checkin-Testcase 3</comment></prop></set></propertyupdate>");
          response = requester.perform(request);
          checkResponseCode(response, 207);

          clear(request);
          request.setMethod("CHECKIN");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          someVersionUri = response.getHeader("Location");
          assertTrue("Location header after CHECKIN missing", someVersionUri != null);

          clear(request);
          request.setMethod("REPORT");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          checkMultistatus("<version-tree xmlns='DAV:'><prop><comment/><version-name/><successor-set/></prop></version-tree>",
            respfile);
          break;
          // revision 2 created

        case 4:
          request.setMethod("GET");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          request.expectsResponseStream(true);
          response = requester.perform(request);
          checkResponseCode(response, 200);
          content = getContent(response.getStream());
          assertTrue ("unexpected content: "+content, CONTENT_2.equals(content));
          break;

        case 5:
          request.setMethod("CHECKOUT");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          response = requester.perform(request);
          checkResponseCode(response, 200);
          break;

        case 6:
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          request.setBody(CONTENT_IGNORED);
          response = requester.perform(request);
          checkResponseCode(response, 200);
          break;

        case 7: {
            clear(request);
            request.setMethod("UNCHECKOUT");
            request.setURI(getLocalpart() + "/versioning/resource.txt");
            response = requester.perform(request);
            checkResponseCode(response, 200);
  
            clear(request);
            request.setMethod("GET");
            request.setURI(getLocalpart() + "/versioning/resource.txt");
            request.expectsResponseStream(true);
            request.setHeader("Translate", "f");
            response = requester.perform(request);
            checkResponseCode(response, 200);
            content = getContent(response.getStream());
            assertTrue("unexpected content: "+content, CONTENT_2.equals(content));
 
            clear(request);
            request.setMethod("REPORT");
            request.setURI(getLocalpart() + "/versioning/resource.txt");
            checkMultistatus("<version-tree xmlns='DAV:'><prop><comment/><version-name/><successor-set/></prop></version-tree>",
              respfile);
  
            clear(request);
            request.setMethod("UNCHECKOUT");
            request.setURI(getLocalpart() + "/versioning/resource.txt");
            response = requester.perform(request);
            checkStatusAndErrorElement(response, 409, DavConditions.N_must_be_checked_out_version_controlled_resource);
        }
          break;
          // still at revision 2

        case 8:
          request.setMethod("GET");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          request.expectsResponseStream(true);
          response = requester.perform(request);
          checkResponseCode(response, 200);
          content = getContent(response.getStream());
          assertTrue ("unexpected content: "+content, CONTENT_2.equals(content));
          break;

        case 9:
          request.setMethod("CHECKIN");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          response = requester.perform(request);
          checkResponseCode(response, 409);
          request.clear();
          request.setMethod("REPORT");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          checkMultistatus("<version-tree xmlns='DAV:'><prop><comment/><version-name/><successor-set/></prop></version-tree>",
            respfile);
          request.clear();
          break;
          // still at revision 2

        case 10:
          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          request.setBody(CONTENT_3);
          response = requester.perform(request);
          checkResponseCode(response, 200);

          request.clear();
          request.setMethod("REPORT");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          checkMultistatus("<version-tree xmlns='DAV:'><prop><comment/><version-name/><successor-set/></prop></version-tree>",
            respfile);
          request.clear();
          break;
          // revision 3 created

        case 11:
          request.setMethod("GET");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          request.expectsResponseStream(true);
          response = requester.perform(request);
          checkResponseCode(response, 200);
          content = getContent(response.getStream());
          assertTrue ("unexpected content: "+content, CONTENT_3.equals(content));
          break;

        case 12:
          request.setMethod("CHECKOUT");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          clear(request);
          request.setMethod("CHECKOUT");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          response = requester.perform(request);
          checkStatusAndErrorElement(response, 409, DavConditions.N_must_be_checked_in);

          clear(request);
          request.setMethod("PROPPATCH");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          request.setBody("<propertyupdate xmlns='DAV:'><set><prop><comment>checkin test case 12</comment></prop></set></propertyupdate>");
          response = requester.perform(request);
          checkResponseCode(response, 207);

          clear(request);
          request.setMethod("CHECKIN");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("CHECKIN");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          response = requester.perform(request);
          checkStatusAndErrorElement(response, 409, DavConditions.N_must_be_checked_out);
          break;
          // revision 4 created

        case 13:
          request.setMethod("REPORT");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          checkMultistatus("<version-tree xmlns='DAV:'><prop><comment/><version-name/><successor-set/></prop></version-tree>",
            respfile);
          break;

        case 14:
          request.setMethod("CHECKOUT");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          // tests that the live property checkout-set is NOT set on the VCR
          clear(request);
          request.setMethod("PROPFIND");
          request.setHeader("Depth", "0");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          checkMultistatus("<propfind xmlns='DAV:'><prop><comment/><predecessor-set/><checkout-set/></prop></propfind>",
            respfile);
          break;

        case 15:
          request.setMethod("PROPFIND");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          request.setHeader("Depth", "0");
          request.setBody("<propfind xmlns='DAV:'><prop><comment/><checked-out/></prop></propfind>");
          request.expectsResponseDocument();
          response = requester.perform(request);
          checkResponseCode(response, 207);
          href = (Element) DP_co_href.matchOrDie (response.getDocument(), 1, 1, "checked-out/href not found").get(0);
          checkedOutUri = DomElement.getText(href);

          // check that the referenced version actually exists
          clear(request);
          request.setMethod("HEAD");
          request.setURI(checkedOutUri);
          response = requester.perform(request);
          checkResponseCode(response, 200);
          break;

        case 16:
          // unused
          break;

        case 17:
          String wrongLocation = "http://greenbytes.de";
          request.setMethod("CHECKIN");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          request.setBody("<checkin xmlns='DAV:'><expected-version-URI xmlns='"+Consts.NS_WDSERVLET+"'><href xmlns='DAV:'>"+wrongLocation+"</href></expected-version-URI></checkin>");
          request.expectsResponseDocument();
          response = requester.perform(request);
          checkResponseCode(response, 201, 403);
          if (response.getStatus() == 201) {
            String loc = response.getHeader("Location");
            assertTrue ("Location header not set", loc != null);
            assertTrue ("wrong expected checkin Uri was accepted", loc.equals(wrongLocation));
          }
          else if (response.getStatus() == 403) {
            Document resp = response.getDocument();
            assertTrue ("expects response doc with 403 answer", resp != null);
            Name nm = new Name (resp.getDocumentElement());
            assertTrue ("document element must be DAV:error, was: " + nm,
              nm.equals(Consts.N_error));
            java.util.List exp = DomElement.getChildren(resp.getDocumentElement(),
              Consts.NS_WDSERVLET, "can-assign-expected-version-URI");
            assertTrue ("response body must contain one can-assign-expected-version-URI element",
              exp.size() == 1);
            exp = DomElement.getChildren(resp.getDocumentElement(),
              Consts.NS_WDSERVLET, "expected-version-URI");
            assertTrue ("response body must contain zero or one expected-version-URI elements",
              exp.size() <= 1);
            if (exp.size() == 1) {
              Set hrefs = DomElement.getHrefs((Element) exp.get(0));
              assertTrue(hrefs.size()==1);
              assertTrue ("expectedUri empty", ((String) hrefs.iterator().next()).length() != 0);
            }
          }
          break;

        case 18:
          if (expectedUri != null) {
            request.setMethod("CHECKIN");
            request.setURI(getLocalpart() + "/versioning/resource.txt");
            request.setBody("<checkin xmlns='DAV:'><expected-version-URI xmlns='"+Consts.NS_WDSERVLET+"'>"+
                            expectedUri + "</expected-version-URI></checkin>");
            request.expectsResponseDocument();
            response = requester.perform(request);
            checkResponseCode(response, 201);
          }
          else {
            request.setMethod("CHECKIN");
            request.setURI(getLocalpart() + "/versioning/resource.txt");
            response = requester.perform(request);
            checkResponseCode(response, 201);
          }
          break;

        case 19:
          request.setMethod("CHECKOUT");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          request.setBody("<checkout xmlns='DAV:'><compute-expected-version-URI xmlns='"+Consts.NS_WDSERVLET+"'/></checkout>");
          request.expectsResponseDocument();
          response = requester.perform(request);
          checkResponseCode(response, 200);
          if (response.getDocument() != null) {
            Document resp = response.getDocument();
            Name nm = new Name (resp.getDocumentElement());
            assertTrue ("document element must be DAV:checkout-response, was: " + nm,
              nm.equals(Consts.N_checkout_response));
            java.util.List exp = DomElement.getChildren(resp.getDocumentElement(),
              Consts.NS_WDSERVLET, "expected-version-URI");
            assertTrue ("response body must contain zero or one expected-version-URI element",
              exp.size() <= 1);
            if (exp.size() == 1) {
              expectedUri = DomElement.getText((Element) exp.get(0));
              assertTrue ("expectedUri empty", expectedUri.length() != 0);
            }
          }
          break;

        case 20:
          request.setMethod("PROPFIND");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          request.setBody("<propfind xmlns='DAV:'><prop><checked-out/></prop></propfind>");
          request.expectsResponseDocument();
          request.setHeader("Depth", "0");
          response = requester.perform(request);
          checkResponseCode(response, 207);
          href = (Element) DP_co_href.matchOrDie (response.getDocument(), 1, 1, "checked-out/href not found").get(0);
          checkedOutUri = DomElement.getText(href);
          break;

        case 21:
          request.setMethod("DELETE");
          request.setURI(checkedOutUri);
          response = requester.perform(request);
          checkResponseCode(response, 403, 409);
          break;

        case 22:
          request.setMethod("REPORT");
          request.setURI(checkedOutUri);
          checkMultistatus("<version-tree xmlns='DAV:'><prop><comment/><version-name/><predecessor-set/><successor-set/></prop></version-tree>",
            respfile);
          break;

        case 23:
          clear(request);
          request.setMethod("CHECKIN");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("DELETE");
          request.setURI(checkedOutUri);
          response = requester.perform(request);
          checkResponseCode(response, 204);

          clear(request);
          request.setMethod("REPORT");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          checkMultistatus("<version-tree xmlns='DAV:'><prop><comment/><version-name/><predecessor-set/><successor-set/></prop></version-tree>",
            respfile);
          break;

        case 24:
          request.setMethod("REPORT");
          request.setURI(getLocalpart() + "/ANSI.TXT");
          request.setBody( "<version-tree xmlns='DAV:'><comment/><prop><version-name/><predecessor-set/><successor-set/></prop></version-tree>");
          response = requester.perform(request);
          checkStatusAndErrorElement(response, 409, DavConditions.N_supported_report);
          break;

        case 25:
          request.setMethod("REPORT");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          checkStatusAndBody(207, "<expand-property xmlns='DAV:'>" +
            "<property name='checked-in'>" +
              "<property name='successor-set'/>" +
              "<property name='predecessor-set'>" +
                "<property name='comment'/><property name='predecessor-set'/><property name='successor-set'/>" +
              "</property>" +
              "<property name='comment'/>" +
            "</property></expand-property>",
            respfile);
          break;

        case 26:
          request.setMethod("REPORT");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          request.setBody("<gibt-es-nicht/>");
          response = requester.perform(request);
          checkStatusAndErrorElement(response, 403, DavConditions.N_supported_report);
          break;

        case 27:
          IHierarchicalUri hur = UriFactory.parseHierarchicalUri(someVersionUri);

          request.setMethod("PUT");
          request.setURI(hur.getPath());
          request.setBody("should fail");
          response = requester.perform(request);
          checkStatusAndErrorElement(response, 403, DavConditions.N_cannot_modify_version);

          clear(request);
          request.setMethod("PROPPATCH");
          request.setURI(hur.getPath());
          request.setBody("<propertyupdate xmlns='DAV:'><set><prop><foobar xmlns=''>dead</foobar></prop></set></propertyupdate>");
          response = requester.perform(request);
          checkStatusAndErrorElement(response, 403, DavConditions.N_cannot_modify_version);
          break;

        case 28:
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/versioning/resource2.txt");
          request.setBody("should fail");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("COPY");
          request.setHeader("Overwrite", "T");
          request.setHeader("Destination", getBaseAndLocalpart() + "/versioning/resource.txt");
          request.setURI(getLocalpart() + "/versioning/resource2.txt");
          response = requester.perform(request);
          checkResponseCode(response, 204);

          clear(request);
          request.setMethod("REPORT");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          checkMultistatus("<version-tree xmlns='DAV:'><prop><comment/><version-name/><predecessor-set/><successor-set/></prop></version-tree>",
            respfile);
          break;

        case 29:
          request.setMethod("COPY");
          request.setHeader("Overwrite", "F");
          request.setHeader("Destination", getBaseAndLocalpart() + "/versioning/resource.txt");
          request.setURI(getLocalpart() + "/versioning/resource2.txt");
          response = requester.perform(request);
          checkResponseCode(response, 412);

          clear(request);
          request.setMethod("REPORT");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          checkMultistatus("<version-tree xmlns='DAV:'><prop><comment/><version-name/><predecessor-set/><successor-set/></prop></version-tree>",
            respfile);
          break;

        case 30:
          setUpTestCollection(getLocalpart() + "/versioning/", "col-1", "res-a", "res-b");
          setUpTestCollection(getLocalpart() + "/versioning/", "col-2", "res-a", "res-c");

          clear(request);
          request.setMethod("COPY");
          request.setHeader("Overwrite", "F");
          request.setHeader("Destination", getBaseAndLocalpart() + "/versioning/col-2");
          request.setURI(getLocalpart() + "/versioning/col-1");
          response = requester.perform(request);
          checkResponseCode(response, 412);

          clear(request);
          request.setMethod("PROPFIND");
          request.setURI(getLocalpart() + "/versioning/col-2");
          request.setHeader("Depth", "1");
          request.setBody("<propfind xmlns='DAV:'><prop><checked-in/></prop></propfind>");
          response = requester.perform(request);
          checkResponseCode(response, 207);
          DP_checked_in.matchOrDie(response.getDocument(), 2, 2, "must have exactly two checked-in members");

          removeTestCollection(getLocalpart() + "/versioning/", "col-1", "res-a", "res-b");
          removeTestCollection(getLocalpart() + "/versioning/", "col-2", "res-a", "res-c");
          break;

        case 31:
          setUpTestCollection(getLocalpart() + "/versioning/", "col-1", "res-a", "res-b");
          setUpTestCollection(getLocalpart() + "/versioning/", "col-2", "res-a", "res-c");

          String resbversion = getRelativeUriFromPropertyHrefs(
            getProperty(getLocalpart() + "/versioning/col-2/res-a", Consts.N_checked_in));

          clear(request);
          request.setMethod("COPY");
          request.setHeader("Overwrite", "T");
          request.setHeader("Destination", getBaseAndLocalpart() + "/versioning/col-2");
          request.setURI(getLocalpart() + "/versioning/col-1");
          response = requester.perform(request);
          checkResponseCode(response, 204);

          // expected post conditions
          // col2/res-a exists, is versioned and got a new version
          // col2/res-c disappears
          // col2/res-b was newly created and is not versioned

          clear(request);
          request.setMethod("PROPFIND");
          request.setURI(getLocalpart() + "/versioning/col-2");
          request.setHeader("Depth", "1");
          request.setBody("<propfind xmlns='DAV:'><prop><checked-in/></prop></propfind>");
          response = requester.perform(request);
          checkResponseCode(response, 207);

          try {
            DP_checked_in.matchOrDie(response.getDocument(), 1, 1, "must have exactly one checked-in member");
            assertTrue("collection must have two members",
              DomElement.getChildren(response.getDocument().getDocumentElement(), Consts.N_response).size() == 3);
          }
          catch (DomPathAssertionFailedException e) {
            assertTrue("assertion failed ("+e.getMessage()
                       +") in " + DomSerializer.toString(response.getDocument(), true), false);
          }

          String resbversionb = getRelativeUriFromPropertyHrefs(
            getProperty(getLocalpart() + "/versioning/col-2/res-a", Consts.N_checked_in));

          assertTrue("check-in URI for res-a did not change", !resbversion.equals(resbversionb));

          removeTestCollection(getLocalpart() + "/versioning/", "col-1", "res-a", "res-b");
          removeTestCollection(getLocalpart() + "/versioning/", "col-2", "res-a", "res-c");
          break;

        case 32:
          setUpTestCollection(getLocalpart() + "/versioning/", "labels", "res-a", "res-b");

          clear(request);
          request.setMethod("OPTIONS");
          request.setURI(getLocalpart() + "/versioning/labels/res-a");
          response = requester.perform(request);
          checkResponseCode(response, 200);
          {
            String t = response.getHeader("DAV");
            assertTrue ("label was reported as feature (shouldn't)", t != null && t.indexOf("label") < 0);
            t = response.getHeader("Allow");
            hasLabelMethod = t.indexOf("LABEL") >= 0;
            super.logComment(hasLabelMethod ? "label:yes" : "label:no");
          }

          break;

        case 33:
          if (hasLabelMethod) {
            clear(request);
            request.setMethod("PROPFIND");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/versioning/labels/res-a");
            request.setBody("<propfind xmlns='DAV:'><prop><supported-method-set/></prop></propfind>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            Document doc = response.getDocument();
            DP_s_m_s_label.matchOrDie(doc, 1, 1, "LABEL method not reported as supported-method");
          }
          break;

        case 34:
          if (hasLabelMethod) {
            clear(request);
            request.setMethod("LABEL");
            request.setURI(getLocalpart() + "/Worddatei.doc");
            request.setBody("<label xmlns='DAV:'><add><label-name>foobar</label-name></add></label>");
            response = requester.perform(request);
            checkResponseCode(response, 409);
          }
          break;

        case 35:
          if (hasLabelMethod) {
            clear(request);
            request.setMethod("LABEL");
            request.setURI(getLocalpart() + "/versioning/labels/res-a");
            request.setBody("<label xmlns='DAV:'><add><label-name>foobar</label-name></add></label>");
            response = requester.perform(request);
            checkResponseCode(response, 200);
          }
          break;

        case 36:
          if (hasLabelMethod) {
            clear(request);
            request.setMethod("LABEL");
            request.setURI(getLocalpart() + "/versioning/labels/res-a");
            request.setBody("<label xmlns='DAV:'><add><label-name>foobar</label-name></add></label>");
            response = requester.perform(request);
            checkStatusAndErrorElement(response, 409, DavConditions.N_add_must_be_new_label);
          }
          break;

        case 37:
          if (hasLabelMethod) {
            clear(request);
            request.setMethod("LABEL");
            request.setURI(getLocalpart() + "/versioning/labels");
            request.setHeader("Depth", "1");
            request.setBody("<label xmlns='DAV:'><add><label-name>foobar</label-name></add></label>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            checkResponseDocument(response, respfile, null);
          }
          break;

        case 38:
          if (hasLabelMethod) {
            clear(request);
            request.setMethod("LABEL");
            request.setURI(getLocalpart() + "/versioning/labels");
            request.setHeader("Depth", "1");
            request.setBody("<label xmlns='DAV:'><add><label-name>from-collection</label-name></add></label>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            checkResponseDocument(response, respfile, null);
          }
          break;

        case 39:
          if (hasLabelMethod) {
            clear(request);
            request.setMethod("CHECKOUT");
            request.setURI(getLocalpart() + "/versioning/labels/res-a");
            response = requester.perform(request);
            checkResponseCode(response, 200);
          }
          break;

        case 40:
          if (hasLabelMethod) {
            clear(request);
            request.setMethod("LABEL");
            request.setURI(getLocalpart() + "/versioning/labels/res-a");
            request.setBody("<label xmlns='DAV:'><add><label-name>from-collecsasasastion</label-name></add></label>");
            response = requester.perform(request);
            checkStatusAndErrorElement(response, 409, DavConditions.N_must_be_checked_in);
          }
          break;

        case 41:
          if (hasLabelMethod) {
            clear(request);
            request.setMethod("CHECKIN");
            request.setURI(getLocalpart() + "/versioning/labels/res-a");
            response = requester.perform(request);
            checkResponseCode(response, 201);
          }
          break;

        case 42:
          if (hasLabelMethod) {
            clear(request);
            request.setMethod("LABEL");
            request.setURI(getLocalpart() + "/versioning/labels");
            request.setHeader("Depth", "1");
            request.setBody("<label xmlns='DAV:'><set><label-name>from-collection</label-name></set></label>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            checkResponseDocument(response, respfile, null);
          }
          break;

        case 43:
          if (hasLabelMethod) {
            clear(request);
            request.setMethod("LABEL");
            request.setURI(getLocalpart() + "/versioning/labels");
            request.setHeader("Depth", "1");
            request.setBody("<label xmlns='DAV:'><remove><label-name>foobar</label-name></remove></label>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            checkResponseDocument(response, respfile, null);
          }
          break;

        case 44:
          if (hasLabelMethod) {
            DomPath hrefs = new DomPath("//D:href", DomPath.DEFAULT_NAMESPACE_BINDINGS);

            clear(request);
            request.setMethod("REPORT");
            request.setURI(getLocalpart() + "/versioning/labels");
            request.setHeader("Depth", "1");
            request.setBody("<expand-property xmlns='DAV:'><property name='version-history'><property name='version-set'><property name='label-name-set'/></property></property></expand-property>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            checkResponseDocument(response, respfile, hrefs);


            clear(request);
            request.setMethod("REPORT");
            request.setURI(getLocalpart() + "/versioning/labels");
            request.setHeader("Depth", "infinity");
            request.setBody("<expand-property xmlns='DAV:'><property name='version-history'><property name='version-set'><property name='label-name-set'/></property></property></expand-property>");
            response = requester.perform(request);
            checkResponseCode(response, 403);
            checkErrorElement(response, DavConditions.N_SAP_depth_infinity_allowed);
          }
          break;

        case 45:
          {
            Element checkedout = getProperty(getLocalpart() + "/versioning/resource.txt", Consts.N_checked_out);
            if (checkedout == null) {
              clear(request);
              request.setMethod("CHECKOUT");
              request.setURI(getLocalpart() + "/versioning/resource.txt");
              response = requester.perform(request);
              checkResponseCode(response, 200, 201);
            }

            clear(request);
            request.setMethod("CHECKIN");
            request.setURI(getLocalpart() + "/versioning/resource.txt");
            request.setBody("<checkin xmlns='DAV:'><keep-checked-out/></checkin>");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            assertTrue(response.getHeader("Location") != null);
            someVersionUri = response.getHeader("Location");
            assertTrue(getProperty(getLocalpart() + "/versioning/resource.txt", Consts.N_checked_in) == null);
          }
          break;

        case 46:
          if (hasLabelMethod) {
            DomPath hrefs = new DomPath("//D:href", DomPath.DEFAULT_NAMESPACE_BINDINGS);

            clear(request);
            request.setMethod("REPORT");
            request.setURI(getLocalpart() + "/versioning/labels");
            request.setHeader("Depth", "1");
            request.setBody("<labeled-version xmlns='DAV:'><label-name>from-collection</label-name><prop><label-name-set/></prop></labeled-version>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            checkResponseDocument(response, respfile, hrefs);
          }
          break;

        case 47:
          clear(request);
          request.setMethod("COPY");
          request.setURI(removeProtocolAndBase(someVersionUri));
          request.setHeader("Destination", this.getBaseAndLocalpart() + "/versioning/thiswasaversion");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          break;

        case 48: {
            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + "/versioning/autovers");
            response = requester.perform(request);
            checkResponseCode(response, 201);
            
            clear(request);
            request.setMethod("VERSION-CONTROL");
            request.setURI(getLocalpart() + "/versioning/autovers");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            // remember VHR URI for later
            vhr = getProperty(getLocalpart() + "/versioning/autovers", Consts.N_version_history);
            vhrs.add(getRelativeUriFromPropertyHrefs(vhr));

            // get autoversioning property
            Element autovers = getProperty(getLocalpart() + "/versioning/autovers", Consts.N_auto_version);
            boolean isAuto = DomElement.hasElementChildrenInDavNs(autovers);
            
            if (isAuto) {
              Name pname = new Name("foo:", "bar");
              setProperty(getLocalpart() + "/versioning/autovers", pname, "testme", false, 200);
              
              // check round-trip
              Element deadprop = getProperty(getLocalpart() + "/versioning/autovers", pname);
              assertTrue("dead propery not persisted", "testme".equals(DomElement.getText(deadprop)));
            }
          }
          break;

        case 49:

          if (hasLabelMethod) {

            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + "/versioning/blanks");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("VERSION-CONTROL");
            request.setURI(getLocalpart() + "/versioning/blanks");
            response = requester.perform(request);
            checkResponseCode(response, 201);
            String firstVersion = getRelativeUriFromLocationHeader(response);

            // remember VHR URI for later
            vhr = getProperty(getLocalpart() + "/versioning/blanks", Consts.N_version_history);
            vhrs.add(getRelativeUriFromPropertyHrefs(vhr));

            clear(request);
            request.setMethod("CHECKOUT");
            request.setURI(getLocalpart() + "/versioning/blanks");
            response = requester.perform(request);
            checkResponseCode(response, 200, 201);

            clear(request);
            request.setMethod("CHECKIN");
            request.setURI(getLocalpart() + "/versioning/blanks");
            request.setBody("<checkin xmlns='DAV:'><keep-checked-out/></checkin>");
            response = requester.perform(request);
            checkResponseCode(response, 201);
            String secondVersion = getRelativeUriFromLocationHeader(response);

            clear(request);
            request.setMethod("CHECKIN");
            request.setURI(getLocalpart() + "/versioning/blanks");
            response = requester.perform(request);
            checkResponseCode(response, 201);
            String thirdVersion = getRelativeUriFromLocationHeader(response);

            clear(request);
            request.setMethod("LABEL");
            request.setURI(firstVersion);
            request.setBody("<label xmlns='DAV:'><set><label-name>testlabel</label-name></set></label>");
            response = requester.perform(request);
            checkResponseCode(response, 200);

            clear(request);
            request.setMethod("LABEL");
            request.setURI(secondVersion);
            request.setBody("<label xmlns='DAV:'><add><label-name>testlabel&#x0020;</label-name></add></label>");
            response = requester.perform(request);
            checkResponseCode(response, 200);

            clear(request);
            request.setMethod("LABEL");
            request.setURI(thirdVersion);
            request.setBody("<label xmlns='DAV:'><add><label-name>&#x0020;testlabel</label-name></add></label>");
            response = requester.perform(request);
            checkResponseCode(response, 200);
          }
          break;

        case 50:

          if (hasLabelMethod) {

            clear(request);
            request.setMethod("LABEL");
            request.setURI(getLocalpart() + "/versioning/blanks");
            request.setBody("<label xmlns='DAV:'><set><label-name>testlabel2</label-name></set></label>");
            response = requester.perform(request);
            checkResponseCode(response, 200);

            clear(request);
            request.setMethod("LABEL");
            request.setURI(getLocalpart() + "/versioning/blanks");
            request.setBody("<label xmlns='DAV:'><add><label-name>testlabel2&#x0020;</label-name></add></label>");
            response = requester.perform(request);
            checkResponseCode(response, 200);

            clear(request);
            request.setMethod("LABEL");
            request.setURI(getLocalpart() + "/versioning/blanks");
            request.setBody("<label xmlns='DAV:'><add><label-name>&#x0020;testlabel2</label-name></add></label>");
            response = requester.perform(request);
            checkResponseCode(response, 200);
          }
          break;

        case 51:
          if (hasLabelMethod) {
            clear(request);
            request.setMethod("REPORT");
            request.expectsResponseDocument(true);
            request.setURI(getLocalpart() + "/versioning/blanks");
            request.setBody("<labeled-version xmlns='DAV:'><label-name>testlabel2</label-name><prop><label-name-set/></prop></labeled-version>");
            response = requester.perform(request);
            checkResponseCode(response, 200);

            DomPath DP_label_name = new DomPath("//D:" + Consts.EN_label_name, DomPath.DEFAULT_NAMESPACE_BINDINGS);
            List labelList = DP_label_name.match(response.getDocument().getDocumentElement());
            assertTrue("wrong number of labels found on version: " + labelList.size() + ", should be: 4",
                labelList.size() == 4);
          }
          break;

        // no, there is no missing break, it is really meant this way ...
        case 52:
        case 53:
          
          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/versioning/deltest");
          response = requester.perform(request);
          checkResponseCode(response, 204, 404);
          
          String[] versions = new String[4];
          
          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/versioning/deltest");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("VERSION-CONTROL");
          request.setURI(getLocalpart() + "/versioning/deltest");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          versions[0] = getRelativeUriFromLocationHeader(response);
          
          // remember VHR URI for later
          vhr = getProperty(getLocalpart() + "/versioning/deltest", Consts.N_version_history);
          vhrs.add(getRelativeUriFromPropertyHrefs(vhr));

          for (int i=1; i<4; i++) {
            clear(request);
            request.setMethod("CHECKOUT");
            request.setURI(getLocalpart() + "/versioning/deltest");
            response = requester.perform(request);
            checkResponseCode(response, 200, 201);
            
            clear(request);
            request.setMethod("CHECKIN");
            request.setURI(getLocalpart() + "/versioning/deltest");          
            response = requester.perform(request);
            checkResponseCode(response, 201);
            versions[i] = getRelativeUriFromLocationHeader(response);
          }

          for (int i=0; i<3; i++) {
            clear(request);
            request.setMethod("DELETE");            
            request.setURI(nr==53 ? versions[i] : versions[2-i]);
            response = requester.perform(request);
            checkResponseCode(response, 204); 
          }          
          break;                 
          
        case 54:

          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/versioning/v52.txt");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("VERSION-CONTROL");
          request.setURI(getLocalpart() + "/versioning/v52.txt");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          String firstVersion = getRelativeUriFromLocationHeader(response);

          // remember VHR URI for later
          vhr = getProperty(getLocalpart() + "/versioning/v52.txt", Consts.N_version_history);
          vhrs.add(getRelativeUriFromPropertyHrefs(vhr));

          clear(request);
          request.setMethod("CHECKOUT");
          request.setURI(getLocalpart() + "/versioning/v52.txt");
          response = requester.perform(request);
          checkResponseCode(response, 200, 201);

          clear(request);
          request.setMethod("CHECKIN");
          request.setURI(getLocalpart() + "/versioning/v52.txt");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          String secondVersion = getRelativeUriFromLocationHeader(response);

          clear(request);
          request.setMethod("CHECKOUT");
          request.setURI(getLocalpart() + "/versioning/v52.txt");
          response = requester.perform(request);
          checkResponseCode(response, 200, 201);

          clear(request);
          request.setMethod("CHECKIN");
          request.setURI(getLocalpart() + "/versioning/v52.txt");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          String thirdVersion = getRelativeUriFromLocationHeader(response);

          clear(request);
          request.setMethod("DELETE");
          request.setURI(secondVersion);
          response = requester.perform(request);
          checkResponseCode(response, 204);

          clear(request);
          request.setMethod("DELETE");
          request.setURI(firstVersion);
          response = requester.perform(request);
          checkResponseCode(response, 204);

          break;
          
        default:
          request.setMethod("CHECKIN");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          response = requester.perform(request);
          checkResponseCode(response, 201, 409);

          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/versioning/resource.txt");
          response = requester.perform(request);
          checkResponseCode(response, 204);

          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/versioning");
          response = requester.perform(request);
          checkResponseCode(response, 204);

          Iterator vhrIter = vhrs.iterator();
          while (vhrIter.hasNext()) {
            String vhrUri = (String) vhrIter.next();
            clear(request);
            request.setMethod("DELETE");
            request.setURI(vhrUri);
            response = requester.perform(request);
            checkResponseCode(response, vhrPreserved() ? 204 : 404);
          }

          break;
      }
    }
    catch (Exception ex) {
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }

  private void setUpTestCollection(String parent, String collname, String res1, String res2)
    throws com.sapportals.wcm.WcmException {

    IResponse response;

    clear(request);
    request.setMethod("MKCOL");
    request.setURI(parent + collname);
    response = requester.perform(request);
    checkResponseCode(response, 201);

    clear(request);
    request.setMethod("PROPPATCH");
    request.setURI(parent + collname);
    request.setBody("<propertyupdate xmlns='DAV:'><set><prop><comment>" +
      collname + "'s dead property" +
      "</comment></prop></set></propertyupdate>");
    response = requester.perform(request);
    checkResponseCode(response, 207);

    clear(request);
    request.setMethod("PUT");
    request.setURI(parent + collname + "/" + res1);
    request.setBody(parent + collname + "/" + res1);
    response = requester.perform(request);
    checkResponseCode(response, 201);

    clear(request);
    request.setMethod("VERSION-CONTROL");
    request.setURI(parent + collname + "/" + res1);
    response = requester.perform(request);
    checkResponseCode(response, 200, 201);

    Element vhr = getProperty(parent + collname + "/" + res1, Consts.N_version_history);
    if (vhr != null) {
      vhrs.add(getRelativeUriFromPropertyHrefs(vhr));
    }

    clear(request);
    request.setMethod("PUT");
    request.setURI(parent + collname + "/" + res2);
    request.setBody(parent + collname + "/" + res2);
    response = requester.perform(request);
    checkResponseCode(response, 201);

    clear(request);
    request.setMethod("VERSION-CONTROL");
    request.setURI(parent + collname + "/" + res2);
    response = requester.perform(request);
    checkResponseCode(response, 200, 201);

    vhr = getProperty(parent + collname + "/" + res2, Consts.N_version_history);
    if (vhr != null) {
      vhrs.add(getRelativeUriFromPropertyHrefs(vhr));
    }

  }

  private void removeTestCollection(String parent, String collname, String res1, String res2)
    throws com.sapportals.wcm.WcmException {

    clear(request);
    request.setMethod("CHECKIN");
    request.setURI(parent + collname + "/" + res1);
    // response =
    requester.perform(request);

    clear(request);
    request.setMethod("CHECKIN");
    request.setURI(parent + collname + "/" + res2);
    // response =
    requester.perform(request);

    clear(request);
    request.setMethod("DELETE");
    request.setURI(parent + collname);
    // response =
    requester.perform(request);
    // checkResponseCode(response, 204);
  }

  private String getContent(InputStream is) throws IOException {
    if (is == null)
      return null;

    byte[] buffer = new byte[64 * 1024];
    ByteArrayOutputStream bos = new ByteArrayOutputStream();
    int len;
    while ((len = is.read(buffer)) != -1) {
      bos.write(buffer, 0, len);
    }

    return bos.toString();
  }
}
