package com.sapportals.wcm.protocol.webdav.server.tests;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.w3c.dom.Element;
import org.w3c.dom.Node;

import com.sapportals.wcm.protocol.webdav.*;
import com.sapportals.wcm.protocol.webdav.dompath.DomPath;
import com.sapportals.wcm.util.http.IResponse;



/**
 * Various tests for Versioned Collections.
 *
 * <p>
 * Copyright (c) SAP AG 2002-2005
 * @author julian.reschke@greenbytes.de
 * @version $Id: VersionedCollectionTest.java,v 1.12 2005/09/20 10:48:35 mko Exp $
 */
public class VersionedCollectionTest extends TestMember
{
  private static final String[] desc = new String[]
  {
    "00 VC Setting up Versioned Collections",
    "01 VC get auto versioning property and remember it",
    "02 VC (non-auto) VC / PP on collection member of checked-in coll",
    "03 VC PUT into checked-out collection",
    "04 VC (non-auto) UN-VERSION-CONTROL inside checked-in collection",
    "05 VC UN-VERSION-CONTROL inside checked-out collection",
    "06 VC UNCHECKOUT",
    "07 VC CHECKOUT, add VCR, CHECKIN, CHECKOUT, delete VCR, create nVCR, UNCHECKOUT, DELETE/HEAD",
    "08 VC CHECKOUT, move VCR away, UNCHECKOUT",
    "09 VC CHECKOUT, delete VCR, UNCHECKOUT",
    "10 VC CHECKOUT, rename VCR, UNCHECKOUT",
    "11 VC UPDATE from collection version while member is checked out",
    "12 VC delete root version",
    "13 VC GET on parent of VC's VHR",
    "   VC cleanup",
  };

  public static final int NrOfTests = desc.length;

  public VersionedCollectionTest(Integer nr)
  {
    super(nr, desc[nr.intValue()]);
  }

  private static boolean hasAutoVersionedCollections = false;
  private static String secondVersionOfCollection;
  public static Set vhrs = new HashSet();

  private static DomPath hrefs = new DomPath("//D:href", DomPath.DEFAULT_NAMESPACE_BINDINGS);

  protected void runTest()
  {
    if (nr == 0) logIdentifier("VC");

    String respfile = "VERSIONED-COLL-" + TestMember.TWODIGITS.format(nr) + ".txt";

    try
    {
      IResponse response;
      String content;

      clear(request);

      switch (nr)
      {
        case 0:
          request.setMethod("MKCOL");
          request.setURI(getLocalpart() + "/vc");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          request.setMethod("OPTIONS");
          request.setURI(getLocalpart() + "/vc");
          response = requester.perform(request);
          checkResponseCode(response, 200);
          assertTrue(response.getHeader("DAV").indexOf("version-controlled-collection") >= 0);

          setProperty(getLocalpart() + "/vc", Consts.N_comment, "Test Case 0", false, DavStatus.SC_OK);

          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/vc/testresource");
          request.setBody("test");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("VERSION-CONTROL");
          request.setURI(getLocalpart() + "/vc");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          
          // remember VHR URI for later
          Element vhr = getProperty(getLocalpart() + "/vc", Consts.N_version_history);
          vhrs.add(getRelativeUriFromPropertyHrefs(vhr));
          break;

        case 1:
          Element av = getProperty(getLocalpart() + "/vc", Consts.N_auto_version);
          hasAutoVersionedCollections = DomElement.hasElementChildrenInDavNs(av);
          this.logComment("auto-vers.:" + (hasAutoVersionedCollections ? "yes" : "no"));
          break;

        case 2:
          if (!hasAutoVersionedCollections) {
            // should fail because parent isn't checked out
            clear(request);
            request.setMethod("VERSION-CONTROL");
            request.setURI(getLocalpart() + "/vc/testresource");
            response = requester.perform(request);
            checkStatusAndErrorElement(response, 409, DavConditions.N_cannot_modify_checked_in_parent);

            clear(request);
            request.setMethod("PROPPATCH");
            request.setURI(getLocalpart() + "/vc");
            request.setBody("<propertyupdate xmlns='DAV:'><set><prop><comment>Test Case 2</comment></prop></set></propertyupdate>");
            response = requester.perform(request);
            checkStatusAndErrorElement(response, 409, DavConditions.N_cannot_modify_version_controlled_property);
          }
          break;

        case 3:
          // same as 2 but checkout parent collection first
          clear(request);
          request.setMethod("CHECKOUT");
          request.setURI(getLocalpart() + "/vc");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          setProperty(getLocalpart() + "/vc", Consts.N_comment, "Test Case 3", false, DavStatus.SC_OK);

          clear(request);
          request.setMethod("VERSION-CONTROL");
          request.setURI(getLocalpart() + "/vc/testresource");
          response = requester.perform(request);
          checkResponseCode(response, 200, 201);

          clear(request);
          request.setMethod("CHECKIN");
          request.setURI(getLocalpart() + "/vc");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          secondVersionOfCollection = getRelativeUriFromLocationHeader(response);

          clear(request);
          request.setMethod("REPORT");
          request.setURI(getLocalpart() + "/vc");
          request.setHeader("Depth", "0");
          checkMultistatus("<expand-property xmlns='DAV:'><property name='comment'/><property name='checked-in'><property name='version-controlled-binding-set'/></property></expand-property>",
            respfile, hrefs);

          break;

        case 4:
          Element vhrn = getProperty(getLocalpart() + "/vc/testresource", Consts.N_version_history);
          String vhruri = getRelativeUriFromPropertyHrefs(vhrn);

//          clear(request);
//          request.setMethod("PROPFIND");
//          request.setHeader("Depth", "0");
//          request.setBody("<propfind xmlns='DAV:'><prop><version-history/></prop></propfind>");
//          request.setURI(getLocalpart() + "/vc/testresource");
//          response = requester.perform(request);
//          checkResponseCode(response, 207);
//          Node vhrn = DP_vhr.matchFirst(response.getDocument());
//          assertTrue(vhrn != null);
//          String vhruri = DomElement.getText(vhrn);

          if (!hasAutoVersionedCollections) {
            // attempt to delete version history
            clear(request);
            request.setMethod("DELETE");
            request.setURI(vhruri);
            response = requester.perform(request);
            checkResponseCode(response, 409);
//            checkStatusAndErrorElement(response, 409, DavConditions.N_cannot_modify_checked_in_parent);
          }
          break;

        case 5:
            // same as 4 with checked-out parent

          clear(request);
          request.setMethod("CHECKOUT");
          request.setURI(getLocalpart() + "/vc");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          vhrn = getProperty(getLocalpart() + "/vc/testresource", Consts.N_version_history);
          vhruri = getRelativeUriFromPropertyHrefs(vhrn);

          // attempt to delete version history
          clear(request);
          request.setMethod("DELETE");
          request.setURI(vhruri);
          response = requester.perform(request);
          checkResponseCode(response, 204);

          clear(request);
          request.setMethod("CHECKIN");
          request.setURI(getLocalpart() + "/vc");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("REPORT");
          request.setURI(getLocalpart() + "/vc");
          request.setHeader("Depth", "0");
          checkMultistatus("<expand-property xmlns='DAV:'><property name='comment'/><property name='checked-in'><property name='version-controlled-binding-set'/></property></expand-property>",
            respfile, hrefs);

          break;

        case 6:
          clear(request);
          request.setMethod("CHECKOUT");
          request.setURI(getLocalpart() + "/vc");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/vc/testresource2");
          request.setBody("test");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          setProperty(getLocalpart() + "/vc", Consts.N_comment, "Test Case 6a", false, 200);

          clear(request);
          request.setMethod("VERSION-CONTROL");
          request.setURI(getLocalpart() + "/vc/testresource2");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          vhr = getProperty(getLocalpart() + "/vc/testresource2", Consts.N_version_history);
          vhrs.add(getRelativeUriFromPropertyHrefs(vhr));

          clear(request);
          request.setMethod("CHECKIN");
          request.setURI(getLocalpart() + "/vc");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("REPORT");
          request.setURI(getLocalpart() + "/vc");
          request.setHeader("Depth", "0");
          checkMultistatus("<expand-property xmlns='DAV:'><property name='comment'/><property name='checked-in'><property name='version-controlled-binding-set'/></property></expand-property>",
            respfile, hrefs);

          clear(request);
          request.setMethod("CHECKOUT");
          request.setURI(getLocalpart() + "/vc");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          setProperty(getLocalpart() + "/vc", Consts.N_comment, "Test Case 6b", false, 200);

          clear(request);
          request.setMethod("VERSION-CONTROL");
          request.setURI(getLocalpart() + "/vc/testresource");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          vhr = getProperty(getLocalpart() + "/vc/testresource", Consts.N_version_history);
          vhrs.add(getRelativeUriFromPropertyHrefs(vhr));

          clear(request);
          request.setMethod("UNCHECKOUT");
          request.setURI(getLocalpart() + "/vc");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          clear(request);
          request.setMethod("REPORT");
          request.setURI(getLocalpart() + "/vc");
          request.setHeader("Depth", "0");
          checkMultistatus("<expand-property xmlns='DAV:'><property name='comment'/><property name='checked-in'><property name='version-controlled-binding-set'/></property></expand-property>",
            respfile, hrefs);

          break;

        // "07 VC (non-auto) CHECKOUT, add VCR, CHECKIN, CHECKOUT, delete VCR, create nVCR, UNCHECKOUT",

        case 7:
          clear(request);
          request.setMethod("CHECKOUT");
          request.setURI(getLocalpart() + "/vc");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/vc/testresource3");
          request.setBody("test3");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          setProperty(getLocalpart() + "/vc", Consts.N_comment, "Test Case 7", false, 200);

          clear(request);
          request.setMethod("VERSION-CONTROL");
          request.setURI(getLocalpart() + "/vc/testresource3");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          vhr = getProperty(getLocalpart() + "/vc/testresource3", Consts.N_version_history);
          vhrs.add(getRelativeUriFromPropertyHrefs(vhr));

          clear(request);
          request.setMethod("CHECKIN");
          request.setURI(getLocalpart() + "/vc");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("CHECKOUT");
          request.setURI(getLocalpart() + "/vc");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          setProperty(getLocalpart() + "/vc", Consts.N_comment, "Test Case 7b", false, 200);

          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/vc/testresource3");
          response = requester.perform(request);
          checkResponseCode(response, 204, 501);

          // we allow NOT IMPLEMENTED for rmfs here

          if (response.getStatus() == 204) {
            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + "/vc/testresource3");
            request.setBody("temp");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("UNCHECKOUT");
            request.setURI(getLocalpart() + "/vc");
            response = requester.perform(request);
            checkResponseCode(response, 200);

            clear(request);
            request.setMethod("PROPFIND");
            request.setURI(getLocalpart() + "/vc");
            request.setHeader("Depth", "1");
            request.setBody("<propfind xmlns='DAV:'><allprop/></propfind>");
            response = requester.perform(request);
            checkResponseCode(response, 207);

            DomPath DP_hrefs = new DomPath("//D:" + Consts.EN_href, DomPath.DEFAULT_NAMESPACE_BINDINGS);
            List hrefList = DP_hrefs.match(response.getDocument().getDocumentElement());
            assertTrue("wrong number of bindings in collection " + getLocalpart() + "/vc: "+ (hrefList.size() - 1),
                       hrefList.size() == 3);

            Element eclipsedSet = getProperty(getLocalpart() + "/vc", Consts.N_eclipsed_set);
            DomPath DP_es = new DomPath("D:" + Consts.EN_binding_name, DomPath.DEFAULT_NAMESPACE_BINDINGS);
            List l = DP_es.matchOrDie(eclipsedSet, 1, 1, "need exactly one binding name");
            String ename = DomElement.getText((Node)l.get(0));
            assertTrue("testresource3".equals(ename));

            clear(request);
            request.setMethod("GET");
            request.setURI(getLocalpart() + "/vc/testresource3");
            request.expectsResponseStream(true);
            response = requester.perform(request);
            checkResponseCode(response, 200);
            content = this.getContent(response.getStream());
            assertTrue ("unexpected content: " + content, content.equals("temp"));

            clear(request);
            request.setMethod("DELETE");
            request.setURI(getLocalpart() + "/vc/testresource3");
            response = requester.perform(request);
            checkResponseCode(response, 204);

            clear(request);
            request.setMethod("PROPFIND");
            request.setHeader("Depth", "0");
            request.setURI(getLocalpart() + "/vc/testresource3");
            response = requester.perform(request);
            checkMultistatus("<propfind xmlns='DAV:'><prop><checked-in/><checked-out/></prop></propfind>",
              respfile, hrefs);

            clear(request);
            request.setMethod("GET");
            request.setURI(getLocalpart() + "/vc/testresource3");
            request.expectsResponseStream(true);
            response = requester.perform(request);
            checkResponseCode(response, 200);
            content = this.getContent(response.getStream());
            assertTrue ("unexpected content: " + content, content.equals("test3"));
          }
          else {
            // cleanup in case we couldn't do the full test
            clear(request);
            request.setMethod("UNCHECKOUT");
            request.setURI(getLocalpart() + "/vc");
            response = requester.perform(request);
            checkResponseCode(response, 200);
          }

          break;

        // "08 VC (non-auto) CHECKOUT, move VCR away, UNCHECKOUT",

        case 8:

          clear(request);
          request.setMethod("CHECKOUT");
          request.setURI(getLocalpart() + "/vc");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          clear(request);
          request.setMethod("MOVE");
          request.setURI(getLocalpart() + "/vc/testresource3");
          request.setHeader("Destination", getLocalpart() + "/testresource3");
          response = requester.perform(request);
          checkResponseCode(response, 201, 409);
          boolean hasMoved = response.getStatus() == 201;

          clear(request);
          request.setMethod("UNCHECKOUT");
          request.setURI(getLocalpart() + "/vc");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          if (hasMoved) {

            Element id1 = getProperty(getLocalpart() + "/vc/testresource3", Consts.N_resource_id);
            Element id2 = getProperty(getLocalpart() + "/testresource3", Consts.N_resource_id);
            assertTrue(!DomSerializer.toString(id1, false).equals(DomSerializer.toString(id2, false)));

            Element v1 = getProperty(getLocalpart() + "/vc/testresource3", Consts.N_checked_in);
            Element v2 = getProperty(getLocalpart() + "/testresource3", Consts.N_checked_in);
            assertTrue(v1 != null);
            assertTrue(v2 != null);
            assertTrue(DomSerializer.toString(v1, false).equals(DomSerializer.toString(v2, false)));
          }

          break;

        //    "09 VC (non-auto) CHECKOUT, delete VCR, UNCHECKOUT",

        case 9:

          clear(request);
          request.setMethod("CHECKOUT");
          request.setURI(getLocalpart() + "/vc");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/vc/testresource3");
          response = requester.perform(request);
          checkResponseCode(response, 204, 501);
          boolean hasDeleted = response.getStatus() == 204;

          clear(request);
          request.setMethod("UNCHECKOUT");
          request.setURI(getLocalpart() + "/vc");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          if (hasDeleted) {
            Element checkedin = getProperty(getLocalpart() + "/vc/testresource3", Consts.N_checked_in);
            assertTrue(checkedin != null);
          }

          break;

        case 10:

          clear(request);
          request.setMethod("CHECKOUT");
          request.setURI(getLocalpart() + "/vc");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          clear(request);
          request.setMethod("MOVE");
          request.setURI(getLocalpart() + "/vc/testresource3");
          request.setHeader("Destination", getLocalpart() + "/vc/testresource3renamed");
          response = requester.perform(request);
          checkResponseCode(response, 201, 409);
          boolean hasRenamed = response.getStatus() == 201;

          clear(request);
          request.setMethod("UNCHECKOUT");
          request.setURI(getLocalpart() + "/vc");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          if (hasRenamed) {
            Element checkedin = getProperty(getLocalpart() + "/vc/testresource3", Consts.N_checked_in);
            assertTrue(checkedin != null);
          }

          break;

        case 11:

          Element versionHistory = getProperty(getLocalpart() + "/vc", Consts.N_version_history);
          String versionHistoryUri = getRelativeUriFromPropertyHrefs(versionHistory);

          Element rootVersion =  getProperty(versionHistoryUri, Consts.N_root_version);
          String rootVersionUri = getRelativeUriFromPropertyHrefs(rootVersion);

          clear(request);
          request.setMethod("UPDATE");
          request.setURI(getLocalpart() + "/vc");
          request.setBody("<update xmlns='DAV:'><version><href>"+
                          rootVersionUri + "</href></version></update>");
          response = requester.perform(request);
          checkResponseCode(response, 207, 501);

          if (response.getStatus() == 207) {
            Element checkedin = getProperty(getLocalpart() + "/vc", Consts.N_checked_in);
            assertTrue(checkedin != null);

            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + "/vc/vcr");
            request.setBody("sdhfzsgdfsa");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("CHECKOUT");
            request.setBody("<checkout xmlns='DAV:'><fork-ok/></checkout>");
            request.setURI(getLocalpart() + "/vc");
            response = requester.perform(request);
            checkResponseCode(response, 200);

            clear(request);
            request.setMethod("VERSION-CONTROL");
            request.setURI(getLocalpart() + "/vc/vcr");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            vhr = getProperty(getLocalpart() + "/vc/vcr", Consts.N_version_history);
            vhrs.add(getRelativeUriFromPropertyHrefs(vhr));

            clear(request);
            request.setMethod("CHECKIN");
            request.setBody("<checkin xmlns='DAV:'><fork-ok/></checkin>");
            request.setURI(getLocalpart() + "/vc");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            Element vcCheckedIn = getProperty(getLocalpart() + "/vc", Consts.N_checked_in);
            String checkedInUri = getRelativeUriFromPropertyHrefs(vcCheckedIn);

            clear(request);
            request.setMethod("CHECKOUT");
            request.setURI(getLocalpart() + "/vc/vcr");
            response = requester.perform(request);
            checkResponseCode(response, 200);

            clear(request);
            request.setMethod("UPDATE");
            request.setURI(getLocalpart() + "/vc");
            request.setBody("<update xmlns='DAV:'><version><href>"
                + checkedInUri
                + "</href></version></update>");
            response = requester.perform(request);
            checkResponseCode(response, 409);
            checkErrorElement(response, DavConditions.N_must_be_checked_in);

            clear(request);
            request.setMethod("UPDATE");
            request.setURI(getLocalpart() + "/vc");
            request.setBody("<update xmlns='DAV:'><prop><resourcetype/></prop><version><href>"
                + rootVersionUri
                + "</href></version></update>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            checkResponseDocument(response, respfile, null);
          }

          break;

        case 12:
          clear(request);
          request.setMethod("DELETE");
          request.setURI(secondVersionOfCollection);
          response = requester.perform(request);
          checkResponseCode(response, 204);
          break;

        case 13:
            
          clear(request);
          request.setMethod("MKCOL");
          request.setURI(getLocalpart() + "/vc13");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          
          clear(request);
          request.setMethod("VERSION-CONTROL");
          request.setURI(getLocalpart() + "/vc13");
          response = requester.perform(request);
          checkResponseCode(response, 201);
          
          // remember VHR URI for later
          vhr = getProperty(getLocalpart() + "/vc13", Consts.N_version_history);
          versionHistoryUri = getRelativeUriFromPropertyHrefs(vhr);
          assertNotNull("version history uri not found", versionHistoryUri);
          vhrs.add(versionHistoryUri);
          
          String vhrParentUri = versionHistoryUri;
          if (vhrParentUri.endsWith("/")) {
            vhrParentUri = vhrParentUri.substring(0, vhrParentUri.length()-1);
          }
          vhrParentUri = vhrParentUri.substring(0, vhrParentUri.lastIndexOf("/"));
          assertNotNull("version history parent uri not found", versionHistoryUri);
          assertTrue("version history parent uri not found", versionHistoryUri.length() > 0);
          
          clear(request);
          request.setMethod("HEAD");
          request.setURI(vhrParentUri);
          response = requester.perform(request);
          checkResponseCode(response, 200);
          
          break;
              
        default:

          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/testresource3");
          response = requester.perform(request);
          checkResponseCode(response, 204, 404);

          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/vc");
          response = requester.perform(request);
          checkResponseCode(response, 204, 404);

          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/vc13");
          response = requester.perform(request);
          checkResponseCode(response, 204, 404);
          
          // cleanup VHRs
          for (Iterator it = vhrs.iterator(); it.hasNext(); ) {
            String uri = it.next().toString();
            clear(request);
            request.setMethod("DELETE");
            request.setURI(uri);
            response = requester.perform(request);
            if (vhrPreserved())
              checkResponseCode(response, 204, 207, 404);
            else
              checkResponseCode(response, 404);
          }
          break;
      }
    }
    catch (Exception ex) {
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }

  private String getContent(InputStream is) throws IOException {
    if (is == null)
      return null;

    byte[] buffer = new byte[64 * 1024];
    ByteArrayOutputStream bos = new ByteArrayOutputStream();
    int len;
    while ((len = is.read(buffer)) != -1) {
      bos.write(buffer, 0, len);
    }
    is.close();

    return bos.toString();
  }
}
