package com.sapportals.wcm.protocol.webdav.server.tests;

import java.util.ArrayList;
import java.util.List;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import com.sapportals.wcm.WcmException;
import com.sapportals.wcm.protocol.webdav.DomElement;
import com.sapportals.wcm.protocol.webdav.DomSerializer;
import com.sapportals.wcm.protocol.webdav.dompath.DomPath;
import com.sapportals.wcm.util.http.HttpRequest;
import com.sapportals.wcm.util.http.IRequester;
import com.sapportals.wcm.util.http.IResponse;
import com.sapportals.wcm.util.uri.URICodec;

/**
 * Helper methods for creating test repository.
 *
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author martin.boettcher@greenbytes.de
 * @author stefan.eissing@greenbytes.de
 * @author julian.reschke@greenbytes.de
 * @version $Id: WDHelper.java,v 1.7 2004/06/18 09:38:52 jre Exp $
 */

public class WDHelper
{
  private static final DomPath DP_hrefs = new DomPath(
    "/D:multistatus/D:response/D:href", DomPath.DEFAULT_NAMESPACE_BINDINGS);
  private static final DomPath DP_locktoken_href = new DomPath(
    "../D:propstat/D:prop/D:lockdiscovery/D:activelock/D:locktoken/D:href", DomPath.DEFAULT_NAMESPACE_BINDINGS);

  private IRequester m_requester;
  private HttpRequest m_request;
  private URICodec m_codec;
  private String m_host;
  private String m_localpart;

  public WDHelper (IRequester requester, String host, String localpart)
  {
    m_requester = requester;
    m_request = new HttpRequest();
    m_codec = new URICodec();
    m_host = host;
    if (localpart != null && localpart.endsWith("\\"))
      localpart = localpart.substring(0, localpart.length()-1 );
    m_localpart = localpart;
  }

  public void clean() throws WcmException
  {
    String ua = System.getProperty("userAgent", null);
    
    // create target if not present
    int status = 503;
    int retries = 25;
    IResponse response;
    
    while (status == 503 && retries-- > 0) {
      m_request.clear();
      m_request.setMethod("MKCOL");
      m_request.setHeader("x-greenbytes", "create target collection");
      if (ua != null) m_request.setHeader("user-agent", ua);
      m_request.setURI(m_localpart);
      response = m_requester.perform(m_request);
      status = response.getStatus();
      
      // System.out.println("status: " + status);

      if (status == 503) {
        try {
          // service not available
          Thread.sleep(20000);
        }
        catch (InterruptedException e) {
          // don't care
        }
      }
    }
    
    m_request.clear();
    m_request.setMethod("PROPFIND");
    m_request.setURI(m_localpart);
    m_request.setHeader("Depth", "1");
    m_request.expectsResponseDocument(true);
    response = m_requester.perform(m_request);

    if (response.getStatus() != 207) {
      throw new WcmException ("PROPFIND returned " + response.getStatus() +
      " instead of 207");
    }

    Document doc = response.getDocument();
    if (doc == null) {
      throw new WcmException ("PROPFIND did not return (well-formed) response body");
    }

    List hrefs = new ArrayList();
    List locks = new ArrayList();
    getURLs(doc, "/", hrefs, locks);
    for (int i = 0; i < hrefs.size(); ++i) {
      String uri = (String)hrefs.get(i);
      String lock = (String)locks.get(i);
      if (lock.length() > 0) {
        m_request.clear();
        m_request.setMethod("UNLOCK");
        m_request.setURI(m_localpart + uri);
        m_request.setHeader("Lock-Token", "<"+lock+">");
        response = m_requester.perform(m_request);
      }
      m_request.clear();
      m_request.setMethod("DELETE");
      m_request.setURI(m_localpart + uri);
      m_request.setHeader("Depth", "Infinity");
      m_request.setHeader("Apply-To-Redirect-Ref", "");
      response = m_requester.perform(m_request);
      if (response.getStatus() != 204 && response.getStatus() != 404) {
        Document ms = response.getDocument();
        throw new WcmException("removing "+uri
                               +" got status "+response.getStatus()
                               + ((ms != null)? DomSerializer.toString(ms, true) : ""));
      }
    }
  }

  public void makeDirStructure() throws WcmException
  {
    String[] rids = new String[] {
      "/leer",
      "/voll",
      "/voll/ist leer",
      "/voll/ist was drin",
    };

    for (int i = 0; i < rids.length; ++i) {
      String rid = rids[i];
      m_request.clear();
      m_request.setMethod("MKCOL");
      m_request.setURI(m_localpart + m_codec.encode(rid));
      IResponse response = m_requester.perform(m_request);
      int status = response.getStatus();
      if (status != 201) { throw new WcmException("mkcol '" + rid + "' returned " + status); }
    }
  }

  private void copy(String source, String destURI, String mimetype, String encoding, String language) throws Exception
  {
    java.io.InputStream fis = this.getClass().getResourceAsStream ("files/" + source);

    int content_length = 0;
    int len;
    byte[] buffer = new byte[32 * 1024];
    while ((len = fis.read(buffer)) != -1) {
      content_length += len;
    }
    fis.close();

    fis = this.getClass().getResourceAsStream ("files/" + source);
    try {
      m_request.clear();
      m_request.setMethod("PUT");
      m_request.setURI(m_codec.encode(m_localpart + destURI));
      if (mimetype != null) {
        if (encoding == null) {
          m_request.setHeader ("Content-Type", mimetype);
        }
        else {
          m_request.setHeader ("Content-Type", mimetype + "; charset=" + encoding);
        }
      }
      if (language != null) m_request.setHeader ("Content-Language", language);
      m_request.setHeader("Content-Length", String.valueOf(content_length));
      m_request.setBody(fis);
      IResponse response = m_requester.perform(m_request);
      int status = response.getStatus();
      if (status != 201) {
        throw new WcmException("put on "+destURI+" returned "+status);
      }
    }
    finally {
      fis.close();
    }
  }

  public void setup() throws Exception
  {
    clean();
    makeDirStructure();
    copy("ANSI.TXT.copyme", "/ANSI.TXT", "text/plain", "ISO-8859-1", "de");
    copy("Euro.txt.copyme", "/Euro\u20ac.txt", "text/plain", null, null);
    copy("Mappe1.xls.copyme", "/Mappe1.xls", "application/vnd.ms-excel", null, null);
    copy("my_dog_and_cat.txt.copyme", "/my_dog_&_cat.txt", "text/plain", null, "us");
    copy("Test.ppt.copyme", "/Test!.ppt", "application/vnd.ms-powerpoint", null, null);
    copy("Umlaute.txt.copyme", "/Umlaute \u00c4\u00d6\u00dc\u00e4\u00f6\u00fc\u00df.txt", "text/plain", "ISO-8859-1", null);
    copy("Unicode-big-endian.txt.copyme", "/Unicode Big Endian.TXT", "text/plain", "UTF-16", null);
    copy("Unicode.TXT.copyme", "/Unicode.TXT", "text/plain", "\"UTF-16\"", null);
    copy("UTF8.TXT.copyme", "/UTF8.TXT", "text/plain", "'UTF-8'", null);
    copy("Worddatei.doc.copyme", "/Worddatei.doc", "application/msword", null, null);
    copy("office.zip.copyme", "/voll/ist was drin/office.zip", "application/zip", null, null);
  }


  private void getURLs(Document doc, String ignore, List result, List resultTokens) {
    Node[] hrefs = DP_hrefs.matchNodes(doc);
    for (int i = 0; i < hrefs.length; ++i) {
      Element href = (Element)hrefs[i];
      String url = DomElement.getText(href);
      if (url.startsWith(m_host)) {
        url = url.substring(m_host.length());
      }
      if (url.startsWith(m_localpart)) {
        url = url.substring(m_localpart.length());
      }
      if (url.length() > 0 && !url.equals(ignore)) {
        result.add(url);
        String tokenString = "";

        Node[] tokens = DP_locktoken_href.matchNodes(href);
        for (int j = 0; j < tokens.length; ++j) {
          Element locktoken = (Element)tokens[j];
          tokenString = DomElement.getText(locktoken);
        }

        resultTokens.add(tokenString);
      }
    }
    //System.err.println("hrefs+tokens: "+result+" "+resultTokens);
  }
}