package com.sapportals.wcm.protocol.webdav.server.tests;

import java.util.*;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import com.sapportals.wcm.protocol.webdav.Consts;
import com.sapportals.wcm.protocol.webdav.DavConditions;
import com.sapportals.wcm.protocol.webdav.DomElement;
import com.sapportals.wcm.util.http.IResponse;

/**
 * Various tests for working resources.
 * <p>
 * Copyright (c) SAP AG 2002-2005
 * @author julian.reschke@greenbytes.de
 * @version $Id: WorkingResourceTest.java,v 1.8 2005/09/07 13:43:47 mbo Exp $
 */
public class WorkingResourceTest extends TestMember
{

  private static final String[] desc = new String[]
  {
    "00 WR Setting up WR",
    "01 WR checkout WR from version, modify, checkin",
    "02 WR checkout WR from version, modify, checkin-keep-checked-out, modify, checkin",
    "03 WR checkout WR from vcr, modify, checkin",
    "04 WR version-control collection",
    "05 WR checkout collection version",
    "06 WR add plain resource to working collection, checkin",
    "07 WR checkout vc with apply-to-version and auto-update (first without fork-ok, then retry), delete vhr from wc, checkin",
    "08 WR undelete of vcr with eclipsed result",
    "09 WR delete eclipsing plain resource",
    "10 WR make sure VCR can be deleted while WR exists (plain)",
    "11 WR REPORT DAV:version-tree on both VCR and version",
    "12 WR checkout version which is already checked out by another vcr, checkout version with descendants",
    "13 WR checkout vcr from version with descendants, checkout vcr which is already checked out by another vcr",
    "   WR cleanup",
  };

  public static final int NrOfTests = desc.length;


  public WorkingResourceTest(Integer nr)
  {
    super(nr, desc[nr.intValue()]);
  }

  private static String wrUri = null;
  private static String wcUri = null;
  private static String wrUri2 = null;
  private static String rootVersion = null;
  private static String lastVersion = null;

  private static boolean workingCollectionsSupported = true;

  private static Map versionHistoryMap = new HashMap();

  protected void runTest()
  {
    if (nr == 0) logIdentifier("WR");

    String collection = "/wrtest";
    String file = "/wrtest/testfile";
    String file2 = "/testfile2";

    try
    {
      IResponse response;

      clear(request);

      switch (nr)
      {
        case 0:
          clear(request);
          request.setMethod("MKCOL");
          request.setURI(getLocalpart() + collection);
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + file);
          request.setBody("created by test case 00 WR - xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("VERSION-CONTROL");
          request.setURI(getLocalpart() + file);
          response = requester.perform(request);
          checkResponseCode(response, 201);
          rootVersion = getRelativeUriFromLocationHeader(response);

          Element vhr = getProperty(getLocalpart() + file, Consts.N_version_history);
          if (vhr != null) {
            versionHistoryMap.put(getLocalpart() + file, getRelativeUriFromPropertyHrefs(vhr));
          }

          break;

        case 1:
          {
            Element checkedin = getProperty(getLocalpart() + file, Consts.N_checked_in);
            String versionuri = getRelativeUriFromPropertyHrefs(checkedin);

            clear(request);
            request.setMethod("CHECKOUT");
            request.setURI(versionuri);
            response = requester.perform(request);
            checkResponseCode(response, 201);

            wrUri = getRelativeUriFromLocationHeader(response);

            Element checkedout = getProperty(wrUri, Consts.N_checked_out);
            String tmp = getRelativeUriFromPropertyHrefs(checkedout);

            assertTrue(tmp.equals(versionuri));

            String comment = "from test case 01 WR";

            clear(request);
            request.setMethod("PROPPATCH");
            request.setURI(wrUri);
            request.setBody("<propertyupdate xmlns='DAV:'><set><prop><comment>"+comment+"</comment></prop></set></propertyupdate>");
            response = requester.perform(request);
            checkResponseCode(response, 207);

            clear(request);
            request.setMethod("CHECKIN");
            request.setURI(wrUri);
            response = requester.perform(request);
            checkResponseCode(response, 201);

            versionuri = getRelativeUriFromLocationHeader(response);

            Element com = getProperty(versionuri, Consts.N_comment);
            assertTrue(comment.equals(DomElement.getText(com)));

            clear(request);
            request.setMethod("HEAD");
            request.setURI(wrUri);
            response = requester.perform(request);
            checkResponseCode(response, 404);
          }
          break;

        case 2:
          {
            Element checkedin = getProperty(getLocalpart() + file, Consts.N_checked_in);
            String versionuri = getRelativeUriFromPropertyHrefs(checkedin);

            clear(request);
            request.setMethod("CHECKOUT");
            request.setBody("<checkout xmlns='DAV:'><fork-ok/></checkout>");
            request.setURI(versionuri);
            response = requester.perform(request);
            checkResponseCode(response, 201);

            wrUri = getRelativeUriFromLocationHeader(response);

            Element checkedout = getProperty(wrUri, Consts.N_checked_out);
            String tmp = getRelativeUriFromPropertyHrefs(checkedout);

            assertTrue(tmp.equals(versionuri));

            String comment = "from test case 02 WR";

            clear(request);
            request.setMethod("PROPPATCH");
            request.setURI(wrUri);
            request.setBody("<propertyupdate xmlns='DAV:'><set><prop><comment>"+comment+"</comment></prop></set></propertyupdate>");
            response = requester.perform(request);
            checkResponseCode(response, 207);

            clear(request);
            request.setMethod("CHECKIN");
            request.setBody("<checkin xmlns='DAV:'><fork-ok/><keep-checked-out/><compute-expected-version-URI xmlns='"+Consts.NS_WDSERVLET+"'/></checkin>");
            request.expectsResponseDocument(true);
            request.setURI(wrUri);
            response = requester.perform(request);
            checkResponseCode(response, 201);

            Document doc = response.getDocument();
            assertTrue(doc != null);
            List uris = DomElement.getChildren(doc.getDocumentElement(), Consts.N_SAP_expected_version_URI);
            assertTrue(uris.size() == 1);
            assertTrue(DomElement.getHrefs(((Element)uris.get(0))).size() == 1);

            versionuri = getRelativeUriFromLocationHeader(response);

            Element com = getProperty(versionuri, Consts.N_comment);
            assertTrue(comment.equals(DomElement.getText(com)));

            // Element base = getProperty(wrUri, Consts.N_checked_in);

            clear(request);
            request.setMethod("PROPPATCH");
            request.setURI(wrUri);
            request.setBody("<propertyupdate xmlns='DAV:'><set><prop><comment>"+comment+"++"+"</comment></prop></set></propertyupdate>");
            response = requester.perform(request);
            checkResponseCode(response, 207);

            clear(request);
            request.setMethod("CHECKIN");
            request.setURI(wrUri);
            response = requester.perform(request);
            checkResponseCode(response, 201);

            versionuri = getRelativeUriFromLocationHeader(response);

            com = getProperty(versionuri, Consts.N_comment);
            assertTrue((comment+"++").equals(DomElement.getText(com)));

            clear(request);
            request.setMethod("HEAD");
            request.setURI(wrUri);
            response = requester.perform(request);
            checkResponseCode(response, 404);
          }
          break;

        case 3:
          {
            clear(request);
            request.setMethod("CHECKOUT");
            request.setURI(getLocalpart() + file);
            request.setBody("<checkout xmlns='DAV:'><apply-to-version/><fork-ok/></checkout>");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            wrUri = getRelativeUriFromLocationHeader(response);

            Element autoupdate = getProperty(wrUri, Consts.N_auto_update);
            String updateUri = getRelativeUriFromPropertyHrefs(autoupdate);
            assertTrue(updateUri.equals(getLocalpart() + file));

            assertMethodNotAllowed(wrUri, "VERSION-CONTROL");

            String comment = "from test case 03 WR";

            clear(request);
            request.setMethod("PROPPATCH");
            request.setURI(wrUri);
            request.setBody("<propertyupdate xmlns='DAV:'><set><prop><comment>"+comment+"</comment></prop></set></propertyupdate>");
            response = requester.perform(request);
            checkResponseCode(response, 207);

            clear(request);
            request.setMethod("CHECKIN");
            request.setBody("<checkin xmlns='DAV:'><fork-ok/></checkin>");
            request.setURI(wrUri);
            response = requester.perform(request);
            checkResponseCode(response, 201);

            Element com = getProperty(getLocalpart() + file, Consts.N_comment);
            assertTrue(com != null);
            assertTrue(comment.equals(DomElement.getText(com)));
          }
          break;

        case 4:
          {
            clear(request);
            request.setMethod("VERSION-CONTROL");
            request.setURI(getLocalpart() + collection);
            response = requester.perform(request);
            checkResponseCode(response, 201, 501);

            workingCollectionsSupported = (response.getStatus() == 201);
            if (workingCollectionsSupported) {
              vhr = getProperty(getLocalpart() + collection, Consts.N_version_history);
              if (vhr != null) {
                versionHistoryMap.put(getLocalpart() + collection, getRelativeUriFromPropertyHrefs(vhr));
              }
            }
          }
          break;

        case 5:
          if (workingCollectionsSupported) {
            Element checkedin = getProperty(getLocalpart() + collection, Consts.N_checked_in);
            String versionuri = getRelativeUriFromPropertyHrefs(checkedin);

            clear(request);
            request.setMethod("CHECKOUT");
            request.setBody("<checkout xmlns='DAV:'><fork-ok/></checkout>");
            request.setURI(versionuri);
            response = requester.perform(request);
            checkResponseCode(response, 201, 403, 501);

            workingCollectionsSupported = (response.getStatus() == 201);

            if (workingCollectionsSupported) {
              wcUri = getRelativeUriFromLocationHeader(response);

              Element resourcetype = getProperty(wcUri, Consts.N_resourcetype);
              assertTrue(DomElement.hasChild(resourcetype, Consts.N_collection));

              Element checkedout = getProperty(wcUri, Consts.N_checked_out);
              assertTrue(checkedout != null);

              // working collection must contain binding to testfile vhr
              Element version_set_of_content = getProperty(wcUri + "/testfile", Consts.N_version_set);
              assertTrue(version_set_of_content != null);
            }
          }
          super.logComment(workingCollectionsSupported ? "working collections:yes" : "working collections:no");
          break;

        case 6:
          if (workingCollectionsSupported) {
            clear(request);
            request.setMethod("PUT");
            request.setURI(wcUri + "/testfile2");
            request.setBody("created by test case 06 WR - jdghsdsdsdhshdg");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("CHECKIN");
            request.setBody("<checkin xmlns='DAV:'><fork-ok/></checkin>");
            request.setURI(wcUri);
            response = requester.perform(request);
            checkResponseCode(response, 201);

            String versionuri = getRelativeUriFromLocationHeader(response);

            Element bindingSetElt = getProperty(versionuri, Consts.N_version_controlled_binding_set);
            assertTrue(bindingSetElt != null);
            assertTrue(bindingSetElt.getChildNodes().getLength() == 2);

            HashSet bindingSetRef = new HashSet();
            bindingSetRef.add("testfile");
            bindingSetRef.add("testfile2");

            Set bindingSetAct = new HashSet();
            Element binding1
                = (Element) DomElement.getChildren(bindingSetElt, Consts.N_version_controlled_binding).get(0);
            Element binding2
                = (Element) DomElement.getChildren(bindingSetElt, Consts.N_version_controlled_binding).get(1);
            bindingSetAct.add(
              DomElement.getText((Node) DomElement.getChildren(binding1, Consts.N_binding_name).get(0)));
            bindingSetAct.add(
              DomElement.getText((Node) DomElement.getChildren(binding2, Consts.N_binding_name).get(0)));

            bindingSetRef.removeAll(bindingSetAct);
            assertTrue(bindingSetRef.isEmpty());
          }
          break;

        case 7:
          if (workingCollectionsSupported) {
            clear(request);
            request.setMethod("CHECKOUT");
            request.setBody("<checkout xmlns='DAV:'><apply-to-version/></checkout>");
            request.setURI(getLocalpart() + collection);
            response = requester.perform(request);
            checkResponseCode(response, 403);

            clear(request);
            request.setMethod("CHECKOUT");
            request.setBody("<checkout xmlns='DAV:'><apply-to-version/><fork-ok/></checkout>");
            request.setURI(getLocalpart() + collection);
            response = requester.perform(request);
            checkResponseCode(response, 201);

            wcUri = getRelativeUriFromLocationHeader(response);

            Element resourcetype = getProperty(wcUri, Consts.N_resourcetype);
            assertTrue(DomElement.hasChild(resourcetype, Consts.N_collection));

            Element checkedout = getProperty(wcUri, Consts.N_checked_out);
            assertTrue(checkedout != null);

            // working collection must contain binding to testfile vhrs
            Element version_set_of_content = getProperty(wcUri + "/testfile", Consts.N_version_set);
            assertTrue(version_set_of_content != null);

            clear(request);
            request.setMethod("DELETE");
            request.setURI(wcUri + "/testfile");
            response = requester.perform(request);
            checkResponseCode(response, 204);

            clear(request);
            request.setMethod("CHECKIN");
            request.setBody("<checkin xmlns='DAV:'><fork-ok/></checkin>");
            request.setURI(wcUri);
            response = requester.perform(request);
            checkResponseCode(response, 201);

            String versionuri = getRelativeUriFromLocationHeader(response);

            clear(request);
            request.setMethod("HEAD");
            request.setURI(getLocalpart() + collection + "/testfile");
            response = requester.perform(request);
            checkResponseCode(response, 404);

            Element vcCheckedIn = getProperty(getLocalpart() + collection, Consts.N_checked_in);
            String checkedInUri = getRelativeUriFromPropertyHrefs(vcCheckedIn);

            assertTrue(checkedInUri.equals(versionuri));
          }
          break;

        case 8:
          if (workingCollectionsSupported) {

            Element checkedin = getProperty(getLocalpart() + collection, Consts.N_checked_in);
            assertTrue(checkedin != null);

            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + collection + "/testfile");
            request.setBody("created by test case 08 WR - dsdsddsdsdsd");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("CHECKOUT");
            request.setBody("<checkout xmlns='DAV:'><apply-to-version/><fork-ok/></checkout>");
            request.setURI(getLocalpart() + collection);
            response = requester.perform(request);
            checkResponseCode(response, 201);

            wcUri = getRelativeUriFromLocationHeader(response);

            Element versionHistory = getProperty(getLocalpart() + collection, Consts.N_version_history);
            String versionHistoryUri = getRelativeUriFromPropertyHrefs(versionHistory);

            Element rootvers =  getProperty(versionHistoryUri, Consts.N_root_version);
            String rootVersionUri = getRelativeUriFromPropertyHrefs(rootvers);

            clear(request);
            request.setMethod("CHECKOUT");
            request.setBody("<checkout xmlns='DAV:'><fork-ok/></checkout>");
            request.setURI(rootVersionUri);
            response = requester.perform(request);
            checkResponseCode(response, 201);

            String wcUriRootVersion = getRelativeUriFromLocationHeader(response);

            Element rootVersionBindings =  getProperty(rootVersionUri, Consts.N_version_controlled_binding_set);
            assertTrue(rootVersionBindings != null);
            List childList = DomElement.getChildren(rootVersionBindings, Consts.N_version_controlled_binding);
            assertTrue(!childList.isEmpty());
            Element rootVersionBinding = (Element) childList.get(0);
            Node bindingNameElt = (Node) DomElement.getChildren(rootVersionBinding, Consts.N_binding_name).get(0);
            String bindingName = DomElement.getText(bindingNameElt);
            assertTrue(bindingName.equals("testfile"));

            clear(request);
            request.setMethod("MOVE");
            request.setHeader("Destination", getBase() + wcUri + "/testfile");
            request.setHeader("Overwrite", "F");
            request.setURI(wcUriRootVersion + "/testfile");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("DELETE");
            request.setURI(wcUriRootVersion);
            response = requester.perform(request);
            checkResponseCode(response, 204);

            clear(request);
            request.setMethod("CHECKIN");
            request.setBody("<checkin xmlns='DAV:'><fork-ok/></checkin>");
            request.setURI(wcUri);
            response = requester.perform(request);
            checkResponseCode(response, 201);
          }
          break;

        case 9:
          if (workingCollectionsSupported) {

            Element vcCheckedIn = getProperty(getLocalpart() + collection, Consts.N_checked_in);
            String checkedInUri = getRelativeUriFromPropertyHrefs(vcCheckedIn);

            Element vcb = getProperty(checkedInUri, Consts.N_version_controlled_binding_set);
            assertTrue(vcb != null);
            List vcbList = DomElement.getChildren(vcb, Consts.N_version_controlled_binding);
            assertTrue(!vcbList.isEmpty());
            Element bindingElt = (Element) vcbList.get(0);
            Node bindingNameElt = (Node) DomElement.getChildren(bindingElt, Consts.N_binding_name).get(0);
            String bindingName = DomElement.getText(bindingNameElt);
            assertTrue(bindingName.equals("testfile"));

            Element ecb = getProperty(getLocalpart() + collection, Consts.N_eclipsed_set);
            assertTrue(ecb != null);
            List ecbList = DomElement.getChildren(ecb, Consts.N_binding_name);
            assertTrue(!ecbList.isEmpty());
            Element eclipsedbindingElt = (Element) ecbList.get(0);
            String eclipsedbindingName = DomElement.getText(eclipsedbindingElt);
            assertTrue(eclipsedbindingName.equals("testfile"));

            String eclUri = getLocalpart() + collection + "/testfile";

            Element eclUriCheckedIn =  getProperty(eclUri, Consts.N_checked_in);
            assertTrue(eclUriCheckedIn == null);

            clear(request);
            request.setMethod("DELETE");
            request.setURI(eclUri);
            response = requester.perform(request);
            checkResponseCode(response, 204);

            eclUriCheckedIn = getProperty(eclUri, Consts.N_checked_in);
            assertTrue(eclUriCheckedIn != null);
          }
          break;

        case 10:
          {
            String delfile = "/delete-plain";

            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + delfile);
            request.setBody("created by test case 10 WR - dfsdjsddddddddddddddddddddddddddddd");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("VERSION-CONTROL");
            request.setURI(getLocalpart() + delfile);
            response = requester.perform(request);
            checkResponseCode(response, 201);

            vhr = getProperty(getLocalpart() + delfile, Consts.N_version_history);
            if (vhr != null) {
              versionHistoryMap.put(getLocalpart() + delfile, getRelativeUriFromPropertyHrefs(vhr));
            }

            Element checkedin = getProperty(getLocalpart() + delfile, Consts.N_checked_in);
            String versionuri = getRelativeUriFromPropertyHrefs(checkedin);

            clear(request);
            request.setMethod("CHECKOUT");
            request.setBody("<checkout xmlns='DAV:'><fork-ok/></checkout>");
            request.setURI(versionuri);
            response = requester.perform(request);
            checkResponseCode(response, 201);

            wrUri2 = getRelativeUriFromLocationHeader(response);
            Element vhrElt = getProperty(versionuri, Consts.N_version_history);
            assertTrue(vhrElt != null);
            String vhrUri = getRelativeUriFromPropertyHrefs(vhrElt);

            // make sure WR is versioned
            assertTrue(null != getProperty(wrUri2, Consts.N_checked_out));

            // attempt the DELETE
            clear(request);
            request.setMethod("DELETE");
            request.setURI(vhrUri);
            response = requester.perform(request);
            checkResponseCode(response, 204);

            // make sure WR not versioned anymore
            assertTrue(null == getProperty(wrUri2, Consts.N_checked_out));

            // cleanup WR
            clear(request);
            request.setMethod("DELETE");
            request.setURI(wrUri2);
            response = requester.perform(request);
            checkResponseCode(response, 204);

            //cleanup
            clear(request);
            request.setMethod("DELETE");
            request.setURI(getLocalpart() + delfile);
            response = requester.perform(request);
            checkResponseCode(response, 204);
          }
          break;

        case 11: {
            Set versions = new HashSet();

            clear(request);
            request.setMethod("REPORT");
            request.setURI(rootVersion);
            request.setHeader("Depth", "0");
            request.setBody("<version-tree xmlns='DAV:'><prop><predecessor-set/><successor-set/></prop></version-tree>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            List responses = DomElement.getChildren(response.getDocument().getDocumentElement(), Consts.N_response);
            for (Iterator it = responses.iterator(); it.hasNext();) {
              Element e = (Element)it.next();
              versions.add(getRelativeUriFromPropertyHrefs(e));
            }

            clear(request);
            request.setMethod("REPORT");
            request.setURI(getLocalpart() + file);
            request.setHeader("Depth", "0");
            request.setBody("<version-tree xmlns='DAV:'><prop><predecessor-set/><successor-set/></prop></version-tree>");
            response = requester.perform(request);
            checkResponseCode(response, 207);
            responses = DomElement.getChildren(response.getDocument().getDocumentElement(), Consts.N_response);
            assertTrue(responses.size() == versions.size());
            for (Iterator it = responses.iterator(); it.hasNext();) {
              Element e = (Element)it.next();
              assertTrue(versions.contains(getRelativeUriFromPropertyHrefs(e)));
            }
          }
          break;

        case 12:

          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + file2);
          request.setBody("created by test case 12 WR - ###################################################");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("VERSION-CONTROL");
          request.setURI(getLocalpart() + file2);
          response = requester.perform(request);
          checkResponseCode(response, 201);
          rootVersion = getRelativeUriFromLocationHeader(response);

          vhr = getProperty(getLocalpart() + file2, Consts.N_version_history);
          if (vhr != null) {
            versionHistoryMap.put(getLocalpart() + file2, getRelativeUriFromPropertyHrefs(vhr));
          }

          clear(request);
          request.setMethod("CHECKOUT");
          request.setURI(rootVersion);
          response = requester.perform(request);
          checkResponseCode(response, 201);
          wrUri = getRelativeUriFromLocationHeader(response);

          clear(request);
          request.setMethod("CHECKOUT");
          request.setURI(rootVersion);
          response = requester.perform(request);
          checkStatusAndErrorElement(response, 403, DavConditions.N_checkout_of_checked_out_version_is_discouraged);

          clear(request);
          request.setMethod("DELETE");
          request.setURI(wrUri);
          response = requester.perform(request);
          checkResponseCode(response, 204);

          clear(request);
          request.setMethod("CHECKOUT");
          request.setURI(getLocalpart() + file2);
          response = requester.perform(request);
          checkResponseCode(response, 200);

          clear(request);
          request.setMethod("CHECKIN");
          request.setURI(getLocalpart() + file2);
          response = requester.perform(request);
          checkResponseCode(response, 201);
          lastVersion = getRelativeUriFromLocationHeader(response);

          clear(request);
          request.setMethod("CHECKOUT");
          request.setURI(rootVersion);
          response = requester.perform(request);
          checkStatusAndErrorElement(response, 403, DavConditions.N_checkout_of_version_with_descendant_is_discouraged);
          break;

        case 13:

          clear(request);
          request.setMethod("UPDATE");
          request.setURI(getLocalpart() + file2);
          request.setBody("<update xmlns='DAV:'><version><href>"
              + rootVersion
              + "</href></version></update>");
          response = requester.perform(request);
          checkResponseCode(response, 207);

          clear(request);
          request.setMethod("CHECKOUT");
          request.setURI(getLocalpart() + file2);
          response = requester.perform(request);
          checkStatusAndErrorElement(response, 403, DavConditions.N_checkout_of_version_with_descendant_is_discouraged);

          clear(request);
          request.setMethod("UPDATE");
          request.setURI(getLocalpart() + file2);
          request.setBody("<update xmlns='DAV:'><version><href>"
              + lastVersion
              + "</href></version></update>");
          response = requester.perform(request);
          checkResponseCode(response, 207);

          clear(request);
          request.setMethod("CHECKOUT");
          request.setURI(lastVersion);
          response = requester.perform(request);
          checkResponseCode(response, 201);
          wrUri = getRelativeUriFromLocationHeader(response);

          clear(request);
          request.setMethod("CHECKOUT");
          request.setURI(getLocalpart() + file2);
          response = requester.perform(request);
          checkStatusAndErrorElement(response, 403, DavConditions.N_checkout_of_checked_out_version_is_discouraged);

          clear(request);
          request.setMethod("DELETE");
          request.setURI(wrUri);
          response = requester.perform(request);
          checkResponseCode(response, 204);
          
          break;

        default:

          if (wcUri != null) {
            clear(request);
            request.setMethod("DELETE");
            request.setURI(wcUri);
            response = requester.perform(request);
            checkResponseCode(response, 204, 404);
          }

          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + collection);
          response = requester.perform(request);
          checkResponseCode(response, 204, 404);

          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + file2);
          response = requester.perform(request);
          checkResponseCode(response, 204, 404);

          Iterator vhrIter = versionHistoryMap.values().iterator();
          while (vhrIter.hasNext()) {
            String vhrUri = (String) vhrIter.next();
            clear(request);
            request.setMethod("DELETE");
            request.setURI(vhrUri);
            response = requester.perform(request);
            if (vhrPreserved()) 
              checkResponseCode(response, 204, 207, 404);
            else
              checkResponseCode(response, 404);
          }
          break;
      }
    }
    catch (Exception ex) {
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }

}
