package com.sapportals.wcm.protocol.webdav.server.tests;

import java.util.*;

import org.w3c.dom.Element;

import com.sapportals.wcm.protocol.webdav.Consts;
import com.sapportals.wcm.protocol.webdav.DomElement;
import com.sapportals.wcm.protocol.webdav.DomSerializer;
import com.sapportals.wcm.util.http.IResponse;

/**
 * Various tests for workspaces and workspace controlled resources.
 * <p>
 * Copyright (c) SAP AG 2002-2004
 * @author manfred.baedke@greenbytes.de
 * @version $Id: WorkspaceTest.java,v 1.7 2005/02/18 12:46:59 sei Exp $
 */
public class WorkspaceTest extends TestMember
{

  private static final String[] desc = new String[]
  {
    "00 WS Setting up WS",
    "01 WS Create second WS, try move between WSs, try bind into foreign WS",
    "02 WS try move WS into non WS collection, rename WS",
    "03 WS version control in WS",
    "04 WS try to create second vcr from the same version history in the same WS, create it in another WS",
    "05 WS WS-internal move/bind",
    "06 WS create version controlled collection in ws",
    "07 WS checkout vcc, move away version controlled binding (workspace internal), uncheckout",
    "08 WS checkout vcc, create version controlled binding, create workingresource (workspace external), delete binding, uncheckout",
    "   WS cleanup",
  };

  public static final int NrOfTests = desc.length;

  public WorkspaceTest(Integer nr)
  {
    super(nr, desc[nr.intValue()]);
  }

  private static boolean wcolSupported = false;

  private static String wspColUri = null;
  private static String wspUriA = null;
  private static String wspUriB = null;
  private static String wspUriC = null;

  public static Set vhrs = new HashSet();

  protected void runTest()
  {
    if (nr == 0) logIdentifier("WS");

    String respfile = "WS-" + TestMember.THREEDIGITS.format(nr) + ".txt";

    try
    {
      IResponse response;

      clear(request);

      switch (nr)
      {
        case 0:
          //workspace property should not be reported on not-workspace-controlled resources
          Element wspElement = this.getProperty(getLocalpart(), Consts.N_workspace);
          assertTrue(wspElement == null);

          //get workspace collection  set
          clear(request);
          request.setMethod("OPTIONS");
          request.setURI(getLocalpart());
          request.expectsResponseDocument(true);
          request.setBody("<options xmlns='DAV:'><workspace-collection-set/></options>");
          response = requester.perform(request);
          checkResponseCode(response, 200);

          assertTrue(response.getDocument() != null);
          List childrenList
              = DomElement.getChildren(response.getDocument().getDocumentElement(), Consts.N_workspace_collection_set);
          assertTrue(childrenList != null);
          assertTrue(childrenList.size() == 1);
          Set wscs = DomElement.getHrefs((Element) childrenList.get(0));
          wspColUri = removeProtocolAndBase((String)wscs.iterator().next());

          //get workspace collection set as live property
          boolean found = false;
          Element wspColElement = this.getProperty(getLocalpart(), Consts.N_workspace_collection_set);
          assertTrue(wspColElement != null);
          Set hrefs = DomElement.getHrefs(wspColElement);
          for (Iterator it = hrefs.iterator(); it.hasNext(); /**/) {
            String href = (String)it.next();
            found |= wspColUri.equals(removeProtocolAndBase(href));
          }
          assertTrue(found);

          assertMethodNotAllowed(getLocalpart(), "MKWORKSPACE");
          // build workspace name from last part of target RID to avoid
          // collisions with concurrently running tests
          wspUriA = wspColUri + getLocalpart().substring(getLocalpart().lastIndexOf('/')) + "wspa";
          assertMethodAllowed(wspUriA, "MKWORKSPACE");

          //make sure that the workspace folder can not be deleted
          clear(request);
          request.setMethod("DELETE");
          request.setURI(wspColUri);
          response = requester.perform(request);
          checkResponseCode(response, 403);          
          
          //create workspace
          clear(request);
          request.setMethod("MKWORKSPACE");
          request.setURI(wspUriA);
          response = requester.perform(request);
          checkResponseCode(response, 201);

          //lookup workspace property
          Element propElt = getProperty(wspUriA, Consts.N_workspace);
          assertTrue(propElt != null);
          assertTrue(wspUriA.equals(getRelativeUriFromPropertyHrefs(propElt)));
          break;

        case 1:

          // build workspace name from last part of target RID to avoid
          // collisions with concurrently running tests
          String wsName = super.getLocalpart().substring(super.getLocalpart().lastIndexOf('/')) + "wspb";
          wspUriB = wspColUri + wsName;

          if (FSInteractionTest.hasCollectionsInFS()) {
            // Use this second create workspace for testing if the creation works when the
            // collection already exists in the file system. This may happen after a backup/restore
            // when db and fs get a little out of synch
            FSInteractionTest.createWorkspaceDir(wsName);
          }
          
          //create second workspace
          clear(request);
          request.setMethod("MKWORKSPACE");
          request.setURI(wspUriB);
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("PUT");
          request.setURI(wspUriA + "/test");
          request.setBody("sadjfkjsf");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("MOVE");
          request.setURI(wspUriA + "/test");
          request.setHeader("Destination", getBase() + wspUriB + "/test");
          request.setHeader("Overwrite", "F");
          response = requester.perform(request);
          checkResponseCode(response, 403);

          boolean bindIsSupported = false;

          clear(request);
          request.setMethod("OPTIONS");
          request.setHeader("x-greenbytes-resource-context", "allow-bind-method=yes");
          request.setURI(wspUriB);
          response = requester.perform(request);
          checkResponseCode(response, 200);

          String allow = response.getHeader("allow");
          assertTrue("Allow header must be present in OPTIONS response", allow != null);


          StringTokenizer st = new StringTokenizer(allow, ", ");
          while (st.hasMoreElements() && !bindIsSupported) {
            String s = st.nextToken();
            if (s.equals("BIND")) {
              bindIsSupported = true;
            }
          }

          clear(request);
          request.setMethod("BIND");
          request.setHeader("x-greenbytes-resource-context", "allow-bind-method=yes");
          request.setURI(wspUriB);
          request.setBody("<bind xmlns='DAV:'>" + "<segment>test</segment>"
                          + "<href>" + wspUriA + "/test" + "</href></bind>");
          response = requester.perform(request);
          checkResponseCode(response, bindIsSupported? 403 : 405);

          break;

        case 2:

          wspUriC = wspColUri + super.getLocalpart().substring(super.getLocalpart().lastIndexOf('/')) + "wspc";

          //move WS into non-workspace collection
          clear(request);
          request.setMethod("MOVE");
          request.setURI(wspUriB);
          request.setHeader("Destination", getBaseAndLocalpart() + "/wspinvalidtarget");
          response = requester.perform(request);
          checkResponseCode(response, 403);

          //rename workspace
          clear(request);
          request.setMethod("MOVE");
          request.setURI(wspUriB);
          request.setHeader("Overwrite", "F");
          request.setHeader("Destination", getBase() + wspUriC);
          response = requester.perform(request);
          checkResponseCode(response, 201);

          break;

        case 3:

          //create 'normal' vcr
          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/wsptestvcr");
          request.setBody("sdkhfkshadfksad");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("VERSION-CONTROL");
          request.setURI(getLocalpart() + "/wsptestvcr");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          // remember VHR URI for later
          Element vhr = getProperty(getLocalpart() + "/wsptestvcr", Consts.N_version_history);
          vhrs.add(getRelativeUriFromPropertyHrefs(vhr));

          propElt = this.getProperty(getLocalpart() + "/wsptestvcr", Consts.N_checked_in);
          assertTrue(propElt != null);
          String version = this.getRelativeUriFromPropertyHrefs(propElt);
          assertTrue(version != null);

          // check precondition (DAV:cannot-add-to-existing-history)
          clear(request);
          request.setMethod("VERSION-CONTROL");
          request.setURI(wspUriA + "/test");
          request.setBody("<version-control xmlns='DAV:'><version>"
                          + "<href>" + version + "</href></version></version-control>");
          response = requester.perform(request);
          checkResponseCode(response, 409);

          // check precondition (DAV:must-be-version)
          clear(request);
          request.setMethod("VERSION-CONTROL");
          request.setURI(wspUriA + "/vcr");
          request.setBody("<version-control xmlns='DAV:'><version>"
                          + "<href>" + getLocalpart() + "/wsptestvcr" + "</href></version></version-control>");
          response = requester.perform(request);
          checkResponseCode(response, 403);

          // create vcr
          clear(request);
          request.setMethod("VERSION-CONTROL");
          request.setURI(wspUriA + "/vcr");
          request.setBody("<version-control xmlns='DAV:'><version>"
                          + "<href>" + version + "</href></version></version-control>");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          //lookup workspace property
          propElt = this.getProperty(wspUriA + "/vcr", Consts.N_workspace);
          assertTrue(propElt != null);
          assertTrue(wspUriA.equals(this.getRelativeUriFromPropertyHrefs(propElt)));

          //check postcondition (DAV:new-version-controlled-resource)
          clear(request);
          request.setMethod("REPORT");
          request.setURI(wspUriA + "/vcr");
          checkMultistatus("<version-tree xmlns='DAV:'><prop><comment/><version-name/><successor-set/></prop></version-tree>",
            respfile);

          break;

        case 4:

          propElt = this.getProperty(wspUriA + "/vcr", Consts.N_checked_in);
          assertTrue(propElt != null);
          version = this.getRelativeUriFromPropertyHrefs(propElt);
          assertTrue(version != null);

          //check precondition (DAV:one-version-controlled-resource-per-history-per-workspace)
          clear(request);
          request.setMethod("VERSION-CONTROL");
          request.setURI(wspUriA + "/vcr2");
          request.setBody("<version-control xmlns='DAV:'><version>"
                          + "<href>" + version + "</href></version></version-control>");
          response = requester.perform(request);
          checkResponseCode(response, 409);

          //create vcr in another WS
          clear(request);
          request.setMethod("VERSION-CONTROL");
          request.setURI(wspUriC + "/vcr");
          request.setBody("<version-control xmlns='DAV:'><version>"
                          + "<href>" + version + "</href></version></version-control>");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          //lookup workspace property
          propElt = this.getProperty(wspUriC + "/vcr", Consts.N_workspace);
          assertTrue(propElt != null);
          assertTrue(wspUriC.equals(this.getRelativeUriFromPropertyHrefs(propElt)));

          //check postcondition (DAV:new-version-controlled-resource)
          clear(request);
          request.setMethod("REPORT");
          request.setURI(wspUriC + "/vcr");
          checkMultistatus("<version-tree xmlns='DAV:'><prop><comment/><version-name/><successor-set/></prop></version-tree>",
            respfile);

          break;

        case 5:

          request.setMethod("MKCOL");
          request.setURI(wspUriC + "/col");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          //workspace internal move
          clear(request);
          request.setMethod("MOVE");
          request.setURI(wspUriC + "/vcr");
          request.setHeader("Destination", getBase() + wspUriC + "/col/vcr");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          //TODO: workspace internal bind

          break;

        case 6:

          //create 'normal' version controlled collection
          clear(request);
          request.setMethod("MKCOL");
          request.setURI(getLocalpart() + "/wsptestcol");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("PUT");
          request.setURI(getLocalpart() + "/wsptestcol/vcr");
          request.setBody("initial content of '/wsptestcol/vcr'");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          clear(request);
          request.setMethod("VERSION-CONTROL");
          request.setURI(getLocalpart() + "/wsptestcol/vcr");
          response = requester.perform(request);
          checkResponseCode(response, 201);

          // remember VHR URI for later
          vhr = getProperty(getLocalpart() + "/wsptestcol/vcr", Consts.N_version_history);
          vhrs.add(getRelativeUriFromPropertyHrefs(vhr));

          clear(request);
          request.setMethod("VERSION-CONTROL");
          request.setURI(getLocalpart() + "/wsptestcol");
          response = requester.perform(request);
          checkResponseCode(response, 201, 501);

          if (response.getStatus() == 201) {

            // remember VHR URI for later
            vhr = getProperty(getLocalpart() + "/wsptestcol", Consts.N_version_history);
            vhrs.add(getRelativeUriFromPropertyHrefs(vhr));

            propElt = this.getProperty(getLocalpart() + "/wsptestcol", Consts.N_checked_in);
            assertTrue(propElt != null);
            version = this.getRelativeUriFromPropertyHrefs(propElt);
            assertTrue(version != null);

            //create WS-controlled version controlled collection
            clear(request);
            request.setMethod("VERSION-CONTROL");
            request.setURI(wspUriA + "/vcc");
            request.setBody("<version-control xmlns='DAV:'><version>"
                            + "<href>" + version + "</href></version></version-control>");
            response = requester.perform(request);
            checkResponseCode(response, 201, 501);

            wcolSupported = (response.getStatus() == 201);

            if (wcolSupported) {
              Element checkedinProp = this.getProperty(wspUriA + "/vcc/vcr", Consts.N_checked_in);
              assertTrue("missing checked-in version of " + wspUriA + "/vcc/vcr", checkedinProp != null);
            }
          }

          break;

        case 7:

          if (wcolSupported) {

            clear(request);
            request.setMethod("CHECKOUT");
            request.setURI(wspUriA + "/vcc");
            response = requester.perform(request);
            checkResponseCode(response, 200);

            clear(request);
            request.setMethod("PUT");
            request.setURI(wspUriA + "/vcc/a");
            request.setBody("initial content of 'a'");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("PUT");
            request.setURI(wspUriA + "/vcc/b");
            request.setBody("initial content of 'b'");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("PUT");
            request.setURI(wspUriA + "/vcc/c");
            request.setBody("initial content of 'c'");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("VERSION-CONTROL");
            request.setURI(wspUriA + "/vcc/a");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            // remember VHR URI for later
            vhr = getProperty(wspUriA + "/vcc/a", Consts.N_version_history);
            vhrs.add(getRelativeUriFromPropertyHrefs(vhr));

            clear(request);
            request.setMethod("VERSION-CONTROL");
            request.setURI(wspUriA + "/vcc/b");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            // remember VHR URI for later
            vhr = getProperty(wspUriA + "/vcc/b", Consts.N_version_history);
            vhrs.add(getRelativeUriFromPropertyHrefs(vhr));

            clear(request);
            request.setMethod("CHECKIN");
            request.setURI(wspUriA + "/vcc");
            request.setBody("<checkin xmlns='DAV:'><keep-checked-out/></checkin>");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("MOVE");
            request.setURI(wspUriA + "/vcc/a");
            request.setHeader("Destination", getBase() + wspUriA + "/a");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("UNCHECKOUT");
            request.setURI(wspUriA + "/vcc");
            response = requester.perform(request);
            checkResponseCode(response, 200);

            Element r1 = getProperty(wspUriA + "/vcc/a", Consts.N_resource_id);
            Element r2 = getProperty(wspUriA + "/a", Consts.N_resource_id);
            assertTrue(DomSerializer.toString(r1, false).equals(DomSerializer.toString(r2, false)));
          }

          break;

        case 8:

          if (wcolSupported) {

            clear(request);
            request.setMethod("CHECKOUT");
            request.setURI(wspUriA + "/vcc");
            response = requester.perform(request);
            checkResponseCode(response, 200);

            clear(request);
            request.setMethod("PUT");
            request.setURI(getLocalpart() + "/justanothervcr");
            request.setBody("just another initial content");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("VERSION-CONTROL");
            request.setURI(getLocalpart() + "/justanothervcr");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            // remember VHR URI for later
            vhr = getProperty(getLocalpart() + "/justanothervcr", Consts.N_version_history);
            vhrs.add(getRelativeUriFromPropertyHrefs(vhr));

            propElt = this.getProperty(getLocalpart() + "/justanothervcr", Consts.N_checked_in);
            assertTrue(propElt != null);
            version = this.getRelativeUriFromPropertyHrefs(propElt);
            assertTrue(version != null);

            clear(request);
            request.setMethod("VERSION-CONTROL");
            request.setURI(wspUriA + "/vcc/d");
            request.setBody("<version-control xmlns='DAV:'><version>"
                            + "<href>" + version + "</href></version></version-control>");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("CHECKIN");
            request.setURI(wspUriA + "/vcc");
            request.setBody("<checkin xmlns='DAV:'><keep-checked-out/></checkin>");
            response = requester.perform(request);
            checkResponseCode(response, 201);

            clear(request);
            request.setMethod("DELETE");
            request.setURI(wspUriA + "/vcc/d");
            response = requester.perform(request);
            checkResponseCode(response, 204);

            clear(request);
            request.setMethod("UNCHECKOUT");
            request.setURI(wspUriA + "/vcc");
            response = requester.perform(request);
            checkResponseCode(response, 200);

            Element r1 = getProperty(wspUriA + "/vcc/d", Consts.N_resource_id);
            Element r2 = getProperty(getLocalpart() + "/justanothervcr", Consts.N_resource_id);
            assertTrue(!DomSerializer.toString(r1, false).equals(DomSerializer.toString(r2, false)));
          }

          break;

        default:

          if (wspUriA != null) {
            clear(request);
            request.setMethod("DELETE");
            request.setURI(wspUriA);
            response = requester.perform(request);
            checkResponseCode(response, 204, 404);
          }
          if (wspUriB != null) {
            clear(request);
            request.setMethod("DELETE");
            request.setURI(wspUriB);
            response = requester.perform(request);
            checkResponseCode(response, 204, 404);
          }
          if (wspUriC != null) {
            clear(request);
            request.setMethod("DELETE");
            request.setURI(wspUriC);
            response = requester.perform(request);
            checkResponseCode(response, 204, 404);
          }

          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/wsptestvcr");
          response = requester.perform(request);
          checkResponseCode(response, 204, 404);

          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/justanothervcr");
          response = requester.perform(request);
          checkResponseCode(response, 204, 404);

          clear(request);
          request.setMethod("DELETE");
          request.setURI(getLocalpart() + "/wsptestcol");
          response = requester.perform(request);
          checkResponseCode(response, 204, 404);
          
          // cleanup VHRs
          for (Iterator it = vhrs.iterator(); it.hasNext(); ) {
            String uri = it.next().toString();
            clear(request);
            request.setMethod("DELETE");
            request.setURI(uri);
            response = requester.perform(request);
            checkResponseCode(response, vhrPreserved() ? 204 : 404);
          }
          break;
      }
    }
    catch (Exception ex) {
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }

}
