package com.sapportals.wcm.protocol.webdav.server.webtests;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Iterator;
import java.util.List;

import junit.framework.TestCase;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import com.sapportals.wcm.protocol.webdav.Consts;
import com.sapportals.wcm.protocol.webdav.DomSerializer;
import com.sapportals.wcm.protocol.webdav.Name;
import com.sapportals.wcm.protocol.webdav.dompath.DomPath;
import com.sapportals.wcm.util.http.*;
import com.sapportals.wcm.util.uri.RID;
import com.sapportals.wcm.util.uri.URICodec;

/**
 * Abstract class. All implementing classes have access to a check of a
 * normalized multistatus responsecode. To compare the results the path to
 * the response-files is needed. It is stored in the system property
 * 'statusPath' and has the default
 * "c:/home/projects/wcm/testcases/JUnitTestResults/".
 *
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author stefan.eissing@greenbytes.de
 * @version $Id: TestMember.java,v 1.4 2004/10/26 10:12:14 sei Exp $
 */
public abstract class TestMember extends TestCase
{
  public static final String FAILURE_EXT = ".failure";
  public static final String UNTRANSFORMED_PREF = "raw-";
  public static final String PROP_NORMALIZE_PATH = "normalizerPath";
  public static final String PROP_STATUS_PATH = "statusPath";
  public static final String PROP_STATUS_FLAVOUR = "statusFlavour";

  protected static Xslt m_xslt = null;

  private String m_testName;

  protected int nr=0;
  protected IRequesterFactory m_factory;
  protected IRequester m_requester;
  protected HttpRequest m_request;
  protected String m_base;
  protected String m_localpart;
  protected String m_statusPath;
  protected String m_statusFlavour;
  protected String m_userid;

  private static String UA = System.getProperty("userAgent", null);

  /**
   * @param nr the current number of the desired test in the subclass.
   * @param s the name of the testclass.
   */
  public TestMember(Integer nr, String s)
  {
    super(s);
    m_testName = s;
    this.nr = nr.intValue();
  }

  public void clear(HttpRequest request) {
		request.clear();
		request.setHeader(Consts.SERVLET_TRACE_HEADER, m_testName);
		if (TestMember.UA != null)
			request.setHeader("user-agent", TestMember.UA);
	}

  public void init(IRequesterFactory factory, IRequester requester,
                   String base, String localpart, String userid) {
    m_factory = factory;
    m_base = base;
    m_localpart = localpart;
    m_requester = requester;
    m_request = new HttpRequest(m_localpart);
    clear(m_request);
    m_request.setURI(m_localpart);

    m_userid = userid;

    m_statusPath = System.getProperty(PROP_STATUS_PATH,
      "c:/projects/WCM3/testcases/JUnitTestResults/");
    m_statusFlavour = System.getProperty(PROP_STATUS_FLAVOUR, null);

    if (!m_statusPath.endsWith("/"))
      m_statusPath += "/";

    if (m_statusFlavour != null && !m_statusFlavour.endsWith("/")) {
      m_statusFlavour += "/";
    }

    String hostpath = m_base;
    if (hostpath.startsWith("http://")) {
      hostpath = hostpath.substring("http://".length());
    }
    RID r = RID.getRID(m_localpart, null);
    r = r.add("/web");
    r = r.add(hostpath);
    if (m_xslt == null) {
      String extbase = URICodec.EncodePath(r.toString());
      r = r.add(m_localpart);
      String extprefix = URICodec.EncodePath(r.toString());
      m_xslt = new Xslt (m_localpart, extbase, extprefix);
    }
  }

  private void writeToFile(String contents, String filename)
  {
    writeToFile(contents.getBytes(), filename);
  }

  private void writeToFile(byte[] contents, String filename)
  {
    System.err.println("Creating response file: " + filename);
    try{
      java.io.FileOutputStream f = new java.io.FileOutputStream(filename);
      f.write( contents );
      f.close();
    }catch(Exception e){
      System.err.println("could not write response to file: " + e.toString() );
    }
  }

  private String readFromFile (java.io.InputStream f)
  {
    if (f == null) return null;

    try{
      ByteArrayOutputStream bos = new ByteArrayOutputStream();
      byte[] buffer = new byte[64 * 1024];
      int len = 0;
      while ((len = f.read(buffer)) != -1) {
        bos.write(buffer, 0, len);
      }
      f.close();
      return bos.toString();
    }
    catch(Exception e){
      System.err.println("could not read response from file: " + e.toString() );
      return null;
    }
  }

  protected byte[] getBytes(InputStream is) {
    try {
      if (is == null)
        return null;

      byte[] buffer = new byte[64 * 1024];
      ByteArrayOutputStream bos = new ByteArrayOutputStream();
      int len;
      while ((len = is.read(buffer)) != -1) {
        bos.write(buffer, 0, len);
      }

      return bos.toByteArray();
    }
    catch (IOException e) {
      System.err.println("could not read response from file: " + e.toString() );
      return null;
    }
  }

  /**
   * Performs the desired m_request and expects a '207' (Multistatus) return code.
   * The returned document will first have all nodes matching the given
   * DomPath removed, then it will be normalized and compared against the
   * content of a file.
   * @param body the desired body of them_request.
   * @param expected the filename (without path) of the "to-be-compared" contents.
   * @param path the DomPath matching Nodes to be removed
   * @return the response of the request.
   */
  protected IResponse checkMultistatus (String body, String expected, DomPath remove)
    throws com.sapportals.wcm.WcmException
  {
    return checkStatusAndBody (207, body, expected, remove, true);
  }

  /**
   * Performs the desired m_request and expects a '207' (Multistatus) return code.
   * The return-value will be normalized and compared to the contents of a file.
   * @param body the desired body of them_request.
   * @param expected the filename (without path) of the "to-be-compared" contents.
   * @return the response of the request.
   */
  protected IResponse checkMultistatus (String body, String expected)
    throws com.sapportals.wcm.WcmException
  {
    return checkStatusAndBody (207, body, expected, null, true);
  }

  /**
   * Performs the desired m_request and expects a '207' (Multistatus) return code.
   * The return-value will be normalized and compared to the contents of a file.
   * @param body the desired body of them_request.
   * @param expected the filename (without path) of the "to-be-compared" contents.
   * @return the response of the request.
   */
  protected IResponse checkMultistatusUnsorted (String body, String expected)
    throws com.sapportals.wcm.WcmException
  {
    return checkStatusAndBody (207, body, expected, null, false);
  }

  private String normalizeLineEnds (String inp)
  {
    StringBuffer s = new StringBuffer("");

    for (int i = 0; i < inp.length(); i++)
    {
      if (inp.charAt(i) != '\r')
        s.append (inp.charAt (i));
    }

    return s.toString();
  }

  /**
   * Performs the desired m_request and expects a specific return code.
   * The return-value will be normalized and compared to the contents of a file.
   * @param body the desired body of them_request.
   * @param expected the filename (without path) of the "to-be-compared" contents.
   * @return the response of them_request.
   */
  protected IResponse checkStatusAndBody(int expectedStatus,
                                           String body,
                                           String expected)
    throws com.sapportals.wcm.WcmException
  {
    return checkStatusAndBody(expectedStatus, body, expected, null, true);
  }

  /**
   * Performs the desired m_request and expects a specific return code.
   * The return-value will be normalized and compared to the contents of a file.
   * @param body the desired body of them_request.
   * @param expected the filename (without path) of the "to-be-compared" contents.
   * @return the response of them_request.
   */
  protected IResponse checkStatusAndBody(int expectedStatus,
                                           String body,
                                           String expected,
                                           boolean sortResponses)
    throws com.sapportals.wcm.WcmException
  {
    return checkStatusAndBody(expectedStatus, body, expected, null, sortResponses);
  }

  /**
   * Performs the desired m_request and expects a specific return code.
   * The return-value will be normalized and compared to the contents of a file.
   * @param body the desired body of them_request.
   * @param expected the filename (without path) of the "to-be-compared" contents.
   * @param path DomPath matchind Nodes to be removed
   * @return the response of them_request.
   */
  protected IResponse checkStatusAndBody(int expectedStatus,
                                           String body,
                                           String expected,
                                           DomPath path,
                                           boolean sortResponses)
    throws com.sapportals.wcm.WcmException
  {
    m_request.setHeader("Content-Type", "text/xml");
    if (body != null) m_request.setBody(body);
    m_request.expectsResponseDocument(true);
    IResponse response = m_requester.perform(m_request);
    assertTrue ("expected " + expectedStatus + " return code, got " + response.getStatus(),
           response.getStatus() == expectedStatus);

    checkResponseDocument(response, expected, path, sortResponses);
    return response;
  }

  protected void checkResponseDocument(IResponse response,
                                       String expected,
                                       DomPath path)
  {
    checkResponseDocument(response, expected, path, true);
  }

  protected void checkResponseDocument(IResponse response,
                                       String expected,
                                       DomPath path,
                                       boolean sortResponses)
  {
    assertTrue ("expected result document, got null", response.hasDocument() );

    Document doc = response.getDocument();
    if (path != null) {
      path.deleteNodes(doc);
    }
    String s = m_xslt.transform(doc, sortResponses);
    assertTrue ("failed to normalize result document", s != null);
    s = normalizeLineEnds (s);

    // Determine the response file we are comparing against
    // First look in the flavour directory, if defined
    // Then look in the status directory
    // Create the response file if it is missing
    // Create a response.failure file if it differs
    // delete old .failure files when everything is ok
    //
    String t = null;
    String filePath = null;
    java.io.InputStream referenceResult = null;

    if (m_statusFlavour != null) {
      referenceResult = this.getClass().getResourceAsStream("results/" + m_statusFlavour + expected);
      filePath = m_statusPath + m_statusFlavour + expected;
      t = readFromFile (referenceResult);
    }
    if (t == null) {
// System.err.println ("results/" + expected);
      referenceResult = this.getClass().getResourceAsStream("results/" + expected);
      filePath = m_statusPath + expected;
      t = readFromFile(referenceResult);
    }

    if (t == null)
    {
      writeToFile(s, filePath);
      try {
        java.io.File toDelete = new java.io.File (filePath + FAILURE_EXT);
        toDelete.delete();
      }
      catch (Exception e) {}
    }
    else if (s != null)
    {
      t = normalizeLineEnds (t);
      if (!s.equals(t)) {
        writeToFile(s, filePath + FAILURE_EXT);
        writeToFile(t, filePath);
        writeToFile(DomSerializer.toString(doc, true), m_statusPath + UNTRANSFORMED_PREF + expected);
        assertTrue ("multistatus response differs from "+expected+": "+s, false);
      }
    }
  }

  protected void checkResponseContent(byte[] result, String expected)
  {
    // Determine the response file we are comparing against
    // First look in the flavour directory, if defined
    // Then look in the status directory
    // Create the response file if it is missing
    // Create a response.failure file if it differs
    // delete old .failure files when everything is ok
    //
    byte[] refResult = null;
    String filePath = null;
    java.io.InputStream referenceResult = null;

    if (m_statusFlavour != null) {
      referenceResult = this.getClass().getResourceAsStream("results/" + m_statusFlavour + expected);
      filePath = m_statusPath + m_statusFlavour + expected;
      refResult = getBytes (referenceResult);
    }

    if (refResult == null) {
      referenceResult = this.getClass().getResourceAsStream("results/" + expected);
      filePath = m_statusPath + expected;
      refResult = getBytes (referenceResult);
    }

    if (refResult == null)
    {
      writeToFile(result, filePath);
      try {
        java.io.File toDelete = new java.io.File (filePath + FAILURE_EXT);
        toDelete.delete();
      }
      catch (Exception e) {}
    }
    else if (result != null)
    {
      if (!same(result, refResult)) {
        writeToFile(result, filePath + FAILURE_EXT);
        writeToFile(refResult, filePath);
        assertTrue ("content of response differs from "+expected+": "+result, false);
      }
    }
  }

  protected void checkResponseCode(IResponse response, int expected) {
    checkResponseCode(response, expected, -1, -1);
  }

  protected void checkResponseCode(IResponse response, int expected1, int expected2) {
    checkResponseCode(response, expected1, expected2, -1);
  }

  protected void checkResponseCode(IResponse response, int expected1,
                                   int expected2, int expected3) {
    int code = response.getStatus();
    if (code != expected1 && code != expected2 && code != expected3) {
      StringBuffer sb = new StringBuffer();
      sb.append("expected response codes (").append(expected1);
      if (expected2 >= 0) {
        sb.append(", ").append(expected2);
        if (expected3 >= 0) {
          sb.append(", ").append(expected3);
        }
      }
      sb.append(") got ").append(code);
      Document doc = response.getDocument();
      if (doc != null) {
        sb.append(" with body\n");
        sb.append(DomSerializer.toString(doc, true));
      }
      assertTrue (sb.toString(), false);
    }
  }

  private boolean same(byte[] a1, byte[] a2) {
    if (a1 == a2) return true;
    if (a1 == null || a2 == null) return false;
    if (a1.length != a2.length) return false;
    for (int i = 0; i < a1.length; ++i) {
      if (a1[i] != a2[i] ) {
        return false;
      }
    }
    return true;
  }

  
  private static String toXml (Name name) {
    return "<" + name.getLocalName() + " xmlns='" + name.getNamespaceUri() + "'/>";
  }

  protected Element getProperty(String requestUri, Name property) throws com.sapportals.wcm.WcmException {

    clear(m_request);
    m_request.setMethod("PROPFIND");
    m_request.setURI(requestUri);
    m_request.setHeader("Depth", "0");
    m_request.setHeader("Translate", "f");
    m_request.expectsResponseDocument(true);
    m_request.setBody("<propfind xmlns='DAV:'><prop>" +
      "<" + property.getLocalName() + " xmlns='" + property.getNamespaceUri() + "'/>" +
      "</prop>" + toXml(Consts.N_SAP_include_hidden_flag) + toXml(Consts.N_SAP_include_protected_flag) +
      "</propfind>");
    IResponse response = m_requester.perform(m_request);
    checkResponseCode(response, 207);

    DomPath propsPath = new DomPath("/D:multistatus/D:response/D:propstat[contains(D:status,'200')]/D:prop/*", DomPath.DEFAULT_NAMESPACE_BINDINGS);
    DomPath propsNfPath = new DomPath("/D:multistatus/D:response/D:propstat[contains(D:status,'404')]/D:prop/*", DomPath.DEFAULT_NAMESPACE_BINDINGS);

    Document doc = response.getDocument();
    if (doc == null) return null;

    List props = propsPath.match(doc);
    for (Iterator it = props.iterator(); it.hasNext(); ) {
      Node n = (Node)it.next();
      if (property.equalsNode(n)) {
        return (Element) n;
      }
    }

    // 404ed?
    props = propsNfPath.match(doc);
    for (Iterator it = props.iterator(); it.hasNext(); ) {
      Node n = (Node)it.next();
      if (property.equalsNode(n)) {
        return null;
      }
    }

    fail("property " + property + " not reported with status 200 or 404: " + DomSerializer.toString(response.getDocument(), true));
    return null;
  }
  
  /**
   * to be implemented by the subclasses.
   */
  protected abstract void runTest();
}
