package com.sapportals.wcm.protocol.webdav.server.webtests;

import java.util.ArrayList;
import java.util.List;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import com.sapportals.wcm.WcmException;
import com.sapportals.wcm.protocol.webdav.DomElement;
import com.sapportals.wcm.protocol.webdav.DomSerializer;
import com.sapportals.wcm.protocol.webdav.dompath.DomPath;
import com.sapportals.wcm.util.http.HttpRequest;
import com.sapportals.wcm.util.http.IRequester;
import com.sapportals.wcm.util.http.IResponse;
import com.sapportals.wcm.util.uri.URICodec;

/**
 * Helper methods for creating test repository.
 *
 * <p>
 * Copyright (c) SAP AG 2001-2003
 * @author stefan.eissing@greenbytes.de
 * @version $Id: WebHelper.java,v 1.4 2005/03/04 14:00:39 sei Exp $
 */

public class WebHelper
{
  private static final DomPath DP_hrefs = new DomPath(
    "/D:multistatus/D:response/D:href", DomPath.DEFAULT_NAMESPACE_BINDINGS);
  private static final DomPath DP_locktoken_href = new DomPath(
    "../D:propstat/D:prop/D:lockdiscovery/D:activelock/D:locktoken/D:href", DomPath.DEFAULT_NAMESPACE_BINDINGS);

  private IRequester m_requester;
  private HttpRequest m_request;
  private URICodec m_codec;
  private String m_host;
  private String m_localpart;

  public WebHelper (IRequester requester, String host, String localpart)
  {
    m_requester = requester;
    m_request = new HttpRequest();
    m_codec = new URICodec();
    m_host = host;
    if (localpart != null && localpart.endsWith("\\"))
      localpart = localpart.substring(0, localpart.length()-1 );
    m_localpart = localpart;
  }

  public void clean() throws WcmException {
    
    String ua = System.getProperty("userAgent", null);

    m_request.clear();
    m_request.setMethod("PROPFIND");
    m_request.setURI(m_localpart);
    m_request.setHeader("Depth", "1");
    if (ua != null) m_request.setHeader("user-agent", ua);
    m_request.expectsResponseDocument(true);
    IResponse response = m_requester.perform(m_request);

    if (response.getStatus() != 207) {
      throw new WcmException ("PROPFIND returned " + response.getStatus() +
      " instead of 207");
    }

    Document doc = response.getDocument();
    if (doc == null) {
      throw new WcmException ("PROPFIND did not return (well-formed) response body");
    }

    List hrefs = new ArrayList();
    List locks = new ArrayList();
    getURLs(doc, "/", hrefs, locks);
    for (int i = 0; i < hrefs.size(); ++i) {
      String uri = (String)hrefs.get(i);
      String lock = (String)locks.get(i);
      if (lock.length() > 0) {
        m_request.clear();
        m_request.setMethod("UNLOCK");
        m_request.setURI(m_localpart + uri);
        m_request.setHeader("Lock-Token", "<"+lock+">");
        response = m_requester.perform(m_request);
      }
      m_request.clear();
      m_request.setMethod("DELETE");
      m_request.setHeader("Apply-To-Redirect-Ref", "");
      m_request.setURI(m_localpart + uri);
      m_request.setHeader("Depth", "Infinity");
      response = m_requester.perform(m_request);
      if (response.getStatus() != 204 && response.getStatus() != 404) {
        Document ms = response.getDocument();
        throw new WcmException("removing "+uri
                               +" got status "+response.getStatus()
                               + ((ms != null)? DomSerializer.toString(ms, true) : ""));
      }
    }
  }

  public void makeDirStructure() throws WcmException
  {
    String[] uris = new String[] {
    };

    for (int i = 0; i < uris.length; ++i) {
      String uri = uris[i];
      m_request.clear();
      m_request.setMethod("MKCOL");
      m_request.setURI(m_codec.encode(m_localpart + uri));
      IResponse response = m_requester.perform(m_request);
      int status = response.getStatus();
      if (status != 201) {
        throw new WcmException("mkcol "+uri+" returned "+status);
      }
    }
  }

  private void copy(String source, String destURI, String mimetype, String encoding, String language) throws Exception
  {
    java.io.InputStream fis = this.getClass().getResourceAsStream ("files/" + source);

    int content_length = 0;
    int len;
    byte[] buffer = new byte[32 * 1024];
    while ((len = fis.read(buffer)) != -1) {
      content_length += len;
    }
    fis.close();

    fis = this.getClass().getResourceAsStream ("files/" + source);
    try {
      m_request.clear();
      m_request.setMethod("PUT");
      m_request.setURI(m_codec.encode(m_localpart + destURI));
      if (mimetype != null) {
        if (encoding == null) {
          m_request.setHeader ("Content-Type", mimetype);
        }
        else {
          m_request.setHeader ("Content-Type", mimetype + "; charset=" + encoding);
        }
      }
      if (language != null) m_request.setHeader ("Content-Language", language);
      // m_request.setHeader("Content-Length", String.valueOf(content_length));
      m_request.setBody(fis);
      IResponse response = m_requester.perform(m_request);
      int status = response.getStatus();
      if (status != 201) {
        throw new WcmException("put on "+destURI+" returned "+status);
      }
    }
    finally {
      fis.close();
    }
  }

  public void setup() throws Exception
  {
    clean();
    makeDirStructure();
    copy ("arabic.copyme",        "/arabic.html", "text/html", null, null);
    copy ("china.copyme",         "/china.html", "text/html", null, null);
    copy ("cnn.copyme",           "/cnn.html", "text/html", null, null);
    copy ("greenbytes.copyme",    "/greenbytes.html", "text/html", null, null);
    copy ("japanese.copyme",      "/japanese.html", "text/html", null, null);
    copy ("metaextract.copyme",   "/metaextract.html", "text/html", null, null);
    copy ("selfhtml.copyme",      "/selfhtml.html", "text/html", null, null);
    copy ("spiegel.copyme",       "/spiegel.html", "text/html", null, null);
    copy ("thainews.copyme",      "/thainews.html", "text/html", null, null);
    copy ("sql92.copyme",         "/sql92.html", "text/html", null, null);
    copy ("sap-error1.copyme",    "/sap-error1.html", "text/html", null, null);
    copy ("location.copyme",      "/location.html", "text/html", null, null);
    copy ("aliases.copyme",       "/aliases.html", "text/html", null, null);
    copy ("synopsys.copyme",       "/synopsys.html", "text/html", null, null);
    copy ("servlet2_2-spec.pdf.copyme", "/servlet2_2.pdf", "application/pdf", null, null);
    copy ("langspec-2.0.html.zip.copyme", "/langspec.zip", "application/zip", null, null);
  }


  private void getURLs(Document doc, String ignore, List result, List resultTokens) {
    Node[] hrefs = DP_hrefs.matchNodes(doc);
    for (int i = 0; i < hrefs.length; ++i) {
      Element href = (Element)hrefs[i];
      String url = DomElement.getText(href);
      if (url.startsWith(m_host)) {
        url = url.substring(m_host.length());
      }
      if (url.startsWith(m_localpart)) {
        url = url.substring(m_localpart.length());
      }
      if (url.length() > 0 && !url.equals(ignore)) {
        result.add(url);
        String tokenString = "";

        Node[] tokens = DP_locktoken_href.matchNodes(href);
        for (int j = 0; j < tokens.length; ++j) {
          Element locktoken = (Element)tokens[j];
          tokenString = DomElement.getText(locktoken);
        }

        resultTokens.add(tokenString);
      }
    }
    System.err.println("hrefs+tokens: "+result+" "+resultTokens);
  }
}
