package com.sapportals.wcm.protocol.webdav.server.webtests;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import com.sapportals.wcm.protocol.webdav.Consts;
import com.sapportals.wcm.protocol.webdav.DomElement;
import com.sapportals.wcm.protocol.webdav.DomSerializer;
import com.sapportals.wcm.protocol.webdav.Name;
import com.sapportals.wcm.protocol.webdav.dompath.DomPath;
import com.sapportals.wcm.util.http.IResponse;

/**
 * Various tests for PROPFIND method.
 * <p>
 * Copyright (c) SAP AG 2001-2003
 * @author stefan.eissing@greenbytes.de
 * @author julian.reschke@greenbytes.de
 * @version $Id: WebPropfindTest.java,v 1.3 2005/03/02 16:32:48 sei Exp $
 */

public class WebPropfindTest extends TestMember
{
  private static java.text.DecimalFormat m_threeDigits = new
    java.text.DecimalFormat ("000");

  private static final String[] desc = new String[]
  {
    "00 PROPFIND with /webcrawl/arabic.html",
    "01 PROPFIND with /webcrawl/china.html",
    "02 PROPFIND with /webcrawl/cnn.html",
    "03 PROPFIND with /webcrawl/greenbytes.html",
    "04 PROPFIND with /webcrawl/japanese.html",
    "05 PROPFIND with /webcrawl/selfhtml.html",
    "06 PROPFIND with /webcrawl/spiegel.html",
    "07 PROPFIND with /webcrawl/thainews.html",
    "08 PROPFIND with /webcrawl/sql92.html",
    "09 PROPFIND root with fixed systems (/webfix)",
    "10 PROPFIND with /sgb/ (https://www.greenbytes.de)",
    "11 PROPFIND with /sgb-proxy/ (https://www.greenbytes.de with proxy)",
    "12 PROPFIND root with dynamic systems (/web)",
    "13 PROPFIND with /webcrawl/location.html (JavaScript)",
    "14 extract DAV:getlastmodified from /webcrawl/metaextract.html",
    "15 extract dcmi:creator from /webcrawl/metaextract.html",
    "16 extract dcmi:language from /webcrawl/metaextract.html",
    "17 check LINK/@href extraction /webcrawl/metaextract.html",
    "18 check links from aliases.html",
    "19 PROPFIND with /webcrawl/SPIEGEL.html",
    "20 PROPFIND case-insensitive lookup (/webFix2)",
  };

  public static int NrOfTests = desc.length;

  private static final DomPath DP_href;

  static {
    Map nsmap = new HashMap(DomPath.DEFAULT_NAMESPACE_BINDINGS);
    nsmap.put("sap", Consts.NS_SAP);
    DP_href = new DomPath("/D:multistatus/D:response/D:href", nsmap);
  }

  private final static String ALLPROPS = "<propfind xmlns='DAV:'><allprop /></propfind>";
  private final static String DISPNAME = "<propfind xmlns='DAV:'><prop><displayname/></prop></propfind>";

  public WebPropfindTest(Integer nr)
  {
    super(nr, desc[nr.intValue()]);
  }

  protected void runTest()
  {

    try
    {
      String respfile = "WEBPROPFIND-" + m_threeDigits.format (nr) + ".txt";
      IResponse response;

      clear(m_request);
      m_request.setMethod("PROPFIND");
      m_request.setHeader("Depth", "1");

      switch (nr)
      {
        case 0:
          m_request.setURI(m_localpart + "/webcrawl/arabic.html");
          super.checkMultistatus(ALLPROPS, respfile);
          break;

        case 1:
          m_request.setURI(m_localpart + "/webcrawl/china.html");
          super.checkMultistatus(ALLPROPS, respfile);
          break;

        case 2:
          m_request.setURI(m_localpart + "/webcrawl/cnn.html");
          super.checkMultistatus(ALLPROPS, respfile);
          break;

        case 3:
          m_request.setURI(m_localpart + "/webcrawl/greenbytes.html");
          super.checkMultistatus(ALLPROPS, respfile);
          break;

        case 4:
          m_request.setURI(m_localpart + "/webcrawl/japanese.html");
          super.checkMultistatus(ALLPROPS, respfile);
          break;

        case 5:
          m_request.setURI(m_localpart + "/webcrawl/selfhtml.html");
          super.checkMultistatus(ALLPROPS, respfile);
          break;

        case 6:
          m_request.setURI(m_localpart + "/webcrawl/spiegel.html");
          super.checkMultistatus(ALLPROPS, respfile);
          break;

        case 7:
          m_request.setURI(m_localpart + "/webcrawl/thainews.html");
          super.checkMultistatus(ALLPROPS, respfile);
          break;

        case 8:
          m_request.setURI(m_localpart + "/webcrawl/sql92.html");
          super.checkMultistatus(ALLPROPS, respfile);
          break;

        case 9:
          m_request.setURI(m_localpart + "/webfix");
          m_request.setHeader("cache-control", "no-cache");
          super.checkMultistatus(DISPNAME, respfile);
          break;

        case 10: {
            Element prop = super.getProperty(m_localpart + "/sgb", new Name(Consts.NS_SAP, "embedded-links"));
            assertTrue("page does not contain any links: /sgb", (prop != null) && ! DomElement.getChildren(prop, Consts.N_href).isEmpty());
          }
          break;

        case 11: {
            Element prop = super.getProperty(m_localpart + "/sgb-proxy", new Name(Consts.NS_SAP, "embedded-links"));
            assertTrue("page does not contain any links: /sgb-proxy", (prop != null) && ! DomElement.getChildren(prop, Consts.N_href).isEmpty());
          }
          break;

        case 12: {
          m_request.setURI(m_localpart + "/web");
          m_request.expectsResponseDocument(true);
          response = m_requester.perform(m_request);
          super.checkResponseCode(response, 207);
          Document doc = response.getDocument();
          assertTrue("responses did not contain document", doc != null);
          Node[] nodes = DP_href.matchNodes(doc);
          assertTrue("responses did not contain at least 3 hrefs: "+DomSerializer.toString(doc, true), nodes.length >= 3);

          List missing = new ArrayList(17);
          missing.add("onvista");
          missing.add("cvs.greenbytes.de");
          for (int i = 0; i < nodes.length; ++i) {
            String ref = DomElement.getText(nodes[i]);
            if (ref != null) {
              if (ref.endsWith("onvista")) {
                missing.remove("onvista");
              }
              else if (ref.endsWith("cvs.greenbytes.de")) {
                missing.remove("cvs.greenbytes.de");
              }
            }
          }
          assertTrue("missing children: "+missing, missing.isEmpty());
          break;
        }

        case 13:
          m_request.setURI(m_localpart + "/webcrawl/location.html");
          super.checkMultistatus(ALLPROPS, respfile);
          break;
          
        case 14: {
            Element prop = super.getProperty(m_localpart + "/webcrawl/metaextract.html", Consts.N_getlastmodified);
            String lastmod = DomElement.getText(prop);
            assertTrue("failed to extract lastmodified date", "Mon, 28 Jul 2003 08:50:34 GMT".equals(lastmod));
          }
          break;

        case 15: {
            Name pn = new Name("http://purl.org/dc/elements/1.1/", "creator");
            DomPath dp = new DomPath("*", DomPath.DEFAULT_NAMESPACE_BINDINGS);
            Element prop = super.getProperty(m_localpart + "/webcrawl/metaextract.html", pn);
            assertTrue(prop != null);
            List children = dp.match(prop);
            assertTrue(children.size() == 2);
          }
          break;

        case 16: {
            Name pn = new Name("http://purl.org/dc/elements/1.1/", "language");
            DomPath dp = new DomPath("*", DomPath.DEFAULT_NAMESPACE_BINDINGS);
            Element prop = super.getProperty(m_localpart + "/webcrawl/metaextract.html", pn);
            assertTrue(prop != null);
            List children = dp.match(prop);
            assertTrue(children.size() == 1);
            assertTrue(prop.getAttributeNS("http://sapportals.com/xmlns/cm/rfc2731", "scheme.0").equals("rfc1766"));
          }
          break;

        case 17: {
            Name pn = new Name(Consts.NS_SAP, "embedded-links");
            Element prop = super.getProperty(m_localpart + "/webcrawl/metaextract.html", pn);
            assertTrue(prop != null);
            String s = DomSerializer.toString(prop, false);
            assertTrue("must contain copyright link", s.indexOf("copyright.asp") >= 0);
            assertTrue("must not contain DC namespace link", s.indexOf("http://purl.org/dc/elements/1.1/") < 0);
          }
          break;

        case 18:
          m_request.setURI(m_localpart + "/webcrawl/aliases.html");
          super.checkMultistatus(ALLPROPS, respfile);
          break;

        case 19:
          m_request.setURI(m_localpart + "/webcrawl/SPIEGEL.html");
          super.checkMultistatus(ALLPROPS, respfile);
          break;

        case 20:
          m_request.setURI(m_localpart + "/webFix2");
          m_request.setHeader("cache-control", "no-cache");
          super.checkMultistatus(DISPNAME, respfile);
          break;

      }

    }
    catch(Exception ex) {
      ex.printStackTrace(System.err);
      java.io.StringWriter sw = new java.io.StringWriter();
      ex.printStackTrace(new java.io.PrintWriter(sw));
      fail(sw.getBuffer().toString());
    }
  }
}
