package com.sapportals.wcm.protocol.webdav.sets;

import java.util.NoSuchElementException;

/**
 * Array storage of int set members
 * <p>
 * Copyright (c) SAP AG 2003
 * @author stefan.eissing@greenbytes.de
 * @version $Id: ArrayIntSet.java,v 1.1.1.1 2004/02/18 13:00:23 sync Exp $
 */
public final class ArrayIntSet implements IntSet {
  
  private static final int DEFAULT_SIZE = 256;
  
  public static ArrayIntSet getInstance() {
    return new ArrayIntSet(DEFAULT_SIZE);
  }
  
  public static ArrayIntSet getInstance(int capacity) {
    return new ArrayIntSet(capacity);
  }
  
  public static ArrayIntSet getInstance(IntSet set) {
    ArrayIntSet s = new ArrayIntSet(set.size());
    for (IntIterator i = set.iterator(); i.hasNext(); /**/) {
      s.add(i.next());
    }
    return s;
  }
  
  private int members[];
  private int size;
  
  private ArrayIntSet() {
    this(DEFAULT_SIZE);
  }
  
  private ArrayIntSet(int capacity) {
    this.members = new int[capacity];
    this.size = 0;
  }
  
  public boolean contains(int i) {
    return (getIndex(i, 0, this.size) >= 0);
  }
  
  public boolean add(int i) {
    int index = getIndex(i, 0, this.size);
    if (index < 0) {
      addAt(-(index+1), i);
      return true;
    }
    return false;
  }
  
  public boolean remove(int i) {
    int index = getIndex(i, 0, this.size);
    if (index >= 0) {
      removeAt(index);
      return true;
    }
    return false;
  }
  
  /**
   * Adds all integers [n, m[, starting from n up to,
   * but not including m to the set
   * @param n lower bound to add
   * @param m upper bound (exclusive) to add
   */
  public void addInterval(int n, int m) {
    for (int i = n; i < m; ++i) {
      add(i);
    }
  }
  
  /**
   * Removes all integers [n, m[, starting from n up to,
   * but not including m from the set
   * @param n lower bound to remove
   * @param m upper bound (exclusive) to remove
   */
  public void removeInterval(int n, int m) {
    for (int i = n; i < m; ++i) {
      remove(i);
    }
  }
  
  public int size() {
    return this.size;
  }
  
  public boolean isEmpty() {
    return (this.size == 0);
  }
  
  public IntIterator iterator() {
    return new MyIterator();
  }
  
  public void clear() {
    this.size = 0;
    shrink();
  }
  
  public String toString() {
    StringBuffer sb = new StringBuffer();
    sb.append("IntSet[");
    for (int i = 0; i < this.size; ++i) {
      if (i != 0) sb.append(", ");
      sb.append(this.members[i]);
    }
    sb.append("]");
    return sb.toString();
  }
  
  public float getStorageEfficiency() {
    if (this.size > 256) {
      int len = this.members[this.size-1] -  this.members[0];
      if (len >= this.size) {
        return 1.0f - (((float)this.size) / len);
      }
      else {
        return 0.0f;
      }
    }
    return 1.0f;
  }
  
  //-------------------- private -----------------------------------------------
  
  /**
   * Get index of value i in interval [min, max[
   */
  private int getIndex(int i, int min, int max) {
    if (min >= max) {
      return (- min - 1);
    }
    int medium = min + ((max - min) / 2);
    int diff = this.members[medium] - i;
    if (diff == 0) {
      return medium;
    }
    else if (diff > 0) {
      return getIndex(i, min, medium);
    }
    else {
      return getIndex(i, medium + 1, max);
    }
  }
  
  private void removeAt(int index) {
    int moveLen = this.size - index - 1;
    if (moveLen > 0) {
      System.arraycopy(this.members, index+1, this.members, index, moveLen);
    }
    --this.size;
    shrink();
  }
  
  private void addAt(int index, int i) {
    int newSize = this.size + 1;
    if (newSize >= this.members.length) {
      grow();
    }
    int moveLen = this.size - index;
    if (moveLen > 0) {
      System.arraycopy(this.members, index, this.members, index+1, moveLen);
    }
    this.members[index] = i;
    this.size = newSize;
  }
  
  private void grow() {
    int nm[] = new int[this.members.length + DEFAULT_SIZE];
    System.arraycopy(this.members, 0, nm, 0, this.size);
    this.members = nm;
  }
  
  private void shrink() {
    if (((this.members.length / 2) - DEFAULT_SIZE) > this.size) {
      int nm[] = new int[this.size + DEFAULT_SIZE];
      System.arraycopy(this.members, 0, nm, 0, this.size);
      this.members = nm;
    }
  }

  private final class MyIterator implements IntIterator {

    private int offset;
        
		public boolean hasNext() {
      return (this.offset < ArrayIntSet.this.size);
		}

		public int next() {
      if (this.offset >= 0 && this.offset < ArrayIntSet.this.size) {
        return ArrayIntSet.this.members[this.offset++];
      }
      throw new NoSuchElementException();
		}

  }
}
