package com.sapportals.wcm.protocol.webdav.sets;

import java.util.NoSuchElementException;

/**
 * Interval storage of int set members
 * <p>
 * Copyright (c) SAP AG 2003
 * @author stefan.eissing@greenbytes.de
 * @version $Id: IntervalIntSet.java,v 1.1.1.1 2004/02/18 13:00:23 sync Exp $
 */
public class IntervalIntSet implements IntSet {
  
  public static IntervalIntSet getInstance() {
    return new IntervalIntSet();
  }
  
  public static IntervalIntSet getInstance(int n, int m) {
    return new IntervalIntSet(n, m);
  }
  
  public static IntervalIntSet getInstance(IntSet set) {
    IntervalIntSet s = new IntervalIntSet();
    for (IntIterator i = set.iterator(); i.hasNext(); /**/) {
      s.add(i.next());
    }
    return s;
  }
  
  private int lower;
  private int upper;
  private IntSet missing;
  
  private IntervalIntSet() {
    this(0, 0);
  }
  
  private IntervalIntSet(int n, int m) {
    this.lower = n;
    this.upper = m;
    this.missing = ArrayIntSet.getInstance();
  }

	public boolean contains(int i) {
    if (i < this.lower) {
      return false;
    }
    else if (i >= this.upper) {
      return false;
    }
    else {
      return !this.missing.contains(i);
    }
	}

	public boolean add(int i) {
    if (i < this.lower) {
      if (this.lower == this.upper) {
        // empty set
        this.lower = i;
        this.upper = i + 1;
      }
      else {
        for (int j = i+1; j < this.lower; ++j) {
          this.missing.add(j);
        }
        this.lower = i;
      }
      return true;
    }
    else if (i >= this.upper) {
      if (this.lower == this.upper) {
        // empty set
        this.lower = i;
        this.upper = i + 1;
      }
      else {
        for (int j = this.upper; j < i; ++j) {
          this.missing.add(j);
        }
        this.upper = i + 1;
      }
      return true;
    }
    else {
      return ! this.missing.remove(i);
    }
	}

	public boolean remove(int i) {
    if (i < this.lower) {
      return false;
    }
    else if (i >= this.upper) {
      return false;
    }
    else if (i + 1 == this.upper) {
      this.upper--;
      boolean retval = ! this.missing.remove(this.upper);
      while (this.missing.contains(this.upper-1)) {
        --this.upper;
        this.missing.remove(this.upper);
      }
      return retval;
    }
    else if (i == this.lower) {
      ++this.lower;
      boolean retval = ! this.missing.remove(i);
      while (this.missing.contains(this.lower)) {
        this.missing.remove(this.lower);
        ++this.lower;
      }
      return retval;
    }
    else {
      return ! this.missing.add(i);
    }
	}

	public void addInterval(int n, int m) {
    for (int i = n; i < m; ++i) {
      this.add(i);    
    }
	}

	public void removeInterval(int n, int m) {
    if (m > this.lower && n < this.upper) {
      // it intersects
      if (n < this.lower) {
        n = this.lower;
      }
      if (m > this.upper) {
        m = this.upper;
      }
      for (int i = n; i < m; ++i) {
        this.remove(i);
      }
    }
	}

	public int size() {
    return (this.upper - this.lower) - this.missing.size();
	}

	public boolean isEmpty() {
    return size() == 0;
	}

	public IntIterator iterator() {
    return new MyIntIterator();
	}

	public void clear() {
    this.missing.clear();
    this.lower = this.upper = 0;
	}

  public String toString() {
    StringBuffer sb = new StringBuffer();
    sb.append("IntervalIntSet[");
    if (!isEmpty()) {
      sb.append(this.lower).append(", ").append(this.upper-1);
      if (this.missing.size() > 0) {
        sb.append(" - except ");
        sb.append(this.missing);
      }
    }
    sb.append("]");
    return sb.toString();
  }
  
  public float getStorageEfficiency() {
    if (this.missing.size() > 256) {
      int len = (this.upper - this.lower);
      if (len > this.missing.size()) {
        return 1.0f - (((float)this.missing.size()) / len);
      }
      else {
        return 0.0f;
      }
    }
    return 1.0f;
  }

  private final class MyIntIterator implements IntIterator {
    
    private boolean needNext;
    private boolean finished;
    private int nextValue;
        
    public MyIntIterator() {
      this.needNext = true;
      this.nextValue = IntervalIntSet.this.lower - 1;
    }
    
    public boolean hasNext() {
      calcValue();
      return (!this.finished);
    }
    
    public int next() {
      calcValue();
      if (this.finished) {
        throw new NoSuchElementException();
      }
      this.needNext = true;
      return this.nextValue;
    }
    
    private void calcValue() {
      while (!this.finished && this.needNext) {
        ++this.nextValue;
        if (this.nextValue >= IntervalIntSet.this.upper) {
          this.finished = true;
        }
        else {
          this.needNext = ! IntervalIntSet.this.contains(this.nextValue);
        }
      }
    }
  }
}
