/*
 * Copyright (c) 2003 by SAP AG. All Rights Reserved.
 *
 * SAP, mySAP, mySAP.com and other SAP products and
 * services mentioned herein as well as their respective
 * logos are trademarks or registered trademarks of
 * SAP AG in Germany and in several other countries all
 * over the world. MarketSet and Enterprise Buyer are
 * jointly owned trademarks of SAP AG and Commerce One.
 * All other product and service names mentioned are
 * trademarks of their respective companies.
 *
 * @version $Id: JDBCDatabaseConnectionUncached.java,v 1.4 2004/05/21 12:09:54 jre Exp $
 */

package com.sapportals.wcm.util.acl.jdbc;
import java.sql.*;
import java.util.*;

import com.sapportals.portal.security.usermanagement.*;
import com.sapportals.wcm.util.acl.*;
import com.sapportals.wcm.util.cache.CacheException;
import com.sapportals.wcm.util.cache.ICache;
import com.sapportals.wcm.util.cache.ICacheEntry;
import com.sapportals.wcm.util.jdbc.connectionpool.JDBCConnectionPool;
import com.sapportals.wcm.util.jdbc.connectionpool.JDBCConnectionPoolManager;
import com.sapportals.wcm.util.jdbc.escape.Escape;
import com.sapportals.wcm.util.logging.LoggingFormatter;
import com.sapportals.wcm.util.usermanagement.WPUMFactory;

/**
 * Uncached access to the ACL database persistence.
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author paul.goetz@sap.com
 */
public final class JDBCDatabaseConnectionUncached {

  // table names
  protected final static String ACL_TABLE_NAME = "wcm_app_acl_acl";
  private final static String ACL_ENTRY_TABLE_NAME = "wcm_app_acl_acl_entry";
  private final static String OWNER_TABLE_NAME = "wcm_app_acl_owner";
  private final static String PERMISSION_TABLE_NAME = "wcm_app_acl_perm";
  private final static String SUPPORTED_PERMISSION_TABLE_NAME = "wcm_app_acl_supp_perm";
  private final static String PERMISSION_MEMBER_TABLE_NAME = "wcm_app_acl_perm_member";

  // columns and their attributes

  private final static String ACL_ID_COL = "RID";

  private final static String ACL_RIDSTART_COL = "RIDStart";
  private final static int RID_START_LENGTH = 900;
  private final static int RID_MAX_ESCAPED_LENGTH = 4000;

  protected final static String ACL_HASH_COL = "Hash";

  private final static String CHANGE_LEVEL_COL = "change_level";

  private final static String OBJECT_TYPE_COL = "object_type";
  private final static int OBJECT_TYPE_LENGTH = 127;

  private final static String ACL_ENTRY_ID_COL = "ID";

  private final static String PRINCIPAL_NAME_COL = "principal_name";
  private final static int PRINCIPAL_NAME_LENGTH = 255;

  private final static String PRINCIPAL_TYPE_COL = "principal_type";
  private final static int PRINCIPAL_TYPE_TAG_USER = 100;
  private final static int PRINCIPAL_TYPE_TAG_GROUP = 101;
  private final static int PRINCIPAL_TYPE_TAG_ROLE = 102;

  private final static String LOCKING_USER_COL = "locking_user";
  private final static int LOCKING_USER_LENGTH = 255;

  private final static String SORT_INDEX_COL = "sort_index";

  private final static String IS_NEGATIVE_COL = "is_negative";

  private final static String IS_PROPAGATED_COL = "is_propagated";

  private final static String PERMISSION_NAME_COL = "permission_name";
  private final static int PERMISSION_NAME_LENGTH = 127;

  private final static String IS_PREDEFINED_COL = "is_predefined";

  private final static String PERMISSION_MEMBER_COL = "member_name";
  private final static int PERMISSION_MEMBER_LENGTH = 127;

  private final static String SEPARATOR = "-";// for the permissions cache
  private final static String KEY_PERMISSION_PREFIX = "p";// key in the permission cache
  private final static String KEY_SUPPORTED_PERMISSIONS_PREFIX = "s";// key in the permission cache

  private static com.sap.tc.logging.Location s_log = com.sap.tc.logging.Location.getLocation(com.sapportals.wcm.util.acl.jdbc.JDBCDatabaseConnectionUncached.class);

  private final String m_applicationID;
  private final String m_managerID;
  private final String m_poolID;
  private final int m_owner;
  private final ICache m_permissionCache;// it's better to cache permissions on this level, because they are heavily used to create ACEs
  private final String m_ApplicationAndManagerSelectionCondition;
  private final boolean generateHashColumns = true;
  private boolean hasOldColumnPopulation = true;

  private Connection m_main_connection = null;

  /**
   * TBD: Description of the class.
   */
  class AclRec {
    private String m_externalAclID;
    private String m_objectTypeName;
    private long m_changeLevel;

    public AclRec(String externalAclID, String objectTypeName, long changeLevel) {
      m_externalAclID = externalAclID;
      m_objectTypeName = objectTypeName;
      m_changeLevel = changeLevel;
    }

    public String getExternalAclID() {
      return m_externalAclID;
    }

    public String getObjectTypeName() {
      return m_objectTypeName;
    }

    public long getChangeLevel() {
      return m_changeLevel;
    }
  }


  /**
   * Construct
   *
   * @param applicationID TBD: Description of the incoming method parameter
   * @param managerID TBD: Description of the incoming method parameter
   * @param poolID TBD: Description of the incoming method parameter
   * @param permissionCache TBD: Description of the incoming method parameter
   * @exception AclLoadClassException Exception raised in failure situation
   * @exception AclPersistenceException Exception raised in failure situation
   */
  public JDBCDatabaseConnectionUncached(String applicationID,
    String managerID,
    String poolID,
    ICache permissionCache)
    throws AclLoadClassException, AclPersistenceException {

    if (applicationID == null || managerID == null || poolID == null) {
      throw new java.lang.IllegalArgumentException();
    }

    m_applicationID = applicationID;
    m_managerID = managerID;
    m_poolID = poolID;

    if (s_log.beDebug()) {
      s_log.debugT("JDBCDatabaseConnectionUncached(158)", "applicationID " + applicationID + ", managerID " + managerID + ", poolID " + poolID + ", permissionCache " + permissionCache.getID());
    }

    m_permissionCache = permissionCache;

    try {
      int owner = 0;
      String selCondition = "1=0";
      m_main_connection = openConnection();

      PreparedStatement st = m_main_connection.prepareStatement("SELECT ID FROM wcm_app_acl_manager_appl WHERE app=? AND manager=?");
      st.setString(1, trim(Escape.escapeSqlString(m_applicationID, 127)));
      st.setString(2, trim(Escape.escapeSqlString(m_managerID, 127)));
      ResultSet rs = st.executeQuery();
      if (rs.next()) {
        selCondition = "OwnerID=?";
        owner = rs.getInt(1);
      }
      else {
        rs.close();
        rs = null;
        st.close();
        st = null;
        st = m_main_connection.prepareStatement("INSERT INTO wcm_app_acl_manager_appl (app, manager) VALUES (?,?)");
        st.setString(1, trim(Escape.escapeSqlString(m_applicationID, 127)));
        st.setString(2, trim(Escape.escapeSqlString(m_managerID, 127)));
        st.executeUpdate();
        st.close();
        st = null;

        st = m_main_connection.prepareStatement("SELECT ID FROM wcm_app_acl_manager_appl WHERE app=? AND manager=?");
        st.setString(1, trim(Escape.escapeSqlString(m_applicationID, 127)));
        st.setString(2, trim(Escape.escapeSqlString(m_managerID, 127)));
        rs = st.executeQuery();
        if (rs.next()) {
          selCondition = "OwnerID=?";
          owner = rs.getInt(1);
        }
        else {
          throw new AclPersistenceException("Can't insert entry for manager='" + managerID + "' and application '" + applicationID
             + " mapped to '" + trim(Escape.escapeSqlString(m_managerID, 127))
             + "'/'" + trim(Escape.escapeSqlString(m_applicationID, 127))
             + "' in poolID '" + poolID + "'");
        }
      }
      rs.close();
      rs = null;
      st.close();
      st = null;

      m_owner = owner;
      m_ApplicationAndManagerSelectionCondition = selCondition;

      ensureHash(m_main_connection, "wcm_app_acl_acl");
      ensureSegments(m_main_connection, "wcm_app_acl_acl");
      ensureHash(m_main_connection, "wcm_app_acl_acl_entry");
      ensureHash(m_main_connection, "wcm_app_acl_owner");

      checkInsertInitialPermissions();
    }
    catch (SQLException e) {
      throw new AclPersistenceException(e);
    }
  }

  private void ensureSegments(Connection con, String table)
    throws SQLException {
    s_log.debugT("ensureSegments(225)", "searching for acls without hash columns...");
    int anz = 0;
    PreparedStatement st = con.prepareStatement("SELECT COUNT(*) FROM " + table + " WHERE Seg1 IS NULL");
    ResultSet rs = st.executeQuery();
    if (rs.next()) {
      anz = rs.getInt(1);
    }
    rs.close();
    rs = null;
    st.close();
    st = null;
    hasOldColumnPopulation = (anz > 0);

    if (!hasOldColumnPopulation) {
      s_log.debugT("ensureSegments(239)", "all acls have correct hash columns!");
      return;
    }

    if (!generateHashColumns) {
      s_log.warningT("ensureSegments(244)", anz + " acls should be converted!");
      return;
    }

    s_log.infoT("ensureSegments(248)", anz + " acl entries are going to be converted in a new thread!");
    try {
      TableColumnPopulator.getInstance().addConnection(openConnection());
    }
    catch (AclPersistenceException e) {
      s_log.errorT("ensureSegments(253)", "Can't start thread to populate acl columns!" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(e));
    }
  }

  private void ensureHash(Connection con, String table)
    throws SQLException {
    PreparedStatement st = con.prepareStatement("SELECT RID FROM " + table + " WHERE Hash IS NULL");
    ResultSet rs = st.executeQuery();
    HashMap l = new HashMap();
    while (rs.next()) {
      String s = rs.getString(1);
      if (s != null) {
        int hash = s.hashCode();
        l.put(s, new Integer(hash));
      }
    }
    rs.close();
    rs = null;
    st.close();
    st = null;

    if (l.isEmpty()) {
      return;
    }

    st = con.prepareStatement("UPDATE " + table + " SET Hash=? WHERE RID=?");
    Iterator it = l.keySet().iterator();
    while (it.hasNext()) {
      String s = (String)it.next();
      st.setInt(1, ((Integer)l.get(s)).intValue());
      st.setString(2, s);
      st.executeUpdate();
    }
    st.close();
    st = null;
  }

  /**
   * @param externalAclID TBD: Description of the incoming method parameter
   * @return the ACL with the specified ID (or null if the ACL doesn't exist)
   * @exception AclPersistenceException Exception raised in failure situation
   */
  public JDBCAcl getAcl(String externalAclID, boolean raw)
    throws AclPersistenceException {

    if (externalAclID == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("getAcl(303)", "externalAclID " + externalAclID);
    }

    synchronized (m_main_connection) {
      try {
        JDBCAcl result = null;

        AclRec acl = uncommittedGetAcl(m_main_connection, externalAclID);
        if (acl != null) {
          UMPrincipalList owners = uncommittedGetOwners(m_main_connection, externalAclID, raw);
          AclEntryList entries = uncommittedGetAclEntries(m_main_connection, externalAclID, raw);
          if(   ( raw )
             && ( owners.containsRawEntries() || entries.containsRawEntries() )
            ) { 
            result = new JDBCAclRaw(this, externalAclID, new ObjectType(acl.getObjectTypeName()), owners, entries, acl.getChangeLevel());
          } else {
            result = new JDBCAcl(this, externalAclID, new ObjectType(acl.getObjectTypeName()), owners, entries, acl.getChangeLevel());
          }
        }

        m_main_connection.commit();
        return result;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * @param externalAclIDs TBD: Description of the incoming method parameter
   * @return the ACLs for the specified IDs (array elements are nuul for ACLs
   *      that doesn't exist)
   * @exception AclPersistenceException Exception raised in failure situation
   */
  public JDBCAcl[] getAcls(String[] externalAclIDs, boolean raw)
    throws AclPersistenceException {

    if (externalAclIDs == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("getAcls(356)", "externalAclIDs-length " + externalAclIDs.length);
    }

    if (externalAclIDs.length == 0) {
      return new JDBCAcl[0];
    }

    synchronized (m_main_connection) {
      try {
        JDBCAcl[] result = new JDBCAcl[externalAclIDs.length];

        // String[] internalIDs = uncommittedGetInternalAclIDs(externalAclIDs);
        // AclRec[] acls = uncommittedGetAcls(m_main_connection, internalIDs, externalAclIDs);
        AclRec[] acls = uncommittedGetAcls(m_main_connection, externalAclIDs, externalAclIDs);
        UMPrincipalList[] owners = uncommittedGetOwners(m_main_connection, acls, raw);
        AclEntryList[] entries = uncommittedGetAclEntries(m_main_connection, acls, raw);

        for (int i = 0; i < externalAclIDs.length; i++) {
          if (acls[i] != null) {
            if(   ( raw )
               && ( owners[i].containsRawEntries() || entries[i].containsRawEntries() )
              ) { 
              result[i] = new JDBCAclRaw(this, externalAclIDs[i], new ObjectType(acls[i].getObjectTypeName()), owners[i], entries[i], acls[i].getChangeLevel());
            } else {
              result[i] = new JDBCAcl(this, externalAclIDs[i], new ObjectType(acls[i].getObjectTypeName()), owners[i], entries[i], acls[i].getChangeLevel());
            }
          }
          else {
            result[i] = null;
          }
        }

        m_main_connection.commit();
        return result;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * @param externalAclID TBD: Description of the incoming method parameter
   * @return the ACEs of the specified ACL
   * @exception AclPersistenceException Exception raised in failure situation
   */
  public AclEntryList getAclEntries(String externalAclID, boolean raw)
    throws AclPersistenceException {

    if (externalAclID == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("getAclEntries(420)", "externalAclID " + externalAclID);
    }

    synchronized (m_main_connection) {
      try {
        AclEntryList result = uncommittedGetAclEntries(m_main_connection, externalAclID, raw);
        m_main_connection.commit();
        return result;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * Set the sort index of an ACE
   *
   * @param aclEntryID aclEntrySortIndex to be set
   * @param sortIndex aclEntrySortIndex to be set
   * @return TBD: Description of the outgoing return value
   * @exception AclPersistenceException Exception raised in failure situation
   */
  public boolean setAclEntrySortIndex(int aclEntryID, int sortIndex)
    throws AclPersistenceException {

    if (s_log.beDebug()) {
      s_log.debugT("setAclEntrySortIndex(463)", "aclEntryID " + aclEntryID + ", sortIndex " + sortIndex);
    }

    synchronized (m_main_connection) {
      try {
        boolean result = uncommittedSetAclEntrySortIndex(m_main_connection, aclEntryID, sortIndex);
        if (result) {
          uncommittedIncreaseAclChangeLevel(m_main_connection, uncommittedGetInternalAclIDFromEntryID(m_main_connection, aclEntryID));
        }

        m_main_connection.commit();
        return result;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * Set the propagate-flag of an ACE
   *
   * @param aclEntryID aclEntryPropagation to be set
   * @param propagate aclEntryPropagation to be set
   * @return TBD: Description of the outgoing return value
   * @exception AclPersistenceException Exception raised in failure situation
   */
  public boolean setAclEntryPropagation(int aclEntryID, boolean propagate)
    throws AclPersistenceException {

    if (s_log.beDebug()) {
      s_log.debugT("setAclEntryPropagation(510)", "aclEntryID " + aclEntryID + ", propagate " + propagate);
    }

    synchronized (m_main_connection) {
      try {
        boolean result = uncommittedSetAclEntryPropagation(m_main_connection, aclEntryID, propagate);
        if (result) {
          uncommittedIncreaseAclChangeLevel(m_main_connection, uncommittedGetInternalAclIDFromEntryID(m_main_connection, aclEntryID));
        }

        m_main_connection.commit();
        return result;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * @return the object types that are supported by the ACL manager
   */
  public IObjectTypeList getSupportedObjectTypes() {

    if (s_log.beDebug()) {
      s_log.debugT("getSupportedObjectTypes(551)", "getSupportedObjectTypes");
    }

    ObjectTypeList objectTypes = new ObjectTypeList();

    objectTypes.add(new ObjectType(IObjectType.OBJECT_TYPE_NODE));
    objectTypes.add(new ObjectType(IObjectType.OBJECT_TYPE_LEAF));

    return objectTypes;
  }

  /**
   * @param objectType TBD: Description of the incoming method parameter
   * @return the permissions which are supported for the specified object type
   * @exception AclPersistenceException Exception raised in failure situation
   */
  public IAclPermissionList getSupportedPermissions(IObjectType objectType)
    throws AclPersistenceException {

    if (objectType == null) {
      throw new java.lang.IllegalArgumentException();
    }

    IAclPermissionList supportedPermissions = cacheGetSupportedPermissions(objectType.getName());
    if (supportedPermissions != null) {
      return supportedPermissions;
    }

    if (s_log.beDebug()) {
      s_log.debugT("getSupportedPermissions(580)", "objectType " + objectType.getName());
    }

    synchronized (m_main_connection) {
      try {
        IAclPermissionList permissions = uncommittedGetSupportedPermissions(m_main_connection, objectType.getName());

        m_main_connection.commit();
        return permissions;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * @param externalAclID TBD: Description of the incoming method parameter
   * @return the owners of the specified ACL
   * @exception AclPersistenceException Exception raised in failure situation
   */
  public UMPrincipalList getOwners(String externalAclID, boolean raw)
    throws AclPersistenceException {

    if (externalAclID == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("getOwners(625)", "externalAclID " + externalAclID);
    }

    synchronized (m_main_connection) {
      try {
        UMPrincipalList result = uncommittedGetOwners(m_main_connection, externalAclID, raw);
        m_main_connection.commit();
        return result;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * @param permissionName TBD: Description of the incoming method parameter
   * @return the members of the permission
   * @exception AclPersistenceException Exception raised in failure situation
   */
  public IAclPermissionList getPermissionMembers(String permissionName)
    throws AclPersistenceException {

    if (permissionName == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("getPermissionMembers(669)", "permissionName " + permissionName);
    }

    synchronized (m_main_connection) {
      try {
        IAclPermissionList members = uncommittedGetPermissionMembers(m_main_connection, permissionName);

        m_main_connection.commit();
        return members;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * @param permission TBD: Description of the incoming method parameter
   * @return true iff the specified permission is used in at least one ACE
   * @exception AclPersistenceException Exception raised in failure situation
   */
  public boolean isPermissionUsedInAcl(IAclPermission permission)
    throws AclPersistenceException {

    if (permission == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("isPermissionUsedInAcl(714)", "permission " + permission.getName());
    }

    synchronized (m_main_connection) {
      try {
        boolean result = uncommittedIsUsedPermission(m_main_connection, permission.getName());

        m_main_connection.commit();
        return result;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * @param permission TBD: Description of the incoming method parameter
   * @return a list of all IDs of ACLs (strings) which use the specified
   *      permission in one of their ACEs
   * @exception AclPersistenceException Exception raised in failure situation
   */
  public List getAffectedAclIDs(IAclPermission permission)
    throws AclPersistenceException {

    if (permission == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("getAffectedAclIDs(760)", "permission " + permission.getName());
    }

    synchronized (m_main_connection) {
      try {
        List result = uncommittedGetAffectedAclIDs(m_main_connection, permission.getName());

        m_main_connection.commit();
        return result;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * @param permissionName TBD: Description of the incoming method parameter
   * @return the IAclPermission object for the specifoed permission name
   * @exception AclPersistenceException Exception raised in failure situation
   */
  public IAclPermission getPermission(String permissionName)
    throws AclPersistenceException {
    if (permissionName == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("getPermission(804)", "permissionName " + permissionName);
    }
    IAclPermission permission = cacheGetPermission(permissionName);
    if (permission != null) {
      return permission;
    }

    synchronized (m_main_connection) {
      try {
        IAclPermission result = uncommittedGetPermission(m_main_connection, permissionName);

        m_main_connection.commit();
        return result;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * @param externalAclID TBD: Description of the incoming method parameter
   * @return true iff the specified ACL is locked
   * @exception AclPersistenceException Exception raised in failure situation
   */
  public boolean isAclLocked(String externalAclID)
    throws AclPersistenceException {

    if (externalAclID == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("isAclLocked(853)", "externalAclID " + externalAclID);
    }

    synchronized (m_main_connection) {
      try {
        boolean result = uncommittedIsAclLocked(m_main_connection, externalAclID);

        m_main_connection.commit();
        return result;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * @param externalAclID TBD: Description of the incoming method parameter
   * @return the user which locks the specified ACL (or null if the ACL is
   *      currently not locked)
   * @exception AclPersistenceException Exception raised in failure situation
   */
  public IUser getAclLockingUser(String externalAclID)
    throws AclPersistenceException {

    if (externalAclID == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("getAclLockingUser(899)", "externalAclID " + externalAclID);
    }

    synchronized (m_main_connection) {
      try {
        IUser result = uncommittedGetAclLockingUser(m_main_connection, externalAclID);
        m_main_connection.commit();
        return result;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * @param externalAclIDRoot TBD: Description of the incoming method parameter
   * @return a list of all IDs of ACLs (strings) which start with the specified
   *      ID
   * @exception AclPersistenceException Exception raised in failure situation
   */
  public List getDescendantsWithAcl(String externalAclIDRoot)
    throws AclPersistenceException {
    if (externalAclIDRoot == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("getDescendantsWithAcl(943)", "externalAclIDRoot " + externalAclIDRoot);
    }

    synchronized (m_main_connection) {
      try {
        List result = uncommittedGetDescendantsWithAcl(m_main_connection, externalAclIDRoot);

        m_main_connection.commit();
        return result;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * @param owner TBD: Description of the incoming method parameter
   * @param externalAclID TBD: Description of the incoming method parameter
   * @param objectType TBD: Description of the incoming method parameter
   * @return a new ACL (create in the database)
   * @exception AclPersistenceException Exception raised in failure situation
   * @exception AclExistsException Exception raised in failure situation
   */
  public JDBCAcl createAcl(IUMPrincipal owner, String externalAclID, IObjectType objectType)
    throws AclPersistenceException, AclExistsException {

    if (owner == null || externalAclID == null || objectType == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("createAcl(991)", "owner " + owner.getId() + ", externalAclID " + externalAclID + ", objectType " + objectType);
    }

    synchronized (m_main_connection) {
      try {

        if (uncommittedCheckAcl(m_main_connection, externalAclID)) {
          throw new AclExistsException();
        }

        // in case the database is inconsistent, entries may exist for a removed acl - remove them
        uncommittedRemoveAclEntries(m_main_connection, externalAclID);

        // in case the database is inconsistent, owners may exist for a removed acl - remove them
        uncommittedRemoveOwners(m_main_connection, externalAclID);

        //check length
        Escape.escapeSqlString(externalAclID, RID_MAX_ESCAPED_LENGTH);

        if (!uncommittedAddOwner(m_main_connection, externalAclID, owner)) {
          throw new Exception("can't create owner");
        }
        if (!uncommittedAddAcl(m_main_connection, externalAclID, objectType)) {
          throw new Exception("can't create acl");
        }

        m_main_connection.commit();
      }
      catch (AclExistsException e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw e;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }

      return getAcl(externalAclID, false); // COOKED
    }
  }

  /**
   * Add a new ACE to the specified ACL (create in database)
   *
   * @param externalAclID aclEntry to be added
   * @param entry aclEntry to be added
   * @return true on success
   * @exception AclPersistenceException Exception raised in failure situation
   * @exception AlreadyAssignedToAclException Exception raised in failure
   *      situation
   */
  public int addAclEntry(String externalAclID, JDBCAclEntry entry)
    throws AclPersistenceException, AlreadyAssignedToAclException {

    if (externalAclID == null || entry == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("addAclEntry(1072)", "externalAclID " + externalAclID + ", entryID " + entry.getID());
    }

    synchronized (m_main_connection) {
      try {
        int result = -1;

        result = uncommittedAddAclEntry(m_main_connection, externalAclID, entry);
        if (result > 0) {
          uncommittedIncreaseAclChangeLevel(m_main_connection, externalAclID);
        }

        m_main_connection.commit();
        return result;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * Remove an ACL
   *
   * @param externalAclID TBD: Description of the incoming method parameter
   * @return true on success
   * @exception AclPersistenceException Exception raised in failure situation
   */
  public boolean removeAcl(String externalAclID)
    throws AclPersistenceException {

    if (externalAclID == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("removeAcl(1124)", "externalAclID " + externalAclID);
    }

    synchronized (m_main_connection) {
      try {
        boolean result = false;

        if (uncommittedGetAcl(m_main_connection, externalAclID) != null) {
          uncommittedRemoveAclEntries(m_main_connection, externalAclID);
          uncommittedRemoveOwners(m_main_connection, externalAclID);
          uncommittedRemoveAcl(m_main_connection, externalAclID);
          result = true;
        }

        m_main_connection.commit();
        return result;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * Remove an ACE from the database
   *
   * @param entry TBD: Description of the incoming method parameter
   * @return true on success
   * @exception AclPersistenceException Exception raised in failure situation
   */
  public boolean removeAclEntry(JDBCAclEntry entry)
    throws AclPersistenceException {

    if (entry == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("removeAclEntry(1178)", "entryID " + entry.getID());
    }

    synchronized (m_main_connection) {
      try {
        boolean result = false;

        String internalAclID = uncommittedGetInternalAclIDFromEntryID(m_main_connection, entry.getID());
        if (internalAclID != null) {
          result = uncommittedRemoveAclEntry(m_main_connection, entry.getID());
          if (result) {
            uncommittedIncreaseAclChangeLevel(m_main_connection, internalAclID);
          }
        }

        m_main_connection.commit();
        return result;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * Add a supported permission for an object type
   *
   * @param objectType supportedPermission to be added
   * @param permission supportedPermission to be added
   * @return true on success
   * @exception AclPersistenceException Exception raised in failure situation
   * @exception PredefinedPermissionException Exception raised in failure
   *      situation
   */
  public boolean addSupportedPermission(IObjectType objectType, IAclPermission permission)
    throws AclPersistenceException, PredefinedPermissionException {

    if (objectType == null || permission == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("addSupportedPermission(1236)", "objectType " + objectType.getName() + ", permission " + permission.getName());
    }

    synchronized (m_main_connection) {
      try {
        if (uncommittedIsPredefinedPermission(m_main_connection, permission.getName())) {
          throw new PredefinedPermissionException();
        }

        boolean result = false;
        if (!uncommittedIsSupportedPermission(m_main_connection, permission.getName(), objectType.getName())) {
          result = uncommittedAddSupportedPermission(m_main_connection, permission.getName(), objectType.getName());
        }

        m_main_connection.commit();
        return result;
      }
      catch (PredefinedPermissionException e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw e;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * Remove a supported permission for an object type
   *
   * @param objectType TBD: Description of the incoming method parameter
   * @param permission TBD: Description of the incoming method parameter
   * @return true on success
   * @exception AclPersistenceException Exception raised in failure situation
   * @exception PredefinedPermissionException Exception raised in failure
   *      situation
   * @exception PermissionUsedException Exception raised in failure situation
   */
  public boolean removeSupportedPermission(IObjectType objectType, IAclPermission permission)
    throws AclPersistenceException, PredefinedPermissionException, PermissionUsedException {

    if (objectType == null || permission == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("removeSupportedPermission(1305)", "objectType " + objectType.getName() + ", permission " + permission.getName());
    }

    synchronized (m_main_connection) {
      try {
        String permissionName = permission.getName();

        if (uncommittedIsPredefinedPermission(m_main_connection, permissionName)) {
          throw new PredefinedPermissionException();
        }
        if (uncommittedIsUsedPermission(m_main_connection, permissionName)) {
          throw new PermissionUsedException();
        }

        boolean result = false;
        if (uncommittedIsSupportedPermission(m_main_connection, permissionName, objectType.getName())) {
          result = uncommittedRemoveSupportedPermission(m_main_connection, permissionName, objectType.getName());
          if (result) {
            cacheRemovePermission(permissionName);
          }
        }

        m_main_connection.commit();
        return result;
      }
      catch (PredefinedPermissionException e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw e;
      }
      catch (PermissionUsedException e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw e;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * @param name TBD: Description of the incoming method parameter
   * @return a new permission (create in database)
   * @exception AclPersistenceException Exception raised in failure situation
   * @exception PermissionExistsException Exception raised in failure situation
   */
  public IAclPermission createPermission(String name)
    throws AclPersistenceException, PermissionExistsException {
    if (name == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("createPermission(1387)", "name " + name);
    }

    synchronized (m_main_connection) {
      try {
        if (uncommittedCheckPermission(m_main_connection, name)) {
          throw new PermissionExistsException();
        }

        // in case the database is inconsistent, entries may exist for a removed permission - remove them
        uncommittedRemovePermissionMembers(m_main_connection, name);
        uncommittedRemoveSupportedPermission(m_main_connection, name, IObjectType.OBJECT_TYPE_LEAF);
        uncommittedRemoveSupportedPermission(m_main_connection, name, IObjectType.OBJECT_TYPE_NODE);

        uncommittedAddPermission(m_main_connection, name, false);
//        uncommittedIncreasePermissionChangeLevel(m_main_connection);

        m_main_connection.commit();
        return new JDBCPermission(this, name, false, new AclPermissionList());
      }
      catch (PermissionExistsException e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw e;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * Remove a permission (from database)
   *
   * @param permission TBD: Description of the incoming method parameter
   * @return true on success
   * @exception AclPersistenceException Exception raised in failure situation
   * @exception PredefinedPermissionException Exception raised in failure
   *      situation
   * @exception PermissionUsedException Exception raised in failure situation
   */
  public boolean removePermission(IAclPermission permission)
    throws AclPersistenceException, PredefinedPermissionException, PermissionUsedException {
    if (permission == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("removePermission(1457)", "permission " + permission.getName());
    }

    synchronized (m_main_connection) {
      try {
        String permissisonName = permission.getName();

        if (uncommittedIsPredefinedPermission(m_main_connection, permissisonName)) {
          throw new PredefinedPermissionException();
        }
        if (uncommittedIsUsedPermission(m_main_connection, permissisonName)) {
          throw new PermissionUsedException();
        }

        uncommittedRemovePermissionMembers(m_main_connection, permissisonName);
        uncommittedRemoveSupportedPermission(m_main_connection, permissisonName, IObjectType.OBJECT_TYPE_LEAF);
        uncommittedRemoveSupportedPermission(m_main_connection, permissisonName, IObjectType.OBJECT_TYPE_NODE);

        boolean result = uncommittedRemovePermission(m_main_connection, permissisonName);
        if (result) {
          cacheRemovePermission(permissisonName);
        }

        m_main_connection.commit();
        return result;
      }
      catch (PredefinedPermissionException e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw e;
      }
      catch (PermissionUsedException e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw e;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * Add a new member to a permission
   *
   * @param permissionName permissionMember to be added
   * @param permission permissionMember to be added
   * @return true on success
   * @exception AclPersistenceException Exception raised in failure situation
   * @exception PermissionExistsException Exception raised in failure situation
   * @exception PredefinedPermissionException Exception raised in failure
   *      situation
   * @exception PermissionUsedException Exception raised in failure situation
   */
  public boolean addPermissionMember(String permissionName, IAclPermission permission)
    throws AclPersistenceException, PermissionExistsException, PredefinedPermissionException, PermissionUsedException {

    if (permissionName == null || permission == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("addPermissionMember(1547)", "permissionName " + permissionName + ", permission " + permission.getName());
    }

    synchronized (m_main_connection) {
      try {
        IAclPermission parent = uncommittedGetPermission(m_main_connection, permissionName);
        if (isPermissionMember(parent, permission)) {
          throw new PermissionExistsException();
        }
        if (isPermissionMember(permission, parent)) {
          throw new PermissionExistsException();
        }
        if (uncommittedIsPredefinedPermission(m_main_connection, permissionName)) {
          throw new PredefinedPermissionException();
        }
        if (uncommittedIsUsedPermission(m_main_connection, permissionName)) {
          throw new PermissionUsedException();
        }

        boolean result = uncommittedAddPermissionMember(m_main_connection, permissionName, permission.getName());
        if (result) {
          cacheRemovePermission(permissionName);
        }

        m_main_connection.commit();
        return result;
      }
      catch (PermissionExistsException e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw e;
      }
      catch (PredefinedPermissionException e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw e;
      }
      catch (PermissionUsedException e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw e;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * Remove a member from a permission
   *
   * @param permissionName TBD: Description of the incoming method parameter
   * @param permission TBD: Description of the incoming method parameter
   * @return true on success
   * @exception AclPersistenceException Exception raised in failure situation
   * @exception PredefinedPermissionException Exception raised in failure
   *      situation
   * @exception PermissionUsedException Exception raised in failure situation
   */
  public boolean removePermissionMember(String permissionName, IAclPermission permission)
    throws AclPersistenceException, PredefinedPermissionException, PermissionUsedException {

    if (permissionName == null || permission == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("removePermissionMember(1648)", "permissionName " + permissionName + ", permission " + permission.getName());
    }

    synchronized (m_main_connection) {
      try {
        if (uncommittedIsPredefinedPermission(m_main_connection, permissionName)) {
          throw new PredefinedPermissionException();
        }
        if (uncommittedIsUsedPermission(m_main_connection, permissionName)) {
          throw new PermissionUsedException();
        }

        boolean result = uncommittedRemovePermissionMember(m_main_connection, permissionName, permission.getName());
        if (result) {
          cacheRemovePermission(permissionName);
        }

        m_main_connection.commit();
        return result;
      }
      catch (PredefinedPermissionException e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw e;
      }
      catch (PermissionUsedException e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw e;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * Add an owner to an ACL
   *
   * @param externalAclID owner to be added
   * @param owner owner to be added
   * @return true on success
   * @exception AclPersistenceException Exception raised in failure situation
   */
  public boolean addOwner(String externalAclID, IUMPrincipal owner)
    throws AclPersistenceException {

    if (externalAclID == null || owner == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("addOwner(1728)", "externalAclID " + externalAclID + ", owner " + owner.getId());
    }

    synchronized (m_main_connection) {
      try {
        boolean result = false;

        if (!uncommittedIsOwner(m_main_connection, externalAclID, owner)) {
          result = uncommittedAddOwner(m_main_connection, externalAclID, owner);
          if (result) {
            uncommittedIncreaseAclChangeLevel(m_main_connection, externalAclID);
          }
        }

        m_main_connection.commit();
        return result;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * Remove an owner from an ACL
   *
   * @param externalAclID TBD: Description of the incoming method parameter
   * @param owner TBD: Description of the incoming method parameter
   * @return true on success
   * @exception AclPersistenceException Exception raised in failure situation
   * @exception LastOwnerException Exception raised in failure situation
   */
  public boolean removeOwner(String externalAclID, IUMPrincipal owner)
    throws AclPersistenceException, LastOwnerException {

    if (externalAclID == null || owner == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("removeOwner(1784)", "externalAclID " + externalAclID + ", owner " + owner.getId());
    }

    synchronized (m_main_connection) {
      try {
        boolean result = false;

        if (uncommittedIsLastOwner(m_main_connection, externalAclID, owner)) {
          throw new LastOwnerException();
        }
        result = uncommittedRemoveOwner(m_main_connection, externalAclID, owner);
        if (result) {
          uncommittedIncreaseAclChangeLevel(m_main_connection, externalAclID);
        }

        m_main_connection.commit();
        return result;
      }
      catch (LastOwnerException e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw e;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * Delete a principal from the tables (is usually called as reaction to an
   * event from the usermanagement)
   *
   * @param principalID TBD: Description of the incoming method parameter
   * @param principalType TBD: Description of the incoming method parameter
   * @exception AclPersistenceException Exception raised in failure situation
   */
  public void deletePrincipal(String principalID, int principalType)
    throws AclPersistenceException {

    if (principalID == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("deletePrincipal(1851)", "principalID " + principalID + ", principalType " + getPrincipalTypeText(principalType));
    }

    synchronized (m_main_connection) {
      try {
        // remove the principal as owner from all ACLs
        uncommittedRemoveOwner(m_main_connection, principalID, principalType);

        // remove all ACEs concerning the principal
        uncommittedRemoveAclEntries(m_main_connection, principalID, principalType);

        m_main_connection.commit();
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * Lock an ACL
   *
   * @param externalAclID TBD: Description of the incoming method parameter
   * @param lockingUser TBD: Description of the incoming method parameter
   * @return true on success
   * @exception AclPersistenceException Exception raised in failure situation
   */
  public boolean lockAcl(String externalAclID, IUMPrincipal lockingUser)
    throws AclPersistenceException {

    if (lockingUser == null || externalAclID == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("lockAcl(1902)", "externalAclID " + externalAclID + ", lockingUser " + lockingUser.getId());
    }

    synchronized (m_main_connection) {
      try {
        boolean result = false;

        if (!uncommittedIsAclLocked(m_main_connection, externalAclID)) {
          result = uncommittedLockAcl(m_main_connection, externalAclID, lockingUser);
          if (result) {
            uncommittedIncreaseAclChangeLevel(m_main_connection, externalAclID);
          }
        }

        m_main_connection.commit();
        return result;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * Unlock an ACL
   *
   * @param externalAclID TBD: Description of the incoming method parameter
   * @return true on success
   * @exception AclPersistenceException Exception raised in failure situation
   */
  public boolean unlockAcl(String externalAclID)
    throws AclPersistenceException {

    if (externalAclID == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("unlockAcl(1956)", "externalAclID " + externalAclID);
    }

    synchronized (m_main_connection) {
      try {
        boolean result = false;

        result = uncommittedUnlockAcl(m_main_connection, externalAclID);
        if (result) {
          uncommittedIncreaseAclChangeLevel(m_main_connection, externalAclID);
        }

        m_main_connection.commit();
        return result;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * Change the ID of an ACL (rename)
   *
   * @param oldExternalAclID TBD: Description of the incoming method parameter
   * @param newExternalAclID TBD: Description of the incoming method parameter
   * @return true on success
   * @exception AclPersistenceException Exception raised in failure situation
   * @exception AclExistsException Exception raised in failure situation
   */
  public boolean changeAclID(String oldExternalAclID, String newExternalAclID)
    throws AclPersistenceException, AclExistsException {

    if (oldExternalAclID == null || newExternalAclID == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("changeAclID(2010)", "oldExternalAclID " + oldExternalAclID + ", newExternalAclID " + newExternalAclID);
    }

    synchronized (m_main_connection) {
      try {
        if (oldExternalAclID.equals(newExternalAclID)) {
          throw new AclExistsException();
        }

        boolean result = false;

        if (uncommittedGetAcl(m_main_connection, newExternalAclID) != null) {
          throw new AclExistsException();
        }

        if (uncommittedCheckAcl(m_main_connection, newExternalAclID)) {
          throw new AclExistsException();
        }
        if (uncommittedCheckAclEntries(m_main_connection, newExternalAclID)) {
          throw new AclExistsException();
        }
        if (uncommittedCheckOwners(m_main_connection, newExternalAclID)) {
          throw new AclExistsException();
        }

        if (!uncommittedChangeAclID_in_AclTable(m_main_connection, oldExternalAclID, newExternalAclID)) {
          throw new Exception("can't change acl id in acl table");
        }
        if (uncommittedCheckAclEntries(m_main_connection, oldExternalAclID)) {
          if (!uncommittedChangeAclID_in_AclEntryTable(m_main_connection, oldExternalAclID, newExternalAclID)) {
            throw new Exception("can't change acl id in acl entry table");
          }
        }
        if (!uncommittedChangeAclID_in_OwnerTable(m_main_connection, oldExternalAclID, newExternalAclID)) {
          throw new Exception("can't change acl id in owner table");
        }

        result = true;

        m_main_connection.commit();
        return result;
      }
      catch (AclExistsException e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw e;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * Change all IDs of ACLs which start with oldAclIdPrefix by replacing it with
   * newAclIdPrefix
   *
   * @param oldAclIdPrefix TBD: Description of the incoming method parameter
   * @param newAclIdPrefix TBD: Description of the incoming method parameter
   * @exception AclPersistenceException Exception raised in failure situation
   * @exception AclExistsException Exception raised in failure situation
   */
  public void changeMultipleAclIDs(String oldAclIdPrefix, String newAclIdPrefix)
    throws AclPersistenceException, AclExistsException {

    if (oldAclIdPrefix == null || newAclIdPrefix == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("changeMultipleAclIDs(2102)", "oldAclIdPrefix " + oldAclIdPrefix + ", newAclIdPrefix " + newAclIdPrefix);
    }

    synchronized (m_main_connection) {
      List descendants = getDescendantsWithAcl(oldAclIdPrefix);
      if (descendants == null) {
        return;
      }

      Iterator iterator = descendants.iterator();
      while (iterator.hasNext()) {
        String id = (String)iterator.next();
        if (id != null) {
          changeAclID(id, newAclIdPrefix + id.substring(oldAclIdPrefix.length()));
        }
      }
    }
  }

  /**
   * Remove all ACLs with IDs starting with aclIdPrefix (remove children)
   *
   * @param aclIdPrefix TBD: Description of the incoming method parameter
   * @exception AclPersistenceException Exception raised in failure situation
   */
  public void removeMultipleAcls(String aclIdPrefix)
    throws AclPersistenceException {

    if (aclIdPrefix == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("removeMultipleAcls(2135)", "aclIdPrefix " + aclIdPrefix);
    }

    synchronized (m_main_connection) {
      try {
        uncommittedRemoveDescendantsWithAcl(m_main_connection, aclIdPrefix);

        m_main_connection.commit();
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  public boolean hasDescendantsWithAcl(String externalAclIDRoot)
    throws AclPersistenceException {
    try {
      List ret = hasDescendantsWithAcl(m_main_connection, externalAclIDRoot);
      return (ret == null ? false : ret.size() > 0);
    }
    catch (Exception e) {
      throw new AclPersistenceException(e);
    }
  }

  protected static Integer[] makeHashSegments(String path) {
    Integer[] ret = new Integer[10];

    if (path.startsWith("/")) {
      path = path.substring(1);
    }
    for (int nrOfSeg = 0; nrOfSeg < 10; nrOfSeg++) {
      if (path == null || path.equals("")) {
        ret[nrOfSeg] = null;
      }
      else {
        int j = path.indexOf('/');
        if (j < 0) {
          ret[nrOfSeg] = new Integer(path.hashCode());
          break;
        }
        ret[nrOfSeg] = new Integer(path.substring(0, j).hashCode());
        path = path.substring(j + 1);
      }
    }

    return ret;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param externalAclIDRoot TBD: Description of the incoming method parameter
   * @return true iff ACLs exist with IDs starting with externalAclIDRoot (has
   *      children)
   * @exception SQLException Exception raised in failure situation
   */
  private List hasDescendantsWithAcl(Connection connection, String externalAclIDRoot)
    throws SQLException {
    List ret = new LinkedList();

    Integer[] segments = makeHashSegments(externalAclIDRoot);
    String sql = " WHERE ((";
    if (hasOldColumnPopulation) {
      sql += "Seg1 IS NULL) OR (";
    }
    for (int i = 0; i < 10; i++) {
      if (segments[i] == null) {
        break;
      }
      if (i > 0) {
        sql += " AND ";
      }
      sql += "(Seg" + (i + 1) + "=?)";
    }
    sql += ")) AND (" + m_ApplicationAndManagerSelectionCondition + ") AND (" + ACL_RIDSTART_COL;
    boolean big = false;

    if (!externalAclIDRoot.endsWith("/")) {
      externalAclIDRoot += "/";
    }
    String s = Escape.escapeSqlString(externalAclIDRoot, 0).trim() + '%';
    String t = s;
    if (t.length() > RID_START_LENGTH) {
      t = t.substring(0, RID_START_LENGTH);
      sql += "=?) AND (" + ACL_ID_COL + " LIKE ?)";
      big = true;
    }
    else {
      sql += " LIKE ?)";
    }

    PreparedStatement st = null;
    ResultSet rs = null;
    try {
      st = connection.prepareStatement("SELECT " + ACL_ID_COL + " FROM " + ACL_TABLE_NAME + sql);
      int i = 1;
      for (int j = 0; j < 10; j++) {
        Integer seg = segments[j];
        if (seg == null) {
          break;
        }
        st.setInt(i++, seg.intValue());
      }
      st.setInt(i++, m_owner);
      st.setString(i++, t);
      if (big) {
        st.setString(i++, s);
      }
      rs = st.executeQuery();
      while (rs.next()) {
        ret.add(rs.getString(1));
      }
      rs.close();
      rs = null;
      st.close();
      st = null;
      return ret;
    }
    finally {
      if (rs != null) {
        rs.close();
      }
      if (st != null) {
        st.close();
      }
    }
  }


  private void uncommittedRemoveDescendantsWithAcl(Connection connection, String externalAclIDRoot)
    throws SQLException {

    //the new call is important, do not change!!!!
    //due to some strange implementations the
    //select count(*) from table where column like 'literal%'
    //is much more performant with an empty result set than the
    //direct select of the desired column!!!!
    List childs = hasDescendantsWithAcl(connection, externalAclIDRoot);
    if (childs == null || childs.size() == 0) {
      return;
    }

    if (!externalAclIDRoot.endsWith("/")) {
      externalAclIDRoot += "/";
    }
    String sql = " WHERE (" + m_ApplicationAndManagerSelectionCondition + ") AND (" + ACL_RIDSTART_COL;
    String s = Escape.escapeSqlString(externalAclIDRoot, 0).trim() + '%';
    boolean big = false;
    String t = s;
    if (t.length() > RID_START_LENGTH) {
      t = t.substring(0, RID_START_LENGTH);
      sql += "=?) AND (" + ACL_ID_COL + " LIKE ?)";
      big = true;
    }
    else {
      sql += " LIKE ?)";
    }

    PreparedStatement st = null;
    ResultSet rs = null;
    try {
      st = connection.prepareStatement("DELETE " + ACL_TABLE_NAME + sql);
      st.setInt(1, m_owner);
      st.setString(2, t);
      if (big) {
        st.setString(3, s);
      }
      st.executeUpdate();
      st.close();
      st = null;

      st = connection.prepareStatement("SELECT COUNT(*) FROM " + ACL_ENTRY_TABLE_NAME + sql);
      st.setInt(1, m_owner);
      st.setString(2, t);
      if (big) {
        st.setString(3, s);
      }
      rs = st.executeQuery();
      int i = 0;
      if (rs.next()) {
        i = rs.getInt(1);
      }
      rs.close();
      rs = null;
      st.close();
      st = null;
      if (i > 0) {
        st = connection.prepareStatement("DELETE " + ACL_ENTRY_TABLE_NAME + sql);
        st.setInt(1, m_owner);
        st.setString(2, t);
        if (big) {
          st.setString(3, s);
        }
        i = st.executeUpdate();
        st.close();
        st = null;
      }

      st = connection.prepareStatement("SELECT COUNT(*) FROM " + OWNER_TABLE_NAME + sql);
      st.setInt(1, m_owner);
      st.setString(2, t);
      if (big) {
        st.setString(3, s);
      }
      rs = st.executeQuery();
      i = 0;
      if (rs.next()) {
        i = rs.getInt(1);
      }
      rs.close();
      rs = null;
      st.close();
      st = null;
      if (i > 0) {
        st = connection.prepareStatement("DELETE " + OWNER_TABLE_NAME + sql);
        st.setInt(1, m_owner);
        st.setString(2, t);
        if (big) {
          st.setString(3, s);
        }
        i = st.executeUpdate();
        st.close();
        st = null;
      }

      st = connection.prepareStatement("UPDATE wcm_acl_version SET counter=counter+1");
      st.executeUpdate();
    }
    finally {
      if (rs != null) {
        rs.close();
      }
      if (st != null) {
        st.close();
      }
    }
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param externalAclIDRoot TBD: Description of the incoming method parameter
   * @return all ACL IDs (strings) in a list which start with the specified
   *      prefix
   * @exception SQLException Exception raised in failure situation
   */
  private List uncommittedGetDescendantsWithAcl(Connection connection, String externalAclIDRoot)
    throws SQLException {

    LinkedList result = new LinkedList();

    List childs = hasDescendantsWithAcl(connection, externalAclIDRoot);
    if (childs != null) {
      Iterator it = childs.listIterator();
      while (it.hasNext()) {
        String externalAclID = trim(Escape.unescapeSqlString((String)it.next()));
        result.add(externalAclID);
      }
    }
    return result;
  }

  /**
   * @param externalAclID TBD: Description of the incoming method parameter
   * @param changeLevel TBD: Description of the incoming method parameter
   * @return true iff the change level of the ACL in the database is not larger
   *      than the specified one
   * @exception AclPersistenceException Exception raised in failure situation
   */
  public boolean isAclUpToDate(String externalAclID, long changeLevel)
    throws AclPersistenceException {

    if (externalAclID == null) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("isAclUpToDate(2428)", "externalAclID " + externalAclID + ", changeLevel " + changeLevel);
    }

    synchronized (m_main_connection) {
      try {
        boolean result = uncommittedIsAclUpToDate(m_main_connection, externalAclID, changeLevel);
        m_main_connection.commit();
        return result;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * @param externalAclIDs TBD: Description of the incoming method parameter
   * @param changeLevels TBD: Description of the incoming method parameter
   * @return up to date flags for ACLs
   * @exception AclPersistenceException Exception raised in failure situation
   */
  public boolean[] areAclsUpToDate(String[] externalAclIDs, long[] changeLevels)
    throws AclPersistenceException {

    if (externalAclIDs == null || changeLevels == null || externalAclIDs.length != changeLevels.length) {
      throw new java.lang.IllegalArgumentException();
    }

    if (s_log.beDebug()) {
      s_log.debugT("areAclsUpToDate(2473)", "externalAclIDs-length " + externalAclIDs.length + ", changeLevels-length " + changeLevels.length);
    }

    synchronized (m_main_connection) {
      try {
        // String[] internalAclIDs = uncommittedGetInternalAclIDs(externalAclIDs);
        // boolean[] result = uncommittedAreAclsUpToDate(m_main_connection, internalAclIDs, changeLevels);
        boolean[] result = uncommittedAreAclsUpToDate(m_main_connection, externalAclIDs, changeLevels);

        m_main_connection.commit();
        return result;
      }
      catch (Exception e) {
        try {
          m_main_connection.rollback();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        try {
          reopenMainConnection();
        }
        catch (SQLException sqlException) {
          /*
           * can only throw one exception at a time
           */
        }
        throw new AclPersistenceException(e);
      }
    }
  }

  /**
   * @return the ID of the used connection pool
   */
  protected String getPoolID() {
    return m_poolID;
  }

  /**
   * @param principal TBD: Description of the incoming method parameter
   * @return the value that's written into the database to identify the type of
   *      the principal
   */
  private int getPrincipalTypeTag(IUMPrincipal principal) {

    return getPrincipalTypeTag(principal.getType());
  }

  /**
   * @param principalType TBD: Description of the incoming method parameter
   * @return the value that's written into the database to identify the
   *      principal type
   */
  private int getPrincipalTypeTag(int principalType) {

    switch (principalType) {
      case IUMPrincipal.IUSER:
        return PRINCIPAL_TYPE_TAG_USER;
      case IUMPrincipal.IGROUP:
        return PRINCIPAL_TYPE_TAG_GROUP;
      case IUMPrincipal.IROLE:
        return PRINCIPAL_TYPE_TAG_ROLE;
    }

    return PRINCIPAL_TYPE_TAG_USER;
  }

  /**
   * @param tag TBD: Description of the incoming method parameter
   * @return the principal type that belongs to the tag (i.e. the value that's
   *      written into the database to identify a principal type)
   */
  private int getPrincipalTypeFromTag(int tag) {

    switch (tag) {
      case PRINCIPAL_TYPE_TAG_USER:
        return IUMPrincipal.IUSER;
      case PRINCIPAL_TYPE_TAG_GROUP:
        return IUMPrincipal.IGROUP;
      case PRINCIPAL_TYPE_TAG_ROLE:
        return IUMPrincipal.IROLE;
    }

    return IUMPrincipal.IUSER;
  }

  /**
   * @param id principal ID
   * @param type principal type
   * @param raw when <code>true</code>, return an object even if the user management doesn't know that principal anymore
   * @return a IUMPrincipal object for the specified principal (<code>null</code> if the
   *      principal does not exist) The principal is NOT populated!
   */
private IUMPrincipal getUMPrincipal(String id, int type, boolean raw) {

    IUMPrincipal result = null;
    
    try {
      switch (type) {
        case IUMPrincipal.IUSER:
          IUserFactory userFactory = WPUMFactory.getUserFactory();
          if (userFactory != null) {
            result = userFactory.getUser(id, IUserFactory.TRUSTED);
          }
          break;
        case IUMPrincipal.IGROUP:
          IGroupFactory groupFactory = WPUMFactory.getGroupFactory();
          if (groupFactory != null) {
            result = groupFactory.getGroup(id);
          }
          break;
        case IUMPrincipal.IROLE:
          IRoleFactory roleFactory = WPUMFactory.getRoleFactory();
          if (roleFactory != null) {
            result = roleFactory.getRole(id);
          }
          break;
      }
    }
    catch (Exception e) {
      s_log.warningT("getUMPrincipal(2593)", "can't get principal " + id + " (" + getPrincipalTypeText(type) + ") from user management: " + LoggingFormatter.extractCallstack(e));
      return null;
    }
    
    if (result == null && raw) {
      try {
        result = new UMPrincipalRaw(type, id);
      }
      catch (Exception e) {
        s_log.warningT("getUMPrincipal", "can't get raw principal " + id + " (" + getPrincipalTypeText(type) + "): " + LoggingFormatter.extractCallstack(e));
      }
    }
    
    return result;
  }

  /**
   * Populate a principal
   *
   * @param principal TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   */
  private IUMPrincipal populate(IUMPrincipal principal) {

    try {
      switch (principal.getType()) {
        case IUMPrincipal.IUSER:
          IUserFactory userFactory = WPUMFactory.getUserFactory();
          if (userFactory != null) {
            return userFactory.getUser(principal.getId());
          }
          break;
        case IUMPrincipal.IGROUP:
          IGroupFactory groupFactory = WPUMFactory.getGroupFactory();
          if (groupFactory != null) {
            return groupFactory.getGroup(principal.getId());
          }
          break;
        case IUMPrincipal.IROLE:
          IRoleFactory roleFactory = WPUMFactory.getRoleFactory();
          if (roleFactory != null) {
            return roleFactory.getRole(principal.getId());
          }
          break;
      }
    }
    catch (Exception e) {
      s_log.warningT("populate(2631)", "can't get populated principal " + principal.getId() + " (" + getPrincipalTypeText(principal.getType()) + ") from user management: " + LoggingFormatter.extractCallstack(e));
      return null;
    }

    return null;
  }

  /**
   * @param type TBD: Description of the incoming method parameter
   * @return a displayname for a principal type (for logging)
   */
  private String getPrincipalTypeText(int type) {

    switch (type) {
      case IUMPrincipal.IUSER:
        return "user";
      case IUMPrincipal.IGROUP:
        return "group";
      case IUMPrincipal.IROLE:
        return "role";
    }

    return "";
  }

  /**
   * Add permissions to the database which are always supported by this ACL
   * manager (i.e. fullcontrol)
   *
   * @exception AclPersistenceException Exception raised in failure situation
   */
  private void checkInsertInitialPermissions()
    throws AclPersistenceException {

    try {
      if (!uncommittedCheckInitialPermissions(m_main_connection)) {
        uncommittedInsertInitialPermissions(m_main_connection);
      }
      if (!uncommittedCheckInitialSupportedPermissions(m_main_connection)) {
        uncommittedInsertInitialSupportedPermissions(m_main_connection);
      }

      m_main_connection.commit();
    }
    catch (Exception e) {
      try {
        m_main_connection.rollback();
      }
      catch (SQLException sqlException) {
        /*
         * can only throw one exception at a time
         */
      }
      throw new AclPersistenceException(e);
    }
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param externalAclID TBD: Description of the incoming method parameter
   * @return an ACL object for an ACL in the database (may be null)
   * @exception SQLException Exception raised in failure situation
   */
  private AclRec uncommittedGetAcl(Connection connection, String externalAclID)
    throws SQLException {

    AclRec result = null;

    PreparedStatement st = connection.prepareStatement(
      "SELECT * FROM " + ACL_TABLE_NAME + " WHERE " + m_ApplicationAndManagerSelectionCondition
       + " AND " + ACL_HASH_COL + "=? AND " + ACL_ID_COL + "=?");

    SqlResult queryResult = query(st, 1, Escape.escapeSqlString(externalAclID, 0), 2);
    if (queryResult != null) {
      try {
        if (queryResult.m_result.next()) {
          String objectType = trim(Escape.unescapeSqlString(queryResult.m_result.getString(OBJECT_TYPE_COL)));
          long changeLevel = queryResult.m_result.getLong(CHANGE_LEVEL_COL);
          result = new AclRec(externalAclID, objectType, changeLevel);
        }
      }
      finally {
        queryResult.close();
      }
    }

    return result;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param entryID TBD: Description of the incoming method parameter
   * @return the ACL ID which an ACE is attached to
   * @exception SQLException Exception raised in failure situation
   */
  private String uncommittedGetInternalAclIDFromEntryID(Connection connection, int entryID)
    throws SQLException {

    String result = null;

    PreparedStatement st = connection.prepareStatement("SELECT " + ACL_ID_COL
       + " FROM " + ACL_ENTRY_TABLE_NAME + " WHERE " + ACL_ENTRY_ID_COL + "=?");

    st.setInt(1, entryID);

    SqlResult queryResult = query(st, 0, null, -1);
    if (queryResult != null) {
      try {
        if (queryResult.m_result.next()) {
          result = trim(Escape.unescapeSqlString(queryResult.m_result.getString(ACL_ID_COL)));
        }
      }
      finally {
        queryResult.close();
      }
    }

    return result;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param internalIDs TBD: Description of the incoming method parameter
   * @param externalIDs TBD: Description of the incoming method parameter
   * @return ACL objects for ACLs in the database (some may be null) (don't get
   *      them all at once to limit the size of the SQL statement)
   * @exception SQLException Exception raised in failure situation
   */
  private AclRec[] uncommittedGetAcls(Connection connection, String[] internalIDs, String[] externalIDs)
    throws SQLException {

    final int MAX_ACLS_PER_QUERY = 100;

    if (internalIDs.length == 0) {
      return new AclRec[0];
    }

    AclRec[] result = new AclRec[internalIDs.length];

    for (int i = 0; i < internalIDs.length; i += MAX_ACLS_PER_QUERY) {
      int takeCount = min(internalIDs.length - i, MAX_ACLS_PER_QUERY);
      String[] internalTakeIDs = new String[takeCount];
      String[] externalTakeIDs = new String[takeCount];
      for (int k = 0; k < takeCount; k++) {
        internalTakeIDs[k] = internalIDs[i + k];
        externalTakeIDs[k] = externalIDs[i + k];
      }
      AclRec[] acls = uncommittedGetAclsFromDB(connection, internalTakeIDs, externalTakeIDs);
      for (int k = 0; k < takeCount; k++) {
        result[i + k] = acls[k];
      }
    }

    return result;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclIDs TBD: Description of the incoming method parameter
   * @param externalAclIDs TBD: Description of the incoming method parameter
   * @return ACL objects for ACLs in the database (some may be null)
   * @exception SQLException Exception raised in failure situation
   */
  private AclRec[] uncommittedGetAclsFromDB(Connection connection, String[] internalAclIDs, String[] externalAclIDs)
    throws SQLException {

    if (internalAclIDs.length == 0) {
      return new AclRec[0];
    }

    String sql = "SELECT * FROM " + ACL_TABLE_NAME + " WHERE " + m_ApplicationAndManagerSelectionCondition
       + " AND ";

    sql += "(";
    HashMap idIndices = new HashMap(internalAclIDs.length);
    for (int i = 0; i < internalAclIDs.length; i++) {
      idIndices.put(internalAclIDs[i], new Integer(i));
      if (i > 0) {
        sql += " OR (";
      }
      String s = Escape.escapeSqlString(internalAclIDs[i], 0);
      sql += ACL_HASH_COL + "=" + s.hashCode() + " AND " + ACL_ID_COL + "='" + s + "'";
      if (i > 0) {
        sql += ")";
      }
    }
    sql += ")";

    AclRec[] result = new AclRec[internalAclIDs.length];
    for (int i = 0; i < result.length; i++) {
      result[i] = null;
    }

    SqlResult queryResult = query(connection, sql);
    if (queryResult != null) {
      try {
        while (queryResult.m_result.next()) {

          String aclID = trim(Escape.unescapeSqlString(queryResult.m_result.getString(ACL_ID_COL)));
          String objectType = trim(Escape.unescapeSqlString(queryResult.m_result.getString(OBJECT_TYPE_COL)));
          long changeLevel = queryResult.m_result.getLong(CHANGE_LEVEL_COL);

          Integer indexObj = (Integer)idIndices.get(aclID);
          if (indexObj != null) {
            int index = indexObj.intValue();
            result[index] = new AclRec(externalAclIDs[index], objectType, changeLevel);
          }
        }
      }
      finally {
        queryResult.close();
      }
    }

    return result;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclID TBD: Description of the incoming method parameter
   * @param changeLevel TBD: Description of the incoming method parameter
   * @return true iff the change level of the ACL in the database is not larger
   *      than the specified one
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedIsAclUpToDate(Connection connection, String internalAclID, long changeLevel)
    throws SQLException {

    boolean result = false;

    PreparedStatement st = connection.prepareStatement(
      "SELECT " + CHANGE_LEVEL_COL + " FROM " + ACL_TABLE_NAME
       + " WHERE " + m_ApplicationAndManagerSelectionCondition + " AND " + ACL_HASH_COL + "=? AND " + ACL_ID_COL + "=?");

    SqlResult queryResult = query(st, 1, Escape.escapeSqlString(internalAclID, 0), 2);
    if (queryResult != null) {
      try {
        if (queryResult.m_result.next()) {
          long dbLevel = queryResult.m_result.getLong(CHANGE_LEVEL_COL);
          result = (dbLevel == changeLevel);
        }
      }
      finally {
        queryResult.close();
      }
    }

    return result;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param internalIDs TBD: Description of the incoming method parameter
   * @param changeLevels TBD: Description of the incoming method parameter
   * @return up to date flags for ACLs (don't get them all at once to limit the
   *      size of the SQL statement)
   * @exception SQLException Exception raised in failure situation
   */
  private boolean[] uncommittedAreAclsUpToDate(Connection connection, String[] internalIDs, long[] changeLevels)
    throws SQLException {

    final int MAX_ACLS_PER_QUERY = 100;

    if (internalIDs.length == 0) {
      return new boolean[0];
    }

    boolean[] result = new boolean[internalIDs.length];

    for (int i = 0; i < internalIDs.length; i += MAX_ACLS_PER_QUERY) {
      int takeCount = min(internalIDs.length - i, MAX_ACLS_PER_QUERY);
      String[] internalTakeIDs = new String[takeCount];
      long[] changeTakeLevels = new long[takeCount];
      for (int k = 0; k < takeCount; k++) {
        internalTakeIDs[k] = internalIDs[i + k];
        changeTakeLevels[k] = changeLevels[i + k];
      }
      boolean[] flags = uncommittedAreAclsUpToDateInDB(connection, internalTakeIDs, changeTakeLevels);
      for (int k = 0; k < takeCount; k++) {
        result[i + k] = flags[k];
      }
    }

    return result;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclIDs TBD: Description of the incoming method parameter
   * @param changeLevels TBD: Description of the incoming method parameter
   * @return up to date flags for ACLs (check change level in the database)
   * @exception SQLException Exception raised in failure situation
   */
  private boolean[] uncommittedAreAclsUpToDateInDB(Connection connection, String[] internalAclIDs, long[] changeLevels)
    throws SQLException {

    if (internalAclIDs.length == 0) {
      return new boolean[0];
    }

    String sql = "SELECT * FROM " + ACL_TABLE_NAME + " WHERE " + m_ApplicationAndManagerSelectionCondition + " AND ";

    sql += "(";
    HashMap idIndices = new HashMap(internalAclIDs.length);
    for (int i = 0; i < internalAclIDs.length; i++) {
      idIndices.put(internalAclIDs[i], new Integer(i));
      if (i > 0) {
        sql += " OR (";
      }
      String s = Escape.escapeSqlString(internalAclIDs[i], 0);
      sql += ACL_HASH_COL + "=" + s.hashCode() + " AND " + ACL_ID_COL + "='" + s + "'";
      if (i > 0) {
        sql += ")";
      }
    }
    sql += ")";

    boolean[] result = new boolean[internalAclIDs.length];
    for (int i = 0; i < result.length; i++) {
      result[i] = false;
    }

    SqlResult queryResult = query(connection, sql);
    if (queryResult != null) {
      try {
        while (queryResult.m_result.next()) {

          String aclID = trim(Escape.unescapeSqlString(queryResult.m_result.getString(ACL_ID_COL)));
          long dbLevel = queryResult.m_result.getLong(CHANGE_LEVEL_COL);

          Integer indexObj = (Integer)idIndices.get(aclID);
          if (indexObj != null) {
            int index = indexObj.intValue();
            result[index] = dbLevel == changeLevels[index];
          }
        }
      }
      finally {
        queryResult.close();
      }
    }

    return result;
  }

  /**
   * Increase the change level of the ACL
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclID TBD: Description of the incoming method parameter
   * @exception SQLException Exception raised in failure situation
   */
  private void uncommittedIncreaseAclChangeLevel(Connection connection, String internalAclID)
    throws SQLException {
    PreparedStatement st = connection.prepareStatement(
      "UPDATE " + ACL_TABLE_NAME + " SET " + CHANGE_LEVEL_COL + "=" + CHANGE_LEVEL_COL + "+1 WHERE " +
      m_ApplicationAndManagerSelectionCondition + " AND " + ACL_HASH_COL + "=? AND " + ACL_ID_COL + "=?");
    String s = Escape.escapeSqlString(internalAclID, 0);
    st.setInt(2, s.hashCode());
    st.setString(3, s);
    update(connection, st);
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclID TBD: Description of the incoming method parameter
   * @return true iff the ACL exists in the database
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedCheckAcl(Connection connection, String internalAclID)
    throws SQLException {
    boolean result = false;
    PreparedStatement st = connection.prepareStatement(
      "SELECT * FROM " + ACL_TABLE_NAME + " WHERE " + m_ApplicationAndManagerSelectionCondition
       + " AND " + ACL_HASH_COL + "=? AND " + ACL_ID_COL + "=?");

    SqlResult queryResult = query(st, 1, Escape.escapeSqlString(internalAclID, 0), 2);
    if (queryResult != null) {
      try {
        if (queryResult.m_result.next()) {
          result = true;
        }
      }
      finally {
        queryResult.close();
      }
    }

    return result;
  }

  /**
   * Set the ACE sort index
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param aclEntryID TBD: Description of the incoming method parameter
   * @param sortIndex TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedSetAclEntrySortIndex(Connection connection, int aclEntryID, int sortIndex)
    throws SQLException {
    PreparedStatement st = connection.prepareStatement(
      "UPDATE " + ACL_ENTRY_TABLE_NAME
       + " SET " + SORT_INDEX_COL + " = ? WHERE " + ACL_ENTRY_ID_COL + " = ?");
    st.setInt(1, sortIndex);
    st.setInt(2, aclEntryID);

    return update(connection, st, 0, null, -1) > 0;
  }

  /**
   * Define the propagate-flag of the ACE
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param aclEntryID TBD: Description of the incoming method parameter
   * @param propagate TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedSetAclEntryPropagation(Connection connection, int aclEntryID, boolean propagate)
    throws SQLException {
    PreparedStatement st = connection.prepareStatement(
      "UPDATE " + ACL_ENTRY_TABLE_NAME
       + " SET " + IS_PROPAGATED_COL + " = ? WHERE "
       + ACL_ENTRY_ID_COL + " = ?");
    st.setInt(1, propagate ? 1 : 0);
    st.setInt(2, aclEntryID);

    return update(connection, st, 0, null, -1) > 0;
  }


  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclID TBD: Description of the incoming method parameter
   * @return true iff the ACL has ACEs
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedCheckAclEntries(Connection connection, String internalAclID)
    throws SQLException {
    boolean result = false;
    PreparedStatement st = connection.prepareStatement(
      "SELECT * FROM " + ACL_ENTRY_TABLE_NAME + " WHERE " + m_ApplicationAndManagerSelectionCondition
       + " AND " + ACL_HASH_COL + "=? AND " + ACL_ID_COL + "=?");

    SqlResult queryResult = query(st, 1, Escape.escapeSqlString(internalAclID, 0), 2);
    if (queryResult != null) {
      try {
        if (queryResult.m_result.next()) {
          result = true;
        }
      }
      finally {
        queryResult.close();
      }
    }

    return result;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param permissionName TBD: Description of the incoming method parameter
   * @return true iff the permission exists
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedCheckPermission(Connection connection, String permissionName)
    throws SQLException {
    boolean result = false;
    PreparedStatement st = connection.prepareStatement(
      "SELECT * FROM " + PERMISSION_TABLE_NAME + " WHERE " + m_ApplicationAndManagerSelectionCondition
       + " AND " + PERMISSION_NAME_COL + " = ?");
    st.setString(2, Escape.escapeSqlString(permissionName, PERMISSION_NAME_LENGTH));

    SqlResult queryResult = query(st);
    if (queryResult != null) {
      try {
        if (queryResult.m_result.next()) {
          result = true;
        }
      }
      finally {
        queryResult.close();
      }
    }

    return result;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclID TBD: Description of the incoming method parameter
   * @return the ACEs of the ACL (principal may be null; otherwise only the
   *      entries belonging to the pricnipal are returned)
   * @exception SQLException Exception raised in failure situation
   */
  private AclEntryList uncommittedGetAclEntries(Connection connection, String internalAclID, boolean raw)
    throws SQLException {
    AclEntryList result = new AclEntryList();

    SqlResult queryResult = uncommittedGetAclEntriesFromDB(connection, internalAclID);// result must be checked for internalAclID case
    if (queryResult != null) {

      class EntryData {
        public int m_aclEntryID;
        public String m_internalAclID;
        public String m_principalName;
        public int m_principalType;
        public int m_sortIndex;
        public boolean m_negative;
        public boolean m_propagated;
        public String m_permissionName;
      }

      List entryDataList = new ArrayList(50);

      try {
        while (queryResult.m_result.next()) {
          EntryData entryData = new EntryData();

          entryData.m_aclEntryID = queryResult.m_result.getInt(ACL_ENTRY_ID_COL);
          entryData.m_internalAclID = trim(Escape.unescapeSqlString(queryResult.m_result.getString(ACL_ID_COL)));
          entryData.m_principalName = trim(Escape.unescapeSqlString(queryResult.m_result.getString(PRINCIPAL_NAME_COL)));
          entryData.m_principalType = queryResult.m_result.getInt(PRINCIPAL_TYPE_COL);
          entryData.m_sortIndex = queryResult.m_result.getInt(SORT_INDEX_COL);
          entryData.m_negative = queryResult.m_result.getInt(IS_NEGATIVE_COL) != 0;
          entryData.m_propagated = queryResult.m_result.getInt(IS_PROPAGATED_COL) != 0;
          entryData.m_permissionName = trim(Escape.unescapeSqlString(queryResult.m_result.getString(PERMISSION_NAME_COL)));

          entryDataList.add(entryData);
        }
      }
      finally {
        queryResult.close();
      }

      Iterator iterator = entryDataList.iterator();
      while (iterator.hasNext()) {
        EntryData entryData = (EntryData)iterator.next();
        try {
          IUMPrincipal newPrincipal = getUMPrincipal(entryData.m_principalName, getPrincipalTypeFromTag(entryData.m_principalType), raw);
          if (newPrincipal != null) {
            // ignore users which are not configured in the user management
            IAclPermission permission = uncommittedGetPermission(connection, entryData.m_permissionName);
            JDBCAclEntry entry = ( raw && ( newPrincipal instanceof IUMPrincipal ) )
                                 ? new JDBCAclEntryRaw(this, newPrincipal, entryData.m_negative, permission, entryData.m_sortIndex, entryData.m_propagated)
                                 : new JDBCAclEntry(this, newPrincipal, entryData.m_negative, permission, entryData.m_sortIndex, entryData.m_propagated);
            entry.setIDs(entryData.m_aclEntryID, entryData.m_internalAclID);
            result.add(entry);
          }
        }
        catch (InvalidClassException e) {
          // will not be thrown; just to keep the exception from the method prototype
        }
      }
    }

    return result;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param acls TBD: Description of the incoming method parameter
   * @return the ACEs of the ACLs (don't get them all at once to limit the size
   *      of the SQL statement) entries of the acls table may be null
   * @exception SQLException Exception raised in failure situation
   */
  private AclEntryList[] uncommittedGetAclEntries(Connection connection, AclRec[] acls, boolean raw)
    throws SQLException {
    final int MAX_IDS_PER_QUERY = 100;
      class EntryData {

      public int m_aclEntryID;
      public String m_internalAclID;
      public String m_principalName;
      public int m_principalType;
      public int m_sortIndex;
      public boolean m_negative;
      public boolean m_propagated;
      public String m_permissionName;
    }

    if (acls.length == 0) {
      return new AclEntryList[0];
    }

    HashMap entriesTable = new HashMap(acls.length);
    for (int i = 0; i < acls.length; i++) {
      if (acls[i] != null) {
        entriesTable.put(acls[i].getExternalAclID(), new LinkedList());
      }
    }

    String[] consolidatedIDs = new String[entriesTable.size()];
    Iterator entriesTableIterator = entriesTable.keySet().iterator();
    int consolidatedIdx = 0;
    while (entriesTableIterator.hasNext()) {
      consolidatedIDs[consolidatedIdx++] = (String)entriesTableIterator.next();
    }

    for (int i = 0; i < consolidatedIDs.length; i += MAX_IDS_PER_QUERY) {
      int takeCount = min(consolidatedIDs.length - i, MAX_IDS_PER_QUERY);
      String[] internalTakeIDs = new String[takeCount];
      for (int k = 0; k < takeCount; k++) {
        internalTakeIDs[k] = consolidatedIDs[i + k];
      }

      SqlResult queryResult = uncommittedGetAclEntriesFromDB(connection, internalTakeIDs);// result must be checked for internalAclID case
      if (queryResult != null) {
        try {
          while (queryResult.m_result.next()) {
            EntryData entryData = new EntryData();
            entryData.m_aclEntryID = queryResult.m_result.getInt(ACL_ENTRY_ID_COL);
            entryData.m_internalAclID = trim(Escape.unescapeSqlString(queryResult.m_result.getString(ACL_ID_COL)));
            entryData.m_principalName = trim(Escape.unescapeSqlString(queryResult.m_result.getString(PRINCIPAL_NAME_COL)));
            entryData.m_principalType = queryResult.m_result.getInt(PRINCIPAL_TYPE_COL);
            entryData.m_sortIndex = queryResult.m_result.getInt(SORT_INDEX_COL);
            entryData.m_negative = queryResult.m_result.getInt(IS_NEGATIVE_COL) != 0;
            entryData.m_propagated = queryResult.m_result.getInt(IS_PROPAGATED_COL) != 0;
            entryData.m_permissionName = trim(Escape.unescapeSqlString(queryResult.m_result.getString(PERMISSION_NAME_COL)));
            LinkedList entries = (LinkedList)entriesTable.get(entryData.m_internalAclID);
            if (entries != null) {
              entries.add(entryData);
            }
          }
        }
        finally {
          queryResult.close();
        }
      }
    }

    AclEntryList[] result = new AclEntryList[acls.length];
    for (int i = 0; i < acls.length; i++) {
      if (acls[i] != null) {
        result[i] = new AclEntryList();
        LinkedList entries = (LinkedList)entriesTable.get(acls[i].getExternalAclID());
        Iterator entriesIterator = entries.iterator();
        while (entriesIterator.hasNext()) {
          EntryData entryData = (EntryData)entriesIterator.next();
          try {
            IUMPrincipal principal = getUMPrincipal(entryData.m_principalName, getPrincipalTypeFromTag(entryData.m_principalType), raw);
            if (principal != null) {
              // ignore users which are not configured in the user management
              IAclPermission permission = uncommittedGetPermission(connection, entryData.m_permissionName);
              JDBCAclEntry aclEntry = ( raw && ( principal instanceof IUMPrincipalRaw ) )
                                      ? new JDBCAclEntry(this, principal, entryData.m_negative, permission, entryData.m_sortIndex, entryData.m_propagated)
                                      : new JDBCAclEntry(this, principal, entryData.m_negative, permission, entryData.m_sortIndex, entryData.m_propagated);
              aclEntry.setIDs(entryData.m_aclEntryID, entryData.m_internalAclID);
              result[i].add(aclEntry);
            }
          }
          catch (InvalidClassException e) {
            // will not be thrown; just to keep the exception from the method prototype
          }
        }
      }
      else {
        result[i] = null;
      }
    }

    return result;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclID TBD: Description of the incoming method parameter
   * @return a result set for the ACEs of the ACL (principal may be null;
   *      otherwise only the entries belonging to the pricnipal are returned)
   *      (result must be checked for internalAclID case)
   * @exception SQLException Exception raised in failure situation
   */
  private SqlResult uncommittedGetAclEntriesFromDB(Connection connection, String internalAclID)
    throws SQLException {
    PreparedStatement st = connection.prepareStatement(
      "SELECT * FROM " + ACL_ENTRY_TABLE_NAME + " WHERE " + m_ApplicationAndManagerSelectionCondition
       + " AND " + ACL_HASH_COL + "=? AND " + ACL_ID_COL + "=? ORDER BY " + SORT_INDEX_COL);

    return query(st, 1, Escape.escapeSqlString(internalAclID, 0), 2);
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclIDs TBD: Description of the incoming method parameter
   * @return a result set for the ACEs of the ACLs
   * @exception SQLException Exception raised in failure situation
   */
  private SqlResult uncommittedGetAclEntriesFromDB(Connection connection, String[] internalAclIDs)
    throws SQLException {

    if (internalAclIDs.length == 0) {
      return null;
    }

    String sql = "SELECT * FROM " + ACL_ENTRY_TABLE_NAME + " WHERE " + m_ApplicationAndManagerSelectionCondition + " AND ";

    sql += "(";
    for (int i = 0; i < internalAclIDs.length; i++) {
      if (i > 0) {
        sql += " OR (";
      }
      String s = Escape.escapeSqlString(internalAclIDs[i], 0);
      sql += ACL_HASH_COL + "=" + s.hashCode() + " AND " + ACL_ID_COL + " = '" + s + "'";
      if (i > 0) {
        sql += ")";
      }
    }
    sql += ") ";
    sql += " ORDER BY " + SORT_INDEX_COL;

    return query(connection, sql);
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param permissionName TBD: Description of the incoming method parameter
   * @return a list with the IDs of the ACLs (strings) which use the specified
   *      permissions in one of their ACEs
   * @exception SQLException Exception raised in failure situation
   */
  private List uncommittedGetAffectedAclIDs(Connection connection, String permissionName)
    throws SQLException {

    LinkedList result = new LinkedList();

    String sql = "SELECT DISTINCT " + ACL_ID_COL + " FROM " + ACL_ENTRY_TABLE_NAME + " WHERE " +
      m_ApplicationAndManagerSelectionCondition + " AND " +
      PERMISSION_NAME_COL + " = '" + Escape.escapeSqlString(permissionName, PERMISSION_NAME_LENGTH) + "'";

    SqlResult queryResult = query(connection, sql);
    if (queryResult != null) {
      try {
        while (queryResult.m_result.next()) {
          result.add(trim(Escape.unescapeSqlString(queryResult.m_result.getString(ACL_ID_COL))));
        }
      }
      finally {
        queryResult.close();
      }
    }

    return result;
  }

  /**
   * Add a new ACL
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclID TBD: Description of the incoming method parameter
   * @param objectType TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedAddAcl(Connection connection, String internalAclID, IObjectType objectType)
    throws SQLException {
    String s = Escape.escapeSqlString(internalAclID, 0);
    String t = s;
    if (t.length() > RID_START_LENGTH) {
      t = t.substring(0, RID_START_LENGTH);
    }

    Integer[] segments = makeHashSegments(internalAclID);

    String sql = "INSERT INTO " + ACL_TABLE_NAME + " (OwnerID, Hash, RID, RIDStart, object_type, locking_user, change_level";

    for (int i = 0; i < 10; i++) {
      if (segments[i] != null) {
        sql += ", Seg" + (i + 1);
      }
    }

    sql += ") VALUES (" + "?," + s.hashCode() + ",'" + s + "','" + t + "',"
       + "'" + Escape.escapeSqlString(objectType.getName(), OBJECT_TYPE_LENGTH) + "',"
       + "'', 0";//no locking user and change level 0

    for (int i = 0; i < 10; i++) {
      if (segments[i] != null) {
        sql += ", " + segments[i].intValue();
      }
    }
    sql += ")";

    return update(connection, sql) > 0;
  }

  /**
   * Add a new ACE
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclID TBD: Description of the incoming method parameter
   * @param aclEntry TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   * @exception SQLException Exception raised in failure situation
   */
  private int uncommittedAddAclEntry(Connection connection, String internalAclID, JDBCAclEntry aclEntry)
    throws SQLException {
    int ret = -1;
    IUMPrincipal principal = populate(aclEntry.getPrincipal());
    if (principal == null) {
      throw new java.lang.IllegalArgumentException();
    }

    JDBCPermission permission = (JDBCPermission)aclEntry.getPermission();
    if (permission == null) {
      throw new java.lang.IllegalArgumentException();
    }

    String s = Escape.escapeSqlString(internalAclID, 0);
    String t = s;
    if (t.length() > RID_START_LENGTH) {
      t = t.substring(0, RID_START_LENGTH);
    }
    CallableStatement cs = null;
    try {
      cs = connection.prepareCall("{CALL wcm_acl_insert_entry(?,?,?,?,?,?,?,?,?,?,?)}");
      cs.registerOutParameter(1, Types.INTEGER);
      cs.setInt(2, m_owner);
      cs.setInt(3, s.hashCode());
      cs.setString(4, s);
      cs.setString(5, t);
      cs.setString(6, Escape.escapeSqlString(principal.getId(), PRINCIPAL_NAME_LENGTH));
      cs.setInt(7, getPrincipalTypeTag(principal));
      cs.setInt(8, aclEntry.getSortIndex());
      cs.setBoolean(9, aclEntry.isNegative());
      cs.setBoolean(10, aclEntry.isPropagated());
      cs.setString(11, Escape.escapeSqlString(permission.getName(), PERMISSION_NAME_LENGTH));
      cs.executeUpdate();
      ret = cs.getInt(1);
      cs.close();
      cs = null;
    }
    catch (AclPersistenceException e) {
      // will not be thrown because the aclEntry is not yet assigned to an acl
    }
    finally {
      if (cs != null) {
        cs.close();
      }
    }

    return ret;
  }

  /**
   * Remove an ACE
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param aclEntryID TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedRemoveAclEntry(Connection connection, int aclEntryID)
    throws SQLException {

    String sql = "DELETE FROM " + ACL_ENTRY_TABLE_NAME + " WHERE " +
      ACL_ENTRY_ID_COL + " = " + aclEntryID;

    return update(connection, sql, 0) > 0;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclID TBD: Description of the incoming method parameter
   * @param owner TBD: Description of the incoming method parameter
   * @return true iff the specified principal is an owner of the ACL
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedIsOwner(Connection connection, String internalAclID, IUMPrincipal owner)
    throws SQLException {
    boolean result = false;
    PreparedStatement st = connection.prepareStatement(
      "SELECT * FROM " + OWNER_TABLE_NAME + " WHERE " + m_ApplicationAndManagerSelectionCondition
       + " AND " + ACL_HASH_COL + "=? AND " + ACL_ID_COL + "=?"
       + " AND " + PRINCIPAL_NAME_COL + " = ?" + " AND " + PRINCIPAL_TYPE_COL + " = ?");
    st.setString(4, Escape.escapeSqlString(owner.getId(), PRINCIPAL_NAME_LENGTH));
    st.setInt(5, getPrincipalTypeTag(owner));

    SqlResult queryResult = query(st, 1, Escape.escapeSqlString(internalAclID, 0), 2);
    if (queryResult != null) {
      try {
        if (queryResult.m_result.next()) {
          result = true;
        }
      }
      finally {
        queryResult.close();
      }
    }

    return result;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclID TBD: Description of the incoming method parameter
   * @param owner TBD: Description of the incoming method parameter
   * @return true iff the specified principal is the last owner of the ACL
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedIsLastOwner(Connection connection, String internalAclID, IUMPrincipal owner)
    throws SQLException {
    return uncommittedIsLastOwner(connection, internalAclID, owner.getId(), owner.getType());
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclID TBD: Description of the incoming method parameter
   * @param ownerID TBD: Description of the incoming method parameter
   * @param ownerType TBD: Description of the incoming method parameter
   * @return true iff the specified principal is the last owner of the ACL
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedIsLastOwner(Connection connection, String internalAclID, String ownerID, int ownerType)
    throws SQLException {
    boolean result = false;
    PreparedStatement st = connection.prepareStatement(
      "SELECT * FROM " + OWNER_TABLE_NAME + " WHERE " + m_ApplicationAndManagerSelectionCondition
       + " AND " + ACL_HASH_COL + "=? AND " + ACL_ID_COL + "=?");

    SqlResult queryResult = query(st, 1, Escape.escapeSqlString(internalAclID, 0), 2);
    if (queryResult != null) {
      try {
        if (queryResult.m_result.next()) {
          String name = trim(Escape.unescapeSqlString(queryResult.m_result.getString(PRINCIPAL_NAME_COL)));
          int type = queryResult.m_result.getInt(PRINCIPAL_TYPE_COL);
          if (ownerID.equals(name) && ownerType == getPrincipalTypeFromTag(type)) {
            if (!queryResult.m_result.next()) {
              result = true;
            }
          }
        }
      }
      finally {
        queryResult.close();
      }
    }

    return result;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclID TBD: Description of the incoming method parameter
   * @return the owners of the ACL
   * @exception SQLException Exception raised in failure situation
   */
  private UMPrincipalList uncommittedGetOwners(Connection connection, String internalAclID, boolean raw)
    throws SQLException {

    UMPrincipalList result = new UMPrincipalList();

    SqlResult queryResult = uncommittedGetOwnersFromDB(connection, internalAclID);// result must be checked for character case
    if (queryResult != null) {
      try {
        while (queryResult.m_result.next()) {
          String principalName = trim(Escape.unescapeSqlString(queryResult.m_result.getString(PRINCIPAL_NAME_COL)));
          int principalType = queryResult.m_result.getInt(PRINCIPAL_TYPE_COL);
          IUMPrincipal owner = getUMPrincipal(principalName, getPrincipalTypeFromTag(principalType), raw);
          if (owner != null) {
            result.add(owner);
          }
        }
      }
      finally {
        queryResult.close();
      }
    }

    return result;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param acls TBD: Description of the incoming method parameter
   * @return the owners of the ACLs (don't get them all at once to limit the
   *      size of the SQL statement; entries of the ACL array may be null)
   * @exception SQLException Exception raised in failure situation
   */
  private UMPrincipalList[] uncommittedGetOwners(Connection connection, AclRec[] acls, boolean raw)
    throws SQLException {

    final int MAX_IDS_PER_QUERY = 100;

    if (acls.length == 0) {
      return new UMPrincipalList[0];
    }

    HashMap ownerTable = new HashMap(acls.length);
    for (int i = 0; i < acls.length; i++) {
      if (acls[i] != null) {
        ownerTable.put(acls[i].getExternalAclID(), new UMPrincipalList());
      }
    }

    String[] consolidatedIDs = new String[ownerTable.size()];
    Iterator ownerTableIterator = ownerTable.keySet().iterator();
    int consolidatedIdx = 0;
    while (ownerTableIterator.hasNext()) {
      consolidatedIDs[consolidatedIdx++] = (String)ownerTableIterator.next();
    }

    for (int i = 0; i < consolidatedIDs.length; i += MAX_IDS_PER_QUERY) {
      int takeCount = min(consolidatedIDs.length - i, MAX_IDS_PER_QUERY);
      String[] internalTakeIDs = new String[takeCount];
      for (int k = 0; k < takeCount; k++) {
        internalTakeIDs[k] = consolidatedIDs[i + k];
      }

      SqlResult queryResult = uncommittedGetOwnersFromDB(connection, internalTakeIDs);
      if (queryResult != null) {
        try {
          while (queryResult.m_result.next()) {
            String id = trim(Escape.unescapeSqlString(queryResult.m_result.getString(ACL_ID_COL)));
            String principalName = trim(Escape.unescapeSqlString(queryResult.m_result.getString(PRINCIPAL_NAME_COL)));
            int principalType = queryResult.m_result.getInt(PRINCIPAL_TYPE_COL);
            IUMPrincipal owner = getUMPrincipal(principalName, getPrincipalTypeFromTag(principalType), raw);
            if (owner != null) {
              UMPrincipalList owners = (UMPrincipalList)ownerTable.get(id);
              if (owners != null) {
                owners.add(owner);
              }
            }
          }
        }
        finally {
          queryResult.close();
        }
      }
    }

    UMPrincipalList[] result = new UMPrincipalList[acls.length];
    for (int i = 0; i < acls.length; i++) {
      if (acls[i] != null) {
        result[i] = (UMPrincipalList)ownerTable.get(acls[i].getExternalAclID());
      }
      else {
        result[i] = null;
      }
    }

    return result;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclID TBD: Description of the incoming method parameter
   * @return true iff the ACL has some owners
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedCheckOwners(Connection connection, String internalAclID)
    throws SQLException {

    boolean result = false;

    SqlResult queryResult = uncommittedGetOwnersFromDB(connection, internalAclID);
    if (queryResult != null) {
      try {
        if (queryResult.m_result.next()) {
          result = true;
        }
      }
      finally {
        queryResult.close();
      }
    }

    return result;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclID TBD: Description of the incoming method parameter
   * @return the owners of the ACL (result must be checked for character case)
   * @exception SQLException Exception raised in failure situation
   */
  private SqlResult uncommittedGetOwnersFromDB(Connection connection, String internalAclID)
    throws SQLException {
    PreparedStatement st = connection.prepareStatement(
      "SELECT * FROM " + OWNER_TABLE_NAME + " WHERE " + m_ApplicationAndManagerSelectionCondition
       + " AND " + ACL_HASH_COL + "=? AND " + ACL_ID_COL + "=?");

    return query(st, 1, Escape.escapeSqlString(internalAclID, 0), 2);
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclIDs TBD: Description of the incoming method parameter
   * @return the owners of the ACLs (result must be checked for character case)
   * @exception SQLException Exception raised in failure situation
   */
  private SqlResult uncommittedGetOwnersFromDB(Connection connection, String[] internalAclIDs)
    throws SQLException {

    if (internalAclIDs.length == 0) {
      return null;
    }

    String sql = "SELECT * FROM " + OWNER_TABLE_NAME + " WHERE " + m_ApplicationAndManagerSelectionCondition + " AND ";

    sql += "(";
    for (int i = 0; i < internalAclIDs.length; i++) {
      if (i > 0) {
        sql += " OR (";
      }
      String s = Escape.escapeSqlString(internalAclIDs[i], 0);
      sql += ACL_HASH_COL + "=" + s.hashCode() + " AND " + ACL_ID_COL + " = '" + s + "'";
      if (i > 0) {
        sql += ")";
      }
    }
    sql += ")";

    return query(connection, sql);
  }

  /**
   * Add an owner to the ACL
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclID TBD: Description of the incoming method parameter
   * @param owner TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedAddOwner(Connection connection, String internalAclID, IUMPrincipal owner)
    throws SQLException {

    owner = populate(owner);
    String s = Escape.escapeSqlString(internalAclID, 0);
    String t = s;
    if (t.length() > RID_START_LENGTH) {
      t = t.substring(0, RID_START_LENGTH);
    }
    String sql = "INSERT INTO " + OWNER_TABLE_NAME + " (OwnerID, Hash, RID, RIDStart, principal_name, principal_type) VALUES (" +
      "?," + s.hashCode() + ",'" + s + "','" + t + "'," +
      "'" + Escape.escapeSqlString(owner.getId(), PRINCIPAL_NAME_LENGTH) + "'," +
      getPrincipalTypeTag(owner) + ")";

    return update(connection, sql) > 0;
  }

  /**
   * Remove an owner from the ACL
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclID TBD: Description of the incoming method parameter
   * @param owner TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedRemoveOwner(Connection connection, String internalAclID, IUMPrincipal owner)
    throws SQLException {

    return uncommittedRemoveOwner(connection, internalAclID, owner.getId(), owner.getType());
  }

  /**
   * Remove an owner from the ACL
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclID TBD: Description of the incoming method parameter
   * @param ownerID TBD: Description of the incoming method parameter
   * @param ownerType TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedRemoveOwner(Connection connection, String internalAclID, String ownerID, int ownerType)
    throws SQLException {
    String s = Escape.escapeSqlString(internalAclID, 0);
    String sql = "DELETE FROM " + OWNER_TABLE_NAME + " WHERE " +
      m_ApplicationAndManagerSelectionCondition + " AND " +
      ACL_HASH_COL + "=" + s.hashCode() + " AND " + ACL_ID_COL + " = '" + s + "' AND " +
      PRINCIPAL_NAME_COL + " = '" + Escape.escapeSqlString(ownerID, PRINCIPAL_NAME_LENGTH) + "' AND " +
      PRINCIPAL_TYPE_COL + " = " + getPrincipalTypeTag(ownerType);

    return update(connection, sql) > 0;
  }

  /**
   * Remove the owner from all ACLs
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param ownerID TBD: Description of the incoming method parameter
   * @param ownerType TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedRemoveOwner(Connection connection, String ownerID, int ownerType)
    throws SQLException {

    String sql = "DELETE FROM " + OWNER_TABLE_NAME + " WHERE " +
      m_ApplicationAndManagerSelectionCondition + " AND " +
      PRINCIPAL_NAME_COL + " = '" + Escape.escapeSqlString(ownerID, PRINCIPAL_NAME_LENGTH) + "' AND " +
      PRINCIPAL_TYPE_COL + " = " + getPrincipalTypeTag(ownerType);

    return update(connection, sql) > 0;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param permissionName TBD: Description of the incoming method parameter
   * @return the members of the permission
   * @exception SQLException Exception raised in failure situation
   */
  private IAclPermissionList uncommittedGetPermissionMembers(Connection connection, String permissionName)
    throws SQLException {

    if (permissionName.equals(IAclPermission.ACL_PERMISSION_FULL_CONTROL)) {
      return uncommittedGetAllPermissionsNotFullControl(connection);
    }

    AclPermissionList result = new AclPermissionList();

    PreparedStatement st = connection.prepareStatement(
      "SELECT * FROM " + PERMISSION_MEMBER_TABLE_NAME + " WHERE " + m_ApplicationAndManagerSelectionCondition
       + " AND " + PERMISSION_NAME_COL + " = ?");
    st.setString(2, Escape.escapeSqlString(permissionName, PERMISSION_NAME_LENGTH));

    SqlResult queryResult = query(st);
    if (queryResult != null) {
      LinkedList members = new LinkedList();

      try {
        while (queryResult.m_result.next()) {
          members.add(trim(Escape.unescapeSqlString(queryResult.m_result.getString(PERMISSION_MEMBER_COL))));
        }
      }
      finally {
        queryResult.close();
      }

      Iterator iterator = members.iterator();
      while (iterator.hasNext()) {
        String memberName = (String)iterator.next();
        IAclPermission permission = uncommittedGetPermission(connection, memberName);
        if (permission != null) {
          result.add(permission);
        }
      }
    }

    return result;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @return a list of all permissions (WITHOUT fullcontrol)
   * @exception SQLException Exception raised in failure situation
   */
  private IAclPermissionList uncommittedGetAllPermissionsNotFullControl(Connection connection)
    throws SQLException {
    AclPermissionList result = new AclPermissionList();
    PreparedStatement st = connection.prepareStatement(
      "SELECT * FROM " + PERMISSION_TABLE_NAME + " WHERE " + m_ApplicationAndManagerSelectionCondition);

    SqlResult queryResult = query(st);
    if (queryResult != null) {

class PermissionRec {

        public String m_name;
        public boolean m_predefined;
      }

      LinkedList foundPermissions = new LinkedList();

      try {
        while (queryResult.m_result.next()) {
          String name = trim(Escape.unescapeSqlString(queryResult.m_result.getString(PERMISSION_NAME_COL)));
          boolean predefined = queryResult.m_result.getInt(IS_PREDEFINED_COL) != 0;

          if (!IAclPermission.ACL_PERMISSION_FULL_CONTROL.equals(name)) {
            PermissionRec rec = new PermissionRec();
            rec.m_name = name;
            rec.m_predefined = predefined;
            foundPermissions.add(rec);
          }
        }
      }
      finally {
        queryResult.close();
      }

      Iterator iterator = foundPermissions.iterator();
      while (iterator.hasNext()) {
        PermissionRec rec = (PermissionRec)iterator.next();
        IAclPermissionList members = uncommittedGetPermissionMembers(connection, rec.m_name);
        result.add(new JDBCPermission(this, rec.m_name, rec.m_predefined, members));
      }
    }

    return result;
  }

  /**
   * Remove all owners of the ACL
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclID TBD: Description of the incoming method parameter
   * @exception SQLException Exception raised in failure situation
   */
  private void uncommittedRemoveOwners(Connection connection, String internalAclID)
    throws SQLException {
    String s = Escape.escapeSqlString(internalAclID, 0);
    String sql = "DELETE FROM " + OWNER_TABLE_NAME + " WHERE " +
      m_ApplicationAndManagerSelectionCondition + " AND " +
      ACL_HASH_COL + "=" + s.hashCode() + " AND " + ACL_ID_COL + " = '" + s + "'";

    update(connection, sql);
  }

  /**
   * Remove all ACE of the ACL
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclID TBD: Description of the incoming method parameter
   * @exception SQLException Exception raised in failure situation
   */
  private void uncommittedRemoveAclEntries(Connection connection, String internalAclID)
    throws SQLException {
    String s = Escape.escapeSqlString(internalAclID, 0);
    String sql = "DELETE FROM " + ACL_ENTRY_TABLE_NAME + " WHERE " +
      m_ApplicationAndManagerSelectionCondition + " AND " +
      ACL_HASH_COL + "=" + s.hashCode() + " AND " + ACL_ID_COL + " = '" + s + "'";

    update(connection, sql);
  }


  /**
   * Remove all ACE of the ACL which belong to the specified principal
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param principalID TBD: Description of the incoming method parameter
   * @param principalType TBD: Description of the incoming method parameter
   * @exception SQLException Exception raised in failure situation
   */
  private void uncommittedRemoveAclEntries(Connection connection, String principalID, int principalType)
    throws SQLException {

    String sql = "DELETE FROM " + ACL_ENTRY_TABLE_NAME + " WHERE " +
      m_ApplicationAndManagerSelectionCondition + " AND " +
      PRINCIPAL_NAME_COL + " = '" + Escape.escapeSqlString(principalID, PRINCIPAL_NAME_LENGTH) + "' AND " +
      PRINCIPAL_TYPE_COL + " = " + getPrincipalTypeTag(principalType);

    update(connection, sql);
  }

  /**
   * Insert the initial permissions (i.e. fullcontrol)
   *
   * @param connection TBD: Description of the incoming method parameter
   * @exception SQLException Exception raised in failure situation
   */
  private void uncommittedInsertInitialPermissions(Connection connection)
    throws SQLException {

    uncommittedAddPermission(connection, IAclPermission.ACL_PERMISSION_FULL_CONTROL, true);
  }

  /**
   * Insert the initial supported permissions (i.e. fullcontrol)
   *
   * @param connection TBD: Description of the incoming method parameter
   * @exception SQLException Exception raised in failure situation
   */
  private void uncommittedInsertInitialSupportedPermissions(Connection connection)
    throws SQLException {

    uncommittedAddSupportedPermission(connection, IAclPermission.ACL_PERMISSION_FULL_CONTROL, IObjectType.OBJECT_TYPE_LEAF);
    uncommittedAddSupportedPermission(connection, IAclPermission.ACL_PERMISSION_FULL_CONTROL, IObjectType.OBJECT_TYPE_NODE);
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @return true iff permissions already have been created
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedCheckInitialPermissions(Connection connection)
    throws SQLException {

    boolean result = false;

    String sql = "SELECT * FROM " + PERMISSION_TABLE_NAME + " WHERE " + m_ApplicationAndManagerSelectionCondition;

    SqlResult queryResult = query(connection, sql);
    if (queryResult != null) {
      try {
        if (queryResult.m_result.next()) {
          result = true;
        }
      }
      finally {
        queryResult.close();
      }
    }

    return result;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @return true iff supported permissions already have been assigned
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedCheckInitialSupportedPermissions(Connection connection)
    throws SQLException {

    boolean result = false;

    String sql = "SELECT * FROM " + SUPPORTED_PERMISSION_TABLE_NAME + " WHERE " + m_ApplicationAndManagerSelectionCondition;

    SqlResult queryResult = query(connection, sql);
    if (queryResult != null) {
      try {
        if (queryResult.m_result.next()) {
          result = true;
        }
      }
      finally {
        queryResult.close();
      }
    }

    return result;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param permissionName TBD: Description of the incoming method parameter
   * @param objectTypeName TBD: Description of the incoming method parameter
   * @return true iff the permission is supported for the object type
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedIsSupportedPermission(Connection connection, String permissionName, String objectTypeName)
    throws SQLException {

    boolean result = false;

    String sql = "SELECT * FROM " + SUPPORTED_PERMISSION_TABLE_NAME + " WHERE " +
      m_ApplicationAndManagerSelectionCondition + " AND " +
      PERMISSION_NAME_COL + " = '" + Escape.escapeSqlString(permissionName, PERMISSION_NAME_LENGTH) + "' AND " +
      OBJECT_TYPE_COL + " = '" + Escape.escapeSqlString(objectTypeName, OBJECT_TYPE_LENGTH) + "'";

    SqlResult queryResult = query(connection, sql);
    if (queryResult != null) {
      try {
        if (queryResult.m_result.next()) {
          result = true;
        }
      }
      finally {
        queryResult.close();
      }
    }

    return result;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param objectTypeName TBD: Description of the incoming method parameter
   * @return the supported permissions for the object type
   * @exception SQLException Exception raised in failure situation
   */
  private IAclPermissionList uncommittedGetSupportedPermissions(Connection connection, String objectTypeName)
    throws SQLException {
    AclPermissionList result = new AclPermissionList();
    PreparedStatement st = connection.prepareStatement(
      "SELECT * FROM " + SUPPORTED_PERMISSION_TABLE_NAME + " WHERE " + m_ApplicationAndManagerSelectionCondition
       + " AND " + OBJECT_TYPE_COL + " = ?");
    st.setString(2, Escape.escapeSqlString(objectTypeName, OBJECT_TYPE_LENGTH));

    SqlResult queryResult = query(st);
    if (queryResult != null) {
      List permissionNames = new ArrayList(20);

      try {
        while (queryResult.m_result.next()) {
          String permissionName = trim(Escape.unescapeSqlString(queryResult.m_result.getString(PERMISSION_NAME_COL)));
          permissionNames.add(permissionName);
        }
      }
      finally {
        queryResult.close();
      }

      for (int i = 0, n = permissionNames.size(); i < n; ++i) {
        String name = (String)permissionNames.get(i);
        IAclPermission permission = uncommittedGetPermission(connection, name);
        if (permission != null) {
          result.add(permission);
        }
      }
    }

    if (result != null) {
      cacheAddSupportedPermissions(objectTypeName, result);
    }

    return result;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param permissionName TBD: Description of the incoming method parameter
   * @return IAclPermission object for the specified permission name
   * @exception SQLException Exception raised in failure situation
   */
  private IAclPermission uncommittedGetPermission(Connection connection, String permissionName)
    throws SQLException {
    IAclPermission permission = cacheGetPermission(permissionName);
    if (permission != null) {
      return permission;
    }

    JDBCPermission result = null;
    PreparedStatement st = connection.prepareStatement(
      "SELECT * FROM " + PERMISSION_TABLE_NAME + " WHERE " + m_ApplicationAndManagerSelectionCondition
       + " AND " + PERMISSION_NAME_COL + " = ?");
    st.setString(2, Escape.escapeSqlString(permissionName, PERMISSION_NAME_LENGTH));

    String name = null;
    boolean predefined = false;

    SqlResult queryResult = query(st);
    if (queryResult != null) {
      try {
        if (queryResult.m_result.next()) {
          name = trim(Escape.unescapeSqlString(queryResult.m_result.getString(PERMISSION_NAME_COL)));
          predefined = queryResult.m_result.getInt(IS_PREDEFINED_COL) != 0;
        }
      }
      finally {
        queryResult.close();
      }
    }

    if (name != null) {
      IAclPermissionList members = uncommittedGetPermissionMembers(connection, name);
      result = new JDBCPermission(this, name, predefined, members);
      cacheAddPermission(result);
    }

    return result;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param permissionName TBD: Description of the incoming method parameter
   * @return true iff the permission is user in some ACL
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedIsUsedPermission(Connection connection, String permissionName)
    throws SQLException {

    boolean result = false;

    String sql = "SELECT * FROM " + ACL_ENTRY_TABLE_NAME + " WHERE " +
      m_ApplicationAndManagerSelectionCondition + " AND " +
      PERMISSION_NAME_COL + " = '" + Escape.escapeSqlString(permissionName, PERMISSION_NAME_LENGTH) + "'";

    SqlResult queryResult = query(connection, sql);
    if (queryResult != null) {
      try {
        if (queryResult.m_result.next()) {
          result = true;
        }
      }
      finally {
        queryResult.close();
      }
    }

    return result;
  }

  /**
   * Create a new permission
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param permissionName TBD: Description of the incoming method parameter
   * @param isPredefined TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedAddPermission(Connection connection, String permissionName, boolean isPredefined)
    throws SQLException {

    String sql = "INSERT INTO " + PERMISSION_TABLE_NAME + " (OwnerID, permission_name, is_predefined) VALUES" +
      " (?,'" + Escape.escapeSqlString(permissionName, PERMISSION_NAME_LENGTH) + "'," +
      (isPredefined ? "1" : "0") + ")";

    return update(connection, sql) > 0;
  }

  /**
   * Remove a permission
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param permissionName TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedRemovePermission(Connection connection, String permissionName)
    throws SQLException {

    cacheRemovePermission(permissionName);

    String sql = "DELETE FROM " + PERMISSION_TABLE_NAME + " WHERE " +
      m_ApplicationAndManagerSelectionCondition + " AND " +
      PERMISSION_NAME_COL + " = '" + Escape.escapeSqlString(permissionName, PERMISSION_NAME_LENGTH) + "'";

    return update(connection, sql) > 0;
  }

  /**
   * Assign a supported permission to an object type
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param permissionName TBD: Description of the incoming method parameter
   * @param objectTypeName TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedAddSupportedPermission(Connection connection, String permissionName, String objectTypeName)
    throws SQLException {

    String sql = "INSERT INTO " + SUPPORTED_PERMISSION_TABLE_NAME + " (OwnerID, permission_name, object_type) VALUES (" +
      "?,'" + Escape.escapeSqlString(permissionName, PERMISSION_NAME_LENGTH) + "'," +
      "'" + Escape.escapeSqlString(objectTypeName, OBJECT_TYPE_LENGTH) + "')";

    return update(connection, sql) > 0;
  }

  /**
   * Remove a supported permission assignment
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param permissionName TBD: Description of the incoming method parameter
   * @param objectTypeName TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedRemoveSupportedPermission(Connection connection, String permissionName, String objectTypeName)
    throws SQLException {

    String sql = "DELETE FROM " + SUPPORTED_PERMISSION_TABLE_NAME + " WHERE " +
      m_ApplicationAndManagerSelectionCondition + " AND " +
      PERMISSION_NAME_COL + " = '" + Escape.escapeSqlString(permissionName, PERMISSION_NAME_LENGTH) + "' AND " +
      OBJECT_TYPE_COL + " = '" + Escape.escapeSqlString(objectTypeName, OBJECT_TYPE_LENGTH) + "'";

    return update(connection, sql) > 0;
  }

  /**
   * Add a new member to a permisson
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param permissionName TBD: Description of the incoming method parameter
   * @param memberName TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedAddPermissionMember(Connection connection, String permissionName, String memberName)
    throws SQLException {

    String sql = "INSERT INTO " + PERMISSION_MEMBER_TABLE_NAME + " (OwnerID, permission_name, member_name) VALUES (" +
      "?,'" + Escape.escapeSqlString(permissionName, PERMISSION_NAME_LENGTH) + "'," +
      "'" + Escape.escapeSqlString(memberName, PERMISSION_MEMBER_LENGTH) + "')";

    return update(connection, sql) > 0;
  }

  /**
   * Remove a member from a permission
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param permissionName TBD: Description of the incoming method parameter
   * @param memberName TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedRemovePermissionMember(Connection connection, String permissionName, String memberName)
    throws SQLException {

    String sql = "DELETE FROM " + PERMISSION_MEMBER_TABLE_NAME + " WHERE " +
      m_ApplicationAndManagerSelectionCondition + " AND " +
      PERMISSION_NAME_COL + " = '" + Escape.escapeSqlString(permissionName, PERMISSION_NAME_LENGTH) + "' AND " +
      PERMISSION_MEMBER_COL + " = '" + Escape.escapeSqlString(memberName, PERMISSION_MEMBER_LENGTH) + "'";

    return update(connection, sql) > 0;
  }

  /**
   * Remove all members from a permission
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param permissionName TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedRemovePermissionMembers(Connection connection, String permissionName)
    throws SQLException {

    String sql = "DELETE FROM " + PERMISSION_MEMBER_TABLE_NAME + " WHERE " +
      m_ApplicationAndManagerSelectionCondition + " AND " +
      PERMISSION_NAME_COL + " = '" + Escape.escapeSqlString(permissionName, PERMISSION_NAME_LENGTH) + "'";

    return update(connection, sql) > 0;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param name TBD: Description of the incoming method parameter
   * @return true iff the permission is predefined (and thus cannot be changed)
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedIsPredefinedPermission(Connection connection, String name)
    throws SQLException {

    boolean result = false;

    String sql = "SELECT * FROM " + PERMISSION_TABLE_NAME + " WHERE " +
      m_ApplicationAndManagerSelectionCondition + " AND " +
      PERMISSION_NAME_COL + " = '" + Escape.escapeSqlString(name, PERMISSION_NAME_LENGTH) + "'";

    SqlResult queryResult = query(connection, sql);
    if (queryResult != null) {
      try {
        if (queryResult.m_result.next()) {
          if (queryResult.m_result.getInt(IS_PREDEFINED_COL) != 0) {
            result = true;
          }
        }
      }
      finally {
        queryResult.close();
      }
    }

    return result;
  }

  /**
   * Remove the ACL
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclID TBD: Description of the incoming method parameter
   */
  private void uncommittedRemoveAcl(Connection connection, String internalAclID) {

    try {
      String s = Escape.escapeSqlString(internalAclID, 0);
      String sql = "DELETE FROM " + ACL_TABLE_NAME + " WHERE " +
        m_ApplicationAndManagerSelectionCondition + " AND " +
        ACL_HASH_COL + "=" + s.hashCode() + " AND " + ACL_ID_COL + " = '" + s + "'";

      update(connection, sql);
    }
    catch (Exception e) {
    }
  }

  /**
   * Lock the ACL
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclID TBD: Description of the incoming method parameter
   * @param lockingUser TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedLockAcl(Connection connection, String internalAclID, IUMPrincipal lockingUser)
    throws SQLException {
    lockingUser = populate(lockingUser);
    PreparedStatement st = connection.prepareStatement(
      "UPDATE " + ACL_TABLE_NAME + " SET " + LOCKING_USER_COL + " = ? "
       + " WHERE " + m_ApplicationAndManagerSelectionCondition
       + " AND " + ACL_HASH_COL + "=? AND " + ACL_ID_COL + "=?");
    st.setString(1, Escape.escapeSqlString(lockingUser.getId(), LOCKING_USER_LENGTH));

    return update(connection, st, 2, Escape.escapeSqlString(internalAclID, 0), 3) > 0;
  }

  /**
   * Unlock the ACL
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclID TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedUnlockAcl(Connection connection, String internalAclID)
    throws SQLException {
    PreparedStatement st = connection.prepareStatement(
      "UPDATE " + ACL_TABLE_NAME + " SET " + LOCKING_USER_COL + " = ? "
       + " WHERE " + m_ApplicationAndManagerSelectionCondition
       + " AND " + ACL_HASH_COL + "=? AND " + ACL_ID_COL + "=?");
    st.setString(1, "");
    return update(connection, st, 2, Escape.escapeSqlString(internalAclID, 0), 3) > 0;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclID TBD: Description of the incoming method parameter
   * @return true iff the ACL is locked
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedIsAclLocked(Connection connection, String internalAclID)
    throws SQLException {
    boolean result = false;
    PreparedStatement st = connection.prepareStatement(
      "SELECT * FROM " + ACL_TABLE_NAME + " WHERE " + m_ApplicationAndManagerSelectionCondition
       + " AND " + ACL_HASH_COL + "=? AND " + ACL_ID_COL + "=?");

    SqlResult queryResult = query(st, 1, Escape.escapeSqlString(internalAclID, 0), 2);
    if (queryResult != null) {
      try {
        if (queryResult.m_result.next()) {
          String userID = trim(Escape.unescapeSqlString(queryResult.m_result.getString(LOCKING_USER_COL)));
          if (userID != null) {
            result = userID.length() > 0;
          }
        }
      }
      finally {
        queryResult.close();
      }
    }

    return result;
  }

  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param internalAclID TBD: Description of the incoming method parameter
   * @return the user which locks the ACL (or null if it is not locked)
   * @exception SQLException Exception raised in failure situation
   */
  private IUser uncommittedGetAclLockingUser(Connection connection, String internalAclID)
    throws SQLException {
    IUser result = null;
    PreparedStatement st = connection.prepareStatement(
      "SELECT * FROM " + ACL_TABLE_NAME + " WHERE " + m_ApplicationAndManagerSelectionCondition
       + " AND " + ACL_HASH_COL + "=? AND " + ACL_ID_COL + " = ?");

    SqlResult queryResult = query(st, 1, Escape.escapeSqlString(internalAclID, 0), 2);
    if (queryResult != null) {
      try {
        if (queryResult.m_result.next()) {
          String userID = trim(Escape.unescapeSqlString(queryResult.m_result.getString(LOCKING_USER_COL)));
          if (userID != null && userID.length() > 0) {
            result = (IUser)getUMPrincipal(userID, IUMPrincipal.IUSER, false);
          }
        }
      }
      finally {
        queryResult.close();
      }
    }

    return result;
  }


  /**
   * @param connection TBD: Description of the incoming method parameter
   * @param oldID TBD: Description of the incoming method parameter
   * @param newID TBD: Description of the incoming method parameter
   * @return the internalAclIDs for the specified externAclIDs
   * @exception SQLException Exception raised in failure situation
   */
  /*
   * not used any more
   * private String[] uncommittedGetInternalAclIDs(String[] externalAclIDs) throws SQLException {
   * if (externalAclIDs == null) throw new java.lang.IllegalArgumentException();
   * if (externalAclIDs.length == 0) return new String[0];
   * String[] result = new String[externalAclIDs.length];
   * / resolve short ids (no db lookup necessary)
   * for (int i = 0; i < externalAclIDs.length; i++) {
   * result[i] = Escape.escapeSqlString(externalAclIDs[i], 0);
   * }
   * return result;
   * }
   */
  /**
   * Change ACL ID in the ACLs table
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param oldID TBD: Description of the incoming method parameter
   * @param newID TBD: Description of the incoming method parameter
   * @return the internalAclIDs for the specified externAclIDs
   * @return the internalAclIDs for the specified externAclIDs
   * @return the internalAclIDs for the specified externAclIDs
   * @return the internalAclIDs for the specified externAclIDs
   * @return the internalAclIDs for the specified externAclIDs
   * @return the internalAclIDs for the specified externAclIDs
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedChangeAclID_in_AclTable(Connection connection, String oldID, String newID)
    throws SQLException {
    String o = Escape.escapeSqlString(oldID, 0);
    String n = Escape.escapeSqlString(newID, 0);

    Integer[] segments = makeHashSegments(n);

    String t = n;
    if (t.length() > RID_START_LENGTH) {
      t = t.substring(0, RID_START_LENGTH);
    }
    String sql = "UPDATE " + ACL_TABLE_NAME + " SET " +
      ACL_HASH_COL + "=" + n.hashCode() + ", " + ACL_ID_COL + " = '" + n + "', "
       + ACL_RIDSTART_COL + "='" + t + "'";

    for (int i = 0; i < 10; i++) {
      sql += ", Seg" + (i + 1) + "=" + (segments[i] == null ? "NULL" : segments[i].toString());
    }
    sql += " WHERE " + m_ApplicationAndManagerSelectionCondition + " AND " +
      ACL_HASH_COL + "=" + o.hashCode() + " AND " + ACL_ID_COL + " = '" + o + "'";

    return update(connection, sql) > 0;
  }

  /**
   * Change ACL ID in the ACEs table
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param oldID TBD: Description of the incoming method parameter
   * @param newID TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedChangeAclID_in_AclEntryTable(Connection connection, String oldID, String newID)
    throws SQLException {
    String o = Escape.escapeSqlString(oldID, 0);
    String n = Escape.escapeSqlString(newID, 0);
    String t = n;
    if (t.length() > RID_START_LENGTH) {
      t = t.substring(0, RID_START_LENGTH);
    }
    String sql = "UPDATE " + ACL_ENTRY_TABLE_NAME + " SET " +
      ACL_HASH_COL + "=" + n.hashCode() + ", " + ACL_ID_COL + " = '" + n + "', "
       + ACL_RIDSTART_COL + "='" + t + "' WHERE " +
      m_ApplicationAndManagerSelectionCondition + " AND " +
      ACL_HASH_COL + "=" + o.hashCode() + " AND " + ACL_ID_COL + " = '" + o + "'";

    return update(connection, sql) > 0;
  }

  /**
   * Change ACL ID in the owners table
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param oldID TBD: Description of the incoming method parameter
   * @param newID TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   * @exception SQLException Exception raised in failure situation
   */
  private boolean uncommittedChangeAclID_in_OwnerTable(Connection connection, String oldID, String newID)
    throws SQLException {
    String o = Escape.escapeSqlString(oldID, 0);
    String n = Escape.escapeSqlString(newID, 0);
    String t = n;
    if (t.length() > RID_START_LENGTH) {
      t = t.substring(0, RID_START_LENGTH);
    }
    String sql = "UPDATE " + OWNER_TABLE_NAME + " SET " +
      ACL_HASH_COL + "=" + n.hashCode() + ", " + ACL_ID_COL + " = '" + n + "', "
       + ACL_RIDSTART_COL + "='" + t + "' WHERE " +
      m_ApplicationAndManagerSelectionCondition + " AND " +
      ACL_HASH_COL + "=" + o.hashCode() + " AND " + ACL_ID_COL + " = '" + o + "'";

    return update(connection, sql) > 0;
  }

  /**
   * Execute an update SQL statement
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param sql TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   * @exception SQLException Exception raised in failure situation
   */
  private int update(Connection connection, String sql)
    throws SQLException {
    return update(connection, sql, 1);
  }

  private int update(Connection connection, String sql, int owner)
    throws SQLException {

    int result = 0;
    PreparedStatement statement = connection.prepareStatement(sql);
    if (statement != null) {
      try {
        if (owner > 0) {
          statement.setInt(owner, m_owner);
        }
        result = statement.executeUpdate();
      }
      finally {
        statement.close();
      }
      statement = null;
      try {
        statement = connection.prepareStatement("UPDATE wcm_acl_version SET counter=counter+1");
        statement.executeUpdate();
      }
      catch (SQLException e) {
        s_log.warningT("update(4640)", "can't update wcm_acl_version...");
      }
      finally {
        if (statement != null) {
          statement.close();
        }
      }
    }
    return result;
  }

  /**
   * Execute an update SQL statement
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param st TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   * @exception SQLException Exception raised in failure situation
   */
  private int update(Connection connection, PreparedStatement st)
    throws SQLException {
    return update(connection, st, 1, null, -1);
  }

  private int update(Connection connection, PreparedStatement st, int owner, String rid, int ridIndex)
    throws SQLException {
    int result = 0;
    if (st != null) {
      try {
        if (owner > 0) {
          st.setInt(owner, m_owner);
        }
        if (rid != null) {
          st.setInt(ridIndex, rid.hashCode());
          st.setString(ridIndex + 1, rid);
        }
        result = st.executeUpdate();
      }
      finally {
        st.close();
      }
      st = null;
      try {
        st = connection.prepareStatement("UPDATE wcm_acl_version SET counter=counter+1");
        st.executeUpdate();
      }
      catch (SQLException e) {
        s_log.warningT("update(4687)", "can't update wcm_acl_version...");
      }
      finally {
        if (st != null) {
          st.close();
        }
      }
    }
    return result;
  }

  /**
   * Returns the Modifycounter of the acl tables or -1 if not possible to
   * retrieve.
   *
   * @return dBVersion
   */
  public long getDBVersion() {
    PreparedStatement st = null;
    ResultSet rs = null;
    try {
      st = m_main_connection.prepareCall("SELECT counter FROM wcm_acl_version");
      rs = st.executeQuery();
      if (rs.next()) {
        return rs.getLong(1);
      }
    }
    catch (SQLException ignore) {
      s_log.errorT("getDBVersion(4715)", "Can't retrieve DB Version!");
    }
    finally {
      try {
        if (rs != null) {
          rs.close();
        }
      }
      catch (Exception ignore) {
      }
      try {
        if (st != null) {
          st.close();
        }
      }
      catch (Exception ignore) {
      }
    }
    return -1;
  }

  /**
   * Container for result sets and statements
   */
  private class SqlResult {
    public Statement m_statement = null;
    public ResultSet m_result = null;

    public void close()
      throws SQLException {
      if (m_result != null) {
        m_result.close();
      }
      if (m_statement != null) {
        m_statement.close();
      }
    }
  }

  /**
   * Execute an query SQL statement
   *
   * @param connection TBD: Description of the incoming method parameter
   * @param sql TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   * @exception SQLException Exception raised in failure situation
   */
  private SqlResult query(Connection connection, String sql)
    throws SQLException {
    return query(connection, sql, 1);
  }

  private SqlResult query(Connection connection, String sql, int owner)
    throws SQLException {
    SqlResult result = new SqlResult();

    try {
      PreparedStatement st = connection.prepareStatement(sql);
      if (owner > 0) {
        st.setInt(owner, m_owner);
      }
      result.m_statement = st;
      result.m_result = st.executeQuery();
    }
    catch (SQLException e) {
      if (result.m_result != null) {
        result.m_result.close();
      }
      if (result.m_statement != null) {
        result.m_statement.close();
      }
      throw e;
    }

    if (result.m_statement == null || result.m_result == null) {
      if (result.m_result != null) {
        result.m_result.close();
      }
      if (result.m_statement != null) {
        result.m_statement.close();
      }
      return null;
    }

    return result;
  }

  /**
   * Execute an query SQL statement
   *
   * @param st TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   * @exception SQLException Exception raised in failure situation
   */
  private SqlResult query(PreparedStatement st)
    throws SQLException {
    return query(st, 1, null, -1);
  }

  private SqlResult query(PreparedStatement st, int owner, String rid, int ridIndex)
    throws SQLException {

    SqlResult result = new SqlResult();

    try {
      if (owner > 0) {
        st.setInt(owner, m_owner);
      }
      if (rid != null) {
        st.setInt(ridIndex, rid.hashCode());
        st.setString(ridIndex + 1, rid);
      }
      result.m_statement = st;
      result.m_result = st.executeQuery();
      ;
    }
    catch (SQLException e) {
      if (result.m_result != null) {
        result.m_result.close();
      }
      if (result.m_statement != null) {
        result.m_statement.close();
      }
      throw e;
    }

    if (result.m_statement == null || result.m_result == null) {
      if (result.m_result != null) {
        result.m_result.close();
      }
      if (result.m_statement != null) {
        result.m_statement.close();
      }
      return null;
    }

    return result;
  }

  /**
   * Open a new database connection
   *
   * @return TBD: Description of the outgoing return value
   * @exception SQLException Exception raised in failure situation
   * @exception AclPersistenceException Exception raised in failure situation
   */
  private Connection openConnection()
    throws SQLException, AclPersistenceException {

    JDBCConnectionPoolManager poolManager = JDBCConnectionPoolManager.getInstance();
    if (poolManager == null) {
      throw new AclPersistenceException("JDBCConnectionPoolManager.getInstance() failed");
    }

    JDBCConnectionPool pool = poolManager.getPool(m_poolID);
    if (pool == null) {
      throw new AclPersistenceException("JDBCConnectionPoolManager.getPool() failed");
    }

    Connection connection = pool.getConnection();
    if (connection == null) {
      throw new AclPersistenceException("JDBCConnectionPool.getConnection() failed");
    }

    connection.setAutoCommit(false);
    connection.setTransactionIsolation(Connection.TRANSACTION_READ_COMMITTED);
    return connection;
  }

  /**
   * Close a database connection
   *
   * @param connection TBD: Description of the incoming method parameter
   * @exception SQLException Exception raised in failure situation
   */
  private void closeConnection(Connection connection)
    throws SQLException {

    if (connection == null) {
      throw new java.lang.IllegalArgumentException();
    }
    connection.close();
  }

  /**
   * Reopen the main database connection (done in case of database errors to
   * survive a database restart)
   *
   * @exception SQLException Exception raised in failure situation
   * @exception AclPersistenceException Exception raised in failure situation
   */
  private void reopenMainConnection()
    throws SQLException, AclPersistenceException {

    closeConnection(m_main_connection);
    m_main_connection = openConnection();
  }

  /**
   * @param name TBD: Description of the incoming method parameter
   * @return a string used as a key in the permissions cache
   */
  private String cacheGetPermissionKey(String name) {
    return m_applicationID + SEPARATOR + m_managerID + SEPARATOR + KEY_PERMISSION_PREFIX + SEPARATOR + name;
  }

  /**
   * Get a permission from the cache
   *
   * @param name TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   */
  private IAclPermission cacheGetPermission(String name) {
    if (m_permissionCache != null) {
      try {
        ICacheEntry cacheEntry = m_permissionCache.getEntry(cacheGetPermissionKey(name));
        if (cacheEntry != null) {
          return (IAclPermission)cacheEntry.getObject();
        }
      }
      catch (CacheException e) {
        s_log.warningT("cacheGetPermission(4936)", "permission cache error: " + LoggingFormatter.extractCallstack(e));
      }
    }

    return null;
  }

  /**
   * Add a permission to the cache
   *
   * @param permission TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   */
  private IAclPermission cacheAddPermission(IAclPermission permission) {
    if (m_permissionCache != null) {
      try {
        m_permissionCache.addEntry(cacheGetPermissionKey(permission.getName()), permission);
      }
      catch (Exception e) {
        s_log.warningT("cacheAddPermission(4955)", "permission cache error: " + LoggingFormatter.extractCallstack(e));
      }
    }

    return permission;
  }

  /**
   * Remove a permission from the permissions cache
   *
   * @param permissionName TBD: Description of the incoming method parameter
   */
  private void cacheRemovePermission(String permissionName) {
    if (m_permissionCache != null) {
      try {
        m_permissionCache.removeEntry(cacheGetPermissionKey(permissionName));
      }
      catch (CacheException e) {
        s_log.warningT("cacheRemovePermission(4973)", "permission cache error: " + LoggingFormatter.extractCallstack(e));
      }
    }
  }

  /**
   * @param objectTypeName TBD: Description of the incoming method parameter
   * @return the key for the permissions that apply to the specified object type
   *      in the permissions cache
   */
  private String cacheGetSupportedPermissionsKey(String objectTypeName) {

    return m_applicationID + SEPARATOR + m_managerID + SEPARATOR + KEY_SUPPORTED_PERMISSIONS_PREFIX + SEPARATOR + objectTypeName;
  }

  /**
   * Add supported permissions to the permissions cache
   *
   * @param objectTypeName TBD: Description of the incoming method parameter
   * @param permissions TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   */
  private IAclPermissionList cacheAddSupportedPermissions(String objectTypeName, IAclPermissionList permissions) {
    if (m_permissionCache != null) {
      try {
        m_permissionCache.addEntry(cacheGetSupportedPermissionsKey(objectTypeName), permissions);
      }
      catch (CacheException e) {
        s_log.warningT("cacheAddSupportedPermissions(5001)", "permission cache error: " + LoggingFormatter.extractCallstack(e));
      }
    }

    return permissions;
  }

  /**
   * @param objectTypeName TBD: Description of the incoming method parameter
   * @return supported permissions as they are stored in the permissions cache
   *      (returns null if they are not stored)
   */
  private IAclPermissionList cacheGetSupportedPermissions(String objectTypeName) {
    if (m_permissionCache != null) {
      try {
        ICacheEntry cacheEntry = m_permissionCache.getEntry(cacheGetSupportedPermissionsKey(objectTypeName));
        if (cacheEntry != null) {
          return (IAclPermissionList)cacheEntry.getObject();
        }
      }
      catch (CacheException e) {
        s_log.warningT("cacheGetSupportedPermissions(5022)", "permission cache error: " + LoggingFormatter.extractCallstack(e));
      }
    }

    return null;
  }

  /**
   * @param parent TBD: Description of the incoming method parameter
   * @param child TBD: Description of the incoming method parameter
   * @return true if the child permission is already contained in the parent
   *      permission
   * @exception AclPersistenceException Exception raised in failure situation
   */
  private boolean isPermissionMember(IAclPermission parent, IAclPermission child)
    throws AclPersistenceException {
    IAclPermissionList children = parent.getMembers();
    IAclPermissionListIterator iterator = children.iterator();
    while (iterator.hasNext()) {
      IAclPermission listChild = iterator.next();
      if (listChild.equals(child)) {
        return true;
      }
      if (isPermissionMember(listChild, child)) {
        return true;
      }
    }
    return false;
  }

  /**
   * @param i TBD: Description of the incoming method parameter
   * @param k TBD: Description of the incoming method parameter
   * @return the minimum of the two values
   */
  private int min(int i, int k) {

    return (i < k) ? i : k;
  }

  /**
   * Remove leading and trailing spaces from a string; if the string is null,
   * return an empty string
   *
   * @param s TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   */
  private String trim(String s) {
    if (s == null) {
      return "";
    }
    return s.trim();
  }
}
