/*
 * Copyright (c) 2003 by SAP AG. All Rights Reserved.
 *
 * SAP, mySAP, mySAP.com and other SAP products and
 * services mentioned herein as well as their respective
 * logos are trademarks or registered trademarks of
 * SAP AG in Germany and in several other countries all
 * over the world. MarketSet and Enterprise Buyer are
 * jointly owned trademarks of SAP AG and Commerce One.
 * All other product and service names mentioned are
 * trademarks of their respective companies.
 *
 * @version $Id: DateFormat.java,v 1.4 2005/06/17 12:07:39 sei Exp $
 */

package com.sapportals.wcm.util.http;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;
import java.util.TimeZone;

/**
 * Utility class for handling date conversions as needed in HTTP headers and
 * WEBDAV response documents. <p>
 *
 * Copyright (c) SAP AG 2001-2002
 *
 * @author julian.reschke@greenbytes.de
 * @author stefan.eissing@greenbytes.de
 */
public class DateFormat {

  private final static String FORMAT_ISO8601 = "yyyy-MM-dd'T'HH:mm:ss'Z'";
  private final static String FORMAT_ISO8601_MS = "yyyy-MM-dd'T'HH:mm:ss.SSS'Z'";
  private final static String FORMAT_ISO8601_SHORT = "yyyy-MM-dd'T'HH:mm:ss";
  private final static String FORMAT_ISO8601_SHORT_MS = "yyyy-MM-dd'T'HH:mm:ss.SSS";

  private final static String FORMAT_HTTP_1 = "EEE, dd MMM yyyy HH:mm:ss 'GMT'";
  private final static String FORMAT_HTTP_2a = "dd-MMM-yy HH:mm:ss 'GMT'";
  private final static String FORMAT_HTTP_2b = "dd-MMM-yyyy HH:mm:ss 'GMT'";
  private final static String FORMAT_HTTP_3a = "EEE MMM dd HH:mm:ss yyyy";
  private final static String FORMAT_HTTP_3b = "EEE MMM  d HH:mm:ss yyyy";

  private final static int FORMAT_ISO8601_LEN = FORMAT_ISO8601.length() - 4;
  private final static int FORMAT_ISO8601_SHORT_LEN = FORMAT_ISO8601_SHORT.length() - 2;
  private final static int FORMAT_ISO8601_SHORT_MS_LEN = FORMAT_ISO8601_SHORT_MS.length() - 2;

  private final static TimeZone GMT = TimeZone.getTimeZone("GMT");

  private SimpleDateFormat m_sdfHTTP_1;
  private SimpleDateFormat m_sdfHTTP_2a;
  private SimpleDateFormat m_sdfHTTP_2b;
  private SimpleDateFormat m_sdfHTTP_3a;
  private SimpleDateFormat m_sdfHTTP_3b;
  private SimpleDateFormat m_sdfISO8601;
  private SimpleDateFormat m_sdfISO8601ms;
  private SimpleDateFormat m_sdfISO8601short;
  private SimpleDateFormat m_sdfISO8601shortms;

  public DateFormat() { }

  /**
   * Convert date to RFC1123 format for use in HTTP headers
   *
   * @param date to convert
   * @return converted date
   */
  public String formatHTTP(Date date) {
    checkHTTP_1();
    return m_sdfHTTP_1.format(date);
  }

  /**
   * Convert date to ISO8601 format for use in WEBDAV responses
   *
   * @param date to convert
   * @return converted date
   */
  public String formatISO8601(Date date) {
    checkISO();
    return m_sdfISO8601.format(date);
  }

  /**
   * Convert string in RFC1123 format to java date.
   *
   * @param text to convert
   * @return converted date
   * @throws java.text.ParseException on wrong format
   */
  public Date parseHTTP(String text) throws java.text.ParseException {
    // For syntax of HTTP dates see RFC 2616 Capter 3.3
    text = text.trim();
    if (text.indexOf('-') > 0) {
      int cindex = text.indexOf(',');
      return parse_rfc850time(text, cindex);
    }
    else {
      int cindex = text.indexOf(',');
      switch (cindex) {
        case -1:
          return parse_asctime(text);
        
        case 3:
          // could be alternative 1 from RFC 2616, Ch. 3.3 
          checkHTTP_1();
          return m_sdfHTTP_1.parse(text);
        
        default:
          break;
      }
    }
    throw new ParseException("Unparseable HTTP date: " + text, 0);
  }

  /**
   * Convert string in ISO8601 format to java date.
   *
   * @param text to convert
   * @return converted date
   * @throws java.text.ParseException on wrong format
   */
  public Date parseISO8601(String text)
    throws java.text.ParseException {
    int len = text.length();
    if (len > 0 && text.charAt(len - 1) == 'Z') {
      if (len > FORMAT_ISO8601_LEN) {
        checkISOms();
        return m_sdfISO8601ms.parse(text);
      }
      else {
        checkISO();
        return m_sdfISO8601.parse(text);
      }
    }
    else {
      Date date;
      int slen;
      if (len == FORMAT_ISO8601_SHORT_MS_LEN + 6) {
        checkISOshortms();
        slen = FORMAT_ISO8601_SHORT_MS_LEN;
        date = m_sdfISO8601shortms.parse(text.substring(0, slen));
      }
      else if (len == FORMAT_ISO8601_SHORT_LEN + 6) {
        checkISOshort();
        slen = FORMAT_ISO8601_SHORT_LEN;
        date = m_sdfISO8601short.parse(text.substring(0, slen));
      }
      else {
        throw new java.text.ParseException("timezone wrong format: (" +
          text + ") expected " + (FORMAT_ISO8601_SHORT_LEN + 6), len - 1);
      }

      char sign = text.charAt(slen);
      if (sign != '+' && sign != '-') {
        throw new java.text.ParseException("timezone offset wrong in " + text, slen);
      }
      
      boolean positive = sign == '+';      

      if (text.charAt(slen + 3) != ':') {
        throw new java.text.ParseException("timezone offset wrong in " + text, slen + 3);
      }

      int hours = 0;
      int minutes = 0;
      try {
        hours = Integer.parseInt(text.substring(slen + 1, slen + 3));
        minutes = Integer.parseInt(text.substring(slen + 4));
      }
      catch (NumberFormatException ex) {
        throw new java.text.ParseException("timezone offset not number in " + text, slen + 1);
      }

      long offset = (minutes * 60 * 1000L) + (hours * 60 * 60 * 1000L);
      if (positive) {
        offset = -offset;
      }
      return new Date(date.getTime() + offset);
    }
  }

  // ------------------------------- private ------------------------------

  private void checkISO() {
    if (m_sdfISO8601 == null) {
      m_sdfISO8601 = new SimpleDateFormat(FORMAT_ISO8601, Locale.ENGLISH);
      m_sdfISO8601.setTimeZone(GMT);
    }
  }

  private void checkISOms() {
    if (m_sdfISO8601ms == null) {
      m_sdfISO8601ms = new SimpleDateFormat(FORMAT_ISO8601_MS, Locale.ENGLISH);
      m_sdfISO8601ms.setTimeZone(GMT);
    }
  }

  private void checkISOshort() {
    if (m_sdfISO8601short == null) {
      m_sdfISO8601short = new SimpleDateFormat(FORMAT_ISO8601_SHORT, Locale.ENGLISH);
      m_sdfISO8601short.setTimeZone(GMT);
    }
  }

  private void checkISOshortms() {
    if (m_sdfISO8601shortms == null) {
      m_sdfISO8601shortms = new SimpleDateFormat(FORMAT_ISO8601_SHORT_MS, Locale.ENGLISH);
      m_sdfISO8601shortms.setTimeZone(GMT);
    }
  }

  private void checkHTTP_1() {
    if (m_sdfHTTP_1 == null) {
      m_sdfHTTP_1 = new SimpleDateFormat(FORMAT_HTTP_1, Locale.ENGLISH);
      m_sdfHTTP_1.setTimeZone(GMT);
    }
  }

  private void checkHTTP_2() {
    if (m_sdfHTTP_2a == null) {
      m_sdfHTTP_2a = new SimpleDateFormat(FORMAT_HTTP_2a, Locale.ENGLISH);
      m_sdfHTTP_2a.setTimeZone(GMT);
    }
    if (m_sdfHTTP_2b == null) {
      m_sdfHTTP_2b = new SimpleDateFormat(FORMAT_HTTP_2b, Locale.ENGLISH);
      m_sdfHTTP_2b.setTimeZone(GMT);
    }
  }

  private void checkHTTP_3() {
    if (m_sdfHTTP_3a == null) {
      m_sdfHTTP_3a = new SimpleDateFormat(FORMAT_HTTP_3a, Locale.ENGLISH);
      m_sdfHTTP_3a.setTimeZone(GMT);
    }
    if (m_sdfHTTP_3b == null) {
      m_sdfHTTP_3b = new SimpleDateFormat(FORMAT_HTTP_3b, Locale.ENGLISH);
      m_sdfHTTP_3b.setTimeZone(GMT);
    }
  }
  
  private static final String[] DAY_NAMES_L = {
    "Monday, ",    
    "Tuesday, ",    
    "Wednesday, ",    
    "Thursday, ",    
    "Friday, ",    
    "Saturday, ",    
    "Sunday, ",    
  };
  
  private static final String[] DAY_NAMES_S = {
      "Mon, ",    
      "Tue, ",    
      "Wed, ",    
      "Thu, ",    
      "Fri, ",    
      "Sat, ",    
      "Sun, ",    
  };
  
  private Date parse_rfc850time(String text, int cindex) throws ParseException {
    // parse alternative 2 from RFC 2616, Ch. 3.3
    String stext = null;
    if (cindex == 3) {
      for (int i = 0; i < DAY_NAMES_S.length; ++i) {
        String day = DAY_NAMES_S[i];
        if (text.startsWith(day)) {
          stext = text.substring(day.length());
          break;
        }
      }
    }
    else if (cindex >= 6) {
      for (int i = 0; i < DAY_NAMES_L.length; ++i) {
        String day = DAY_NAMES_L[i];
        if (text.startsWith(day)) {
          stext = text.substring(day.length());
          break;
        }
      }
    }
    
    if (stext != null) {
      checkHTTP_2();
      try {
        return m_sdfHTTP_2a.parse(stext);
      }
      catch (ParseException ex) {
        // $JL-EXC$
        return m_sdfHTTP_2b.parse(stext);
      }
    }
    throw new ParseException("Unparseable HTTP date(cindex="+cindex+"): " + text, 0);
  }

  private Date parse_asctime(String text) throws ParseException {
    // parse alternative 3 from RFC 2616, Ch. 3.3
    checkHTTP_3();
    try {
      return m_sdfHTTP_3a.parse(text);
    }
    catch (ParseException ex) {
      // $JL-EXC$
      return m_sdfHTTP_3b.parse(text);
    }
  }

}
