/*
 * Copyright (c) 2003 by SAP AG. All Rights Reserved.
 *
 * SAP, mySAP, mySAP.com and other SAP products and
 * services mentioned herein as well as their respective
 * logos are trademarks or registered trademarks of
 * SAP AG in Germany and in several other countries all
 * over the world. MarketSet and Enterprise Buyer are
 * jointly owned trademarks of SAP AG and Commerce One.
 * All other product and service names mentioned are
 * trademarks of their respective companies.
 *
 * @version $Id: JDBCConnectionPool.java,v 1.4 2005/01/24 12:49:45 jre Exp $
 */

package com.sapportals.wcm.util.jdbc.connectionpool;

import com.merant.datadirect.jdbc.extensions.ExtEmbeddedConnection;
import com.sapportals.wcm.*;
import java.sql.*;
import java.util.*;

/**
 * This class implements a pool for JDBC Connections. See constructor for
 * connection and configuration options. <p>
 *
 * Copyright (c) SAP AG 2001-2002
 *
 * @author m.breitenfelder@sapportals.com
 * @version $Id: //javabas/com.sapportals.wcm/50_COR/src/java/util/api/com/sapportals/wcm/util/jdbc/connectionpool/JDBCConnectionPool.java#22
 *      $
 */
public class JDBCConnectionPool {

  public final static int MSSQL_LOCKTIMEOUT = 10000;

  public final static int DB_UNKNOWN = 0;
  public final static int DB_MSSQL = 1;
  public final static int DB_ORACLE = 2;

  private final static String DRIVER_MSSQL = "com.sap.portals.jdbc.sqlserver.SQLServerDriver";
  private final static String DRIVER_ORACLE = "com.sap.portals.jdbc.oracle.OracleDriver";

  private static com.sap.tc.logging.Location log = com.sap.tc.logging.Location.getLocation(com.sapportals.wcm.util.jdbc.connectionpool.JDBCConnectionPool.class);

  protected final static int NO_DEFAULT_TRANSACTION_ISOLATION = -1;

  private static boolean versionLogged = false;
  private static long MAX_CONN_AGE = 10 * 60 * 1000 ; /* 10 minutes minimum */

//  private static Class ddSpyDriverClass;
//  static {
//    try {
//      JDBCConnectionPool.ddSpyDriverClass = Class.forName("com.ddtek.jdbcspy.SpyConnection");
//      if (JDBCConnectionPool.ddSpyDriverClass != null) {
//        JDBCConnectionPool.log.infoT("Datadirect JDBC Spy Driver is loaded");
//      }
//    }
//    catch (Exception ignored) {}
//  }

  private String id;
  private String dbDriver;
  private String dbUrl;
  private String dbUser;
  private String dbPass;
  private int maxConns;
  private int timeOut;
  private int defaultTransactionIsolation = NO_DEFAULT_TRANSACTION_ISOLATION;
  private boolean defaultAutoCommit;
  private boolean cacheCallableStatements;

  private boolean supportsTransactionIsolation = false;

  private int checkedOutCount;
  private LinkedList connections = new LinkedList();
  private Map ageMap = Collections.synchronizedMap( new HashMap()) ;

  private int dbType = JDBCConnectionPool.DB_UNKNOWN;

  /**
   * Only for debugging
   */
  private int debug = 0;


  /**
   * Create a new connection pool
   *
   * @param dbUrl The connection url string
   * @param dbUser The user name
   * @param dbPass The password
   * @param initConns The initial number of connections in the pool
   * @param maxConns The maximum numbers of connections in the pool
   * @param timeOut The time in seconds the calling thread will wait if the max.
   *      number of connections has been reached.
   * @param defaultTransactionIsolation
   * @param defaultAutoCommit
   * @param cacheCallableStatements
   * @param id TBD: Description of the incoming method parameter
   * @param dbDriver TBD: Description of the incoming method parameter
   * @exception WcmException Exception raised in failure situation
   */
  public JDBCConnectionPool(
    String id,
    String dbDriver,
    String dbUrl,
    String dbUser,
    String dbPass,
    int initConns,
    int maxConns,
    int timeOut,
    int defaultTransactionIsolation,
    boolean defaultAutoCommit,
    boolean cacheCallableStatements)
    throws WcmException {

    this.id = id;
    this.dbDriver = dbDriver;
    this.dbUrl = dbUrl;
    this.dbUser = dbUser;
    this.dbPass = dbPass;
    this.maxConns = maxConns;
    this.defaultAutoCommit = defaultAutoCommit;
    this.defaultTransactionIsolation = defaultTransactionIsolation;
    this.cacheCallableStatements = cacheCallableStatements;
    this.timeOut = timeOut > 0 ? timeOut : 5;

    if (this.dbDriver.equals(JDBCConnectionPool.DRIVER_MSSQL)) {
      this.dbType = JDBCConnectionPool.DB_MSSQL;
    }
    else if (this.dbDriver.equals(JDBCConnectionPool.DRIVER_ORACLE)) {
      this.dbType = JDBCConnectionPool.DB_ORACLE;
    }

    try {
      Class.forName(this.dbDriver).newInstance();
    }
    catch (Exception ex) {
      throw new WcmException("Error loading class: " + this.dbDriver, ex);
    }
    log.debugT("JDBCConnectionPool(144)", "Loaded driver: " + this.dbDriver);

    // Check if transaction isolation levels are supported
    // and log version info
    Connection conn = null;
    try {
      if (this.dbUser == null) {
        conn = DriverManager.getConnection(this.dbUrl);
      }
      else {
        conn = DriverManager.getConnection(this.dbUrl, this.dbUser, this.dbPass);
      }
      DatabaseMetaData md = conn.getMetaData();
      if (this.dbType == JDBCConnectionPool.DB_ORACLE) {
        /*
         * Oracle doesn't support READ_UNCOMMITTED; therefore, set the flag based
         * on READ_COMMITTED; if supportsTransactionIsolation is true then
         * we need a reasonable level; if the configured level is still at
         * NO_DEFAULT_TRANSACTION_ISOLATION then set it to the default level reported
         * by the database
         */
        this.supportsTransactionIsolation =
          md.supportsTransactionIsolationLevel(Connection.TRANSACTION_READ_COMMITTED);
        if (this.supportsTransactionIsolation
           && (this.defaultTransactionIsolation == NO_DEFAULT_TRANSACTION_ISOLATION)) {
          this.defaultTransactionIsolation = md.getDefaultTransactionIsolation();
        }
      }
      else {
        this.supportsTransactionIsolation =
          md.supportsTransactionIsolationLevel(Connection.TRANSACTION_READ_UNCOMMITTED);
      }
      log.debugT("JDBCConnectionPool(176)", "Support for transaction isolation: " + this.supportsTransactionIsolation);

      if (!JDBCConnectionPool.versionLogged) {
        JDBCConnectionPool.log.infoT("JDBCConnectionPool(179)", "JDBC driver version: " + md.getDriverVersion());
        JDBCConnectionPool.versionLogged = true;
      }
    }
    catch (SQLException ex) {
      log.errorT("JDBCConnectionPool(184)", ex.getMessage() + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(ex));
      throw new WcmException(ex.getMessage(), ex);
    }
    try {
      conn.close();
    }
    catch (SQLException ignored) {
    }

    initPool(initConns);
    log.debugT("JDBCConnectionPool(194)", "Created pool with " + initConns + " connections");
  }

  private void initPool(int initConns)
    throws WcmException {
    for (int i = 0; i < initConns; i++) {
      try {
        this.connections.add(newConnection());
      }
      catch (SQLException ex) {
        throw new WcmException(ex.getMessage(), ex);
      }
    }
  }

  //////////////////////////////////////////////////////////////////////////////
  // public
  //////////////////////////////////////////////////////////////////////////////

  public String getId() {
    return this.id;
  }

  public String toString() {
    return "JDBCConnectionPool[" + this.id + "]";
  }

  /**
   * Get a connection from the pool
   *
   * @return connection
   * @exception SQLException Exception raised in failure situation
   */
  public Connection getConnection()
    throws SQLException {
    Connection conn = getConnection(this.timeOut * 1000);
    return new JDBCConnectionWrapper(conn, this);
  }


  /**
   * Release all connections in the pool
   */
  public synchronized void release() {
    Iterator it = this.connections.iterator();
    while (it.hasNext()) {
      Connection con = (Connection)it.next();
      try {
        con.close();
      }
      catch (SQLException ignored) {
      }
    }
    this.connections.clear();
    this.ageMap.clear() ;
  }


  public void returnCleanConnection(Connection conn)
    throws SQLException {
    if (conn instanceof JDBCConnectionWrapper) {
      ((JDBCConnectionWrapper)conn).returnToPool(true);
    }
    else {
      conn.close();
    }
  }

  public boolean isDeadlock(SQLException e) {
    String state = e.getSQLState();
    if (state != null && state.startsWith("44")) {
      return true;
    }
    return (this.dbType == JDBCConnectionPool.DB_MSSQL && (e.getErrorCode() == 1205 || e.getErrorCode() == 1222))
       || (this.dbType == JDBCConnectionPool.DB_ORACLE && e.getErrorCode() == 60);
  }


  //////////////////////////////////////////////////////////////////////////////
  // protected
  //////////////////////////////////////////////////////////////////////////////


  /**
   * Returns a connection into the pool - called by close() in the connection
   * wrapper
   *
   * @param conn a connection retrieved from the connection pool
   * @param cleanState specifies whether to run cleanup SQL statements on the
   *      connection
   * @param invalidConnection specifies whether any of the statements executed
   *      on the connection threw an exception indicating problems with the
   *      underlying communications channel to the database
   */
  protected synchronized void returnConnection(Connection conn,
    boolean cleanState, boolean invalidConnection) {
    if (conn == null) {
      return;
    }

    /*
     * the previous implementation would just start with a local
     * variable invalidConnection set to 'false'; unfortunately, none
     * of the methods called here detects whether the underlying
     * network connection has been lost as no network communication
     * is involved, and because of that no exception is raised and
     * invalidConnection is never changed to 'true' when the
     * server has gone done and become unreachable. Consequently,
     * the connection is returned to the pool, and handed out again
     * when the pool receives the next request for a connection.
     * Upon usage of the connection the calling code will receive
     * an SQLException, basically telling that nothing worked, and
     * very likely the app just calls 'returnConnection' to return
     * it to the pool manager, and once again the broken state of
     * the connection is not detected, it joins the pool, and deja vu.
     *
     * Ok, one way out of this situation is to send a dummy select
     * statement to the server just before adding the connection
     * to the pool again; that would make sure that communication
     * with the server works. Unfortunately, that adds a performance
     * penalty to every good-case connection return as well.
     *
     * Alternative solution, though not perfect, is to monitor all
     * sql exceptions that are caught in the ConnectionWrapper
     * layer and mark down the state for later retrieval. This
     * is what's implemented here.
     *
     * Why isn't it perfect? Well, there are a bunch of other ways
     * to talk to the server that simply bypass the ConnectionWrapper.
     * Unfortunately, starting to wrap those alternate routes doesn't
     * seem to work out as it means to wrap cursor, result set
     * iterators, etc as well. So, handed-out connections that
     * experience networking problems while accessing a result set
     * would receive the SQLException but without the ConnectionWrapper
     * taking notice. This connection would go back into the pool and
     * the bad networking condition would only be detected when the
     * connection is handed out the next time. Still, that's good
     * enough to make sure that dead connections are not kept in the
     * pool forever.
     */
    if (!cleanState && !invalidConnection) {
      try {
        conn.clearWarnings();

        // Undo all uncommited changes
        //conn.rollback();

        // Reset to defaults
        if (this.supportsTransactionIsolation && this.defaultTransactionIsolation != NO_DEFAULT_TRANSACTION_ISOLATION) {
          if (conn.getTransactionIsolation() != this.defaultTransactionIsolation) {
            conn.setTransactionIsolation(this.defaultTransactionIsolation);
          }
        }
        if (conn.getAutoCommit() != this.defaultAutoCommit) {
          conn.setAutoCommit(this.defaultAutoCommit);
        }
      }
      catch (Exception ex) {
        log.debugT("returnConnection(356)", "Exception while cleaning connection returned to the pool: " + ex.getMessage());

        invalidConnection = true;
      }
    }

    if (!invalidConnection) {
      this.connections.addLast(conn);
    }
    else {
      ageMap.remove ( conn ) ;
      try {
        if (!conn.isClosed()) {
          conn.close();
        }
      }
      catch (SQLException e) {
        /*
         * no logging here as this condition has already been
         * logged
         */
      }
    }
    this.checkedOutCount--;
    // DEBUG
    //System.out.println((++this.debug) + " <<<<<<<<<< returnConnection: " + conn);
    //new Exception((++this.debug) + " <<<<<<<<<< returnConnection: " + conn).printStackTrace();

    // Notify all threads that are waiting if the max. number of connections was reached
    notifyAll();
  }

  protected synchronized void removeConnection(Connection conn) {
    if (conn == null) {
      return;
    }
    ageMap.remove ( conn ) ;
    try {
      conn.rollback();
    }
    catch (SQLException ignore) {
    }
    try {
      conn.close();
    }
    catch (SQLException ignore) {
    }

    this.checkedOutCount--;
    notifyAll();
  }

  protected synchronized void releaseFinalizedConnection() {
    this.checkedOutCount--;
  }

  /**
   * Used by connection wrapper: CallableStatement instances will be cached or
   * not
   *
   * @return cacheCallableStatements
   */
  protected boolean getCacheCallableStatements() {
    return this.cacheCallableStatements;
  }


  //////////////////////////////////////////////////////////////////////////////
  // private
  //////////////////////////////////////////////////////////////////////////////


  /**
   * Get a pooled Connection from the cache or a new one. Wait if all are
   * checked out and the max limit has been reached.
   *
   * @param timeout TBD: Description of the incoming method parameter
   * @return connection
   * @exception SQLException Exception raised in failure situation
   */
  private synchronized Connection getConnection(long timeout)
    throws SQLException {

    long startTime = System.currentTimeMillis();
    long remaining = timeout;
    Connection conn = null;

    // Wait if the max. number of connections was reached
    while ((conn = getPooledConnection()) == null) {
      /*
      * only start waiting if the pool has run out of connections; if 'null'
      * is returned because the underlying physical connection has gotten
      * too old and was dropped then the 'null' only indicates that this loop
      * should retry right away instead of waiting for some application code
      * to return a connection to this pool, of course, all unless the pool
      * is empty
      */
      if (this.connections.size() == 0) {
        try  {
          wait(remaining);
        }
        catch (InterruptedException e) {}
        remaining = timeout - (System.currentTimeMillis() - startTime);
        if (remaining <= 0) {
          // Timed out
          throw new SQLException("getConnection() timed out");
        }
      }
    }

    // Check if the connection is not OK retry with the remaining time
    //
    // greenbytes: this seems an unnecessary burden
    if (false && !isConnectionOK(conn)) {
      return getConnection(remaining);
    }
    this.checkedOutCount++;
    return conn;
  }

  private StringBuffer debugOut = new StringBuffer();

  private boolean isConnectionOK(Connection conn) {
    Statement testStmt = null;
    try {
      if (!conn.isClosed()) {
        // isClosed is not enough. Call createStatement to see if the connection is really alive
        testStmt = conn.createStatement();
        testStmt.close();
      }
      else {
        return false;
      }
    }
    catch (SQLException e) {
      log.debugT("isConnectionOK(493)", e.toString());
      if (testStmt != null) {
        try {
          testStmt.close();
        }
        catch (SQLException ignored) {
        }
      }
      return false;
    }
    return true;
  }

  /**
   * Remove a connection from the pool or create a new one
   *
   * @return pooledConnection
   * @exception SQLException Exception raised in failure situation
   */
  private Connection getPooledConnection() throws SQLException {
    Connection conn = null;
    if (this.connections.size() > 0) {
      conn = (Connection)this.connections.removeFirst();
      /*
      * check connection's age
      */
      Object o = this.ageMap.get(conn) ;
      if ( o == null ) {
        conn.close() ;
        return null ;
      }
      long maxAge = ((Long)o).longValue() ;
      if ( maxAge < System.currentTimeMillis()) {
        /* too old, close and drop it */
        ageMap.remove ( conn ) ;
        conn.close() ;
        if (JDBCConnectionPool.log.beDebug()) {
          log.debugT("closed real connection instance <" + conn.toString() + "> because it has exceeded its maximum lifetime");
        }
        /*
        * if we've just drained the connection list here then let's not
        * return null as that would drive the calling code into the timeout
        * handling section; instead let's take the effort to create a real
        * new connection right away. NOTE that we're not exceeding the maxconn
        * limit as we've just removed one connection here.
        */
        if ( this.connections.isEmpty()) {
          conn = newConnection();
        }
        else {
          conn = null ;
        }
      }
    }
    else if (this.maxConns == 0 || this.checkedOutCount < this.maxConns) {
      conn = newConnection();
    }
    return conn;
  }


  /**
   * Creates a new connection
   *
   * @return TBD: Description of the outgoing return value
   * @exception SQLException Exception raised in failure situation
   */
  private Connection newConnection()
    throws SQLException {
    Connection conn = null;
    if (this.dbUser == null) {
      conn = DriverManager.getConnection(this.dbUrl);
    }
    else {
      conn = DriverManager.getConnection(this.dbUrl, this.dbUser, this.dbPass);
    }

    //log.errorT("getting new connection for "+this.getId());
//    if (JDBCConnectionPool.ddSpyDriverClass != null &&
//        JDBCConnectionPool.ddSpyDriverClass.isInstance(conn)) {
//      boolean unlocked = ((com.ddtek.jdbcspy.SpyConnection)conn).unlock("SAPPORTALSEP50");
//      if (!unlocked) {
//        throw new SQLException("SpyConnection.unlock() returned false");
//      }
//      JDBCConnectionPool.log.infoT("Pool " + this.id + ": Datadirect JDBC Spy Driver is active");
//    }
//    else if (conn instanceof ExtEmbeddedConnection) {

    if (conn instanceof ExtEmbeddedConnection) {
      boolean unlocked = ((ExtEmbeddedConnection)conn).unlock("SAPPORTALSEP50");
      if (!unlocked) {
        throw new SQLException("ExtEmbeddedConnection.unlock() returned false");
      }
    }

    if (this.supportsTransactionIsolation && this.defaultTransactionIsolation != NO_DEFAULT_TRANSACTION_ISOLATION) {
      conn.setTransactionIsolation(this.defaultTransactionIsolation);
    }
    conn.setAutoCommit(this.defaultAutoCommit);

    if (this.dbType == JDBCConnectionPool.DB_MSSQL) {
      Statement stmt = null;
      try {
        stmt = conn.createStatement();
        stmt.executeUpdate("SET LOCK_TIMEOUT " + JDBCConnectionPool.MSSQL_LOCKTIMEOUT);
      }
      finally {
        if (stmt != null) {
          stmt.close();
        }
      }
    }

    /*
    * set the expiration time for this physical db connection to MAX_CONN_AGE
    * plus an offset of 0 to 10 minutes in order to avoid all connections
    * becoming invalid at the same point in time; this way the effective
    * lifetime of the underlying connection is somewhere between 10 and 20
    * minutes.
    */
    this.ageMap.put ( conn, new Long ( Math.round ( Math.random() * 600000.0 )
      + MAX_CONN_AGE + System.currentTimeMillis())) ;
    
    if (JDBCConnectionPool.log.beDebug()) {
      JDBCConnectionPool.log.debugT("Created new connection instance <" + conn.toString() + ">");
    }
    
    return conn;
  }

}
