/*
 * Copyright (c) 2003 by SAP AG. All Rights Reserved.
 *
 * SAP, mySAP, mySAP.com and other SAP products and
 * services mentioned herein as well as their respective
 * logos are trademarks or registered trademarks of
 * SAP AG in Germany and in several other countries all
 * over the world. MarketSet and Enterprise Buyer are
 * jointly owned trademarks of SAP AG and Commerce One.
 * All other product and service names mentioned are
 * trademarks of their respective companies.
 *
 * @version $Id: JDBCConnectionPoolManager.java,v 1.3 2006/01/04 22:44:29 jre Exp $
 */

package com.sapportals.wcm.util.jdbc.connectionpool;

import java.security.AccessController;
import java.security.PrivilegedActionException;
import java.security.PrivilegedExceptionAction;
import java.sql.Connection;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.Set;
import java.util.StringTokenizer;

import com.sap.security.toolkit.login.Credentials;
import com.sapportals.config.fwk.Configuration;
import com.sapportals.config.fwk.IConfigClientContext;
import com.sapportals.config.fwk.IConfigManager;
import com.sapportals.config.fwk.IConfigPlugin;
import com.sapportals.config.fwk.IConfigurable;
import com.sapportals.config.fwk.LockConfigException;
import com.sapportals.portal.security.usermanagement.AbstractBackendConnection;
import com.sapportals.portal.security.usermanagement.ILogonDataBroker;
import com.sapportals.portal.security.usermanagement.IServiceUserFactory;
import com.sapportals.portal.security.usermanagement.IUser;
import com.sapportals.portal.security.usermanagement.UserManagementException;
import com.sapportals.wcm.WcmException;
import com.sapportals.wcm.util.config.ConfigurationProxy;
import com.sapportals.wcm.util.logging.LoggingFormatter;
import com.sapportals.wcm.util.usermanagement.WPUMFactory;

/**
 * Manages all pools defined in config <p>
 *
 * Copyright (c) SAP AG 2001-2002
 *
 * @author m.breitenfelder@sapportals.com
 * @version $Id: //javabas/com.sapportals.wcm/50_COR/src/java/util/api/com/sapportals/wcm/util/jdbc/connectionpool/JDBCConnectionPoolManager.java#16
 *      $
 */
public class JDBCConnectionPoolManager {

  private static com.sap.tc.logging.Location log = com.sap.tc.logging.Location.getLocation(com.sapportals.wcm.util.jdbc.connectionpool.JDBCConnectionPoolManager.class);

  private final static String DRIVER_DATADIRECT_BRANDED_MSSQL = "com.sap.portals.jdbc.sqlserver.SQLServerDriver";
  private final static String DRIVER_DATADIRECT_BRANDED_ORACLE = "com.sap.portals.jdbc.oracle.OracleDriver";
  private final static String DRIVER_DATADIRECT_MSSQL = "com.merant.datadirect.jdbc.sqlserver.SQLServerDriver";
  private final static String DRIVER_DATADIRECT_ORACLE = "com.merant.datadirect.jdbc.oracle.OracleDriver";

  private final static String CONFIG_DB_DRIVER = "driver";
  private final static String CONFIG_DB_USER = "user";
  private final static String CONFIG_DB_PASS = "pass";
  private final static String CONFIG_DB_URL = "url";
  private final static String CONFIG_MIN = "min";
  private final static String CONFIG_MAX = "max";
  private final static String CONFIG_TIMEOUT = "timeout";
  private final static String CONFIG_AUTO_COMMIT = "default-autocommit";
  private final static String CONFIG_ISOLATIONLEVEL = "default-isolationlevel";
  private final static String CONFIG_CACHE_CSTM = "enable-stm-caching";

  private final static String CONFIG_DRIVER_DEFAULT_ISO = "driver_default";

  private final static String CONFIG_MAPUSER = "mapuser";
  private final static String CONFIG_SYSTEM = "system_id";

  private final static String TIL_NONE = "none";
  private final static String TIL_READ_COMMITED = "read_commited";
  private final static String TIL_READ_UNCOMMITED = "read_uncommited";
  private final static String TIL_REPEATABLE_READ = "repeatable_read";
  private final static String TIL_SERIALIZABLE = "serializable";

  private final static int DEFAULT_MIN = 10;
  private final static int DEFAULT_MAX = 50;
  private final static int DEFAULT_TIMEOUT = 15;
  private final static boolean DEFAULT_AUTO_COMMIT = true;
  private final static int DEFAULT_ISOLATIONLEVEL = Connection.TRANSACTION_NONE;
  private final static boolean DEFAULT_CACHE_CSTM = false;

  private static JDBCConnectionPoolManager instance = null;

  private static HashMap poolMap = null;

  //for service user security; added by d034127
  //only used in getServiceUser
  private String serviceUserID = null;

  private JDBCConnectionPoolManager() {
    JDBCConnectionPoolManager.poolMap = new HashMap();
    this.initFromConfig();
  }

  public static synchronized JDBCConnectionPoolManager getInstance() {
    if (JDBCConnectionPoolManager.instance == null) {
      JDBCConnectionPoolManager.initConfig();
      JDBCConnectionPoolManager.instance = new JDBCConnectionPoolManager();
    }
    return JDBCConnectionPoolManager.instance;
  }


  /**
   * Returns the instance of an existing pool for the specified Jdbc URL and
   * user name Returns null if a suitable pool was not found.
   *
   * @param id TBD: Description of the incoming method parameter
   * @return pool
   */
  public JDBCConnectionPool getPool(String id) {
    return (JDBCConnectionPool)JDBCConnectionPoolManager.poolMap.get(id);
  }

  /**
   * Calls release() for all pools
   */
  public synchronized void releaseAll() {
    Iterator it = JDBCConnectionPoolManager.poolMap.values().iterator();
    while (it.hasNext()) {
      JDBCConnectionPool p = (JDBCConnectionPool)it.next();
      p.release();
    }
  }

  public synchronized void resetInstance() {
    this.releaseAll();
    JDBCConnectionPoolManager.instance = null;
  }

  /**
   * Creates a new pool. see constructor of JDBCConnectionPool
   *
   * @param id TBD: Description of the incoming method parameter
   * @param dbDriver TBD: Description of the incoming method parameter
   * @param dbUrl TBD: Description of the incoming method parameter
   * @param dbUser TBD: Description of the incoming method parameter
   * @param dbPass TBD: Description of the incoming method parameter
   * @param initConns TBD: Description of the incoming method parameter
   * @param maxConns TBD: Description of the incoming method parameter
   * @param timeOut TBD: Description of the incoming method parameter
   * @param defaultTransactionIsolation TBD: Description of the incoming method
   *      parameter
   * @param defaultAutoCommit TBD: Description of the incoming method parameter
   * @param cacheCallableStatements TBD: Description of the incoming method
   *      parameter
   * @return TBD: Description of the outgoing return value
   * @exception WcmException If a JDBC errors occured while initializing the new
   *      pool's connections If a pool with the specified id already exists
   */
  public synchronized JDBCConnectionPool createPool(
    String id,
    String dbDriver,
    String dbUrl,
    String dbUser,
    String dbPass,
    int initConns,
    int maxConns,
    int timeOut,
    int defaultTransactionIsolation,
    boolean defaultAutoCommit,
    boolean cacheCallableStatements)
    throws WcmException {

    if (!dbDriver.equals(DRIVER_DATADIRECT_BRANDED_MSSQL) &&
      !dbDriver.equals(DRIVER_DATADIRECT_BRANDED_ORACLE) &&
      !dbDriver.equals(DRIVER_DATADIRECT_MSSQL) &&
      !dbDriver.equals(DRIVER_DATADIRECT_ORACLE)) {
      throw new WcmException("Unsupported JDBC driver: " + dbDriver);
    }

    if (JDBCConnectionPoolManager.poolMap.containsKey(id)) {
      throw new WcmException("Pool already exists: " + id);
    }

    JDBCConnectionPool newPool = new JDBCConnectionPool(id, dbDriver, dbUrl, dbUser, dbPass,
      initConns, maxConns, timeOut,
      defaultTransactionIsolation,
      defaultAutoCommit, cacheCallableStatements);
    JDBCConnectionPoolManager.poolMap.put(id, newPool);
    return newPool;
  }

  public synchronized String[] getPoolIds() {
    Set ks = JDBCConnectionPoolManager.poolMap.keySet();
    String[] rv = new String[ks.size()];
    ks.toArray(rv);
    return rv;
  }


  //////////////////////////////////////////////////////////////////////////////

  private void initFromConfig() {
    try {
      IConfigurable[] configurables = ConfigurationProxy.getInstance().getPluginConfigurables(ConfigurationProxy.CFG_PLUGIN_CM_UTILITIES_CONNECTION_POOLS);
      for (int i = 0; i < configurables.length; i++) {

        String poolid = configurables[i].getId().getValueAsString();
        Properties poolProps = ConfigurationProxy.getConfigurableProperties(configurables[i]);

        if (poolProps != null && poolProps.size() > 0) {

          // Read the parameters
          int imin;

          // Read the parameters
          int imax;

          // Read the parameters
          int itimeout;

          // Read the parameters
          int til;
          boolean ac;
          boolean cstm;

          String driver = poolProps.getProperty(CONFIG_DB_DRIVER);
          String user = poolProps.getProperty(CONFIG_DB_USER);
          String pass = poolProps.getProperty(CONFIG_DB_PASS);
          String url = poolProps.getProperty(CONFIG_DB_URL);
          String min = poolProps.getProperty(CONFIG_MIN);
          String max = poolProps.getProperty(CONFIG_MAX);
          String timeout = poolProps.getProperty(CONFIG_TIMEOUT);
          String autocommit = poolProps.getProperty(CONFIG_AUTO_COMMIT);
          String isolation = poolProps.getProperty(CONFIG_ISOLATIONLEVEL);
          String cachecstm = poolProps.getProperty(CONFIG_CACHE_CSTM);

          String mapuser = poolProps.getProperty(CONFIG_MAPUSER);
          String system_id = poolProps.getProperty(CONFIG_SYSTEM);

          // Validate the parameters
          if (driver == null || driver.length() < 1) {
            log.errorT("initFromConfig(236)", "Missing configuration parameter driver for pool id: " + poolid);
            continue;
          }

          if (mapuser != null && mapuser.length() > 0) {
            if (system_id == null || system_id.length() < 1) {
              log.errorT("initFromConfig(242)", "Missing configuration parameter system_id for pool id: " + poolid);
              continue;
            }
            String userpass[] = new String[2];
            if (!getUserAndPassword(mapuser, system_id, userpass, poolid)) {
              continue;
            }
            user = userpass[0];
            pass = userpass[1];
          }

          if (user == null || user.length() < 1) {
            log.errorT("initFromConfig(254)", "Missing configuration parameter user for pool id: " + poolid);
            continue;
          }
          else {
            if (pass == null || pass.length() < 1) {
              log.errorT("initFromConfig(259)", "Missing configuration parameter pass for pool id: " + poolid);
              continue;
            }
          }
          if (url == null || url.length() < 1) {
            log.errorT("initFromConfig(264)", "Missing configuration parameter url for pool id: " + poolid);
            continue;
          }

          // Optional parameters

          if (min == null || min.length() < 1) {
            imin = DEFAULT_MIN;
          }
          else {
            imin = Integer.parseInt(min);
          }
          if (max == null || max.length() < 1) {
            imax = DEFAULT_MAX;
          }
          else {
            imax = Integer.parseInt(max);
          }
          if (timeout == null || timeout.length() < 1) {
            itimeout = DEFAULT_TIMEOUT;
          }
          else {
            itimeout = Integer.parseInt(timeout);
          }
          if (autocommit == null || autocommit.length() < 1) {
            ac = DEFAULT_AUTO_COMMIT;
          }
          else {
            ac = new Boolean(autocommit).booleanValue();
          }
          if (isolation == null || isolation.length() < 1 || isolation.equals(CONFIG_DRIVER_DEFAULT_ISO)) {
            til = JDBCConnectionPool.NO_DEFAULT_TRANSACTION_ISOLATION;
          }
          else {
            if (isolation.equals(this.TIL_NONE)) {
              til = Connection.TRANSACTION_NONE;
            }
            else if (isolation.equals(TIL_READ_COMMITED)) {
              til = Connection.TRANSACTION_READ_COMMITTED;
            }
            else if (isolation.equals(TIL_READ_UNCOMMITED)) {
              til = Connection.TRANSACTION_READ_UNCOMMITTED;
            }
            else if (isolation.equals(TIL_REPEATABLE_READ)) {
              til = Connection.TRANSACTION_REPEATABLE_READ;
            }
            else if (isolation.equals(TIL_SERIALIZABLE)) {
              til = Connection.TRANSACTION_SERIALIZABLE;
            }
            else {
              til = Connection.TRANSACTION_NONE;
            }
          }
          if (cachecstm == null || cachecstm.length() < 1) {
            cstm = DEFAULT_CACHE_CSTM;
          }
          else {
            cstm = new Boolean(cachecstm).booleanValue();
          }

          // Create JDBCConnectionPool
          try {
            JDBCConnectionPool newPool = new JDBCConnectionPool(poolid, driver, url, user, pass, imin, imax, itimeout, til, ac, cstm);
            JDBCConnectionPoolManager.poolMap.put(poolid, newPool);
            log.infoT("initFromConfig(328)", "JDBCConnectionPool created: " + poolid);
          }
          catch (Exception ex) {
            // jre: catch all exceptions, so that a runtime exception won't
            // abort the loop over all configurables
            log.errorT("initFromConfig(333)", "Exception createing JDBCConnectionPool " + poolid + " :" + ex.getMessage() + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(ex));
          }
        }
        else {
          log.errorT("initFromConfig(337)", "Missing configuration for poolID: " + poolid);
        }
      }
    }
    catch (Exception ex) {
      log.errorT("initFromConfig(342)", "configuration error" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(ex));
    }
  }


  /**
   * Tokenize a string with a separator char into a List
   *
   * @param values TBD: Description of the incoming method parameter
   * @param separator TBD: Description of the incoming method parameter
   * @return TBD: Description of the outgoing return value
   */
  private static List parseValueList(String values, String separator) {
    List list = new ArrayList();
    if (values == null) {
      return list;
    }
    StringTokenizer st = new StringTokenizer(values, separator);
    while (st.hasMoreTokens()) {
      list.add(st.nextToken().trim());
    }
    return list;
  }

  /**
   * TBD: Description of the class.
   */
  public class BackendConnection extends AbstractBackendConnection {
    private final static String USER_KEY = "user";
    private final static String PASSWORD_KEY = "mappedpassword";

    private final static String TYPE_KEY = "Type";
    private final static String PROMPT_KEY = "Prompt";

    private final static String TEXT_TYPE = "text";
    private final static String PASSWORD_TYPE = "password";

    private final static String USER_TEXT = "User";
    private final static String PASSWORD_TEXT = "Password";

    private String m_systemId = null;
    private String m_logonUser = null;
    private String m_logonPassword = null;

    public BackendConnection(String systemId) {
      m_systemId = systemId;
    }

    public String getRemoteApplication() {
      return m_systemId;
    }

    public Credentials getRequiredData() {
      Properties credentials = new Properties();

      credentials.put(USER_KEY + '.' + TYPE_KEY, TEXT_TYPE);
      credentials.put(USER_KEY + '.' + PROMPT_KEY, USER_TEXT);

      credentials.put(PASSWORD_KEY + '.' + TYPE_KEY, PASSWORD_TYPE);
      credentials.put(PASSWORD_KEY + '.' + PROMPT_KEY, PASSWORD_TEXT);

      return new Credentials("", credentials, "");
    }

    public void setLogonData(HashMap env) {
      m_logonUser = (String)env.get(USER_KEY);
      m_logonPassword = (String)env.get(PASSWORD_KEY);
    }

    public String getUser() {
      return m_logonUser;
    }

    public String getPassword() {
      return m_logonPassword;
    }

    public void connect()
      throws UserManagementException { }

    public void logoff() { }
  }

  private boolean getUserAndPassword(String mapuser, String system_id, String[] userpass, String poolid) {
    try {
      IUser user = getServiceUser(mapuser);
      if (user == null) {
        log.errorT("getUserAndPassword(429)", "user " + mapuser + " does not exist in user management - pool " + poolid);
        return false;
      }

      ILogonDataBroker databroker = user.getMappingInfo();
      if (databroker == null) {
        log.errorT("getUserAndPassword(435)", "no mapping info for user " + mapuser + " in user management - pool " + poolid);
        return false;
      }

      BackendConnection connection = new BackendConnection(system_id);
      if (!databroker.enrich(connection)) {
        log.errorT("getUserAndPassword(441)", "no mapping data for user " + mapuser + " in user management - pool " + poolid);
        return false;
      }

      if ((userpass[0] = connection.getUser()) == null) {
        log.errorT("getUserAndPassword(446)", "no mapping data (user) for user " + mapuser + " in user management - pool " + poolid);
      }
      if ((userpass[1] = connection.getPassword()) == null) {
        log.errorT("getUserAndPassword(449)", "no mapping data (password) for user " + mapuser + " in user management - pool " + poolid);
      }

      return true;
    }
    catch (Exception e) {
      log.errorT("getUserAndPassword(455)", "failed to retrieve mapping data for user " + mapuser + " - pool " + poolid + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(e));
      return false;
    }
  }

  private IUser getServiceUserInternal()
    throws UserManagementException {
    IServiceUserFactory userFactory = WPUMFactory.getServiceUserFactory();
    return userFactory == null ? null : userFactory.getServiceUser(this.serviceUserID);
  }

  private IUser getServiceUser(String id)
    throws UserManagementException {
    Object serviceUser = null;
    //variable only for transport of userID to getServiceUserInternal
    this.serviceUserID = id;
    try {
      serviceUser = AccessController.doPrivileged(
        new PrivilegedExceptionAction() {
          public Object run()
            throws UserManagementException {
            //call of secure method
            return JDBCConnectionPoolManager.this.getServiceUserInternal();
          }
        }
        );
    }
    catch (PrivilegedActionException ex) {
      throw new UserManagementException(ex.getMessage());
    }
    //variable only for transport of userID to getServiceUserInternal
    this.serviceUserID = null;
    return (IUser)serviceUser;
  }

  /**
   * If DB type is "oracle" then copy oracle pools as default and delete them.
   */
  private static void initConfig() {
    final String CFG_NAME = "local";
    final String CFG_DBTYPE_CONFIGURABLE = "/cm/utilities/connection_pools/db_type/default";
    final String CFG_TYPE_PARAM = "type";
    final String CFG_INIT_PARAM = "initialized";
    final String MSSQL = "mssql";
    final String ORACLE = "oracle";
    final String CFG_POOL = "/cm/utilities/connection_pools";

    final String CFG_POOL_DEFAULT_1 = "dbcon_rep";
    final String CFG_POOL_DEFAULT_2 = "dbcon_acl";
    final String CFG_POOL_DEFAULT_3 = "dbcon_wcmdb_cm";

    final String CFG_POOL_MSSQL_1 = "dbcon_rep_mssql";
    final String CFG_POOL_MSSQL_2 = "dbcon_acl_mssql";
    final String CFG_POOL_MSSQL_3 = "dbcon_wcmdb_cm_mssql";

    final String CFG_POOL_ORACLE_1 = "dbcon_rep_ora";
    final String CFG_POOL_ORACLE_2 = "dbcon_acl_ora";
    final String CFG_POOL_ORACLE_3 = "dbcon_wcmdb_cm_ora";

    IConfigClientContext context = null;
    IConfigManager cfg = null;
    IConfigPlugin plugin = null;
    try {
      context = IConfigClientContext.createContext(JDBCConnectionPoolManager
        .getConfigServiceUser(), CFG_NAME);
      cfg = Configuration.getInstance().getConfigManager(context);
      plugin = cfg.getAndLockConfigPlugin(CFG_POOL, 60 * 60 * 1000L, 1000L);
      if (plugin == null) {
        throw new RuntimeException(
          "Failed to init. connection pool configuration: missing plugin "
            + CFG_POOL);
      }

      IConfigurable oraPool1 = cfg.getConfigurable(CFG_POOL + "/"
        + CFG_POOL_ORACLE_1);
      IConfigurable oraPool2 = cfg.getConfigurable(CFG_POOL + "/"
        + CFG_POOL_ORACLE_2);
      IConfigurable oraPool3 = cfg.getConfigurable(CFG_POOL + "/"
        + CFG_POOL_ORACLE_3);

      IConfigurable defaultType = cfg.getConfigurable(CFG_DBTYPE_CONFIGURABLE);
      if (defaultType == null) {
        throw new RuntimeException(
          "Failed to init. connection pool configuration: missing configurable "
            + CFG_DBTYPE_CONFIGURABLE);
      }
      String valueType = defaultType.getPropertyValue(CFG_TYPE_PARAM);
      String valueInit = defaultType.getPropertyValue(CFG_INIT_PARAM);

      if (valueInit == null) {
        throw new RuntimeException(
          "Failed to init. connection pool configuration: missing parameter: "
            + CFG_INIT_PARAM);
      }
      else if (valueType == null) {
        throw new RuntimeException(
          "Failed to init. connection pool configuration: missing parameter: "
            + CFG_TYPE_PARAM);
      }
      else if (!valueType.equals(ORACLE) && !valueType.equals(MSSQL)) {
        throw new RuntimeException(
          "Failed to init. connection pool configuration: invalid parameter value: "
            + valueType);
      }

      if (valueInit.equals("false")) {
        if (valueType.equals(ORACLE)) {
          if (oraPool1 != null) {
            plugin.replaceConfigurable(oraPool1.doClone(CFG_POOL_DEFAULT_1));
          }
          if (oraPool2 != null) {
            plugin.replaceConfigurable(oraPool2.doClone(CFG_POOL_DEFAULT_2));
          }
          if (oraPool3 != null) {
            plugin.replaceConfigurable(oraPool3.doClone(CFG_POOL_DEFAULT_3));
          }
        }

        defaultType.setPropertyValue(CFG_INIT_PARAM, "true");
        defaultType.commit();
      }

      // Delete oracle pools
      if (oraPool1 != null) {
        plugin.removeConfigurable(oraPool1.doClone(CFG_POOL_ORACLE_1));
      }
      if (oraPool2 != null) {
        plugin.removeConfigurable(oraPool2.doClone(CFG_POOL_ORACLE_2));
      }
      if (oraPool3 != null) {
        plugin.removeConfigurable(oraPool3.doClone(CFG_POOL_ORACLE_3));
      }
    }
    catch (Exception ex) {
      JDBCConnectionPoolManager.log.fatalT("initConfig(565)", LoggingFormatter
        .extractCallstack(ex));
      throw new RuntimeException(
        "Failed to init. connection pool configuration: " + ex.getMessage());
    }
    finally {
      if (plugin != null) {
        try {
          plugin.unlock(context);
        }
        catch (LockConfigException ignored) {
        }
      }
    }
  }

  private static IUser getConfigServiceUser()
    throws Exception {
    Object serviceUser = null;
    serviceUser = AccessController.doPrivileged(
      new PrivilegedExceptionAction() {
        public Object run() throws Exception {
          ConfigurationProxy configProxy = ConfigurationProxy.getInstance();
          return configProxy.getConfigServiceUser();
        }
      }
      );
    return (IUser)serviceUser;
  }
}
