/*
 * Copyright (c) 2003 by SAP AG. All Rights Reserved.
 *
 * SAP, mySAP, mySAP.com and other SAP products and
 * services mentioned herein as well as their respective
 * logos are trademarks or registered trademarks of
 * SAP AG in Germany and in several other countries all
 * over the world. MarketSet and Enterprise Buyer are
 * jointly owned trademarks of SAP AG and Commerce One.
 * All other product and service names mentioned are
 * trademarks of their respective companies.
 *
 * @version $Id: JDBCConnectionWrapper.java,v 1.2 2005/01/24 12:49:45 jre Exp $
 */

package com.sapportals.wcm.util.jdbc.connectionpool;

import com.sapportals.wcm.util.logging.LoggingFormatter;

import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.SQLWarning;
import java.sql.Statement;
import java.util.HashMap;
import java.util.Map;

/**
 * This is a wrapper for a JDBC Connection.In the close method it will clean up
 * the used connection and put it back into the pool. Also, it will prevent the
 * client from using a connection that it has returned into the pool. <code>
 * CallableStatment</code> instances are cached. They must no longer be used
 * after the connection has been put back into the pool. <p>
 *
 * Copyright (c) SAP AG 2001-2002
 *
 * @author m.breitenfelder@sapportals.com
 * @version $Id: //javabas/com.sapportals.wcm/50_COR/src/java/util/api/com/sapportals/wcm/util/jdbc/connectionpool/JDBCConnectionWrapper.java#17
 *      $
 */
public class JDBCConnectionWrapper implements Connection {

  private final static com.sap.tc.logging.Location log = com.sap.tc.logging.Location.getLocation(com.sapportals.wcm.util.jdbc.connectionpool.JDBCConnectionWrapper.class);

  private final Connection realConn;
  private final JDBCConnectionPool pool;
  private boolean isClosed = false;
  private boolean lostNetworkConnection = false;
  private boolean isAutoCommitting;

  /**
   * Cache for CallableStatement instances
   */
  private HashMap statementCache;

  /**
   * debugging finalizer
   */
  private final Exception ex;

  private int pendingStmtCnt = 0;

  public JDBCConnectionWrapper(Connection realConn, JDBCConnectionPool pool)
    throws SQLException {
    this.realConn = realConn;
    this.isAutoCommitting = realConn.getAutoCommit();
    this.pool = pool;
    if (this.pool.getCacheCallableStatements()) {
      this.statementCache = new HashMap();
    }
    this.ex = new Exception();
    this.pendingStmtCnt = 0;
  }

  public JDBCConnectionPool getPool() {
    return this.pool;
  }

  public void close() throws SQLException {
    if ( this.isClosed ) {
      log.warningT ( "close() of a connection that's already closed" ) ;
      return ;
    }
    if ((this.pendingStmtCnt > 0) && (this.isAutoCommitting == false)) {
      log.errorT("close(85)",
        "returning non-autocommit connection with pending transaction: "
        + LoggingFormatter.extractCallstack(this.ex));
    }
    returnToPool(false);
  }

  public boolean isClosed()
    throws SQLException {
    return this.isClosed;
  }

  public void finalize() {
    if (!this.isClosed) {
      log.errorT("finalize(98)", "finalizing JDBCConnectionWrapper on unclosed connection!" + " - " + LoggingFormatter.extractCallstack(this.ex));
      this.pool.releaseFinalizedConnection();
    }
  }

  // ---------- Connection -------------

  public void clearWarnings()
    throws SQLException {
    if (this.isClosed) {
      throw new SQLException("Pooled connection is closed");
    }
    try {
      this.realConn.clearWarnings();
    }
    catch (SQLException e) {
      evalException(e);
      throw e;
    }
  }

  public void commit()
    throws SQLException {
    if (this.isClosed) {
      throw new SQLException("Pooled connection is closed");
    }
    try {
      if (this.isAutoCommitting) {
        /*
         * jdbc driver silently ignores the commit() when a
         * tx is still open on a connection that has been
         * changed to autocommit mode before callig commit();
         * e.g. the following sequence:
         * getConnection();setAutoCommit(false);sql(delete);
         * setAutoCommit(true);commit() ;
         */
        this.realConn.setAutoCommit(false);
        this.realConn.commit();
        this.realConn.setAutoCommit(true);
      }
      else {
        this.realConn.commit();
      }
      this.pendingStmtCnt = 0;
    }
    catch (SQLException e) {
      evalException(e);
      throw e;
    }
  }

  public Statement createStatement()
    throws SQLException {
    if (this.isClosed) {
      throw new SQLException("Pooled connection is closed");
    }
    try {
      this.pendingStmtCnt++;
      return new JDBCStatementWrapper(this.realConn.createStatement(), this);
    }
    catch (SQLException e) {
      evalException(e);
      throw e;
    }
  }

  public Statement createStatement(int resultsetType, int resultSetConcurrency)
    throws SQLException {
    if (this.isClosed) {
      throw new SQLException("Pooled connection is closed");
    }
    try {
      this.pendingStmtCnt++;
      return new JDBCStatementWrapper(
        this.realConn.createStatement(resultsetType, resultSetConcurrency),
        this);
    }
    catch (SQLException e) {
      evalException(e);
      throw e;
    }
  }

  public boolean getAutoCommit()
    throws SQLException {
    if (this.isClosed) {
      throw new SQLException("Pooled connection is closed");
    }
    return this.isAutoCommitting;
  }

  public String getCatalog()
    throws SQLException {
    if (this.isClosed) {
      throw new SQLException("Pooled connection is closed");
    }
    try {
      return this.realConn.getCatalog();
    }
    catch (SQLException e) {
      evalException(e);
      throw e;
    }
  }

  public DatabaseMetaData getMetaData()
    throws SQLException {
    if (this.isClosed) {
      throw new SQLException("Pooled connection is closed");
    }
    try {
      return this.realConn.getMetaData();
    }
    catch (SQLException e) {
      evalException(e);
      throw e;
    }
  }

  public int getTransactionIsolation()
    throws SQLException {
    if (this.isClosed) {
      throw new SQLException("Pooled connection is closed");
    }
    try {
      return this.realConn.getTransactionIsolation();
    }
    catch (SQLException e) {
      evalException(e);
      throw e;
    }
  }

  public SQLWarning getWarnings()
    throws SQLException {
    if (this.isClosed) {
      throw new SQLException("Pooled connection is closed");
    }
    try {
      return this.realConn.getWarnings();
    }
    catch (SQLException e) {
      evalException(e);
      throw e;
    }
  }

  public boolean isReadOnly()
    throws SQLException {
    if (this.isClosed) {
      throw new SQLException("Pooled connection is closed");
    }
    try {
      return this.realConn.isReadOnly();
    }
    catch (SQLException e) {
      evalException(e);
      throw e;
    }
  }

  public String nativeSQL(String sql)
    throws SQLException {
    if (this.isClosed) {
      throw new SQLException("Pooled connection is closed");
    }
    try {
      this.pendingStmtCnt++;
      return this.realConn.nativeSQL(sql);
    }
    catch (SQLException e) {
      evalException(e);
      throw e;
    }
  }

  public CallableStatement prepareCall(String sql)
    throws SQLException {
    if (this.isClosed) {
      throw new SQLException("Pooled connection is closed");
    }

    CallableStatement statement = null;
    try {
      if (pool.getCacheCallableStatements()) {
        // Use caching for callable statements
        statement = (CallableStatement)this.statementCache.get(sql);
        if (statement == null) {
          statement = new JDBCCallableStatementWrapper(this.realConn.prepareCall(sql), this);
          this.statementCache.put(sql, statement);
        }
        else {
          statement.clearParameters();
          statement.clearWarnings();
          statement.clearBatch();
        }
      }
      else {
        statement = this.realConn.prepareCall(sql);
      }

      this.pendingStmtCnt++;
      return statement;
    }
    catch (SQLException e) {
      evalException(e);
      throw e;
    }
  }

  public CallableStatement prepareCall(String sql, int resultSetType, int resultSetConcurrency)
    throws SQLException {
    if (this.isClosed) {
      throw new SQLException("Pooled connection is closed");
    }

    CallableStatement statement = null;
    try {
      if (pool.getCacheCallableStatements()) {
        // Use caching for callable statements
        statement = (CallableStatement)this.statementCache.get(sql + "~" + resultSetType + "~" + resultSetConcurrency);
        if (statement == null) {
          statement = new JDBCCallableStatementWrapper(this.realConn.prepareCall(sql, resultSetType, resultSetConcurrency), this);
          this.statementCache.put(sql + "~" + resultSetType + "~" + resultSetConcurrency, statement);
        }
        else {
          statement.clearParameters();
          statement.clearWarnings();
          statement.clearBatch();
        }
      }
      else {
        statement = this.realConn.prepareCall(sql, resultSetType, resultSetConcurrency);
      }

      this.pendingStmtCnt++;
      return statement;
    }
    catch (SQLException e) {
      evalException(e);
      throw e;
    }
  }

  public PreparedStatement prepareStatement(String sql)
    throws SQLException {
    if (this.isClosed) {
      throw new SQLException("Pooled connection is closed");
    }
    try {
      this.pendingStmtCnt++;
      return new JDBCPreparedStatementWrapper(
        this.realConn.prepareStatement(sql), this);
    }
    catch (SQLException e) {
      evalException(e);
      throw e;
    }
  }

  public PreparedStatement prepareStatement(String sql, int resultSetType, int resultSetConcurrency)
    throws SQLException {
    if (this.isClosed) {
      throw new SQLException("Pooled connection is closed");
    }
    try {
      this.pendingStmtCnt++;
      return new JDBCPreparedStatementWrapper(
        this.realConn.prepareStatement(sql, resultSetType, resultSetConcurrency),
        this);
    }
    catch (SQLException e) {
      evalException(e);
      throw e;
    }
  }

  public void rollback()
    throws SQLException {
    if (this.isClosed) {
      throw new SQLException("Pooled connection is closed");
    }
    try {
      if (this.isAutoCommitting) {
        /*
         * jdbc driver silently ignores the rollback() when a
         * tx is still open on a connection that has been
         * changed to autocommit mode before callig rollback();
         * e.g. the following sequence:
         * getConnection();setAutoCommit(false);sql(delete);
         * setAutoCommit(true);rollback() ;
         */
        this.realConn.setAutoCommit(false);
        this.realConn.rollback();
        this.realConn.setAutoCommit(true);
      }
      else {
        this.realConn.rollback();
      }
      this.pendingStmtCnt = 0;
    }
    catch (SQLException e) {
      evalException(e);
      throw e;
    }
  }

  public void setAutoCommit(boolean autoCommit)
    throws SQLException {
    if (this.isClosed) {
      throw new SQLException("Pooled connection is closed");
    }
    try {
      if (autoCommit != this.isAutoCommitting) {
        /*
         * complain if switching to autocommit mode while statement
         * is pending on the connection
         */
        if ((this.pendingStmtCnt > 0) && (this.isAutoCommitting == false)) {
          log.errorT(
            "setAutoCommit(418)", "changing connection with pending transaction to autocommit mode!");
        }
        /*
         * if turning autocommit off then the statement counter needs to
         * be reset as the counting is done unconditionally to avoid
         * branch prediction failures
         */
        else if (this.isAutoCommitting == true) {
          this.pendingStmtCnt = 0;
        }
        this.isAutoCommitting = autoCommit;
        this.realConn.setAutoCommit(autoCommit);
      }
    }
    catch (SQLException e) {
      evalException(e);
      throw e;
    }
  }

  public void setCatalog(String catalog)
    throws SQLException {
    if (this.isClosed) {
      throw new SQLException("Pooled connection is closed");
    }
    try {
      this.realConn.setCatalog(catalog);
    }
    catch (SQLException e) {
      evalException(e);
      throw e;
    }
  }

  public void setReadOnly(boolean readOnly)
    throws SQLException {
    if (this.isClosed) {
      throw new SQLException("Pooled connection is closed");
    }
    try {
      this.realConn.setReadOnly(readOnly);
    }
    catch (SQLException e) {
      evalException(e);
      throw e;
    }
  }

  public void setTransactionIsolation(int level)
    throws SQLException {
    if (this.isClosed) {
      throw new SQLException("Pooled connection is closed");
    }
    try {
      if (this.realConn.getTransactionIsolation() != level) {
        this.realConn.setTransactionIsolation(level);
      }
    }
    catch (SQLException e) {
      evalException(e);
      throw e;
    }
  }

  public Map getTypeMap()
    throws SQLException {
    if (this.isClosed) {
      throw new SQLException("Pooled connection is closed");
    }
    try {
      return this.realConn.getTypeMap();
    }
    catch (SQLException e) {
      evalException(e);
      throw e;
    }
  }

  public void setTypeMap(Map map)
    throws SQLException {
    if (this.isClosed) {
      throw new SQLException("Pooled connection is closed");
    }
    try {
      this.realConn.setTypeMap(map);
    }
    catch (SQLException e) {
      evalException(e);
      throw e;
    }
  }

  public void remove() {
    synchronized ( this ) {
      if ( this.isClosed ) {
        log.warningT ( "remove() of a connection that's already closed" ) ;
        return ;
      }
      this.isClosed = true;
    }
    // Put myself back into the pool
    this.pool.removeConnection(this.realConn);
  }

  // -------------------------- package private -----------------------------

  void returnToPool(boolean cleanState) throws SQLException {
    synchronized ( this ) {
      if ( this.isClosed ) {
        log.warningT ( "returnToPool() of a connection that's already closed" );
        return ;
      }
      this.isClosed = true;
    }
    // Put myself back into the pool
    this.pool.returnConnection(this.realConn, cleanState, this.lostNetworkConnection);
  }

  void evalException(SQLException e) {
    if ((e != null) && (!lostNetworkConnection)) {
      String state = e.getSQLState();
      if (state != null && (state.startsWith("08S0")
         || state.startsWith("0800") || state.equals("HY000"))) {
        this.lostNetworkConnection = true;
      }
    }
  }
}
