/**
 * Copyright (c) SAP AG 2001-2002
 *
 * @author Dirk Sabiwalsky
 * @author Helmut Cossmann
 * @version $Id: IDMapper.java,v 1.3 2006/01/04 22:44:29 jre Exp $
 */
package com.sapportals.wcm.util.jdbc.idmapper;

import java.io.UnsupportedEncodingException;
import java.sql.*;
import java.util.*;
import java.util.zip.CRC32;

import com.sapportals.wcm.util.jdbc.connectionpool.*;
import com.sapportals.wcm.util.uuid.*;
import com.sapportals.wcm.util.cache.*;


/**
 * Maps external IDs of arbitrary length to internal IDs with a maximum length of 255. <p>
 * - the mapping is persisted in a JDBC database<br>
 * - each mapping consists of a constant internal ID and a variable internal ID<br>
 * - if an external ID is renamed, the constant internal ID remains unchanged and the variable internal ID changes<br>
 * <p>
 */
public class IDMapper {

  public final static String IDLIST = "IDMapper.ext_id_const_id_list" ;

  public final static int CREATE_CONST_INDEX = 0;
  public final static int CREATE_VAR_INDEX   = 1;

  private final static  String   TABLE_NAME                     = "wcm_id_mapping";
  private final static  String   HASH_COL_NAME                  = "hash";
  private final static  String   EXTERNAL_ID_COL_NAME           = "external";
  private final static  String   CONST_INTERNAL_ID_COL_NAME     = "const_internal";
  private final static  String   VAR_INTERNAL_ID_COL_NAME       = "var_internal";

  private final static  int       EXTERNAL_CONST_INTERNAL_MAP   = 0;
  private final static  int       EXTERNAL_VAR_INTERNAL_MAP     = 1;
  private final static  int       CONST_INTERNAL_EXTERNAL_MAP   = 2;
  private final static  int       VAR_INTERNAL_EXTERNAL_MAP     = 3;
  // private final static  char   CACHE_SEPARATOR               = '~';

  private final static  int       HASH_LENGTH                   = 255;
  private final static  int       EXTERNAL_ID_LENGTH            = 4000;
  private final static  int       INTERNAL_ID_LENGTH            = 255;

  private static        com.sap.tc.logging.Location s_log = com.sap.tc.logging.Location.getLocation(IDMapper.class.getName());

  private static boolean logDebugInfo = false ;

  /*
  * prepared statements:
  */
  private final static String PS_GET_MULTIPLE_ID_MAPPINGS =
    "SELECT LTRIM(RTRIM(" +EXTERNAL_ID_COL_NAME+")),LTRIM(RTRIM("
    +CONST_INTERNAL_ID_COL_NAME+")) FROM " + TABLE_NAME
    +" WHERE " + EXTERNAL_ID_COL_NAME + " = ?"
    +" OR " + EXTERNAL_ID_COL_NAME + " LIKE ?";
  private final static String PS_REMOVE_MULTIPLE_ID_MAPPINGS1 =
    "DELETE FROM " + TABLE_NAME + " WHERE " + HASH_COL_NAME + " = ? AND "
    + EXTERNAL_ID_COL_NAME + " = ?" ;
  private final static String PS_REMOVE_MULTIPLE_ID_MAPPINGS2 =
    "DELETE FROM " + TABLE_NAME + " WHERE " + EXTERNAL_ID_COL_NAME + " LIKE ?";
  private final static String PS_RENAME_MULTIPLE_EXTERNAL_IDS =
    "SELECT "+HASH_COL_NAME+","+EXTERNAL_ID_COL_NAME+","
    +VAR_INTERNAL_ID_COL_NAME+" FROM " + TABLE_NAME + " WHERE "
    +EXTERNAL_ID_COL_NAME + " LIKE ?";
  private final static String PS_CREATE_ID_MAPPING =
    "INSERT INTO " + TABLE_NAME + " VALUES (?,?,?,?)" ;
  private final static String PS_GET_CONST_INTERNAL_ID =
    "SELECT LTRIM(RTRIM(" +EXTERNAL_ID_COL_NAME+")),LTRIM(RTRIM("
    +CONST_INTERNAL_ID_COL_NAME+")) FROM " + TABLE_NAME
    +" WHERE " + HASH_COL_NAME + " = ? AND " + EXTERNAL_ID_COL_NAME + " = ?" ;
  private final static String PS_GET_VAR_INTERNAL_ID =
    "SELECT LTRIM(RTRIM(" +EXTERNAL_ID_COL_NAME+")),LTRIM(RTRIM("
    +VAR_INTERNAL_ID_COL_NAME+")) FROM " + TABLE_NAME + " WHERE "
    +HASH_COL_NAME + " = ? AND " + EXTERNAL_ID_COL_NAME + " = ?" ;
  private final static String PS_INTERNAL_RENAME_EXTERNAL_ID =
    "UPDATE " + TABLE_NAME + " SET " + HASH_COL_NAME + " = ?, "
    +EXTERNAL_ID_COL_NAME + " = ?, " + VAR_INTERNAL_ID_COL_NAME
    +" = ? WHERE " + CONST_INTERNAL_ID_COL_NAME + " = ? " ;
  private final static String PS_REMOVE_ID_MAPPING =
    "DELETE FROM " + TABLE_NAME
    +" WHERE " + HASH_COL_NAME + " = ? AND " + EXTERNAL_ID_COL_NAME + " = ?" ;
  private final static String PS_GET_EXTERNAL_ID_BY_VAR =
    "SELECT LTRIM(RTRIM("+VAR_INTERNAL_ID_COL_NAME +")), LTRIM(RTRIM("
    +EXTERNAL_ID_COL_NAME+")) FROM " + TABLE_NAME + " WHERE "
    +VAR_INTERNAL_ID_COL_NAME +" = ? " ;
  private final static String PS_GET_EXTERNAL_ID_BY_CONST =
    "SELECT LTRIM(RTRIM("+CONST_INTERNAL_ID_COL_NAME +")), LTRIM(RTRIM("
    +EXTERNAL_ID_COL_NAME+")) FROM " + TABLE_NAME + " WHERE "
    +CONST_INTERNAL_ID_COL_NAME +" = ? " ;

  /*
  * statement and result set column indices:
  */
  private final static int PS_GET_MUL_EXT_CI = 1 ;
  private final static int PS_GET_MUL_CINT_CI = 2 ;
  private final static int PS_REM_MUL2_EXT_CI = 1 ;
  private final static int PS_REM_HSH_CI = 1 ;
  private final static int PS_REM_EXT_CI = 2 ;
  private final static int PS_REM_MUL1_HSH_CI = 1 ;
  private final static int PS_REM_MUL1_EXT_CI = 2 ;
  private final static int PS_REN_MUL_EXT_CI = 1 ;
  private final static int RS_REN_MUL_HSH_CI = 1 ;
  private final static int RS_REN_MUL_EXT_CI = 2 ;
  private final static int RS_REN_MUL_VINT_CI = 3 ;
  private final static int PS_CRT_HSH_CI = 1 ;
  private final static int PS_CRT_EXT_CI = 2 ;
  private final static int PS_CRT_CINT_CI = 3 ;
  private final static int PS_CRT_VINT_CI = 4 ;
  private final static int PS_GETCONST_HSH_CI = 1 ;
  private final static int PS_GETCONST_EXT_CI = 2 ;
  private final static int RS_GETCONST_EXT_CI = 1 ;
  private final static int RS_GETCONST_CINT_CI = 2 ;
  private final static int PS_GETVAR_HSH_CI = 1 ;
  private final static int PS_GETVAR_EXT_CI = 2 ;
  private final static int RS_GETVAR_EXT_CI = 1 ;
  private final static int RS_GETVAR_VINT_CI = 2 ;
  private final static int RS_GETDB_INT_CI = 1 ;
  private final static int RS_GETDB_EXT_CI = 2 ;
  private final static int PS_GETEXTBYCONST_CINT_CI = 1 ;
  private final static int RS_GETEXTBYCONST_CINT_CI = 1 ;
  private final static int RS_GETEXTBYCONST_EXT_CI = 2 ;
  private final static int PS_GETEXTBYVAR_VINT_CI = 1 ;
  private final static int RS_GETEXTBYVAR_CINT_CI = 1 ;
  private final static int RS_GETEXTBYVAR_EXT_CI = 2 ;
  private final static int PS_REN_EXT_HSH_CI = 1 ;
  private final static int PS_REN_EXT_EXT_CI = 2 ;
  private final static int PS_REN_EXT_VINT_CI = 3 ;
  private final static int PS_REN_EXT_CINT_CI = 4 ;

  private final static int LOOKUP_BY_CONST_INT = 1 ;
  private final static int LOOKUP_BY_VAR_INT = 2 ;

  private final static String[] PS_GET_CONST_INTERNAL_IDS_FROM_DB ;
  private final static String[] PS_GET_VAR_INTERNAL_IDS_FROM_DB ;


  /*
  * thread to handle long running database operations that
  * would negatively affect user experience when executed
  * synchronously
  */
  private static ArrayList jobList = new ArrayList();

  private static class JobEntry {
    JDBCConnectionPool p ;
    String s ;

    public JobEntry ( JDBCConnectionPool p, String s ) {
      this.p = p ;
      this.s = s ;
    }
  }

  private static JobThread jobThread ;

  private static class JobThread extends Thread {
    boolean active ;

    public JobThread() {
      active = true ;
    }

    public void run() {
      synchronized ( jobList ) {
        jobList.ensureCapacity ( 128 ) ;
      }
      ArrayList localList = null ;
      int lls, i ;
      Connection c = null ;
      PreparedStatement ps = null ;
      while  ( active ) {
        /*
        * read size unsynchronized as we can live with not exactly
        * seeing all entries at this point in time
        */
        if (( lls = jobList.size()) > 0 ) {
          synchronized ( jobList ) {
            localList = (ArrayList)(jobList.clone()) ;
            jobList.subList(0,lls).clear() ;
          }
          for ( i = 0 ; i < localList.size() ; i++ ) {
            JobEntry je = (JobEntry)localList.get ( i ) ;
            try {
              c = getConnection ( je.p ) ;
              c.setAutoCommit ( false ) ;
              IDDeadlockHandler deadlockHandler = new IDDeadlockHandler(je.p, c);
              SQLException exception;
              long uc = 0, t0 = System.currentTimeMillis();
              do {
                try {
                  exception = null;
                  if (ps != null) {
                    //retry handling; the connection has already been rolled back!
                    try {
                      ps.close();
                    }
                    catch (SQLException ignored) {
                      //$JLIN-EXC
                      /*
                      * forced close of statement as we're retrying with a new
                      * statement
                      */
                    }
                  }
                  ps = c.prepareStatement ( PS_REMOVE_MULTIPLE_ID_MAPPINGS2 ) ;
                  ps.setString ( PS_REM_MUL2_EXT_CI, je.s + "/%" ) ;
                  uc = ps.executeUpdate();
                }
                catch (SQLException exception2) {
                  exception = exception2;
                }
              } while(deadlockHandler.isTransientDeadlock(exception));
              c.commit() ;
              t0 = System.currentTimeMillis() - t0;
              if (s_log.beDebug()) {
                s_log.debugT("mapper thread: " + uc + " deep entries for prefix ["
                  + je.s + "/%] deleted in " + t0 + " ms");
              }
            }
            catch ( Exception e ) {
              try {
                c.rollback() ;
              }
              catch ( SQLException e2 ) {
                s_log.errorT( "rollback error" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(e2));
              }
              s_log.errorT( "mapper thread exception" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(e));
            }
            finally {
              try {
                if ( ps != null ) {
                  ps.close() ;
                  ps = null ;
                }
                if ( c != null ) {
                  c.close() ;
                }
              }
              catch ( SQLException e ) {
                s_log.errorT( "error closing prepared statement" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(e));
              }
            }
          }
          /*
          * had some work to do which means we've spent some time
          * which comes with a certain probability of more work
          * having piled up; so check the queue again right away
          */
        }
        else {
          /*
          * nothing found on queue, take a nap
          */
          try {
            synchronized ( jobList ) {
              jobList.wait() ;
            }
          }
          catch ( InterruptedException e ) {
          }
        }
      }
    }
  }

  static {
    /*
    * keep at least 100 prepared statement strings available as
    * getInternalIDsFromDB() is using chunks of 100
    */
    PS_GET_CONST_INTERNAL_IDS_FROM_DB = new String[128] ;
    PS_GET_CONST_INTERNAL_IDS_FROM_DB[0] = "SELECT LTRIM(RTRIM("
      +CONST_INTERNAL_ID_COL_NAME+")), LTRIM(RTRIM("+EXTERNAL_ID_COL_NAME
      +")) FROM "+TABLE_NAME+" WHERE ("+HASH_COL_NAME+" = ? AND "
      +EXTERNAL_ID_COL_NAME+" = ?)";
    for ( int i = 1 ; i < PS_GET_CONST_INTERNAL_IDS_FROM_DB.length ; i++ ) {
      PS_GET_CONST_INTERNAL_IDS_FROM_DB[i] =
        PS_GET_CONST_INTERNAL_IDS_FROM_DB[i-1] +" OR (" + HASH_COL_NAME
        +" = ? AND " + EXTERNAL_ID_COL_NAME + " = ?) " ;
    }
    PS_GET_VAR_INTERNAL_IDS_FROM_DB =
      new String[PS_GET_CONST_INTERNAL_IDS_FROM_DB.length] ;
    PS_GET_VAR_INTERNAL_IDS_FROM_DB[0] = "SELECT LTRIM(RTRIM("
      +VAR_INTERNAL_ID_COL_NAME+")), LTRIM(RTRIM("+EXTERNAL_ID_COL_NAME
      +")) FROM " + TABLE_NAME+" WHERE (" + HASH_COL_NAME + " = ? AND "
      +EXTERNAL_ID_COL_NAME+" = ?)" ;
    for ( int i = 1 ; i < PS_GET_VAR_INTERNAL_IDS_FROM_DB.length ; i++ ) {
      PS_GET_VAR_INTERNAL_IDS_FROM_DB[i] =
        PS_GET_VAR_INTERNAL_IDS_FROM_DB[i-1] +" OR (" + HASH_COL_NAME
        +" = ? AND " + EXTERNAL_ID_COL_NAME + " = ?) " ;
    }
    jobThread = new JobThread() ;
    jobThread.setDaemon ( true ) ;
    jobThread.start() ;
  }

  // ------------------------------------------------------------------------
  public static void configure(JDBCConnectionPool pool,
  IClusterCacheArea cacheArea) {
  }

  // ------------------------------------------------------------------------
  public static void terminate() {
    jobThread.active = false;
    jobThread.interrupt();
  }

  /**
   * @param externalID  a external ID
   * @return            the hash code for an external ID as a long
   */
  public static long getHashCodeAsLong(String externalID) {
    if(externalID == null) {
      throw new java.lang.IllegalArgumentException();
    }
    CRC32  crc  = new CRC32();
    try {
      crc.update(externalID.getBytes("UTF-8"));
    }
    catch (UnsupportedEncodingException ex) {
      throw new RuntimeException(ex.getMessage());
    }
    return crc.getValue() ;
  }

  /**
   * @param externalID  a external ID
   * @return            the hash code for an external ID
   */
  public static String getHashCode(String externalID) {
    if(externalID == null) {
      throw new java.lang.IllegalArgumentException();
    }
    CRC32  crc  = new CRC32();
    try {
      crc.update(externalID.getBytes("UTF-8"));
    }
    catch (UnsupportedEncodingException ex) {
      throw new RuntimeException(ex.getMessage());
    }    
    return new Long(crc.getValue()).toString();
  }

  public static String getConstInternalID ( JDBCConnectionPool pool,
  String externalID) throws IDMapperException {
    if(pool == null || externalID == null) {
      throw new java.lang.IllegalArgumentException();
    }
    Connection connection = getConnection(pool);
    try {
      connection.setAutoCommit(false);
      String result = null;
      IDDeadlockHandler deadlockHandler = new IDDeadlockHandler(pool, connection);
      IDMapperException exception;
      do {
        try {
          exception = null;
          result = getConstInternalID ( connection, externalID ) ;
        }
        catch (IDMapperException exception2) {
          exception = exception2;
        }
      } while (deadlockHandler.isTransientDeadlock(exception));
      connection.commit();
      return result;
    }
    catch (IDMapperException e) {
      try {
        connection.rollback();
      }
      catch (SQLException sqlException) {
        s_log.errorT("rollback error" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
      throw e;
    }
    catch (Exception e) {
      try {
        connection.rollback();
      }
      catch (SQLException sqlException) {
        s_log.errorT("rollback error" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
      throw new IDMapperException(e);
    }
    finally {
      try {
        connection.close();
      }
      catch (SQLException sqlException) {
        s_log.errorT("error closing connection" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
    }
  }

  public static String[] getConstInternalIDs ( JDBCConnectionPool pool,
  String[] externalIDs ) throws IDMapperException {
    return getInternalIDs ( pool, externalIDs, LOOKUP_BY_CONST_INT ) ;
  }

  public static String[] getVarInternalIDs ( JDBCConnectionPool pool,
  String[] externalIDs ) throws IDMapperException {
    return getInternalIDs ( pool, externalIDs, LOOKUP_BY_VAR_INT ) ;
  }

  static String[] getInternalIDs ( JDBCConnectionPool pool,
  String[] externalIDs, int lookup ) throws IDMapperException {
    if(pool == null || externalIDs == null) {
      throw new java.lang.IllegalArgumentException();
    }
    if(externalIDs.length == 0) {
      return new String[0];
    }
    String[] localResult = null ;
    Connection connection = getConnection(pool);
    try {
      connection.setAutoCommit(false);
      IDDeadlockHandler deadlockHandler = new IDDeadlockHandler(pool, connection);
      IDMapperException exception;
      do {
        try {
          exception = null;
          localResult = getInternalIDs(connection, externalIDs, lookup);
        }
        catch (IDMapperException exception2) {
          exception = exception2;
        }
      } while (deadlockHandler.isTransientDeadlock(exception));
      connection.commit();
    }
    catch (IDMapperException e) {
      try {
        connection.rollback();
      }
      catch (SQLException sqlException) {
        s_log.errorT("rollback error" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
      throw e;
    }
    catch (Exception e) {
      try {
        connection.rollback();
      }
      catch (SQLException sqlException) {
        s_log.errorT("rollback error" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
      throw new IDMapperException(e);
    }
    finally {
      try {
        connection.close();
      }
      catch (SQLException sqlException) {
        s_log.errorT("error closing connection" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
    }
    return localResult ;
  }

  /**
   * @param externalID             the external ID
   * @param pool                   @todo: Description of the incoming method parameter
   * @return                       the constant internal ID for an external ID; create a new mapping if it does not
   *      exist yet
   * @exception IDMapperException  a database error occurred
   */
  public static synchronized String getCreateConstInternalID (
  JDBCConnectionPool pool, String externalID ) throws IDMapperException {
    if(pool == null || externalID == null) {
      throw new java.lang.IllegalArgumentException();
    }
    Connection connection = getConnection(pool);
    try {
      connection.setAutoCommit(false);
      String result = null;
      IDDeadlockHandler deadlockHandler = new IDDeadlockHandler(pool, connection);
      IDMapperException exception;
      do {
        try {
          exception = null;
          String constInternalID = getConstInternalID ( connection, externalID ) ;
          if (constInternalID != null) {
            result = constInternalID ;
          }
          else {
            result = createIDMapping(connection, externalID)[CREATE_CONST_INDEX];
          }
        }
        catch (IDMapperException exception2) {
          exception = exception2;
        }
      } while (deadlockHandler.isTransientDeadlock(exception));
      connection.commit();
      return result ;
    }
    catch (IDMapperException e) {
      try {
        connection.rollback();
      }
      catch (SQLException sqlException) {
        s_log.errorT("rollback error" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
      throw e;
    }
    catch (Exception e) {
      try {
        connection.rollback();
      }
      catch (SQLException sqlException) {
        s_log.errorT("rollback errort" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(e));
      }
      throw new IDMapperException(e);
    }
    finally {
      try {
        connection.close();
      }
      catch (SQLException sqlException) {
        s_log.errorT("error closing connection" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
    }
  }


  /**
   * @param externalID             the external ID
   * @param pool                   @todo: Description of the incoming method parameter
   * @return                       the variable internal ID for an external ID (null if mapping doesn't exists)
   * @exception IDMapperException  a database error occurred
   */
  public static String getVarInternalID ( JDBCConnectionPool pool,
  String externalID ) throws IDMapperException {
    if(pool == null || externalID == null) {
      throw new java.lang.IllegalArgumentException();
    }
    Connection connection = getConnection(pool);
    try {
      connection.setAutoCommit(false);
      String result = null;
      IDDeadlockHandler deadlockHandler = new IDDeadlockHandler(pool, connection);
      IDMapperException exception;
      do {
        try {
          exception = null;
          result = getVarInternalID ( connection, externalID ) ;
        }
        catch (IDMapperException exception2) {
          exception = exception2;
        }
      } while (deadlockHandler.isTransientDeadlock(exception));
      connection.commit();
      return result;
    }
    catch (IDMapperException e) {
      try {
        connection.rollback();
      }
      catch (SQLException sqlException) {
        s_log.errorT("rollback error" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
      throw e;
    }
    catch (Exception e) {
      try {
        connection.rollback();
      }
      catch (SQLException sqlException) {
        s_log.errorT("rollback error" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
      throw new IDMapperException(e);
    }
    finally {
      try {
        connection.close();
      }
      catch (SQLException sqlException) {
        s_log.errorT("error closing connection" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
    }
  }


  /**
   * @param externalID             the external ID
   * @param pool                   @todo: Description of the incoming method parameter
   * @return                       the variable internal ID for an external ID; create a new mapping if it does not
   *      exist yet
   * @exception IDMapperException  a database error occurred
   */
  public static synchronized String getCreateVarInternalID (
  JDBCConnectionPool pool, String externalID ) throws IDMapperException {
    if(pool == null || externalID == null) {
      throw new java.lang.IllegalArgumentException();
    }
    Connection connection = getConnection(pool);
    try {
      connection.setAutoCommit(false);
      String result = null;
      IDDeadlockHandler deadlockHandler = new IDDeadlockHandler(pool, connection);
      IDMapperException exception;
      do {
        try {
          exception = null;
          String varInternalID = getVarInternalID ( connection, externalID ) ;
          if ( varInternalID != null ) {
            result = varInternalID ;
          }
          else {
            result = createIDMapping(connection, externalID)[CREATE_VAR_INDEX];
          }
        }
        catch (IDMapperException exception2) {
          exception = exception2;
        }
      } while (deadlockHandler.isTransientDeadlock(exception));
      connection.commit();
      return result;
    }
    catch (IDMapperException e) {
      try {
        connection.rollback();
      }
      catch (SQLException sqlException) {
        s_log.errorT("rollback error" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
      throw e;
    }
    catch (Exception e) {
      try {
        connection.rollback();
      }
      catch (SQLException sqlException) {
        s_log.errorT("rollback error" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
      throw new IDMapperException(e);
    }
    finally {
      try {
        connection.close();
      }
      catch (SQLException sqlException) {
        s_log.errorT("error closing connection" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
    }
  }

  /**
   * @param constInternalID        the constant internal ID
   * @param pool                   @todo: Description of the incoming method parameter
   * @return                       the external ID for a constant internal ID (null if mapping doesn't exists)
   * @exception IDMapperException  a database error occurred
   */
  public static String getExternalIDFromConst ( JDBCConnectionPool pool,
  String constInternalID ) throws IDMapperException {
    if(pool == null || constInternalID == null) {
      throw new java.lang.IllegalArgumentException();
    }
    Connection connection = getConnection(pool);
    try {
      connection.setAutoCommit(false);
      String result = null;
      IDDeadlockHandler deadlockHandler = new IDDeadlockHandler(pool, connection);
      IDMapperException exception;
      do {
        try {
          exception = null;
          result = getExternalIDByConst ( connection, constInternalID ) ;
        }
        catch (IDMapperException exception2) {
          exception = exception2;
        }
      } while (deadlockHandler.isTransientDeadlock(exception));
      connection.commit();
      return result;
    }
    catch (IDMapperException e) {
      try {
        connection.rollback();
      }
      catch (SQLException sqlException) {
        s_log.errorT("rollback error" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
      throw e;
    }
    catch (Exception e) {
      try {
        connection.rollback();
      }
      catch (SQLException sqlException) {
        s_log.errorT("rollback error" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
      throw new IDMapperException(e);
    }
    finally {
      try {
        connection.close();
      }
      catch (SQLException sqlException) {
        s_log.errorT("error closing connection" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
    }
  }


  /**
   * @param pool                   @todo: Description of the incoming method parameter
   * @param varInternalID          @todo: Description of the incoming method parameter
   * @return                       the external ID for a variable internal ID (null if mapping doesn't exists)
   * @exception IDMapperException  a database error occurred
   */
  public static String getExternalIDFromVar(JDBCConnectionPool pool,
  String varInternalID) throws IDMapperException {
    if(pool == null || varInternalID == null) {
      throw new java.lang.IllegalArgumentException();
    }
    Connection connection = getConnection(pool);
    try {
      connection.setAutoCommit(false);
      String result = null;
      IDDeadlockHandler deadlockHandler = new IDDeadlockHandler(pool, connection);
      IDMapperException exception;
      do {
        try {
          exception = null;
          result = getExternalIDByVar(connection, varInternalID);
        }
        catch (IDMapperException exception2) {
          exception = exception2;
        }
      } while (deadlockHandler.isTransientDeadlock(exception));
      connection.commit();
      return result;
    }
    catch (IDMapperException e) {
      try {
        connection.rollback();
      }
      catch (SQLException sqlException) {
        s_log.errorT("rollback error" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
      throw e;
    }
    catch (Exception e) {
      try {
        connection.rollback();
      }
      catch (SQLException sqlException) {
        s_log.errorT("rollback error" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
      throw new IDMapperException(e);
    }
    finally {
      try {
        connection.close();
      }
      catch (SQLException sqlException) {
        s_log.errorT("error closing connection" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
    }
  }

  /**
   * remove a mapping for an external ID
   *
   * @param externalID             the external ID
   * @param pool                   @todo: Description of the incoming method parameter
   * @return                       true iff successful
   * @exception IDMapperException  a database error occurred
   */
  public static synchronized boolean removeIDMapping(JDBCConnectionPool pool,
  String externalID) throws IDMapperException {
    /*
    * previous implementation would run two database queries to
    * pickup const and var internal ids in order to run the delete
    * statement based on the const internal; that's kind of an overkill:
    * apparently it's at least as efficient to run the delete on the
    * [external,hash(external)] tupel directly, thereby reducing the
    * number of database statements from a maximum of 3 to 1
    */
    if(pool == null || externalID == null) {
      throw new java.lang.IllegalArgumentException();
    }
    Connection connection = getConnection(pool);
    PreparedStatement ps = null ;
    try {
      connection.setAutoCommit(false);
      boolean rv = false;
      IDDeadlockHandler deadlockHandler = new IDDeadlockHandler(pool, connection);
      SQLException exception;
      do {
        try {
          exception = null;
          if (ps != null) {
            //retry handling; the connection has already been rolled back!
            try {
              ps.close();
            }
            catch (SQLException ignored) {
              //$JLIN-EXC
              //forced close of statement as we're retrying with a new statement
            }
          }
          // direct jdbc methods
          ps = connection.prepareStatement(PS_REMOVE_ID_MAPPING);
          ps.setLong ( PS_REM_HSH_CI, getHashCodeAsLong(externalID)) ;
          ps.setString ( PS_REM_EXT_CI, externalID );
          rv = ps.executeUpdate() > 0 ;
        }
        catch (SQLException exception2) {
          exception = exception2;
        }
      } while (deadlockHandler.isTransientDeadlock(exception));
      connection.commit();
      return rv ;
    }
    catch (Exception e) {
      try {
        connection.rollback();
      }
      catch (SQLException sqlException) {
        s_log.errorT("rollback error" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
      throw new IDMapperException(e);
    }
    finally {
      try {
        if ( ps != null ) {
          ps.close() ;
        }
        connection.close();
      }
      catch (SQLException sqlException) {
        s_log.errorT("error closing connection" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
    }
  }

  /**
  * get all mappings starting with a specific external ID
  *
  * @param pool                   reference to the connection pool
  * @param externalID             the external ID
  * @exception IDMapperException  a database error occurred
  */
  public static Map getMultipleIDMappings ( JDBCConnectionPool pool,
  String externalID) throws IDMapperException {
    if(pool == null || externalID == null) {
      throw new java.lang.IllegalArgumentException();
    }
    if(logDebugInfo = s_log.beDebug()) {
      s_log.debugT("["+pool.getId()+"] getMultipleIDMappings ( "
        +externalID+"*)");
    }
    Connection connection = getConnection(pool);
    try {
      ResultSet rs = null ;
      PreparedStatement ps = null ;
      connection.setAutoCommit(false);
      try {
        HashMap rv = new HashMap() ;
        IDDeadlockHandler deadlockHandler = new IDDeadlockHandler(pool, connection);
        SQLException exception;
        do {
          try {
            exception = null;
            if (ps != null) {
              //retry handling; the connection has already been rolled back!
              if (rs != null) {
                try {
                  rs.close();
                }
                catch (SQLException ignored) {
                  //$JLIN-EXC
                  //forced close of statement as we're retrying with a new statement
                }
                rs = null;
              }
              try {
                ps.close();
              }
              catch (SQLException ignored) {
                //$JLIN-EXC
                //forced close of statement as we're retrying with a new statement
              }
              rv.clear();
            }
            // direct jdbc methods
            ps = connection.prepareStatement ( PS_GET_MULTIPLE_ID_MAPPINGS,
              ResultSet.TYPE_FORWARD_ONLY, ResultSet.CONCUR_READ_ONLY ) ;
            ps.setString ( 1, externalID ) ;
            ps.setString ( 2, externalID + "/%" ) ;
            rs = ps.executeQuery() ;
            while (rs.next()) {
              rv.put ( rs.getString ( PS_GET_MUL_EXT_CI ).trim(),
                rs.getString ( PS_GET_MUL_CINT_CI ).trim() ) ;
            }
          }
          catch (SQLException exception2) {
            exception = exception2;
          }
        } while (deadlockHandler.isTransientDeadlock(exception));
        connection.commit() ;
        return ( rv ) ;
      }
      catch (SQLException e) {
        throw new IDMapperException(e);
      }
      finally {
        try {
          if ( rs != null ) {
            rs.close() ;
          }
          if ( ps != null ) {
            ps.close() ;
          }
        }
        catch (SQLException e) {
          s_log.errorT("error closing result set" + " - "
            + com.sapportals.wcm.util.logging.LoggingFormatter
            .extractCallstack(e));
        }
      }
    }
    catch (IDMapperException e) {
      try {
        connection.rollback();
      }
      catch (SQLException sqlException) {
        s_log.errorT("rollback error" + " - "
          + com.sapportals.wcm.util.logging.LoggingFormatter
          .extractCallstack(sqlException));
      }
      throw e;
    }
    catch (Exception e) {
      try {
        connection.rollback();
      }
      catch (SQLException sqlException) {
        s_log.errorT("rollback error" + " - "
          + com.sapportals.wcm.util.logging.LoggingFormatter
          .extractCallstack(sqlException));
      }
      throw new IDMapperException(e);
    }
    finally {
      try {
        connection.close();
      }
      catch (SQLException sqlException) {
        s_log.errorT("error closing connection" + " - "
          + com.sapportals.wcm.util.logging.LoggingFormatter
          .extractCallstack(sqlException));
      }
    }
  }

  /**
   * remove all mappings starting with a specific external ID (followed by a slash) (including the mapping for the given
   * external ID)
   *
   * @param externalID             the external ID
   * @param pool                   @todo: Description of the incoming method parameter
   * @exception IDMapperException  a database error occurred
   */
  public static void removeMultipleIDMappings(JDBCConnectionPool pool,
  String externalID) throws IDMapperException {
    if(pool == null || externalID == null) {
      throw new java.lang.IllegalArgumentException();
    }
    if(logDebugInfo = s_log.beDebug()) {
      s_log.debugT("["+pool.getId()+"] removeMultipleIDMappings(id "
        +externalID+")");
    }
    Connection connection = getConnection(pool);
    PreparedStatement ps = null ;
    try {
      connection.setAutoCommit(false);
      int uc = 0;
      IDDeadlockHandler deadlockHandler = new IDDeadlockHandler(pool, connection);
      SQLException exception;
      do {
        try {
          exception = null;
          if (ps != null) {
            //retry handling; the connection has already been rolled back!
            try {
              ps.close();
            }
            catch (SQLException ignored) {
              //$JLIN-EXC
              //forced close of statement as we're retrying with a new statement
            }
          }
          // direct jdbc methods
          ps = connection.prepareStatement ( PS_REMOVE_MULTIPLE_ID_MAPPINGS1 ) ;
          ps.setLong ( PS_REM_MUL1_HSH_CI, getHashCodeAsLong(externalID));
          ps.setString ( PS_REM_MUL1_EXT_CI, externalID ) ;
          uc = ps.executeUpdate() ;
        }
        catch (SQLException exception2) {
          exception = exception2;
        }
      } while (deadlockHandler.isTransientDeadlock(exception));
      connection.commit();
      /*
      * fix for csn 1494286: we can't run the following code as it's dependent
      * on full trees, which is, for each leaf node all nodes of the full path
      * to the root node do exist. This assumption is wrong; especially
      * collaboration does create sparse idmapper trees, so that an update count
      * of zero, returned from the database does NOT imply that no subtree nodes
      * do exist.
      * 
      *      if ( uc <= 0 ) {
      *        return ;
      *      }
      */
      synchronized ( jobList ) {
        jobList.add ( new JobEntry ( pool, externalID )) ;
        jobList.notify() ;
      }
    }
    catch (Exception e) {
      try {
        connection.rollback();
      }
      catch (SQLException sqlException) {
        s_log.errorT("rollback error" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
      throw new IDMapperException(e);
    }
    finally {
      try {
        if ( ps != null ) {
          ps.close() ;
        }
        connection.close();
      }
      catch (SQLException sqlException) {
        s_log.errorT("error closing connection" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
    }
  }

  /**
   * rename a single external ID: the constant internal ID remains constant, the variable internal ID changes
   *
   * @param fromExternalID         the old external ID
   * @param toExternalID           the new external ID
   * @param pool                   @todo: Description of the incoming method parameter
   * @return                       true iff successful
   * @exception IDMapperException  a database error occurred
   */
  public static synchronized boolean renameSingleExternalID(
  JDBCConnectionPool pool, String fromExternalID, String toExternalID)
  throws IDMapperException {

    if(pool == null || fromExternalID == null || toExternalID == null) {
      throw new java.lang.IllegalArgumentException();
    }
    Connection connection = getConnection(pool);
    try {
      connection.setAutoCommit(false);
      boolean result = false ;
      IDDeadlockHandler deadlockHandler = new IDDeadlockHandler(pool, connection);
      IDMapperException exception;
      do {
        try {
          exception = null;
          //indirect jdbc method throwing IDMapperException
          String constIntID = getConstInternalID ( connection, fromExternalID ) ;
          if ( constIntID != null ) {
            result = internalRenameExternalID ( connection, constIntID, toExternalID ) ;
          }
        }
        catch (IDMapperException exception2) {
          exception = exception2;
        }
      } while (deadlockHandler.isTransientDeadlock(exception));
      connection.commit();
      return result;
    }
    catch (IDMapperException e) {
      try {
        connection.rollback();
      }
      catch (SQLException sqlException) {
        s_log.errorT("rollback error" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
      throw e;
    }
    catch (Exception e) {
      try {
        connection.rollback();
      }
      catch (SQLException sqlException) {
        s_log.errorT("rollback error" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
      throw new IDMapperException(e);
    }
    finally {
      try {
        connection.close();
      }
      catch (SQLException sqlException) {
        s_log.errorT("error closing connection" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
    }
  }


  public static synchronized void renameMultipleExternalIDs(
  JDBCConnectionPool pool, String fromExternalID, String toExternalID)
  throws IDMapperException {
    renameMultipleExternalIDs(pool, fromExternalID, toExternalID, null);
  }


  /**
   * rename a multiple external IDs: the constant internal IDs remain constant, the variable internal IDs change - all
   * external IDs starting with <fromExternalID> are changed
   *
   * @param fromExternalID         the old external ID
   * @param toExternalID           the new external ID
   * @param pool                   @todo: Description of the incoming method parameter
   * @exception IDMapperException  a database error occurred
   */
  public static synchronized void renameMultipleExternalIDs(
  JDBCConnectionPool pool, String fromExternalID, String toExternalID, IDMapperRenameCallback callback)
  throws IDMapperException {

    if(pool == null || fromExternalID == null || toExternalID == null) {
      throw new java.lang.IllegalArgumentException();
    }
    if(logDebugInfo = s_log.beDebug()) {
      s_log.debugT("["+pool.getId()+"] renameMultipleExternalIDs(from "
        +fromExternalID+", to "+toExternalID+")");
    }
    Connection connection = getConnection(pool);
    try {
      ResultSet rs = null ;
      PreparedStatement ps = null ;
      try {
        connection.setAutoCommit(false);
        IDDeadlockHandler deadlockHandler = new IDDeadlockHandler(pool, connection);
        SQLException exception;
        do {
          try {
            exception = null;
            if (ps != null) {
              //retry handling; the connection has already been rolled back!
              if (rs != null) {
                try {
                  rs.close();
                }
                catch (SQLException ignored) {
                  //$JLIN-EXC
                  //forced close of statement as we're retrying with a new statement
                }
                rs = null;
              }
              try {
                ps.close();
              }
              catch (SQLException ignored) {
                //$JLIN-EXC
                //forced close of statement as we're retrying with a new statement
              }
            }
            // direct jdbc methods
            ps = connection.prepareStatement ( PS_RENAME_MULTIPLE_EXTERNAL_IDS,
              ResultSet.TYPE_FORWARD_ONLY, ResultSet.CONCUR_UPDATABLE ) ;
            ps.setString ( PS_REN_MUL_EXT_CI, fromExternalID + "%" ) ;
            rs = ps.executeQuery() ;
            while (rs.next()) {
              String currentExternalID = rs.getString(RS_REN_MUL_EXT_CI).trim();
              if ( currentExternalID.startsWith ( fromExternalID )) {
                String newExtID ;
                if (currentExternalID.length() > fromExternalID.length()) {
                  newExtID = toExternalID
                    + currentExternalID.substring(fromExternalID.length());
                }
                else {
                  newExtID = toExternalID;
                }
                rs.updateLong ( RS_REN_MUL_HSH_CI, getHashCodeAsLong(newExtID));
                rs.updateString ( RS_REN_MUL_EXT_CI, newExtID );
                rs.updateString ( RS_REN_MUL_VINT_CI, new UUID().toString() );
                rs.updateRow();
                if( callback != null ) {
                  callback.rename(currentExternalID, newExtID);
                }
              }
            }
          }
          catch (SQLException exception2) {
            exception = exception2;
          }
        } while (deadlockHandler.isTransientDeadlock(exception));
        connection.commit() ;
      }
      catch (SQLException e) {
        throw new IDMapperException(e);
      }
      finally {
        try {
          if ( rs != null ) {
            rs.close() ;
          }
          if ( ps != null ) {
            ps.close() ;
          }
        }
        catch (SQLException e) {
          s_log.errorT("error closing result set" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(e));
        }
      }
    }
    catch (IDMapperException e) {
      try {
        connection.rollback();
      }
      catch (SQLException sqlException) {
        s_log.errorT("rollback error" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
      throw e;
    }
    catch (Exception e) {
      try {
        connection.rollback();
      }
      catch (SQLException sqlException) {
        s_log.errorT("rollback error" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
      throw new IDMapperException(e);
    }
    finally {
      try {
        connection.close();
      }
      catch (SQLException sqlException) {
        s_log.errorT("error closing connection" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
    }
  }


  /**
   * create the mapping table (if not already existing)
   *
   * @param pool                   @todo: Description of the incoming method parameter
   * @exception IDMapperException  a database error occurred
   */
  public static synchronized void checkCreateTable(JDBCConnectionPool pool)
    throws IDMapperException {

    if(pool == null) {
      throw new java.lang.IllegalArgumentException();
    }

    Connection  connection  = getConnection(pool);
    Statement s = null ;
    try {
      connection.setAutoCommit(false);

      if (!checkTable(connection, TABLE_NAME)) {
        String  sql  = "CREATE TABLE " + TABLE_NAME + " (" +
          sqlLongCol(HASH_COL_NAME) + "," +
          sqlLongStrCol(EXTERNAL_ID_COL_NAME, EXTERNAL_ID_LENGTH) + "," +
          sqlStr255Col(CONST_INTERNAL_ID_COL_NAME, INTERNAL_ID_LENGTH) + "," +
          sqlStr255Col(VAR_INTERNAL_ID_COL_NAME, INTERNAL_ID_LENGTH) + "," +
          "PRIMARY KEY (" + CONST_INTERNAL_ID_COL_NAME + "))";
        //update(connection, sql);
        s = connection.createStatement() ;
        s.execute ( sql ) ;
      }

      connection.commit();
    }
    catch (Exception e) {
      try {
        connection.rollback();
      }
      catch (SQLException sqlException) {
        s_log.errorT("rollback error" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
      throw new IDMapperException(e);
    }
    finally {
      try {
        if ( s != null ) {
          s.close() ;
        }
        connection.close();
      }
      catch (SQLException sqlException) {
        s_log.errorT("error closing connection" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
    }
  }

  /**
   * delete the mapping table
   *
   * @param pool                   @todo: Description of the incoming method parameter
   * @exception IDMapperException  a database error occurred
   */
  public static void deleteTable(JDBCConnectionPool pool)
    throws IDMapperException {

    if(pool == null) {
      throw new java.lang.IllegalArgumentException();
    }

    Connection  connection  = getConnection(pool);
    Statement s = null ;
    try {
      connection.setAutoCommit(false);

      String  sql  = "DROP TABLE " + TABLE_NAME;
      //update(connection, sql);
      s = connection.createStatement() ;
      s.execute ( sql ) ;

      connection.commit();
    }
    catch (Exception e) {
      try {
        connection.rollback();
      }
      catch (SQLException sqlException) {
        s_log.errorT("rollback error" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
      throw new IDMapperException(e);
    }
    finally {
      try {
        if ( s != null ) {
          s.close() ;
        }
        connection.close();
      }
      catch (SQLException sqlException) {
        s_log.errorT("error closing connection" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(sqlException));
      }
    }
  }

  /**
   * @todo: Description of the Method.
   *
   * @return String array - firstr element: const id, second element: var id
   * @param connection             @todo: Description of the incoming method parameter
   * @param externalID             @todo: Description of the incoming method parameter
   * @exception IDMapperException  Exception raised in failure situation
   */
  public static String[] createIDMapping(Connection connection,
  String externalID) throws IDMapperException {

    if(connection == null || externalID == null) {
      throw new java.lang.IllegalArgumentException();
    }
    if(logDebugInfo = s_log.beDebug()) {
      s_log.debugT("["+getPoolID(connection)+
        "] createIDMapping(id "+externalID+")");
    }
    String[] result = new String[2];
    result[CREATE_CONST_INDEX] = new UUID().toString();
    result[CREATE_VAR_INDEX]   = new UUID().toString();
    PreparedStatement ps = null ;
    try {
      ps = connection.prepareStatement ( PS_CREATE_ID_MAPPING ) ;
      ps.setLong ( PS_CRT_HSH_CI, getHashCodeAsLong(externalID) ) ;
      ps.setString ( PS_CRT_EXT_CI, externalID ) ;
      ps.setString ( PS_CRT_CINT_CI, result[CREATE_CONST_INDEX] ) ;
      ps.setString ( PS_CRT_VINT_CI, result[CREATE_VAR_INDEX] ) ;
      if ( ps.executeUpdate() <= 0 ) {
        throw new IDMapperException("can't create new ID mapping");
      }
      return result;
    }
    catch (SQLException e) {
      throw new IDMapperException(e);
    }
    finally {
      try {
        if ( ps != null ) {
          ps.close() ;
        }
      }
      catch (SQLException e) {
        s_log.errorT("error closing result set" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(e));
      }
    }
  }

  private static String getConstInternalID ( Connection connection,
  String externalID ) throws IDMapperException {

    if(logDebugInfo) {
      s_log.debugT("["+getPoolID(connection)+"] getConstInternalID(extid "
        +externalID+", intidcol "+CONST_INTERNAL_ID_COL_NAME+")");
    }
    PreparedStatement ps = null ;
    ResultSet rs = null ;
    try {
      ps = connection.prepareStatement(PS_GET_CONST_INTERNAL_ID);
      ps.setLong ( PS_GETCONST_HSH_CI, getHashCodeAsLong(externalID)) ;
      ps.setString ( PS_GETCONST_EXT_CI, externalID );
      rs = ps.executeQuery() ;
      while (rs.next()) {
        if (rs.getString(RS_GETCONST_EXT_CI).equals(externalID)) {
          return rs.getString(RS_GETCONST_CINT_CI);
        }
      }
      return null ;
    }
    catch (SQLException e) {
      throw new IDMapperException(e);
    }
    finally {
      try {
        if ( rs != null ) {
          rs.close() ;
        }
        if ( ps != null ) {
          ps.close() ;
        }
      }
      catch (SQLException e) {
        s_log.errorT("error closing result set" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(e));
      }
    }
  }

  private static String getVarInternalID ( Connection connection,
  String externalID ) throws IDMapperException {

    if(logDebugInfo) {
      s_log.debugT("["+getPoolID(connection)+
        "] getVarInternalID(extid "+externalID+
        ", intidcol "+VAR_INTERNAL_ID_COL_NAME+")");
    }
    PreparedStatement ps = null ;
    ResultSet rs = null ;
    try {
      ps = connection.prepareStatement(PS_GET_VAR_INTERNAL_ID);
      ps.setLong ( PS_GETVAR_HSH_CI, getHashCodeAsLong(externalID)) ;
      ps.setString ( PS_GETVAR_EXT_CI, externalID );
      rs = ps.executeQuery() ;
      while (rs.next()) {
        if (rs.getString(RS_GETVAR_EXT_CI).equals(externalID)) {
          return rs.getString(RS_GETVAR_VINT_CI);
        }
      }
      return null ;
    }
    catch (SQLException e) {
      throw new IDMapperException(e);
    }
    finally {
      try {
        if ( rs != null ) {
          rs.close() ;
        }
        if ( ps != null ) {
          ps.close() ;
        }
      }
      catch (SQLException e) {
        s_log.errorT("error closing result set" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(e));
      }
    }
  }

  private static final int MAX_IDS_PER_QUERY = 100;

  private static String[] getInternalIDs(Connection connection,
  String[] externalIDs, int lookup ) throws IDMapperException {
    if(logDebugInfo) {
      String internalIDColName ;
      if ( lookup == LOOKUP_BY_CONST_INT ) {
        internalIDColName = CONST_INTERNAL_ID_COL_NAME ;
      }
      else if ( lookup == LOOKUP_BY_VAR_INT ) {
        internalIDColName = VAR_INTERNAL_ID_COL_NAME ;
      }
      else {
        internalIDColName = "???" ;
      }
      s_log.debugT("["+getPoolID(connection)
        +"] getInternalIDs(..., intidcol "+internalIDColName+")");
    }
    if(externalIDs.length == 0) {
      return new String[0];
    }
    String[] result ;
    if(externalIDs.length <= MAX_IDS_PER_QUERY ) {
      result = getInternalIDsFromDB ( connection, externalIDs, lookup ) ;
    }
    else {
      result = new String[externalIDs.length];
      String[] externalTakeIDs = new String[MAX_IDS_PER_QUERY];
      for(int i = 0; i < externalIDs.length; i += MAX_IDS_PER_QUERY) {
        int takeCount = Math.min(externalIDs.length - i, MAX_IDS_PER_QUERY);
        if ( takeCount != MAX_IDS_PER_QUERY ) {
          externalTakeIDs = new String[takeCount];
        }
        for(int k = 0; k < takeCount; k++) {
          externalTakeIDs[k] = externalIDs[i+k];
        }
        String[] internalIDs = getInternalIDsFromDB (
          connection, externalTakeIDs, lookup ) ;
        for(int k = 0; k < takeCount; k++) {
          result[i+k] = internalIDs[k];
        }
      }
    }
    return result;
  }


  private static String[] getInternalIDsFromDB ( Connection connection,
  String[] extIDs, int lookup ) throws IDMapperException {
    if(logDebugInfo) {
      s_log.debugT("["+getPoolID(connection)+"] getInternalIDsFromDB (...)");
    }
    if (extIDs.length == 0) {
      return new String[0] ;
    }
    ResultSet rs = null ;
    PreparedStatement ps = null ;
    try {
      int i ;
      String PS_GET_INTERNAL_IDS_FROM_DB_LOCAL ;
      if ( lookup == LOOKUP_BY_CONST_INT ) {
        i = Math.min(PS_GET_CONST_INTERNAL_IDS_FROM_DB.length,extIDs.length) ;
        PS_GET_INTERNAL_IDS_FROM_DB_LOCAL = PS_GET_CONST_INTERNAL_IDS_FROM_DB[i - 1] ;
        for ( i = PS_GET_CONST_INTERNAL_IDS_FROM_DB.length ; i < extIDs.length ; i++ ) {
          PS_GET_INTERNAL_IDS_FROM_DB_LOCAL +=
            (" OR (" + HASH_COL_NAME + " = ? AND " + EXTERNAL_ID_COL_NAME + " = ?) ") ;
        }
      }
      else if ( lookup == LOOKUP_BY_VAR_INT ) {
        i = Math.min(PS_GET_VAR_INTERNAL_IDS_FROM_DB.length,extIDs.length) ;
        PS_GET_INTERNAL_IDS_FROM_DB_LOCAL = PS_GET_VAR_INTERNAL_IDS_FROM_DB[i - 1] ;
        for ( i = PS_GET_VAR_INTERNAL_IDS_FROM_DB.length ; i < extIDs.length ; i++ ) {
          PS_GET_INTERNAL_IDS_FROM_DB_LOCAL +=
            (" OR (" + HASH_COL_NAME + " = ? AND " + EXTERNAL_ID_COL_NAME + " = ?) ") ;
        }
      }
      else {
        throw new IDMapperException("unknown internal column name");
      }
      ps = connection.prepareStatement ( PS_GET_INTERNAL_IDS_FROM_DB_LOCAL ) ;
      for ( i = 0 ; i < extIDs.length ; i++ ) {
        int j = i<<1 ;
        ps.setLong ( j+1, getHashCodeAsLong(extIDs[i]) ) ;
        ps.setString ( j+2, extIDs[i] ) ;
      }
      rs = ps.executeQuery() ;
      /*
      * the previous implementation would fill in the result array with
      * the help of a hash map; basically, the getString(3) table entry
      * was stored in there as a key with the integer index into result[]
      * as the value;
      */
      String[] result = new String[extIDs.length] ;
      int startndx = 0 ;
      while ( rs.next() ) {
        String s = rs.getString(RS_GETDB_EXT_CI);
        for ( i = startndx ; i < extIDs.length ; i++ ) {
          if ( result[i] != null ) {
            if ( startndx == i ) {
              startndx++ ;
            }
          }
          else if ( s.equals(extIDs[i])) {
            result[i] = rs.getString(RS_GETDB_INT_CI);
            if ( startndx == i ) {
              startndx++ ;
            }
            break ;
          }
        }
      }
      return result ;
    }
    catch (SQLException e) {
      throw new IDMapperException(e);
    }
    finally {
      try {
        if ( rs != null ) {
          rs.close() ;
        }
        if ( ps != null ) {
          ps.close() ;
        }
      }
      catch (SQLException e) {
        s_log.errorT("error closing result set" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(e));
      }
    }
  }


  private static String getExternalIDByConst ( Connection connection,
  String internalID) throws IDMapperException {
    if(logDebugInfo) {
      s_log.debugT("["+getPoolID(connection)+"] getExternalIDByConst (intid "
        +internalID+", intidcol "+CONST_INTERNAL_ID_COL_NAME +")");
    }
    ResultSet rs = null ;
    PreparedStatement ps = null ;
    try {
      ps = connection.prepareStatement ( PS_GET_EXTERNAL_ID_BY_CONST ) ;
      ps.setString ( PS_GETEXTBYCONST_CINT_CI, internalID ) ;
      rs = ps.executeQuery() ;
      while ( rs.next() ) {
        /* verify that column value match by case */
        if ( rs.getString(RS_GETEXTBYCONST_CINT_CI).equals(internalID) ) {
          return rs.getString(RS_GETEXTBYCONST_EXT_CI);
        }
      }
    }
    catch (SQLException e) {
      throw new IDMapperException(e);
    }
    finally {
      try {
        if ( rs != null ) {
          rs.close() ;
        }
        if ( ps != null ) {
          ps.close() ;
        }
      }
      catch (SQLException e) {
        s_log.errorT("error closing result set" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(e));
      }
    }
    return null ;
  }


  private static String getExternalIDByVar ( Connection connection,
  String internalID) throws IDMapperException {
    if(logDebugInfo) {
      s_log.debugT("["+getPoolID(connection)+"] getExternalIDByVar (intid "
        + internalID+", intidcol "+VAR_INTERNAL_ID_COL_NAME +")");
    }
    ResultSet rs = null ;
    PreparedStatement ps = null ;
    try {
      ps = connection.prepareStatement ( PS_GET_EXTERNAL_ID_BY_VAR ) ;
      ps.setString ( PS_GETEXTBYVAR_VINT_CI, internalID ) ;
      rs = ps.executeQuery() ;
      while ( rs.next() ) {
        /* verify that column value match by case */
        if ( rs.getString(RS_GETEXTBYVAR_CINT_CI).equals(internalID) ) {
          return rs.getString(RS_GETEXTBYVAR_EXT_CI);
        }
      }
    }
    catch (SQLException e) {
      throw new IDMapperException(e);
    }
    finally {
      try {
        if ( rs != null ) {
          rs.close() ;
        }
        if ( ps != null ) {
          ps.close() ;
        }
      }
      catch (SQLException e) {
        s_log.errorT("error closing result set" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(e));
      }
    }
    return null ;
  }


  /**
   * Get the Connection attribute of the IDMapper class.
   *
   * @param pool                   @todo: Description of the incoming method parameter
   * @return                       The Connection value
   * @exception IDMapperException  Exception raised in failure situation
   */
  private static Connection getConnection(JDBCConnectionPool pool)
    throws IDMapperException {
    try {
      return pool.getConnection();
    }
    catch (SQLException e) {
      throw new IDMapperException(e);
    }
  }

  private static String sqlLongCol(String name ) {
    return name + " BIGINT ";
    // mssql
  }
  /**
   * @todo: Description of the Method.
   *
   * @param name       @todo: Description of the incoming method parameter
   * @param maxLength  @todo: Description of the incoming method parameter
   * @return           @todo: Description of the outgoing return value
   */
  private static String sqlLongStrCol(String name, int maxLength) {
    return name + " NVARCHAR(" + maxLength + ")";
    // mssql
  }

  /**
   * @todo: Description of the Method.
   *
   * @param name  @todo: Description of the incoming method parameter
   * @return      @todo: Description of the outgoing return value
   */
  private static String sqlLongStrCol_MYSQL(String name) {
    return name + " BLOB";
  }

  /**
   * @todo: Description of the Method.
   *
   * @param name       @todo: Description of the incoming method parameter
   * @param maxLength  @todo: Description of the incoming method parameter
   * @return           @todo: Description of the outgoing return value
   */
  private static String sqlStr255Col(String name, int maxLength) {
    // length is limited to 255 bytes

    return name + " CHAR(" + maxLength + ") NOT NULL";
  }

  /**
   * @todo: Description of the Method.
   *
   * @param connection  @todo: Description of the incoming method parameter
   * @param table       @todo: Description of the incoming method parameter
   * @return            @todo: Description of the outgoing return value
   */
  private static boolean checkTable(Connection con, String table) {
    try {
      return con.getMetaData().getTables ( null, null, table, null ).next() ;
    }
    catch (Exception e) {
    }
    return false ;
  }


  /**
   * @todo: Description of the Method.
   *
   * @param connection             @todo: Description of the incoming method parameter
   * @param constInternalID        @todo: Description of the incoming method parameter
   * @param toExternalID           @todo: Description of the incoming method parameter
   * @return                       @todo: Description of the outgoing return value
   * @exception IDMapperException  Exception raised in failure situation
   */
  private static boolean internalRenameExternalID ( Connection connection,
  String constInternalID, String toExternalID) throws IDMapperException {
    if(logDebugInfo) {
      s_log.debugT("["+getPoolID(connection)
        +"] internalRenameExternalID(constid "
        +constInternalID+", toextid "+toExternalID+")");
    }
    PreparedStatement ps = null ;
    try {
      ps = connection.prepareStatement ( PS_INTERNAL_RENAME_EXTERNAL_ID ) ;
      ps.setLong ( PS_REN_EXT_HSH_CI, getHashCodeAsLong(toExternalID)) ;
      ps.setString ( PS_REN_EXT_EXT_CI, toExternalID ) ;
      ps.setString ( PS_REN_EXT_VINT_CI, new UUID().toString() ) ;
      ps.setString ( PS_REN_EXT_CINT_CI, constInternalID ) ;
      return ps.executeUpdate() > 0 ;
    }
    catch (SQLException e) {
      throw new IDMapperException(e);
    }
    finally {
      try {
        if ( ps != null ) {
          ps.close() ;
        }
      }
      catch (SQLException e) {
        s_log.errorT("error closing result set" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(e));
      }
    }
  }


   /**
    * get the id of the pool the connection belongs to
    */
   private static String getPoolID(Connection connection) {
     if(connection instanceof JDBCConnectionWrapper) {
       return ((JDBCConnectionWrapper)connection).getPool().getId();
     }
     return "?";
   }
}
