/**
 * Copyright (c) SAP AG 2001-2002
 * @author Dirk Sabiwalsky
 * @author Helmut Cossmann
 * @version $Id: IDMapperCached.java,v 1.3 2004/09/09 15:40:20 jre Exp $
 */
package com.sapportals.wcm.util.jdbc.idmapper;

import java.sql.*;
import java.util.*;
import java.util.zip.CRC32;

import com.sapportals.config.fwk.*;
import com.sapportals.wcm.WcmException;
import com.sapportals.wcm.util.config.ConfigurationProxy;
import com.sapportals.wcm.util.cache.*;
import com.sapportals.wcm.util.jdbc.connectionpool.*;
import com.sap.tc.logging.Location;
import com.sapportals.wcm.util.logging.LoggingFormatter;
import com.sapportals.wcm.util.uuid.UUID;

/**
 * Maps external IDs of arbitrary length to internal IDs with a maximum length of 255. <p>
 * - the mapping is persisted in a JDBC database<br>
 * - each mapping consists of a constant internal ID and a variable internal ID<br>
 * - if an external ID is renamed, the constant internal ID remains unchanged and the variable internal ID changes <p>
 */
public final class IDMapperCached
extends IDMapper {

  // =========================
  // INNER CLASS CacheCallback ================================================
  // =========================

  static class CacheCallback
  implements IDMapperRenameCallback {

    String poolid;

    CacheCallback(String poolid) {
      this.poolid = poolid;
    }

    public void rename(String oldExternalID,
                       String newExternalID) {
      IDMapperCached.cacheRenameExternalID(this.poolid, oldExternalID, newExternalID);
    }

  }


  // =========
  // CONSTANTS ================================================================
  // =========

  private final static int LOOKUP_BY_CONST_INT          = 1 ;
  private final static int LOOKUP_BY_VAR_INT            = 2 ;

  private final static int EXTERNAL_CONST_INTERNAL_MAP  = 0;
  private final static int EXTERNAL_VAR_INTERNAL_MAP    = 1;
  private final static int CONST_INTERNAL_EXTERNAL_MAP  = 2;
  private final static int VAR_INTERNAL_EXTERNAL_MAP    = 3;

  private final static String CFG_CACHE_KEY             = "cache";


  // ===============
  // CLASS VARIABLES ==========================================================
  // ===============

  private static com.sap.tc.logging.Location log = com.sap.tc.logging.Location.getLocation(IDMapperCached.class.getName());

  private static Hashtable cacheTables[];
  private static IClusterCacheArea cacheArea;


  // ===============
  // PRIVATE METHODS ==========================================================
  // ===============

  // ------------------------------------------------------------------------
  private static ICache createCache() {
    try {
      String cacheid;
      IConfigurable configurable = ConfigurationProxy.getInstance().getFirstPluginConfigurable(ConfigurationProxy.CFG_PLUGIN_CM_UTILITIES_IDMAPPER);
      if( configurable != null ) {
        Properties idMapperProperties = ConfigurationProxy.getConfigurableProperties(configurable);
        idMapperProperties.setProperty(CacheFactory.CFG_AUTO_DELAY_EXPIRATION_KEY, new Boolean(false).toString());
        cacheid = idMapperProperties.getProperty(CFG_CACHE_KEY);
      } else {
        cacheid = null;
      }
      if( cacheid == null ) { // use default cache setting
        IDMapperCached.log.warningT("cache not configured - using defaults");
        Properties cacheProperties = new Properties();
        cacheProperties.setProperty(CacheFactory.CFG_CACHE_TYPE_KEY, ICache.DEFAULT_CACHE_TYPE);
        cacheProperties.setProperty(CacheFactory.CFG_CACHE_CLASS_KEY, ICache.DEFAULT_CACHE_CLASS);
        cacheProperties.setProperty(CacheFactory.CFG_CAPACITY_KEY, new Integer(ICache.DEFAULT_CAPACITY).toString());
        cacheProperties.setProperty(CacheFactory.CFG_MAX_CACHE_SIZE_KEY, new Long(ICache.DEFAULT_MAX_CACHE_SIZE).toString());
        cacheProperties.setProperty(CacheFactory.CFG_MAX_ENTRY_SIZE_KEY, new Long(ICache.DEFAULT_MAX_ENTRY_SIZE).toString());
        cacheProperties.setProperty(CacheFactory.CFG_DEFAULT_TIME_TO_LIVE_KEY, new Integer(ICache.DEFAULT_TIME_TO_LIVE).toString());
        cacheProperties.setProperty(CacheFactory.CFG_AVERAGE_ENTRY_SIZE_KEY, new Long(ICache.DEFAULT_AVG_ENTRY_SIZE).toString());
        cacheProperties.setProperty(CacheFactory.CFG_SINGLETON_KEY, new Boolean(true).toString());
        cacheProperties.setProperty(CacheFactory.CFG_AUTO_DELAY_EXPIRATION_KEY, new Boolean(false).toString());
        return CacheFactory.getInstance().getCache(CacheFactory.getUniqueCacheID(),cacheProperties);
      } else { // use configured cache
        return CacheFactory.getInstance().getCache(cacheid);
      }
    }
    catch( Exception e ) {
      IDMapperCached.log.errorT("cache disabled due to exception - " + LoggingFormatter.extractCallstack(e));
      return null;
    }
  }


  // ------------------------------------------------------------------------
  private static ICache findCache(int index,
                                  String poolid) {
    if( IDMapperCached.cacheTables == null ) return null; // not yet initialized
    Hashtable table = IDMapperCached.cacheTables[index];
    if( table == null ) return null; // not yet initialized
    Object object = table.get(poolid);
    if( object == null ) {
      ICache cache = createCache();
      if( cache == null ) {
        table.put(poolid, new Object());
      } else {
        table.put(poolid, cache);
      }
      return cache;
    } else {
      if( object instanceof ICache ) {
        return (ICache)object;
      } else {
        return null;
      }
    }
  }


  // ------------------------------------------------------------------------
  private static String cacheGetEntry(ICache cache,
                                      String key) {
    if( cache == null ) return null;
    try {
      ICacheEntry entry = cache.getEntry(key);
      if( entry == null ) return null;
      return (String)entry.getObject();
    }
    catch( Exception e ) {
      return null;
    }
  }


  // ------------------------------------------------------------------------
  private static void cachePutEntry(ICache cache,
                                    String key,
                                    String value) {
    if( cache != null ) {
      try {
        cache.addEntry(key, value);
      }
      catch( Exception e ) {
        // ignore
      }
    }
  }


  // ------------------------------------------------------------------------
  private static void cacheRemoveEntry(ICache cache,
                                       String key) {
    if( cache != null ) {
      try {
        cache.removeEntry(key);
      }
      catch( Exception e ) {
        // ignore
      }
    }
  }


  // ------------------------------------------------------------------------
  private static void cacheClear() {
    if( IDMapperCached.cacheTables == null ) return; // not yet initialized
    for( int i = 0; i < IDMapperCached.cacheTables.length; i++ ) {
      Hashtable table = IDMapperCached.cacheTables[i];
      if( table == null ) continue; // not yet initialized
      String key;
      Object entry;
      Iterator iterator = table.keySet().iterator();
      while( iterator.hasNext() ) {
        key = (String)iterator.next();
        entry = table.get(key);
        if(   ( entry != null )
           && ( entry instanceof ICache )
          ) {
          try {
            ((ICache)entry).clearCache();
          }
          catch(Exception e) {
            // ignore
          }
        }
      }
    }
  }


  // ------------------------------------------------------------------------
  private static boolean checkCache() {
    if( IDMapperCached.cacheArea == null ) {
      return true;
    } else if( IDMapperCached.cacheArea.isCluster() ) {
      if( IDMapperCached.cacheArea.isValid() ) {
        return true;
      } else {
        cacheClear();
        IDMapperCached.cacheArea.validate();
        return false;
      }
    } else {
      return true;
    }
  }


  // ------------------------------------------------------------------------
  private static String cacheGetConstInternalID(String poolid,
                                                String externalID) {
    return cacheGetEntry(findCache(EXTERNAL_CONST_INTERNAL_MAP, poolid), externalID);
  }


  // ------------------------------------------------------------------------
  private static String cacheGetVarInternalID(String poolid,
                                              String externalID) {
    return cacheGetEntry(findCache(EXTERNAL_VAR_INTERNAL_MAP, poolid), externalID);
  }


  // ------------------------------------------------------------------------
  private static String cacheGetExternalIDFromConst(String poolid,
                                                    String constInternalID) {
    return cacheGetEntry(findCache(CONST_INTERNAL_EXTERNAL_MAP, poolid), constInternalID);
  }


  // ------------------------------------------------------------------------
  private static String cacheGetExternalIDFromVar(String poolid,
                                                  String varInternalID) {
    return cacheGetEntry(findCache(VAR_INTERNAL_EXTERNAL_MAP, poolid), varInternalID);
  }


  // ------------------------------------------------------------------------
  private static void cachePutConstInternalID(String poolid,
                                              String externalID,
                                              String constInternalID) {
    cachePutEntry(findCache(EXTERNAL_CONST_INTERNAL_MAP, poolid), externalID, constInternalID);
    cachePutEntry(findCache(CONST_INTERNAL_EXTERNAL_MAP, poolid), constInternalID, externalID);
  }


  // ------------------------------------------------------------------------
  private static void cachePutVarInternalID(String poolid,
                                            String externalID,
                                            String varInternalID) {
    cachePutEntry(findCache(EXTERNAL_VAR_INTERNAL_MAP, poolid), externalID, varInternalID);
    cachePutEntry(findCache(VAR_INTERNAL_EXTERNAL_MAP, poolid), varInternalID, externalID);
  }


  // ------------------------------------------------------------------------
  private static void cachePutExternalIDFromConst(String poolid,
                                                  String constInternalID,
                                                  String externalID) {
    cachePutEntry(findCache(CONST_INTERNAL_EXTERNAL_MAP, poolid), constInternalID, externalID);
    cachePutEntry(findCache(EXTERNAL_CONST_INTERNAL_MAP, poolid), externalID, constInternalID);
  }


  // ------------------------------------------------------------------------
  private static void cachePutExternalIDFromVar(String poolid,
                                                String varInternalID,
                                                String externalID) {
    cachePutEntry(findCache(VAR_INTERNAL_EXTERNAL_MAP, poolid), varInternalID, externalID);
    cachePutEntry(findCache(EXTERNAL_VAR_INTERNAL_MAP, poolid), externalID, varInternalID);
  }


  // ------------------------------------------------------------------------
  private static void cacheRemoveExternalID(String poolid,
                                            String externalID) {
    String mapping;
    mapping = cacheGetConstInternalID(poolid, externalID);
    if( mapping != null ) {
      cacheRemoveEntry(findCache(EXTERNAL_CONST_INTERNAL_MAP, poolid), externalID);
      cacheRemoveEntry(findCache(CONST_INTERNAL_EXTERNAL_MAP, poolid), mapping);
    }
    mapping = cacheGetVarInternalID(poolid, externalID);
    if( mapping != null ) {
      cacheRemoveEntry(findCache(EXTERNAL_VAR_INTERNAL_MAP, poolid), externalID);
      cacheRemoveEntry(findCache(VAR_INTERNAL_EXTERNAL_MAP, poolid), mapping);
    }
  }


  // ------------------------------------------------------------------------
  private static void cacheRenameExternalID(String poolid,
                                            String oldExternalID,
                                            String newExternalID) {
    String mapping;
    mapping = cacheGetConstInternalID(poolid, oldExternalID);
    if( mapping != null ) {
      cacheRemoveEntry(findCache(EXTERNAL_CONST_INTERNAL_MAP, poolid), oldExternalID);
      cachePutEntry(findCache(EXTERNAL_CONST_INTERNAL_MAP, poolid), newExternalID, mapping);
      cachePutEntry(findCache(CONST_INTERNAL_EXTERNAL_MAP, poolid), mapping, newExternalID);
    }
    mapping = cacheGetVarInternalID(poolid, oldExternalID);
    if( mapping != null ) {
      cacheRemoveEntry(findCache(EXTERNAL_VAR_INTERNAL_MAP, poolid), oldExternalID);
      cachePutEntry(findCache(EXTERNAL_VAR_INTERNAL_MAP, poolid), newExternalID, mapping);
      cachePutEntry(findCache(VAR_INTERNAL_EXTERNAL_MAP, poolid), mapping, newExternalID);
    }
  }


  // ------------------------------------------------------------------------
  static String[] getInternalIDs(JDBCConnectionPool pool,
                                 String[] externalIDs,
                                 int lookup)
                          throws IDMapperException {

    if(   ( pool == null )
       || ( externalIDs == null )
      ) {
      throw new java.lang.IllegalArgumentException();
    }
    if( externalIDs.length == 0 ) {
      return new String[0];
    }
    // get the cached entries
    checkCache();
    String poolid = pool.getId();
    int i = 0;
    int cacheMissCount = 0;
    String[] result = new String[externalIDs.length];
    String cachedID;
    switch( lookup ) {
      case LOOKUP_BY_CONST_INT :
        for( i = 0; i < externalIDs.length; i++ ) {
          cachedID = cacheGetConstInternalID(poolid, externalIDs[i]);
          if( cachedID != null ) {
            result[i] = cachedID;
          } else {
            cacheMissCount++;
          }
        }
        break;
      case LOOKUP_BY_VAR_INT :
        for( i = 0; i < externalIDs.length; i++ ) {
          cachedID = cacheGetVarInternalID(poolid, externalIDs[i]);
          if( cachedID != null ) {
            result[i] = cachedID;
          } else {
            cacheMissCount++;
          }
        }
        break;
      default :
        break;
    }
    // if all entries found from cache -> done
    if( cacheMissCount == 0 ) {
      return result;
    }
    // get remaining entries from database
    String[] remainingIDs;
    if( cacheMissCount == externalIDs.length ) { // nothing was found in cache
      remainingIDs = externalIDs;
    } else { // some entries where cached -> extract the ids to get
      remainingIDs = new String[cacheMissCount];
      for( i--; i >= 0; i-- ) {
        if( result[i] == null ) remainingIDs[--cacheMissCount] = externalIDs[i];
      }
    }
    String[] localResult = IDMapper.getInternalIDs(pool, remainingIDs, lookup);
    // merge the cached entries and the database entries
    if( cacheMissCount == externalIDs.length ) { // nothing was found in cache
      switch ( lookup ) {
        case LOOKUP_BY_CONST_INT :
          for( i = 0; i < externalIDs.length; i++ ) {
            cachePutConstInternalID(poolid, externalIDs[i], localResult[i]);
          }
          break ;
        case LOOKUP_BY_VAR_INT :
          for( i = 0; i < externalIDs.length; i++ ) {
            cachePutVarInternalID(poolid, externalIDs[i], localResult[i]);
          }
          break ;
        default :
          break;
      }
    } else { // some entries where cached -> put uncached entries to cache and result
      switch ( lookup ) {
        case LOOKUP_BY_CONST_INT :
          for( i = 0; i < externalIDs.length; i++ ) {
            if( result[i] == null ) {
              result[i] = localResult[cacheMissCount++];
              cachePutConstInternalID(poolid, externalIDs[i], result[i]);
            }
          }
	  break ;
        case LOOKUP_BY_VAR_INT :
          for( i = 0; i < externalIDs.length; i++ ) {
            if( result[i] == null ) {
              result[i] = localResult[cacheMissCount++];
              cachePutVarInternalID(poolid, externalIDs[i], result[i]);
            }
          }
          break ;
        default :
          break ;
      }
      localResult = result;
    }
    // done
    return localResult;

  }


  // =========
  // IIDMapper ================================================================
  // =========

  // ------------------------------------------------------------------------
  /**
   * Initialize the <code>IDMapper</code>.
   * @param pool the connection pool to use.
   * @param clusterFlag <code>true</code> if running in cluster mode,
   *        <code>false</code> if single-node installation.
   * @param clusterCacheTimeout the clustered cache timeout to use in cluster
   *        mode.
   */
  // ------------------------------------------------------------------------
  public static void configure(JDBCConnectionPool pool,
                               IClusterCacheArea cacheArea) {

    IDMapperCached.cacheArea = cacheArea;

    IDMapperCached.cacheTables = new Hashtable[4];
    IDMapperCached.cacheTables[EXTERNAL_CONST_INTERNAL_MAP] = new Hashtable();
    IDMapperCached.cacheTables[EXTERNAL_VAR_INTERNAL_MAP] = new Hashtable();
    IDMapperCached.cacheTables[CONST_INTERNAL_EXTERNAL_MAP] = new Hashtable();
    IDMapperCached.cacheTables[VAR_INTERNAL_EXTERNAL_MAP] = new Hashtable();

    String poolid = pool.getId();
    findCache(EXTERNAL_CONST_INTERNAL_MAP, poolid);
    findCache(EXTERNAL_VAR_INTERNAL_MAP, poolid);
    findCache(CONST_INTERNAL_EXTERNAL_MAP, poolid);
    findCache(VAR_INTERNAL_EXTERNAL_MAP, poolid);

  }


  // ------------------------------------------------------------------------
  /**
   * Get the mapping for an external ID (RID) to its constant internal ID or
   * create a mapping, if it does not exist.
   * @param pool the connection pool to use.
   * @param externalID the external ID to get/create the mapping for.
   * @return the mapped constant internal ID for the given external ID.
   * @exception IDMapperException if a database error occurred
   */
  public static synchronized String getCreateConstInternalID(JDBCConnectionPool pool,
                                                             String externalID)
                                                      throws IDMapperException {

    if(   ( pool == null )
       || ( externalID == null )
      ) {
      throw new java.lang.IllegalArgumentException();
    }
    // cache lookup
    checkCache();
    String poolid = pool.getId() ;
    String result = cacheGetConstInternalID(poolid, externalID);
    if( result != null ) {
      return result;
    }
    // cache miss
    result = IDMapper.getCreateConstInternalID(pool, externalID);
    cachePutConstInternalID(poolid, externalID, result);
    return result;

  }


  // ------------------------------------------------------------------------
  /**
   * Get the mapping for an external ID (RID) to its constant internal ID.
   * @param pool the connection pool to use.
   * @param externalID the external ID to get the mapping for.
   * @return the mapped constant internal ID for the given external ID or
   *        <code>null</code> if the mapping doesn't exist.
   * @exception IDMapperException if a database error occurred
   */
  public static String getConstInternalID(JDBCConnectionPool pool,
                                          String externalID)
                                   throws IDMapperException {

    if(   ( pool == null )
       || ( externalID == null )
      ) {
      throw new java.lang.IllegalArgumentException();
    }
    // cache lookup
    checkCache();
    String poolid = pool.getId();
    String result = cacheGetConstInternalID(poolid, externalID);
    if( result != null ) {
      return result;
    }
    // cache miss
    result = IDMapper.getConstInternalID(pool, externalID);
    if( result != null ) {
      cachePutConstInternalID(poolid, externalID, result);
    }
    return result;

  }


  // ------------------------------------------------------------------------
  /**
   * Get the mappings for a given array of external IDs (RIDs) to their
   * constant internal IDs.
   * @param pool the connection pool to use.
   * @param externalIDs the array of external IDs to get the mappings for.
   * @return the array of mapped constant internal IDs for the given external
   *        IDs (which might contain <code>null</code>s for those external IDs
   *        with non-existing mapping).
   * @exception IDMapperException if a database error occurred
   */
  public static String[] getConstInternalIDs(JDBCConnectionPool pool,
                                             String[] externalIDs)
                                      throws IDMapperException {

    return getInternalIDs(pool, externalIDs, LOOKUP_BY_CONST_INT);

  }


  // ------------------------------------------------------------------------
  /**
   * Get the mapping for an external ID (RID) to its variable internal ID or
   * create a mapping, if it does not exist.
   * @param pool the connection pool to use.
   * @param externalID the external ID to get/create the mapping for.
   * @return the mapped variable internal ID for the given external ID.
   * @exception IDMapperException if a database error occurred
   */
  public static synchronized String getCreateVarInternalID(JDBCConnectionPool pool,
                                                           String externalID)
                                                    throws IDMapperException {

    if(   ( pool == null )
       || ( externalID == null )
      ) {
      throw new java.lang.IllegalArgumentException();
    }
    // cache lookup
    checkCache();
    String poolid = pool.getId() ;
    String result = cacheGetVarInternalID(poolid, externalID);
    if( result != null ) {
      return result;
    }
    // cache miss
    result = IDMapper.getCreateVarInternalID(pool, externalID);
    cachePutVarInternalID(poolid, externalID, result);
    return result;

  }


  // ------------------------------------------------------------------------
  /**
   * Get the mapping for an external ID (RID) to its variable internal ID.
   * @param pool the connection pool to use.
   * @param externalID the external ID to get the mapping for.
   * @return the mapped variable internal ID for the given external ID or
   *        <code>null</code> if the mapping doesn't exist.
   * @exception IDMapperException if a database error occurred
   */
  public static String getVarInternalID(JDBCConnectionPool pool,
                                        String externalID)
                                 throws IDMapperException {

    if(   ( pool == null )
       || ( externalID == null )
      ) {
      throw new java.lang.IllegalArgumentException();
    }
    // cache lookup
    checkCache();
    String poolid = pool.getId() ;
    String result = cacheGetVarInternalID(poolid, externalID);
    if( result != null ) {
      return result;
    }
    // cache miss
    result = IDMapper.getVarInternalID(pool, externalID);
    if( result != null ) {
      cachePutVarInternalID(poolid, externalID, result);
    }
    return result;
  }


  // ------------------------------------------------------------------------
  /**
   * Get the mappings for a given array of external IDs (RIDs) to their
   * variable internal IDs.
   * @param pool the connection pool to use.
   * @param externalIDs the array of external IDs to get the mappings for.
   * @return the array of mapped constant internal IDs for the given external
   *        IDs (which might contain <code>null</code>s for those external IDs
   *        with non-existing mapping).
   * @exception IDMapperException if a database error occurred
   */
  public static String[] getVarInternalIDs (JDBCConnectionPool pool,
                                            String[] externalIDs)
                                     throws IDMapperException {

    return getInternalIDs(pool, externalIDs, LOOKUP_BY_VAR_INT);

  }


  // ------------------------------------------------------------------------
  /**
   * Get the mapping for a given const internal IDs to its external ID (RID).
   * @param pool the connection pool to use.
   * @param constInternalID the const internal ID to get the mapping for.
   * @return the mapped external internal ID for the given const internal
   *        ID, or <code>null</code>s if the mapping doesn't exist.
   * @exception IDMapperException if a database error occurred
   */
  public static String getExternalIDFromConst(JDBCConnectionPool pool,
                                              String constInternalID)
                                       throws IDMapperException {

    if(   ( pool == null )
       || ( constInternalID == null )
      ) {
      throw new java.lang.IllegalArgumentException();
    }
    // cache lookup
    checkCache();
    String poolid = pool.getId() ;
    String result = cacheGetExternalIDFromConst(poolid, constInternalID);
    if( result != null ) {
      return result;
    }
    // cache miss
    result = IDMapper.getExternalIDFromConst(pool, constInternalID);
    cachePutExternalIDFromConst(poolid, constInternalID, result);
    return result;

  }


  // ------------------------------------------------------------------------
  /**
   * Get the mapping for a given var internal IDs to its external ID (RID).
   * @param pool the connection pool to use.
   * @param constInternalID the const var ID to get the mapping for.
   * @return the mapped external internal ID for the given var internal
   *        ID, or <code>null</code>s if the mapping doesn't exist.
   * @exception IDMapperException if a database error occurred
   */
  public static String getExternalIDFromVar(JDBCConnectionPool pool,
                                            String varInternalID)
                                     throws IDMapperException {

    if(   ( pool == null )
       || ( varInternalID == null )
      ) {
      throw new java.lang.IllegalArgumentException();
    }
    // cache lookup
    checkCache();
    String poolid = pool.getId() ;
    String result = cacheGetExternalIDFromVar(poolid, varInternalID);
    if ( result != null ) {
      return result;
    }
    // cache miss
    result = IDMapper.getExternalIDFromVar(pool, varInternalID);
    cachePutExternalIDFromVar(poolid, varInternalID, result);
    return result;

  }


  // ------------------------------------------------------------------------
  /**
   * Remove the mapping(s) for a given external ID (RID).
   * @param pool the connection pool to use.
   * @param externalID the external ID to remove the mappings for.
   * @return <code>true</code> if any mappings have been removed,
   *        <code>false</code> if no mapping had to be removed.
   * @exception IDMapperException if a database error occurred
   */
  public static synchronized boolean removeIDMapping(JDBCConnectionPool pool,
                                                     String externalID)
                                              throws IDMapperException {
    if(   ( pool == null )
       || ( externalID == null )
      ) {
      throw new java.lang.IllegalArgumentException();
    }
    boolean result = IDMapper.removeIDMapping(pool, externalID);
    checkCache();
    cacheRemoveExternalID(pool.getId(), externalID);
    cacheArea.invalidate();
    return result;

  }

  // ------------------------------------------------------------------------
  /**
   * Get all the mapping(s) for a those external IDs (RIDs) which begin with
   * the given externalID (externalID and externalID/*).
   * @param pool the connection pool to use.
   * @param externalIDpath the prefix of those external ID to get the
   *        mappings for.
   * @return a <code>Map</code> of mappings, indexed by their external ID.
   * @exception IDMapperException if a database error occurred
   */
  public static Map getMultipleIDMappings(JDBCConnectionPool pool,
                                          String externalIDpath)
                                   throws IDMapperException {
    if(   ( pool == null )
       || ( externalIDpath == null )
      ) {
      throw new java.lang.IllegalArgumentException();
    }
    // unable to retrieve all matching entries from cache -> use database directly
    return IDMapper.getMultipleIDMappings(pool, externalIDpath);

  }


  // ------------------------------------------------------------------------
  /**
   * Remove all the mapping(s) for the external IDs (RIDs) for the given
   * path (externalID and externalID/*).
   * @param pool the connection pool to use.
   * @param externalIDpath the path of the external IDs to remove the mappings
   *        for.
   * @exception IDMapperException if a database error occurred
   */
  public static void removeMultipleIDMappings(JDBCConnectionPool pool,
                                              String externalIDpath)
                                       throws IDMapperException {

    if(   ( pool == null )
       || ( externalIDpath == null )
      ) {
      throw new java.lang.IllegalArgumentException();
    }
    IDMapper.removeMultipleIDMappings(pool, externalIDpath);
    if( IDMapperCached.cacheTables == null ) return; // not yet initialized
    // cacheClear(); would be brute force
    checkCache();
    String poolid = pool.getId();
    HashSet idSet = new HashSet(IDMapperCached.cacheTables[EXTERNAL_CONST_INTERNAL_MAP].keySet());
    idSet.addAll(IDMapperCached.cacheTables[EXTERNAL_VAR_INTERNAL_MAP].keySet());
    String idPath = externalIDpath + "/";
    int idPathLength = externalIDpath.length();
    Iterator i = idSet.iterator();
    String id;
    while( i.hasNext() ) {
      id = (String)i.next();
      if( id.length() == idPathLength ) {
        if( id.equals(externalIDpath) ) { // is given ID
          cacheRemoveExternalID(poolid, id);
        }
      } else if( id.length() > idPathLength ) {
        if( id.startsWith(idPath) ) { // starts with given path
          cacheRemoveExternalID(poolid, id);
        }
      }
    }
    cacheArea.invalidate();

  }


  // ------------------------------------------------------------------------
  /**
   * Rename all the mapping(s) for the given external ID (RID) to the new
   * external ID.
   * @param pool the connection pool to use.
   * @param fromExternalID the old external IDs to be renamed from.
   * @param toExternalID the new external IDs to be renamed to.
   * @return <code>true</code> if successful (if a mapping existed).
   * @exception IDMapperException if a database error occurred
   */
  public static synchronized boolean renameSingleExternalID(JDBCConnectionPool pool,
                                                            String fromExternalID,
                                                            String toExternalID)
                                                     throws IDMapperException {

    if(   ( pool == null )
       || ( fromExternalID == null )
       || ( toExternalID == null )
      ) {
      throw new java.lang.IllegalArgumentException();
    }
    boolean result = IDMapper.renameSingleExternalID(pool, fromExternalID, toExternalID);
    checkCache();
    cacheRenameExternalID(pool.getId(), fromExternalID, toExternalID);
    cacheArea.invalidate();
    return result;

  }


  // ------------------------------------------------------------------------
  /**
   * Rename all the mapping(s) for the given external ID (RID) to the new
   * external ID.
   * @param pool the connection pool to use.
   * @param fromExternalID the old external IDs to be renamed from.
   * @param toExternalID the new external IDs to be renamed to.
   * @return <code>true</code> if successful (if a mapping existed).
   * @exception IDMapperException if a database error occurred
   */
  public static synchronized void renameMultipleExternalIDs(JDBCConnectionPool pool,
                                                            String fromExternalIDpath,
                                                            String toExternalIDpath)
                                                     throws IDMapperException {

    if(   ( pool == null )
       || ( fromExternalIDpath == null )
       || ( toExternalIDpath == null )
      ) {
      throw new java.lang.IllegalArgumentException();
    }
    IDMapper.renameMultipleExternalIDs(pool, fromExternalIDpath, toExternalIDpath, new CacheCallback(pool.getId()));
    cacheArea.invalidate();
  }


}
