/*
 * Copyright (c) 2003 by SAP AG. All Rights Reserved.
 *
 * SAP, mySAP, mySAP.com and other SAP products and
 * services mentioned herein as well as their respective
 * logos are trademarks or registered trademarks of
 * SAP AG in Germany and in several other countries all
 * over the world. MarketSet and Enterprise Buyer are
 * jointly owned trademarks of SAP AG and Commerce One.
 * All other product and service names mentioned are
 * trademarks of their respective companies.
 *
 * @version $Id: LoggingConfigurator.java,v 1.1.1.1 2004/02/18 12:56:47 sync Exp $
 */

package com.sapportals.wcm.util.logging;

import com.sap.tc.logging.*;

import java.io.*;
import java.util.*;

/**
 * Load, activate and update logging configuration when changed; log files will
 * be placed in parallel with the configuration file if supplied with home based
 * file path.
 *
 * @created 21. Mrz 2002
 */
public class LoggingConfigurator extends PropertiesConfigurator {
  private static Location logger =
    Location.getLocation(com.sapportals.wcm.util.logging.LoggingConfigurator.class);

  private File loggingConfigurationFile = null;
  private Class loggingConfigurationResourceLoader = null;
  private String loggingConfigurationResourceName = null;
  private Thread periodicityThread = null;
  private int periodicityUpdateIntervall = 1000;

  /**
   * Load and activate logging configuration. Set this class class loader as
   * logging class loader so that this package logging formatter is found when
   * requested. Note: The imput parameters must be valid and must not be null!
   *
   * @param loggingConfigurationFile file holding logging configuration
   */
  public LoggingConfigurator(File loggingConfigurationFile) {
    super(new Properties(), LoggingFormatter.class.getClassLoader());
    this.loggingConfigurationFile = loggingConfigurationFile;
  }

  /**
   * Load and activate logging configuration. Set this class class loader as
   * logging class loader so that this package logging formatter is found when
   * requested. Note: The imput parameters must be valid and must not be null!
   *
   * @param loggingConfigurationResourceClassLoader file holding logging
   *      configuration
   * @param loggingConfigurationResourceName file holding logging configuration
   */
  public LoggingConfigurator(
    Class loggingConfigurationResourceClassLoader,
    String loggingConfigurationResourceName) {
    super(new Properties(), LoggingFormatter.class.getClassLoader());
    this.loggingConfigurationResourceLoader =
      loggingConfigurationResourceClassLoader;
    this.loggingConfigurationResourceName =
      loggingConfigurationResourceName;
  }

  /**
   * Configure logging settings using the logging configuration file and home
   * based log file locations to be in parallel with the configuration file.
   */
  public void configure() {
    // Read logging configuration properties
    Properties properties = new Properties();
    try {
      // Read logging configuration properties from stream
      InputStream loggingConfigurationStream = null;
      if (loggingConfigurationFile != null) {
        loggingConfigurationStream =
          new FileInputStream(loggingConfigurationFile);
      }
      else if (
        (loggingConfigurationResourceLoader != null)
         && (loggingConfigurationResourceName != null)) {
        loggingConfigurationStream =
          loggingConfigurationResourceLoader.getResourceAsStream(
          loggingConfigurationResourceName);
      }
      properties.load(loggingConfigurationStream);
      loggingConfigurationStream.close();
    }
    catch (Exception exception) {
      throw new IllegalArgumentException("Configuration file couldn't be read!");
    }

    // Manipulate log file locations to be in parallel with the configuration file
    //if ( loggingConfigurationFile != null )
    //{
    //    for( Enumeration enum = properties.propertyNames();
    //        enum.hasMoreElements(); )
    //    {
    //        String key = ( ( String )enum.nextElement() ).trim();
    //        if( key.endsWith( "].pattern" ) )
    //        {
    //            String value = properties.getProperty( key );
    //            int percentIndex = 0;
    //            while( ( ( percentIndex + 1 ) < value.length() )
    //                 && ( ( percentIndex = value.indexOf( '%', percentIndex ) ) != -1 ) )
    //            {
    //                if( ( ( percentIndex + 1 ) < value.length() )
    //                     && ( value.charAt( percentIndex + 1 ) == 'h' ) )
    //                {
    //                    properties.setProperty(
    //                        key,
    //                        value.substring( 0, percentIndex )
    //                         + loggingConfigurationFile.getParent()
    //                         + value.substring( percentIndex + 2 ) );
    //                }
    //                percentIndex += 2;
    //            }
    //        }
    //    }
    //}

    // Set logging configuration properties
    setProperties(properties);

    // Configure logging now based on read logging configuration
    super.configure();
  }

  /**
   * Set time interval to update logging configuration. This methods overwrites
   * the library implementation which is based on minutes. It is checking the
   * file modification date of the given file every updateIntervall milliseconds
   * and only reads and activates the configuration when it has changed. If
   * applied multiple times, the former defintion get ovcerwritten by the later.
   *
   * @param updateIntervall update intervall in millicseconds
   */
  public synchronized void setPeriodicity(final int updateIntervall) {
    if (loggingConfigurationFile != null) {
      periodicityUpdateIntervall = updateIntervall;
      if (periodicityThread == null) {
        periodicityThread = new Thread(
          new Runnable() {
            /**
             * Main processing method for the LoggingConfigurator object.
             */
            public void run() {
              long loggingConfigurationTime =
                loggingConfigurationFile.lastModified();
              while (Thread.currentThread() == periodicityThread) {
                try {
                  // Configure logging if changed
                  if (loggingConfigurationFile.lastModified()
                     > loggingConfigurationTime) {
                    loggingConfigurationTime =
                      loggingConfigurationFile.lastModified();
                    logger.infoT(
                      "setPeriodicity(164)",
                      "Configure logging because configuraion has changed on "
                       + new Date(loggingConfigurationTime));
                    configure();
                  }

                  // Go to sleep
                  Thread.sleep(periodicityUpdateIntervall);
                }
                catch (Exception exception) {
                  logger.errorT(
                    "setPeriodicity(175)",
                    "Update logging configuration thread catched exception: "
                     + LoggingFormatter.extractCallstack(
                    exception));
                }
              }
            }
          });
        periodicityThread.setDaemon(true);
        periodicityThread.start();
      }
    }
    else {
      throw new IllegalArgumentException("Periodicity can only be set for file configured logging!");
    }
  }
}
