/*
 * Copyright (c) 2003 by SAP AG. All Rights Reserved.
 *
 * SAP, mySAP, mySAP.com and other SAP products and
 * services mentioned herein as well as their respective
 * logos are trademarks or registered trademarks of
 * SAP AG in Germany and in several other countries all
 * over the world. MarketSet and Enterprise Buyer are
 * jointly owned trademarks of SAP AG and Commerce One.
 * All other product and service names mentioned are
 * trademarks of their respective companies.
 *
 * @version $Id: LoggingInitializer.java,v 1.1.1.1 2004/02/18 12:56:48 sync Exp $
 */

package com.sapportals.wcm.util.logging;

import com.sap.tc.logging.*;

import java.io.*;
import java.net.*;
import java.util.*;
// import javax.servlet.http.HttpServletRequest;

/**
 * Trigger initialization of logging.
 *
 * @created 21. M�rz 2002
 */
public class LoggingInitializer {
  private static LoggingInitializer inst = null;

  private static boolean isConfigurationRegistered = false;

  private Map confFiles;

  /**
   * Construct LoggingInitializer.
   */
  private LoggingInitializer() {
    // Initialize map of configuration files
    confFiles = new HashMap();
  }

  /**
   * Get singleton instance of this class.
   *
   * @return Singleton instance
   */
  public static synchronized LoggingInitializer getInstance() {
    // Check if singleton wasn't instantiated yet
    if (inst == null) {
      // Instantiate singleton instance now
      try {
        inst = new LoggingInitializer();
      } catch (Exception exception) {
        // Nothing to do except for logging/tracing since inst wasn't assigned yet
      }
    }

    // Return singleton instance
    return inst;
  }

  /**
   * Register file based configuration. Reregistering an already registered
   * configuration will be ignored.
   *
   * @param configurationFile configuration file
   */
  public synchronized void registerConfiguration(File configurationFile) {
    // Safeguard operations
    try {
      // Check argument validity
      if (configurationFile == null) {
        throw new IllegalArgumentException("Configuration file may not be null!");
      }
      if (!configurationFile.exists()) {
        throw new IllegalArgumentException("Configuration file does not exist!");
      }

      // Register configuration file if not already done
      if (!confFiles.containsKey(configurationFile)) {
        LoggingConfigurator loggingConfigurator = new LoggingConfigurator(configurationFile);
        loggingConfigurator.configure();
        loggingConfigurator.setPeriodicity(1000);
        isConfigurationRegistered = true;
        confFiles.put(configurationFile, loggingConfigurator);
        Location logger = Location.getLocation(LoggingInitializer.class);
        System.out.println("Successfully initialized logging based on the configuration file " + configurationFile.getAbsolutePath() + ".");
        logger.infoT("registerConfiguration(92)", "Successfully initialized logging based on the configuration file " + configurationFile.getAbsolutePath());
      } else {
        Location logger = Location.getLocation(LoggingInitializer.class);
        System.out.println("Ignored logging re-initialization based on the configuration file " + configurationFile.getAbsolutePath() + ".");
        logger.warningT(
          "registerConfiguration(97)",
          "Ignored logging re-initialization based on the configuration file "
            + configurationFile.getAbsolutePath()
            + " in: "
            + LoggingFormatter.extractCallstack(new Exception()));
      }
    } catch (Error error) {
      // Handle caught throwable
      System.err.println("Failed to initialize logging by configuration file " + configurationFile + ": " + LoggingFormatter.extractCallstack(error));
      throw error;
    } catch (RuntimeException runtimeException) {
      // Handle caught throwable
      System.err.println(
        "Failed to initialize logging by configuration file " + configurationFile + ": " + LoggingFormatter.extractCallstack(runtimeException));
      throw runtimeException;
    } catch (Throwable throwable) {
      // Handle caught throwable
      System.err.println("Failed to initialize logging by configuration file " + configurationFile + ": " + LoggingFormatter.extractCallstack(throwable));
      throw new RuntimeException(
        "Failed to initialize logging by configuration file " + configurationFile + ": " + LoggingFormatter.extractCallstack(throwable));
    }
  }

  /**
   * Register file URL based configuration (only file URLs are supported).
   * Reregistering an already registered configuration will be ignored.
   *
   * @param configurationURL file URL of the configuration file
   */
  public synchronized void registerConfiguration(URL configurationURL) {
    // Safeguard operations
    try {
      // Check argument validity
      if (configurationURL == null) {
        throw new IllegalArgumentException("Configuration URL may not be null!");
      }
      if (!"file".equalsIgnoreCase(configurationURL.getProtocol())) {
        throw new IllegalArgumentException("Configuration URL is not of FILE protocol!");
      }

      // Call the file based registration version
      String fileName = configurationURL.getFile();
      if (fileName == null) {
        throw new RuntimeException("Configuration file identified by URL can't be found!");
      }
      registerConfiguration(new File(fileName));
    } catch (Error error) {
      // Handle caught throwable
      System.err.println("Failed to initialize logging by configuration URL " + configurationURL + ": " + LoggingFormatter.extractCallstack(error));
      throw error;
    } catch (RuntimeException runtimeException) {
      // Handle caught throwable
      System.err.println("Failed to initialize logging by configuration URL " + configurationURL + ": " + LoggingFormatter.extractCallstack(runtimeException));
      throw runtimeException;
    } catch (Throwable throwable) {
      // Handle caught throwable
      System.err.println("Failed to initialize logging by configuration URL " + configurationURL + ": " + LoggingFormatter.extractCallstack(throwable));
      throw new RuntimeException("Failed to initialize logging by configuration URL " + configurationURL + ": " + LoggingFormatter.extractCallstack(throwable));
    }
  }

  /**
   * Register resource name based configuration (the resource file must be
   * accessible from the classpath of this class). Reregistering an already
   * registered configuration will be ignored.
   *
   * @param resourceName resource name of the configuration file
   */
  public synchronized void registerConfiguration(String resourceName) {
    registerConfiguration(this.getClass(), resourceName);
  }

  /**
   * Register resource name based configuration (the resource file must be
   * accessible from the classpath of the given class loader). Reregistering an
   * already registered configuration will be ignored.
   *
   * @param resourceLoader class being able to load the resource
   * @param resourceName resource name of the configuration file
   */
  public synchronized void registerConfiguration(Class resourceLoader, String resourceName) {
    // Safeguard operations
    try {
      // Check argument validity
      if (resourceLoader == null) {
        throw new IllegalArgumentException("Resource loader may not be null!");
      }
      if (resourceName == null) {
        throw new IllegalArgumentException("Resource name may not be null!");
      }

      // Try the file based registration version
      URL resourceUrl = resourceLoader.getResource(resourceName);
      String fileName = null;
      if (resourceUrl != null) {
        fileName = resourceUrl.getFile();
      } else {
        fileName = findConfigurationFilePath(getClassPath(resourceLoader), resourceName) + resourceName;
      }
      if (fileName == null) {
        throw new RuntimeException("Configuration resource file for logging is not stored in the classpath!");
      }

      File configurationFile = new File(fileName);
      if (configurationFile == null) {
        throw new IllegalArgumentException("Configuration resource file for logging is not stored in the classpath!");
      }
      if (!configurationFile.exists()) {
        // Try to use resource based configuration
        if (resourceLoader.getResourceAsStream(resourceName) == null) {
          throw new IllegalArgumentException("Configuration resource file for logging is not stored in the classpath!");
        }

        // Register configuration resource file if not done already done
        if (!confFiles.containsKey(configurationFile)) {
          LoggingConfigurator loggingConfigurator = new LoggingConfigurator(resourceLoader, resourceName);
          loggingConfigurator.configure();
          isConfigurationRegistered = true;
          confFiles.put(configurationFile, loggingConfigurator);
          Location logger = Location.getLocation(LoggingInitializer.class);
          System.out.println("Successfully initialized logging based on the configuration resource file " + configurationFile.getAbsolutePath() + ".");
          logger.infoT(
            "registerConfiguration(215)",
            "Successfully initialized logging based on the configuration resource file " + configurationFile.getAbsolutePath());
        } else {
          Location logger = Location.getLocation(LoggingInitializer.class);
          System.out.println("Ignored logging re-initialization based on the configuration resource file " + configurationFile.getAbsolutePath() + ".");
          logger.warningT(
            "registerConfiguration(220)",
            "Ignored logging re-initialization based on the configuration resource file "
              + configurationFile.getAbsolutePath()
              + " in: "
              + LoggingFormatter.extractCallstack(new Exception()));
        }
      } else {
        // Use normal file based configuration in order to benefit from
        // its advantages like periodic configuration file reloading
        registerConfiguration(configurationFile);
      }
    } catch (Error error) {
      // Handle caught throwable
      System.err.println("Failed to initialize logging by configuration resource name " + resourceName + ": " + LoggingFormatter.extractCallstack(error));
      throw error;
    } catch (RuntimeException runtimeException) {
      // Handle caught throwable
      System.err.println(
        "Failed to initialize logging by configuration resource name " + resourceName + ": " + LoggingFormatter.extractCallstack(runtimeException));
      throw runtimeException;
    } catch (Throwable throwable) {
      // Handle caught throwable
      System.err.println("Failed to initialize logging by configuration resource name " + resourceName + ": " + LoggingFormatter.extractCallstack(throwable));
      throw new RuntimeException(
        "Failed to initialize logging by configuration resource name " + resourceName + ": " + LoggingFormatter.extractCallstack(throwable));
    }
  }

  private String getClassPath(Class aClass) {
    final String JAR_FILE_PREFIX = "jar:file:";
    final String FILE_PREFIX = "file:";

    try {
      String className = aClass.getName().replace('.', '/') + ".class";
      ClassLoader classLoader = aClass.getClassLoader();
      String url = classLoader.getResource(className).toString();

      if (url.startsWith(JAR_FILE_PREFIX)) {
        return url.substring(JAR_FILE_PREFIX.length(), url.length() - className.length() - 2);
      }
      // ???
      if (url.startsWith(FILE_PREFIX)) {
        return url.substring(FILE_PREFIX.length());
      }

      return null;
    } catch (Exception e) {
      return null;
    }
  }

  private String findConfigurationFilePath(String classpath, String confFile) {
    boolean found = false;
    String path = classpath;

    // first search for the config file at the location of the jar file
    if (path != null) {
      String dir = new File(path).getParent();
      found = new File(dir, confFile).exists();
      if (found) {
        return dir;
      }
    }

    // second search for the config file in the classpath
    StringTokenizer paths = new StringTokenizer(System.getProperty("java.class.path"), System.getProperty("path.separator"));
    confFile = (String)System.getProperty("file.separator") + confFile;

    while (!found && paths.hasMoreTokens()) {
      path = paths.nextToken();
      File cFile = new File(path);
      if (cFile.isFile()) {
        File nFile = new File(cFile.getParent(), confFile);
        if (nFile.exists()) {
          return nFile.getParent();
        }
      } else {
        File nFile = new File(cFile.getAbsolutePath(), confFile);
        if (nFile.exists()) {
          return nFile.getParent();
        }
      }
    }

    return null;
  }

  /**
   * Check whether or not a configuration was already registered.
   *
   * @return true when a configuration was already registered
   */
  public static boolean isConfigurationRegistered() {
    return isConfigurationRegistered;
  }
}
