/*
 * Copyright (c) 2003 by SAP AG. All Rights Reserved.
 *
 * SAP, mySAP, mySAP.com and other SAP products and
 * services mentioned herein as well as their respective
 * logos are trademarks or registered trademarks of
 * SAP AG in Germany and in several other countries all
 * over the world. MarketSet and Enterprise Buyer are
 * jointly owned trademarks of SAP AG and Commerce One.
 * All other product and service names mentioned are
 * trademarks of their respective companies.
 *
 * @version $Id: LoggingResolver.java,v 1.1.1.1 2004/02/18 12:56:41 sync Exp $
 */

package com.sapportals.wcm.util.logging;

import java.io.*;
import java.util.*;

import org.apache.oro.text.regex.*;

/**
 * Class scanning for java files and therein location method calls and fixing
 * the method name if invalid.
 *
 * @created 19. Oktober 2001
 */
public class LoggingResolver {
  private static File sourceFileDirectory = null;
  private static PrintWriter protocolFile = null;
  private static boolean usePerforce = false;
  private static boolean modifySource = false;
  private static boolean protocolStdOut = false;
  private static boolean protocolAnalysis = false;

  private static Collection sourceFiles = new ArrayList();

  private static PatternCompiler compiler = new Perl5Compiler();
  private static PatternMatcher matcher = new Perl5Matcher();
  private static MatchResult result = null;
  private static Pattern patternFor_PackStat_ClassBegin_MethodBegin_LocInit_QuotBegin_BlockBegin_ClassMethodBlockEnd = null;
  private static Pattern patternFor_LocStat_QuoteBegin_BlockBegin_MethodBlockEnd = null;
  private static Pattern patternFor_SRemarkEnd = null;
  private static Pattern patternFor_CRemarkEnd = null;
  private static Pattern patternFor_SQuotEsc_SQuotEnd = null;
  private static Pattern patternFor_CQuotEsc_CQuotEnd = null;

  /**
   * Run this program.
   *
   * @param args The command line arguments
   */
  public static void main(String[] args) {
    // Evaluate command line parameters
    try {
      for (int i = 0; i < args.length; i++) {
        if (args[i].indexOf("=") != -1) {
          StringTokenizer argumentTokenizer = new StringTokenizer(args[i], "=");
          String key = argumentTokenizer.nextToken();
          String val = argumentTokenizer.nextToken();
          if ((key != null) && (key.toUpperCase().indexOf("SRC") != -1)) {
            key = null;
            sourceFileDirectory = new File(val);
            if (!sourceFileDirectory.exists()) {
              throw new Exception("Java source file/directory path invalid!");
            }
          }
          if ((key != null) && (key.toUpperCase().indexOf("LOG") != -1)) {
            key = null;
            protocolFile = new PrintWriter(new FileOutputStream(new File(val)));
          }
          if (key != null) {
            throw new Exception("Invalid parameter " + key + " specified!");
          }
        }
        else {
          if ((args[i] != null) && (args[i].substring(1, args[i].length()).equalsIgnoreCase("PERFORCE"))) {
            usePerforce = true;
            args[i] = null;
          }
          if ((args[i] != null) && (args[i].substring(1, args[i].length()).equalsIgnoreCase("MODIFY"))) {
            modifySource = true;
            args[i] = null;
          }
          if ((args[i] != null) && (args[i].substring(1, args[i].length()).equalsIgnoreCase("STDOUT"))) {
            protocolStdOut = true;
            args[i] = null;
          }
          if ((args[i] != null) && (args[i].substring(1, args[i].length()).equalsIgnoreCase("ANALYSIS"))) {
            protocolAnalysis = true;
            args[i] = null;
          }
          if ((args[i] != null) && ((args[i].substring(1, args[i].length()).equalsIgnoreCase("HELP"))
             || (args[i].substring(1, args[i].length()).equalsIgnoreCase("H"))
             || (args[i].substring(1, args[i].length()).equalsIgnoreCase("?")))) {
            throw new Exception("Help requested!");
          }
          if (args[i] != null) {
            throw new Exception("Invalid parameter " + args[i] + " specified!");
          }
        }
      }
      if (sourceFileDirectory == null) {
        throw new Exception("Mandatory parameter for java source file/dir path not provided!");
      }
    }
    catch (Exception exception) {
      exception.printStackTrace(System.err);
      System.err.println();
      System.err.println("Usage: java <package>.LoggingResolver /SRC=<java source file/dir path>");
      System.err.println("                                     [/LOG=<protocol file path>]");
      System.err.println("                                     [/PERFORCE] [/MODIFY]");
      System.err.println("                                     [/STDOUT] [/ANALYSIS]");
      System.err.println();
      System.err.println("Either the specified source file or all *.java files inside the specified source");
      System.err.println("directory will be parsed for location constructions and statements. If no or an");
      System.err.println("illegal class/method name is found, the class/method name will be fixed.");
      System.err.println("If the PERFORCE flag is set, the file will be checked out from Perforce (no sync");
      System.err.println("will be done; just edit). If and only if the MODIFY flag is set, the file will");
      System.err.println("be modified and written back to disk.");
      System.err.println("The program will print out a list of all modified files. A protocol file, if the");
      System.err.println("the parameter is provided, will contain a more verbose protocol. If the ANALYSIS");
      System.err.println("flag is set, a detailed parse protocol will be written. Using the STDOUT flag,");
      System.err.println("all output will also be written to your console.");
      System.exit(-1);
    }

    // Open protocol
    openProtocol();

    // Scan for java files
    try {
      if (sourceFileDirectory.isDirectory()) {
        protocolMessage("");
        protocolMessage("Identifying java sources:");
        scanDirectory(sourceFileDirectory);
      }
      else {
        protocolMessage("");
        protocolMessage("Requested java source:");
        String fileName = null;
        try {
          fileName = sourceFileDirectory.getCanonicalPath();
        }
        catch (IOException exception) {
          fileName = sourceFileDirectory.getAbsolutePath();
        }
        // Ensure canonical absolute path
        sourceFiles.add(new File(fileName));
        protocolMessage(fileName);
      }
    }
    catch (Exception exception) {
      protocolMessage("Failed to locate java source files in given path " + sourceFileDirectory.getAbsolutePath() + ": " + exception);
      if (!protocolStdOut) {
        System.err.println("Failed to locate java source files in given path " + sourceFileDirectory.getAbsolutePath() + ": " + exception);
      }
      System.exit(-1);
    }

    // Resolve java files
    try {
      // Pattern for package statement, class begin, method begin, location construction, simple/complex remark begin, quotation begin, block begin or class/method/block end
      patternFor_PackStat_ClassBegin_MethodBegin_LocInit_QuotBegin_BlockBegin_ClassMethodBlockEnd = compiler.compile(
        "(package\\s*([_$a-zA-Z][\\._$a-zA-Z0-9]*)\\s*;)|(\\s*(interface|class)\\s+([^\\s=;\"\\{]+)\\s*[^=;\"\\{]*\\{)|(([_$a-zA-Z][_$a-zA-Z0-9]*)\\s*=\\s*(new\\s+Location|[^;]+\\.getLocation)\\s*\\(\\s*((('([^\\\\']*|\\\\'|\\\\[^'])*')|(\"([^\\\\\"]*|\\\\\"|\\\\[^\"])*\")|([^,;'\"])*)*)\\s*(,|;))|(\\s*((static)\\s*|(([_$a-zA-Z][_$a-zA-Z0-9]*)\\s*\\([^\\)]*\\)[^=;\"\\{]*))\\{)|(//)|(/\\*)|(')|(\")|(\\{)|(\\})");

      // Pattern for location statement, simple/complex remark begin, quotation begin, block begin or method/block end
      patternFor_LocStat_QuoteBegin_BlockBegin_MethodBlockEnd = compiler.compile(
        "([_$a-zA-Z][_$a-zA-Z0-9]*\\.(entering|debugT|infoT|warningT|errorT|fatalT|assertion|throwing)\\s*\\(\\s*((('([^\\\\']*|\\\\'|\\\\[^'])*')|(\"([^\\\\\"]*|\\\\\"|\\\\[^\"])*\")|([^,;'\"])*)*)\\s*(,|;))|(//)|(/\\*)|(')|(\")|(\\{)|(\\})");

      // Pattern for simple remark end
      patternFor_SRemarkEnd = compiler.compile(
        "\\n");

      // Pattern for complex remark end
      patternFor_CRemarkEnd = compiler.compile(
        "\\*/");

      // Pattern for escaped quotation or quotation end
      patternFor_SQuotEsc_SQuotEnd = compiler.compile(
        "(\\\\\\\\)|(\\\\')|(')");

      // Pattern for escaped quotation or quotation end
      patternFor_CQuotEsc_CQuotEnd = compiler.compile(
        "([^\\\\\\\"]*\\\\\\\\)|([^\\\\\\\"]*\\\\\")|(\")");

      protocolMessage("");
      protocolMessage("Resolving java sources:");
      for (Iterator iter = sourceFiles.iterator(); iter.hasNext(); ) {
        File sourceFile = (File)iter.next();
        resolveSource(sourceFile);
      }
    }
    catch (Exception exception) {
      protocolMessage("Failed to resolve java source files: " + exception);
      if (!protocolStdOut) {
        System.err.println("Failed to resolve java source files: " + exception);
      }
      System.exit(-1);
    }

    // Close protocol
    closeProtocol();
  }


  /**
   * Store modified java source file.
   *
   * @param sourceFile java source file
   * @param content java source file content
   */
  public static void storeSource(File sourceFile, StringBuffer content) {
    // Check if content should be modified
    if (modifySource) {
      try {
        // Checkout from perforce
        if (usePerforce) {
          try {
            Process perforceProcess = Runtime.getRuntime().exec("p4 edit " + sourceFile.getAbsolutePath());
            perforceProcess.waitFor();
            if (perforceProcess.exitValue() != 0) {
              throw new Exception("Process error code " + perforceProcess.exitValue() + " not ok.");
            }
          }
          catch (Exception exception) {
            protocolMessage("  Failed to checkout java source file " + sourceFile.getAbsolutePath() + " from perforce: " + exception);
            throw new Exception("Failed to checkout java source file " + sourceFile.getAbsolutePath() + " from perforce: " + exception);
          }
        }

        // Store content
        BufferedWriter outWriter = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(sourceFile)));
        outWriter.write(content.toString());
        outWriter.close();
      }
      catch (Exception exception) {
        protocolMessage("Failed to store java source file " + sourceFile.getAbsolutePath() + ": " + exception);
        if (!protocolStdOut) {
          System.err.println("Failed to store java source file " + sourceFile.getAbsolutePath() + ": " + exception);
        }
      }
    }
  }


  /**
   * Compute line number from buffer and offset.
   *
   * @param patternMatcherBuffer character buffer
   * @param lastLine last line number based on last offset
   * @param lastOffset last offset used to compute last line number
   * @param offset actual offset to compute actual line number
   * @return actual line number
   */
  private static int getLine(char[] patternMatcherBuffer, int lastLine, int lastOffset, int offset) {
    for (int i = lastOffset; i < offset; i++) {
      if (patternMatcherBuffer[i] == '\n') {
        lastLine++;
      }
    }
    return lastLine;
  }


  /**
   * Scan diectory for java source files.
   *
   * @param directory directory to be scanned
   */
  private static void scanDirectory(File directory) {
    File[] filesArray = directory.listFiles();
    File file;
    if (filesArray != null) {
      for (int i = 0; i < filesArray.length; i++) {
        file = filesArray[i];
        if (file.isDirectory()) {
          scanDirectory(file);
        }
        if (file.isFile()) {
          if (file.getName().endsWith(".java")) {
            String fileName = null;
            try {
              fileName = file.getCanonicalPath();
            }
            catch (IOException exception) {
              fileName = file.getAbsolutePath();
            }
            // Ensure canonical absolute path
            sourceFiles.add(new File(fileName));
            protocolMessage(fileName);
          }
        }
      }
    }
  }


  /**
   * Resolve java source file.
   *
   * @param sourceFile java source file
   */
  private static void resolveSource(File sourceFile) {
    final int FILE = 0;
    final int CLASS = 1;
    final int METHOD = 2;
    final int SREMARK = 3;
    final int CREMARK = 4;
    final int SQUOT = 5;
    final int CQUOT = 6;

    try {
      // Read content
      StringBuffer content = new StringBuffer();
      BufferedInputStream inStream = new BufferedInputStream(new FileInputStream(sourceFile));
      byte[] buf = new byte[4096];
      int len = 0;
      while ((len = inStream.read(buf)) != -1) {
        content.append(new String(buf, 0, len));
      }
      if (content.length() > 0) {
        if (content.charAt(content.length() - 1) != '\n') {
          content.append("\n");
        }
      }
      PatternMatcherInput patternMatcherInput = new PatternMatcherInput(content.toString());
      char[] patternMatcherBuffer = patternMatcherInput.getBuffer();
      boolean modified = false;
      int delta = 0;

      // Init and start state machine
      int line = 1;
      int offset = 0;
      String pack = null;
      StateStack stack = new StateStack();
      stack.pushState(FILE);
      while (!stack.isEmpty()) {
        switch (stack.peekStateID()) {
          case FILE:
          case CLASS:
            if (matcher.contains(patternMatcherInput,
              patternFor_PackStat_ClassBegin_MethodBegin_LocInit_QuotBegin_BlockBegin_ClassMethodBlockEnd)) {
              line = getLine(patternMatcherBuffer, line, offset, patternMatcherInput.getCurrentOffset());
              offset = patternMatcherInput.getCurrentOffset();
              result = matcher.getMatch();
              if (result.group(2) != null) {
                // Found package statement
                pack = result.group(2);
                protocolAnalysis("Line " + line + ": package statement (" + pack + ")");
              }
              if (result.group(5) != null) {
                // Found class begin
                stack.pushState(CLASS, ((pack == null) ? "" : pack + ".") + result.group(5));
                protocolAnalysis("Line " + line + ": class begin (" + stack.peekStateArg() + ")");
              }
              if ((result.group(7) != null) && (result.group(9) != null)) {
                // Found location construction
                protocolAnalysis("Line " + line + ": location construction (" + result.group(7) + ")");
                String classString = stack.peekStateArg() + ".class ";
                if (result.group(9).indexOf(classString) == -1) {
                  int closingBracketPos = result.group(9).lastIndexOf(')');
                  if (closingBracketPos != -1) {
                    protocolAnalysis("Line " + line + ": replacing class name " + result.group(9).substring(0, closingBracketPos) + " in between positions " + result.beginOffset(9) + "-" + (result.endOffset(9) - (result.group(9).length() - closingBracketPos)) + " with " + classString);
                    if (!modified) {
                      modified = true;
                      protocolMessage(sourceFile.getAbsolutePath());
                      if (modifySource) {
                        if (!protocolStdOut) {
                          System.out.println(sourceFile.getAbsolutePath());
                        }
                      }
                    }
                    protocolMessage("  Modifying " + result.group(7) + " construction class name " + result.group(9).substring(0, closingBracketPos) + " to " + classString);
                    content.replace(result.beginOffset(9) + delta, (result.endOffset(9) - (result.group(9).length() - closingBracketPos)) + delta, classString);
                    delta += classString.length() - closingBracketPos;
                  }
                }
              }
              if (result.group(19) != null) {
                // Found static method begin
                stack.pushState(METHOD, "<cinit>");
                protocolAnalysis("Line " + line + ": method begin (" + stack.peekStateArg() + ")");
              }
              if (result.group(21) != null) {
                // Found method begin
                stack.pushState(METHOD, result.group(21));
                protocolAnalysis("Line " + line + ": method begin (" + stack.peekStateArg() + ")");
              }
              if (result.group(22) != null) {
                // Found simple remark begin
                protocolAnalysis("Line " + line + ": simple remark begin");
                stack.pushState(SREMARK);
              }
              if (result.group(23) != null) {
                // Found complex remark begin
                protocolAnalysis("Line " + line + ": complex remark begin");
                stack.pushState(CREMARK);
              }
              if (result.group(24) != null) {
                // Found simple quotation begin
                protocolAnalysis("Line " + line + ": simple quotation begin");
                stack.pushState(SQUOT);
              }
              if (result.group(25) != null) {
                // Found quotation begin
                protocolAnalysis("Line " + line + ": complex quotation begin");
                stack.pushState(CQUOT);
              }
              if (result.group(26) != null) {
                // Found block begin
                protocolAnalysis("Line " + line + ": block begin");
                stack.pushState(stack.peekState());
              }
              if (result.group(27) != null) {
                // Found block end
                protocolAnalysis("Line " + line + ": block end");
                stack.popState();
              }
            }
            else {
              // Found nothing more
              stack.popState();
              if (stack.isEmpty()) {
                if (modified) {
                  inStream.close();
                  storeSource(sourceFile, content);
                  modified = false;
                }
              }
              else {
                throw new Exception("Java source malformed in or past line " + line + " and position " + patternMatcherInput.getCurrentOffset());
              }
            }
            break;
          case METHOD:
            if (matcher.contains(patternMatcherInput,
              patternFor_LocStat_QuoteBegin_BlockBegin_MethodBlockEnd)) {
              line = getLine(patternMatcherBuffer, line, offset, patternMatcherInput.getCurrentOffset());
              offset = patternMatcherInput.getCurrentOffset();
              result = matcher.getMatch();
              if ((result.group(2) != null) && (result.group(3) != null) && (result.group(10) != null)) {
                // Found location statement
                protocolAnalysis("Line " + line + ": location statement (" + result.group(2) + ")");
                String methodString = "\"" + stack.peekStateArg() + "(" + line + ")\"";
                if (!result.group(3).trim().startsWith(methodString)) {
                  // Inserting method name
                  if ((result.group(10).equals(";")) && (result.group(2).endsWith("T"))) {
                    protocolAnalysis("Line " + line + ": inserting method name " + methodString + " at position " + result.beginOffset(3));
                    if (!modified) {
                      modified = true;
                      protocolMessage(sourceFile.getAbsolutePath());
                      if (modifySource) {
                        if (!protocolStdOut) {
                          System.out.println(sourceFile.getAbsolutePath());
                        }
                      }
                    }
                    protocolMessage("  Modifying " + result.group(2) + " location statement method name " + methodString);
                    methodString += ", ";
                    content.insert(result.beginOffset(3) + delta, methodString);
                    delta += methodString.length();
                  }

                  // Replacing method name
                  if (result.group(10).equals(",")) {
                    protocolAnalysis("Line " + line + ": replacing method name " + result.group(3) + " in between positions " + result.beginOffset(3) + "-" + result.endOffset(3) + " with " + methodString);
                    if (!modified) {
                      modified = true;
                      protocolMessage(sourceFile.getAbsolutePath());
                      if (modifySource) {
                        if (!protocolStdOut) {
                          System.out.println(sourceFile.getAbsolutePath());
                        }
                      }
                    }
                    protocolMessage("  Modifying " + result.group(2) + " location statement method name " + result.group(3) + " to " + methodString);
                    content.replace(result.beginOffset(3) + delta, result.beginOffset(10) + delta, methodString);
                    delta += methodString.length() - (result.beginOffset(10) - result.beginOffset(3));
                  }
                }
              }
              if (result.group(11) != null) {
                // Found simple remark begin
                protocolAnalysis("Line " + line + ": simple remark begin");
                stack.pushState(SREMARK);
              }
              if (result.group(12) != null) {
                // Found complex remark begin
                protocolAnalysis("Line " + line + ": complex remark begin");
                stack.pushState(CREMARK);
              }
              if (result.group(13) != null) {
                // Found simple quotation begin
                protocolAnalysis("Line " + line + ": simple quotation begin");
                stack.pushState(SQUOT);
              }
              if (result.group(14) != null) {
                // Found quotation begin
                protocolAnalysis("Line " + line + ": complex quotation begin");
                stack.pushState(CQUOT);
              }
              if (result.group(15) != null) {
                // Found block begin
                protocolAnalysis("Line " + line + ": block begin");
                stack.pushState(stack.peekState());
              }
              if (result.group(16) != null) {
                // Found block end
                protocolAnalysis("Line " + line + ": block end");
                stack.popState();
              }
            }
            else {
              // Found nothing more
              throw new Exception("Java source malformed in or past line " + line + " and position " + patternMatcherInput.getCurrentOffset());
            }
            break;
          case SREMARK:
            if (matcher.contains(patternMatcherInput,
              patternFor_SRemarkEnd)) {
              line = getLine(patternMatcherBuffer, line, offset, patternMatcherInput.getCurrentOffset());
              offset = patternMatcherInput.getCurrentOffset();

              // Found simple remark end
              protocolAnalysis("Line " + line + ": simple remark end");
              stack.popState();
            }
            else {
              // Found nothing more
              throw new Exception("Java source malformed in or past line " + line + " and position " + patternMatcherInput.getCurrentOffset());
            }
            break;
          case CREMARK:
            if (matcher.contains(patternMatcherInput,
              patternFor_CRemarkEnd)) {
              line = getLine(patternMatcherBuffer, line, offset, patternMatcherInput.getCurrentOffset());
              offset = patternMatcherInput.getCurrentOffset();

              // Found complex remark end
              protocolAnalysis("Line " + line + ": complex remark end");
              stack.popState();
            }
            else {
              // Found nothing more
              throw new Exception("Java source malformed in or past line " + line + " and position " + patternMatcherInput.getCurrentOffset());
            }
            break;
          case SQUOT:
            if (matcher.contains(patternMatcherInput,
              patternFor_SQuotEsc_SQuotEnd)) {
              line = getLine(patternMatcherBuffer, line, offset, patternMatcherInput.getCurrentOffset());
              offset = patternMatcherInput.getCurrentOffset();
              result = matcher.getMatch();
              if (result.group(1) != null) {
                // Found escaped simple quotation escape
                protocolAnalysis("Line " + line + ": escaped simple quotation escape");
              }
              if (result.group(2) != null) {
                // Found escaped simple quotation
                protocolAnalysis("Line " + line + ": escaped simple quotation");
              }
              if (result.group(3) != null) {
                // Found simple quotation end
                protocolAnalysis("Line " + line + ": simple quotation end");
                stack.popState();
              }
            }
            else {
              // Found nothing more
              throw new Exception("Java source malformed in or past line " + line + " and position " + patternMatcherInput.getCurrentOffset());
            }
            break;
          case CQUOT:
            if (matcher.contains(patternMatcherInput,
              patternFor_CQuotEsc_CQuotEnd)) {
              line = getLine(patternMatcherBuffer, line, offset, patternMatcherInput.getCurrentOffset());
              offset = patternMatcherInput.getCurrentOffset();
              result = matcher.getMatch();
              if (result.group(1) != null) {
                // Found escaped complex quotation escape
                protocolAnalysis("Line " + line + ": escaped complex quotation escape");
              }
              if (result.group(2) != null) {
                // Found escaped complex quotation
                protocolAnalysis("Line " + line + ": escaped complex quotation");
              }
              if (result.group(3) != null) {
                // Found complex quotation end
                protocolAnalysis("Line " + line + ": complex quotation end");
                stack.popState();
              }
            }
            else {
              // Found nothing more
              throw new Exception("Java source malformed in or past line " + line + " and position " + patternMatcherInput.getCurrentOffset());
            }
            break;
        }
      }
      if (modified) {
        // File not stored
        throw new Exception("Java source malformed. Mismatch of undefined reason at undefined line!");
      }
    }
    catch (Exception exception) {
      protocolMessage("Failed to resolve java source file " + sourceFile.getAbsolutePath() + ": " + exception);
      if (!protocolStdOut) {
        System.err.println("Failed to resolve java source file " + sourceFile.getAbsolutePath() + ": " + exception);
      }
    }
  }


  /**
   * Protocol analysis.
   *
   * @param analysis analysis
   */
  private static void protocolAnalysis(String analysis) {
    if (protocolAnalysis) {
      protocolMessage(analysis);
    }
  }


  /**
   * Protocol message.
   *
   * @param message message
   */
  private static void protocolMessage(String message) {
    if (protocolFile != null) {
      protocolFile.println(message);
    }
    if (protocolStdOut) {
      System.out.println(message);
    }
  }


  /**
   * Open protocol.
   */
  private static void openProtocol() {
    protocolMessage("Protocol opened on " + new Date() + " to resolve file/directory " + sourceFileDirectory.getAbsolutePath() + ((usePerforce && modifySource) ? " using Perforce" : "") + " [" + (modifySource ? "WARNING: MODIFY MODE ACTIVATED" : "SIMULATION ONLY") + "].");
  }


  /**
   * Close protocol.
   */
  private static void closeProtocol() {
    protocolMessage("");
    protocolMessage("Protocol closed on " + new Date() + ".");
    protocolMessage("");
    if (protocolFile != null) {
      protocolFile.close();
    }
  }
}

/**
 * State.
 *
 * @author Vedran
 * @created 19. Oktober 2001
 */
class State {


  private int id = 0;
  private String arg = null;


  /**
   * Construct object of class State.
   *
   * @param id state id
   */
  public State(int id) {
    this.id = id;
  }


  /**
   * Construct object of class State.
   *
   * @param id state id
   * @param arg state argument
   */
  public State(int id, String arg) {
    this.id = id;
    this.arg = arg;
  }


  /**
   * Get state id.
   *
   * @return state id
   */
  public int getID() {
    return id;
  }


  /**
   * Get state argument.
   *
   * @return state argument
   */
  public String getArg() {
    return arg;
  }
}

/**
 * Stack for states.
 *
 * @author Vedran
 * @created 19. Oktober 2001
 */
class StateStack extends Stack {


  /**
   * Looks at the object at the top of this stack without removing it from the
   * stack.
   *
   * @return state
   */
  public State peekState() {
    return (State)super.peek();
  }


  /**
   * Looks at the object at the top of this stack without removing it from the
   * stack.
   *
   * @return state id
   */
  public int peekStateID() {
    return ((State)super.peek()).getID();
  }


  /**
   * Looks at the object at the top of this stack without removing it from the
   * stack.
   *
   * @return state argument
   */
  public String peekStateArg() {
    return ((State)super.peek()).getArg();
  }


  /**
   * Removes the object at the top of this stack and returns that object as the
   * value of this function.
   *
   * @return state
   */
  public State popState() {
    return (State)super.pop();
  }


  /**
   * Removes the object at the top of this stack and returns that object as the
   * value of this function.
   *
   * @return state ID
   */
  public int popStateID() {
    return ((State)super.pop()).getID();
  }


  /**
   * Removes the object at the top of this stack and returns that object as the
   * value of this function.
   *
   * @return state argument
   */
  public String popStateArg() {
    return ((State)super.pop()).getArg();
  }


  /**
   * Pushes an item onto the top of this stack.
   *
   * @param state state
   * @return state (the same)
   */
  public State pushState(State state) {
    return (State)super.push(state);
  }


  /**
   * Pushes an item onto the top of this stack.
   *
   * @param stateID state id
   * @return state (constructed)
   */
  public State pushState(int stateID) {
    return (State)super.push(new State(stateID));
  }


  /**
   * Pushes an item onto the top of this stack.
   *
   * @param stateID state id
   * @param stateArg state argument
   * @return state (constructed)
   */
  public State pushState(int stateID, String stateArg) {
    return (State)super.push(new State(stateID, stateArg));
  }
}
