/*
 * Copyright (c) 2003 by SAP AG. All Rights Reserved.
 *
 * SAP, mySAP, mySAP.com and other SAP products and
 * services mentioned herein as well as their respective
 * logos are trademarks or registered trademarks of
 * SAP AG in Germany and in several other countries all
 * over the world. MarketSet and Enterprise Buyer are
 * jointly owned trademarks of SAP AG and Commerce One.
 * All other product and service names mentioned are
 * trademarks of their respective companies.
 *
 * @version $Id: Name.java,v 1.2 2004/02/24 10:28:30 sync Exp $
 */

package com.sapportals.wcm.util.name;

import java.io.Serializable;

/**
 * Generic class for the representation of namespace-qualified names. <p>
 *
 * Copyright (c) SAP AG 2001-2003
 *
 * @author julian.reschke@greenbytes.de
 * @version $Id: //javabas/com.sapportals.wcm/dev/src/java/util/api/com/sapportals/wcm/util/name/Name.java#9
 *      $
 */

public class Name implements IName, Serializable {

  private final String namespace;
  private final String name;
  private final String string;
  private final int hash;

  /**
   * @param namespace namespaceUri. Note that both "" (the empty string) and
   *      <code>null</code> are internally treated identically (as a
   *      non-namespace-qualified name).
   * @param name <a href="http://www.w3.org/TR/1999/REC-xml-names-19990114/#NT-NCName">
   *      NCName</a>
   */
  public Name(String namespace, String name) {
    // expand null into empty string for consistency
    this.namespace = (namespace == null) ? "" : namespace;

    this.name = name;
    this.string = "{" + this.namespace + "}" + this.name;
    this.hash = this.string.hashCode();
  }

  /**
   * @param prefixedName namespace & name in {namespace}name format
   * @exception Exception Exception raised in failure situation
   */
  public Name(String prefixedName)
    throws Exception {

    if (prefixedName.charAt(0) != '{') {
      throw new Exception("illegal format: " + prefixedName);
    }

    int rightBracket = prefixedName.lastIndexOf('}');
    if (rightBracket < 0) {
      throw new Exception("illegal format: " + prefixedName);
    }

    this.namespace = prefixedName.substring(1, rightBracket);
    this.name = prefixedName.substring(rightBracket + 1);
    this.string = prefixedName;
    this.hash = this.string.hashCode();
  }


  /**
   * @return the namespace URI (or the empty string when no namespace was given)
   */

  public final String getNamespace() {
    return this.namespace;
  }

  /**
   * @return the namespace URI (or <code>null</code> when no namespace was
   *      given)
   */
  public final String getNamespaceOrNull() {
    return "".equals(this.namespace) ? null : this.namespace;
  }

  /**
   * return get the local name
   *
   * @return name
   */

  public final String getName() {
    return this.name;
  }


  public final boolean equals(Object other) {

    if (other == null) {
      return false;
    }

    if (!(other instanceof Name)) {
      return false;
    }

    if (this.hash != ((Name)other).hash) {
      return false;
    }

    return this.string.equals(((Name)other).string);
  }

  public final int hashCode() {
    return this.hash;
  }


  public final String toString() {
    return this.string;
  }
}
