/*
 * Copyright (c) 2003 by SAP AG. All Rights Reserved.
 *
 * SAP, mySAP, mySAP.com and other SAP products and
 * services mentioned herein as well as their respective
 * logos are trademarks or registered trademarks of
 * SAP AG in Germany and in several other countries all
 * over the world. MarketSet and Enterprise Buyer are
 * jointly owned trademarks of SAP AG and Commerce One.
 * All other product and service names mentioned are
 * trademarks of their respective companies.
 *
 * @version $Id: SystemConfig.java,v 1.3 2004/06/07 07:36:44 mbo Exp $
 */

package com.sapportals.wcm.util.systemconfig;
import com.sap.tc.logging.Location;

import com.sapportals.config.fwk.*;

import com.sapportals.portal.security.usermanagement.*;

import com.sapportals.wcm.util.config.ConfigurationProxy;
import com.sapportals.wcm.util.logging.LoggingFormatter;
import com.sapportals.wcm.util.string.StrUtil;
//for service user security
import java.security.AccessController;
import java.security.PrivilegedActionException;
import java.security.PrivilegedExceptionAction;

import java.util.*;

/**
 * Main entry point for retrieving the <code>ISystemPrincipal</code> s. <br>
 *
 */
public class SystemConfig {

  private final static String ACL_PERMISSION_READ = "read";
  private final static String ACL_PERMISSION_WRITE = "write";
  private final static String ACL_PERMISSION_READWRITE = "readwrite";
  private final static String ACL_PERMISSION_DELETE = "delete";
  private final static String ACL_PERMISSION_FULLCONTROL = "fullcontrol";

  private final static String PERMISSION_READ = "read";
  private final static String PERMISSION_WRITE = "write";
  private final static String PERMISSION_DELETE = "delete";

  private final static String CONFIG_UPDATED = "updated_to_fwk_permissions";
  private final static String CONFIG_ACL_PERMISSIONS = "acl_permissions";
  private final static String CONFIG_ACL_OWNER = "acl_owner";
  private final static String CONFIG_PERMISSIONS = "permissions";
  private final static String CONFIG_CHANGE_PERMISSIONS = "change_permissions";
  private final static String CONFIG_LOCK_BREAKER = "lock_breaker";
  private final static String CONFIG_SERVICEACLS = "serviceacls";
  private final static String CONFIG_PRINCIPAL_NAME = "principal_name";

  private final static String SYSTEM_USERS_CONFIG_CLASS = "SystemUser";
  private final static String SYSTEM_GROUPS_CONFIG_CLASS = "SystemGroup";
  private final static String SYSTEM_ROLES_CONFIG_CLASS = "SystemRole";

  private final static char CFGFWK_ENUM_LIST_SEPARATOR = ',';
  private final static char CFGFWK_STRING_LIST_SEPARATOR = ',';

  private final static String PERM_CACHE_IN_USER = "--km-permissions--";
  private final static String CACL_CACHE_IN_USER = "--km-change-permission--";
  private final static String HASSACL_CACHE_IN_USER = "--km-has-service-acls--";

  private final static String UNDEFINED_PRINCIPAL_NAME = "undefined";

  private final static com.sap.tc.logging.Location s_log = com.sap.tc.logging.Location.getLocation(com.sapportals.wcm.util.systemconfig.SystemConfig.class);

  private static SystemConfig s_systemConfig = null;

  private final SystemPrincipalList m_systemUsers;
  private final SystemPrincipalList m_systemGroups;
  private final SystemPrincipalList m_systemRoles;
  private final SystemPrincipalList m_systemPrincipals;

  private final HashMap m_permissions = new HashMap();

  /**
   * Get an instance of the class (singleton)
   *
   * @return instance
   * @exception Exception Exception raised in failure situation
   */
  public static synchronized SystemConfig getInstance()
    throws Exception {
    if (s_systemConfig == null) {
      s_systemConfig = new SystemConfig();
    }
    return s_systemConfig;
  }

  /**
   * Construct system config
   *
   * @exception Exception Exception raised in failure situation
   */
  private SystemConfig()
    throws Exception {

    // migrate system configuration in configuration framework (map acl permissions to system config permissions)
    migrateSystemConfiguration();

    // create lists of system principals
    List users = getSystemPrincipalsFromConfig(SYSTEM_USERS_CONFIG_CLASS, IUMPrincipal.IUSER);
    List groups = getSystemPrincipalsFromConfig(SYSTEM_GROUPS_CONFIG_CLASS, IUMPrincipal.IGROUP);
    List roles = getSystemPrincipalsFromConfig(SYSTEM_ROLES_CONFIG_CLASS, IUMPrincipal.IROLE);

    List principals = new ArrayList(users.size() + groups.size() + roles.size());
    principals.addAll(users);
    principals.addAll(groups);
    principals.addAll(roles);

    m_systemUsers = new SystemPrincipalList(users);
    m_systemGroups = new SystemPrincipalList(groups);
    m_systemRoles = new SystemPrincipalList(roles);
    m_systemPrincipals = new SystemPrincipalList(principals);

    // create inverse cache for permissions
    ISystemPrincipalListIterator iter = m_systemPrincipals.iterator();
    while (iter.hasNext()) {
      ISystemPrincipal principal = (ISystemPrincipal)iter.next();
      Set permissions = principal.getMappedPermissions();
      Iterator permiter = permissions.iterator();

      while (permiter.hasNext()) {
        String permission = (String)permiter.next();
        ArrayList pl;

        if (!m_permissions.containsKey(permission)) {
          pl = new ArrayList();
          m_permissions.put(permission, pl);
        }
        else {
          pl = (ArrayList)m_permissions.get(permission);
        }

        pl.add(principal);
      }
    }
  }

  /**
   * @return all system users
   */
  public ISystemPrincipalList getSystemUsers() {
    return m_systemUsers;
  }

  /**
   * @return all system groups
   */
  public ISystemPrincipalList getSystemGroups() {
    return m_systemGroups;
  }

  /**
   * @return all system roles
   */
  public ISystemPrincipalList getSystemRoles() {
    return m_systemRoles;
  }

  /**
   * @return all system principals
   */
  public ISystemPrincipalList getSystemPrincipals() {
    return m_systemPrincipals;
  }

  /**
   * @param user TBD: Description of the incoming method parameter
   * @param permission TBD: Description of the incoming method parameter
   * @return true iff the user is a system principal which has the specified
   *      permission
   */
  public boolean hasPermission(IUser user, String permission) {

    HashSet permissions = (HashSet)user.getTransientAttribute(PERM_CACHE_IN_USER);

    // system permissions not set on user - fill permission set as HashSet
    // and store it as transient user property
    if (null == permissions) {
      permissions = new HashSet();
      user.setTransientAttribute(PERM_CACHE_IN_USER, permissions);

      ISystemPrincipalListIterator iter = m_systemPrincipals.iterator();
      while (iter.hasNext()) {
        ISystemPrincipal systemPrincipal = (ISystemPrincipal)iter.next();

        if (containsPrincipal(systemPrincipal.getPrincipal(), user)) {
          permissions.addAll(systemPrincipal.getMappedPermissions());
        }
      }
      user.setTransientAttribute(PERM_CACHE_IN_USER, permissions);
    }

    return permissions.contains(permission);
  }

  /**
   * @param user TBD: Description of the incoming method parameter
   * @param permissions TBD: Description of the incoming method parameter
   * @return true iff the user is a system principal which has the specified
   *      permissions
   */
  public boolean hasPermissions(IUser user, String[] permissions) {
    for (int i = 0, n = permissions.length; i < n; ++i) {
      if (!hasPermission(user, permissions[i])) {
        return false;
      }
    }
    return true;
  }

  /**
   * @param principal TBD: Description of the incoming method parameter
   * @return true iff the user is a system principal which is authorized to
   *      change permissions
   */
  public boolean canChangePermissions(IUMPrincipal principal) {
    IUser user = (IUser) principal;
    Boolean canChangePermmision = (Boolean)user.getTransientAttribute(CACL_CACHE_IN_USER);
    if (null != canChangePermmision) {
      return canChangePermmision.booleanValue();
    }
    try {
      for (int i = 0, n = m_systemPrincipals.principals.length; i < n; ++i) {
        ISystemPrincipal systemPrincipal = m_systemPrincipals.principals[i];
        if (containsPrincipal(systemPrincipal.getPrincipal(), principal)) {
          if (systemPrincipal.canChangePermissions()) {
            user.setTransientAttribute(CACL_CACHE_IN_USER, Boolean.TRUE);
            return true;
          }
        }
      }
    }
    catch (Exception e) {
      s_log.warningT("canChangePermissions(233)", "failed to check system principals" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(e));
    }

    user.setTransientAttribute(CACL_CACHE_IN_USER, Boolean.FALSE);
    return false;
  }

  /**
   * Check, if a given principal has all service acls.
   
   * @param principal the <code>IUMPrincipal</code> to check
   * @return <code>true</code> if the user is a system principal which is
   *        authorized to all service acls.
   */
  public boolean hasServiceAcls(IUMPrincipal principal) {

    IUser user = (IUser)principal;
    Boolean hasServiceAcls = (Boolean)user.getTransientAttribute(HASSACL_CACHE_IN_USER);

    if (hasServiceAcls != null) {
      return hasServiceAcls.booleanValue();
    }
    
    // change permission not set on principal: fill flag as Boolean
    // and store it as transient user property
    
    try {
      for (int i = 0, n = m_systemPrincipals.principals.length; i < n; ++i) {
        ISystemPrincipal systemPrincipal = m_systemPrincipals.principals[i];
        if (containsPrincipal(systemPrincipal.getPrincipal(), principal)) {
          if (systemPrincipal.hasServiceAcls()) {
            user.setTransientAttribute(HASSACL_CACHE_IN_USER, Boolean.TRUE);
            return true;
          }
        }
      }
    }
    catch (Exception e) {
      s_log.warningT("hasServiceAcls(797)", "failed to check system principals - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(e));
    }

    user.setTransientAttribute(HASSACL_CACHE_IN_USER, Boolean.FALSE);
    return false;
    
  }

  /**
   * @param configClass TBD: Description of the incoming method parameter
   * @param type TBD: Description of the incoming method parameter
   * @return a list of system principals of a specific type (get from the
   *      configuration framework)
   * @exception Exception Exception raised in failure situation
   */
  private List getSystemPrincipalsFromConfig(String configClass, int type)
    throws Exception {

    // retrieve system principals configurables from configuration framework
    IConfigurable[] configurables = ConfigurationProxy.getInstance().getPluginConfigurables(ConfigurationProxy.CFG_PLUGIN_CM_UTILITIES_SYSTEM_CONFIG);

    // generate system principal list from the configurables
    return getSystemPrincipals(configurables, configClass, type);
  }

  /**
   * @param configurables TBD: Description of the incoming method parameter
   * @param configClass TBD: Description of the incoming method parameter
   * @param type TBD: Description of the incoming method parameter
   * @return a list of system principals of a specific type (generate using a
   *      list of system principal configurables)
   * @exception Exception Exception raised in failure situation
   */
  private List getSystemPrincipals(IConfigurable[] configurables, String configClass, int type)
    throws Exception {

    // create result list
    LinkedList result = new LinkedList();

    // loop over the configurables
    for (int i = 0; i < configurables.length; i++) {

      // check if config class matches the requested type
      if (configClass.equals(configurables[i].getConfigClass().getName())) {

        // get system principal ID
        String id = configurables[i].getPropertyValue(CONFIG_PRINCIPAL_NAME);
        try {
          // create system principal object
          result.add(new SystemPrincipal(id, type, getLockBreakerFlag(configurables[i]), getServiceAclsFlag(configurables[i]), getMappedPermissions(configurables[i]), getChangePermissionsFlag(configurables[i])));
        }
        catch (Exception e) {
          s_log.errorT("getSystemPrincipals(283)", "Failed to create system principal" + " - " + com.sapportals.wcm.util.logging.LoggingFormatter.extractCallstack(e));
        }
      }
    }

    return result;
  }

  /**
   * @param configurable TBD: Description of the incoming method parameter
   * @return the CM framework permissions of a system principal as string array
   *      (generate using a configurable)
   * @exception Exception Exception raised in failure situation
   */
  private HashSet getMappedPermissions(IConfigurable configurable)
    throws Exception {

    // get the system config permissions in a string
    String permissionsString = configurable.getPropertyValue(CONFIG_PERMISSIONS);

    // if no permissions are assigned to the system principal, return an empty array (no mapped CM framework permissions)
    if (permissionsString == null) {
      return new HashSet();
    }

    // make array from string components (system config permissions)
    String[] permissionsArray = StrUtil.extractStringComponentsToArray(permissionsString, CFGFWK_ENUM_LIST_SEPARATOR, true, '\\');

    // get the permission mappings (a system config permissions can be mapped to multiple CM framework permissions)
    HashMap mappings = getPermissionMappings();

    // create a set to collect the mapped permissions
    HashSet result = new HashSet();

    // insert the mappings to the set (for all system config permissions)
    for (int i = 0; i < permissionsArray.length; i++) {

      // get the mappings for the current system config permissions
      String[] mappedPermissions = (String[])mappings.get(permissionsArray[i]);
      if (mappedPermissions != null) {

        // insert the mapped permissions (CM framework permissions) to the set with all mapped permissions
        for (int k = 0; k < mappedPermissions.length; k++) {
          result.add(mappedPermissions[k]);
        }
      }
    }

    return result;
  }

  /**
   * @return a hash map containing the mappings of system config permissions to
   *      CM framework permissions as they are configured in the configuration
   *      framework key: system config permission name (string) value: mapped CM
   *      framework permissions (string array)
   * @exception Exception Exception raised in failure situation
   */
  private HashMap getPermissionMappings()
    throws Exception {

    final String CONFIG_FROM = "from";
    final String CONFIG_TO = "to";

    // create result
    HashMap result = new HashMap();

    // get all configurables (configured permission mappings)
    IConfigurable[] configurables = ConfigurationProxy.getInstance().getPluginConfigurables(ConfigurationProxy.CFG_PLUGIN_CM_UTILITIES_SYSTEM_CONFIG_MAPPINGS);

    // loop over the mappings
    for (int i = 0; i < configurables.length; i++) {

      // get the mapping values
      String from = configurables[i].getPropertyValue(CONFIG_FROM);
      String to = configurables[i].getPropertyValue(CONFIG_TO);

      // put the mapping to the result
      if (from != null && to != null) {
        String[] toArray = StrUtil.extractStringComponentsToArray(to, CFGFWK_STRING_LIST_SEPARATOR, true);
        result.put(from, toArray);
      }
    }

    return result;
  }

  /**
   * @param configurable TBD: Description of the incoming method parameter
   * @return the value of the change_permissions attribute of the configurable
   * @exception Exception Exception raised in failure situation
   */
  private boolean getChangePermissionsFlag(IConfigurable configurable)
    throws Exception {
    String appLogViewerFlag = configurable.getPropertyValue(CONFIG_CHANGE_PERMISSIONS);
    if (appLogViewerFlag == null) {
      return false;
    }

    return new Boolean(appLogViewerFlag).booleanValue();
  }

  /**
   * @param configurable TBD: Description of the incoming method parameter
   * @return the value of the lock_breaker attribute of the configurable
   * @exception Exception Exception raised in failure situation
   */
  private boolean getLockBreakerFlag(IConfigurable configurable)
    throws Exception {
    String lockBreakerFlag = configurable.getPropertyValue(CONFIG_LOCK_BREAKER);
    if (lockBreakerFlag == null) {
      return false;
    }

    return lockBreakerFlag.equals(Boolean.TRUE.toString());    
  }

  /**
   * get the 'lock_breaker' attribute value from a given configurable.
   * @return the value of the lock_breaker attribute of the configurable
   * @exception Exception Exception raised in failure situation
   */
  private boolean getServiceAclsFlag(IConfigurable configurable)
                                     throws Exception {
    String serviceAclsFlag = configurable.getPropertyValue(CONFIG_SERVICEACLS);
    if (serviceAclsFlag == null) {
      return false;
    }
    return serviceAclsFlag.equals(Boolean.TRUE.toString());    
  }

  /**
   * Update old versions of the configuration 1. map old acl-permissions to new
   * permissions property (containing the permissions read, write, and delete)
   * 2. map old acl-owner flag to new change_permissions property 3. if the
   * principal_name property is "undefined" copy the configurable name to the
   * principal_name property
   *
   * @exception Exception Exception raised in failure situation
   */
  private void migrateSystemConfiguration()
    throws Exception {

    final long CONFIG_LOCK_TIMEOUT = 10L * 60L * 1000L;// 10 minutes
    final long CONFIG_WAIT_TIMEOUT = 60L * 1000L;// 1 minute

    // get a config manager
    IConfigClientContext context = IConfigClientContext.createContext(this.getConfigServiceUser());
    IConfigManager configManager = com.sapportals.config.fwk.Configuration.getInstance().getConfigManager(context);

    // get and lock the system config plugin
    IConfigPlugin plugin = configManager.getAndLockConfigPlugin(ConfigurationProxy.CFG_PLUGIN_CM_UTILITIES_SYSTEM_CONFIG, context, CONFIG_LOCK_TIMEOUT, CONFIG_WAIT_TIMEOUT);
    try {
      // get all configurables (configured system principals)
      IMutableConfigurable[] configurables = plugin.getConfigurables();

      // loop over the system principals
      for (int i = 0; i < configurables.length; i++) {

        // get the system principal property which tells us whether the principal already has been updated
        IConfigProperty updatedProperty = configurables[i].getProperty(CONFIG_UPDATED);

        // if the property is unset or false, update the configurable
        if (updatedProperty == null || !((Boolean)updatedProperty.getValueAsObject()).booleanValue()) {

          // get the old acl_permissions property
          IConfigProperty aclPermissionsPropery = configurables[i].getProperty(CONFIG_ACL_PERMISSIONS);

          // set the old acl_owner flag as change_permissions flag
          boolean isOwner = false;
          IConfigProperty aclOwnerPropery = configurables[i].getProperty(CONFIG_ACL_OWNER);
          if (aclOwnerPropery != null && ((Boolean)aclOwnerPropery.getValueAsObject()).booleanValue()) {
            configurables[i].setPropertyValue(CONFIG_CHANGE_PERMISSIONS, "true");
            isOwner = true;
          }

          // if the property exists map the acl_permissions to the new permissions which can only contain read, write, and delete
          if (aclPermissionsPropery != null) {
            String aclPermissionsString = configurables[i].getProperty(CONFIG_ACL_PERMISSIONS).getValueAsString();
            if (aclPermissionsString != null && aclPermissionsString.length() > 0) {
              configurables[i].setPropertyValue(CONFIG_PERMISSIONS, mapAclPermissionsToSystemConfigPermissions(aclPermissionsString, isOwner));
            }
          }

          // set the updated property to true
          configurables[i].setPropertyValue(CONFIG_UPDATED, "true");

          // commit the changes of the configurable
          plugin.replaceConfigurable(configurables[i]);
        }

        // if the principal_name property is "undefined" copy the configurable name to the principal_name property
        if (UNDEFINED_PRINCIPAL_NAME.equals(configurables[i].getPropertyValue(CONFIG_PRINCIPAL_NAME))) {

          // set the new value of the principal_name property to true
          configurables[i].setPropertyValue(CONFIG_PRINCIPAL_NAME, configurables[i].getId().getValueAsString());

          // commit the changes of the configurable
          plugin.replaceConfigurable(configurables[i]);
        }
      }
    }
    finally {
      plugin.unlock(context);
    }
  }

  /**
   * Map the old acl_permissions contained in aclPermissionsString (separated by
   * colons) to the new system principal permissions
   *
   * @param aclPermissionsString TBD: Description of the incoming method
   *      parameter
   * @param isOwner TBD: Description of the incoming method parameter
   * @return a string containing the system principal permissions separated by
   *      colons
   */
  private String mapAclPermissionsToSystemConfigPermissions(String aclPermissionsString, boolean isOwner) {

    // create a set to collect the mapped permissions
    HashSet permissions = new HashSet();

    if (isOwner) {
      // if the system principal is an owner, all permissions are granted
      permissions.add(PERMISSION_READ);
      permissions.add(PERMISSION_WRITE);
      permissions.add(PERMISSION_DELETE);
    }
    else {
      // get array with acl_permissions
      String[] aclPermissionsArray = StrUtil.extractStringComponentsToArray(aclPermissionsString, CFGFWK_ENUM_LIST_SEPARATOR, true);

      // map every single acl_permission to the corresponding framework permissions
      for (int i = 0; i < aclPermissionsArray.length; i++) {
        if (aclPermissionsArray[i].equals(ACL_PERMISSION_READ)) {
          permissions.add(PERMISSION_READ);
        }
        if (aclPermissionsArray[i].equals(ACL_PERMISSION_WRITE)) {
          permissions.add(PERMISSION_WRITE);
        }
        if (aclPermissionsArray[i].equals(ACL_PERMISSION_READWRITE)) {
          permissions.add(PERMISSION_READ);
          permissions.add(PERMISSION_WRITE);
        }
        if (aclPermissionsArray[i].equals(ACL_PERMISSION_FULLCONTROL)) {
          permissions.add(PERMISSION_READ);
          permissions.add(PERMISSION_WRITE);
          permissions.add(PERMISSION_DELETE);
        }
      }
    }

    // generate the result string
    StringBuffer result = new StringBuffer();
    Iterator it = permissions.iterator();
    while (it.hasNext()) {
      if (result.length() > 0) {
        result.append(CFGFWK_ENUM_LIST_SEPARATOR);
      }
      result.append((String)it.next());
    }

    return result.toString();
  }

  // HJR 8-11-02: cache results for a while...
  private final static long lifetime = 60000;// ms ~ 1 min
  static long lastRefresh = 0;
  static Hashtable containsPrincipalCache = new Hashtable();


  private String getPrincipalsHashKey(IUMPrincipal p1, IUMPrincipal p2) {

    char tp1 = '-';

    char tp2 = '-';

    // avoid unnessecary object generation, so convert int -> char
    switch (p1.getType()) {
      case IUMPrincipal.IROLE:
        tp1 = 'R';
      case IUMPrincipal.IGROUP:
        tp1 = 'G';
      case IUMPrincipal.IUSER:
        tp1 = 'U';
    }

    switch (p2.getType()) {
      case IUMPrincipal.IROLE:
        tp2 = 'R';
      case IUMPrincipal.IGROUP:
        tp2 = 'G';
      case IUMPrincipal.IUSER:
        tp2 = 'U';
    }

    StringBuffer sb = new StringBuffer(50);
    sb.append(tp1);
    sb.append(p1.getId());
    sb.append(tp2);
    sb.append(p2.getId());
    return sb.toString();
  }

  /**
   * @param p1 TBD: Description of the incoming method parameter
   * @param p2 TBD: Description of the incoming method parameter
   * @return true iff the principal p1 equals or contains the principal p2 (as
   *      group or role member, recursive check)
   */
  private boolean containsPrincipal(IUMPrincipal p1, IUMPrincipal p2) {
    boolean result = false;

    if (p1 == p2 || p1.equals(p2)) {
      // check the cheap case first
      result = true;
    }
    else {
      long now = System.currentTimeMillis();

      if (now - lastRefresh > lifetime) {
        containsPrincipalCache.clear();
        lastRefresh = now;
      }

      String hashkey = getPrincipalsHashKey(p1, p2);
      Boolean res = (Boolean)containsPrincipalCache.get(hashkey);

      if (res == null) {
        switch (p2.getType()) {
          case IUMPrincipal.IUSER:
            if ((p1.getType() == IUMPrincipal.IGROUP && ((IGroup)p1).containsUser(p2.getId()))
               || (p1.getType() == IUMPrincipal.IROLE && ((IRole)p1).containsUser(p2.getId()))
              ) {
              result = true;
            }
            break;
          case IUMPrincipal.IGROUP:
            if ((p1.getType() == IUMPrincipal.IGROUP && ((IGroup)p1).containsGroup(p2.getId()))
               || (p1.getType() == IUMPrincipal.IROLE && ((IRole)p1).containsGroup(p2.getId()))
              ) {
              result = true;
            }
            break;
        }
        containsPrincipalCache.put(hashkey, result ? Boolean.TRUE : Boolean.FALSE);
      }
      else {
        result = res.booleanValue();
      }
    }

    return result;
  }

  /**
   * AccessController.doPrivileged stops the security check on this level of the
   * call stack. This allows customer code to call this method. Otherwise the
   * execution of foreign code would be forbidden.
   *
   * @return configServiceUser
   * @exception Exception Exception raised in failure situation
   */
  private IUser getConfigServiceUser()
    throws Exception {
    Object serviceUser = null;
    serviceUser = AccessController.doPrivileged(
      new PrivilegedExceptionAction() {
        public Object run()
          throws Exception {
          //call of secure method
          ConfigurationProxy configProxy = ConfigurationProxy.getInstance();
          return configProxy.getConfigServiceUser();
        }
      }
      );
    return (IUser)serviceUser;
  }
}
