/*
 * Copyright (c) 2003 by SAP AG. All Rights Reserved.
 *
 * SAP, mySAP, mySAP.com and other SAP products and
 * services mentioned herein as well as their respective
 * logos are trademarks or registered trademarks of
 * SAP AG in Germany and in several other countries all
 * over the world. MarketSet and Enterprise Buyer are
 * jointly owned trademarks of SAP AG and Commerce One.
 * All other product and service names mentioned are
 * trademarks of their respective companies.
 *
 * @version $Id: IHierarchicalUri.java,v 1.1.1.1 2004/02/18 12:56:48 sync Exp $
 */

package com.sapportals.wcm.util.uri;

/**
 * A <a href="http://www.ietf.org/rfc/rfc2396.txt">RFC2396</a> URI interface.
 * <pre>
 *    remainder   ::= '//' authority abspath query
 *    abspath     ::= '/' pathsegment | empty
 *    pathsegment ::= name | name '/' pathsegment | empty
 *    query       ::= '?' uri-characters | empty
 * </pre> Copyright (c) SAP AG 2001-2002
 *
 * @author julian.reschke@greenbytes.de
 * @author stefan.eissing@greenbytes.de
 * @version $Id: IHierarchicalUri.java,v 1.1.1.1 2004/02/18 12:56:48 sync Exp $
 */

public interface IHierarchicalUri extends IUri {

  /**
   * Append the given path segment to the current path of this uri. Takes care
   * of leading, trailing slashes. Note that the path <b>must</b> be uri encoded
   * or the resulting uri is not valid.
   *
   * @param path TBD: Description of the incoming method parameter
   * @return new uri with concatenated path
   */
  public IHierarchicalUri appendPath(String path);

  /**
   * Return authority part of this hierarchical Uri.
   *
   * @return authority part of this hierarchical Uri
   */
  public String getAuthority();

  /**
   * Return path of this Uri or, if not there, the emtpy string
   *
   * @return path of this uri
   */
  public String getPath();

  /**
   * Return the segments of the uri path
   *
   * @return segments of uri path
   */
  public String[] getPathSegments();

  /**
   * Return query of this Uri in RFC 2396 encoded form or, if not there, null
   *
   * @return query of this uri in RFC 2396 encoded form or null
   */
  public String getQuery();

  /**
   * Returns the uri of the root collection (path = "/").
   *
   * @return uri of server root
   */
  public IHierarchicalUri getRoot();

  /**
   * Determine if this Uri is ancestor of other uri
   *
   * @param other TBD: Description of the incoming method parameter
   * @return if this Uri is ancestor of other uri
   */
  public boolean isAncestorOf(IHierarchicalUri other);

  /**
   * Convert the given uri to a relative uri reference with this uri as base. If
   * this uri is no ancestor of the other uri, an absolute uri refernce is
   * returned.
   *
   * @param other uri to unresolve
   * @param fragment to add to result uri reference
   * @return uri reference relative to this uri
   */
  public IUriReference unresolve(IHierarchicalUri other, String fragment);

  /**
   * Given this Uri as the base of the WCM repository, map the Uri Reference to
   * the corresponsing WCM Path in the framework. <br>
   * <pre>
   * Example: "http://host/docs" + "/docs/test%203", gives "/test 3"
   * Example: "http://host/docs" + "/xxx/test%203", gives null
   * </pre>
   *
   * @param ref the uri reference to map
   * @return the mapped wcm path or <code>null</code> if not possible
   */
  public URI mapToWcmPath(IUriReference ref);

  public RID mapToResourceID(IUriReference ref);

  /**
   * Given this Uri as the base of the WCM repository, map the Uri Reference to
   * the corresponsing WCM Path in the framework. <br>
   * <pre>
   * Example: "http://host/docs" + "/docs/test%203", gives "/test 3"
   * Example: "http://host/docs" + "/xxx/test%203", gives null
   * </pre>
   *
   * @param ref the uri reference to map
   * @param codec to use for uri decoding
   * @return the mapped wcm path or <code>null</code> if not possible
   */
  public URI mapToWcmPath(IUriReference ref, URICodec codec);

  public RID mapToResourceID(IUriReference ref, URICodec codec);

  /**
   * Given this Uri as the base of the WCM repository, map the WCM Path to a uri
   * reference with absolute path component. <br>
   * <pre>
   * Example: "http://host/docs" + "test 3", gives "/docs/test%203"
   * Example: "http://host/docs" + "/fs/test", gives "/docs/fs/test"
   * </pre>
   *
   * @param wcmPath to map
   * @return the mapped uri reference
   */
  public IUriReference mapToAbsolutePath(URI wcmPath);

  public IUriReference mapToAbsolutePath(RID wcmPath);

  /**
   * Given this Uri as the base of the WCM repository, map the WCM Path to a uri
   * reference with absolute path component. <br>
   * <pre>
   * Example: "http://host/docs" + "test 3", gives "/docs/test%203"
   * Example: "http://host/docs" + "/fs/test", gives "/docs/fs/test"
   * </pre>
   *
   * @param wcmPath to map
   * @param codec to use for uri encoding
   * @return the mapped uri reference
   */
  public IUriReference mapToAbsolutePath(URI wcmPath, URICodec codec);

  public IUriReference mapToAbsolutePath(RID wcmPath, URICodec codec);

  /**
   * Given this Uri as the base, map the uri reference to a uri reference with
   * absolute path component. <br>
   * <pre>
   * Example: "http://host/docs" + "test%203", gives "/docs/test%203"
   * Example: "http://host/docs" + "http://another/fs/test", gives "http://another/fs/test"
   * </pre>
   *
   * @param ref TBD: Description of the incoming method parameter
   * @return the mapped uri reference
   */
  public IUriReference mapToAbsolutePath(IUriReference ref);

  /**
   * Given this Uri as the base of the WCM repository, map the WCM Path to a
   * absolute uri. <br>
   * <pre>
   * Example: "http://host/docs" + "test 3", gives "http://host/docs/test%203"
   * Example: "http://host/docs" + "/fs/test", gives "http://host/docs/fs/test"
   * </pre>
   *
   * @param wcmPath to map
   * @return the mapped uri
   */
  public IHierarchicalUri mapToAbsoluteUri(URI wcmPath);

  public IHierarchicalUri mapToAbsoluteUri(RID wcmPath);

  /**
   * Given this Uri as the base of the WCM repository, map the WCM Path to a
   * absolute uri. <br>
   * <pre>
   * Example: "http://host/docs" + "test 3", gives "http://host/docs/test%203"
   * Example: "http://host/docs" + "/fs/test", gives "http://host/docs/fs/test"
   * </pre>
   *
   * @param wcmPath to map
   * @param codec to use for uri encoding
   * @return the mapped uri
   */
  public IHierarchicalUri mapToAbsoluteUri(URI wcmPath, URICodec codec);

  public IHierarchicalUri mapToAbsoluteUri(RID wcmPath, URICodec codec);

  /**
   * Construct a new object of same type with the given query. If query is
   * <code>null</code> , the new uri will be without query. The query string
   * must be in RFC 2396 encoded form. See {@link UriQuery} for manipulation of
   * query strings.
   *
   * @param query encoded query string for returned object
   * @return hierarchical uri of same type as this
   */
  public IHierarchicalUri setQuery(String query);

  /**
   * Construct a new object of same type with the given path. If path is <code>
   * null</code> , the new uri will have the root path /. The path string must
   * be in RFC 2396 encoded form. See {@link URICodec}. The query of the
   * resulting uri will always be empty.
   *
   * @param path encoded path string for returned object
   * @return hierarchical uri of same type as this
   */
  public IHierarchicalUri setPath(String path);
}
