/*
 * Copyright (c) 2003 by SAP AG. All Rights Reserved.
 *
 * SAP, mySAP, mySAP.com and other SAP products and
 * services mentioned herein as well as their respective
 * logos are trademarks or registered trademarks of
 * SAP AG in Germany and in several other countries all
 * over the world. MarketSet and Enterprise Buyer are
 * jointly owned trademarks of SAP AG and Commerce One.
 * All other product and service names mentioned are
 * trademarks of their respective companies.
 *
 * @version $Id: DomBuilder.java,v 1.2 2004/06/28 18:06:05 jre Exp $
 */

package com.sapportals.wcm.util.xml;

import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.*;
import org.xml.sax.EntityResolver;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import com.sap.tc.logging.Location;
import com.sapportals.wcm.util.logging.LoggingFormatter;

/**
 * Convenience functions for building {@link Document}s from XML.
 * <p>
 * Copyright (c) SAP AG 2001-2004
 * @author julian.reschke@greenbytes.de
 * @version $Id: DomBuilder.java,v 1.2 2004/06/28 18:06:05 jre Exp $
 */

public class DomBuilder {

  private final static Location log = Location.getLocation(DomBuilder.class);

  final static DocumentBuilderFactory Factory = DocumentBuilderFactory.newInstance();

  final static EntityResolver noEntityResolver = new NoEntityResolver();
  final static ErrorHandler strictErrorHandler = new StrictErrorHandler();


  public static Document newDocument() {
    return getBuilder(true).newDocument();
  }

  public static Document parse(String input, boolean namespaceAware)
    throws SAXException, java.io.IOException {
    return parse(new InputSource(new StringReader(input)), namespaceAware);
  }

  public static Document parse(InputSource is, boolean namespaceAware)
    throws SAXException {
    try {
      DocumentBuilder builder = getBuilder(namespaceAware);
      return builder.parse(is);
    }
    catch (IOException ex) {
      log.debugT("parse(59)", "while parsing document" + " - " + LoggingFormatter.extractCallstack(ex));
      throw new SAXException(ex);
    }
  }

  /**
   * Parse a string into a {@link Document}.
   *
   * @param input the input stream
   * @return the Document
   */
  public static Document parse(String input)
    throws SAXException, java.io.IOException {
    return parse(new InputSource(new StringReader(input)), true);
  }


  /**
   * Parse an input stream into a {@link Document}.
   *
   * @param input the input stream
   * @param tolerateNSerrors when set to <code>true</code> , elements that
   *      aren't XML-NS-wellformed but which <i>are</i> XML-wellformed will be
   *      tolerated (and dropped) (<em>unsupported in releases later than NW04</em>)
   * @return the Document
   */
  public static Document parse(InputStream input, boolean tolerateNSerrors)
    throws SAXException, IOException {
    Document doc = parse(new InputSource(input), !tolerateNSerrors);

    if (tolerateNSerrors) {
      // jre: hack has been removed in releases later than NW04
      if (log.beInfo()) {
        log.infoT("parse(98)", SimpleSerializer.toString(doc, true, false));
      }
      return parse(SimpleSerializer.toString(doc, false, false));
    }

    // HACK
    try {
      doc.getDocumentElement().getNamespaceURI();
    }
    catch (NoSuchMethodError ex) {
      log.fatalT("parse(108)", "-------------------------------------------------------");
      log.fatalT("parse(109)", "Couldn't invoke: getDocumentElement().getNamespaceURI()");
      log.fatalT("parse(110)", "This is likely caused by outdated JAR files in the");
      log.fatalT("parse(111)", "CLASSPATH preceding the DOM-Level-2 that is needed.");
      log.fatalT("parse(112)", "Current DOM impl is: " + doc.getClass().getName());
      log.fatalT("parse(113)", "-------------------------------------------------------");
      throw ex;
    }

    return doc;
  }

  private static DocumentBuilder getBuilder(boolean namespaceAware) {

    try {
      DocumentBuilder builder = null;
      synchronized (Factory) {
        Factory.setNamespaceAware(namespaceAware);
        builder = Factory.newDocumentBuilder();
      }
      builder.setErrorHandler(strictErrorHandler);
      builder.setEntityResolver(noEntityResolver);
      return builder;
    }
    catch (ParserConfigurationException ex) {
      log.fatalT("getBuilder(133)", "creating DocumentBuilder" + " - " + LoggingFormatter.extractCallstack(ex));
      return null;
    }

  }
}
